#include "Engine/AbstractScene.hpp"

#include "Engine/Game.hpp"
#include "Engine/Input/Input.hpp"
#include "Engine/GUI/GUIWidget.hpp"
#include "Town/World.hpp"
#include "Utility/Assert.hpp"
#include "Utility/Vector.hpp"

#ifdef SL_DEBUG
	#include "Engine/GUI/FontManager.hpp"
	#include "Engine/Graphics/SpriteFont.hpp"
#endif // SL_DEBUG

namespace sl
{

AbstractScene::AbstractScene(Game *game, int viewWidth, int viewHeight, int depth)
	:viewWidth(viewWidth)
	,viewHeight(viewHeight)
	,input(game->getRawInput(), *this)
	,world(nullptr)
	,running(true)
	,visible(true)
	,game(game)
	,isInUpdate(false)
	,timestamp(0)
	,depth(depth)
	,playerController(input, game->getPlayerControls())
{
#ifdef SL_DEBUG
	debugRenderRect.setOutlineThickness(1);
	debugRenderCircle.setOutlineThickness(1);
#endif
}

AbstractScene::~AbstractScene()
{
	ASSERT(widgets.empty());
}

void AbstractScene::stop()
{
	running = false;
}

void AbstractScene::start()
{
	running = true;
}

void AbstractScene::update(Timestamp timestamp)
{
	if (running)
	{
		ASSERT(!isInUpdate);
		isInUpdate = true;
#ifdef SL_DEBUG
		// I don't know why I was doing this before, but there must have been a reason:
		//for (auto& rects : debugRects)
		//{
		//	rects.second.clear();
		//}
		//for (auto& lines : debugLines)
		//{
		//	lines.second.clear();
		//}
		debugRects.clear();
		debugLines.clear();
		debugStrings.clear();
		debugCircles.clear();
		// remove all cached SpriteText instances whose lifetime has expired (<= 0)
		for (auto it = cachedSpriteTexts.begin(); it != cachedSpriteTexts.end(); /*incr in body*/)
		{
			if (--it->second.second <= 0)
			{
				it = cachedSpriteTexts.erase(it);
			}
			else
			{
				++it;
			}
		}
#endif
		onUpdate(timestamp);
		this->timestamp = timestamp;

		isInUpdate = false;
		for (gui::GUIWidget *widget : toRegister)
		{
			registerGUIWidget(widget);
		}
		toRegister.clear();
	}
}

void AbstractScene::hide()
{
	visible = false;
}

void AbstractScene::show()
{
	visible = true;
}

void AbstractScene::draw(RenderTarget& target)
{
	if (visible)
	{
		onDraw(target);
#ifdef SL_DEBUG
		const Settings& settings = game->getSettings();
		for (auto it = debugRects.begin(); it != debugRects.end(); ++it)
		{
			if (settings.get(it->first))
			{
				for (int i = 0; i < it->second.size(); ++i)
				//for (const auto& rectColCol : it->second)
				{
					const auto& rectColCol = it->second[i];
					const Rect<int>& rect = std::get<0>(rectColCol);
					debugRenderRect.setPosition(rect.left, rect.top);
					debugRenderRect.setSize(sf::Vector2f(rect.width, rect.height));
					debugRenderRect.setOutlineColor(std::get<1>(rectColCol));
					debugRenderRect.setFillColor(std::get<2>(rectColCol));
					target.getSFMLTarget().draw(debugRenderRect);
				}
				it->second.clear();
			}
		}

		for (auto& pair : debugLines)
		{
			if (settings.get(pair.first))
			{
				for (const sf::VertexArray& line : pair.second)
				{
					target.getSFMLTarget().draw(line);
				}
				pair.second.clear();
			}
		}
		for (auto& pair : debugStrings)
		{
			if (settings.get(pair.first))
			{
				for (const auto& strCol : pair.second)
				{
					// only use this as a fall-back (right now it's not used)
					//static SpriteText dText("default", Rect<int>(500, 100), 8, "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
					
					auto it = cachedSpriteTexts.find(strCol.str);
					const int framesToKeep = 120;
					if (it == cachedSpriteTexts.end())
					{
						
						auto sf = unique<SpriteText>("default", Rect<int>(500, 100), 8, 1, strCol.str);
						it = cachedSpriteTexts.insert(std::make_pair(strCol.str, std::make_pair(std::move(sf), framesToKeep))).first;
					}
					it->second.second = framesToKeep;
					it->second.first->update(strCol.pos.x, strCol.pos.y);
					it->second.first->setFontColor(strCol.col);
					it->second.first->draw(target.getSFMLTarget());
					//cachedSpriteTexts.clear(); // TODO: remove?!?!
				}
				pair.second.clear();
			}
		}

		for (auto it = debugCircles.begin(); it != debugCircles.end(); ++it)
		{
			if (settings.get(it->first))
			{
				for (int i = 0; i < it->second.size(); ++i)
				{
					const auto& posRadColCol = it->second[i];
					const Vector2f& middle = std::get<0>(posRadColCol);
					const float radius = std::get<1>(posRadColCol);
					debugRenderCircle.setPosition(middle.x, middle.y);
					debugRenderCircle.setOrigin(radius, radius);
					debugRenderCircle.setRadius(radius);
					debugRenderCircle.setOutlineColor(std::get<2>(posRadColCol));
					debugRenderCircle.setFillColor(std::get<3>(posRadColCol));
					target.getSFMLTarget().draw(debugRenderCircle);
				}
				it->second.clear();
			}
		}
#endif
	}
	else
	{
#ifdef SL_DEBUG
		debugRects.clear();
		debugLines.clear();
		debugStrings.clear();
		debugCircles.clear();
#endif // SL_DEBUG
	}
}

void AbstractScene::drawGUI(RenderTarget& target)
{
	if (visible)
	{
		for (gui::GUIWidget *widget : widgets)
		{
			widget->draw(target.getSFMLTarget());
		}
		onDrawGUI(target);
	}
}

void AbstractScene::registerGlobalEvent(std::unique_ptr<const Event> event)
{
	ASSERT(event->getCategory() != Event::Category::NonBroadcasted);
	game->registerEvent(std::move(event));
}

void AbstractScene::registerLocalEvent(std::unique_ptr<const Event> event)
{
	ASSERT(event->getCategory() != Event::Category::NonBroadcasted);
	eventQueue.addEvent(std::move(event));
}

void AbstractScene::handleEvents(const EventQueue& events)
{
	onHandleEvents(events);
	onHandleEvents(eventQueue);
	const Event *ev = nullptr;
	for (int i = 0; i < Event::Category::EventTypeNum; ++i)
	{
		{
			EventIterator eventIt = events.iterator(static_cast<Event::Category>(i));
			while (eventIt.poll(ev))
			{
				handleEvent(*ev);
			}
		}
		{
			EventIterator eventIt = eventQueue.iterator(static_cast<Event::Category>(i));
			while (eventIt.poll(ev))
			{
				handleEvent(*ev);
			}
		}
	}
	eventQueue.reset();
}

void AbstractScene::registerGUIWidget(gui::GUIWidget *widget)
{
	if (isInUpdate)
	{
		toRegister.push_back(widget);
	}
	else
	{
		for (gui::GUIWidget *w : widgets)
		{
			ASSERT(w != widget);
		}
		widgets.push_back(widget);
		widget->setScene(this);
	}
}

void AbstractScene::unregisterGUIWidget(gui::GUIWidget *widget)
{
	for (gui::GUIWidget*& w : widgets)
	{
		if (w == widget)
		{
			widget->setScene(nullptr);
			w = widgets.back();
			widgets.pop_back();
			return;
		}
	}
	for (gui::GUIWidget*& w : toRegister)
	{
		if (w == widget)
		{
			widget->setScene(nullptr);
			w = toRegister.back();
			toRegister.pop_back();
			return;
		}
	}
	ASSERT(false);
}

const sf::Vector2f AbstractScene::getMousePos() const
{
	return game->getRawInput().mousePosF() + this->getView() - sf::Vector2f(viewWidth / 2, viewHeight / 2);
}

void AbstractScene::setView(int x, int y)
{
	ERROR("Implement this properly you fuck");
}

Rect<int> AbstractScene::getViewRect() const
{
	const sf::Vector2f pos(this->getView());
	return Rect<int>(pos.x - viewWidth / 2, pos.y - viewHeight / 2, viewWidth, viewHeight);
}

int AbstractScene::getViewWidth() const
{
	return viewWidth;
}

int AbstractScene::getViewHeight() const
{
	return viewHeight;
}

Game * AbstractScene::getGame()
{
	return game;
}

const Game * AbstractScene::getGame() const
{
	return game;
}

World * AbstractScene::getWorld()
{
	return world;
}

const World * AbstractScene::getWorld() const
{
	return world;
}

const InputHandle& AbstractScene::getInputHandle() const
{
	return input;
}

const Controller& AbstractScene::getPlayerController() const
{
	return playerController;
}

void AbstractScene::debugDrawRect(const std::string& category, const Rect<int>& rect, sf::Color outline, sf::Color fill)
{
#ifdef SL_DEBUG
	debugRects[category].push_back(std::make_tuple(rect, outline, fill));
#endif
}

void AbstractScene::debugDrawLine(const std::string& category, const Vector2f& p1, const Vector2f& p2, sf::Color color)
{
#ifdef SL_DEBUG
	sf::VertexArray line(sf::PrimitiveType::Lines, 2);
	line[0].color = line[1].color = color;
	line[0].position = sf::Vector2f(p1.x, p1.y);
	line[1].position = sf::Vector2f(p2.x, p2.y);
	debugLines[category].push_back(std::move(line));
#endif
}

void AbstractScene::debugDrawText(const std::string& category, const Vector2f& pos, const std::string& text, const sf::Color& colour)
{
#ifdef SL_DEBUG
	DebugText debugText = {text, colour, pos};
	debugStrings[category].push_back(std::move(debugText));
#endif // SL_DEBUG
}

void AbstractScene::debugDrawCircle(const std::string& category, const Vector2f& middle, float radius, sf::Color outline, sf::Color fill)
{
#ifdef SL_DEBUG
	debugCircles[category].push_back(std::make_tuple(middle, radius, outline, fill));
#endif // SL_DEBUG
}

bool AbstractScene::isActive() const
{
	Timestamp diff = game->getTimestamp() - this->getTimestamp();
	return diff == 0 || diff == 1;
}

AbstractScene::Timestamp AbstractScene::getTimestamp() const
{
	return timestamp;
}

void AbstractScene::destroy()
{
	stop();
	hide();
	game->removeScene(this);
}

int AbstractScene::getDepth() const
{
	return depth;
}

/*		protected		*/

void AbstractScene::onDrawGUI(RenderTarget& target)
{
	//	empty by default, if you want unique behaviour
	//	then override it in a base class or something
}

void AbstractScene::handleEvent(const Event& event)
{
	// empty on purpose
}

void AbstractScene::onHandleEvents(const EventQueue& events)
{
	// empty on purpose
}

void AbstractScene::onWorldAdd()
{
	// empty on purpose
}

void AbstractScene::serializeAbstractScene(Serialize& out) const
{
	out.startObject("AbstractScene");
	// view ignored
	// input ignored
	SAVEPTR(world);
	SAVEVAR(b, running);
	SAVEVAR(b, visible);
	// game ignored
	// events ignored (TODO: is this correct?)
	// ignore all widget stuff
	ASSERT(!isInUpdate);
	// ignore debug draw stuff
	// ignore timestamp?
	SAVEVAR(i, depth);
	// ignore playerController
	out.endObject("AbstractScene");
}

void AbstractScene::deserializeAbstractScene(Deserialize& in)
{
	in.startObject("AbstractScene");
	// view ignored
	// input ignored
	LOADPTR(world);
	LOADVAR(b, running);
	LOADVAR(b, visible);
	// game ignored
	// events ignored (TODO: is this correct?)
	// ignore all widget stuff
	ASSERT(!isInUpdate);
	// ignore debug draw stuff
	// ignore timestamp?
	LOADVAR(i, depth);
	// ignore playerController
	in.endObject("AbstractScene");
}

}// End of sl namespace
