/**
 * @section DESCRIPTION
 * This is the abstract base class for all Scenes(states, levels) in the game.
 * Use this for non-physical scenes like menus, pause screens, but use Scene for levels like a town, dungeon, whatever.
 */
#ifndef SL_ABSTRACTSCENE_HPP
#define SL_ABSTRACTSCENE_HPP

#include <cstdint>
#include <memory>
#include <vector>

#include "Engine/Events/EventQueue.hpp"
#include "Engine/Input/Controller.hpp"
#include "Engine/Input/InputHandle.hpp"
#include "Engine/Serialization/Serializable.hpp"
#include "Engine/Graphics/RenderTarget.hpp"
#include "Utility/Rect.hpp"
#include "Utility/Vector.hpp"

#ifdef SL_DEBUG
	#include <SFML/Graphics/Color.hpp>
	#include <SFML/Graphics/RectangleShape.hpp>
	#include <SFML/Graphics/VertexArray.hpp>
	#include <tuple>
	#include "Engine/Graphics/SpriteText.hpp"
#endif

namespace sl
{

class Event;
class Game;
class World;

namespace gui
{
class GUIWidget;
} // gui

class AbstractScene : public Serializable
{
public:
	typedef int64_t Timestamp;

	virtual ~AbstractScene();
	/**
	 * Sets the AbstractScene to stop updating
	 */
	void stop();
	/**
	 * Sets the AbstractScene to start updating
	 */
	void start();
	/**
	 * Updates the AbstractScene
	 * @param timestamp The current time (in seconds since game start)
	 */
	void update(int64_t timestamp);
	/**
	 * Marks the AbstractScene as invisible
	 */
	void hide();
	/**
	 * Marks the AbstractScene as visible
	 */
	void show();
	/**
	 * Renders the AbstractScene. Calls onDraw()
	 * @param target The RenderTarget that the AbstractScene will be rendered to
	 */
	void draw(RenderTarget& target);
	/**
	 * Renders all GUI-related things over top of the AbstractScene. Calls onDrawGUI()
	 * @param target The RenderTarget that the AbstractScene will be rendered to
	 */
	void drawGUI(RenderTarget& target);
	/**
	 * Passes an event up to the Game for it to register
	 * @param event The event to be registered
	 */
	void registerGlobalEvent(std::unique_ptr<const Event> event);
	/**
	 * Registers an event in the local EventQueue
	 * @param event The event to be registered
	 */
	void registerLocalEvent(std::unique_ptr<const Event> event);

	void handleEvents(const EventQueue& events);

	void registerGUIWidget(gui::GUIWidget *widget);

	void unregisterGUIWidget(gui::GUIWidget *widget);

	const sf::Vector2f getMousePos() const;

	/**
	 * Returns where the camera should look at the AbstractScene Abstract method, must be defined in subclasses.
	 * @return The vector representing where the camera should be rendered.
	 */
	virtual const sf::Vector2f getView() const = 0;
	/**
	 * Sets where teh camera should look at the AbstractScene. Unless overridden, this method does nothing.
	 * @param x The x position to change the view to
	 * @param y The y position to change the view to
	 */
	virtual void setView(int x, int y);

	Rect<int> getViewRect() const;

	int getViewWidth() const;

	int getViewHeight() const;

	Game * getGame();

	const Game * getGame() const;

	World * getWorld();

	const World * getWorld() const;

	const InputHandle& getInputHandle() const;

	const Controller& getPlayerController() const;

	void debugDrawRect(const std::string& category, const Rect<int>& rect, sf::Color outline, sf::Color fill = sf::Color::Transparent);

	void debugDrawLine(const std::string& category, const Vector2f& p1, const Vector2f& p2, sf::Color color);

	template <typename T>
	void debugDrawLines(const std::string& category, const T& points, sf::Color color);

	void debugDrawText(const std::string& category, const Vector2f& pos, const std::string& text, const sf::Color& colour = sf::Color::White);

	void debugDrawCircle(const std::string& category, const Vector2f& middle, float radius, sf::Color outline, sf::Color fill = sf::Color::Transparent);

	bool isActive() const;

	Timestamp getTimestamp() const;

	void destroy();

	int getDepth() const;

protected:
	/**
	 * Constructor to create a AbstractScene. Can only be accessed by subclasses since this is an abstract class.
	 * @param game The game this AbstractScene belongs to
	 */
	AbstractScene(Game *game, int viewWidth = 0, int viewHeight = 0, int depth = 0);

	/**
	 * This method gets called on every update() and defines any unique behaviour to be handled in subclasses.
	 * Abstract method, must be defined in subclasses.
	 * @param timestamp The current time (in seconds since game start)
	 */
	virtual void onUpdate(Timestamp timestamp) = 0;
	/**
	 * This method gets called on every draw() and defines any unique behaviour to be handled in subclasses.
	 * Abstract method, must be defined in subclasses.
	 * @param target The RenderTarget to draw to
	 */
	virtual void onDraw(RenderTarget& target) = 0;
	/**
	 * Draws any screen overlay in the AbstractScene. Called every drawGUI()
	 * @param target The RenderTarget that the overlay will be rendered to
	 */
	virtual void onDrawGUI(RenderTarget& target);

	virtual void handleEvent(const Event& event);

	virtual void onHandleEvents(const EventQueue& events);

	virtual void onWorldAdd();

	void serializeAbstractScene(Serialize& out) const;
	
	void deserializeAbstractScene(Deserialize& in);




	//!The width of the camera in the scene in pixels
	const int viewWidth;
	//!The height of the camera in the scene in pixels
	const int viewHeight;

	InputHandle input;

private:

	//!The World this AbstractScene belongs to (null by default unless added to one)
	World *world;
	//!Shows if the AbstractScene will update()
	bool running;
	//!Shows if the AbstractScene will draw()
	bool visible;
	//!The Game this AbstractScene belongs to
	Game *game;
	//!A queue of all new Events
	EventQueue eventQueue;
	//!All widgets registered to the AbstractScene
	std::vector<gui::GUIWidget*> widgets;
	//! Whether this scene is currently inside the update() method
	bool isInUpdate;
	//! To solve some bugs, we don't add widgets until after the scene is done updating
	std::vector<gui::GUIWidget*> toRegister;
#ifdef SL_DEBUG
	std::map<std::string, std::vector<std::tuple<Rect<int>, sf::Color, sf::Color>>> debugRects;
	sf::RectangleShape debugRenderRect;
	std::map<std::string, std::vector<sf::VertexArray>> debugLines;
	struct DebugText
	{
		std::string str;
		sf::Color col;
		Vector2f pos;
	};
	std::map<std::string, std::vector<DebugText>> debugStrings;
	//! cache of all recently used sprite texsts so we can avoid re-rendering, and also share texts. <text, <drawable, stepsUntilRemove>>
	std::map<std::string, std::pair<std::unique_ptr<SpriteText>, int>> cachedSpriteTexts;
	std::map<std::string, std::vector<std::tuple<Vector2f, float, sf::Color, sf::Color>>> debugCircles;
	sf::CircleShape debugRenderCircle;
#endif
	Timestamp timestamp;
	//!Depth the scene gets drawn at if there are multiple scenes being drawn. Lower is on top
	int depth;
	Controller playerController;


	friend class World;
};

template <typename T>
void AbstractScene::debugDrawLines(const std::string& category, const T& points, sf::Color color)
{
#ifdef SL_DEBUG
	sf::VertexArray lines(sf::PrimitiveType::LinesStrip);
	for (const auto& point : points)
	{
		lines.append(sf::Vertex(sf::Vector2f(point.x, point.y), color));
	}
	debugLines[category].push_back(std::move(lines));
#endif // SL_DEBUG
}

}// End of sl namespace

#endif // SL_ABSTRACTSCENE_HPP
