#include "Engine/Console.hpp"

#include <sstream>

#include "Engine/Game.hpp"
#include "Engine/Events/Event.hpp"
#include "Engine/Input/Input.hpp"
#include "Engine/GUI/FontManager.hpp"
#include "Parsing/Binding/Bindings.hpp"
#include "Parsing/Parsers/ExpressionParser.hpp"
#include "Parsing/Errors/SyntaxError.hpp"
#include "Town/Cars/CarSpecs.hpp"
#include "Utility/Memory.hpp"

const int xoffset = 16;
const int yoffset = 8;
const int height = 32;

namespace sl
{

Console::Console(Game& game)
	:game(game)
	,context(game.getContext())
	,input(game.getRawInput())
	,background()
	,inputText()
	,outputText()
	,enabled(false)
	,history()
	,historyIndex(0)
	,backspaceCooldown(0)
{
}

void Console::init()
{
	const int cWidth = game.getWindowWidth() - 2 * xoffset;
	background.setSize(sf::Vector2f(cWidth, height));
	background.setPosition(xoffset, yoffset);
	background.setFillColor(sf::Color(0, 0, 0, 128));
	background.setOutlineColor(sf::Color(255, 255, 255, 192));
	background.setOutlineThickness(1);

	inputText.setPosition(xoffset, yoffset);
	inputText.setFillColor(sf::Color(255, 255, 255, 192));
	inputText.setOutlineColor(sf::Color(255, 255, 255, 192));
	inputText.setFont(gui::FontManager::getFont("resources/arial.ttf"));
	inputText.setCharacterSize(12);

	outputText.setPosition(xoffset, yoffset + 16);
	outputText.setFillColor(sf::Color(255, 255, 255, 192));
	outputText.setOutlineColor(sf::Color(255, 255, 255, 192));
	outputText.setFont(gui::FontManager::getFont("resources/arial.ttf"));
	outputText.setCharacterSize(12);
}

void Console::update()
{
	if (enabled)
	{
		/*			INPUT			*/
		const bool shift = input.isKeyHeld(Input::KeyboardKey::LShift) || input.isKeyHeld(Input::KeyboardKey::RShift);
		const char numberAlts[10] = {')', '!', '@', '#', '$', '%', '^', '&', '*', '('};
		for (int i = 0; i < 10; ++i)
		{
			if (input.isKeyPressed((Input::KeyboardKey)((int) Input::KeyboardKey::Num0 + i)) || input.isKeyPressed((Input::KeyboardKey)((int)Input::KeyboardKey::Numpad0 + i)))
			{
				inputBuffer += shift ? numberAlts[i] : ('0' + (char) i);
			}
		}
		for (int i = 0; i < 26; ++i)
		{
			if (input.isKeyPressed((Input::KeyboardKey)((int) Input::KeyboardKey::A + i)))
			{
				inputBuffer += 'a' + (char) i + (shift ? (int)('A' - 'a') : 0);
			}
		}
		if (input.isKeyPressed(Input::KeyboardKey::LBracket))
		{
			inputBuffer += shift ? '{' : '[';
		}
		if (input.isKeyPressed(Input::KeyboardKey::RBracket))
		{
			inputBuffer += shift ? '}' : ']';
		}
		if (input.isKeyPressed(Input::KeyboardKey::SemiColon))
		{
			inputBuffer += shift ? ':' : ';';
		}
		if (input.isKeyPressed(Input::KeyboardKey::Comma))
		{
			inputBuffer += shift ? '<' : ',';
		}
		if (input.isKeyPressed(Input::KeyboardKey::Period))
		{
			inputBuffer += shift ? '>' : '.';
		}
		if (input.isKeyPressed(Input::KeyboardKey::Quote))
		{
			inputBuffer += shift ? '"' : '\'';
		}
		if (input.isKeyPressed(Input::KeyboardKey::Slash))
		{
			inputBuffer += shift ? '?' : '/';
		}
		if (input.isKeyPressed(Input::KeyboardKey::BackSlash))
		{
			inputBuffer += shift ? '|' : '\\';
		}
		if (input.isKeyPressed(Input::KeyboardKey::Equal))
		{
			inputBuffer += shift ? '+' : '=';
		}
		if (input.isKeyPressed(Input::KeyboardKey::Dash))
		{
			inputBuffer += shift ? '_' : '-';
		}
		if (input.isKeyPressed(Input::KeyboardKey::Space))
		{
			inputBuffer += ' ';
		}
		if (input.isKeyPressed(Input::KeyboardKey::Add))
		{
			inputBuffer += '+';
		}
		if (input.isKeyPressed(Input::KeyboardKey::Subtract))
		{
			inputBuffer += '-';
		}
		if (input.isKeyPressed(Input::KeyboardKey::Multiply))
		{
			inputBuffer += '*';
		}
		if (input.isKeyPressed(Input::KeyboardKey::Divide))
		{
			inputBuffer += '/';
		}


		/*		CONTROL INPUT		*/
		if (input.isKeyPressed(Input::KeyboardKey::Return))
		{
			this->execute();
			history.push_back(inputBuffer);
			historyIndex = history.size();
			inputBuffer.clear();
		}
		if (input.isKeyPressed(Input::KeyboardKey::BackSpace))
		{
			if (!inputBuffer.empty())
			{
				inputBuffer.pop_back();
				backspaceCooldown = 12; // bigger initial delay than once held.
			}
		}
		else if (input.isKeyHeld(Input::KeyboardKey::BackSpace) && --backspaceCooldown < 0)
		{
			if (!inputBuffer.empty())
			{
				inputBuffer.pop_back();
				backspaceCooldown = 4;
			}
		}
		else if (input.isKeyPressed(Input::KeyboardKey::Up) && historyIndex > 0)
		{
			inputBuffer = history[--historyIndex];
		}
		else if (input.isKeyPressed(Input::KeyboardKey::Down) && historyIndex < history.size() - 1)
		{
			inputBuffer = history[++historyIndex];
		}
	}

	if (input.isKeyPressed(Input::KeyboardKey::Tilde))
	{
		if (enabled)
		{
			enabled = false;
			outputText.setString("");
			//background.setSize(background.getSize().x, height);
		}
		else
		{
			enabled = true;
		}
	}
	inputText.setString(">" + inputBuffer);
}

void Console::execute()
{
	std::stringstream output;
	try
	{
		const char *evalStr = "eval";
		const int evalLen = 4;
		const char *setStr = "set";
		const int setLen = 3;
		const char *spawnStr = "spawn";
		const int spawnLen = 5;
		ls::ExpressionParser expressionParser("n/a");
		if (inputBuffer.compare(0, evalLen, evalStr) == 0)
		{
			std::unique_ptr<ls::Expression> exp(expressionParser.parseEntireExpression(inputBuffer.c_str() + evalLen));
			output << exp->evaluate(context).asString() << std::endl;
		}
		else if (inputBuffer.compare(0, setLen, setStr) == 0)
		{
			const int startIndex = inputBuffer.find_first_not_of(' ', setLen);
			const int endIndex = inputBuffer.find(' ',  startIndex);
			const std::string name(inputBuffer.begin() + startIndex, inputBuffer.begin() + endIndex);
			if (game.getSettings().exists(name))
			{
				std::unique_ptr<ls::Expression> exp(expressionParser.parseEntireExpression(inputBuffer.c_str() + endIndex));
				//output << "Evaluatng [" << (inputBuffer.c_str() + endIndex) << "]" << std::endl;
				const int value = exp->evaluate(context).asInt();
				try
				{	
					game.getSettings().set(name, value);
					output << "Set '" << name << "' to " << value << "\n";
				}
				catch (std::exception& e)
				{
					output << "Couldn't set '" << name << "' to '" << value << "' due to conversion errors or something. Error: " << e.what() << std::endl;
				}
			}
			else
			{
				output << "Setting '" << name << "' doesn't even exist!\n";
			}
		}
		else if (inputBuffer.compare(0, sizeof("show settings"), "show settings") == 0)
		{
			for (auto& p : game.getSettings().hackyIterator())
			{
				output << p.first << "=" << p.second.value << ", ";
			}
		}
		else if (inputBuffer.compare(0, sizeof("toggle debug btree"), "toggle debug btree") == 0)
		{
			game.registerEvent(unique<Event>(Event::Category::Debug, "player's toggle debug btree"));
		}
		else if (inputBuffer.compare(0, spawnLen, spawnStr) == 0)
		{
			std::stringstream ss(inputBuffer.c_str() + spawnLen);
			std::string type;
			ss >> type;
			
			if (type == "loli")
			{
				std::string nationality;
				ss >> nationality;
				if (nationality.empty())
				{
					nationality = "random";
				}
				auto ev = std::make_unique<Event>(Event::Category::Debug, "spawn_loli");
				ev->setFlag("nationality", nationality);
				game.registerEvent(std::move(ev));
			}
			else if (type == "cop")
			{
			}
			else if (type == "car")
			{
				std::map<std::string, CarSpecs::Model> carModels;
				for (int i = 0; i < CarSpecs::Model::NumOfModels; ++i)
				{
					const CarSpecs::Model model = static_cast<CarSpecs::Model>(i);
					carModels[CarSpecs::getSpecs(model).filename] = model;
				}
				std::string modelName;
				ss >> modelName;
				if (modelName.empty())
				{
					modelName = "sportscar"; // default value
				}
				auto it = carModels.find(modelName);
				if (it != carModels.end())
				{
					auto ev = std::make_unique<Event>(Event::Category::Debug, "spawn_car");
					ev->setFlag("model", it->second);
					game.registerEvent(std::move(ev));
				}
				else
				{
					output << "unknown car type '" << modelName << "'" << std::endl;
				}
			}
			else
			{
				output << "Could not spawn a '" << type << "'. Unknown object name." << std::endl;
			}
		}
		else
		{
			output << "Invalid input. Try: 'eval <expression>', 'set <variable> <value>'" << std::endl;
		}
	}
	catch (ls::SyntaxError& see)
	{
		output << "Syntax Error Occured:\n" << see.what() << std::endl;
	}
	catch (ls::RuntimeError& riee)
	{
		output << "Runtime Error Occured:\n" << riee.what() << std::endl;
	}
	if (!output.str().empty())
	{
		outputText.setString("<< " + output.str());
	}
	else
	{
		outputText.setString("");
	}
}

void Console::draw(sf::RenderTarget& target) const
{
	if (enabled)
	{
		target.draw(background);
		target.draw(inputText);
		target.draw(outputText);
	}
}

bool Console::isEnabled() const
{
	return enabled;
}

}
