#include "Engine/Door.hpp"

#include "Engine/Serialization/AutoDeserialize.hpp"
#include "Engine/Serialization/AutoSerialize.hpp"
#include "Engine/Serialization/Serialization.hpp"
#include "Engine/Game.hpp"
#include "Engine/Scene.hpp"
#include "Town/World.hpp"
#include "Utility/TypeToString.hpp"
#include "Utility/Assert.hpp"

#include "Town/StaticGeometry.hpp"

namespace sl
{

DECLARE_TYPE_TO_STRING(Door);

Door::Door(DeserializeConstructor)
	:Entity(0, 0, 0, 0, "Door")
	,target()
	,targetScene(nullptr)
	,entranceNode(0, 0, nullptr)
	,exitNode(0, 0, nullptr)
{
}

Door::Door(Vector2f pos, int w, int h, Vector2f target, Scene *targetScene)
	:Door(pos, w, h, target, targetScene, pf::NodeID(-1, -1, nullptr), pf::NodeID(target.to<int>() / targetScene->getTileSize(), targetScene))
{
}

Door::Door(Vector2f pos, int w, int h, Vector2f target, Scene *targetScene, const pf::NodeID& entrance, const pf::NodeID& exit)
	:Entity(pos.x, pos.y, w, h, "Door")
	,target(target)
	,targetScene(targetScene)
	,entranceNode(entrance)
	,exitNode(exit)
{
	ASSERT(targetScene != nullptr);
}

const Types::Type Door::getType() const
{
	return Entity::makeType("Door");
}

bool Door::isStatic() const
{
	return true;
}

void Door::onUpdate()
{
#ifdef SL_DEBUG
	if (scene->getGame()->getSettings().get("draw_portals"))
	{
		const Vector2f entrancePos((getEntranceNodeID().getX() + 0.5f) * scene->getTileSize(), (getEntranceNodeID().getY() + 0.5f) * scene->getTileSize());
		scene->debugDrawCircle("draw_portals", entrancePos, scene->getTileSize() / 2, sf::Color::Green, sf::Color(16, 255, 64, 64));
		const Vector2f exitPos((getExitNodeID().getX() + 0.5f) * targetScene->getTileSize(), (getExitNodeID().getY() + 0.5f) * targetScene->getTileSize());
		targetScene->debugDrawCircle("draw_portals", exitPos, targetScene->getTileSize() / 2, sf::Color::Red, sf::Color(255, 64, 16, 64));

		scene->debugDrawCircle("draw_portals", Vector2f(pos.x, pos.y), 2.f, sf::Color::Green);
		targetScene->debugDrawCircle("draw_portals", target, 2.f, sf::Color::Red);
	}
#endif // SL_DEBUG
}

void Door::onEnter()
{
	// @HACK-oween
	if (scene->getType() == "Town")
	{
		//scene->addEntityToGrid(new StaticGeometry(getPos().x + 16, getPos().y, 16, 16, std::make_unique<Sprite>("pumpkin.png")));
	}
	if (entranceNode.getScene() == nullptr)
	{
		entranceNode = pf::NodeID(getPos().to<int>() / targetScene->getTileSize(), scene);
	}
}

void Door::enter()
{
	scene->getWorld()->setCurrentScene(targetScene);
}

void Door::transfer(Entity *entity)
{
	targetScene->transferEntity(entity);
	entity->setPos(target);
}

pf::NodeID Door::getEntranceNodeID() const
{
	ASSERT(scene->getTileSize() == entranceNode.getScene()->getTileSize());
	const int ts = scene->getTileSize();
	//scene->debugDrawRect("draw_npc_path_graph", Rect<int>(entranceNode.getX() * ts + 4, entranceNode.getY() * ts + 4, 24, 24), sf::Color::Green);
	return entranceNode;//pf::NodeID(pos.x / ts, pos.y / ts, scene);
}

pf::NodeID Door::getExitNodeID() const
{
	const int ts = scene->getTileSize();
	//scene->debugDrawRect("draw_npc_path_graph", Rect<int>(exitNode.getX() * ts + 2, exitNode.getY() * ts + 2, 28, 28), sf::Color::Red);
	return exitNode;//pf::NodeID(targetX / ts, targetY / ts, target);
}

void Door::serialize(Serialize& out) const
{
	serializeEntity(out);
	AutoSerialize::serialize(out, mask.getBoundingBox());
	AutoSerialize::serialize(out, target);
	SAVEPTR(targetScene);
	AutoSerialize::serialize(out, entranceNode);
	AutoSerialize::serialize(out, exitNode);
}

void Door::deserialize(Deserialize& in)
{
	deserializeEntity(in);
	Rect<int> bounds;
	AutoDeserialize::deserialize(in, bounds);
	mask.set(bounds.left, bounds.top, bounds.width, bounds.height);
	AutoDeserialize::deserialize(in, target);
	LOADPTR(targetScene);
	entranceNode = AutoDeserialize::create<pf::NodeID>(in);
	exitNode = AutoDeserialize::create<pf::NodeID>(in);
}

} // sl