#include "Engine/Entity.hpp"

#include "Engine/Serialization/AutoDeserialize.hpp"
#include "Engine/Serialization/AutoSerialize.hpp"
#include "Engine/Physics/PhysicsComponent.hpp"
#include "Engine/Scene.hpp"

namespace sl
{

#ifdef SL_DEBUG
std::map<std::string, std::pair<unsigned int, unsigned int> > Entity::entityCount;
#endif

Entity::Entity()
	:pos()
	,mask(0, 0, 0, 0, false)
	,scene(nullptr)
	,listNode(*this)
	,multiLink(*this)
	,type("NonPhysical")
	,dead(false)
	,active(true)
	,visible(true)
	,physicsComponent(nullptr)
	,storageType(StorageType::InNone)
	,height(0)
{
#ifdef SL_DEBUG
	entityCount[type].first++;
	entityCount[type].second++;
#endif // SL_DEBUG
}

Entity::Entity(int x, int y, int w, int h, const std::string& type)
	:pos(x, y)
	,mask(x, y, w, h)
	,scene(nullptr)
	,listNode(*this)
	,multiLink(*this)
	,type(type)
	,dead(false)
	,active(true)
	,visible(true)
	,physicsComponent(nullptr)
	,storageType(StorageType::InNone)
	,height(0)
{
	this->init();
#ifdef SL_DEBUG
	entityCount[type].first++;
	entityCount[type].second++;
#endif // SL_DEBUG
}

Entity::~Entity()
{
#ifdef SL_DEBUG
	entityCount[type].first--;
#endif // SL_DEBUG
}

void Entity::update()
{
	for (std::unique_ptr<const Event>& event : events)
	{
		handleEvent(*event);
	}
	events.clear();

	onUpdate();
	mask.updatePosition(pos.x, pos.y);

	if (physicsComponent)
	{
		physicsComponent->update();
	}

	updateDrawables();
	updateBounds();
}

void Entity::draw(RenderTarget& target) const
{
	drawablesToDelete.clear();
	if (!dead && visible)
	{
		//for (std::vector<Drawable*>::const_iterator it = drawables.begin(); it != drawables.end(); ++it)
		//{
		//	target.draw(**it);
		//}
		for (int i = 0; i < drawables.size(); ++i)
			target.draw(*drawables[i]);
		auto mit = drawableSets.find(currentSet);
		if (mit != drawableSets.cend())
		{
			auto& drawable = mit->second;
			//for (const std::unique_ptr<Drawable>& drawable : mit->second)
			{
				target.draw(*drawable);
			}
		}
#ifdef SL_DEBUG
		if (mask.isEnabled())
		{
			scene->debugDrawRect("draw_collision_masks", mask.getBoundingBox(), sf::Color::Yellow);
		}
		for (const std::unique_ptr<Drawable>& drawable : drawables)
		{
			scene->debugDrawRect("draw_drawable_outlines", drawable->getCurrentBounds(), sf::Color::Blue);
		}
		auto it = drawableSets.find(currentSet);
		if (it != drawableSets.end())
		{
			auto& drawable = it->second;
			//for (const std::unique_ptr<Drawable>& drawable : it->second)
			{
				scene->debugDrawRect("draw_drawable_outlines", drawable->getCurrentBounds(), sf::Color::Blue);
			}
		}
#endif
	}
}

bool Entity::isStatic() const
{
	return false;
}

const Vector2<float>& Entity::getPos() const
{
	return pos;
}

void Entity::setPos(const Vector2<float>& pos)
{
	this->pos = pos;
	mask.updatePosition(pos.x, pos.y);
}

void Entity::setPos(float x, float y)
{
	this->pos.x = x;
	this->pos.y = y;
	mask.updatePosition(x, y);
}

const CollisionMask& Entity::getMask() const
{
	return mask;
}

const Rect<int>& Entity::getAbsoluteBounds() const
{
	return absoluteBounds;
}

const Rect<int>& Entity::getDrawableBounds() const
{
	return drawableBounds;
}

void Entity::destroy()
{
	if (!dead)
	{
		// TODO: this didn't used to be here - but it seems like it should. However, the game crashes when you walk indoors if this is here.
		//if (scene)
		//{
		//	onExit();
		//}
		onDestroy();
		dead = true;
		if (scene)
		{
			scene->deleteEntity(this);
		}
		else
		{
			delete this;
		}
	}
}

bool Entity::isDead() const
{
	return dead;
}

bool Entity::isActive() const
{
	return active;
}

const Refable& Entity::getAliveRef() const
{
	return refable;
}

void Entity::setVisible(bool visible)
{
	this->visible = visible;
}

void Entity::setCollisionEnabled(bool collisionEnabled)
{
	mask.setEnabled(collisionEnabled);
}

void Entity::setActive(bool active)
{
	if (active != this->active)
	{
		if (active)
		{
			// TODO: should this be empty? wtf was this here for? (been here for years)
		}
		this->active = active;
	}
}

IntrusiveListNode<Entity>& Entity::getLink()
{
	return listNode;
}

IntrusiveMultiListNode<Entity>& Entity::getMultiLink()
{
	return multiLink;
}

Scene* Entity::getScene() const
{
	return scene;
}

World* Entity::getWorld() const
{
	return getScene()->getWorld();
}

void Entity::handleEvent(const Event& event)
{
	//	override in base class
}

void Entity::handleEventNextTurn(std::unique_ptr<const Event> event)
{
	events.push_back(std::move(event));
}

int Entity::getHeight() const
{
	return height;
}

//const Drawable* Entity::getCurrentAnimation() const
//{
//	if (currentAnim != "none")
//	{
//		const auto it = animations.find(currentAnim);
//		if (it != animations.end())
//		{
//			return it->second.get();
//		}
//	}
//	return nullptr;
//}

//void Entity::setAnimation(const std::string& animName, std::unique_ptr<Drawable>& anim)
//{
//	ASSERT(animName != "none");
//	animations[animName] = std::move(anim);
//}
//
//void Entity::setCurrentAnimation(const std::string& animName)
//{
//	currentAnim = animName;
//}

void Entity::tryMoveRaycast(const Vector2f& motion)
{
	Vector2f end;
	if (scene->raycast(*this, { "StaticGeometry" }, motion, end) != Scene::RaycastResult::DidNotMove)
	{
		setPos(end);
	}
}

/*  protected   */
void Entity::onEnter()
{
	//	empty unless overridden in base classes
}

void Entity::onExit()
{
	//	empty unless overridden in base classes
}

void Entity::addDrawable(std::unique_ptr<Drawable> toAdd)
{
	toAdd->update(pos.x, pos.y);
	drawables.push_back(std::move(toAdd));
}

//void Entity::addDrawable(std::unique_ptr<Drawable>&& toAdd)
//{
//	toAdd->update(pos.x, pos.y);
//	drawables.push_back(toAdd);
//}

void Entity::addDrawableToSet(const std::string& key, std::unique_ptr<Drawable> toAdd)
{
	// @todo remove assert
	//ASSERT(drawableSets.count(key) == 0);
	toAdd->update(pos.x, pos.y);
	drawableSets[key] = std::move(toAdd);
}

void Entity::clearNonSetDrawables()
{
	std::move(drawables.begin(), drawables.end(), std::back_inserter(drawablesToDelete));
	drawables.clear();
}

bool Entity::removeNonSetDrawable(Drawable *toRemove)
{
	for (std::unique_ptr<Drawable>& drawable : drawables)
	{
		if (drawable.get() == toRemove)
		{
			std::swap(drawable, drawables.back());
			drawablesToDelete.push_back(std::move(drawables.back()));
			drawables.pop_back();
			return true;
		}
	}
	return false;
}

void Entity::clearDrawablesInSet(const std::string& set)
{
	auto it = drawableSets.find(set);
	if (it != drawableSets.end())
	{
		drawablesToDelete.push_back(std::move(it->second));
		drawableSets.erase(it);
	}
	
}

void Entity::setAnimationSpeed(float animationSpeed)
{
	auto it = drawableSets.find(currentSet);
	if (it != drawableSets.end())
	{
		it->second->setAnimationSpeed(animationSpeed);
	}
	for (std::unique_ptr<Drawable>& drawable : drawables)
	{
		drawable->setAnimationSpeed(animationSpeed);
	}
}

void Entity::setAnimationSpeed(const std::string& key, float animationSpeed)
{
	auto it = drawableSets.find(key);
	if (it == drawableSets.end())
	{
		ERROR("Trying to set animation speed on set " + key + " that does not exist");
	}
	else
	{
		it->second->setAnimationSpeed(animationSpeed);
	}
}

void Entity::setDrawableSet(const std::string& key)
{
	currentSet = key;
}

const Drawable* Entity::getCurrentDrawable() const
{
	const auto it = drawableSets.find(currentSet);
	if (it != drawableSets.end())
	{
		return it->second.get();
	}
	return nullptr;
}

Drawable* Entity::getDrawable(const std::string& key)
{
	const auto it = drawableSets.find(key);
	if (it != drawableSets.end())
	{
		return it->second.get();
	}
	return nullptr;
}

const std::string& Entity::getCurrentDrawableSet() const
{
	return currentSet;
}

void Entity::enablePhysics()
{
	if (!physicsComponent)
	{
		physicsComponent = std::make_unique<PhysicsComponent>(*this, scene);
	}
	else
	{
		//	say something?
	}
}

PhysicsComponent& Entity::getPhysics()
{
	this->enablePhysics();
	return *physicsComponent;
}

void Entity::calculateHeight()
{
	height = scene->getHeightAt(pos.x, pos.y);
}

void Entity::setHeight(int h)
{
	height = h;
}

void Entity::setCollisionMask(CollisionMask mask)
{
	this->mask = std::move(mask);
}

void Entity::onDestroy()
{
	// override in base class
}

void Entity::serializeEntity(Serialize& out) const
{
	AutoSerialize::serialize(out, pos);
	// mask?
	SAVEVAR(ptr, scene);
	// drawables handled separately
	//listNode.serialize(out);
	//multilink?
	//type ignored
	SAVEVAR(b, dead);
	SAVEVAR(b, active);
	SAVEVAR(b, visible);
	AutoSerialize::serialize(out, absoluteBounds);
	AutoSerialize::serialize(out, drawableBounds);
	// ignored physics component?
	out.u8("storageType", static_cast<uint8_t>(storageType));
	SAVEVAR(i, height);
	AutoSerialize::serialize(out, origin);
}

void Entity::deserializeEntity(Deserialize& in)
{
	// @SAVELOADHASIGNORED
	AutoDeserialize::deserialize(in, pos);
	// mask?
	scene = static_cast<Scene*>(in.ptr("scene"));
	// drawables handled separately
	//listNode.deserialize(in);
	//multilink?
	//type ignored
	LOADVAR(b, dead);
	LOADVAR(b, active);
	LOADVAR(b, visible);
	AutoDeserialize::deserialize(in, absoluteBounds);
	AutoDeserialize::deserialize(in, drawableBounds);
	// ignored physics component?
	storageType = static_cast<StorageType>(in.u8("storageType"));
	LOADVAR(i, height);
	AutoDeserialize::deserialize(in, origin);
}

void Entity::serializeDrawables(Serialize& out) const
{
	AutoSerialize::serialize(out, drawables);
	AutoSerialize::serialize(out, drawableSets);
	SAVEVAR(s, currentSet);
}

void Entity::deserializeDrawables(Deserialize& in)
{
	AutoDeserialize::deserialize(in, drawables);
	AutoDeserialize::deserialize(in, drawableSets);
	LOADVAR(s, currentSet);
	init();
}

void Entity::forceTypeChangeForDeserialization(const Deserialize&, std::string newType)
{
	type = std::move(newType);
	if (scene)
	{
		const StorageType storage = storageType;
		storageType = StorageType::InNone;
		scene->addImmediately(this, storage);
	}
}

/*		private		*/
void Entity::init()
{
	mask.updatePosition(pos.x, pos.y);

	if (physicsComponent)
	{
		physicsComponent->update();
	}
	// Update sprite/other drawable position(s) and other stuff here maybe
	for (std::unique_ptr<Drawable>& drawable : drawables)
	{
		drawable->update(pos.x, pos.y);
	}
	auto mit = drawableSets.find(currentSet);
	if (mit != drawableSets.end())
	{
		mit->second->update(pos.x, pos.y);
	}
	this->updateBounds();
}

void Entity::updateBounds()
{
	Rect<int> temp;
	for (std::unique_ptr<Drawable>& drawable : drawables)
	{
		drawable->update(pos.x, pos.y - drawable->getHeight());
		temp = drawable->getCurrentBounds();
		if (temp.left < absoluteBounds.left)
		{
			absoluteBounds.left = temp.left;
		}
		if (temp.top < absoluteBounds.top)
		{
			absoluteBounds.top = temp.top;
		}
		if (temp.left + temp.width > absoluteBounds.left + absoluteBounds.width)
		{
			absoluteBounds.width = temp.left + temp.width - absoluteBounds.left;
		}
		if (temp.top + temp.height > absoluteBounds.top + absoluteBounds.height)
		{
			absoluteBounds.height = temp.top + temp.height - absoluteBounds.top;
		}
	}
	auto mit = drawableSets.find(currentSet);
	if (mit != drawableSets.end())
	{
		auto& drawable = mit->second;
		//for (std::unique_ptr<Drawable>& drawable : drawables)
		{
			drawable->update(pos.x, pos.y - drawable->getHeight());
			temp = drawable->getCurrentBounds();
			if (temp.left < absoluteBounds.left)
			{
				absoluteBounds.left = temp.left;
			}
			if (temp.top < absoluteBounds.top)
			{
				absoluteBounds.top = temp.top;
			}
			if (temp.left + temp.width > absoluteBounds.left + absoluteBounds.width)
			{
				absoluteBounds.width = temp.left + temp.width - absoluteBounds.left;
			}
			if (temp.top + temp.height > absoluteBounds.top + absoluteBounds.height)
			{
				absoluteBounds.height = temp.top + temp.height - absoluteBounds.top;
			}
		}
	}
	drawableBounds = temp;
	temp = mask.getBoundingBox();
	if (temp.left < absoluteBounds.left)
	{
		absoluteBounds.left = temp.left;
	}
	if (temp.top < absoluteBounds.top)
	{
		absoluteBounds.top = temp.top;
	}
	if (temp.left + temp.width > absoluteBounds.left + absoluteBounds.width)
	{
		absoluteBounds.width = temp.left + temp.width - absoluteBounds.left;
	}
	if (temp.top + temp.height > absoluteBounds.top + absoluteBounds.height)
	{
		absoluteBounds.height = temp.top + temp.height - absoluteBounds.top;
	}
	absoluteBounds = temp;
}

void Entity::updateDrawables()
{
	// Update sprite/other drawable position(s) and other stuff here maybe
	for (std::unique_ptr<Drawable>& drawable : drawables)
	{
		const int heightDelta = height - drawable->getHeight();
		drawable->setHeight(height);
		//drawable->update(pos.x, pos.y - height);
	}
	auto mit = drawableSets.find(currentSet);
	if (mit != drawableSets.end())
	{
		auto& drawable = mit->second;
		//for (std::unique_ptr<Drawable>& drawable : mit->second) 
		{
			const int heightDelta = height - drawable->getHeight();
			drawable->setHeight(height);
			//drawable->update(pos.x, pos.y - height);
		}
	}
}

/*  public static   */
Types::Type Entity::makeType(const char *str)
{
	return str;
}

std::string Entity::typeToString(Types::Type type)
{
	//  change this when the type system doesn't suck dicks
	return std::string(type);
}

}// End of sl namespace
