/**
 * @section DESCRIPTION
 * The base class for all physical entites in PhysicsScenes.
 */
#ifndef SL_ENTITY_HPP
#define SL_ENTITY_HPP

#include <SFML/Graphics/RenderTarget.hpp>
#include <SFML/Graphics/RectangleShape.hpp>
#include <SFML/Graphics/Rect.hpp>
#include <SFML/System/Vector2.hpp>

#include <map>
#include <memory>
#include <initializer_list>
#include <string>
#include <vector>

#include "Engine/Serialization/Serializable.hpp"
#include "Engine/Refable.hpp"
#include "Engine/Types.hpp"
#include "Engine/Graphics/RenderTarget.hpp"
#include "Engine/Physics/CollisionMask.hpp"
#include "Utility/IntrusiveList.hpp"
#include "Utility/IntrusiveMultiList.hpp"
#include "Utility/Vector.hpp"

namespace sl
{

class Event;
class Scene;
class PhysicsComponent;
class World;

enum StorageType : char;

class Entity : public Serializable
{
public:
#ifdef SL_DEBUG
	static std::map<std::string, std::pair<unsigned int, unsigned int> > entityCount;
#endif
	//!An Entity Type to distinguish concrete versions of entities from each other.
	typedef std::string Type;

	/**
	 * Polymorphic destructor that performs cleanup on the Entity and deletes all Drawables.
	 */
	virtual ~Entity();

	/**
	 * Updates the Entity by one cycle and calls the onUpdate() method on the subclass.
	 */
	void update();

	/**
	 * Renders all the Entity's Drawables.
	 * @param target RenderTarget to render the Drawables to
	 */
	void draw(RenderTarget& target) const;

	/**
	 * Checks if the Entity is static (doesn't move)
	 * @return Returns true if the Entity doesn't move
	 */
	virtual bool isStatic() const;

	/**
	 * A pure virtual method that gets called by update(). Used to update subclass-specific things.
	 */
	virtual void onUpdate() = 0;

	/**
	 * Checks the Entity's dynamic type.
	 * @return Returns a Types::Type that is the given Entity instance's true type.
	 */
	virtual const Type getType() const = 0;

	/**
	 * Gets the Entity's current position.
	 * @return Returns a const reference to the Entity's current position.
	 */
	const Vector2<float>& getPos() const;

	/**
	 * Sets the Entity's position (and moves the mask to keep it valid)
	 * @param pos The new position
	 */
	void setPos(const Vector2<float>& pos);

	/**
	 * Sets the Entity's position (and moves the mask to keep it valid)
	 * @param x The new x position
	 * @param y The new y position
	 */
	void setPos(float x, float y);

	/**
	 * Gets the Entity's CollisionMask
	 * @return Returns a constant reference to the Entity's CollisionMask
	 */
	const CollisionMask& getMask() const;

	/**
	 * Gets the absolute (union of CollisionMask+all Drawables) bounding box of this Entity
	 * @return The bounding box
	 */
	const Rect<int>& getAbsoluteBounds() const;

	/**
	 * Gets the current bounds of this Entity's visual footprint
	 * @return The bounding box
	 */
	const Rect<int>& getDrawableBounds() const;

	/**
	 * Marks this Entity for removal
	 */
	void destroy();

	/**
	 * Checks if an instance is marked for removal
	 * @return whether it's marked for removal
	 */
	bool isDead() const;

	/**
	 * Checks if an instance is active (drawn/updated/collision detected)
	 * @return whether the entity is active
	 */
	bool isActive() const;

	/**
	 * @return Obtains an alive reference to the Entity. This should ONLY be used from Ref<Entity> or anything owning an Entity to check if it's okay to delete.
	 */
	const Refable& getAliveRef() const;

	/**
	 * Sets whether or not the Entity is rendered in the AbstractScene
	 * @param visible True for rendered, False for not
	 */
	void setVisible(bool visible);

	void setCollisionEnabled(bool collisionEnabled);

	void setActive(bool active);

	/**
	 * Gets the node of this object in whatever list it's stored in. (Used by Scene)
	 */
	IntrusiveListNode<Entity>& getLink();

	/**
	 * @return node for multi-list nodes for whatever non-owning lists it's stored in (Used by other systems)
	 */
	IntrusiveMultiListNode<Entity>& getMultiLink();

	/**
	 * Creates a Types::Type from a given string
	 * @param str String key to convert to a Type
	 * @return Returns the generated Types::Type
	 */
	static Types::Type makeType(const char *str);

	static std::string typeToString(Type type);


	/**
	 * Adds a Drawable to the Entity post-creation.
	 * @param toAdd The Drawable to add to the Entity.
	 */
	void addDrawable(std::unique_ptr<Drawable> toAdd);
	//void addDrawable(std::unique_ptr<Drawable>&& toAdd);

	//@todo change this to like setDrawableSet()
	/**
	 * Adds a Drawable into a set of Drawables referenced by a key
	 * @param key The set to insert into
	 * @param toAdd The Drawable to add to it
	 */
	void addDrawableToSet(const std::string& key, std::unique_ptr<Drawable> toAdd);

	/**
	 * Deletes all non-set Drawables attached to this Entity
	 */
	void clearNonSetDrawables();

	/**
	 * @param toRemove The drawable to remove from the non-set Drawables
	 * @return True if it was removed, False if it was not found
	 */
	bool removeNonSetDrawable(Drawable *toRemove);

	/**
	 * Deletes all Drawables in the given set
	 * @param set The set to clear
	 */
	void clearDrawablesInSet(const std::string& set);

	/**
	 * Sets which set of Drawables that the Entity currently renders
	 * @param key The set to render from now on
	 */
	void setDrawableSet(const std::string& key);

	/**
	 * Sets the animation speed of the currently rendered Drawables (only matters for Animations)
	 * @param animationSpeed The speed to animate in terms of <frames advanced in animation>/<frame of gameplay>
	 */
	void setAnimationSpeed(float animationSpeed);
	/**
	 * Sets the animation speed for a certain set (only matters for Animations)
	 * @param animationSpeed The speed to animate in terms of <frames advanced in animation>/<frame of gameplay>
	 */
	void setAnimationSpeed(const std::string& key, float animationSpeed);

	const Drawable* getCurrentDrawable() const;

	Drawable* getDrawable(const std::string& key);

	const std::string& getCurrentDrawableSet() const;

	/**
	 * Returns the AbstractScene this Entity exists within
	 * @return The scene
	 */
	Scene* getScene() const;

	World* getWorld() const;

	/**
	 * Reacts to an event getting sent to this Entity
	 * @param event The event to handle
	 */
	virtual void handleEvent(const Event& event);

	/**
	 * Handles an event the next turn. This is here to get around issues where handling
	 * it right away would cause issues.
	 * @param even The event to handle
	 */
	void handleEventNextTurn(std::unique_ptr<const Event> event);

	int getHeight() const;

	///**
	// * @return the currently playing Animation or nullptr if no current animation
	// */
	//const Animation* getCurrentAnimation() const;

	//void setAnimation(const std::string& animName, std::unique_ptr<Animation>& anim);

	//void setCurrentAnimation(const std::string& animName);

	/**
	 * Raytraces to try and move this entity while avoiding the specified
	 */
	void tryMoveRaycast(const Vector2f& motion);

protected:
	/**
	 * This constructor is for entities that won't have any physical prescence
	 */
	Entity();
	/**
	 * Entity constructor for use by subclasses using the bareminimum of needed information.
	 * @param x The x position to create the Entity at.
	 * @param y The y positiion to creaste the Entity at.
	 * @param w The width this Entity will take up in the Scene. Used for the CollisionMask.
	 * @param h The height this Entity will take up in the Scene. Used for the CollisionMask.
	 * @param type The concrete type of the entity.
	 */
	Entity(int x, int y, int w, int h, const std::string& type);



	/**
	 * Code that gets called when an Entity enters a scene
	 * Note: before this is called for the first time, scene will be null
	 */
	virtual void onEnter();

	/**
	 * Code that gets called when an Entity leaves a scene
	 */
	virtual void onExit();

	/*				Physics stuff				*/
	/**
	 * Enables Physics which creates a PhysicsComponent for this Entity
	 */
	void enablePhysics();
	/**
	 * Gets a reference to the PhysicsComponent, creating it if it doesn't exist.
	 */
	PhysicsComponent& getPhysics();
	/**
	 * Calculates height height of this Entity by checking the AbstractScene's HeightMaps
	 */
	void calculateHeight();
	/**
	 * Manually sets the Entity's height
	 */
	void setHeight(int h);
	/**
	 * Resets the collision mask for this Entity
	 * @param mask The new mask to use
	 */
	void setCollisionMask(CollisionMask mask);



	/**
	 * A method where you should put stuff that happens when an instance is marked for removal
	 * that isn't memory cleanup - leave the memory cleanup for the destructor.
	 */
	virtual void onDestroy();


	void serializeEntity(Serialize& out) const;

	void deserializeEntity(Deserialize& in);

	void serializeDrawables(Serialize& out) const;

	void deserializeDrawables(Deserialize& in);

	/**
	 * This should probably almost never be used. Right now the only use-case is deserialization of marker-type Entities
	 * This changes the Type of the Entity, while also re-inserting it into the right Scene structure (if in one)
	 * @param in Just to verify tha that people aren't abusing this, this is a dummy parameter
	 */
	void forceTypeChangeForDeserialization(const Deserialize&, std::string newType);

	//!A 2D vector used to keep track of the Entity's position in the Scene
	Vector2<float> pos;
	//!The CollisionMask used to represent the Entity's physical presence in a Scene
	CollisionMask mask;
	//!The Scene in which the Entity lives in.
	Scene *scene;

private:
	Entity(const Entity&);
	Entity& operator=(const Entity&);
	/**
	 * Sets up everything properly so that there are no issues with grids
	 * (basically moves drawables/masks to right location)
	 */
	void init();

	/**
	 * Updates the drawable + absolute bounds to keep them valid for the current frame
	 */
	void updateBounds();

	void updateDrawables();



	//!Contains all the Drawables for rendering this Entity.
	std::vector<std::unique_ptr<Drawable>> drawables;
	//!List of sets of Drawables that are available
	std::map<std::string, std::unique_ptr<Drawable>> drawableSets;
	//!ID of the current set to draw
	std::string currentSet;
	////!Animations
	//std::map<std::string, std::unique_ptr<Animation>> animations;
	////!Current animation
	//std::string currentAnim;
	//!Intrusive list node pointer to this Entity
	IntrusiveListNode<Entity> listNode;
	//!Additional lists the object might be in
	IntrusiveMultiListNode<Entity> multiLink;
	//!The concrete type of an Entity
	std::string type;
	//!Flag for removal
	bool dead;
	//!Whether the Entity is active (rendered/updated/collision detected)
	bool active;
	//!Whether or not the Entity is rendered
	bool visible;
	//!This is the maximum bounds of this object as a union of both CollisionMask and all Drawables
	Rect<int> absoluteBounds;
	//!This is the current bounds for the union of the bounds of all Drawables currently being drawn
	Rect<int> drawableBounds;
	//!The physics model for this Entity.
	std::unique_ptr<PhysicsComponent> physicsComponent;
	//Which kind of structure this Entity is stored in in its AbstractScene. (List, QuadTree, etc)
	StorageType storageType;
	//!The elevation of this object. Only affects rendering. See HeightMap for more info
	int height;
	//!List of events sent specifically to this Entity which will be sent at the start of an update
	std::vector<std::unique_ptr<const Event>> events;
	//!Origin that the collision mask is offset by
	Vector2i origin;
	//! To avoid deleting drawables that were already sent to the RenderTarget, simply store them for deletion before the next draw call
	mutable std::vector<std::unique_ptr<Drawable>> drawablesToDelete;
	//!
	Refable refable;

	friend class Scene;	//	because I can't friend private methods in Scene
};

}// End of sl namespace

#endif
