/**
 * @section DESCRIPTION
 * A basic event class for sending to event handlers.
 */
#ifndef SL_EVENT_HPP
#define SL_EVENT_HPP

#include <functional>
#include <map>
#include <string>

#include "Parsing/Binding/Value.hpp"
#include "Utility/Assert.hpp"
#include "Utility/TypeToString.hpp"
#include "Utility/Vector.hpp"

namespace sl
{

class Event
{
public:
	typedef std::function<void(const std::string&)> ResponseFunction;

	enum Category : int
	{
		General = 0,
		NPCInteraction,
		Debug,
		//Crime,
		NonBroadcasted,

		EventTypeNum
	};
	/**
	 * Constructs a basic Event
	 * @param category The category of event this is
	 * @param name The name identifier for the Event
	 */
	Event(Category category, const std::string& name);
	/**
	* Constructs a basic Event
	* @param category The category of event this is
	* @param name The name identifier for the Event
	* @paran location The locaiton this event is associated with
	*/
	Event(Category category, const std::string& name, const Vector2f& location);
	/**
	 * Constructs a basic Event that can be responded to
	 * @param category The category of event this is
	 * @param name The name identifier for the event
	 * @param responder The function to be called when someone responds
	 */
	Event(Category category, const std::string& name, ResponseFunction responder);
	/**
	* Constructs a basic Event that can be responded to
	* @param category The category of event this is
	* @param name The name identifier for the event
	* @param responder The function to be called when someone responds
	* @paran location The locaiton this event is associated with
	*/
	Event(Category category, const std::string& name, ResponseFunction responder, const Vector2f& location);
	
	// for some reason MSVC++ isn't wanting to generate a move ctor for this class
	// yet a naive implementation works? whyyyyyyyyyyyyy
	Event(Event&& other);
	Event& operator=(Event&& rhs);

	/**
	 * @param flag The flag to set
	 * @param value The value to set it to
	 */
	void setFlag(const std::string& flag, ls::Value value);

	template <typename T>
	void setFlag(const std::string& flag, const T& value)
	{
		setFlag(flag, ls::Value(value));
	}

	/**
	 * Gets the category this event is registered in
	 * @return The event Category
	 */
	Category getCategory() const;
	/**
	 * Getter method for the event name
	 * @return the event name
	 */
	const std::string& getName() const;
	/**
	 * @return The location of the event, or nullptr of no location associated with it
	 */
	const Vector2f* getLocation() const;

	/**
	 * Responds to an event and sends the message back to the originating callback
	 * @param msg The answer to the event
	 */
	void respond(const std::string& msg) const;

	/**
	 * @param flag The flag to get
	 * @return The value of the flag
	 */
	const ls::Value& getFlag(const std::string& flag) const;

	static Event copy(const Event& ev);

private:
	Event(const Event&); // only callable from copy()
	Event& operator=(const Event&) = delete;//forbid

	//!The Category this event is in
	Category category;
	//!Name identifier for the event
	std::string name;
	//!The function to call if anyone responds to the event
	ResponseFunction responder;
	//!
	std::map<std::string, ls::Value> flags;
	//! Location of the event - set to NO_LOCATION otherwise
	Vector2f location;

	static const std::string invalidString;

	static const Vector2f NO_LOCATION;
};

} // sl

#endif // SL_EVENT_HPP
