#include "Engine/GUI/BasicMenuBuilder.hpp"

#include "Engine/GUI/Button.hpp"
#include "Engine/GUI/NormalBackground.hpp"
#include "Engine/GUI/TextLabel.hpp"

namespace sl
{
namespace gui
{

BasicMenuBuilder::BasicMenuBuilder(int menuWidth, int maxHeight, int buttonGap, int minButtonSize, int fontSize)
	:menuWidth(menuWidth)
	,maxHeight(maxHeight)
	,buttonGap(buttonGap)
	,heightSoFar(0)
	,minButtonSize(minButtonSize)
	,fontSize(fontSize)
	,controller(unique<MenuController>(Rect<int>(0, 0, menuWidth, maxHeight)))
{
}

BasicMenuBuilder::~BasicMenuBuilder()
{
}

void BasicMenuBuilder::addButton(const std::string& text, const std::function<void()>& function)
{
	addDynamicButton(text, [function](TextLabel*){
		function();
	});
}

void BasicMenuBuilder::addDynamicButton(const std::string& text, const std::function<void(TextLabel*)>& function)
{
	const Rect<int> buttonBox(0, heightSoFar, menuWidth, 160);
	std::unique_ptr<TextLabel> textLabel = unique<TextLabel>(buttonBox, text, TextLabel::Config().centered().pad(8, 2));
	textLabel->setFontSize(fontSize);
	TextLabel *textLabelPtr = textLabel.get();

	textLabel->extendVerticallyToFitText(minButtonSize);
	std::unique_ptr<Button> button(new Button(std::move(textLabel),
					                sf::Color(192, 192, 192), sf::Color::White,
					                NormalBackground(sf::Color::Black, sf::Color(192, 192, 192)),
					                NormalBackground(sf::Color(32, 32, 32), sf::Color::White)));
	heightSoFar += buttonGap + button->getBox().height;
	auto callback = std::bind(function, textLabelPtr);
	button->registerMouseInputEvent(Key::GUIClick, GUIWidget::InputEventType::Released, callback);
	
	controller->addButton(std::move(button), std::move(callback));

	controller->setSize(menuWidth, heightSoFar);
}

void BasicMenuBuilder::addBackButton(std::function<void()> function)
{
	addButton("Back", function);
	controller->registerInputEvent(Key::GUIBack, GUIWidget::InputEventType::Released, std::move(function));
}


std::unique_ptr<GUIWidget> BasicMenuBuilder::create()
{
	return std::move(controller);
}



// ---------------------- MenuController ------------------------------
BasicMenuBuilder::MenuController::MenuController(const Rect<int>& box)
	:GUIWidget(box)
	,hoverIndex(0)
{
	registerInputEvent(Key::GUIUp, InputEventType::Released, [this] {
		if (hoverIndex > 0)
		{
			select(hoverIndex - 1);
		}
	});
	registerInputEvent(Key::GUIDown, InputEventType::Released, [this] {
		if (hoverIndex < callbacks.size() - 1)
		{
			select(hoverIndex + 1);
		}
	});
	registerInputEvent(Key::GUIAccept, InputEventType::Released, [this] {
		callbacks[hoverIndex]();
	});
}

void BasicMenuBuilder::MenuController::addButton(std::unique_ptr<GUIWidget> button, std::function<void()> callback)
{
	buttons.push_back(button.get());
	callbacks.push_back(std::move(callback));
	addWidget(std::move(button));
}

// private
void BasicMenuBuilder::MenuController::select(int index)
{
	buttons[hoverIndex]->setFocus(false);
	buttons[index]->setFocus(true);
	hoverIndex = index;
}

} // gui
} // sl
