#include "Engine/GUI/Button.hpp"

#include "Engine/GUI/NormalBackground.hpp"
#include "Engine/GUI/WidgetBackground.hpp"

namespace sl
{
namespace gui
{

Button::Button(const Rect<int>& box, std::unique_ptr<TextLabel> text, sf::Color mouseOffColour, sf::Color mouseOnColour,
			   const WidgetBackground& mouseOffBackground, const WidgetBackground& mouseOnBackground)
	:GUIWidget(box)
	,text(*text)
	,mouseOffColour(std::move(mouseOffColour))
	,mouseOnColour(std::move(mouseOnColour))
	,mouseOffBackground(mouseOffBackground.clone())
	,mouseOnBackground(mouseOnBackground.clone())
	,selected(false)
{
	this->addWidget(std::move(text));
	this->onFocusLostOrMouseLeave(); // to initialize the background
}

Button::Button(std::unique_ptr<TextLabel> text, sf::Color mouseOffColour, sf::Color mouseOnColour,
			   const WidgetBackground& mouseOffBackground, const WidgetBackground& mouseOnBackground)
	:GUIWidget(text->getBox())
	,text(*text)
	,mouseOffColour(std::move(mouseOffColour))
	,mouseOnColour(std::move(mouseOnColour))
	,mouseOffBackground(mouseOffBackground.clone())
	,mouseOnBackground(mouseOnBackground.clone())
	,selected(false)
{
	this->addWidget(std::move(text));
	this->onFocusLostOrMouseLeave(); // to initialize the background
}

bool Button::isSelected() const
{
	return selected;
}

TextLabel& Button::getText()
{
	return text;
}

void Button::setOnHover(std::function<void()> onHover)
{
	this->onHover = onHover;
}


/*	  GUIWidget private methods	   */
void Button::onDraw(sf::RenderTarget& target) const
{
}

void Button::onFocusGainedOrMouseEnter()
{
	selected = true;
	if (mouseOnBackground)
	{
		this->applyBackground(*mouseOnBackground);
	}
	text.setColour(mouseOnColour);
	if (onHover)
	{
		onHover();
	}
}

void Button::onFocusLostOrMouseLeave()
{
	selected = false;
	if (mouseOffBackground)
	{
		this->applyBackground(*mouseOffBackground);
	}
	text.setColour(mouseOffColour);
}

// static
std::unique_ptr<Button> Button::createGenericButton(const Rect<int>& box, const std::string& text, std::function<void()> callback, int fontSize, TextLabel::HorizontalPolicy halign)
{
	std::unique_ptr<TextLabel> textLabel = unique<TextLabel>(box, text, TextLabel::Config().halign(halign).valign(TextLabel::Middle).fontSize(fontSize));
	auto button = unique<Button>(std::move(textLabel),
		sf::Color(192, 192, 192), sf::Color::White,
		NormalBackground(sf::Color::Black, sf::Color(192, 192, 192)),
		NormalBackground(sf::Color(32, 32, 32), sf::Color::White)
	);
	button->registerMouseInputEvent(Key::GUIClick, GUIWidget::InputEventType::Pressed, callback);
	return std::move(button);
}

} // gui
} // sl
