#include "Engine/GUI/GUIWidget.hpp"

#include "Utility/Assert.hpp"
#include "Engine/GUI/WidgetBackground.hpp"
#include "Engine/AbstractScene.hpp"
#include "Engine/Game.hpp"

namespace sl
{
namespace gui
{

GUIWidget::GUIWidget(const Rect<int>& box)
	:box(box)
	,scene(nullptr)
	,parent(nullptr)
	,widgets()
	,mouseWasOverLastStep(false)
	,handlesMouseEvents(false)
	,depth(0)
	,background(nullptr)
	,toMove(0, 0)
	,isExecutingEvent(false)
	,updating(false)
	,focused(false)
{
}

GUIWidget::~GUIWidget()
{
	//	can't do range-based or iterators here or else iterators get invalidated
	//for (int i = 0; i < callLater.size(); ++i)
	//{
	//	callLater[i]();
	//}
	//callLater.clear();
}

/*			GUIWidget public methods			*/
void GUIWidget::update()
{
	update(box.left, box.top);
}

void GUIWidget::update(int x, int y)
{
	const int xdif = x - box.left + toMove.x;
	const int ydif = y - box.top + toMove.y;

	updating = true;

	this->onUpdate(x + toMove.x, y + toMove.y);

	if (xdif || ydif)
	{
		box.left = x + toMove.x;
		box.top = y + toMove.y;
		if (background)
		{
			background->move(box.left, box.top);
		}
	}
	if (!parent)
	{
		if (scene)
		{
			const sf::Vector2f& mouse = scene->getGame()->getRawInput().mousePosF();
			std::set<Key> handled;
			handleInputEvents(handled);
			handleMouseInputEvents(handled, mouse, box);
		}
	}
	// update could cause this vector to reallocate self and invalidate iterators, hence regular loop
	for (unsigned int i = 0; i < widgets.size(); ++i)
	{
		widgets[i]->update(widgets[i]->box.left + xdif, widgets[i]->box.top + ydif);
	}
	toMove.x = 0;
	toMove.y = 0;

	updating = false;
	for (int i = 0; i < callLater.size(); ++i)
	{
		callLater[i]();
	}
	callLater.clear();
	// woo smart pointers
	deleteLater.clear();
}

void GUIWidget::move(int x, int y)
{
	toMove.x = x - box.left;
	toMove.y = y - box.top;
	for (auto& widget : widgets)
	{
		//widget->move(widget->box.left + toMove.x, widget->box.top + toMove.y);
	}
}

void GUIWidget::moveBy(int x, int y)
{
	toMove.x += x;
	toMove.y += y;
}

void GUIWidget::draw(sf::RenderTarget& target) const
{
	this->beforeDraw(target);
	if (background)
	{
		background->draw(target);
	}
	this->onDraw(target);
	for (const auto& widget : widgets)
	{
		widget->draw(target);
	}
}

int GUIWidget::getDepth() const
{
	return parent ? parent->getDepth() + 1 : depth;
}

void GUIWidget::setDepth(int depth)
{
	ASSERT(parent == nullptr);
	this->depth = depth;
}

const Rect<int>& GUIWidget::getBox() const
{
	return box;
}

void GUIWidget::setSize(int width, int height)
{
	ASSERT(width >= 0 && height >= 0);
	box.width = width;
	box.height = height;
	if (background)
	{
		background->setSize(width, height);
	}
	this->onResize(width, height);
}

Vector2<int> GUIWidget::getSizeHint() const
{
	return Vector2<int>(box.width, box.height);
}

void GUIWidget::addWidget(std::unique_ptr<GUIWidget> widget)
{
	this->adopt(*widget);
	widgets.push_back(std::move(widget));
}

bool GUIWidget::removeWidget(GUIWidget& widget)
{
	for (std::size_t i = 0; i < widgets.size(); ++i)
	{
		if (&widget == &*widgets[i])
		{
#ifdef SL_DEBUG
			for (const std::unique_ptr<GUIWidget>& alreadyIn : deleteLater)
			{
				if (alreadyIn == widgets[i])
				{
					ERROR("Trying to double remove a Widget from its parent");
				}
			}
#endif // SL_DEBUG
			deleteLater.push_back(std::move(widgets[i]));
			widgets[i] = std::move(widgets.back());
			widgets.pop_back();
			return true;
		}
	}
	return false;
}

void GUIWidget::destroy()
{
	ASSERT(parent);
	const bool removed = parent->removeWidget(*this);
	ASSERT(removed);
}

void GUIWidget::registerMouseInputEventWithMousePos(Key key, InputEventType type, std::function<void(int, int)> function)
{
	mouseEventHandles[std::make_pair(key, type)] = std::move(function);
	captureMouseEvents();
}

void GUIWidget::registerMouseInputEvent(Key key, InputEventType type, std::function<void()> function)
{
	registerMouseInputEventWithMousePos(key, type, [f = std::move(function)](int, int) { f(); });
}

void GUIWidget::registerInputEvent(Key key, InputEventType type, std::function<void()> function)
{
	eventHandles[std::make_pair(key, type)] = std::move(function);
}

//void GUIWidget::unregisterMouseEvent(MouseEventType type, Input::MouseButton key)
//{
//	if (eventHandles[type][key])
//	{
//		//	assign to empty function
//		eventHandles[type][key] = std::function<void(int, int)>();
//	}
//}

void GUIWidget::captureMouseEvents()
{
	handlesMouseEvents = true;
}

void GUIWidget::applyBackground(const WidgetBackground& background)
{
	this->background = background.clone();
	if (this->background)
	{
		this->background->move(box.left, box.top);
		this->background->setSize(box.width, box.height);
	}
}

void GUIWidget::applyBackgroundRecursively(const WidgetBackground& background)
{
	for (auto& widget : widgets)
	{
		widget->applyBackgroundRecursively(background);
	}
	this->applyBackground(background);
}

void GUIWidget::adopt(GUIWidget& child)
{
	ASSERT(child.parent == nullptr);
	child.parent = this;
	child.setScene(this->scene);

}

void GUIWidget::setScene(AbstractScene *scene)
{
	if (this->scene != nullptr)
	{
		onSceneExit();
	}
	this->scene = scene;
	if (scene)
	{
		onSceneEnter();
	}
	for (auto& widget : widgets)
	{
		widget->setScene(scene);
	}
}

AbstractScene* GUIWidget::getScene()
{
	return scene;
}

bool GUIWidget::isFocused() const
{
	return focused;
}

void GUIWidget::setFocus(bool focus)
{
	if (focused && !focus)
	{
		onFocusLost();
		onFocusLostOrMouseLeave();
	}
	if (!focused && focus)
	{
		onFocusGained();
		onFocusGainedOrMouseEnter();
	}
	focused = focus;
	for (const auto& child : widgets)
	{
		child->setFocus(focus);
	}
}

/*			GUIWidget protected methods			*/
/*static*/void GUIWidget::handleInputEventsOn(GUIWidget& target, std::set<Key>& handled)
{
	target.handleInputEvents(handled);
}
/*static*/void GUIWidget::handleMouseInputEventsOn(GUIWidget& target, std::set<Key>& handled, const sf::Vector2f& mouse, const Rect<int>& cutoff)
{
	target.handleMouseInputEvents(handled, mouse, cutoff);
}

bool GUIWidget::isUpdating() const
{
	return updating;
}

void GUIWidget::callThisLater(std::function<void()> func)
{
	callLater.push_back(func);
}

const GUIWidget* GUIWidget::getParent() const
{
	return parent;
}

GUIWidget* GUIWidget::getParent()
{
	return parent;
}

static inline bool checkInputHelper(const Controller& controls, GUIWidget::InputEventType type, Key key)
{
	switch (type)
	{
	case GUIWidget::InputEventType::Held:
		return controls.held(key);
	case GUIWidget::InputEventType::Pressed:
		return controls.pressed(key);
	case GUIWidget::InputEventType::Released:
		return controls.released(key);
	}
	// should not hit
	return false;
}

// private
void GUIWidget::handleInputEvents(std::set<Key>& handled)
{
	for (int i = widgets.size() - 1; i >= 0; --i)
	{
		widgets[i]->handleInputEvents(handled);
	}
	onHandleInputEvents(handled);
	if (scene/* && focused*/)// TODO: figure out wtf focused even means, if it doesn't mean responds to input. If it means input, then figure out something else to do for 'selection' ie WidgetList
	{
		const Controller& controls = scene->getPlayerController();
		std::vector<Key> locallyHandledKeys;
		for (auto& eventHandler : eventHandles)
		{
			const Key key = eventHandler.first.first;
			const InputEventType type = eventHandler.first.second;
			if (checkInputHelper(controls, type, key) && handled.count(key) == 0)
			{
				isExecutingEvent = true;
				this->callThisLater(eventHandler.second);
				isExecutingEvent = false;
				// mark all button events as handled so that siblings don't handle an event too
				// since it's stupid to have a widget handle something when it's below another widget
				locallyHandledKeys.push_back(key);
			}
		}
		// but wait until we've handled all local events, or else having both a Held and Pressed event for example could cancel each other out
		handled.insert(locallyHandledKeys.begin(), locallyHandledKeys.end());
	}
}
void GUIWidget::handleMouseInputEvents(std::set<Key>& handled, const sf::Vector2f& mouse, const Rect<int>& cutoff)
{
	for (int i = widgets.size() - 1; i >= 0; --i)
	{
		widgets[i]->handleMouseInputEvents(handled, mouse, cutoff);
	}
	onHandleMouseInputEvents(handled, mouse, cutoff);

	const bool mouseIsOverNow = handled.size() == 0 && box.containsPoint(mouse.x, mouse.y) && cutoff.containsPoint(mouse.x, mouse.y);
	if (mouseWasOverLastStep)
	{
		if (!mouseIsOverNow)
		{
			onMouseLeave(mouse.x, mouse.y);
			onFocusLostOrMouseLeave();
		}
	}
	else if (mouseIsOverNow)
	{
		onMouseEnter(mouse.x, mouse.y);
		onFocusGainedOrMouseEnter();
	}
	if (mouseIsOverNow)
	{
		onMouseHover(mouse.x, mouse.y);
	}
	mouseWasOverLastStep = mouseIsOverNow;

	if (scene && mouseIsOverNow && handlesMouseEvents)
	{
		const Controller& controls = scene->getPlayerController();
		std::vector<Key> locallyHandledKeys;
		for (auto& eventHandler : mouseEventHandles)
		{
			const Key key = eventHandler.first.first;
			const InputEventType type = eventHandler.first.second;
			if (checkInputHelper(controls, type, key) && handled.size() == 0)//handled.count(key) == 0)
			{
				isExecutingEvent = true;
				this->callThisLater(std::bind(eventHandler.second, mouse.x, mouse.y));
				isExecutingEvent = false;
				// mark all button events as handled so that siblings don't handle an event too
				// since it's stupid to have a widget handle something when it's below another widget
				locallyHandledKeys.push_back(key);
			}
		}
		// but wait until we've handled all local events, or else having both a Held and Pressed event for example could cancel each other out
		handled.insert(locallyHandledKeys.begin(), locallyHandledKeys.end());
		handled.insert(Key::Accelerate); // hack
	}
}

} // gui
} // sl
