#ifndef SL_GUI_GUIWIDGET_HPP
#define SL_GUI_GUIWIDGET_HPP

#include <functional>
#include <set>
#include <string>
#include <vector>

#include "Engine/Input/Input.hpp"
#include "Engine/Input/Key.hpp"
#include "Engine/Graphics/RenderTarget.hpp"
#include "Engine/Graphics/Drawable.hpp"
#include "Engine/GUI/WidgetBackground.hpp"
#include "Utility/Memory.hpp"
#include "Utility/Rect.hpp"
#include "Utility/Vector.hpp"

namespace sl
{

class AbstractScene;

namespace gui
{

class GUIWidget
{
public:
	enum class InputEventType
	{
		Held,
		Pressed,
		Released
	};
	GUIWidget(const Rect<int>& box);

	virtual ~GUIWidget();

	/*		GUIWidget public methods		*/
	void update();

	void update(int x, int y);

	void move(int x, int y);

	void moveBy(int x, int y);

	void draw(sf::RenderTarget& target) const;

	/**
	 * Gets the depth this GUIWidget is to rendered at
	 * @return the depth - higher is drawn later (I think?)
	 */
	int getDepth() const;

	/**
	 * Sets the render depth of this GUIWidget
	 * @param depth The new depth
	 */
	void setDepth(int depth);

	/**
	 * Gets the bounds this widget has
	 * Note: does not factor in subwidgets
	 * @return The bounds
	 */
	const Rect<int>& getBox() const;

	void setSize(int width, int height);

	/**
	 * Gets a size hint for this widget's ideal size.
	 * Clients don't have to respect this, but it is advised.
	 * @return A vector specifying the ideal size of this widget.
	 */
	virtual Vector2<int> getSizeHint() const;

	/**
	 * Adds in a widget attached to this widget. Whenever this widget is drawn
	 * or updated the subwidget will be too
	 * @param widget The widget to be added (ownership is taken here)
	 */
	void addWidget(std::unique_ptr<GUIWidget>widget);

	/**
	 * Removes a widget attached to this widget if it exists.
	 * @param widget A pointer to the widget to remove
	 * @return Whether or not the widget was removed (ie whether it even existed)
	 */
	bool removeWidget(GUIWidget& widget);

	/**
	 * Removes this widget from its parent.
	 */
	void destroy();

	/**
	 * This will invoke only if the mouse is hovered over the widget.
	 * @param key Input event th act on 
	 * @param type The type (Held, Released, Pressed) of event
	 * @param function A functor wrapping the function/method/whatever (ownership is taken here)
	 */
	void registerMouseInputEventWithMousePos(Key key, InputEventType type, std::function<void(int, int)> function);
	/**
	 * Adapter for the above if you don't care about mouse coords
	 * @param key Input event th act on
	 * @param type The type (Held, Released, Pressed) of event
	 * @param function A functor wrapping the function/method/whatever (ownership is taken here)
	 */
	void registerMouseInputEvent(Key key, InputEventType type, std::function<void()> function);
	/**
	 * This will invoke regardless of if the cursor is hoered over.
	 * @param key Input event th act on
	 * @param type The type (Held, Released, Pressed) of event
	 * @param function A functor wrapping the function/method/whatever (ownership is taken here)
	 */
	void registerInputEvent(Key key, InputEventType type, std::function<void()> function);

	/**
	// * Unregisters a mouse event.
	// * @param type The type of event (held, pressed, released)
	// * @param key The mouse button
	// */
	//void unregisterInputEvent(Key type, Input::MouseButton key);

	/**
	 * Marks this Widget as capturing mouse events - this way you can't click behind it.
	 * This is automatically invoked whenever someone calls registerMouseInputEvent*()
	 */
	void captureMouseEvents();

	void applyBackground(const WidgetBackground& background);

	void applyBackgroundRecursively(const WidgetBackground& background);

	void adopt(GUIWidget& child);

	void setScene(AbstractScene *scene);

	AbstractScene* getScene();

	bool isFocused() const;

	/**
	 * Recursively sets the focus on this widget and all its children.
	 * @param focus New focus value
	 */
	void setFocus(bool focus);

	/**
	 * @return The parent of this widget
	 */
	const GUIWidget* getParent() const;
	/**
	 * @return The parent of this widget
	 */
	GUIWidget* getParent();

protected:
	/**
	 * Since you can't invoke protected on other GUIWidget subclasses.
	 * Handles both regular input events and mouse input ones, and does mouse ones first.
	 */
	static void handleInputEventsOn(GUIWidget& target, std::set<Key>& handled);
	static void handleMouseInputEventsOn(GUIWidget& target, std::set<Key>& handled, const sf::Vector2f& mouse, const Rect<int>& cutoff);
	// These are here for Scrollbar mostly right now. Override these if you have sub-widgets that aren't inside widgets.
	virtual void onHandleInputEvents(std::set<Key>& handled) {}
	virtual void onHandleMouseInputEvents(std::set<Key>& handled, const sf::Vector2f& mouse, const Rect<int>& cutoff) {}
	
	
	virtual void onDraw(sf::RenderTarget& target) const {}
	virtual void beforeDraw(sf::RenderTarget& target) const {}
	virtual void onUpdate(int x, int y) {}
	virtual void onResize(int width, int height) {}

	virtual void onMouseEnter(int x, int y) {}
	virtual void onMouseLeave(int x, int y) {}
	virtual void onMouseHover(int x, int y) {}

	virtual void onFocusGained() {}
	virtual void onFocusLost() {}

	virtual void onFocusGainedOrMouseEnter() {}
	virtual void onFocusLostOrMouseLeave() {}

	virtual void onSceneEnter() {}
	virtual void onSceneExit() {}
	/**
	 * To contain code
	 * If the subclass uses a RenderTarget or anything that would need updating when
	 * this widget changes backgrounds then implement this method in that class, as it
	 * is called when a background is changed.
	 */
	virtual void onBackgroundChange() {}




	bool isUpdating() const;
	/**
	 * Anything here will get called when this Widget finishes updating
	 */
	void callThisLater(std::function<void()> func);


	Rect<int> box;
	AbstractScene *scene;

private:

	void handleInputEvents(std::set<Key>& handled);
	void handleMouseInputEvents(std::set<Key>& handled, const sf::Vector2f& mouse, const Rect<int>& cutoff);
	


	GUIWidget *parent;
	std::vector<std::unique_ptr<GUIWidget>> widgets;
	bool mouseWasOverLastStep;
	bool handlesMouseEvents;
	std::map<std::pair<Key, InputEventType>, std::function<void()>> eventHandles;
	std::map<std::pair<Key, InputEventType>, std::function<void(int, int)>> mouseEventHandles;
	int depth;
	std::unique_ptr<WidgetBackground> background;
	Vector2<int> toMove;
	bool isExecutingEvent;
	std::vector<std::function<void()>> callLater;
	bool updating;
	std::vector<std::unique_ptr<GUIWidget>> deleteLater;
	bool focused;
};

} // gui
} // sl

#endif // SL_GUI_GUIWIDGET_HPP
