#include "Engine/GUI/Scrollbar.hpp"

#include "Engine/Game.hpp"
#include "Engine/AbstractScene.hpp"
#include "Engine/Settings.hpp"
#include "Engine/GUI/WidgetGrid.hpp"
#include "Utility/Assert.hpp"
#include "Utility/Logger.hpp"


namespace sl
{
namespace gui
{

Scrollbar::Scrollbar(const Rect<int>&box, std::unique_ptr<GUIWidget> scrollWidget)
	:GUIWidget(box)
	,scrollWidget()
	,scrollbarWidth(16)
	,scrollbarHeight(0)
	,widthbarScroll(0)
	,heightbarScroll(0)
	,scrollingWidth(false)
	,scrollingHeight(false)
	,clickYPos(0)
	,renderTextureNeedsUnfucking(true)
{
	renderTexture.create(box.width, box.height);
	renderSprite.setTexture(renderTexture.getTexture());
	if (this->scrollWidget)
	{
		this->adopt(*this->scrollWidget);
	}
	scrollbar.setFillColor(sf::Color::Transparent);
	scrollbar.setOutlineColor(sf::Color::Transparent);//Green);
	scrollbar.setOutlineThickness(1);
	for (int i = 0; i < 100; ++i)
	{
		buf1[i] = buf2[i] = 0;
	}

	setScrollArea(std::move(scrollWidget));
}

void Scrollbar::setScrollArea(std::unique_ptr<GUIWidget> widget)
{
	if (this->scrollWidget)
	{
		//	it shouldn't be in the GUIWidget's list of parents anyway (wtf)
		ASSERT(!this->removeWidget(*scrollWidget));
	}
	scrollWidget = std::move(widget);
	if (scrollWidget)
	{
		this->adopt(*scrollWidget);
		if (isFocused() != scrollWidget->isFocused())
		{
			scrollWidget->setFocus(isFocused());
		}
	}
}

void Scrollbar::setVerticalScroll(float scroll)
{
	ASSERT(scroll >= 0.f && scroll <= 1.f);
	heightbarScroll = scroll;
}

float Scrollbar::getVerticalScroll() const
{
	return heightbarScroll;
}

int Scrollbar::getScrollbarWidth() const
{
	return scrollbarWidth;
}

Rect<int> Scrollbar::getVisibleRegion() const
{
	return Rect<int>(
		0,
		(scrollWidget->getBox().height - box.height) * heightbarScroll,
		box.width - scrollbarWidth,
		box.height - scrollbarHeight);
}

void Scrollbar::setVisiblePosition(Vector2i topLeft)
{
	ASSERT(topLeft.x == 0);
	const int denom = scrollWidget->getBox().height - box.height;
	const float scroll = denom <= 0 ? 0.f : topLeft.y / (float)denom;
	setVerticalScroll(scroll);
}



/*	  GUIWidget private methods	   */
void Scrollbar::onDraw(sf::RenderTarget& target) const
{
	renderTexture.clear(sf::Color::Transparent);
	if (scrollWidget)
	{
		scrollWidget->draw(renderTexture);
	}
	renderTexture.draw(scrollbar);
	renderTexture.display();
	target.draw(renderSprite);
	//renderTexture.display();

	// This needed to go AFTER the stuff above here... interesting. No idea why.
	// It wasn't working when I put it at the start of the function.
	//if (renderTextureNeedsUnfucking)
	//{
	//	unfuckTextureFlipping();
	//	renderTextureNeedsUnfucking = false;
	//}
}

void Scrollbar::onUpdate(int x, int y)
{
	renderSprite.setPosition(x, y);
	box.left = x;
	box.top = y;
	if (scene && scrollWidget)
	{
		int xdif = x - box.left;
		int ydif = y - box.top;

		
		scrollWidget->update(0, 0 - (scrollWidget->getBox().height - box.height) * heightbarScroll);
		if (!scene) return;// TODO: figure out why this is necessary (caused in the towngen setting menus when you change ethnicity and it deletes the old widget)
		const sf::Vector2f mouse = scene->getGame()->getRawInput().mousePosF() - sf::Vector2f(box.left, box.top);

		//	scrolling logic here
		const bool canScroll = scrollWidget->getBox().height > box.height;
		const int barsize = canScroll ?
						(box.height - scrollbarHeight) * (box.height - scrollbarHeight) / scrollWidget->getBox().height
						: box.height;
		const Rect<int> vScrollBox(box.width - scrollbarWidth, heightbarScroll * (box.height - scrollbarHeight - barsize), scrollbarWidth, barsize);

		if (canScroll)
		{
			if (scene->getPlayerController().released(Key::GUIClick))
			{
				scrollingWidth = false;
				scrollingHeight = false;
			}

			if (scrollingHeight)
			{
				heightbarScroll = 1 - ((box.height - scrollbarHeight - barsize) - (mouse.y - clickYPos)) / (box.height - scrollbarHeight - barsize);
				if (heightbarScroll < 0)
				{
					heightbarScroll = 0;
				}
				else if (heightbarScroll > 1)
				{
					heightbarScroll = 1;
				}
				Logger::log(Logger::GUI, Logger::Debug) << "heighbarScroll = " << heightbarScroll << " | mouse.y = " << mouse.y << " | clickYPos = " << clickYPos << std::endl;
				scrollbar.setOutlineColor(sf::Color::White);
				scrollbar.setFillColor(sf::Color(64, 64, 64));
			}
			else
			{
				if (vScrollBox.containsPoint(mouse.x, mouse.y))
				{
					if (scene->getPlayerController().pressed(Key::GUIClick))
					{
						clickYPos = mouse.y - vScrollBox.top;
						scrollingHeight = true;
					}
					scrollbar.setOutlineColor(sf::Color::White);
					scrollbar.setFillColor(sf::Color(32, 32, 32));
				}
				else
				{
					scrollbar.setOutlineColor(sf::Color(192, 192, 192));
					scrollbar.setFillColor(sf::Color::Black);
				}
			}
		}
		else
		{
			heightbarScroll = 0;
		}
		scrollbar.setPosition(vScrollBox.left + 1, vScrollBox.top + 1);
		scrollbar.setSize(sf::Vector2f(vScrollBox.width - 2, vScrollBox.height));
	}
}

void Scrollbar::onSceneEnter()
{
	//renderTextureNeedsUnfucking = true;
	unfuckTextureFlipping();
	if (scrollWidget)
	{
		scrollWidget->setScene(scene);
	}
}

void Scrollbar::onSceneExit()
{
	if (scrollWidget)
	{
		scrollWidget->setScene(nullptr);
	}
}

void Scrollbar::onFocusGained()
{
	if (scrollWidget)
	{
		scrollWidget->setFocus(true);
	}
}

void Scrollbar::onFocusLost()
{
	if (scrollWidget)
	{
		scrollWidget->setFocus(false);
	}
}

void Scrollbar::onHandleInputEvents(std::set<Key>& handled)
{
	if (scrollWidget)
	{
		handleInputEventsOn(*scrollWidget, handled);
	}
}

void Scrollbar::onHandleMouseInputEvents(std::set<Key>& handled, const sf::Vector2f& mouse, const Rect<int>& cutoff)
{
	if (scrollWidget)
	{
		const sf::Vector2f mouseInArea = mouse - sf::Vector2f(box.left, box.top);
		const Rect<int> innerbox(box.width - scrollbarWidth, box.height - scrollbarHeight);
		handleMouseInputEventsOn(*scrollWidget, handled, mouseInArea, innerbox);
	}
}

void Scrollbar::unfuckTextureFlipping() const
{
	// for SOME fucking reason the texture flipping doesn't always work properly
	// if you have never previous called RenderTexture::display() on it before
	// so we must do that here to figure out if we need to unfuck it
	renderTexture.display();
	// test whether or not textures will be flipped
	sf::RectangleShape redRect(sf::Vector2f(renderTexture.getSize().x, renderTexture.getSize().y / 2));
	redRect.setFillColor(sf::Color::Red);
	renderTexture.clear(sf::Color::Green);
	renderTexture.draw(redRect);
	sf::Image testResultImage = renderTexture.getTexture().copyToImage();
	const bool testFailed = testResultImage.getPixel(0, 0) != sf::Color::Red;
	if (testFailed)
	{
		renderSprite.setOrigin(0, renderTexture.getSize().y);
		renderSprite.setScale(1.f, -1.f);
	}
	else
	{
		renderSprite.setOrigin(0, 0);
		renderSprite.setScale(1.f, 1.f);
	}
}

} // gui
} // sl
