#include "Engine/GUI/TextLabel.hpp"

#include "Engine/Graphics/SFTextBox.hpp"
#include "Engine/Graphics/SpriteText.hpp"
#include "Engine/GUI/FontManager.hpp"
#include "Utility/Assert.hpp"
#include "Utility/Memory.hpp"

namespace sl
{
namespace gui
{

TextLabel::TextLabel(const Rect<int>& box, const Config& config)
	:TextLabel(box, "", config)
{
}

TextLabel::TextLabel(const Rect<int>& box, const std::string &text)
	:TextLabel(box, text, Config())
{
}
	
TextLabel::TextLabel(const Rect<int>& box)
	:TextLabel(box, "", Config())
{
}

TextLabel::TextLabel(const Rect<int>& box, const std::string& text, const Config& config)
	:GUIWidget(box)
	,text()
	,hPolicy(config.hPolicy)
	,vPolicy(config.vPolicy)
	,writePolicy(config.writePolicy)
	,writeSpeed(config.speed)
	,writtenIndex(0.f)
	,colSize(config.colSize)
	,headerSize(config.headerSize)
	,callFinalize(config.callFinalize)
{
	const int taWidth = box.width - 2 * colSize;
	const int taHeight = box.height - 2 * headerSize;
	const Rect<int> textArea(box.left + colSize, box.top + headerSize, taWidth > 0 ? taWidth : 1, taHeight > 0 ? taHeight : 1);
	switch (config.fontType)
	{
	case FontType::Regular:
		textBox = unique<SFTextBox>(textArea);
		textBox->setFont("resources/arial.ttf");
		textBox->setFontSize(config.fontScale);
		break;
	case FontType::Sprite:
		textBox = unique<SpriteText>("default", textArea, 8, config.fontScale);
		break;
	}

	setString(text);
	refresh();
	writeAll();
}

void TextLabel::setFont(const std::string& fontname)
{
	textBox->setFont(fontname);
}

void TextLabel::setFontSize(int newSize)
{
	textBox->setFontSize(newSize);
}

void TextLabel::setString(const std::string& text)
{
	this->text = textBox->insertNewlines(text);
	if (writePolicy == Instant)
	{
		textBox->setString(text);
	}
	else
	{
		textBox->setString("");
		writtenIndex = 0.f;
	}
}

const std::string& TextLabel::getString() const
{
	return text;
}

void TextLabel::setColour(const sf::Color& colour)
{
	textBox->setFontColor(colour);
}

void TextLabel::write()
{
	const int before = static_cast<int>(writtenIndex);
	writtenIndex += writeSpeed;
	const int after = static_cast<int>(writtenIndex);

	std::string toAppend;
	for (int i = before; i < after && i < text.size(); ++i)
	{
		toAppend += text[i];
	}
	if (!toAppend.empty())
	{
		textBox->appendString(toAppend);
	}
	if (!text.empty() && callFinalize && isDoneWriting())
	{
		textBox->finalize();
	}
}

void TextLabel::writeAll()
{
	writtenIndex = text.size() - 1;
	textBox->setString(text);
	if (!text.empty() && callFinalize)
	{
		textBox->finalize();
	}
}

bool TextLabel::isDoneWriting() const
{
	return writtenIndex >= text.size();
}

/*		GUIWidget private methods		*/
void TextLabel::onDraw(sf::RenderTarget& target) const
{
	textBox->draw(target);
	//target.draw(dbg);
}

void TextLabel::onUpdate(int x, int y)
{
	sf::Vector2f offset(x - box.left, y - box.top);
	//textBox->update(textBox->getCurrentBounds().left + offset.x, textBox->getCurrentBounds().top + offset.y);
	if (writePolicy == TextWritePolicy::Letter && text.size() > static_cast<int>(writtenIndex))
	{
		write();
	}
	const auto& cb = textBox->getCurrentBounds();
	dbg.setSize(sf::Vector2f(cb.width, cb.height));
	dbg.setPosition(cb.left, cb.top);
	dbg.setFillColor(sf::Color::Transparent);
	dbg.setOutlineColor(sf::Color::Red);
	dbg.setOutlineThickness(1);
	refresh();
}

/*		TextLabel private methods		*/
void TextLabel::refresh()
{
	//	Now we need to position the text in the box with respect to the
	//	horizontal/vertical spacing policies as well as the header/column sizes
	int xOffset = 0;
	int yOffset = 0;
	const int textWidth = textBox->getCurrentBounds().width;
	const int textHeight = textBox->getCurrentBounds().height;
	const int width = box.width - 2 * colSize;
	const int height = box.height - 2 * headerSize;

	if (hPolicy == Center)
	{
		xOffset = (width - textWidth) / 2;
	}
	else if (hPolicy == Right)
	{
		xOffset = (width - textWidth);
	}
	else
	{
		ASSERT(hPolicy == Left && "Invalid horizontal text policy");
	}

	if (vPolicy == Middle)
	{
		yOffset = (height - textHeight) / 2;
	}
	else if (vPolicy == Bottom)
	{
		yOffset = (height - textHeight);
	}
	else
	{
		ASSERT(vPolicy == Top && "Invalid vertical text policy");
	}

	textBox->update(box.left + colSize + xOffset, box.top + headerSize + yOffset);
}

void TextLabel::extendVerticallyToFitText()
{
	int requiredHeight = headerSize * 2 + textBox->getMinimumHeight();
	//if (requiredHeight > box.height)
	{
		setSize(box.width, requiredHeight);
		textBox->setArea(box.width, requiredHeight);
	}
}

void TextLabel::extendVerticallyToFitText(int minSize)
{
	int requiredHeight = std::max(minSize, headerSize * 2 + textBox->getMinimumHeight());
	//if (requiredHeight > box.height)
	{
		setSize(box.width, requiredHeight);
		textBox->setArea(box.width, requiredHeight);
	}
}

void TextLabel::shrinkHorizontally()
{
	int requiredWidth = colSize * 2 + textBox->getCurrentBounds().width;
	if (requiredWidth < box.width)
	{
		setSize(requiredWidth, box.height);
		textBox->setArea(requiredWidth, box.height);
	}
}

void TextLabel::setWritePolicy(TextWritePolicy writePolicy, float speed)
{
	this->writePolicy = writePolicy;
	writeSpeed = speed;
	writtenIndex = 0.f;
	textBox->setString("");
}

} // gui
} // sl
