#ifndef SL_GUI_TEXTLABEL_HPP
#define SL_GUI_TEXTLABEL_HPP

#include <string>

#include "Engine/Graphics/AbstractTextBox.hpp"
#include "Engine/GUI/GUIWidget.hpp"
#include "Utility/Memory.hpp"

namespace sl
{
namespace gui
{

class TextLabel : public GUIWidget
{
public:
	enum HorizontalPolicy
	{
		Left   = 0,
		Center = 1,
		Right  = 2
	};

	enum VerticalPolicy
	{
		Top    = 3, // starts at 3 so we can differentiate HorizontalPolicy from VerticalPolicy in case people mix up Center/Middle
		Middle = 4,
		Bottom = 5
	};

	enum TextWritePolicy
	{
		Instant,
		Word,
		Letter
	};

	enum class FontType
	{
		Regular,
		Sprite
	};

	//! Builder for the parameters for constructor - used since we (possibly) need to finalize the text after the constructor for optimization purposes
	class Config
	{
	public:
		Config& halign(HorizontalPolicy hPolicy)
		{
			this->hPolicy = hPolicy;
			return *this;
		}
		Config& valign(VerticalPolicy vPolicy)
		{
			this->vPolicy = vPolicy;
			return *this;
		}
		Config&	centered()
		{
			hPolicy = Center;
			vPolicy = Middle;
			return *this;
		}
		Config& padx(int colSize)
		{
			this->colSize = colSize;
			return *this;
		}
		Config& pady(int headerSize)
		{
			this->headerSize = headerSize;
			return *this;
		}
		Config& pad(int colSize, int headerSize)
		{
			this->colSize = colSize;
			this->headerSize = headerSize;
			return *this;
		}
		Config& setFontType(FontType fontType)
		{
			this->fontType = fontType;
			return *this;
		}
		Config& fontSize(int size)
		{
			fontScale = size;
			return *this;
		}
		Config& modifedOften()
		{
			callFinalize = false;
			return *this;
		}
		Config& setWritePolicy(TextWritePolicy writePolicy, float speed)
		{
			this->writePolicy = writePolicy;
			this->speed = speed;
			return *this;
		}
	private:
		//! Whether the text aligns to the left, center or right of the widget
		HorizontalPolicy hPolicy = Left;
		//! Whether the text aligns to the top, middle or bottom of the widget
		VerticalPolicy vPolicy = Top;
		//! The minimum amount of space to keep on each side of the text (horizontally)
		int colSize = 0;
		//! The minimum amount of space to keep on the top or bottom of the text (will be violated if text string is too long obviously)
		int headerSize = 0;
		//! The type of font to use
		FontType fontType = FontType::Sprite;
		//! font size
		int fontScale = 1;
		//! Whether to call finalize() on the SpriteText (only disable if you're going to be modifying this text a lot, ie every frame or so)
		bool callFinalize = true;
		//! Whether text immediately shows up, or gets written out letter at a time/etc
		TextWritePolicy writePolicy = Instant;
		//! How many characters(or words) per update to append (should only be set of writePolicy is set to Word or Letter)
		float speed = 0.f;

		friend class TextLabel;
	};

	TextLabel(const Rect<int>& box, const std::string& text, const Config& config);

	TextLabel(const Rect<int>& box, const Config& config);

	TextLabel(const Rect<int>& box, const std::string &text);
	
	TextLabel(const Rect<int>& box);
	
	/**
	 * Sets the font to be used (gets from FontManager)
	 * @param fontname The filename of the font
	 */
	void setFont(const std::string& fontname);
	/**
	 * Sets the font size of the font
	 * @param newSize The size of the font
	 */
	void setFontSize(int newSize);
	/**
	 * Sets the string this label should display
	 * @param text The new string to display
	 */
	void setString(const std::string& text);

	/**
	 * Gets the string that is currently being displayed
	 * @return A reference to the string
	 */
	const std::string& getString() const;

	/**
	 * Sets the colour of the text
	 * @colour The new colour to use
	 */
	void setColour(const sf::Color& colour);


	/**
	 * Extends (or shrinks) the widget to exactly fit the text vertically
	 */
	void extendVerticallyToFitText();

	void extendVerticallyToFitText(int minSize);

	/**
	 * Shrinks the widget horizontally to fit if possible (widget is 1 line and text is smaller than width)
	 */
	void shrinkHorizontally();

	/**
	 * Sets the write policy and speed at which things are written out.
	 * @param writePolicy Whether text immediately shows up, or gets written out letter at a time/etc
	 * @param speed How many units per frame get drawn
	 */
	void setWritePolicy(TextWritePolicy writePolicy, float speed);

	/**
	 * Advances one unit of text. Automatically called once per frame.
	 */
	void write();

	/**
	 * Writes all text. Effectively just makes writtenText = text
	 */
	void writeAll();

	/**
	 * @return Whether or not all text has been written yet
	 */
	bool isDoneWriting() const;


private:
	/*		GUIWidget private methods		*/
	void onDraw(sf::RenderTarget& target) const override;

	void onUpdate(int x, int y) override;



	std::string text;

	HorizontalPolicy hPolicy;

	VerticalPolicy vPolicy;

	TextWritePolicy writePolicy;

	float writeSpeed;

	float writtenIndex;

	int colSize;

	int headerSize;

	void refresh();

	std::unique_ptr<AbstractTextBox> textBox;

	sf::RectangleShape dbg;

	bool callFinalize;
};

} // gui
} // sl

#endif // SL_GUI_TEXTLABEL_HPP
