#include "Engine/GUI/WidgetGrid.hpp"

#include "Utility/Assert.hpp"
#include "Utility/Logger.hpp"
#include <cstdlib>

namespace sl
{
namespace gui
{

static const int emptyCellSize = 16;

WidgetGrid::WidgetGrid(int x, int y, int columns, int rows)
	:GUIWidget(Rect<int>(x, y, emptyCellSize, emptyCellSize))
	,grid()
	,colWidths(columns, emptyCellSize)
	,rowHeights(rows, emptyCellSize)
	,colWidthSoFar(columns)
	,rowHeightSoFar(rows)
	,columns(columns)
	,rows(rows)
	,unfuckSelfCountdown(0)
{
	ASSERT(columns > 0 && rows > 0);
	for (int i = 0; i < columns; ++i)
	{
		grid.push_back(std::vector<GUIWidget*>(rows, nullptr));
		colWidthSoFar[i] = emptyCellSize * i;
	}
	for (int i = 0; i < rows; ++i)
	{
		rowHeightSoFar[i] = emptyCellSize * i;
	}
	this->updateSize();
}

void WidgetGrid::setWidgetHelper(int x, int y, GUIWidget *widget)
{
	this->setWidget(x, y, std::unique_ptr<GUIWidget>(widget));
}
/*			WidgetGrid public methods		*/
void WidgetGrid::setWidget(int x, int y, std::unique_ptr<GUIWidget> widget)
{
	if (this->isUpdating())
	{
		this->callThisLater(std::bind(&WidgetGrid::setWidgetHelper, this, x, y, widget.release()));
	}
	else
	{
		ASSERT(x >= 0 && x < columns && y >= 0 && y < rows);
		//	the width and height differences between the old widget (or lack thereof)
		//	and the new widget (of lack thereof) occupying cell (x, y)
		const int wdif = (widget ? widget->getBox().width : emptyCellSize) - colWidths[x];

		const int hdif = (widget ? widget->getBox().height : emptyCellSize) - rowHeights[y];

		const bool positionsNeedUpdating = (wdif || hdif);

		if (grid[x][y])
		{
			this->removeWidget(*grid[x][y]);
		}
		grid[x][y] = widget.get();	//	no ownership assumes by grid!
		if (widget)
		{
			this->addWidget(std::move(widget));
		}

		if (wdif)
		{
			colWidths[x] += wdif;
		}

		if (hdif)
		{
			rowHeights[y] += hdif;
		}

		for (int i = x + 1; i < columns; ++i)
		{
			colWidthSoFar[i] += wdif;
		}
		for (int i = y + 1; i < rows; ++i)
		{
			rowHeightSoFar[i] += hdif;
		}

		if (positionsNeedUpdating)
		{
			this->repositionEntities(x, y);
		}
		this->updateSize();
	}
}

void WidgetGrid::setRows(const int rows)
{
	ASSERT(rows >= 0);
	if (this->isUpdating())
	{
		this->callThisLater(std::bind(&WidgetGrid::setRows, this, rows));
	}
	else
	{
		if (rows > this->rows)
		{
			for (auto& col : grid)
			{
				col.resize(rows, nullptr);
			}
			rowHeights.resize(rows, emptyCellSize);
			for (int i = this->rows; i < rows; ++i)
			{
				if (rowHeightSoFar.empty())
				{
					rowHeightSoFar.push_back(0);
				}
				else
				{
					rowHeightSoFar.push_back(rowHeightSoFar.back() + emptyCellSize);
				}
			}
		}
		else if (rows < this->rows)
		{
			for (auto& col : grid)
			{
				for (int i = 0; i < this->rows - rows; ++i)
				{
					if (col.back())
					{
						this->removeWidget(*col.back());
					}
					col.pop_back();
				}
			}
			rowHeights.resize(rows);
			rowHeightSoFar.resize(rows);
		}
		this->rows = rows;
		this->updateSize();
		this->repositionEntities(0, 0);

		ASSERT(rowHeights.size() == rowHeightSoFar.size() && rowHeightSoFar.size() == rows);
		for (const auto& row : grid)
		{
			ASSERT(row.size() == rows);
		}
	}
}

void WidgetGrid::setCols(int cols)
{
	ASSERT (cols >= 0);
	if (this->isUpdating())
	{
		this->callThisLater(std::bind(&WidgetGrid::setCols, this, cols));
	}
	else
	{
		if (cols > columns)
		{
			for (int i = columns; i < cols; ++i)
			{
				grid.push_back(std::vector<GUIWidget*>(rows, nullptr));
				colWidths.push_back(emptyCellSize);
				if (colWidthSoFar.empty())
				{
					colWidthSoFar.push_back(0);
				}
				else
				{
					colWidthSoFar.push_back(colWidthSoFar.back() + emptyCellSize);
				}
			}
			columns = cols;
		}
		else if (cols < columns)
		{
			while (columns-- > cols)
			{
				for (GUIWidget *widget : grid.back())
				{
					this->removeWidget(*widget);
				}
				grid.pop_back();
				colWidths.pop_back();
				colWidthSoFar.pop_back();
			}
		}
		this->updateSize();
		this->repositionEntities(0, 0);

		ASSERT(colWidths.size() == colWidthSoFar.size() && colWidthSoFar.size() == cols);
		ASSERT(grid.size() == rows);
		ASSERT(columns == cols);
	}
}


int WidgetGrid::getRows() const
{
	return rows;
}

int WidgetGrid::getColumns() const
{
	return columns;
}

const GUIWidget* WidgetGrid::at(int x, int y) const
{
	ASSERT(x >= 0 && x < columns && y >= 0 && y < rows);
	return grid[x][y];
}

GUIWidget* WidgetGrid::at(int x, int y)
{
	ASSERT(x >= 0 && x < columns && y >= 0 && y < rows);
	return grid[x][y];
}

/*			priavate methods		*/
void WidgetGrid::repositionEntities(int xStart, int yStart)
{
	for (int x = xStart; x < grid.size(); ++x)
	{
		for (int y = yStart; y < grid[x].size(); ++y)
		{
			if (grid[x][y])
			{
				Logger::log(Logger::GUI, Logger::Debug) << "moving button to " << colWidthSoFar[x] << ", " << rowHeightSoFar[y] << std::endl;
				Logger::log(Logger::GUI, Logger::Debug) << "from " << grid[x][y]->getBox().left << ", " << grid[x][y]->getBox().top << std::endl;
				grid[x][y]->move(colWidthSoFar[x], rowHeightSoFar[y]);
			}
		}
	}
	unfuckSelfCountdown = 2;
}

void WidgetGrid::updateSize()
{
	Logger::log(Logger::GUI, Logger::Debug) << "before: " << box.width << " x " << box.height << std::endl;
	this->box.width = columns ? colWidthSoFar.back() + colWidths.back() : 1;
	this->box.height = rows ? rowHeightSoFar.back() + rowHeights.back() : 1;
	Logger::log(Logger::GUI, Logger::Debug) << "now: " << box.width << " x " << box.height << std::endl;
}

void WidgetGrid::onUpdate(int x, int y)
{
	if (unfuckSelfCountdown > 0 && --unfuckSelfCountdown == 0)
	{
		for (int i = 0; i < grid.size(); ++i)
		{
			for (int j = 0; j < grid[i].size(); ++j)
			{
				if (grid[i][j])
				{
					grid[i][j]->move(colWidthSoFar[i] + x, rowHeightSoFar[j] + y);
				}
			}
		}
	}
}

} // gui
} // sl
