#ifndef SL_GUI_WIDGETGRID_HPP
#define SL_GUI_WIDGETGRID_HPP

#include <functional>
#include "Engine/GUI/GUIWidget.hpp"
#include "Utility/Memory.hpp"

namespace sl
{
namespace gui
{

class WidgetGrid : public GUIWidget
{
public:
	/**
	 * Constructs a matrix/grid of GUIWidgets
	 * @param box The area the grid takes up
	 * @param columns How many widgets across the grid is
	 * @param rows How many widgets high the gird is
	 */
	WidgetGrid(int x, int y, int columns, int rows);
	
	/**
	 * Places a GUIWidget into the grid
	 * @param x The column to place the widget in
	 * @param y The row to place the widget in
	 * @param widget The GUIWidget to insert
	 */
	void setWidget(int x, int y, std::unique_ptr<GUIWidget> widget);
	
	/**
	 * Sets how many rows are in the grid.
	 * If this shrinks the grid then it will delete any cut off widgets
	 * @param rows The new number of rows
	 */
	void setRows(const int rows);
	
	/**
	 * Sets how many columns are in the grid.
	 * If this shrinks the gird then it will delete any cut off widgets
	 * @param cols The new number of columns
	 */
	void setCols(int cols);

	/**
	 * Changes the width of a certain column.
	 * Note: resizes entire grid's overall size potentially.
	 * @param colIndex The column to resize
	 * @param newWidth The new width to use
	 */
	void setWidthAtCol(int colIndex, int newWidth);

	/**
	 * Changes the height of a certain row.
	 * Note: resizes entire grid's overall size potentially.
	 * @param rowIndex The row to resize
	 * @param newHeight The new height to use
	 */
	void setHeightAtRow(int rowIndex, int newHeight);

	int getRows() const;

	int getColumns() const;

	const GUIWidget* at(int x, int y) const;

	GUIWidget* at(int x, int y);
	
private:
	void repositionEntities(int xStart = 0, int yStart = 0);

	void updateSize();

	void onUpdate(int x, int y) override;
	/*
	 * Helper function called only by setWidget() for deferred calls
	 * This exists because std::unique_ptr is non-copyable and any resulting 
	 * std::function would be too. No memory can be leaked here though since
	 * all callThisLaters are called upon GUIWidget's destructor anyway.
	 * @param x X position of widget
	 * @param y Y position of widget
	 * @param widget The Widget itself
	 */
	void setWidgetHelper(int x, int y, GUIWidget *widget);


	//!A 2D array/matrix of all the pointers to GUIWidgets
	std::vector<std::vector<GUIWidget*>> grid;
	//!A lookup table of the widths for a given column
	std::vector<int> colWidths;
	//!A lookup table of the heights for a given row
	std::vector<int> rowHeights;
	std::vector<int> colWidthSoFar;
	std::vector<int> rowHeightSoFar;
	//!The number of columns
	int columns;
	//!The number of rows
	int rows;
	int unfuckSelfCountdown;
};

} // gui
} // sl

#endif // SL_GUI_WIDGETGRID_HPP