#ifndef SL_GUI_WIDGETHANDLE_HPP
#define SL_GUI_WIDGETHANDLE_HPP

#include "Engine/AbstractScene.hpp"
#include "Engine/GUI/GUIWidget.hpp"
#include "Utility/Assert.hpp"
#include "Utility/Memory.hpp"

namespace sl
{
namespace gui
{

template <typename WidgetType = GUIWidget>
class WidgetHandle
{
public:
	/**
	 * Creates a new handle
	 */
	WidgetHandle();
	/**
	 * Destructor... unregisters/frees/etc needed stuff
	 */
	~WidgetHandle();

	/**
	 * Copy constructor - transfers ownership away from copy to this
	 */
	WidgetHandle(WidgetHandle<WidgetType>& copy);

	/**
	 * Sets the widget this handle wraps
	 * takes care of unregistering/freeing the previous one
	 * @param widget The new widget for this to represent
	 */
	void setHandle(std::unique_ptr<WidgetType> widget);

	/**
	 * Sets the AbstractScene to have the Widget be registered to
	 * @param scene The scene to register to
	 */
	void setScene(AbstractScene *scene);

	/**
	 * Releases the widget (unregisters from scene + destroys)
	 */
	void releaseHandle();

	/**
	 * Enables the widget (will register with scene if unregistered)
	 */
	void enable();

	/**
	 * Disables the widget (will unregister with scene if registered)
	 */
	void disable();
	/**
	 * Returns whether or not the handle is enabled)
	 * @return True if enabled, False if not enabled
	 */
	bool isEnabled() const;
	/**
	 * Tests whether or not a widget is valid (non-null)
	 */
	explicit operator bool() const;

	/**
	 * Dereferences the interal widget - throws something if null
	 * @return reference to the internal widget.
	 */
	WidgetType& operator*();

	/**
	 * Derefernces the internal widget - throws something if null
	 * @return const reference to the internal widget
	 */
	const WidgetType& operator*() const;

	/**
	 * Dereferences the internal widget for member accessing
	 */
	WidgetType* operator->();
	/**
	 * Dereferences the internal widget for member accessing
	 */
	const WidgetType* operator->() const;

	std::unique_ptr<WidgetType> take();

private:
	//!The widget this handle wraps
	std::unique_ptr<WidgetType> widget;
	//!The scene this handle registers/unregisters widgets in
	AbstractScene *scene;
	//!Whether or not this widget is enabled (registered in the current scene)
	bool enabled;
};


/*			WidgetHandle<T> public method definitions		*/
template <typename WidgetType>
WidgetHandle<WidgetType>::WidgetHandle()
	:widget(nullptr)
	,scene(nullptr)
	,enabled(false)
{
}

template <typename WidgetType>
WidgetHandle<WidgetType>::WidgetHandle(WidgetHandle<WidgetType>& copy)
	:widget(copy.widget.release())
	,scene(copy.scene)
	,enabled(copy.enabled)
{
	copy.scene = nullptr;
	copy.enabled = false;
}


template <typename WidgetType>
WidgetHandle<WidgetType>::~WidgetHandle()
{
	this->releaseHandle();
}


template <typename WidgetType>
void WidgetHandle<WidgetType>::setHandle(std::unique_ptr<WidgetType> widget)
{
	ASSERT(widget != nullptr);
	if (this->widget == widget)	//	why
	{
		return;
	}
	if (this->widget)
	{
		if (enabled)
		{
			scene->unregisterGUIWidget(this->widget.get());
		}
	}
	this->widget = std::move(widget);
	if (enabled)
	{
		scene->registerGUIWidget(this->widget.get());
	}
}

template <typename WidgetType>
void WidgetHandle<WidgetType>::setScene(AbstractScene *scene)
{
	if (this->scene != scene)
	{
		if (widget && enabled)
		{
			this->scene->unregisterGUIWidget(widget.get());
		}
		this->scene = scene;
		if (scene)
		{
			if (enabled && scene)
			{
				scene->registerGUIWidget(widget.get());
			}
		}
		else
		{
			enabled = false;
		}
	}
}


template <typename WidgetType>
void WidgetHandle<WidgetType>::releaseHandle()
{
	if (widget)
	{
		if (enabled)
		{
			scene->unregisterGUIWidget(widget.get());
		}
		widget.reset();
	}
}


template <typename WidgetType>
void WidgetHandle<WidgetType>::enable()
{
	ASSERT(scene != nullptr);
	if (!enabled)
	{
		scene->registerGUIWidget(widget.get());
		enabled = true;
	}
}


template <typename WidgetType>
void WidgetHandle<WidgetType>::disable()
{
	if (enabled)
	{
		scene->unregisterGUIWidget(widget.get());
		enabled = false;
	}
}

template <typename WidgetType>
bool WidgetHandle<WidgetType>::isEnabled() const
{
	return enabled;
}

template <typename WidgetType>
WidgetHandle<WidgetType>::operator bool() const
{
	return widget != nullptr;
}


template <typename WidgetType>
WidgetType& WidgetHandle<WidgetType>::operator*()
{
	return *widget;
}


template <typename WidgetType>
const WidgetType& WidgetHandle<WidgetType>::operator*() const
{
	return *widget;
}


template <typename WidgetType>
WidgetType* WidgetHandle<WidgetType>::operator->()
{
	return widget.get();
}


template <typename WidgetType>
const WidgetType* WidgetHandle<WidgetType>::operator->() const
{
	return widget.get();
}

template <typename WidgetType>
std::unique_ptr<WidgetType> WidgetHandle<WidgetType>::take()
{
	if (widget && enabled)
	{
		scene->unregisterGUIWidget(widget.get());
	}

	std::unique_ptr<WidgetType> ret = std::move(widget);

	widget.reset(nullptr);

	enabled = false;

	return ret;
}

} // gui
} // sl

#endif // SL_GUI_WIDGETHANDLE_HPP
