#ifndef SL_GUI_WIDGETLIST_HPP
#define SL_GUI_WIDGETLIST_HPP

#include <iterator>
#include <functional>
#include <vector>

#include "Engine/GUI/GUIWidget.hpp"
#include "Engine/GUI/Button.hpp"
#include "Engine/GUI/WidgetGrid.hpp"
#include "Engine/GUI/Scrollbar.hpp"

namespace sl
{
namespace gui
{

template <typename T>
class WidgetList : public GUIWidget
{
public:
	typedef std::function<std::unique_ptr<GUIWidget>(const T&)> PreviewCreator;
	typedef std::function<std::string(const T&)> NameCreator;
	typedef std::function<void(const T&)> Callback;

	WidgetList(const Rect<int>& box, const NameCreator& nameCreator, const Callback& onSelect = nullptr);


	WidgetList& setPreview(const PreviewCreator& creator);

	WidgetList& setOnSelect(const Callback& callback);

	WidgetList& setOnDoubleClick(const Callback& callback);

	WidgetList& setOnHover(const Callback& callback);

	WidgetList& setHorizAlignment(TextLabel::HorizontalPolicy halign);

	WidgetList& setButtonHeight(int height);

	template <typename Container>
	WidgetList& updateItems(const Container& container);

	WidgetList& registerAcceptButton(Key key);

	/**
	 * Adjusts the list box to exactly fit the items, if they don't fit.
	 */
	void trim();

	/**
	 * Re-creates the buttons using the name/previews
	 */
	void refresh();
private:

	void onClicked(int index);


	PreviewCreator previewCreator;
	NameCreator nameCreator;
	Callback onDoubleClick;
	Callback onSelect;
	Callback onHover;
	WidgetGrid *grid;
	Scrollbar *scrollbar;
	int selectedIndex;
	//! The index of the one hovered over by KB/gamepad controls (NOT mouse! this is independent!)
	int hoverIndex;
	std::vector<T> data;
	TextLabel::HorizontalPolicy halign;
	int buttonHeight;

	static constexpr int nothingSelected = -1;
};

template <typename T>
WidgetList<T>::WidgetList(const Rect<int>& box, const NameCreator& nameCreator, const Callback& onSelect)
	:GUIWidget(box)
	,previewCreator()
	,nameCreator(nameCreator)
	,onDoubleClick()
	,onSelect(onSelect)
	,grid(nullptr)
	,scrollbar(nullptr)
	,selectedIndex(nothingSelected)
	,hoverIndex(0)
	,halign(TextLabel::Center)
	,buttonHeight(32)
{
	auto widgetGrid = unique<WidgetGrid>(0, 0, 1, 1);
	grid = widgetGrid.get();

	auto scrollArea = unique<Scrollbar>(box);
	scrollArea->setScrollArea(std::move(widgetGrid));
	scrollbar = scrollArea.get();
	addWidget(std::move(scrollArea));

	registerInputEvent(Key::GUIUp, InputEventType::Pressed, [this] {
		if (hoverIndex > 0)
		{
			grid->at(0, hoverIndex)->setFocus(false);
			--hoverIndex;
			grid->at(0, hoverIndex)->setFocus(true);
			// Since scrolling transforms the x/y coordinates of buttons, we must compare it to the 1st button.
			const int originalTop = grid->at(0, hoverIndex)->getBox().top - grid->at(0, 0)->getBox().top;
			if (originalTop < scrollbar->getVisibleRegion().top)
			{
				scrollbar->setVisiblePosition(Vector2i(0, originalTop));
			}
		}
	});
	registerInputEvent(Key::GUIDown, InputEventType::Pressed, [this] {
		if (hoverIndex < grid->getRows() - 1)
		{
			grid->at(0, hoverIndex)->setFocus(false);
			++hoverIndex;
			grid->at(0, hoverIndex)->setFocus(true);
			// Since scrolling transforms the x/y coordinates of buttons, we must compare it to the 1st button.
			const int originalBottom = grid->at(0, hoverIndex)->getBox().bottom() - grid->at(0, 0)->getBox().top;
			if (originalBottom >= scrollbar->getVisibleRegion().bottom())
			{
				scrollbar->setVisiblePosition(Vector2i(0, originalBottom - scrollbar->getBox().height));
			}
		}
	});
	registerAcceptButton(Key::GUIAccept);
}

template <typename T>
WidgetList<T>& WidgetList<T>::setPreview(const PreviewCreator& creator)
{
	previewCreator = creator;
	return *this;
}

template <typename T>
WidgetList<T>& WidgetList<T>::setOnSelect(const Callback& callback)
{
	onSelect = callback;
	return *this;
}

template <typename T>
WidgetList<T>& WidgetList<T>::setOnDoubleClick(const Callback& callback)
{
	onDoubleClick = callback;
	return *this;
}

template <typename T>
WidgetList<T>& WidgetList<T>::setOnHover(const Callback& callback)
{
	onHover = callback;
	return *this;
}


template <typename T>
WidgetList<T>& WidgetList<T>::setHorizAlignment(TextLabel::HorizontalPolicy halign)
{
	this->halign = halign;
	// TODO: re-create all buttons with new alignment? or keep it this way?
	return *this;
}

template <typename T>
WidgetList<T>& WidgetList<T>::setButtonHeight(int height)
{
	buttonHeight = height;
	return *this;
}

template <typename T>
template <typename Container>
WidgetList<T>& WidgetList<T>::updateItems(const Container& container)
{
	const int n = std::size(container);
	grid->setRows(n);
	data.clear();
	data.reserve(n);
	for (const T& item : container)
	{
		data.push_back(item);
	}
	selectedIndex = nothingSelected;
	hoverIndex = 0;
	refresh();
	return *this;
}

template <typename T>
WidgetList<T>& WidgetList<T>::registerAcceptButton(Key key)
{
	registerInputEvent(key, InputEventType::Released, [this] {
		onClicked(hoverIndex);
	});
	return *this;
}

template <typename T>
void WidgetList<T>::trim()
{
	const int listHeight = data.size() * buttonHeight;
	if (box.height > listHeight)
	{
		this->setSize(box.width, listHeight);
	}
}

template <typename T>
void WidgetList<T>::refresh()
{
	// we need to do reset to the on-creation position since children are relative
	const float vertScroll = scrollbar->getVerticalScroll();
	scrollbar->setVerticalScroll(0.f);
	// then force-move it since the above code (same with move()) won't get called until next update
	grid->update(0, 0);

	int index = 0;
	for (const T& item : data)
	{
		std::unique_ptr<GUIWidget> preview;
		if (previewCreator)
		{
			preview = previewCreator(item);
		}
		auto itemWidget = unique<GUIWidget>(Rect<int>(0, 0, box.width - scrollbar->getScrollbarWidth(), buttonHeight));
		const int previewWidth = preview ? preview->getBox().width : 0;
		const Rect<int> buttonBox(previewWidth, 0, box.width - scrollbar->getScrollbarWidth() - previewWidth, buttonHeight);
		auto button = Button::createGenericButton(buttonBox, nameCreator(item), std::bind(&WidgetList<T>::onClicked, this, index), 1, halign);
		if (onHover)
		{
			button->setOnHover([this, index] {
				onHover(data[index]);
			});
		}
		if (preview)
		{
			preview->move(0, (buttonHeight - preview->getBox().height) / 2);
			itemWidget->addWidget(std::move(preview));
		}
		itemWidget->addWidget(std::move(button));
		itemWidget->moveBy(0, buttonHeight * index);
		grid->setWidget(0, index, std::move(itemWidget));
		++index;
	}
	grid->setFocus(false);
	grid->at(0, hoverIndex)->setFocus(true);

	// and then focus back to where it was
	scrollbar->setVerticalScroll(vertScroll);
}

// private
template <typename T>
void WidgetList<T>::onClicked(int index)
{
	if (selectedIndex == index && onDoubleClick)
	{
		onDoubleClick(data[index]);
	}
	else if (onSelect)
	{
		onSelect(data[index]);
	}
	selectedIndex = index;
}

} // gui
} // sl

#endif // SL_GUI_WIDGETLIST_HPP