#include "Engine/Game.hpp"

#include <algorithm>
#include <iostream>
#include <sstream>
#include <fstream>

#include "Engine/Input/Input.hpp"
#include "Engine/Graphics/RenderTarget.hpp"
#include "Engine/Graphics/TextureManager.hpp"
#include "Engine/GUI/FontManager.hpp"
#include "Engine/Serialization/BinaryDeserialize.hpp"
#include "Engine/Serialization/BinarySerialize.hpp"
#include "Menus/Main/MainMenu.hpp"
#include "Engine/Entity.hpp"
#include "Engine/AbstractScene.hpp"
#include "Utility/Assert.hpp"


namespace sl
{

TextureManager *TextureManager::instance = nullptr;

const char *drawCollisions = "draw_collision_masks";

const int windowWidth = 512;
const int windowHeight = 384;

Game::Game()
	:towngenSeed(0)
	,input()
	,settings()
	,width(windowWidth)
	,height(windowHeight)
	,frameRate(30)
	,isRunning(false)
	,window(sf::VideoMode(width, height), "")
	,renderTarget(window)
	,view()
	,console(*this)
	,playerControls()
	,timestamp(0)
{
	TextureManager::init();
	gui::FontManager::create();
	srand(time(0));
	//---------------------- Debug Settings
	settings.createSystemSetting(drawCollisions, 0);
	settings.createSystemSetting("draw_car_path_graph", 0);
	settings.createSystemSetting("draw_npc_path_graph", 0);
	settings.createSystemSetting("draw_drawable_outlines", 0);
	settings.createSystemSetting("draw_heightmaps", 0);
	settings.createSystemSetting("draw_noise", 0);
	settings.createSystemSetting("loli_mode", 0);
	settings.createSystemSetting("draw_btree", 0);
	settings.createSystemSetting("draw_activities", 0);
	settings.createSystemSetting("sight_test", 0);
	settings.createSystemSetting("CarConfig", 1);
	settings.createSystemSetting("static_geo_grid_draw", 0);
	settings.createSystemSetting("draw_portals", 0);
	settings.createSystemSetting("draw_y_bottom", 0);
#ifdef SL_DEBUG
	settings.createSystemSetting("unlimited_sprint", 1);
#else
	settings.createSystemSetting("unlimited_sprint", 0);
#endif // SL_DEBUG

	//---------------------- System Settings
	settings.createSystemSetting("player_sneak_speed", 5);
	settings.createSystemSetting("player_walk_speed", 23);
	settings.createSystemSetting("player_sprint_speed", 69);
	


	//-------------------- User-visible Settings
	settings.createSetting("metric", "Measurements", 0);
	settings.setValuesByRange("metric", 0, 1);
	settings.setNameFunc("metric",
		[](int value) -> std::string
		{
			ASSERT(value == 0 || value == 1);
			return value ? "Metric" : "Imperial";
		}
	);


	settings.createSetting("res_scale", "Resolution", maxResScale());
	settings.setOnChangeCallback("res_scale", [&](int newScale) {
		window.setSize(sf::Vector2u(windowWidth * newScale, windowHeight * newScale));
	});
	settings.setValuesCusom("res_scale",
		[]
		{
			return 1;
		},
		[&]
		{
			return maxResScale();
		}
	);
	settings.setNameFunc("res_scale", [&](int scale) -> std::string
	{
		std::stringstream ss;
		ss << scale * getWindowWidth() << "x" << scale * getWindowHeight();
		return ss.str();
	});


	settings.createSetting("blood_visible", "Blood", true);
	settings.setValuesByRange("blood_visible", 0, 1);
	settings.setNameFunc("blood_visible", [](int b) -> std::string
	{
		return b ? "On" : "Off";
	});

	settings.createSetting("portraits_visible", "Dialogue Portraits", true);
	settings.setValuesByRange("portraits_visible", 0, 1);
	settings.setNameFunc("portraits_visible", [](int b) -> std::string
	{
		return b ? "On" : "Off";
	});

	settings.createSetting("bind_first_use", "Action First Use", false);
	settings.setValuesByRange("bind_first_use", 0, 1);
	settings.setNameFunc("bind_first_use", [](int b) -> std::string
	{
		return b ? "Perform" : "Switch";
	});

	settings.load();

	Logger::setDefaultLogger(&logger); // so everyone can be lazy fucks

#ifdef SL_DEBUG
	logger.setOutputLocation(Logger::Useless, Logger::File);
	logger.setOutputLocation(Logger::Debug, Logger::Console);
	logger.setOutputLocation(Logger::Serious, Logger::Popup);
	logger.setOutputLocation(Logger::UltraSerious, Logger::Popup);

	logger.setMinimumPriority(Logger::Dialogue, Logger::Debug);
	logger.setMinimumPriority(Logger::Graphics, Logger::Serious);
	logger.setMinimumPriority(Logger::Parsing, Logger::Serious);
	logger.setMinimumPriority(Logger::NPC, Logger::Debug);
	logger.setMinimumPriority(Logger::GUI, Logger::Serious);
	logger.setMinimumPriority(Logger::Physics, Logger::Serious);
	logger.setMinimumPriority(Logger::Serialization, Logger::Debug);
	logger.setMinimumPriority(Logger::Town, Logger::Debug);
	logger.setMinimumPriority(Logger::Temp, Logger::Debug);
	logger.setMinimumPriority(Logger::Pathfinding, Logger::Serious);//Logger::Useless);

	logger.setOutputLocation(Logger::Graphics, Logger::Serious, Logger::Console);
#else // SL_DEBUG
	logger.setMinimumPriority(Logger::Priority::Serious);
#endif // ELSE SL_DEBUG

	// movement
	playerControls.setKeybinds(Key::MoveUp,
		KeyBind(Input::KeyboardKey::W),
		KeyBind(Input::KeyboardKey::Up),
		XBOX360::LeftStickUp
	);
	playerControls.setKeybinds(Key::MoveDown,
		KeyBind(Input::KeyboardKey::S),
		KeyBind(Input::KeyboardKey::Down),
		XBOX360::LeftStickDown
	);
	playerControls.setKeybinds(Key::MoveRight,
		KeyBind(Input::KeyboardKey::D),
		KeyBind(Input::KeyboardKey::Right),
		XBOX360::LeftStickRight
	);
	playerControls.setKeybinds(Key::MoveLeft,
		KeyBind(Input::KeyboardKey::A),
		KeyBind(Input::KeyboardKey::Left),
		XBOX360::LeftStickLeft
	);
	playerControls.setKeybinds(Key::Interact,
		KeyBind(Input::KeyboardKey::E),
		XBOX360::A
	);
	playerControls.setKeybinds(Key::Sprint,
		KeyBind(Input::KeyboardKey::LShift),
		XBOX360::B
	);
	playerControls.setKeybinds(Key::Sneak,
		KeyBind(Input::KeyboardKey::LControl),
		KeyBind(Input::KeyboardKey::Z), // since I have ctrl+arrow keys as hotkeys in programs I use, let's have this work too...
		XBOX360::LeftStickPressed
	);
	playerControls.setKeybinds(Key::CycleCallout,
		KeyBind(Input::KeyboardKey::Tab),
		XBOX360::X
	);
	playerControls.setKeybinds(Key::ActionChooser,
		KeyBind(Input::KeyboardKey::Q),
		XBOX360::Y
	);

	// driving
	playerControls.setKeybinds(Key::ExitCar,
		KeyBind(Input::KeyboardKey::E),
		XBOX360::B
	);
	playerControls.setKeybinds(Key::Brake,
		KeyBind(Input::KeyboardKey::LControl),
		KeyBind(Input::AxisButton::Z),
		XBOX360::X
	);
	playerControls.setKeybinds(Key::Accelerate,
		KeyBind(Input::KeyboardKey::W),
		KeyBind(Input::KeyboardKey::Up),
		XBOX360::RightTrigger
	);
	playerControls.setKeybinds(Key::Reverse,
		KeyBind(Input::KeyboardKey::S),
		KeyBind(Input::KeyboardKey::Down),
		XBOX360::LeftTrigger
	);
	playerControls.setKeybinds(Key::SteerLeft,
		KeyBind(Input::KeyboardKey::A),
		KeyBind(Input::KeyboardKey::Left),
		XBOX360::LeftStickLeft
	);
	playerControls.setKeybinds(Key::SteerRight,
		KeyBind(Input::KeyboardKey::D),
		KeyBind(Input::KeyboardKey::Right),
		XBOX360::LeftStickRight
	);
	playerControls.addKeybind(Key::Cruise, KeyBind(Input::KeyboardKey::Q));
	//TODO: remove
	playerControls.addKeybind(Key::Autopilot, KeyBind(Input::KeyboardKey::P));

	playerControls.setKeybinds(Key::Pause,
		KeyBind(Input::KeyboardKey::Escape),
		XBOX360::Start
	);
	playerControls.setKeybinds(Key::OpenInventory,
		KeyBind(Input::KeyboardKey::I),
		XBOX360::Back
	);
	playerControls.setKeybinds(Key::ToggleHud,
		KeyBind(Input::KeyboardKey::H)
	);
	playerControls.setKeybinds(Key::ToggleMap,
		KeyBind(Input::KeyboardKey::M),
		XBOX360::RightStickPressed
	);
	// GUI
	playerControls.setKeybinds(Key::GUIClick, KeyBind(Input::MouseButton::Left));
	playerControls.setKeybinds(Key::GUIRightClick, KeyBind(Input::MouseButton::Right));
	playerControls.setKeybinds(Key::GUIUp,
		KeyBind(Input::KeyboardKey::W),
		KeyBind(Input::KeyboardKey::Up),
		XBOX360::LeftStickUp,
		XBOX360::DPadUp
	);
	playerControls.setKeybinds(Key::GUIDown,
		KeyBind(Input::KeyboardKey::S),
		KeyBind(Input::KeyboardKey::Down),
		XBOX360::LeftStickDown,
		XBOX360::DPadDown
	);
	playerControls.setKeybinds(Key::GUIAccept,
		KeyBind(Input::KeyboardKey::Space),
		KeyBind(Input::KeyboardKey::Return),
		XBOX360::A
	);
	playerControls.setKeybinds(Key::GUIBack,
		KeyBind(Input::KeyboardKey::BackSpace),
		XBOX360::B
	);
	playerControls.setKeybinds(Key::GUIRight,
		KeyBind(Input::KeyboardKey::D),
		KeyBind(Input::KeyboardKey::Right),
		XBOX360::LeftStickRight,
		XBOX360::DPadRight
	);
	playerControls.setKeybinds(Key::GUILeft,
		KeyBind(Input::KeyboardKey::A),
		KeyBind(Input::KeyboardKey::Left),
		XBOX360::LeftStickLeft,
		XBOX360::DPadLeft
	);
}

Game::~Game()
{
	scenes.clear();
	scenesToAdd.clear();
#ifdef SL_DEBUG
	bool fine = true;
	for (std::map<std::string, std::pair<unsigned int, unsigned int>>::iterator it = Entity::entityCount.begin(); it != Entity::entityCount.end(); ++it)
	{
		if (it->second.first != 0)
		{
			std::cerr << "Memory leak found: " << it->second.first << " / " << it->second.second <<  " undestroyed Entities of subtype " << it->first << std::endl;
			fine = false;
		}
	}
	ASSERT(fine);
#endif
	// DO NOT CALL THIS AFTER TextureManager::destroy()
	gui::FontManager::destroy();

	// do not call this before scenes/entities are destroyed
	TextureManager::destroy();
}

int Game::run()
{
	sf::Image icon;
	icon.loadFromFile("resources/icon.png");
	window.setIcon(icon.getSize().x, icon.getSize().y, icon.getPixelsPtr());
	window.setFramerateLimit(frameRate);

	input.setWindow(window);

	//view.setSize(townWidth, townHeight);
	view.setSize(width, height);
	//view.setSize(10240, 6144);

	std::size_t seconds = time(0), frames = 0;

	console.init();

	addScene(unique<MainMenu>(this));

	isRunning = true;

	while (isRunning && window.isOpen())
	{
		sf::Event event;
		while (window.pollEvent(event))
		{
			switch (event.type)
			{
			case sf::Event::KeyPressed:
				switch (event.key.code)
				{
				//case sf::Keyboard::Escape:
				//	window.close();
				//	break;
				case sf::Keyboard::F7:
					settings.set(drawCollisions, settings.get(drawCollisions) == 0 ? 1 : 0);
					break;
				case sf::Keyboard::F4:
					{
						const auto& setting = settings.getSetting("res_scale");
						settings.set("res_scale", setting.value == setting.end() ? setting.begin() : setting.next(setting.value));
					}
					break;
				default:
					break;
				}
				break;
			case sf::Event::GainedFocus:
				input.setFocus(true);
				break;
			case sf::Event::LostFocus:
				input.setFocus(false);
				break;
			case sf::Event::Closed:
				window.close();
				break;
			default:
				break;
			}
		}

		//Update input and scenes
		input.update();

		if (console.isEnabled() && input.isEnabled())
		{
			input.disable();
		}
		else if (!console.isEnabled() && !input.isEnabled())
		{
			input.enable();
		}

		for (auto &scene : scenes)
		{
			scene->handleEvents(eventQueue);
		}
		eventQueue.reset();
		for (auto &scene : scenes)
		{
			scene->update(timestamp);
		}

		bool wasEnabled = input.isEnabled();

		if (!wasEnabled && console.isEnabled())
		{
			input.enable();
		}

		console.update();

		if (!wasEnabled && input.isEnabled())
		{
			input.isEnabled();
		}

		redraw();


		//Update frame and title
		if (time(0) != (int) seconds)
		{
			std::stringstream ss;
			//ss << "Game";
			ss << "\t\t\tSim Loli PRE ALPHA 0.4.1   ";
			if (towngenSeed != 0)
			{
				ss << "-   townseed = "
				   << towngenSeed;
			}
			ss << "   (FPS: "
			   << (frames)
			   << "/"
			   << frameRate
			   << ")"
			   << std::endl;

			window.setTitle(ss.str());
			frames = 0;
			seconds = time(0);
		}
		else
		{
			++frames;
		}

		//Update and reorganize scene vector if necessary
		const bool updateSort = clearRemovedScenes() || addNewScenes();
		if (updateSort)
		{
			sortScenesByDepth();
		}

		for (const auto& s : scheduledSerializations)
		{
			BinarySerialize out(s.first);
			out.saveSerializable("root", s.second);
		}
		scheduledSerializations.clear();

		++timestamp;
	}

	return EXIT_SUCCESS;
}

void Game::addScene(std::unique_ptr<AbstractScene> scene)
{
	scenesToAdd.push_back(std::move(scene));
}

void Game::removeScene(AbstractScene* scene)
{
	scenesToDelete.push_back(scene);
}

void Game::registerEvent(std::unique_ptr<const Event> event)
{
	ASSERT(event->getCategory() != Event::Category::NonBroadcasted);
	eventQueue.addEvent(std::move(event));
}

ls::Bindings& Game::getContext()
{
	return context;
}

const Input& Game::getRawInput() const
{
	return input;
}

int Game::getWindowWidth() const
{
	return width;
}

int Game::getWindowHeight() const
{
	return height;
}

AbstractScene::Timestamp Game::getTimestamp() const
{
	return timestamp;
}

Settings& Game::getSettings()
{
	return settings;
}

const Settings& Game::getSettings() const
{
	return settings;
}

const Logger& Game::getLogger() const
{
	return logger;
}

void Game::end()
{
	isRunning = false;
}

void Game::redraw()
{
	window.clear(sf::Color(0, 0, 0));

	for (std::unique_ptr<AbstractScene> &scene : scenes)
	{
		// @SAVEMAP
		view.setCenter(scene->getView());
		//view.setCenter(sf::Vector2f(scene->getViewWidth() / 2.f, scene->getViewHeight() / 2.f));

		window.setView(view);
		if (false)//if (settings.exists("save_level_image") && settings.get("save_level_image"))
		{
			sf::RenderTexture fileTex;
			RenderTarget fileTarget(fileTex);
			fileTex.create(10240, 6144);//scene->getViewWidth(), scene->getViewHeight());
			scene->draw(fileTarget);
			fileTarget.render();
			fileTex.getTexture().copyToImage().saveToFile("level_image.png");
		}
		scene->draw(renderTarget);

		window.setView(window.getDefaultView());

		scene->drawGUI(renderTarget);
	}

	console.draw(window);

	window.display();
}


const ControllerBinds& Game::getPlayerControls() const
{
	return playerControls;
}

ControllerBinds& Game::getMutablePlayerControls()
{
	return playerControls;
}

void Game::scheduleSerialization(const std::string& filename, const Serializable *serializable)
{
	scheduledSerializations.push_back(std::make_pair(filename, serializable));
}

/*				private					*/
bool Game::clearRemovedScenes()
{
	bool hasRemoved = false;

	for (const AbstractScene *toDelete : scenesToDelete)
	{
		for (int i = 0; i < scenes.size(); ++i)
		{
			if (toDelete == scenes[i].get())
			{
				hasRemoved = true;
				std::swap(scenes[i],scenes.back());
				scenes.pop_back();
			}
		}
	}
	scenesToDelete.clear();
	return hasRemoved;
}

bool Game::addNewScenes()
{
	bool hasAdded = !scenesToAdd.empty();

	for (auto &toAdd : scenesToAdd)
	{
		scenes.push_back(std::move(toAdd));
	}
	scenesToAdd.clear();
	return hasAdded;
}

void Game::sortScenesByDepth()
{
	std::sort(scenes.begin(), scenes.end(), [](const std::unique_ptr<AbstractScene> &a, const std::unique_ptr<AbstractScene> &b)
	{
		return a->getDepth() > b->getDepth();
	});
}

int Game::maxResScale() const
{
	return std::min(sf::VideoMode::getDesktopMode().width / getWindowWidth(), sf::VideoMode::getDesktopMode().height / getWindowHeight());
}


}// End of sl namespace
