/**
 * @section DESCRIPTION
 * A class representing the entire game.
 */
#ifndef SL_GAME_HPP
#define SL_GAME_HPP

#include <vector>

#include <SFML/Graphics.hpp>

#include "Engine/Events/Event.hpp"
#include "Engine/Events/EventQueue.hpp"
#include "Engine/Console.hpp"
#include "Engine/Input/ControllerBinds.hpp"
#include "Engine/Input/Input.hpp"
#include "Engine/AbstractScene.hpp"
#include "Engine/Settings.hpp"
#include "Parsing/Binding/Bindings.hpp"
#include "Utility/Logger.hpp"

namespace sl
{

class Console;
class Serializable;

class Game
{
public:
	/**
	 * Constructs an instance of Sim Loli
	 */
	Game();
	/**
	 * Destroys Sim Loli
	 */
	~Game();
	/**
	 * Runs Sim Loli
	 * @return An error code if the game encounters a runtime error.
	 */
	int run();
	/**
	 * Adds a AbstractScene to the game's scene manager
	 * @param scene The scene to be added
	 */
	void addScene(std::unique_ptr<AbstractScene> scene);
	/**
	 * Removes a scene from the game's scene manager
	 * @param scene The scene to be removed
	 */
	void removeScene(AbstractScene* scene);
	/**
	 * Registers a global event
	 * @param event The event to be handled
	 */
	void registerEvent(std::unique_ptr<const Event> event);

	ls::Bindings& getContext();

	const Input& getRawInput() const;

	int getWindowWidth() const;

	int getWindowHeight() const;

	AbstractScene::Timestamp getTimestamp() const;

	Settings& getSettings();

	const Settings& getSettings() const;

	const Logger& getLogger() const;

	void end();

	void redraw();

	const ControllerBinds& getPlayerControls() const;

	ControllerBinds& getMutablePlayerControls();

	/**
	 * Schedules a save to happen happen after the end of an update. This is to ensure that Scenes/etc are completely consistent and to reduce complexity.
	 * @param filename Where to save the file to.
	 * @param serializable The serializable to serialize.
	 */
	void scheduleSerialization(const std::string& filename, const Serializable *serializable);

	// @HACK @TODO remove after demo - only here so people can report the seed for towngen issues
	unsigned int towngenSeed;

private:
	bool clearRemovedScenes();

	bool addNewScenes();

	void sortScenesByDepth();

	int maxResScale() const;



	//! A list of all the current scenes in the game
	std::vector<std::unique_ptr<AbstractScene>> scenes;
	std::vector<std::unique_ptr<AbstractScene>> scenesToAdd;
	std::vector<AbstractScene*> scenesToDelete;

	//! A queue of all new events
	EventQueue eventQueue;

	ls::Bindings context;

	Input input;

	Settings settings;

	Logger logger;

	const int width;
	const int height;
	const int frameRate;

	bool isRunning;

	sf::RenderWindow window;
	RenderTarget renderTarget;
	sf::View view;
	Console console;
	ControllerBinds playerControls;

	AbstractScene::Timestamp timestamp;

	std::vector<std::pair<const std::string, const Serializable*>> scheduledSerializations;
};

}// End of sl namespace

#endif
