/**
 * @section DESCRIPTION
 * This class represents a palette swap on a sprite as an ID to facilitate easy
 * comparing and mapping of textures that have been swapped
 */
#ifndef SL_COLORSWAPID_HPP
#define SL_COLORSWAPID_HPP

#include <SFML/Graphics/Color.hpp>

#include <algorithm>
#include <cstdint>
#include <initializer_list>
#include <ostream>
#include <utility>
#include <vector>

namespace sl
{

class Deserialize;
class Serialize;

class ColorSwapID
{
public:
	/**
	 * Constucts a ColorSwapID that has had no colours swapped on its palette
	 */
	ColorSwapID();
	/**
	 * Constructs a ColorSwapID with a variable amount of colours swapped in its palette
	 * @oaram swaps A list of all swaps in the form of {pair(from, to), pair(from, to)...}
	 */
	template <typename T>
	explicit ColorSwapID(const T& swaps)
		:id(swaps.size() * 2)
	{
		std::vector<uint32_t>::iterator vit = id.begin();
		for (const std::pair<sf::Color, sf::Color>& swap : swaps)
		{
			*vit = (swap.first.r << 24) | (swap.first.g << 16) | (swap.first.b << 8) | (swap.first.a);
			++vit;
			*vit = (swap.second.r << 24) | (swap.second.g << 16) | (swap.second.b << 8) | (swap.second.a);
			++vit;
		}
		//std::sort(id.begin(), id.end());
	}

	explicit ColorSwapID(const std::initializer_list<std::pair<sf::Color, sf::Color>> swaps)
		:id(swaps.size() * 2)
	{
		std::vector<uint32_t>::iterator vit = id.begin();
		for (const std::pair<sf::Color, sf::Color>& swap : swaps)
		{
			*vit = (swap.first.r << 24) | (swap.first.g << 16) | (swap.first.b << 8) | (swap.first.a);
			++vit;
			*vit = (swap.second.r << 24) | (swap.second.g << 16) | (swap.second.b << 8) | (swap.second.a);
			++vit;
		}
		//std::sort(id.begin(), id.end());
	}

	/**
	 * Compares two ColorSwapIDs so that it can be used in std::map and the like
	 */
	bool operator<(const ColorSwapID& rhs) const;

	/**
	 * Extracts the sf::Color data used to construct this ID back and inserts
	 * it into the input vector.
	 * @param colors The vector to extract to
	 */
	void getColors(std::vector<std::pair<sf::Color, sf::Color>>& colors) const;

	/**
	 * Checks whether or not this is a non-empty SwapID
	 * @return true if swaps, false if no swaps
	 */
	explicit operator bool() const;

	void serialize(Serialize& out) const;

	void deserialize(Deserialize& in);

private:
	//!An list of easily comparable ids for each colour swap (represents an sequence of 24bit
	//!integers alternating between original and swapped RGB values)
	std::vector<uint32_t> id;

	friend std::ostream& operator<<(std::ostream& os, const ColorSwapID& swapID);
	
	friend ColorSwapID operator|(const ColorSwapID& lhs, const ColorSwapID& rhs);
};

std::ostream& operator<<(std::ostream& os, const sl::ColorSwapID& swapID);

ColorSwapID operator|(const ColorSwapID& lhs, const ColorSwapID& rhs);

} // sl

#endif
