/**
 * @section DESCRIPTION
 * This is an abstract interface for all drawable components that belong to entities.
 */
#ifndef SL_DRAWABLE_HPP
#define SL_DRAWABLE_HPP

#include <SFML/Graphics/RenderTarget.hpp>

#include "Engine/Serialization/Serializable.hpp"
#include "Utility/Rect.hpp"

namespace sl
{

class RenderTarget;

//! Making this serializable was an attempt to serialize StaticGeometry in sane way. In general, you shoudl reconstruct these rather than deserialize them.
class Drawable : public Serializable
{
public:
	virtual ~Drawable();
	/**
	 * The interface for moving the Drawablea
	 * @param x The new x position
	 * @param y The new y position
	 */
	virtual void update(int x, int y) = 0;
	/**
	 * returns the visual depth of the object, or which order (lower = rendered on top)
	 * to render stuff in.
	 * @return The depth of the Drawable...
	 */
	int getDepth() const;

	void setDepth(int depth);


	/**
	 * Gets the bounds of the animation right now.
	 * This is relative to the Drawable's position, not the origin.
	 * This is used for culling before drawing.
	 */
	virtual const Rect<int>& getCurrentBounds() const = 0;

	// @TODO move this to Animation?
	/**
	 * Gets the speed at which Animation plays
	 * (returns 0 (non-animating) by default, classes like Animation override it to do stuff)
	 * @return The animation speed (1 = normal, 0 = still, negative = reversed)
	 */
	virtual float getAnimationSpeed() const;

	// @TODO move this to Animation?
	/**
	 * Sets the speed at which this Animation plays
	 * (does nothing by defaut, classes like Animation override it to do stuff)
	 * @param speed The animation speed (1 = normal, 0 = still, negative = reversed)
	 */
	virtual void setAnimationSpeed(float animationSpeed);

	// @TODO move this to Animation?
	/**
	 * @return Whether the class is 'done' drawing - only relevant for Animations and related drawables - returns false by default
	 */
	virtual bool isDone() const;

	/**
	* Resets an animation to the first frame
	*/
	virtual void reset() {}

	virtual void setProgress(float progress) {}

	virtual float getProgress() const { return 0.f; }

	/**
	 * Interface for drawing
	 * @param target The RenderTarget to draw to
	 */
	virtual void draw(sf::RenderTarget& target) const = 0;

	virtual void rotate(float newRotation);

	virtual void setColor(const sf::Color& color);

	void setHeight(int height);

	int getHeight() const;

	int getDepthConflictPower() const;

	/**
	 * Sets the offset from the bottom of the bounds for which we use as the bottom
	 * of the image for depth calculations
	 * @param offset The offset. Positive = down, negative = up
	 */
	void setBottomPointOffset(int offset);

	int getBottomPointOffset() const;

protected:
	Drawable();

	void serializeDrawable(Serialize& out) const;

	void deserializeDrawable(Deserialize& in);

private:
	const int depthConflictPower;
	int height;
	int depth;
	int bottomPointOffset;


	static int depthConflictCounter;
};

} // sl

#endif
