#include "Engine/Graphics/DrawableAnimation.hpp"

#include "Engine/Graphics/RenderTarget.hpp"
#include "Engine/Graphics/Sprite.hpp"
#include "Engine/Graphics/TextureManager.hpp"
#include "Utility/Assert.hpp"

#include <numeric>

namespace sl
{

DrawableAnimation::DrawableAnimation(std::vector<std::pair<std::unique_ptr<Drawable>, std::size_t> > frames, bool repeats)
	:frames(std::move(frames))
	,currentFrame(0)
	,interFrameProgress(0)
	,animationSpeed(1)
	,frameNums(frames.size())
	,repeats(repeats)
{
}

DrawableAnimation::DrawableAnimation(std::vector<std::unique_ptr<Drawable>> drawables, std::size_t framerate, bool repeats)
	:frames()
	,currentFrame(0)
	,interFrameProgress(0)
	,animationSpeed(1)
	,frameNums(drawables.size())
	,repeats(repeats)
{
	// Allocate memory beforehand so push_back doesn't do an assload of syscalls
	this->frames.reserve(drawables.size());
	for (auto it = drawables.begin(); it != drawables.end(); ++it)
	{
		this->frames.push_back(std::make_pair(std::unique_ptr<Drawable>(it->release()), framerate));
	}
}

DrawableAnimation::DrawableAnimation(const std::string& fname, int cellWidth, std::size_t framerate, bool repeats, const ColorSwapID& swapID, int offX, int offY, bool mirrored)
	:frames()
	,currentFrame(0)
	,interFrameProgress(0)
	,animationSpeed(1)
	,frameNums(0)
	,currentBounds(cellWidth, -1)
	,repeats(repeats)
{
	ASSERT(cellWidth > 0);

	Texture texture(TextureManager::getTexture(fname));

	const sf::Vector2u& textureSize = texture.get().getSize();
	currentBounds.height = textureSize.y;

	frameNums = textureSize.x / cellWidth;
	frames.reserve(frameNums);

	for (Rect<int> subrect(0, 0, cellWidth, textureSize.y);
	     subrect.left < (int)textureSize.x;
	     subrect.left += cellWidth)
	{
		auto sprite = std::make_unique<Sprite>(fname.c_str(), swapID, offX, offY);
		if (mirrored)
		{
			sprite->mirrorHorizontally();
		}
		sprite->setRenderedSubrect(subrect);
		frames.push_back(std::make_pair(std::move(sprite), framerate));
	}
}


/*    DrawableAnimation public methods   */

void DrawableAnimation::reset()
{
	currentFrame = 0;
	interFrameProgress = 0;
}


bool DrawableAnimation::isRepeating() const
{
	return repeats;
}

void DrawableAnimation::setProgress(float progress)
{
	if (std::accumulate(frames.begin(), frames.end(), 0, [](int a, const auto& f) { return a + f.second; }) != 0)
	{
		progress -= interFrameProgress;
		for (currentFrame = 0; progress > 0.f; currentFrame = (currentFrame + 1) % frameNums)
		{
			ASSERT(frames[currentFrame].second > 0);
			progress -= frames[currentFrame].second;
		}
	}
}

float DrawableAnimation::getProgress() const
{
	float p = interFrameProgress;
	for (int i = 0; i < currentFrame; ++i)
	{
		p += frames[i].second;
	}
	return p;
}

void DrawableAnimation::rotate(float newAngle)
{
	for (auto& pair : frames)
	{
		// @hack for demo
		((Sprite*)pair.first.get())->rotate(newAngle);
	}
}

void DrawableAnimation::setFrame(int frameIndex)
{
	currentFrame = frameIndex;
}

int DrawableAnimation::getFrameCount() const
{
	return frames.size();
}

/*	  Drawable public methods		 */
void DrawableAnimation::update(int x, int y)
{
	if (!isDone())
	{
		if (animationSpeed > 0)
		{
			interFrameProgress += animationSpeed;

			if (interFrameProgress > frames[currentFrame].second)
			{
				interFrameProgress -= frames[currentFrame].second;
				++currentFrame;
				if (currentFrame >= frameNums && repeats)
				{
					currentFrame = 0;
				}
			}
		}
		else
		{
			interFrameProgress -= animationSpeed;

			if (interFrameProgress > frames[currentFrame].second)
			{
				interFrameProgress -= frames[currentFrame].second;
				--currentFrame;
				if (currentFrame < 0 && repeats)
				{
					currentFrame = frameNums - 1;
				}
			}
		}
	}

	frames[displayableFrame()].first->update(x, y);
	currentBounds.left = frames[displayableFrame()].first->getCurrentBounds().left;
	currentBounds.top = frames[displayableFrame()].first->getCurrentBounds().top;
}

const Rect<int>& DrawableAnimation::getCurrentBounds() const
{
	return currentBounds;//frames[displayableFrame()].first->getCurrentBounds();
}

float DrawableAnimation::getAnimationSpeed() const
{
	return animationSpeed;
}

void DrawableAnimation::setAnimationSpeed(float speed)
{
	animationSpeed = speed;
}

bool DrawableAnimation::isDone() const
{
	return !isRepeating() && currentFrame >= frameNums;
}

/*	  Drawable private methods		*/
void DrawableAnimation::draw(sf::RenderTarget& target) const
{
	frames[displayableFrame()].first->draw(target);
	//target.draw(*(frames[currentFrame].first));
}


int DrawableAnimation::displayableFrame() const
{
	return currentFrame < frameNums ? currentFrame : frameNums - 1;
}

}
