/**
 * @section DESCRIPTION
 * A sequence of Drawables that play in an animation.
 */
#ifndef SL_DRAWABLEANIMATION_HPP
#define SL_DRAWABLEANIMATION_HPP

#include <vector>
#include <initializer_list>
#include <utility>

#include "Engine/Graphics/Animation.hpp"
#include "Engine/Graphics/ColorSwapID.hpp"


namespace sl
{

class DrawableAnimation : public Animation //  Yo dawg, I heard you liked animating, so we made your DrawableAnimation
{                                          //  an Animation, so you can animiate while you animate.
public:
	/**
	 * Constructs an DrawableAnimation from a series of Drawable-frame length pairs
	 * so you can specifiy the length in frames that each Drawable is displayed
	 * for individually.
	 * @param frames The list of Drawable-length pairs
	 * @param repeats Whether the animation should start over from the first frame when it finishes
	 */
	DrawableAnimation(std::vector<std::pair<std::unique_ptr<Drawable>, std::size_t> > frames, bool repeats = true);

	/**
	 * Constructs an DrawableAnimation from a series of Drawables and a single frame-length
	 * so that each Drawable will display for exactly framerate frames each
	 * (assuming animationSpeed is still 1 and isn't set otherwise)
	 * @param frames The list of Drawables
	 * @param framerate The number of frames to display each Drawable for
	 * @param repeats Whether the animation should start over from the first frame when it finishes
	 */
	DrawableAnimation(std::vector<std::unique_ptr<Drawable>> frames, std::size_t framerate, bool repeats = true);

	/**
	 * Constructs an DrawableAnimation from a strip-formatted sprite filename with each
	 * frame in the strip displaying for exactly framerate frames each
	 * @param fname The filename of the sprite strip
	 * @param cellWidth The width of each cell. eg there are sprite's width / cellWidth cells total
	 * @param framerate The number of frames to display each cell for
	 * @param repeats Whether the animation should start over from the first frame when it finishes
	 * @param swapID Any palette swapping that should be done (optional - if not specified, none is done)
	 */
	DrawableAnimation(const std::string& fname, int cellWidth, std::size_t framerate, bool repeats, const ColorSwapID& swapID = ColorSwapID(), int offX = 0, int offY = 0, bool mirrored = false);

	/*	  Drawable public methods	 */
	void update(int x, int y) override;

	const Rect<int>& getCurrentBounds() const override;

	float getAnimationSpeed() const override;

	void setAnimationSpeed(float speed) override;

	/**
	 * Checks whether an animation is done
	 * @return true if done last frame and NOT repeating
	 */
	bool isDone() const override;

	/**
	 * Resets an animation to the first frame
	 */
	void reset() override;

	/**
	 * Checks if the animation repeats or not
	 * @return true if repeats
	 */
	bool isRepeating() const override;

	void setProgress(float progress) override;

	float getProgress() const override;

	void rotate(float newAngle) override;

	void setFrame(int frameIndex) override;;

	int getFrameCount() const override;

private:
	/*	  Drawable public methods	 */
	void draw(sf::RenderTarget& target) const override;

	int displayableFrame() const;


	//!All the drawables that compromise this animation as well as how many frames long they are
	std::vector<std::pair<std::unique_ptr<Drawable>, std::size_t>> frames;
	//!The current Drawable the animation is on
	std::size_t currentFrame;
	//!How far through the current Drawable is in frames
	float interFrameProgress;
	//!How many frames per update the DrawableAnimation plays at (1=normal)
	float animationSpeed;
	//!How many Drawables compromise the DrawableAnimation (same as frames.size())
	std::size_t frameNums;
	//! The current frame's bounds
	Rect<int> currentBounds;
	//!Whether the animation repeats
	bool repeats;
};

} // sl

#endif // SL_DRAWABLEANIMATION_HPP
