#include "Engine/Graphics/DrawableCollection.hpp"

#include <algorithm>

#include "Engine/Serialization/AutoDeserialize.hpp"
#include "Engine/Serialization/AutoSerialize.hpp"
#include "Utility/Assert.hpp"
#include "Utility/Logger.hpp"

namespace sl
{

DrawableCollection::DrawableCollection()
	:masterAnimSpeed(1.f)
{
}


void DrawableCollection::update(int x, int y)
{
	auto it = drawables.begin();
	auto end = drawables.end();
	if (it != end)
	{
		it->first->update(x, y);
		currentBounds = it->first->getCurrentBounds();
		int bottom = currentBounds.bottom() + it->first->getBottomPointOffset();
		++it;
		for (; it != end; ++it)
		{
			it->first->update(x, y);
			const int newRight = std::max(currentBounds.left + currentBounds.width, it->first->getCurrentBounds().left + it->first->getCurrentBounds().width);
			const int newBottom = std::max(currentBounds.top + currentBounds.height, it->first->getCurrentBounds().top + it->first->getCurrentBounds().height);
			currentBounds.left = std::min(it->first->getCurrentBounds().left, currentBounds.left);
			currentBounds.top = std::min(it->first->getCurrentBounds().top, currentBounds.top);
			currentBounds.width = newRight - currentBounds.left;
			currentBounds.height = newBottom - currentBounds.top;
			bottom = std::max(bottom, it->first->getCurrentBounds().bottom() + it->first->getBottomPointOffset());

			// cache the animation speed for ??? reason
			it->second = it->first->getAnimationSpeed();
		}
		// TODO: what to do if someone calls setBottomPointOffset() on us from the outside? should we handle that?
		setBottomPointOffset(bottom - currentBounds.bottom());
	}
}


const Rect<int>& DrawableCollection::getCurrentBounds() const
{
	return currentBounds;
}

float DrawableCollection::getAnimationSpeed() const
{
	return masterAnimSpeed;
}

void DrawableCollection::setAnimationSpeed(float animationSpeed)
{
	masterAnimSpeed = animationSpeed;
	for (auto& drawable : drawables)
	{
		drawable.first->setAnimationSpeed(/*drawable.second * */animationSpeed);
	}
}

bool DrawableCollection::isDone() const
{
	for (const auto& drawable : drawables)
	{
		if (!drawable.first->isDone())
		{
			return false;
		}
	}
	return true;
}

void DrawableCollection::reset()
{
	for (auto& drawable : drawables)
	{
		drawable.first->reset();
	}
}

void DrawableCollection::setProgress(float progress)
{
	for (auto& drawable : drawables)
	{
		drawable.first->setProgress(progress);
	}
}

// TODO: move to /Utility/
static constexpr bool floatEqEps(float x, float y)
{
	return x >= y - std::numeric_limits<float>::epsilon() && x <= y + std::numeric_limits<float>::epsilon();
}

float DrawableCollection::getProgress() const
{
	// Drawable::getProgress() defaults to 0.f, so let's just take one arbitrarily
	float p = 0.f;
	for (const auto& drawable : drawables)
	{
		const float dp = drawable.first->getProgress();
		if (dp != 0.f)
		{
			if (p != 0.f && !floatEqEps(p, dp))
			{
				Logger::log(Logger::Graphics, Logger::Debug) << "DrawableContainer not homogenous progress but tried to call getProgres() [" << p << "] vs [" << dp << "]" << std::endl;
			}
			p = dp;
		}
	}
	return p;
}

void DrawableCollection::addDrawable(std::unique_ptr<Drawable> drawable)
{
	const float animSpeed = drawable->getAnimationSpeed();
	drawables.push_back(std::make_pair(std::move(drawable), animSpeed));
}

void DrawableCollection::rotate(float newRotation)
{
	for (auto& drawable : drawables)
	{
		drawable.first->rotate(newRotation);
	}
}

void DrawableCollection::setColor(const sf::Color& color)
{
	for (auto& drawable : drawables)
	{
		drawable.first->setColor(color);
	}
}

void DrawableCollection::serialize(Serialize& out) const
{
	serializeDrawable(out);
	SAVEVAR(f, masterAnimSpeed);
	AutoSerialize::serialize(out, currentBounds);
	//out.startArray("drawables", drawables.size());
	//for (const std::pair<std::unique_ptr<Drawable>, float>& drawable : drawables)
	//{
	//	const bool shouldSerialize = drawable.first->shouldSerialize();
	//	ASSERT(shouldSerialize);
	//	if (shouldSerialize)
	//	{
	//		out.saveSerializable("drawable", drawable.first.get());
	//		out.f("speed", drawable.second);
	//	}
	//}
	//out.endArray("drawables");
	AutoSerialize::serialize(out, drawables);
}

void DrawableCollection::deserialize(Deserialize& in)
{
	deserializeDrawable(in);
	LOADVAR(f, masterAnimSpeed);
	AutoDeserialize::deserialize(in, currentBounds);
	//drawables.resize(in.startArray("drawables"));
	//for (std::pair<std::unique_ptr<Drawable>, float>& drawable : drawables)
	//{
	//	drawable.first.reset(static_cast<Drawable*>(in.loadSerializable("drawable").release()));
	//	drawable.second = in.f("speed");
	//}
	//in.endArray("drawables");
	AutoDeserialize::deserialize(in, drawables);
}

std::string DrawableCollection::serializationType() const
{
	return "DrawableCollection";
}

bool DrawableCollection::shouldSerialize() const
{
	return true;
}

//private
void DrawableCollection::draw(sf::RenderTarget& target) const
{
	for (const auto& drawable : drawables)
	{
		drawable.first->draw(target);
	}
}

} // sl
