/**
 * @section DESCRIPTION
 * A set of drawables to draw all at once and treat as if it's one single drawable
 */
#ifndef SL_DRAWABLECONTAINER_HPP
#define SL_DRAWABLECONTAINER_HPP

#include <memory>
#include <vector>

#include "Engine/Graphics/Drawable.hpp"

namespace sl
{

class DrawableCollection : public Drawable
{
public:
	DrawableCollection();

	/**
	 * The interface for moving the Drawablea
	 * @param x The new x position
	 * @param y The new y position
	 */
	void update(int x, int y) override;


	/**
	 * Gets the bounds of the animation right now.
	 * This is relative to the Drawable's position, not the origin.
	 * This is used for culling before drawing.
	 */
	const Rect<int>& getCurrentBounds() const override;

	/**
	 * Gets the speed at which Animation plays
	 * (returns 0 (non-animating) by default, classes like Animation override it to do stuff)
	 * @return The animation speed (1 = normal, 0 = still, negative = reversed)
	 */
	float getAnimationSpeed() const override;

	/**
	 * Sets the speed at which this Animation plays
	 * (does nothing by defaut, classes like Animation override it to do stuff)
	 * @param speed The animation speed (1 = normal, 0 = still, negative = reversed)
	 */
	void setAnimationSpeed(float animationSpeed) override;

	/**
	 * @return Whether the class is 'done' drawing - only relevant for Animations and related drawables - returns false by default
	 */
	bool isDone() const override;

	void reset() override;

	void setProgress(float progress) override;

	float getProgress() const override;

	/**
	 * Adds a Drawable to this collection. It will be rendered in the order it is added here.
	 * @param drawable The Drawable to add
	 */
	void addDrawable(std::unique_ptr<Drawable> drawable);

	void rotate(float newAngle) override;

	void setColor(const sf::Color& color) override;

	void serialize(Serialize& out) const override;

	void deserialize(Deserialize& in) override;

	std::string serializationType() const override;

	bool shouldSerialize() const override;

private:

	void draw(sf::RenderTarget& target) const override;


	float masterAnimSpeed;
	Rect<int> currentBounds;
	std::vector<std::pair<std::unique_ptr<Drawable>, float>> drawables;
};

} // sl

#endif // SL_DRAWABLECONTAINER_HPP
