#include "Engine/Graphics/SpriteAnimation.hpp"

#include "Engine/Graphics/Framerate.hpp"
#include "Engine/Graphics/RenderTarget.hpp"
#include "Engine/Graphics/TextureManager.hpp"
#include "Engine/Serialization/AutoDeserialize.hpp"
#include "Engine/Serialization/AutoSerialize.hpp"
#include "Engine/Serialization/Impl/SerializeFrame.hpp"
#include "Utility/Assert.hpp"

#include <numeric>

namespace sl
{

SpriteAnimation::SpriteAnimation(const std::string& fname, int cellWidth, const Framerate& framerate, bool repeats, const ColorSwapID& swapID, int offX, int offY, bool mirrored)
	:sprite(fname.c_str(), swapID, offX, offY)
	,frames(framerate.makeAnimationSpeeds(fname, cellWidth))
	,currentFrame(0)
	,interFrameProgress(0)
	,animationSpeed(1)
	,repeats(repeats)
	,cellWidth(cellWidth)
{
	ASSERT(cellWidth > 0);
	if (mirrored)
	{
		sprite.mirrorHorizontally();
	}
	setPosition(0, 0);
	//sprite.setRenderedSubrect(Rect<int>(0, 0, cellWidth, sprite.getCurrentBounds().height));
	updateSpriteSubrect();
}


/*    SpriteAnimation public methods   */
void SpriteAnimation::reset()
{
	currentFrame = 0;
	interFrameProgress = 0;
}


bool SpriteAnimation::isRepeating() const
{
	return repeats;
}

void SpriteAnimation::setProgress(float progress)
{
	// protect against non-animating ones
	if (std::accumulate(frames.begin(), frames.end(), 0, [](int a, const auto& f) { return a + f.duration; }) != 0)
	{
		for (currentFrame = 0; progress > frames[currentFrame].duration; currentFrame = (currentFrame + 1) % frames.size())
		{
			ASSERT(frames[currentFrame].duration > 0);
			progress -= frames[currentFrame].duration;
		}
		interFrameProgress = progress;
		updateSpriteSubrect();
	}
}

float SpriteAnimation::getProgress() const
{
	float p = interFrameProgress;
	for (int i = 0; i < currentFrame; ++i)
	{
		p += frames[i].duration;
	}
	return p;
}

void SpriteAnimation::rotate(float newAngle)
{
	sprite.rotate(newAngle);
}

void SpriteAnimation::setColor(const sf::Color& color)
{
	sprite.setColor(color);
}

/*	  Drawable public methods		 */
void SpriteAnimation::update(int x, int y)
{
	if (!isDone())
	{
		if (animationSpeed > 0)
		{
			interFrameProgress += animationSpeed;

			if (interFrameProgress > frames[currentFrame].duration)
			{
				interFrameProgress -= frames[currentFrame].duration;
				++currentFrame;
				if (currentFrame >= getFrameCount() && repeats)
				{
					currentFrame = 0;
				}
				updateSpriteSubrect();
			}
		}
		else if (animationSpeed < 0)
		{
			interFrameProgress -= animationSpeed;

			if (interFrameProgress > frames[currentFrame].duration)
			{
				interFrameProgress -= frames[currentFrame].duration;
				--currentFrame;
				if (currentFrame < 0 && repeats)
				{
					currentFrame = getFrameCount() - 1;
				}
				updateSpriteSubrect();
			}
		}
	}
	setPosition(x, y);
}

const Rect<int>& SpriteAnimation::getCurrentBounds() const
{
	return sprite.getCurrentBounds();
}

float SpriteAnimation::getAnimationSpeed() const
{
	return animationSpeed;
}

void SpriteAnimation::setAnimationSpeed(float speed)
{
	animationSpeed = speed;
}

bool SpriteAnimation::isDone() const
{
	return !isRepeating() && currentFrame >= getFrameCount();
}


void SpriteAnimation::setFrame(int frameIndex)
{
	currentFrame = frameIndex;
}

int SpriteAnimation::getFrameCount() const
{
	return frames.size();
}


void SpriteAnimation::setPosition(int x, int y)
{
	sprite.update(x, y);
}

void SpriteAnimation::setOrigin(const Vector2i& offset)
{
	sprite.setOrigin(offset);
}

void SpriteAnimation::serialize(Serialize& out) const
{
	serializeDrawable(out);
	sprite.serialize(out);
	AutoSerialize::serialize(out, frames);
	SAVEVAR(i, currentFrame);
	SAVEVAR(f, interFrameProgress);
	SAVEVAR(f, animationSpeed);
	SAVEVAR(b, repeats);
	SAVEVAR(i, cellWidth);
}

void SpriteAnimation::deserialize(Deserialize& in)
{
	deserializeDrawable(in);
	sprite.deserialize(in);
	AutoDeserialize::deserialize(in, frames);
	LOADVAR(i, currentFrame);
	LOADVAR(f, interFrameProgress);
	LOADVAR(f, animationSpeed);
	LOADVAR(b, repeats);
	LOADVAR(i, cellWidth);
}

std::string SpriteAnimation::serializationType() const
{
	return "SpriteAnimation";
}

bool SpriteAnimation::shouldSerialize() const
{
	return true;
}

/*	  Drawable private methods		*/
void SpriteAnimation::draw(sf::RenderTarget& target) const
{
	sprite.draw(target);
}


int SpriteAnimation::displayableFrame() const
{
	return frames[currentFrame < getFrameCount() ? currentFrame : getFrameCount() - 1].index;
}

void SpriteAnimation::updateSpriteSubrect()
{
	const int frameShown = cellWidth != sprite.getTexture().get().getSize().x ? displayableFrame() : 0;
	sprite.setRenderedSubrect(Rect<int>(frameShown * cellWidth, 0, cellWidth, sprite.getCurrentBounds().height));
}

} // sl
