/**
 * @section DESCRIPTION
 * A spritesheet as an animation - animates from left to right
 */
#ifndef SL_SPRITEANIMATION_HPP
#define SL_SPRITEANIMATION_HPP

#include <vector>

#include "Engine/Graphics/Animation.hpp"
#include "Engine/Graphics/ColorSwapID.hpp"
#include "Engine/Graphics/Framerate.hpp"
#include "Engine/Graphics/Sprite.hpp"

namespace sl
{

class Framerate;

class SpriteAnimation : public Animation
{
public:
	/**
	 * Constructs an Animation from a strip-formatted sprite filename
	 * @param fname The filename of the sprite strip
	 * @param cellWidth The width of each cell. eg there are sprite's width / cellWidth cells total
	 * @param framerate The number of frames to display each cell for
	 * @param repeats Whether the animation should start over from the first frame when it finishes
	 * @param swapID Any palette swapping that should be done (optional - if not specified, none is done)
	 */
	SpriteAnimation(const std::string& fname, int cellWidth, const Framerate& framerate, bool repeats, const ColorSwapID& swapID = ColorSwapID(), int offX = 0, int offY = 0, bool mirrored = false);
	
	/*	  Drawable public methods	 */
	void update(int x, int y) override;

	const Rect<int>& getCurrentBounds() const override;

	float getAnimationSpeed() const override;

	void setAnimationSpeed(float speed) override;

	/**
	 * Checks whether an animation is done
	 * @return true if done last frame and NOT repeating
	 */
	bool isDone() const override;

	/*    Animation public methods   */
	
	/**
	 * Resets an animation to the first frame
	 */
	void reset() override;

	/**
	 * Checks if the animation repeats or not
	 * @return true if repeats
	 */
	bool isRepeating() const override;

	void setProgress(float progress) override;

	float getProgress() const override;

	void rotate(float newAngle) override;

	void setColor(const sf::Color& color) override;

	void setFrame(int frameIndex) override;

	int getFrameCount() const override;

	void setPosition(int x, int y);

	void setOrigin(const Vector2i& offset);



	void serialize(Serialize& out) const override;

	void deserialize(Deserialize& in) override;

	std::string serializationType() const override;

	bool shouldSerialize() const override;

private:
	/*	  Drawable public methods	 */
	void draw(sf::RenderTarget& target) const override;

	int displayableFrame() const;

	void updateSpriteSubrect();


	//! The sprite that contains the current frame
	Sprite sprite;
	//!How long each frame takes
	std::vector<Framerate::Frame> frames;
	//!The current frame INDEX (into frames) the animation is on
	int currentFrame;
	//!How far through the current Drawable is in frames
	float interFrameProgress;
	//!How many frames per update the Animation plays at (1=normal)
	float animationSpeed;
	//!Whether the animation repeats
	bool repeats;
	//!The width of the frame in pixels
	int cellWidth;
};

} // sl

#endif // SL_SPRITEANIMATION_HPP
