#include "Engine/Graphics/SpriteFont.hpp"

#include "Engine/Graphics/TextureManager.hpp"
#include "Utility/Assert.hpp"

namespace sl
{

SpriteFont::SpriteFont(const std::string& fname, int charWidth, int charHeight, int vSpace, int hSpace, int specialCount, std::map<char, int>&& yOffsets)
	:charWidth(charWidth)
	,charHeight(charHeight)
	,vSpace(vSpace)
	,hSpace(hSpace)
	,specialCount(specialCount)
	,texture(TextureManager::getTexture("fonts/" + fname + ".png"))
	,sprite(texture.get())
	,tilesAcross(texture.get().getSize().x / charWidth)
	,yOffsets(yOffsets)
{
}


int SpriteFont::getSymbolWidth() const
{
	return charWidth + hSpace;
}

int SpriteFont::getSymbolHeight() const
{
	return charHeight + vSpace;
}

// private
const sf::Sprite& SpriteFont::getCharImpl(char c, const sf::Vector2f& pos) const
{
	if (c == '\t' || c == ' ') c = ' '; // @todo implement tabs properly
	const int index = static_cast<int>(static_cast<unsigned char>(c) - 32);
	ASSERT(index >= 0 && index <= (127 - 32) + specialCount);
	const int xOffset = (index % tilesAcross) * charWidth;
	const int yOffset = (index / tilesAcross) * charHeight;
	sprite.setTextureRect(sf::IntRect(xOffset, yOffset, charWidth, charHeight));
	auto it = yOffsets.find(c);
	sprite.setPosition(sf::Vector2f(pos.x, pos.y + (it == yOffsets.end() ? 0 : it->second) * sprite.getScale().y));
	return sprite;
}





//----------------------------- Writer impl -----------------------------------
SpriteFont::Writer::Writer(const SpriteFont& font, float scalar, const sf::Color& color)
	:font(font)
{
	font.sprite.setScale(scalar, scalar);
	font.sprite.setColor(color);
}


const sf::Sprite& SpriteFont::Writer::getChar(char c, const sf::Vector2f& pos) const
{
	return font.getCharImpl(c, pos);
}

//--------------------------- RainbowWriter impl ------------------------------

SpriteFont::RainbowWriter::RainbowWriter(const SpriteFont& font, float scalar)
	:font(font)
	,colours()
	,it()
{
	const sf::Color rainbowBasic[] = {
		sf::Color::Red,
		sf::Color::Yellow,
		sf::Color::Green,
		sf::Color::Blue,
		sf::Color::Magenta
	};
	for (const sf::Color& color : rainbowBasic)
	{
		colours.push_back(color);
	}
	it = colours.cbegin();
	font.sprite.setScale(scalar, scalar);
}


const sf::Sprite& SpriteFont::RainbowWriter::getChar(char c, const sf::Vector2f& pos) const
{
	if (c != ' ')
	{
		font.sprite.setColor(*it);
		if (++it == colours.cend())
		{
			it = colours.cbegin();
		}
	}
	return font.getCharImpl(c, pos);
}

} // sl
