#include "Engine/Graphics/SpriteText.hpp"

#include <algorithm>

#include "Engine/Graphics/SpriteFont.hpp"
#include "Engine/GUI/FontManager.hpp"
#include "Utility/Assert.hpp"
#include "Utility/Logger.hpp" 

namespace sl
{

static int renderTargets = 0;
SpriteText::~SpriteText()
{
	--renderTargets;
	Logger::log(Logger::Area::GUI, Logger::Priority::Debug) << "renderTargets = " << renderTargets << std::endl;
}

SpriteText::SpriteText(const std::string& fontName, const Rect<int>& box, int spaceGap, int scale, const std::string& str)
	:AbstractTextBox()
	,font(nullptr)
	,scale(scale)
	,color(sf::Color::White)
	,text()
	,bounds(box.left, box.top, 0, 0)
	,scaledBounds(bounds)
	,spaceGap(spaceGap)
	,symbolsAcross()
	,render(std::make_unique<sf::RenderTexture>())
	,sprite()
	,cursor(0, 0)
	,textureSize(box.width, box.height)
#ifdef SL_DEBUG
	,inputTextHadNewlines(false)
#endif //SL_DEBUG
{
	sprite.setScale(scale, scale);
	setFont(fontName);
	setArea(box.width, box.height);
	if (!str.empty())
	{
		setString(str);
	}
	++renderTargets;
	Logger::log(Logger::Area::GUI, Logger::Priority::Debug) << "renderTargets = " << renderTargets << std::endl;
}


void SpriteText::update(int x, int y)
{
	sprite.setPosition(x, y);
	bounds.left = x;
	bounds.top = y;
}

const Rect<int>& SpriteText::getCurrentBounds() const
{
	return scaledBounds;
}

void SpriteText::setString(const std::string& str)
{
	remakeRenderTexture();
	render->clear(sf::Color::Transparent);
	text.clear();
	cursor.x = 0;
	cursor.y = 0;
	appendString(insertNewlines(str));
}

void SpriteText::appendString(const std::string& str)
{
	ASSERT(render);
	const int sw = font->getSymbolWidth();
	const int sh = (font->getSymbolHeight() + spaceGap);
	SpriteFont::Writer writer(*font, 1, color);
	//SpriteFont::RainbowWriter writer(*font, 1);//, color);
	for (auto it = str.begin(); it != str.end(); ++it)
	{
		// also, to reduce bugs between the two, handle *ALL* newline insertion inside of insertNewlines()
		// but only when assertOnMissingNewline is enabled (which happens when you call insertNewlines())
		const char c = *it;
		if (c == '\t')
		{
			cursor.x += 4;
		}
		else if (c == '\n')
		{
#ifdef SL_DEBUG
			ASSERT(inputTextHadNewlines);
#endif // SL_DEBUG
			cursor.x = 0;
			++cursor.y;
		}
		else if (c == ' ')
		{
			++cursor.x;
		}
		else
		{
			render->draw(writer.getChar(c, sf::Vector2f(cursor.x * sw, cursor.y * sh)));
			++cursor.x;
		}
		if (cursor.x >= symbolsAcross)
		{
#ifdef SL_DEBUG
			ASSERT(!inputTextHadNewlines);
#endif // SL_DEBUG
			cursor.x = 0;
			++cursor.y;
		}
	}
	text += str;
	render->display();
	sprite.setTexture(render->getTexture());
	bounds.width = std::min(symbolsAcross, (int)text.size()) * sw;
	const int rows = ceilf(text.size() / (float)symbolsAcross);
	bounds.height = (rows * font->getSymbolHeight() + std::max(0, rows - 1) * spaceGap);
	scaledBounds = bounds * scale;
}

const std::string& SpriteText::getString() const
{
	return text;
}

bool SpriteText::backspace(int numCharsToRemove)
{
	ERROR("unimplemented");
	return false;
}

void SpriteText::setArea(int width, int height)
{
	textureSize.x = width;
	textureSize.y = height;
	remakeRenderTexture();
	ASSERT(render->create(width, height));
	sprite.setTexture(render->getTexture());

	if (width != bounds.width)
	{
		symbolsAcross = std::max(1, width / (font->getSymbolWidth()));
	}
	refresh();
}


void SpriteText::setFont(const std::string& fontName)
{
	font = gui::FontManager::getSpriteFont(fontName);
	ASSERT(font);
}

void SpriteText::setFontSize(int size)
{
	if (size != scale)
	{
		scale = size;
		sprite.setScale(size, size);
		setArea(textureSize.x / size, textureSize.y / size);
	}
}

void SpriteText::setFontColor(const sf::Color& color)
{
	this->color = color;
}

int SpriteText::getMinimumHeight() const
{
	return ceilf(text.size() / (float)symbolsAcross) * (font->getSymbolHeight() + spaceGap);
}


std::string SpriteText::insertNewlines(std::string str) const
{
#ifdef SL_DEBUG
	inputTextHadNewlines = true;
#endif // SL_DEBUG
	int cx = 0;
	auto lastSpace = str.end();
	// always call str.end() to avoid iterator invalidation
	for (auto it = str.begin(); it != str.end(); ++it)
	{
		if (*it == '\t')
		{
			cx += 4;
		}
		else if (*it == '\n')
		{
			cx = 0;
			lastSpace = str.end();
		}
		else if (*it == ' ')
		{
			lastSpace = it;
			++cx;
		}
		else
		{
			++cx;
		}

		if (cx >= symbolsAcross)
		{
			if (lastSpace != str.end())
			{
				cx = std::distance(lastSpace, it);
				*lastSpace = '\n';
			}
			else
			{
				// no worry about iterator invalidation here since we reset lastSpace at the end
				auto next = std::next(it);
				const bool split = next == str.end() || *next != ' ' && *next != '\t' && *next != '\n';
				// word is split in half
				// this should happen rarely, since most should hit the above trigger and just put a space before it.
				if (split)
				{
					ASSERT(it != str.begin());
					auto last = std::prev(it);
					const char insertSequence[2] = { '\n', *last };
					*last = '-';
					it = std::next(str.insert(it, std::begin(insertSequence), std::end(insertSequence)), 2);
				}
				else if (next != str.end() && (*next == ' ' || *next == '\n'))
				{
					// begins with a space (which is ugly). but if they have a tab, then I think we should respect that?
					// we can always change it if it's an issue and make this an else.
					// also newline should definitely be reomved to stop a double-newline.
					it = str.erase(next);
					if (it != str.end())
					{
						ASSERT(it != str.begin());
						--it;
					}
				}
			}
			lastSpace = str.end();
		}
	}
	return str;
}

void SpriteText::finalize()
{
	if (render)
	{
		texture.loadFromImage(render->getTexture().copyToImage());
		sprite.setTexture(texture);
		#ifdef SPRITETEXT_PRINT
		static int fuck = 0;
		texture.copyToImage().saveToFile("finalized texture " + std::to_string(fuck++) + ".png");
		#endif
		render.reset();
		--renderTargets;
		Logger::log(Logger::Area::GUI, Logger::Priority::Debug) << "[finalized()] renderTargets = " << renderTargets << std::endl;
	}
}

// private
void SpriteText::draw(sf::RenderTarget& target) const
{
	target.draw(sprite);
}

void SpriteText::refresh()
{
	//if (!text.empty())
	{
		std::string temp;
		std::swap(temp, text);
		setString(temp);
	}
}

void SpriteText::remakeRenderTexture()
{
	if (!render)
	{
		render = std::make_unique<sf::RenderTexture>();
		ASSERT(render->create(textureSize.x, textureSize.y));
		sprite.setTexture(render->getTexture());
		++renderTargets;
		Logger::log(Logger::Area::GUI, Logger::Priority::Debug) << "[reallocated] renderTargets = " << renderTargets << std::endl;
	}
}

} // sl
