#ifndef SL_TEXTURE_HPP
#define SL_TEXTURE_HPP

#include "Engine/Graphics/ColorSwapID.hpp"
#include "Engine/Graphics/TextureManager.hpp"
#include "Engine/Serialization/DeserializeConstructor.hpp"

namespace sl
{

class TextureManager;

class Texture
{
public:
	Texture(DeserializeConstructor);
	/**
	 * Constructs a reference-counted Texture. Called by TextureManager only.
	 * @param owner The TextureManager this belongs to
	 * @param id The filename id of the texture
	 * @param textureData The actual texture data
	 */
	Texture(TextureManager *owner, const std::string& id, const sf::Texture *textureData);
	
	/**
	 * Constructs a reference-counted colour-swapped Texture. Called by TextureManager only.
	 * @param owner The TextureManager this belongs to
	 * @param id the filename of the texture
	 * @param swapID The colour swap identifier used to create this texture
	 * @param textureData The actual texture data
	 */
	Texture(TextureManager *owner, const std::string& id, const ColorSwapID& swapID, const sf::Texture *textureData);
	
	/**
	 * Copy constructor for Texture. (increases reference)
	 * @param copy The Texture to copy
	 */
	Texture(const Texture& copy);
	
	/**
	 * Destructor (decreases reference)
	 */
	~Texture();

	/**
	 * Assigns another texture to this one.
	 * (decreases this texture's reference and increases the new texture's one)
	 * @param copy The new texture
	 * @return A reference to this texture
	 */
	const Texture& operator=(const Texture& copy);

	/**
	 * Accesses the internal sf::Texture data in a read-only way
	 * @return A reference to the internal sf:Texture
	 */
	const sf::Texture& get() const;

	const std::string& filename() const;

private:

	//!The TextureManager that owns this Texture's texture data
	TextureManager *owner;
	//!The filename id of the texture
	std::string id;
	//!The colour-swap applied to the texture
	ColorSwapID swapID;
	//!Whether a colour-swap was applied
	bool isAColorSwap;
	//!The actual texture data
	const sf::Texture *textureData;
};

} // sl

#endif // SL_TEXTURE_HPP
