#include "Engine/Graphics/TextureManager.hpp"

#include "Utility/Assert.hpp"
#include "Utility/Logger.hpp"

namespace sl
{

TextureManager::TextureManager()
	:textureDirectory("resources/img/")
{
	instance = this;
}

TextureManager::~TextureManager()
{
	bool unloadedTextures = false;
	for (std::map<std::string, TextureBank>::iterator it = textures.begin(); it != textures.end(); ++it)
	{
		//ASSERT(false);  //  I don't actually think this sould happen...
#ifdef SL_DEBUG
		//Logger::log(Logger::Graphics, Logger::UltraSerious) << "[THIS SHOULDN'T HAPPEN]Unloading: " << it->first << std::endl;
#endif
		delete std::get<0>(it->second);
		for (ColorSwaps::iterator tbi = std::get<2>(it->second).begin(); tbi != std::get<2>(it->second).end(); ++tbi)
		{
			delete tbi->second.first;
		}
		unloadedTextures = true;
	}
	//ASSERT(unloadedTextures == false);
	instance = nullptr;
}

void TextureManager::init()
{
	if (!instance)
	{
		instance = new TextureManager();
	}
}

void TextureManager::destroy()
{
	if (instance)
	{
		delete instance;
		instance = nullptr;
	}
}

Texture TextureManager::getTexture(const std::string& file)
{
	std::map<std::string, TextureBank>::iterator it = instance->textures.find(file);
	if (it == instance->textures.end())
	{
		sf::Texture *tex = new sf::Texture();
		if (!tex->loadFromFile(getDirectory() + file))
		{
			Logger::log(Logger::Graphics, Logger::Serious) << "ERROR: Couldn't load resources/img/" << file << std::endl;
		}
		instance->textures.insert(std::make_pair(file, std::make_tuple(tex, 0, ColorSwaps())));
#ifdef SL_DEBUG
		Logger::log(Logger::Graphics, Logger::Debug) << "Loading: " << file << std::endl;
#endif
		return Texture(instance, file, tex);
	}
	else
	{
		return Texture(instance, file, std::get<0>(it->second));
	}
}

Texture TextureManager::getTexture(const std::string& file, const ColorSwapID& swapID)
{
	if (swapID) // check if swap is empty
	{
		std::map<std::string, TextureBank>::iterator it = instance->textures.find(file);
		if (it == instance->textures.end())
		{
			sf::Texture *tex = new sf::Texture();
			if (!tex->loadFromFile(getDirectory() + file))
			{
				Logger::log(Logger::Graphics, Logger::Serious) << "ERROR: Couldn't load resources/img/" << file << std::endl;
			}
			it = instance->textures.insert(std::make_pair(file, std::make_tuple(tex, 0, ColorSwaps()))).first;
#ifdef SL_DEBUG
			Logger::log(Logger::Graphics, Logger::Debug) << "Loading: " << file << std::endl;
#endif
		}
		ColorSwaps::iterator it2 = std::get<2>(it->second).find(swapID);
		if (it2 == std::get<2>(it->second).end())
		{
			/*sf::Image image(std::get<0>(it->second)->copyToImage());
			int totalPixels = image.getSize().x * image.getSize().y;
			sf::Uint8 *pixels = new sf::Uint8[totalPixels];
			sf::Uint8 *pit = pixels;
			std::vector<std::pair<sf::Color, sf::Color> > colors;
			swapID.getColors(colors);
			for (int i = 0; i < totalPixels; ++i, pit += 4)
			{
				if (*(pit + 3))	//	if pixel isn't transparent
				{
					for (auto c = colors.begin(); c != colors.end(); ++c)
					{
						if (*pit == c->first.r && *(pit + 1) == c->first.g && *(pit + 2) == c->first.b)
						{
							*pit = c->second.r;
							*(pit + 1) = c->second.g;
							*(pit + 2) = c->second.b;
							break;
						}
					}
				}
			}
			sf::Texture *tex = new sf::Texture();
			tex->create(image.getSize().x, image.getSize().y);
			tex->update(pixels);
			delete[] pixels*/

			sf::Image image(std::get<0>(it->second)->copyToImage());

			sf::Vector2u dimensions(image.getSize());

			std::vector<std::pair<sf::Color, sf::Color> > colors;
			swapID.getColors(colors);

			for (unsigned int x = 0; x < dimensions.x; ++x)
			{
				for (unsigned int y = 0; y < dimensions.y; ++y)
				{
					sf::Color colour = image.getPixel(x, y);
					if (colour.a)
					{
						for (auto it = colors.begin(); it != colors.end(); ++it)
						{
							if (colour == it->first)
							{
								image.setPixel(x, y, it->second);
							}
						}
					}
				}
			}

			sf::Texture *tex = new sf::Texture();
			tex->loadFromImage(image);
			it2 = std::get<2>(it->second).insert(std::make_pair(swapID, std::make_pair(tex, 0))).first;
#ifdef SL_DEBUG
			Logger::log(Logger::Graphics) << "Creating swap " << swapID << " of " << file << std::endl;
#endif
		}
		return Texture(instance, file, swapID, it2->second.first);
	}
	else // the swap is empty
	{
		return getTexture(file);
	}
}

void TextureManager::addReference(const std::string& id)
{
	std::map<std::string, TextureBank>::iterator it = instance->textures.find(id);
	ASSERT(it != instance->textures.end());   // God knows how the fuck this happened
	++std::get<1>(it->second);
	//std::cout << "Increased reference for " << id << " to " << std::get<1>(it->second) << std::endl;
}

void TextureManager::removeReference(const std::string& id)
{
	std::map<std::string, TextureBank>::iterator it = instance->textures.find(id);
	ASSERT(it != instance->textures.end());   // God knows how the fuck this happened
	if (--std::get<1>(it->second) <= 0)
	{
		delete std::get<0>(it->second);
		textures.erase(id);
#ifdef SL_DEBUG
		Logger::log(Logger::Graphics, Logger::Debug) << "Unloading: " << id << std::endl;
#endif
	}
	//std::cout << "Decreased reference for " << id << " to " << std::get<1>(it->second) << std::endl;
}

void TextureManager::addReference(const std::string& id, const ColorSwapID& color)
{
	std::map<std::string, TextureBank>::iterator it = instance->textures.find(id);
	ASSERT(it != instance->textures.end());   // God knows how the fuck this happened
	ColorSwaps::iterator it2 = std::get<2>(it->second).find(color);
	ASSERT(it2 != std::get<2>(it->second).end());   //  Not even God knows about this one...
	++it2->second.second;
	//std::cout << "Increasing reference for swap " << color << " on " << id << " to " << it2->second.second << std::endl;
}

void TextureManager::removeReference(const std::string& id, const ColorSwapID& color)
{
	std::map<std::string, TextureBank>::iterator it = instance->textures.find(id);
	ASSERT(it != instance->textures.end());   // God knows how the fuck this happened
	ColorSwaps::iterator it2 = std::get<2>(it->second).find(color);
	ASSERT(it2 != std::get<2>(it->second).end());   //  Not even God knows about this one...
	if (--it2->second.second <= 0)
	{
		delete it2->second.first;
		std::get<2>(it->second).erase(color);
#ifdef SL_DEBUG
		//std::cout << "\nUnloading: " << color << " swap on " << id << std::endl << std::endl;
#endif
	}
	//std::cout << "Decreasing reference for swap " << color << " on " << id << " to " << it2->second.second << std::endl;
}

const std::string& TextureManager::getDirectory()
{
	return instance->textureDirectory;
}

} // sl
