/**
 * @section DESCRIPTION
 * This class manages sf::Texture objects so that only one sf::Texture exists per image file to reduce redundant resource usage.
 * Call init() to initialize this class before using it.
 */
#ifndef SL_TEXTUREMANAGER_HPP
#define SL_TEXTUREMANAGER_HPP

#include <SFML/Graphics.hpp>
#include <map>
#include <tuple>
#include "Engine/Graphics/Texture.hpp"
#include "Engine/Graphics/ColorSwapID.hpp"

namespace sl
{

class Texture;

class TextureManager
{
public:
	/**
	 * Gets a Texture from the manager
	 * @param file The filename of the image
	 * @return A reference-counting Texture class containing the texture
	 */
	static Texture getTexture(const std::string& file);
	/**
	 * Gets a colour-swapped Texture from the manager
	 * @param file The filename of the image
	 * @param swapID The colour-swap to be done to the texture
	 * @return A referencing-counting Texture class containing the colour-swapped texture
	 */
	static Texture getTexture(const std::string& file, const ColorSwapID& swapID);
	/**
	 * Initializes TextureManager
	 */
	static void init();
	/**
	 * Destroys TextureManager and performs garbage cleanup on all images
	 */
	static void destroy();
	/**
	 * Increases the reference count for a given texture. Called by Texture
	 * @param id The texture name of the texture
	 */
	void addReference(const std::string& id);
	 /**
	 * Reduces the reference count for a given texture. Called by Texture
	 * @param id The texture name of the texture
	 */
	void removeReference(const std::string& id);
	/**
	 * Increases the reference count for a given colour-swapped texture. Called by Texture
	 * @param id The filename
	 */
	void addReference(const std::string& id, const ColorSwapID& color);
	/**
	 * Reduces the reference count for a given colour-swapped Texture. Called by Texture
	 * @param id The filename
	 * @param color The colour-swap identifier
	 */
	void removeReference(const std::string& id, const ColorSwapID& color);

	/**
	 * @return The directory from which all textures are loaded from
	 */
	static const std::string& getDirectory();

private:
	//  first = Texture, second = how many other places currently reference it
	typedef std::pair<sf::Texture*, int> RefCountedTex;
	//  first = id of colour-swap, second = colour-swapped texture
	typedef std::map<ColorSwapID, RefCountedTex> ColorSwaps;
	//  first = Texture, second = how many reference there are to it, third = map of colour-swapped versions
	typedef std::tuple<sf::Texture*, int, ColorSwaps> TextureBank;
	/**
	 * Constructs the Manager
	 */
	TextureManager();
	/**
	 * Destructs the Manager
	 */
	~TextureManager();

	//!A hashmap of textures
	std::map<std::string, TextureBank> textures;
	//!Directory to load images from
	std::string textureDirectory;



	//!The Manager
	static TextureManager *instance;
};

}// End of sl namespace

#endif

