#include "Engine/Graphics/TileGrid.hpp"

#include <SFML/Graphics/RenderTarget.hpp>
#include <SFML/Graphics/Sprite.hpp>
#include <SFML/Graphics/Texture.hpp>

#include "Utility/Assert.hpp"

namespace sl
{

TileGrid::TileGrid(int xOffset, int yOffset, int width, int height, int cellSizeX, int cellSizeY)
{
	this->set(xOffset, yOffset, width, height, cellSizeX, cellSizeY);
	this->setDepth(999999);//	default it to below pretty much everything
}


/*	  Drawable public methods		*/
void TileGrid::update(int x, int y)
{
	const int diffX = x - left;
	const int diffY = y - top;
	for (sf::Sprite& tile : tiles)
	{
		tile.move(diffX, diffY);
	}
	left = x;
	top = y;

	this->recalculateVisibleTiles();
}

const Rect<int>& TileGrid::getCurrentBounds() const
{
	return visibleTilesByPixels;
}


/*	  TileGrid public methods		*/
void TileGrid::setTexture(int x, int y, const sf::Texture& texture)
{
	tiles(x, y).setTexture(texture);
}

void TileGrid::setTexture(int x, int y, const sf::Texture& texture, const sf::IntRect& textureRect)
{
	tiles(x, y).setTexture(texture);
	tiles(x, y).setTextureRect(textureRect);
	tiles(x, y).setPosition(left + x * cellSizeX, top + y * cellSizeY);
}

void TileGrid::resetTextures()
{
	for (sf::Sprite& tile : tiles)
	{
		tile = sf::Sprite();
	}
}

void TileGrid::setVisibleArea(const Rect<int>& bBox)
{
	if (this->bBox != bBox)
	{
		this->bBox = bBox;
		this->recalculateVisibleTiles();
	}
}

void TileGrid::set(int xOffset, int yOffset, int width, int height, int cellSizeX, int cellSizeY)
{
	this->left = xOffset;
	this->top = yOffset;
	this->cellSizeX = cellSizeX;
	this->cellSizeY = cellSizeY;
	this->bBox = Rect<int>(xOffset, yOffset, width * cellSizeX, height * cellSizeY);
	tiles = Grid<sf::Sprite>(width, height);
	for (int y = 0; y < height; ++y)
	{
		for (int x = 0; x < width; ++x)
		{
			tiles(x, y).setPosition(xOffset + x * cellSizeX, yOffset + y * cellSizeY);
		}
	}
	this->recalculateVisibleTiles();
}

/*	  Drawable private methods		*/
void TileGrid::draw(sf::RenderTarget& target) const
{

	//  Actually draw stuff
	for (int j = visibleTilesByIndices.top; j < visibleTilesByIndices.height; ++j)
	{
		for (int i = visibleTilesByIndices.left; i < visibleTilesByIndices.width; ++i)
		{
			target.draw(tiles(i, j));
		}
	}
}

/*	  TileGrid private methods		*/
void TileGrid::recalculateVisibleTiles()
{
	visibleTilesByIndices.left = 0;
	visibleTilesByIndices.top = 0;
	visibleTilesByIndices.width = tiles.getWidth();
	visibleTilesByIndices.height = tiles.getHeight();

	//  If the new minimum/maxiumums are valid (not outside the grid) then update them
	if (bBox.left - left > 0)
	{
		visibleTilesByIndices.left = (bBox.left - left) / cellSizeX;
	}

	if ((bBox.left + bBox.width - left) / cellSizeX < tiles.getWidth())
	{
		visibleTilesByIndices.width = (bBox.left + bBox.width - left) / cellSizeX + 1;
	}

	if (bBox.top - top > 0)
	{
		visibleTilesByIndices.top = (bBox.top - top) / cellSizeY;
	}

	if ((bBox.top + bBox.height - top) / cellSizeY < tiles.getHeight())
	{
		visibleTilesByIndices.height = (bBox.top + bBox.height - top) / cellSizeY + 1;
	}

	visibleTilesByPixels.left = left + visibleTilesByIndices.left * cellSizeX;
	visibleTilesByPixels.top = top + visibleTilesByIndices.top * cellSizeY;
	visibleTilesByPixels.width = visibleTilesByIndices.width * cellSizeX;
	visibleTilesByPixels.height = visibleTilesByIndices.height * cellSizeY;
}

} // sl
