#include "Engine/Input/Controller.hpp"

#include "Engine/Input/ControllerBinds.hpp"
#include "Engine/Input/InputHandle.hpp"
#include "Utility/Assert.hpp"
#include "Utility/Logger.hpp"

namespace sl
{

Controller::Controller(const InputHandle& input, const ControllerBinds& binds, unsigned int gamepadId)
	:input(&input)
	,binds(&binds)
	,gamepadId(gamepadId)
{
}


//void Controller::setInputHandle(const InputHandle& input)
//{
//	this->input = &input;
//}
//
//void Controller::unsetInputHandle()
//{
//	this->input = nullptr;
//}


bool Controller::pressed(Key key) const
{
	if (!input)
	{
		return false;
	}
	auto keys = binds->getBinds(key);
	if (keys)
	{
		for (const KeyBind& bind : *keys)
		{
			if (checkPressed(bind))
			{
				return true;
			}
		}
	}
	return false;
}

bool Controller::released(Key key) const
{
	if (!input)
	{
		return false;
	}
	auto keys = binds->getBinds(key);
	if (keys)
	{
		for (const KeyBind& bind : *keys)
		{
			if (checkReleased(bind))
			{
				return true;
			}
		}
	}
	return false;
}

bool Controller::held(Key key) const
{
	if (!input)
	{
		return false;
	}
	auto keys = binds->getBinds(key);
	if (keys)
	{
		for (const KeyBind& bind : *keys)
		{
			if (checkHeld(bind))
			{
				return true;
			}
		}
	}
	return false;
}

void Controller::setJoystickID(unsigned int id)
{
	gamepadId = id;
}

float Controller::axis(Input::Axis axis) const
{
	if (input)
	{
		return input->axis(gamepadId, axis);
	}
	// print warning in warning log when I implement that probably...
	return 0.f;
}

std::unique_ptr<KeyBind> Controller::getJustPressedUnboundKey() const
{
	std::unique_ptr<KeyBind> freeBind;
	if (input)
	{
		std::vector<KeyBind> justPressed = input->getJustPressed(gamepadId);
		for (const KeyBind& keyBind : justPressed)
		{
			if (binds->isBound(keyBind))
			{
				freeBind = std::make_unique<KeyBind>(keyBind);
				break;
			}
		}
	}
	return freeBind;
}

CustomAction Controller::customActionPressed() const
{
	for (const std::pair<KeyBind, CustomAction>& customAction : binds->getCustomActions())
	{
		if (checkPressed(customAction.first))
		{
			return customAction.second;
		}
	}
	return NO_CUSTOM_ACTION;
}

// private
bool Controller::checkPressed(const KeyBind& bind) const
{
	switch (bind.type())
	{
	case KeyBind::DeviceType::Keyboard:
		return input->isKeyPressed(static_cast<sf::Keyboard::Key>(bind.code()));
	case KeyBind::DeviceType::Mouse:
		return input->isMousePressed(static_cast<sf::Mouse::Button>(bind.code()));
	case KeyBind::DeviceType::GamepadButton:
		return input->isGamepadButtonPressed(gamepadId, bind.code());
	case KeyBind::DeviceType::GamepadAxis:
		return input->isGamepadAxisPressed(gamepadId, static_cast<Input::AxisButton>(bind.code()));
	}
	ERROR("should not hit");
	return false;
}

bool Controller::checkReleased(const KeyBind& bind) const
{
	switch (bind.type())
	{
	case KeyBind::DeviceType::Keyboard:
		return input->isKeyReleased(static_cast<sf::Keyboard::Key>(bind.code()));
	case KeyBind::DeviceType::Mouse:
		return input->isMouseReleased(static_cast<sf::Mouse::Button>(bind.code()));
	case KeyBind::DeviceType::GamepadButton:
		return input->isGamepadButtonReleased(gamepadId, bind.code());
	case KeyBind::DeviceType::GamepadAxis:
		return input->isGamepadAxisReleased(gamepadId, static_cast<Input::AxisButton>(bind.code()));
	}
	ERROR("should not hit");
	return false;
}

bool Controller::checkHeld(const KeyBind& bind) const
{
	switch (bind.type())
	{
	case KeyBind::DeviceType::Keyboard:
		return input->isKeyHeld(static_cast<sf::Keyboard::Key>(bind.code()));
	case KeyBind::DeviceType::Mouse:
		return input->isMouseHeld((sf::Mouse::Button) bind.code());
	case KeyBind::DeviceType::GamepadButton:
		return input->isGamepadButtonHeld(gamepadId, bind.code());
	case KeyBind::DeviceType::GamepadAxis:
		return input->isGamepadAxisHeld(gamepadId, static_cast<Input::AxisButton>(bind.code()));
	}
	ERROR("should not hit");
	return false;
}

} // sl
