#include "Engine/Input/ControllerBinds.hpp"

#include "Utility/Assert.hpp"
#include "Utility/Logger.hpp"

namespace sl
{

void ControllerBinds::addKeybind(Key key, KeyBind bind)
{
	++regularBindCount[bind];
	keybinds[key].push_back(bind);
}


void ControllerBinds::clearKeybinds(Key key)
{
	auto keybind = keybinds.find(key);
	if (keybind != keybinds.end())
	{
		for (const KeyBind& bind : keybind->second)
		{
			auto it = regularBindCount.find(bind);
			ASSERT(it != regularBindCount.end());
			if (--it->second <= 0)
			{
				regularBindCount.erase(it);
			}
		}
		keybinds.erase(keybind);
	}
}

void ControllerBinds::clearKeybinds()
{
	for (const auto& keybind : keybinds)
	{
		for (const KeyBind& bind : keybind.second)
		{
			auto it = regularBindCount.find(bind);
			ASSERT(it != regularBindCount.end());
			if (--it->second <= 0)
			{
				regularBindCount.erase(it);
			}
		}
	}
}

void ControllerBinds::registerCustomAction(CustomAction action, KeyBind bind)
{
	ASSERT(action != NO_CUSTOM_ACTION);
	ASSERT(regularBindCount.count(bind) == 0);
	customActions[bind] = action;
}

void ControllerBinds::removeCustomActionKeybind(KeyBind bind)
{
	auto it = customActions.find(bind);
	if (it != customActions.end())
	{
		customActions.erase(it);
	}
}

bool ControllerBinds::isBound(KeyBind keyBind) const
{
	return regularBindCount.count(keyBind) == 0;
}

const std::vector<KeyBind>* ControllerBinds::getBinds(Key key) const
{
	auto it = keybinds.find(key);
	return it != keybinds.end() ? &(it->second) : nullptr;
}

const std::map<KeyBind, CustomAction>& ControllerBinds::getCustomActions() const
{
	return customActions;
}

std::string ControllerBinds::getBindsString(Key key, bool excludeDisconnectedGamepads) const
{
	auto binds = keybinds.find(key);
	if (binds != keybinds.end())
	{
		ASSERT(!binds->second.empty());
		std::stringstream ss;
		bool first = true;
		for (const KeyBind& bind : binds->second)
		{
			// TODO: figure out what to do for non-0 gamepads (multiple gamepads)
			if (!excludeDisconnectedGamepads || (bind.type() != KeyBind::DeviceType::GamepadAxis && bind.type() != KeyBind::DeviceType::GamepadButton) || sf::Joystick::isConnected(0))
			{
				if (first)
				{
					first = false;
				}
				else
				{
					ss << ", ";
				}
				ss << bind.string();
			}
		}
		return ss.str();
	}
	return "unassigned";
}

} // sl