#include "Engine/Input/Input.hpp"

#include "Engine/Input/KeyBind.hpp"
#include "Utility/Assert.hpp"
#include "Utility/JsonUtil.hpp"
#include "Utility/Logger.hpp"

#include <algorithm>
#include <iostream>

namespace sl
{

/*static*/Input::AxisButton Input::fromSF(sf::Joystick::Axis sfAxis, bool negative)
{
	if (!negative)
	{
		switch (sfAxis)
		{
		case sf::Joystick::Axis::X:
			return AxisButton::X;
		case sf::Joystick::Axis::Y:
			return AxisButton::Y;
		case sf::Joystick::Axis::Z:
			return AxisButton::Z;
		case sf::Joystick::Axis::R:
			return AxisButton::R;
		case sf::Joystick::Axis::U:
			return AxisButton::U;
		case sf::Joystick::Axis::V:
			return AxisButton::V;
		case sf::Joystick::Axis::PovX:
			return AxisButton::PovX;
		case sf::Joystick::Axis::PovY:
			return AxisButton::PovY;
		}
	}
	else
	{
		switch (sfAxis)
		{
		case sf::Joystick::Axis::X:
			return AxisButton::X_Neg;
		case sf::Joystick::Axis::Y:
			return AxisButton::Y_Neg;
		case sf::Joystick::Axis::Z:
			return AxisButton::Z_Neg;
		case sf::Joystick::Axis::R:
			return AxisButton::R_Neg;
		case sf::Joystick::Axis::U:
			return AxisButton::U_Neg;
		case sf::Joystick::Axis::V:
			return AxisButton::V_Neg;
		case sf::Joystick::Axis::PovX:
			return AxisButton::PovX_Neg;
		case sf::Joystick::Axis::PovY:
			return AxisButton::PovY_Neg;
		}
	}
	// stupid warnings
	ERROR("Actually unreachable");
	return AxisButton::X;
}

Input::Input()
	:window(nullptr)
	,focused(true)
	,enabled(true)
{
	std::fill(std::begin(mouseButtonUp), std::end(mouseButtonUp), 0);
	std::fill(std::begin(mouseButtonDown), std::end(mouseButtonDown), 0);
	std::fill(std::begin(keyUp), std::end(keyUp), 0);
	std::fill(std::begin(keyDown), std::end(keyDown), 0);
	for (int joystick = 0; joystick < sf::Joystick::Count; ++joystick)
	{
		std::fill(std::begin(gamepadButtonUp[joystick]), std::end(gamepadButtonUp[joystick]), 0);
		std::fill(std::begin(gamepadButtonDown[joystick]), std::end(gamepadButtonDown[joystick]), 0);
		std::fill(std::begin(gamepadAxisUp[joystick]), std::end(gamepadAxisUp[joystick]), 0);
		std::fill(std::begin(gamepadAxisDown[joystick]), std::end(gamepadAxisDown[joystick]), 0);
	}
}

void Input::update()
{
	for (int key = 0; key < sf::Keyboard::KeyCount; ++key)
	{
		if (sf::Keyboard::isKeyPressed((sf::Keyboard::Key) key))
		{
			keyDown[key] = 2;
			if (keyUp[key] > 0)
			{
				--keyUp[key];
			}
		}
		else
		{
			keyUp[key] = 2;
			if (keyDown[key] > 0)
			{
				--keyDown[key];
			}
		}
	}
	for (int button = 0; button < sf::Mouse::ButtonCount; ++button)
	{
		if (sf::Mouse::isButtonPressed((sf::Mouse::Button) button))
		{
			mouseButtonDown[button] = 2;
			if (mouseButtonUp[button] > 0)
			{
				--mouseButtonUp[button];
			}
		}
		else
		{
			mouseButtonUp[button] = 2;
			if (mouseButtonDown[button] > 0)
			{
				--mouseButtonDown[button];
			}
		}
	}
	for (int gamepad = 0; gamepad < sf::Joystick::Count; ++gamepad)
	{
		if (sf::Joystick::isConnected(gamepad))
		{
			for (int button = 0; button < sf::Joystick::ButtonCount; ++button)
			{
				if (sf::Joystick::isButtonPressed(gamepad, button))
				{
					gamepadButtonDown[gamepad][button] = 2;
					if (gamepadButtonUp[gamepad][button] > 0)
					{
						--gamepadButtonUp[gamepad][button];
					}
				}
				else
				{
					gamepadButtonUp[gamepad][button] = 2;
					if (gamepadButtonDown[gamepad][button] > 0)
					{
						--gamepadButtonDown[gamepad][button];
					}
				}
			}
			for (int axis = 0; axis < sf::Joystick::AxisCount; ++axis)
			{
				if (sf::Joystick::hasAxis(gamepad, static_cast<sf::Joystick::Axis>(axis)))
				{
					constexpr float directionalModifier[2] = {
						1.f / 100.f,
						-1.f / 100.f
					};
					for (int direction = 0; direction < 2; ++direction)
					{
						const int index = (direction * sf::Joystick::AxisCount) + axis;
						if (directionalModifier[direction] * sf::Joystick::getAxisPosition(gamepad, static_cast<Input::Axis>(axis)) >= gamepadAxisThreshold)
						{
							//const char *axisStr[sf::Joystick::AxisCount] = { "X", "Y", "Z", "R", "U", "V", "PovX", "PovY" };
							//std::cout << "Axis [" << axisStr[axis] << " = " << sf::Joystick::getAxisPosition(gamepad, static_cast<Input::Axis>(axis)) << std::endl;
							gamepadAxisDown[gamepad][index] = 2;
							if (gamepadAxisUp[gamepad][index] > 0)
							{
								--gamepadAxisUp[gamepad][index];
							}
						}
						else
						{
							gamepadAxisUp[gamepad][index] = 2;
							if (gamepadAxisDown[gamepad][index] > 0)
							{
								--gamepadAxisDown[gamepad][index];
							}
						}
					}
				}
			}
		}
	}

	if (window)
	{
		currentMousePosI = sf::Mouse::getPosition(*window);
		currentMousePosF.x = currentMousePosI.x;
		currentMousePosF.y = currentMousePosI.y;

		currentMousePosTransformedF = window->mapPixelToCoords(currentMousePosI);
		currentMousePosTransformedI.x = currentMousePosTransformedF.x;
		currentMousePosTransformedI.y = currentMousePosTransformedF.y;
	}
}

void Input::setFocus(bool focus)
{
	focused = focus;
	Logger::log(Logger::Core, Logger::Debug) << "focused = " << (focused ? "True" : "False") << std::endl;
}

void Input::setWindow(const sf::RenderWindow& window)
{
	this->window = &window;
}

void Input::enable()
{
	enabled = true;
}

void Input::disable()
{
	enabled = false;
}

bool Input::isEnabled() const
{
	return enabled;
}

/*			keyboard			*/
bool Input::isKeyPressed(KeyboardKey key) const
{
	return enabled && focused && keyUp[key] == 1;
}

bool Input::isKeyReleased(KeyboardKey key) const
{
	return enabled && focused && keyDown[key] == 1;
}

bool Input::isKeyHeld(KeyboardKey key) const
{
	return enabled && focused && keyDown[key] == 2;
}

/*			mouse				*/
bool Input::isMousePressed(MouseButton button) const
{
	return enabled && focused && mouseButtonUp[button] == 1;
}

bool Input::isMouseReleased(MouseButton button) const
{
	return enabled && focused && mouseButtonDown[button] == 1;
}

bool Input::isMouseHeld(MouseButton button) const
{
	return enabled && focused && mouseButtonDown[button] == 2;
}

/*			gamepad			*/
bool Input::isGamepadButtonPressed(unsigned int gamepadId, GamepadButton button) const
{
	return enabled && focused && gamepadButtonUp[gamepadId][button] == 1;
}

bool Input::isGamepadButtonReleased(unsigned int gamepadId, GamepadButton button) const
{
	return enabled && focused && gamepadButtonDown[gamepadId][button] == 1;
}

bool Input::isGamepadButtonHeld(unsigned int gamepadId, GamepadButton button) const
{
	return enabled && focused && gamepadButtonDown[gamepadId][button] == 2;
}

bool Input::isGamepadAxisPressed(unsigned int gamepadId, AxisButton axis) const
{
	return enabled && focused && gamepadAxisUp[gamepadId][static_cast<int>(axis)] == 1;
}
bool Input::isGamepadAxisReleased(unsigned int gamepadId, AxisButton axis) const
{
	return enabled && focused && gamepadAxisDown[gamepadId][static_cast<int>(axis)] == 1;
}
bool Input::isGamepadAxisHeld(unsigned int gamepadId, AxisButton axis) const
{
	return enabled && focused && gamepadAxisDown[gamepadId][static_cast<int>(axis)] == 2;
}

float Input::axis(int gamepadId, Input::Axis axis) const
{
	if (enabled && focused)
	{
		return sf::Joystick::getAxisPosition(gamepadId, axis) / 100.f;
	}
	return 0.f;
}

const sf::Vector2i& Input::mousePos() const
{
	return currentMousePosTransformedI;
}

const sf::Vector2f& Input::mousePosF() const
{
	return currentMousePosTransformedF;
}

const sf::Vector2i& Input::mousePosAbsolute() const
{
	return currentMousePosI;
}

const sf::Vector2f& Input::mousePosAbsoluteF() const
{
	return currentMousePosF;
}

std::vector<std::string> Input::buttonNames;
std::vector<Input::AxisNames> Input::axesNames;

/*static*/std::string Input::name(KeyboardKey key)
{
	switch (key)
	{
	case KeyboardKey::Escape:
		return "Esc";
	case KeyboardKey::LControl:
		return "Ctrl";
	case KeyboardKey::LShift:
		return "Shift";
	case KeyboardKey::LAlt:
		return "Alt";
	case KeyboardKey::LSystem:
		return "Sys";
	case KeyboardKey::RControl:
		return "Ctrl (R)";
	case KeyboardKey::RShift:
		return "Shift (R)";
	case KeyboardKey::RAlt:
		return "Alt (R)";
	case KeyboardKey::RSystem:
		return "Sys (R)";
	case KeyboardKey::Menu:
		return "Menu";
	case KeyboardKey::LBracket:
		return "[";
	case KeyboardKey::RBracket:
		return "]";
	case KeyboardKey::SemiColon:
		return ";";
	case KeyboardKey::Comma:
		return ",";
	case KeyboardKey::Period:
		return ".";
	case KeyboardKey::Quote:
		return "'";
	case KeyboardKey::Slash:
		return "/";
	case KeyboardKey::BackSlash:
		return "\\";
	case KeyboardKey::Tilde:
		return "~";
	case KeyboardKey::Equal:
		return "=";
	case KeyboardKey::Dash:
		return "-";
	case KeyboardKey::Space:
		return "Space";
	case KeyboardKey::Return:
		return "Enter";
	case KeyboardKey::BackSpace:
		return "Backspace";
	case KeyboardKey::Tab:
		return "Tab";
	case KeyboardKey::PageUp:
		return "Page Up";
	case KeyboardKey::PageDown:
		return "Page Down";
	case KeyboardKey::End:
		return "End";
	case KeyboardKey::Home:
		return "Home";
	case KeyboardKey::Insert:
		return "Insert";
	case KeyboardKey::Delete:
		return "Del";
	case KeyboardKey::Add:
		return "+";
	case KeyboardKey::Subtract:
		return "-";
	case KeyboardKey::Multiply:
		return "*";
	case KeyboardKey::Divide:
		return "/";
	case KeyboardKey::Left:
		return "Left";
	case KeyboardKey::Right:
		return "Right";
	case KeyboardKey::Up:
		return "Up";
	case KeyboardKey::Down:
		return "Down";
	case KeyboardKey::Pause:
		return "Pause";
	default:
		break;
	}
	if (key >= KeyboardKey::A && key <= KeyboardKey::Z)
	{
		return std::string(1, static_cast<char>('A' + (key - KeyboardKey::A)));
	}
	else if (key >= KeyboardKey::Num0 && key <= KeyboardKey::Num9)
	{
		return std::string(1, static_cast<char>('0' + (key - KeyboardKey::Num0)));
	}
	else if (key >= KeyboardKey::Numpad0 && key <= KeyboardKey::Numpad9)
	{
		return std::string(1, static_cast<char>('0' + (key - KeyboardKey::Numpad0)));
	}
	else if (key >= KeyboardKey::F1 && key <= KeyboardKey::F15)
	{
		return "F" + std::to_string(1 + static_cast<int>(key - KeyboardKey::F1));
	}
	ERROR("unhandled key");
	return "[INVALID KEY]";
}

/*static*/std::string Input::name(MouseButton button)
{
	switch (button)
	{
	case MouseButton::Left:
		return "Left Click";
	case MouseButton::Right:
		return "Right Click";
	case MouseButton::Middle:
		return "Middle";
	case MouseButton::XButton1:
		return "Exra Mouse 1";
	case MouseButton::XButton2:
		return "Extra Mouse 2";
	}
	ERROR("unhandled button");
	return "[INVALID MOUSE BUTTON]";
}

/*static*/std::string Input::name(GamepadButton button, int gamepadId)
{
	loadGamepadStringsFile();
	if (button >= buttonNames.size())
	{
		return "GP " + std::to_string(button + 1);
	}
	return buttonNames[static_cast<int>(button)];
	//const sf::Joystick::Identification id = sf::Joystick::getIdentification(gamepadId);
}

/*static*/std::string Input::name(AxisButton button, int gamepadId)
{
	loadGamepadStringsFile();
	const int axisIndex = static_cast<int>(button) % sf::Joystick::AxisCount;
	const bool pos = static_cast<int>(button) < sf::Joystick::AxisCount;
	if (axisIndex >= axesNames.size())
	{
		return "Axis " + std::to_string(axisIndex + 1) + (pos ? "+" : "-");
	}
	return pos ? axesNames.at(axisIndex).pos : axesNames.at(axisIndex).neg;
}

/*static*/std::string Input::name(Axis axis, int gamepadId)
{
	loadGamepadStringsFile();
	const int axisIndex = static_cast<int>(axis) % sf::Joystick::AxisCount;
	if (axisIndex >= axesNames.size())
	{
		return "Axis " + std::to_string(axisIndex + 1);
	}
	return axesNames.at(axisIndex).axis;
}

static const char *gamepadStringsFile = "data/gamepad_strings.json";

/*static*/ void Input::loadGamepadStringsFile()
{
	if (buttonNames.empty() && axesNames.empty())
	{
		auto file = tryReadJsonFile(gamepadStringsFile);
		const json11::Json::object& obj = file.object_items().at(file.object_items().at("used").string_value()).object_items();
		for (const json11::Json& button : obj.at("buttons").array_items())
		{
			buttonNames.push_back(button.string_value());
		}
		const auto& axes = obj.at("axes").array_items();
		ASSERT(axes.size() == sf::Joystick::AxisCount);
		for (const json11::Json& button : axes)
		{
			axesNames.push_back({
				button.array_items().at(0).string_value(), // axis name
				button.array_items().at(1).string_value(), // neg name
				button.array_items().at(2).string_value()  // pos name
			});
		}
	}
}

/*static*/void Input::saveGamepadCustomNames(const std::map<KeyBind, std::string>& translations)
{
	int highestButton = 0;
	int highestAxis = 0;
	for (const std::pair<KeyBind, std::string>& translation : translations)
	{	
		switch (translation.first.type())
		{
		case KeyBind::DeviceType::Keyboard:
		case KeyBind::DeviceType::Mouse:
			ERROR("only gamepad allowed");
			break;
		case KeyBind::DeviceType::GamepadAxis:
			highestAxis = std::max<int>(highestAxis, translation.first.code() % sf::Joystick::AxisCount);
			break;
		case KeyBind::DeviceType::GamepadButton:
			highestButton = std::max<int>(highestButton, translation.first.code());
			break;
		}
	}
	
	static const std::string invalid = "???";
	auto orInvalid = [&translations](const KeyBind& bind) -> std::string {
		auto it = translations.find(bind);
		return it != translations.end() ? it->second : invalid;
	};
	
	buttonNames.clear();
	for (int button = 0; button <= highestButton; ++button)
	{
		buttonNames.push_back(orInvalid(KeyBind(static_cast<GamepadButton>(button))));
	}

	axesNames.clear();
	for (int axis = 0; axis <= highestAxis; ++axis)
	{
		axesNames.push_back({
			invalid,
			orInvalid(KeyBind(fromSF(static_cast<Axis>(axis), true))),
			orInvalid(KeyBind(fromSF(static_cast<Axis>(axis), false)))
		});
	}
	json11::Json::object rootOut;
	json11::Json::object customOut;
	json11::Json::array buttonsOut;
	for (const std::string& button : buttonNames)
	{
		buttonsOut.push_back(button);
	}
	customOut["buttons"] = std::move(buttonsOut);
	json11::Json::array axesOut;
	for (const AxisNames& axis : axesNames)
	{
		json11::Json::array axisOut = { axis.axis, axis.neg, axis.pos };
		axesOut.push_back(std::move(axisOut));
	}
	customOut["axes"] = std::move(axesOut);
	rootOut["custom"] = std::move(customOut);
	rootOut["used"] = "custom";
	auto file = tryReadJsonFile(gamepadStringsFile);
	for (const std::pair<std::string, json11::Json>& pair : file.object_items())
	{
		if (pair.first != "used" && pair.first != "custom")
		{
			rootOut[pair.first] = pair.second;
		}
	}
	std::string err = saveJsonFile(gamepadStringsFile, json11::Json(std::move(rootOut)));
	if (!err.empty())
	{
		ERROR(err);
	}
}

}// End of sl namespace
