/**
* @section DESCRIPTION
* A wrapper class for handling keyboard and mouse input that also provides additional
* features. But mostly so you can access a bunch of stuff outside of the main event poll.
*/
#ifndef SL_INPUT_HPP
#define SL_INPUT_HPP

#include <map>

#include <SFML/Window/Mouse.hpp>
#include <SFML/Window/Keyboard.hpp>
#include <SFML/Window/Joystick.hpp>
#include <SFML/System/Vector2.hpp>
#include <SFML/Graphics/RenderWindow.hpp> // for mouse coordinates

namespace sl
{

class KeyBind;

class Input
{
public:
	typedef sf::Keyboard::Key KeyboardKey;
	typedef sf::Mouse::Button MouseButton;
	typedef unsigned int GamepadButton;
	typedef sf::Joystick::Axis Axis;
	enum class AxisButton
	{
		X = 0,    ///< The X axis
		Y,        ///< The Y axis
		Z,        ///< The Z axis
		R,        ///< The R axis
		U,        ///< The U axis
		V,        ///< The V axis
		PovX,     ///< The X axis of the point-of-view hat
		PovY,     ///< The Y axis of the point-of-view hat
		X_Neg,    ///< The X axis (negative)
		Y_Neg,    ///< The Y axis (negative)
		Z_Neg,    ///< The Z axis (negative)
		R_Neg,    ///< The R axis (negative)
		U_Neg,    ///< The U axis (negative)
		V_Neg,    ///< The V axis (negative)
		PovX_Neg, ///< The X axis of the point-of-view hat (negative)
		PovY_Neg  ///< The Y axis of the point-of-view hat (negative)
	};
	static AxisButton fromSF(sf::Joystick::Axis sfAxis, bool negative = false);
	/**
	 * Constructs an Input class with no associated Window. Must call setWindow
	 * before expecting any mouse input stuff to work
	 */
	Input();

	/**
	 * Updates the state of the Input objects.
	 */
	void update();
	/**
	 * Sets whether the Window is focused or not. This is to be called via
	 * SFML events only, since there's no way to check if a sf::Window is focused
	 * otherwise.
	 * @param focus If the Window is focused
	 */
	void setFocus(bool focus);

	void setWindow(const sf::RenderWindow& window);

	void enable();

	void disable();

	bool isEnabled() const;


	/**
	 * Check if a key has just been released this update
	 * @ param key The key to check for. See the Key enum
	 * @ return whether or not it was released
	 */
	bool isKeyReleased(KeyboardKey key) const;
	/**
	 * Check if a key has just been pressed this update
	 * @ param key The key to check for. See the Key enum
	 * @ return whether or not it was pressed
	 */
	bool isKeyPressed(KeyboardKey key) const;
	/**
	 * Check if a key is being held this update
	 * @ param key The key to check for. See the Key enum
	 * @ return whether or not the key is held down
	 */
	bool isKeyHeld(KeyboardKey key) const;
	/**
	 * Check if a mouse button has just been released this update
	 * @ param button The button to check for. See the Button enum
	 * @ return whether or not it was released
	 */
	bool isMouseReleased(MouseButton button) const;
	/**
	 * Check if a mouse button has just been pressed this update
	 * @ param button The button to check for. See the Button enum
	 * @ return whether or not it was pressed
	 */
	bool isMousePressed(MouseButton button) const;
	/**
	 * Check if a mouse button is being held down this update
	 * @ param button The button to check for. See the Button enum
	 * @ return whether or not it is held down
	 */
	bool isMouseHeld(MouseButton button) const;


	bool isGamepadButtonPressed(unsigned int gamepadId, GamepadButton button) const;

	bool isGamepadButtonReleased(unsigned int gamepadId, GamepadButton button) const;

	bool isGamepadButtonHeld(unsigned int gamepadId, GamepadButton button) const;

	bool isGamepadAxisPressed(unsigned int gamepadId, AxisButton axis) const;

	bool isGamepadAxisReleased(unsigned int gamepadId, AxisButton axis) const;

	bool isGamepadAxisHeld(unsigned int gamepadId, AxisButton axis) const;

	float axis(int gamepadId, Axis axis) const;

	

	/**
	 * Gets the position of the mouse relative to the current Window
	 * even after any transformations have occured, so this will be in game space,
	 * not screen space.
	 * @return The mouse position as an integer vector
	 */
	const sf::Vector2i& mousePos() const;
	/**
	 * Gets the position of the mouse relative to the current Window
	 * even after any transformations have occured, so this will be in game space,
	 * not screen space.
	 * @return The mouse position as a floating point vector
	 */
	const sf::Vector2f& mousePosF() const;
	/**
	 * Gets the absolute positioj of the mouse relative to the top left of the screen.
	 * This does not take into consideration anything about the window.
	 * @return The mouse position as an integer vector
	 */
	const sf::Vector2i& mousePosAbsolute() const;
	/**
	 * Gets the absolute positioj of the mouse relative to the top left of the screen.
	 * This does not take into consideration anything about the window.
	 * @return The mouse position as a floating point vector
	 */
	const sf::Vector2f& mousePosAbsoluteF() const;

	static constexpr float gamepadAxisThreshold = 0.3f;

	static std::string name(KeyboardKey key);

	static std::string name(MouseButton button);

	static std::string name(GamepadButton button, int gamepadId);

	static std::string name(AxisButton button, int gamepadId);

	static std::string name(Axis axis, int gamepadId);

	static void saveGamepadCustomNames(const std::map<KeyBind, std::string>& translations);

private:

	//!Keeps track of when a key was released.
	int mouseButtonUp[sf::Mouse::ButtonCount];
	//!Keeps track of when a key was pressed
	int mouseButtonDown[sf::Mouse::ButtonCount];
	//!Keeps track of when a button was released
	int keyUp[sf::Keyboard::KeyCount];
	//!Keeps track of when a button was pressed
	int keyDown[sf::Keyboard::KeyCount];
	//!Keeps track of when a gamepad button was released
	int gamepadButtonUp[sf::Joystick::Count][sf::Joystick::ButtonCount];
	//!Keeps track of when a gamepad button was pressed
	int gamepadButtonDown[sf::Joystick::Count][sf::Joystick::ButtonCount];
	//!Keeps track of when a gamepad axis was moved back
	int gamepadAxisUp[sf::Joystick::Count][sf::Joystick::AxisCount * 2];
	//!Keeps track of when a gamepad axis was moved into
	int gamepadAxisDown[sf::Joystick::Count][sf::Joystick::AxisCount * 2];

	//!The current absolute float position of the mouse
	sf::Vector2f currentMousePosF;
	//!The current float position of the mouse relative to the window and transformed
	sf::Vector2f currentMousePosTransformedF;
	//!The current absolute integer position of the mouse
	sf::Vector2i currentMousePosI;
	//!The current integer position of the mouse relative to the window and transformed
	sf::Vector2i currentMousePosTransformedI;
	//!The window to use for mouse position transforms
	const sf::RenderWindow *window;
	//!Whether or not the window is focused
	bool focused;
	//!Whether or not the Input is enabled (will return false/0 for all if disabled)
	bool enabled;

	static void loadGamepadStringsFile();

	static std::vector<std::string> buttonNames;
	struct AxisNames
	{
		std::string axis;
		std::string neg;
		std::string pos;
	};
	static std::vector<AxisNames> axesNames;
};

}// End of sl namespace

#endif
