#include "Engine/Physics/CollisionMask.hpp"

#include <algorithm>
#include <cmath>

#include "Utility/Assert.hpp"

namespace sl
{

CollisionMask::CollisionMask(int x, int y, int w, int h, bool enabled)
	:bBox(x, y, w, h)
	,enabled(enabled)
	,origin(0, 0)
{
}

void CollisionMask::set(int x, int y, int w, int h)
{
	bBox.left = x - origin.x;
	bBox.top = y - origin.y;
	bBox.width = w;
	bBox.height = h;
}


void CollisionMask::updatePosition(int x, int y)
{
	bBox.left = x - origin.x;
	bBox.top = y - origin.y;
}

bool CollisionMask::intersects(const Rect<int>& check) const
{
	if (enabled && bBox.intersects(check))
	{
		return true;
	}
	return false;
}

bool CollisionMask::intersects(const CollisionMask& other) const
{
	const Rect<int>& check = other.bBox;
	if (enabled && bBox.intersects(check))
	{
		return true;
	}
	return false;
}

bool CollisionMask::intersects(const Vector2f& p1, const Vector2f& p2) const
{
	// infinite slope case
	if (p1.x == p2.x)
	{
		return bBox.left < p1.x && bBox.left + bBox.width >= p1.x;
	}

	// regular slope case
	const float slope = (p2.y - p1.y) / (p2.x - p1.x);
	const float y1 = p1.y + slope * (bBox.left - p1.x);
	const float y2 = p1.y + slope * (bBox.left + bBox.width - p1.x);
	// I don't even 100% know if this works, but if both points seem to be above or both below the box
	// then the line doesn't seem to intersect the box. seems cheaper than the other method of 4 line-line checks
	return !((y1 < bBox.top && y2 < bBox.top) || (y1 >= bBox.top + bBox.height && y2 >= bBox.top + bBox.height));
}

bool CollisionMask::intersects(const Vector2f& p1, const Vector2f& p2, Vector2f& intersection) const
{
	if (p1.x == p2.x)
	{
		// infinite slope case
		if (bBox.left < p1.x && bBox.left + bBox.width >= p1.x)
		{
			intersection.x = p1.x;
			if (p1.y < p2.y)
			{
				intersection.y = bBox.top;
			}
			else
			{
				intersection.y = bBox.top + bBox.height;
			}
			return true;
		}
	}
	else if (p1.y == p2.y)
	{
		// 0 slope case
		if (bBox.top < p1.y && bBox.top + bBox.height >= p1.y)
		{
			if (p1.x < p2.x)
			{
				intersection.x = bBox.left;
			}
			else
			{
				intersection.x = bBox.left + bBox.width;
			}
			intersection.y = p1.y;
			return true;
		}
	}
	else
	{
		// regular slope case
		const float slope = (p2.y - p1.y) / (p2.x - p1.x);
		bool first = true;
		auto updateIntersectionIfValidAndCloser = [&](float px, float py)
		{
			const float eps = 0.001f;
			// is the point in the box? (a bit wasteful but less code duplication than coding separate horiz/vert checks)
			if (px >= bBox.left - eps && px < bBox.left + bBox.width + eps && py >= bBox.top - eps && py < bBox.top + bBox.height + eps)
			{
				// is it closer than the previous? (two cases depending if line is up or down - can't be straight since that case is earlier)
				if (first || (p1.y < p2.y && py < intersection.y) || (p1.y > p2.y && py > intersection.y))
				{
					first = false;
					intersection.x = px;
					intersection.y = py;
				}
			}
		};
		// check each side
		updateIntersectionIfValidAndCloser(bBox.left, p1.y + slope * (bBox.left - p1.x));
		updateIntersectionIfValidAndCloser(bBox.left + bBox.width - 1, p1.y + slope * (bBox.left + bBox.width - 1 - p1.x));
		updateIntersectionIfValidAndCloser(p1.x + (1.f / slope) * (bBox.top - p1.y), bBox.top);
		updateIntersectionIfValidAndCloser(p1.x + (1.f / slope) * (bBox.top + bBox.height - 1 - p1.y), bBox.top + bBox.height - 1);
		return !first;
	}
	return false;
}

const Rect<int>& CollisionMask::getBoundingBox() const
{
	return bBox;
}

void CollisionMask::setEnabled(bool enabled)
{
	this->enabled = enabled;
}

bool CollisionMask::isEnabled() const
{
	return enabled;
}

void CollisionMask::setOrigin(int x, int y)
{
	origin.x = x;
	origin.y = y;
}

const Vector2i& CollisionMask::getOrigin() const
{
	return origin;
}

Vector2i CollisionMask::getPos() const
{
	return Vector2i(bBox.left + origin.x, bBox.top + origin.y);
}

} // sl
