/**
 * @section DESCRIPTION
 * This class represents a collison mask, which is a way of checking collisions between objects
 * via multiple methods like bounding boxes and spritemasks.
 */
#ifndef SL_COLLISIONMASK_HPP
#define SL_COLLISIONMASK_HPP

//#include <SFML/Graphics/Rect.hpp>
#include <string>

#include "Utility/Rect.hpp"
#include "Utility/Vector.hpp"

namespace sl
{

class CollisionMask
{
public:
	/**
	 * Constructs a CollisionMask with only a bounding box
	 * @param x The x position of the mask
	 * @param y The y position of the mask
	 * @param w The width of the mask's bounding box
	 * @param h The height of the mask's bounding box
	 * @param enabled Whether or not the bounding box should exist
	 */
	CollisionMask(int x = 0, int y = 0, int w = 0, int h = 0, bool enabled = true);

	/**
	 * Sets the mask with a bounding box only
	 * @param x The x position of the mask
	 * @param y The y position of the mask
	 * @param w The width of the mask's bounding box
	 * @param h The height of the mask's bounding box
	 */
	void set(int x, int y, int w, int h);
	/**
	 * Checks if the CollisionMask and a rectangle intersect
	 * @param check The bounding box to check for collisions against
	 * @return Whether or not they intersect
	 */
	bool intersects(const Rect<int>& check) const;
	/**
	 * Checks if the CollisionMask and another CollisionMask intersect
	 * @param check The CollisionMask to check for collisions against
	 * @return Whether or not they intersect
	 */
	bool intersects(const CollisionMask& other) const;
	/**
	 * Checks whether or not a (infinite) line intersects the mask
	 * @param p1 One point on line
	 * @param p2 Another point on line
	 * @return True if the line intersects the collision mask
	 */
	bool intersects(const Vector2f& p1, const Vector2f& p2) const;
	/**
	* Checks whether or not a (infinite) line intersects the mask and returns the first point of intersection
	* @param p1 One point on line
	* @param p2 Another point on line
	* @param intersects The first point at which the line intersects the mask
	* @return True if the line intersects the collision mask
	 */
	bool intersects(const Vector2f& p1, const Vector2f& p2, Vector2f& intersection) const;
	/**
	 * Updates the position of the mask
	 * @param x The new x position
	 * @param y The new y position
	 */
	void updatePosition(int x, int y);
	/**
	 * Gets the bounding box around the mask
	 * @return A reference to the bounding box
	 */
	const Rect<int>& getBoundingBox() const;

	void setEnabled(bool enabled);

	bool isEnabled() const;
	
	void setOrigin(int x, int y);

	const Vector2i& getOrigin() const;

	Vector2i getPos() const;


private:
	//!The bounding box around the mask
	Rect<int> bBox;
	//!Whether the bounding box exists/is enabled. If this is false, all checks will return false
	bool enabled;
	//!The origin of the mask
	Vector2i origin;
};

} // sl

#endif
