#include "Engine/Physics/EntityGrid.hpp"

#include <cmath>
#include <set>

#include "Utility/Assert.hpp"
#include "Engine/Entity.hpp"

namespace sl
{

EntityGrid::EntityGrid(unsigned int w, unsigned int h, unsigned int gridSize)
	:grid(w, h)
	,w(w)
	,h(h)
	,gridSize(gridSize)
{
}

EntityGrid::EntityGrid(EntityGrid&& other)
	:grid(std::move(other.grid))
	,entities(std::move(other.entities))
	,w(other.w)
	,h(other.h)
	,gridSize(other.gridSize)
{
}

EntityGrid& EntityGrid::operator=(EntityGrid&& rhs)
{
	grid = std::move(rhs.grid);
	entities = std::move(rhs.entities);
	w = rhs.w;
	h = rhs.h;
	gridSize = rhs.gridSize;
	return *this;
}


void EntityGrid::insert(Entity *obj)
{
	ASSERT(obj->isStatic());
	int ix = obj->getAbsoluteBounds().left / gridSize;
	int iy = obj->getAbsoluteBounds().top / gridSize;
	//  It might be wise to instead place the entity in the nearest grid cell, rather than just crashing.
	/*  Like this:
		int ix = obj->getPos().x / gridSize;
		int iy = obj->getPos().y / gridSize;
		if (ix < 0)
		{
			ix = 0;
		}
		if (ix >= w)
		{
			ix = w - 1;
		}
		if (iy < 0)
		{
			iy = 0;
		}
		if (iy >= h)
		{
			iy = h - 1;
		}
		//  Basically, use this block insetad of the above if the ASSERT below
		//  ever happens from regular use or something.
	*/
	ASSERT(ix >= 0 && iy >= 0 && ix < w && iy < h);
	int ew = ceil(obj->getAbsoluteBounds().width / (float) gridSize);
	int eh = ceil(obj->getAbsoluteBounds().height / (float) gridSize);
	for (int i = ix; i <= ix + ew && i < w; ++i)
	{
		for (int j = iy; j <= iy + eh && j < h; ++j)
		{
			if (i >= 0 && i < w && j >= 0 && j < h)
			{
				grid.at(i, j).insert(obj->getMultiLink());
			}
		}
	}
	entities.insert(obj->getLink());
}

Entity* EntityGrid::findFirst(const Rect<int>& bBox)
{
	const int minX = std::max(0, bBox.left / gridSize - 1);
	const int minY = std::max(0, bBox.top / gridSize - 1);
	const int maxX = std::min(w - 1, bBox.right() / gridSize);
	const int maxY = std::min(h - 1, bBox.bottom() / gridSize);
	for (int x = minX; x <= maxX; ++x)
	{
		for (int y = minY; y <= maxY; ++y)
		{
			for (Entity& entity : grid.at(x, y))
			{
				if (entity.getMask().getBoundingBox().intersects(bBox))
				{
					return &entity;
				}
			}
		}
	}
	return nullptr;
}

Entity* EntityGrid::findFirst(const Entity& obj)
{
	const int minX = std::max(0, obj.getMask().getBoundingBox().left / gridSize - 1);
	const int minY = std::max(0, obj.getMask().getBoundingBox().top / gridSize - 1);
	const int maxX = std::min(w - 1, obj.getMask().getBoundingBox().right() / gridSize);
	const int maxY = std::min(h - 1, obj.getMask().getBoundingBox().bottom() / gridSize);
	for (int x = minX; x <= maxX; ++x)
	{
		for (int y = minY; y <= maxY; ++y)
		{
			for (Entity& entity : grid.at(x, y))
			{
				if (&obj != &entity && entity.getMask().intersects(obj.getMask()))
				{
					return &entity;
				}
			}
		}
	}
	return nullptr;
}

void EntityGrid::cull(std::list<Entity*>& results, const Rect<int>& bBox)
{
	std::set<const Entity*> hasSelected;
	const int minX = std::max(0, bBox.left / gridSize - 1);
	const int minY = std::max(0, bBox.top / gridSize - 1);
	const int maxX = std::min(w - 1, bBox.right() / gridSize);
	const int maxY = std::min(h - 1, bBox.bottom() / gridSize);
	for (int x = minX; x <= maxX; ++x)
	{
		for (int y = minY; y <= maxY; ++y)
		{
			for (Entity& entity : grid.at(x, y))
			{
				if (entity.getMask().intersects(bBox))
				{
					auto res = hasSelected.insert(&entity);
					if (res.second)
					{
						results.push_front(&entity);
					}
				}
			}
		}
	}
}

void EntityGrid::cull(std::list<Entity*>& results, const Entity& obj)
{
	std::set<const Entity*> hasSelected;
	const int minX = std::max(0, obj.getMask().getBoundingBox().left / gridSize - 1);
	const int minY = std::max(0, obj.getMask().getBoundingBox().top / gridSize - 1);
	const int maxX = std::min(w - 1, obj.getMask().getBoundingBox().right() / gridSize);
	const int maxY = std::min(h - 1, obj.getMask().getBoundingBox().bottom() / gridSize);
	for (int x = minX; x <= maxX; ++x)
	{
		for (int y = minY; y <= maxY; ++y)
		{
			for (Entity& entity : grid.at(x, y))
			{
				if (entity.getMask().intersects(obj.getMask()))
				{
					auto res = hasSelected.insert(&entity);
					if (res.second)
					{
						results.push_front(&entity);
					}
				}
			}
		}
	}
}

void EntityGrid::update()
{
	for (Entity& entity : entities)
	{
		entity.update();
	}
}

void EntityGrid::draw(RenderTarget& target, const Rect<int>& bBox)
{
	std::set<const Entity*> hasDrawn;
	const int minX = std::max(0, bBox.left / gridSize - 1);
	const int minY = std::max(0, bBox.top / gridSize - 1);
	const int maxX = std::min(w - 1, bBox.right() / gridSize);
	const int maxY = std::min(h - 1, bBox.bottom() / gridSize);
	for (int x = minX; x <= maxX; ++x)
	{
		for (int y = minY; y <= maxY; ++y)
		{
			for (const Entity& entity : grid.at(x, y))
			{
				auto res = hasDrawn.insert(&entity);
				if (res.second)
				{
					entity.draw(target);
				}
			}
		}
	}
}

int EntityGrid::count() const
{
	return entities.size();
}

} // sl
