/**
 * @section DESCRIPTION
 * This class represents a grid of static Entity objects held in Scene objects.
 * Store Entities in here if they don't move. Basically, it splits the entire area up into
 * a bunch of cells, so when you go to do culling/collision detection in an area, you only
 * have to check the cells which overlap the area you're checking for, which is pretty much
 * constant time collision detection unless shit is really crowded.
 * Also, Entities must have  bool Entity::isStatic() return true to be allowed to be stored
 * in here.
 */
#ifndef SL_ENTITYGRID_HPP
#define SL_ENTITYGRID_HPP

#include <list>
#include <vector>

#include "Engine/Graphics/RenderTarget.hpp"
#include "Utility/IntrusiveList.hpp"
#include "Utility/IntrusiveMultiList.hpp"
#include "Utility/Grid.hpp"

namespace sl
{

class Entity;

class EntityGrid
{
public:
	/**
	 * Creates a grid of the given size
	 * @param w How many cells the EntityGrid spans horizontally
	 * @param h How many cells the EntityGrid spans vertically
	 * @param gridSize How big those cells are
	 */
	EntityGrid(unsigned int w, unsigned int h, unsigned int gridSize);

	EntityGrid(EntityGrid&& other);

	EntityGrid& operator=(EntityGrid&& rhs);

	/**
	 * Inserts an Entity into the EntityGrid
	 * @param obj The Entity to be inserted
	 */
	void insert(Entity* obj);

	/**
	 * Finds the first Entity that resides within the given bounding box
	 * @param bBox the bounding box to check with
	 * @return the first Entity found (null if nothing found)
	 */
	Entity* findFirst(const Rect<int>& bBox);

	/**
	 * Finds the first Entity that intersects the mask of the input Entity
	 * @param obj The Entity to check with
	 * @return the first Entity found (null if nothing found)
	 */
	Entity* findFirst(const Entity& obj);

	/**
	 * Fills up a list with any Entities stored in the grid which intersect with bBox
	 * @param results The list where Entities are to be stored
	 * @param bBox The bounding box to check Entities against
	 */
	void cull(std::list<Entity*>& results, const Rect<int>& bBox);
	/**
	 * Fills up a list with any Entities stored in the grid which intersect with an Entity
	 * @param results The list where Entities are to be stored
	 * @param obj The Entity to check Entities against
	 */
	void cull(std::list<Entity*>& results, const Entity& obj);

	/**
	 * Updates all Entities stored within the EntityGrid
	 */
	void update();
	/**
	 * Draws all Entities stored within the grid that intersect frustrum
	 * @param target The RenderTarget to render the Entities to
	 * @param bBox the bounding box to check intersections against
	 */
	void draw(RenderTarget& target, const Rect<int>& bBox);

	/**
	 * @return Number of entities stored in this grid
	 */
	int count() const;

private:
	EntityGrid(const EntityGrid&) = delete;
	EntityGrid& operator=(const EntityGrid&) = delete;


	//!A 2D array of lists of the Entities of the EntityGrid
	Grid<IntrusiveMultiList<Entity>> grid;
	//! The actual entity list
	IntrusiveList<Entity, IntrusiveListOwnershipType::Owned> entities;
	//!How many cells wide the EntityGrid is
	int w;
	//!How many cells high the EntityGrid is
	int h;
	//!How big the cells are
	int gridSize;
};

} // sl

#endif // SL_ENTITYGRID_HPP
