#include "Engine/Physics/HeightRect.hpp"

#include "Utility/Assert.hpp"

#ifdef SL_DEBUG
	#include <initializer_list> // for vertices in debug line drawing
	#include "Engine/AbstractScene.hpp" // for debug line drawing
#endif

namespace sl
{

HeightRect::HeightRect(const Rect<int>& bounds, int topLeftHeight, int topRightHeight, int bottomLeftHeight, int bottomRightHeight)
	:HeightMap(bounds)
	,topLeftHeight(topLeftHeight)
	,topRightHeight(topRightHeight)
	,bottomLeftHeight(bottomLeftHeight)
	,bottomRightHeight(bottomRightHeight)
{
}

HeightRect::HeightRect(const HeightRect& copy)
	:HeightMap(copy.getBounds())
	,topLeftHeight(copy.topLeftHeight)
	,topRightHeight(copy.topRightHeight)
	,bottomLeftHeight(copy.bottomLeftHeight)
	,bottomRightHeight(copy.bottomRightHeight)
{
}

HeightRect::~HeightRect()
{
}

HeightRect& HeightRect::operator=(const HeightRect& rhs)
{
	bounds = rhs.bounds;
	topLeftHeight = rhs.topLeftHeight;
	topRightHeight = rhs.topRightHeight;
	bottomLeftHeight = rhs.bottomLeftHeight;
	bottomRightHeight = rhs.bottomRightHeight;
	return *this;
}

bool HeightRect::affects(int x, int y) const
{
	return bounds.containsPoint(x, y);
}

int HeightRect::getHeightAt(int x, int y) const
{
	if (bounds.containsPoint(x, y))
	{
		const float rightFactor = (x - bounds.left) / (float) bounds.width;
		const float leftFactor = 1 - rightFactor;
		const float bottomFactor = (y - bounds.top) / (float) bounds.height;
		const float topFactor = 1 - bottomFactor;
		return (topFactor * topLeftHeight + bottomFactor * bottomLeftHeight) * leftFactor +
				(topFactor * topRightHeight + bottomFactor * bottomRightHeight) * rightFactor;
	}
	return 0;
}

int HeightRect::getHeightAt(const Rect<int>& area) const
{
	ERROR("not implemented");
	return 0;
}

#ifdef SL_DEBUG
void HeightRect::debugDraw(AbstractScene& scene, RenderTarget& renderTarget) const
{
	const int x2 = bounds.left + bounds.width;
	const int y2 = bounds.top + bounds.height;
	const Vector2f topLeft(bounds.left, bounds.top - topLeftHeight);
	const Vector2f topRight(x2, bounds.top - topRightHeight);
	const Vector2f bottomRight(x2, y2 - bottomRightHeight);
	const Vector2f bottomLeft(bounds.left, y2 - bottomLeftHeight);
	std::initializer_list<Vector2f> points = {
		topLeft,
		topRight,
		bottomRight,
		bottomLeft,
		topLeft
	};
	const std::string category = "draw_heightmaps";
	scene.debugDrawRect(category, bounds, sf::Color(255, 128, 0), sf::Color(255, 128, 0, 64));
	scene.debugDrawLine(category, Vector2f(bounds.left, y2), topLeft, sf::Color(255, 128, 0, 128));
	scene.debugDrawLine(category, Vector2f(bounds.left + bounds.width, y2), topRight, sf::Color(255, 128, 0, 128));
	scene.debugDrawLines(category, points, sf::Color::Red);
}
#endif

}