/**
 * @section DESCRIPTION
 * This class is an auto-balancing region-based hierarchy quad tree. Essentially, this class allows for O(log(n)) object querying in
 * a 2D space, rather than O(n). Objects instantiated of this class must have as a parameter a pointer to a class which implements the following methods:
 * void update();
 * void draw(sf::RenderTarget&);
 * Vector2<float>& getPos() const;
 * CollisionMask& getMask() const; (or any return class that implements a bool intersects(Rect<int>&) const, bool intersects(CollisionMask&) and Rect<int>& getBoundingBox() const)
 * IntrusiveListNode<T>& getLink();
 */
#ifndef SL_QUADTREE_HPP
#define SL_QUADTREE_HPP

#include <array>
#include <list>
#include <memory>

//#include <SFML/Graphics/Rect.hpp>
#ifdef SL_DEBUG
	#include <SFML/Graphics/RectangleShape.hpp>
	#include <SFML/Graphics.hpp>
	#include "Engine/Input/Input.hpp"
#endif // SL_DEBUG
#include "Utility/Rect.hpp"
#include "Engine/Physics/CollisionMask.hpp"
#include "Utility/IntrusiveList.hpp"
#include "Engine/Graphics/RenderTarget.hpp"

namespace sl
{

/* Definitions */

template<class T, IntrusiveListOwnershipType O = IntrusiveListOwnershipType::Owned>
class QuadTree
{
public:
	/**
	 * Creates a QuadTree
	 * @param x Where the top left x position of the QuadTree is
	 * @param y Where the top left y position of the QuadTree is
	 * @param w How wide the QuadTree is
	 * @param h How tall the QuadTree is
	 * @param bucketSize How many things a QuadTree can hold before it tries to subdivide
	 * @param parent The parent of the QuadTree in the Quadtree. nullptr means it's the root
	 */
	QuadTree(int x, int y, int w, int h, std::size_t bucketSize = 32, QuadTree<T, O> *parent = nullptr);

	QuadTree(QuadTree<T, O>&& other) = default;

	QuadTree<T, O>& operator=(QuadTree<T, O>&& rhs) = default;


	/**
	 * Inserts something into the QuadTree
	 * @param obj The thing to be inserted
	 */
	void insert(T *obj);

	/**
	 * Queries the QuadTree via bounding box and fills a list with the results
	 * @param results The list to fill
	 * @param bBox the bounding box to query with
	 */
	void cull(std::list<T*>& results, const Rect<int>& bBox);
	/**
	 * Queries the QuadTree via a mask and fills a list with the results
	 * @param results The list to fill
	 * @param obj The thing to query with (needs to implement getMask())
	 */
	void cull(std::list<T*>& results, const T *obj);
	/**
	 * Queries the QuadTree for things that intersect Rect Include but not Exclude
	 * this is used in conjunction with pathfinding voodoo to find nearest things
	 * @param results The list to fill
	 * @param include The bounding box to query with
	 * @param exclude The bounding box to exclude an area from querying
	 */
	 void cull(std::list<T*>& results, const Rect<int>& include, const Rect<int>& exclude);
	/**
	 * Returns the first thing found in the tree that intersects with the bounding box
	 * @param bBox The bounding box to query with
	 * @return Either the first thing to intersect, or nullptr if nothing was found
	 */
	T* findFirst(const Rect<int>& bBox);
	/**
	 * Returns the first thing found in the tree that intersects with the bounding box that isn't notThis
	 * @param bBox The bounding box to query with
	 * @param notThis A thing to ignore if it comes up as the first thing to intersect
	 * @return Either the first thing to intersect, or nullptr if nothing was found
	 */
	T* findFirst(const Rect<int>& bBox, const T *notThis);
	/**
	 * Returns the first thing found in the tree that intersects with a thing
	 * @param obj The thing to query with (needs getMask())
	 * @return Either the first thing to intersect, or nullptr if nothing was found
	 */
	T* findFirst(const T *obj);
	/**
	 * Returns the first thing found in the tree that intersects with the thing that isn't notThis
	 * @param obj The thing to query with
	 * @param notThis A thing to ignore if it comes up as the first thing to intersect
	 * @return Either the first thing to intersect, or nullptr if nothing was found
	 */
	T* findFirst(const T *obj, const T *notThis);
	/**
	 * Checks if a bounding box fits entirely within the QuadTree
	 * @param bBox the bounding box
	 * @return Whether or not it fits
	 */
	bool fits(const Rect<int>& bBox) const;
	/**
	 * Checks if a bounding box intersects with thw QuadTree
	 * @param bBox The bounding box
	 * @return Whether it intersects
	 */
	bool intersects(const Rect<int>& bBox) const;
	/**
	 * Returns the number of elements in the QuadTree and all of its children recursively
	 * @return The number of elements
	 */
	std::size_t size() const;

	/* usage-specific stuff */
	/**
	 * Calls update on the entities. If an Entity updates its position in here then it is floated
	 * up the tree until it is valid. It isn't sunk back down until it's at the lowest quad it can be
	 * though, so you must call balance() once in a while to make the tree optimal (ie sink stuff down)
	 * The reason this isn't done in update() is for performance reasons, as floating stuff down can
	 * take a decent amount of performance, so instead you should periodically call balance() once every
	 * few seconds or whenever to sink stuff.
	 *
	 * tl;dr: the tree will always be VALID even if stuff moves due to updates, but it won't always be
	 * OPTIMAL, so you'll have to call balance() once in a while.
	 */
	void update();
	/**
	 * Tries to sink all entities down the tree to balance it, which may be neccessary for optimal
	 * querying performance if stuff moved in update().
	 */
	void balance();
	/**
	 * Draws all the things in the quad that interesct with the bounding box
	 * @param target The RenderTarget to draw to
	 * @param bBox The bounding box
	 */
	void draw(RenderTarget& target, const Rect<int>& bBox) const;

	void forEach(const std::function<void(const T*)>& function) const;
	/**
	 * Apply a function to every element in the quadtree
	 */
	void forEach(const std::function<void(T*)>& function);
	/**
	 * Apply a function to every element within a certain area
	 * @param function The function to apply
	 * @param bBox The area the things have to exist in to have the function applied
	 */
	void forEach(const std::function<void(T*)>& function, const Rect<int>& bBox);

	//void draw(RenderTarget& target) const;
#ifdef SL_DEBUG
	void drawAsTree(RenderTarget& target, const sf::Vector2f& pos, int width, int height, const Rect<int>& bBox);
#endif
private:
	//nocopy
	QuadTree(const QuadTree<T, O>&) = delete;
	QuadTree<T, O>& operator=(const QuadTree<T, O>&) = delete;

	/**
	 * Unsubdivides and deletes unneccesary leaf quads
	 */
	void absorbChildren();
	/**
	 * Subdivides and pushes objects into leaf nodes
	 */
	void subdivide();

	//!The parent QuadTree
	QuadTree<T, O> *parent;
	//!The children
	std::unique_ptr<std::array<QuadTree<T, O>, 4>> children;
	//!All things in the QuadTree
	IntrusiveList<T, O> contents;
	//!The number of things in the QuadTree
	std::size_t elements;
	//!The maximum number of things a QuadTree can hold before it tries to subdivide
	std::size_t bucketSize;
	//!The area the QuadTree covers
	Rect<int> area;

#ifdef SL_DEBUG
	sf::RectangleShape box;
	sf::CircleShape node;
	sf::Vertex line[4][2];
	sf::Text txt;
#endif
};

/* Implementations */

template <class T, IntrusiveListOwnershipType O>
QuadTree<T, O>::QuadTree(int x, int y, int w, int h, std::size_t bucketSize, QuadTree<T, O> *parent)
	:parent(parent)
	,contents()
	,elements(0)
	,bucketSize(bucketSize)
	,area(x, y, w, h)
#ifdef SL_DEBUG
	,box(sf::Vector2f(w, h))
	,node(8, 16)
#endif
{
#ifdef SL_DEBUG
	box.setPosition(x, y);
	box.setFillColor(sf::Color::Transparent);
	box.setOutlineColor(sf::Color::Color(128, 128, 128));
	box.setOutlineThickness(1);
	node.setFillColor(sf::Color::Black);
	node.setOutlineColor(sf::Color::White);
	node.setOutlineThickness(1);
	txt.setCharacterSize(16);
#endif
}

/* Public interface */
template <class T, IntrusiveListOwnershipType O>
void QuadTree<T, O>::insert(T *obj)
{
	if (size() < bucketSize) // Are we full?
	{
		contents.insert(obj->getLink());
		return;
	}
	// Okay, we're full, so that means we have to subdivide
	subdivide();
	ASSERT(children);
	// We're either not a leaf, or we're a leaf that was full and subdivided
	for (QuadTree<T, O>& child : *children)
	{
		if (child.fits(obj->getMask().getBoundingBox()))
		{
			child.insert(obj);
			return;
		}
	}
	// It's too big to fit into the children
	contents.insert(obj->getLink());
}

template <class T, IntrusiveListOwnershipType O>
void QuadTree<T, O>::cull(typename std::list<T*>& results, const Rect<int>& bBox)
{
	if (area.intersects(bBox))
	{
		// Check for intersections on all contents
		for (typename IntrusiveList<T, O>::Iterator it = contents.begin(); it != contents.end(); ++it)
		{
			if (it->getMask().intersects(bBox))
			{
				results.push_front(&*it);
			}
		}
		// Now cull the children
		if (children)
		{
			for (QuadTree<T, O>& child : *children)
			{
				child.cull(results, bBox);
			}
		}
	}
}

template <class T, IntrusiveListOwnershipType O>
void QuadTree<T, O>::cull(std::list<T*>& results, const Rect<int>& include, const Rect<int>& exclude)
{
	if (area.intersects(include))
	{
		// Check for intersections on all contents
		for (typename IntrusiveList<T, O>::Iterator it = contents.begin(); it != contents.end(); ++it)
		{
			if (it->getMask().intersects(include) && !it->getMask().intersects(exclude))
			{
				results.push_front(&*it);
			}
		}
		// Now cull the children
		if (children)
		{
			for (QuadTree<T, O>& child : *children)
			{
				child.cull(results, include, exclude);
			}
		}
	}
}

template <class T, IntrusiveListOwnershipType O>
void QuadTree<T, O>::cull(typename std::list<T*>& results, const T *obj)
{
	if (area.intersects(obj->getMask().getBoundingBox()))
	{
		// Check for intersections on all contents
		for (typename IntrusiveList<T, O>::Iterator it = contents.begin(); it != contents.end(); ++it)
		{
			if (&*it != obj && it->getMask().intersects(obj->getMask()))
			{
				results.push_front(&*it);
			}
		}
		// Now cull the children
		if (children)
		{
			for (QuadTree<T, O>& child : *children)
			{
				child.cull(results, obj);
			}
		}
	}
}

template <class T, IntrusiveListOwnershipType O>
T* QuadTree<T, O>::findFirst(const Rect<int>& bBox)
{
	if (area.intersects(bBox))
	{
		// Check for intersections on all contents
		for (typename IntrusiveList<T, O>::Iterator it = contents.begin(); it != contents.end(); ++it)
		{
			if (it->getMask().intersects(bBox))
			{
				return &*it;
			}
		}
		// Now cull the children
		if (children)
		{
			for (QuadTree<T,O>& child : *children)
			{
				T *found = child.findFirst(bBox);
				if (found)
				{
					return found;
				}
			}
		}
	}
	return nullptr;
}

template <class T, IntrusiveListOwnershipType O>
T* QuadTree<T, O>::findFirst(const Rect<int>& bBox, const T *notThis)
{
	if (area.intersects(bBox))
	{
		// Check for intersections on all contents
		for (typename IntrusiveList<T, O>::Iterator it = contents.begin(); it != contents.end(); ++it)
		{
			if (it->getMask().intersects(bBox) && &*it != notThis)
			{
				return &*it;

			}
		}
		// Now cull the children
		if (children)
		{
			for (QuadTree<T, O>& child : *children)
			{
				child.findFirst(bBox);
			}
		}
	}
	return nullptr;
}

template <class T, IntrusiveListOwnershipType O>
T* QuadTree<T, O>::findFirst(const T *obj)
{
	if (area.intersects(obj->getMask().getBoundingBox()))
	{
		// Check for intersections on all contents
		for (typename IntrusiveList<T, O>::Iterator it = contents.begin(); it != contents.end(); ++it)
		{
			if (it->getMask().intersects(obj->getMask()))
			{
				return &*it;
			}
		}
		// Now cull the children
		if (children)
		{
			for (QuadTree<T, O>& child : *children)
			{
				T *found = child.findFirst(obj);
				if (found)
				{
					return found;
				}
			}
		}
	}
	return nullptr;
}

template <class T, IntrusiveListOwnershipType O>
T* QuadTree<T, O>::findFirst(const T *obj, const T *notThis)
{
	if (area.intersects(obj->getMask().getBoundingBox()))
	{
		// Check for intersections on all contents
		for (typename IntrusiveList<T, O>::Iterator it = contents.begin(); it != contents.end(); ++it)
		{
			if (it->getMask().intersects(obj->getMask()) && &*it != notThis)
			{
				return &*it;
			}
		}
		// Now cull the children
		if (children)
		{
			for (QuadTree<T, O>& child : *children)
			{
				T *found = child.findFirst(obj);
				if (found)
				{
					return found;
				}
			}
		}
	}
	return nullptr;
}

template <class T, IntrusiveListOwnershipType O>
bool QuadTree<T, O>::fits(const Rect<int>& bBox) const
{
	return (bBox.left >= area.left && bBox.left + bBox.width < area.left + area.width && bBox.top >= area.top && bBox.top + bBox.height < area.top + area.height);
}

template <class T, IntrusiveListOwnershipType O>
bool QuadTree<T, O>::intersects(const Rect<int>& bBox) const
{
	return area.intersects(bBox);
}

template <class T, IntrusiveListOwnershipType O>
std::size_t QuadTree<T, O>::size() const
{
	std::size_t elementsFound = contents.size();
	if (children)
	{
		for (QuadTree<T, O>& child : *children)
		{
			elementsFound += child.size();
		}
	}
	return elementsFound;
}
/* usage specific stuff */
template <class T, IntrusiveListOwnershipType O>
void QuadTree<T, O>::update()
{
	int traversed = 0;
	QuadTree<T, O> *par; // variable used to float things up
	// update entities
	bool advance;
	for (typename IntrusiveList<T, O>::Iterator it = contents.begin(); it != contents.end(); )
	{
		advance = true;
		it->update();
		// if they moved then this might no longer be a valid quadtree, so fix that
		if (!fits(it->getMask().getBoundingBox()) && parent)
		{
			par = parent;
			while (par->parent)
			{
				if (!par->parent->fits(it->getMask().getBoundingBox())) // Move up until we hit the top or fit into the quad
				{
					par = par->parent;
				}
				else
				{
					break;
				}
			}
			if (par)
			{
				//	advance iterator then remove old node so as to not invalidate iterator
				//	then remember to not increase the iterator automatically since we already
				//	went to the next one here.
				T &obj = *it;
				++it;
				advance = false;
				par->contents.insert(obj.getLink());
			}
			if (children)
			{
				if (size() < bucketSize)
				{
					absorbChildren();
				}
			}
		}
		++traversed;
		if (advance)
		{
			++it;
		}
	}
	if (children)
	{
		for (QuadTree<T, O>& child : *children)
		{
			child.update();
		}
	}
}

template <class T, IntrusiveListOwnershipType O>
void QuadTree<T, O>::balance()
{
	if (children)
	{
		bool advance;
		for (typename IntrusiveList<T, O>::Iterator it = contents.begin(); it != contents.end();)
		{
			advance = true;
			for (QuadTree<T, O>& child : *children)
			{
				if (child.fits(it->getMask().getBoundingBox()))
				{
					//	advance iterator then link so we don't invalidate iterator
					//	then remember to not automatically advance it later since
					//	we just did that
					advance = false;
					T& obj = *it;
					++it;
					child.contents.insert(obj.getLink());
					if (child.size() > bucketSize)
					{
						child.subdivide();
					}
					break;
				}
			}
			if (advance)
			{
				++it;
			}
		}
		for (QuadTree<T, O>& child : *children)
		{
			child.balance();
		}
	}
}

template <class T, IntrusiveListOwnershipType O>
void QuadTree<T, O>::draw(RenderTarget& target, const Rect<int>& bBox) const
{
	if (area.intersects(bBox))
	{
		for (typename IntrusiveList<T, O>::ConstIterator it = contents.cbegin(); it != contents.cend(); ++it)
		{
			//if (bBox.intersects(it->getDrawableBounds()))
			{
				it->draw(target);
			}
		}
		if (children)
		{
			for (const QuadTree<T, O>& child : *children)
			{
				child.draw(target, bBox);
			}
		}
	}
}

template <class T, IntrusiveListOwnershipType O>
void QuadTree<T, O>::forEach(const std::function<void(const T*)>& function) const
{
	for (const T& elem : contents)
	{
		function(&elem);
	}
	if (children)
	{
		for (const QuadTree<T, O>& child : *children)
		{
			child.forEach(function);
		}
	}
}

template <class T, IntrusiveListOwnershipType O>
void QuadTree<T, O>::forEach(const std::function<void(T*)>& function)
{
	for (T& elem : contents)
	{
		function(&elem);
	}
	if (children)
	{
		for (QuadTree<T, O>& child : *children)
		{
			child.forEach(function);
		}
	}
}

template <class T, IntrusiveListOwnershipType O>
void QuadTree<T, O>::forEach(const std::function<void(T*)>& function, const Rect<int>& bBox)
{
	if (area.intersects(bBox))
	{
		for (T& elem : contents)
		{
			if (bBox.intersects(elem.getBoundingBox()))
			{
				function(&elem);
			}
		}
		if (children)
		{
			for (QuadTree<T, O>& child : *children)
			{
				child.forEach(function, bBox);
			}
		}
	}
}

#ifdef SL_DEBUG
template <class T, IntrusiveListOwnershipType O>
void QuadTree<T, O>::drawAsTree(RenderTarget& target, const sf::Vector2f& pos, int width, int height, const Rect<int>& bBox)
{
	int depth = 0;
	QuadTree<T, O> *par = parent;
	while (par)
	{
		par = par->parent;
		++depth;
	}
	//txt.setString(makeString(depth));
	sf::Color colour2(bBox.intersects(area) ? sf::Color::White : sf::Color::Color(128, 128, 128));
	sf::Color colour(bBox.intersects(area) ? sf::Color::Green : sf::Color::Red);
	//if (Input::isKeyHeld(Input::LShift))
	//{
	//	for (typename IntrusiveList<T, O>::ConstIterator it = contents.cbegin(); it != contents.cend(); ++it)
	//	{
	//		line[0][0] = sf::Vertex(sf::Vector2f(area.left + area.width / 2, area.top + area.height / 2), colour);
	//		line[0][1] = sf::Vertex(sf::Vector2f(it->getPos().x, it->getPos().y), colour);
	//		target.target.draw(line[0], 2, sf::Lines);
	//		//txt.setPosition((*it)->getPos().x, (*it)->getPos().y - 16);
	//		//target.draw(txt);
	//	}
	//}
	node.setOutlineColor(colour);
	node.setOutlineThickness(1 + 2 * area.intersects(bBox));
	sf::Vector2f cPos(pos.x - width / 2, pos.y + height);
	if (children)
	{
		for (QuadTree<T, O>& child : *children)
		{
			colour = child.area.intersects(bBox) ? sf::Color::Green : sf::Color::Red;
			line[i][0] = sf::Vertex(pos, colour);
			line[i][1] = sf::Vertex(cPos, colour);
			target.draw(line[i], 2, sf::Lines);
			child.drawAsTree(target, cPos, width / 4, height, bBox);
		}
		cPos.x += width / 3;
	}
	box.setOutlineColor(colour2);
	target.draw(box);
	const int elements = size();
	if (elements >= bucketSize)
	{
		node.setFillColor(sf::Color::Yellow);
		node.setRadius(12);
	}
	else if (elements == bucketSize)
	{
		node.setFillColor(sf::Color::White);
		node.setRadius(8);
	}
	else if (elements >= bucketSize / 3 * 2)
	{
		node.setFillColor(sf::Color::Color(192, 192, 192));
		node.setRadius(8);
	}
	else if (elements >= bucketSize / 3)
	{
		node.setFillColor(sf::Color::Color(128, 128, 128));
		node.setRadius(8);
	}
	else if (elements > 0)
	{
		node.setFillColor(sf::Color::Color(64, 64, 64));
		node.setRadius(8);
	}
	else
	{
		node.setFillColor(sf::Color::Red);
		node.setRadius(4);
	}
	node.setPosition(pos.x - node.getRadius(), pos.y - node.getRadius());
	target.draw(node);
	//txt.setString(makeString(elements));
	//txt.setPosition(pos.x + 16, pos.y + 16);
	//target.draw(txt);
}
#endif
/* private methods */
template <class T, IntrusiveListOwnershipType O>
void QuadTree<T, O>::absorbChildren()
{
	if (children)
	{
		for (QuadTree<T, O>& child : *children)
		{
			child.absorbChildren();
			elements += child.elements;// Do this or we will think that we contain anti-matter
			contents.splice(child.contents);
		}
		children.reset();
	}
}

template <class T, IntrusiveListOwnershipType O>
void QuadTree<T, O>::subdivide()
{
	if (children)
	{
		return;
	}
	children = std::make_unique<std::array<QuadTree<T, O>, 4>>(std::array<QuadTree<T, O>, 4>{{
		QuadTree<T, O>(area.left + (0 % 2) * (area.width / 2), area.top + (0 / 2) * (area.height / 2), area.width / 2, area.height / 2, bucketSize, this),
		QuadTree<T, O>(area.left + (1 % 2) * (area.width / 2), area.top + (1 / 2) * (area.height / 2), area.width / 2, area.height / 2, bucketSize, this),
		QuadTree<T, O>(area.left + (2 % 2) * (area.width / 2), area.top + (2 / 2) * (area.height / 2), area.width / 2, area.height / 2, bucketSize, this),
		QuadTree<T, O>(area.left + (3 % 2) * (area.width / 2), area.top + (3 / 2) * (area.height / 2), area.width / 2, area.height / 2, bucketSize, this)
	}});
	for (typename IntrusiveList<T, O>::Iterator it = contents.begin(); it != contents.end(); ++it)
	{
		for (QuadTree<T, O>& child : *children)
		{
			if (child.fits(it->getMask().getBoundingBox()))
			{
				//	don't fuck up our iterator!
				T& temp = *it;
				--it;
				child.insert(&temp);
				break;
			}
		}
	}
}

}// End of sl namespace

#endif // SL_QUADTREE_HPP
