/**
 * @DESCRIPTION This functions more or less similar to an intrusive version of std::shared_ptr.
 * It provides a more direct interface where there is no need to always lock the pointer before using it,
 * but a locking mechanism is still provided, although any owning containers will have to be diligent about using it.
 *
 * Motivations to use this instead of std::shared_ptr<T>:
 * 1) It provides less indirection than storing containers of std::shared_ptrs.
 * 2) Locking is really only necessary when modifying from other threads. A mutex could have been used,
 *    but by simply checking if the alive counter before deletion we allow the main thread to continue
 *    and just skip this entity, instead of potentially locking up when another thread (ie pathing) needs access to an Entity/etc.
 * 3) A Ref<T> is directly obtainable from a T rather than requiring every single handling site to keep it as a std::shared_ptr<T>
 *
 * Neutral:
 * 1) The interface here also allows more natural conversions to raw pointers when needed, unlike std::shared_ptr.
 *
 * Downsides:
 * 1) More diligence is needed when handling storage of Refables.
 *    This should be largely limited, and Entity storage is mostly only done via intrusive containres anyway.
 * 2) It's intrusive, thus is stored directly in the type. This isn't too big of an issue considering our main use case is Entity,
 *    which is already specialized and has an IntrusiveListNode in it anyway.
 */
#ifndef SL_REF_HPP
#define SL_REF_HPP

#include <memory>

#include "Engine/Refable.hpp"
#include "Utility/Assert.hpp"

namespace sl
{

namespace ref_impl
{
// to get around inheritance access fuckery (Ref<T> can't access Ref<U>'s protected members...)
template <bool Owning>
class RefBase
{
protected:
	RefBase() = default;

	template <typename T>
	RefBase(T *ptr)
		:alive(ptr ? ptr->getAliveRef().alive : nullptr)
	{
		incr();
	}

	RefBase(Refable& refable)
		:alive(refable.alive)
	{
		incr();
	}

	template <bool U>
	RefBase(const RefBase<U>& other)
		:alive(other.alive)
	{
		incr();
	}

	~RefBase()
	{
		decr();
	}

	template <bool U>
	void reset(const RefBase<U>& other)
	{
		decr();
		alive = other.alive;
		incr();
	}

	template <typename T>
	void reset(T *ptr)
	{
		decr();
		if (ptr)
		{
			alive = ptr->getAliveRef().alive;
			incr();
		}
		else
		{
			alive = nullptr;
		}
	}

	// this is an alive counter - the owner sets it to 1
	// and anyone who wishes to have an owning reference increments it
	std::shared_ptr<int> alive;

private:
	void incr()
	{
		if (Owning && alive)
		{
			++*alive;
		}
	}

	void decr()
	{
		if (Owning && alive)
		{
			--*alive;
		}
	}
};
} // ref_impl

// requirements on T:
// 1) Refable getAliveRef()
template <typename T, bool Owning = false>
class Ref : public ref_impl::RefBase<Owning>
{
public:
	Ref()
		:RefBase()
		,ptr(nullptr)
	{
	}

	Ref(T *ptr)
		:RefBase(ptr)
		,ptr(static_cast<T*>(ptr))
	{
	}

	template <typename U, bool V>
	Ref(const Ref<U, V>& other)
		:RefBase(other)
		,ptr(static_cast<T*>(other.get()))
	{
	}

	template <typename U, bool V>
	Ref<T, Owning>& operator=(const Ref<U, V>& rhs)
	{
		this->RefBase<Owning>::reset(rhs);
		ptr = static_cast<T*>(rhs.get());
		return *this;
	}

	template <typename U>
	Ref<T>& operator=(std::nullptr_t)
	{
		alive = nullptr;
		ptr = nullptr;
		return *this;
	}

	template <typename U>
	Ref<T, Owning>& operator=(U *rawPtr)
	{
		this->ref_impl::RefBase<Owning>::reset(rawPtr);
		ptr = static_cast<T*>(rawPtr);
		return *this;
	}

	~Ref() = default;

	template <typename U>
	bool operator==(const Ref<U>& rhs) const
	{
		if (alive != rhs.alive)
		{
			return false;
		}
		if (alive)
		{
			return get() != static_cast<T*>(rhs.get());
		}
		return true;
	}

	bool operator==(std::nullptr_t)
	{
		return get() == nullptr;
	}

	template <typename U>
	bool operator==(const U *rawPtr)
	{
		return get() == static_cast<T*>(rawPtr);
	}

	template <typename U>
	bool operator!=(const Ref<U>& rhs) const
	{
		return !this->operator==(rhs);
	}

	bool operator!=(std::nullptr_t)
	{
		return get() != nullptr;
	}

	template <typename U>
	bool operator!=(const U *rawPtr)
	{
		return get() != static_cast<T*>(rawPtr);
	}

	explicit operator bool() const
	{
		return get() != nullptr;
	}

	operator T*() const
	{
		return get();
	}

	T& operator*() const
	{
		ASSERT(isAlive());
		return *get();
	}

	T* operator->() const
	{
		ASSERT(isAlive());
		return get();
	}

	T* get() const
	{
		return isAlive() ? ptr : nullptr;
	}

	Ref<T, true> lock()
	{
		return Ref<T, true>(get());
	}

	bool isAlive() const
	{
		return alive && *alive != 0;
	}

private:
	T *ptr;
};

} // sl

#endif // SL_REF_HPP