#include "Engine/AbstractScene.hpp"

#include <algorithm>
#include <cmath>
#include <limits>

#include "Engine/Serialization/AutoDeserialize.hpp"
#include "Engine/Serialization/AutoSerialize.hpp"
#include "Engine/Input/Input.hpp"
#include "Town/World.hpp"
#include "Pathfinding/GridPathManager.hpp"
#include "Pathfinding/RealTimePathGrid.hpp"
#include "Engine/Physics/CollisionMask.hpp"
#include "Engine/Entity.hpp"
#include "Engine/Physics/PhysicsComponent.hpp"
#include "Player/Player.hpp"
#include "Town/Map/Map.hpp"
#include "Town/Map/MapData.hpp"
#include "Utility/Assert.hpp"
#include "Utility/Logger.hpp"
#include "Utility/Memory.hpp"

namespace sl
{

// purposely not set between saves (@TODO reset this when someone loads?)
int Scene::idCount = 0;

Scene::Scene(Game *game, unsigned int width, unsigned int height, unsigned int viewWidth, unsigned int viewHeight, const std::string& type)
	:AbstractScene(game, viewWidth, viewHeight)
	,width(width)
	,height(height)
	,gridSize(64)
	,grids()
	,quads()
	,lists()
	,bucketSize(64)
	,balanceRate(60)
	,viewPos(width / 2, height / 2)
	,balanceStep(0)
	,entitiesToBalance(1)   //  These are set to 1 instead of 0 to avoid a division by zero error without
	,entitiesBalanced(1)	//  wasting performance checking if entitiesToBalance > 0 every single frame.
	,quadsBalanced(0)
	,balanceIterator(quads.begin())
	,isUpdating(false)
	,player(nullptr)
	,type(type)
	,worldCoords(0, 0, 0)
	,worldScale(1.f)
	,id(idCount++)
{
	ASSERT(id >= 0 && id < (1 << 11)); // pathfinding assumes 0 <= id <= so it can fit in 10 bits for it's NodeID hash
}

Scene::~Scene()
{
	clearDeletedEntityList();
	ASSERT(deleteList.empty());

	// TODO: figure out whether I need to delayAddEntities()
	// since theres some crashes in delayAddEntities. This would have the implication that
	// Entities could expect to be in a scene upon deletion - is that a valid assumption
	// that we should keep? Does it even exist right now? Not really...
	//delayAddEntities();

	isUpdating = false;
	for (std::function<void()>& f : callAtEndOfUpdate)
	{
		f();
	}
}

/*
	-----------------------------------------------------------------------------------
	-------------------------	ENTITY QUERYING		-----------------------------------
	-----------------------------------------------------------------------------------
*/

Entity* Scene::checkCollision(const Entity& e, const Types::Type& type)
{
	auto it = entitiesLookup.find(type);
	if (it != entitiesLookup.end())
	{
		Entity *temp;
		if (it->second.grid && (temp = it->second.grid->findFirst(e)))
		{
			return temp;
		}
		if (it->second.quad && (temp = it->second.quad->findFirst(&e)))
		{
			return temp;
		}
		if (it->second.list)
		{
			for (Entity& entity : *it->second.list)
			{
				if (&entity != &e && entity.getMask().intersects(e.getMask()))
				{
					return &entity;
				}
			}
		}
	}
	//else
	//{
	//	Logger::log(Logger::Physics, Logger::Debug) << "Couldn't find type {" << type << "}" << std::endl;
	//}
	return nullptr;
}

Entity* Scene::checkCollision(const Rect<int>& bBox, const Types::Type& type)
{
	auto it = entitiesLookup.find(type);
	if (it != entitiesLookup.end())
	{
		Entity *temp = nullptr;
		if (it->second.grid && (temp = it->second.grid->findFirst(bBox)))
		{
			return temp;
		}
		if (it->second.quad && (temp = it->second.quad->findFirst(bBox)))
		{
			return temp;
		}
		if (it->second.list)
		{
			for (Entity& entity : *it->second.list)
			{
				if (entity.getMask().intersects(bBox))
				{
					return &entity;
				}
			}
		}
	}
	//else
	//{
	//	Logger::log(Logger::Physics, Logger::Debug) << "Couldn't find type {" << type << "}" << std::endl;
	//}
	return nullptr;
}

void Scene::cull(std::list<Entity*>& results, const Rect<int>& bBox, const Types::Type& type)
{
	auto it = entitiesLookup.find(type);
	if (it != entitiesLookup.end())
	{
		if (it->second.grid)
		{
			it->second.grid->cull(results, bBox);
		}
		if (it->second.quad)
		{
			it->second.quad->cull(results, bBox);
		}
		if (it->second.list)
		{
			for (Entity& entity : *it->second.list)
			{
				if (entity.getMask().getBoundingBox().intersects(bBox))
				{
					results.push_front(&entity);
				}
			}
		}
	}
	//else
	//{
	//	Logger::log(Logger::Physics, Logger::Debug) << "Couldn't find type {" << type << "}" << std::endl;
	//}
}

void Scene::cull(std::list<Entity*>& results, const Entity& e, const Types::Type& type)
{
	auto it = entitiesLookup.find(type);
	if (it != entitiesLookup.end())
	{
		if (it->second.grid)
		{
			it->second.grid->cull(results, e);
		}
		if (it->second.quad)
		{
			it->second.quad->cull(results, &e);
		}
		if (it->second.list)
		{
			for (Entity& entity : *it->second.list)
			{
				if (entity.getMask().intersects(e.getMask()))
				{
					results.push_front(&entity);
				}
			}
		}
	}
	//else
	//{
	//	Logger::log(Logger::Physics, Logger::Debug) << "Couldn't find type {" << type << "}" << std::endl;
	//}
}

void Scene::cullFiniteLine(std::list<Entity*>& results, const Vector2f& p1, const Vector2f& p2, const Types::Type& type)
{
	std::list<Entity*> cullBuffer;
	Vector2f why;
	const float minLineY = std::min(p1.y, p2.y);
	const float maxLineY = std::max(p1.y, p2.y);
	const float minLineX = std::min(p1.x, p2.x);
	const float maxLineX = std::max(p1.x, p2.x);
	const Rect<int> lineBox(minLineX, minLineY, maxLineX - minLineX, maxLineY - minLineY);
	cull(cullBuffer, lineBox, type);
	for (Entity *entity : cullBuffer)
	{
		// check if they pass the infinite line check, but since they already passed the axis-aligned bounding box
		// check in the earlier cull() to get here this means that a finite line between p1 and p2 intersects them too
		if (entity->getMask().intersects(p1, p2, why))
		{
			results.push_front(entity);
		}
	}
}

Scene::RaycastResult Scene::raycast(const Entity& e, const std::initializer_list<const char*>& types, const Vector2f& motion, Vector2f& result)
{
	if (motion.length() <= std::numeric_limits<float>::epsilon())
	{
		result = e.getPos();
		return RaycastResult::DidNotMove;
	}
	Rect<int> rayBox = e.getMask().getBoundingBox();
	if (motion.x < 0)
	{
		rayBox.left += motion.x;
	}
	if (motion.y < 0)
	{
		rayBox.top += motion.y;
	}
	rayBox.width += abs(motion.x);
	rayBox.height += abs(motion.y);
	std::list<Entity*> query;
	for (const auto& type : types)
	{
		cull(query, rayBox, type);
	}
	int steps = ceilf(motion.length());
	ASSERT(steps > 0);
	const Vector2f step = motion * (1.f / steps);
	CollisionMask stepMask = e.getMask();
	Vector2f queryPos = e.getPos() - step;
	for (int i = 0; i < steps; ++i)
	{
		queryPos += step;
		stepMask.updatePosition(queryPos.x, queryPos.y);
		
		for (Entity *entity : query)
		{
			if (entity->getMask().intersects(stepMask))
			{
				if (i == 0)
				{
					result = e.getPos();
					return RaycastResult::DidNotMove;
				}
				else
				{
					result = queryPos - step;
					return RaycastResult::MovedPartially;
				}
			}
		}
	}
	result = queryPos;
	return RaycastResult::MovedFully;
}

Scene::RaycastResult Scene::raycast(const Vector2f& pos, const std::initializer_list<const char*>& types, const Vector2f& motion, Vector2f& result)
{
	if (motion.length() <= std::numeric_limits<float>::epsilon())
	{
		result = pos;
		return RaycastResult::DidNotMove;
	}
	const Vector2f p2 = pos + motion;
	std::vector<Vector2f> intersectionPoints;
	const int minLineY = std::min(pos.y, p2.y);
	const int maxLineY = std::max(pos.y, p2.y);
	const int minLineX = std::min(pos.x, p2.x);
	const int maxLineX = std::max(pos.x, p2.x);
	const Rect<int> lineBox(minLineX, minLineY, std::max(1, maxLineX - minLineX), std::max(1, maxLineY - minLineY));
	for (const auto& type : types)
	{
		std::list<Entity*> cullBuffer;
		cull(cullBuffer, lineBox, type);
		Vector2f intersection;
		for (const Entity *entity : cullBuffer)
		{
			// check if they pass the infinite line check, but since they already passed the axis-aligned bounding box
			// check in the earlier cull() to get here this means that a finite line between p1 and p2 intersects them too
			if (entity->getMask().intersects(pos, p2, intersection))
			{
				intersectionPoints.push_back(intersection);
			}
		}
	}
	if (!intersectionPoints.empty())
	{
		auto smallest = intersectionPoints.begin();
		for (auto it = smallest + 1; it != intersectionPoints.end(); ++it)
		{
			// since they're all on a line, if we just check the x-difference from p1 we know that the
			// ones with the smallest xdiff will have the smallest distance on line too
			const float xdif = fabsf(it->x - pos.x);
			if (xdif < fabsf(smallest->x - pos.x))
			{
				smallest = it;
			}
		}
		result = *smallest;
		return RaycastResult::MovedPartially;
	}
	result = p2;
	return RaycastResult::MovedFully;
}


/*
	-----------------------------------------------------------------------------------
	-------------------------	ENTITY INSERTING	-----------------------------------
	-----------------------------------------------------------------------------------
*/

void Scene::addEntityToGrid(Entity *obj)
{
	delayedAddEntities[StorageType::InGrid].insert(obj->getLink());
}

void Scene::addEntityToGridNoUpdate(Entity *obj)
{
	delayedAddEntities[StorageType::InGridNoUpdate].insert(obj->getLink());
}

void Scene::addEntityToQuad(Entity *obj)
{
	delayedAddEntities[StorageType::InQuad].insert(obj->getLink());
}

void Scene::addEntityToList(Entity *obj)
{
	delayedAddEntities[StorageType::InList].insert(obj->getLink());
}

void Scene::addImmediately(Entity *obj, StorageType storageType)
{
	ASSERT(obj != nullptr);
	ASSERT(obj->storageType == StorageType::InNone);
	const Types::Type& type = obj->getType();
	switch (storageType)
	{
		case StorageType::InGrid:
		case StorageType::InGridNoUpdate:
		{
			if (storageType == StorageType::InGrid)
			{
				updateableGridEntities.push_back(obj);
			}
			else
			{
				nonUpdateableGridEntities.push_back(obj);
			}
			auto it = grids.find(type);
			if (it == grids.end())
			{
				it = grids.insert(std::make_pair(type, EntityGrid(ceil(width / gridSize), ceil(height / gridSize), gridSize))).first;
				entitiesLookup[type].grid = &it->second;
#ifdef SL_DEBUG
				Logger::log(Logger::Physics, Logger::Debug) << "Added entity type \"" << type << "\" to the grid list" << std::endl;
#endif
			}
			it->second.insert(obj);
		}
		break;

		case StorageType::InQuad:
		{
			auto it = quads.find(type);
			if (it == quads.end())
			{
				it = quads.insert(std::make_pair(type, QuadTree<Entity>(0, 0, width, height, bucketSize))).first;
				entitiesLookup[type].quad = &it->second;
#ifdef SL_DEBUG
				Logger::log(Logger::Physics, Logger::Debug) << "Added entity type \"" << type << "\" to the quad list" << std::endl;
#endif
				rectifyBalanceIterator();
			}
			it->second.insert(obj);
		}
		break;

		case StorageType::InList:
		{
			auto& list = lists[type];
			entitiesLookup[type].list = &list;
			list.insert(obj->getLink());
		}
		break;

		default:
		{
			ERROR("Trying to add Entity to unsupported structure");
		}
		break;
	}

	obj->storageType = storageType;

	if (obj->scene)
	{
		obj->onExit();
	}
	obj->scene = this;
	if (obj->physicsComponent)
	{
		obj->physicsComponent->scene = this;
	}
	obj->onEnter();
	recomputeAreaAroundEntityIfNecessary(*obj);
}


int Scene::getWidth() const
{
	return width;
}

int Scene::getHeight() const
{
	return height;
}

int Scene::getTileSize() const
{
	// @TILESIZE
	return 16;
}

const sf::Vector2f Scene::getView() const
{
	return viewPos;
}

void Scene::setView(int x, int y)
{
	const int halfViewWidth = viewWidth / 2;
	const int halfViewHeight = viewHeight / 2;
	if (x < halfViewWidth)
	{
		x = halfViewWidth;
	}
	else if (x > (int) width - halfViewWidth)
	{
		x = width - halfViewWidth;
	}
	if (y < halfViewHeight)
	{
		y = halfViewHeight;
	}
	else if (y > (int) height - halfViewHeight)
	{
		y = height - halfViewHeight;
	}
	viewPos.x = x;
	viewPos.y = y;
}

void Scene::deleteEntity(Entity *entity)
{
	ASSERT(entity->isDead());
	deleteList.push_back(entity);
}

void Scene::activateEntity(Entity *entity)
{
	ASSERT(entity != nullptr);
	ASSERT(entity->storageType != InNone);//Trying to activate an Entity with no valid previous data structure
	StorageType storageType = entity->storageType;
	entity->storageType = InNone;
	addImmediately(entity, storageType);
}

void Scene::deactivateEntity(Entity *entity)
{
	// TODO fix this. Right now when you deactivate an entity it doesn't remove its multilist stuff.
	// You will need to remove it from the EntityGrid and possibly other places as well.
	// I am not sure how to handle the other lists that may exist in the future and this needs to be addressed.
	ASSERT(entity->storageType != StorageType::InGrid && entity->storageType != StorageType::InGridNoUpdate);

	deactivationQueue.push_back(entity);
}

void Scene::transferEntity(Entity *entity)
{
	if (entity->scene == nullptr)
	{
		ERROR("Trying to move an Entity to a different Scene when it never belonged in one before");
	}
	if (entity->scene->isUpdating)
	{
		entity->scene->callAtEndOfUpdate.push_back(std::bind(&Scene::transferEntity, this, entity));
	}
	else if (this->isUpdating)
	{
		callAtEndOfUpdate.push_back(std::bind(&Scene::transferEntity, this, entity));
	}
	else
	{
		this->activateEntity(entity);
	}
}

void Scene::addHeightMap(HeightMap *heightMap)
{
	ASSERT(heightMap != nullptr);
	heightMaps.insert(heightMap->getLink());
}

int Scene::getHeightAt(int x, int y) const
{
	int height = 0;

	for (const HeightMap& map : heightMaps)
	{
		const int ret = map.getHeightAt(x, y);
		if (map.affects(x, y) && ret)
		{
			height = ret;
		}
	}

	return height;
}

int Scene::getHeightAt(const Rect<int>& area) const
{
	ERROR("implement this before you call it you faggot");
	return 0;
}

Player* Scene::getPlayer() const
{
	return player;
}

void Scene::setPlayer(Player *player)
{
	this->player = player;
}

void Scene::pause(const std::string& type)
{
	pauseList.insert(type);
}

void Scene::unpause(const std::string& type)
{
	pauseList.erase(type);
}

map::Map* Scene::getMap() const
{
	return mapDrawable.get();
}

void Scene::makeMap(const map::MapData& mapData)
{
	createMapDrawable();
	mapDrawable->setData(mapData);
}

void Scene::makeMap(const sf::Image& mapImage)
{
	createMapDrawable();
	mapDrawable->setData(mapImage);
}

void Scene::createMapDrawable()
{
	constexpr int mapWidth = 512 - 32;
	constexpr int mapHeight = 384 - 32;
	// mapWidth/mapHeight are in PIXELS, but in the default zoom 2 pixels = 1 mapspace coordinate.
	mapDrawable = unique<map::Map>(getTileSize(), Vector2i(128, 128), Vector2i(mapWidth / 2, mapHeight / 2));
}

const std::string& Scene::getType() const
{
	return type;
}

const sf::Vector3i& Scene::getWorldCoords() const
{
	return worldCoords;
}

float Scene::getWorldScale() const
{
	return worldScale;
}

int Scene::getSceneID() const
{
	return id;
}

int Scene::instanceCount(const std::string& id) const
{
	int count = 0;
	auto it = entitiesLookup.find(id);
	if (it != entitiesLookup.end())
	{
		if (it->second.grid)
		{
			count += it->second.grid->count();
		}
		if (it->second.quad)
		{
			count += it->second.quad->size();
		}
	}
	return count;
}


/*  protected   */
void Scene::updateEntities()
{
#ifdef SL_DEBUG
	if (input.isKeyHeld(Input::KeyboardKey::LShift) && input.isKeyPressed(Input::KeyboardKey::P))
	{
		for (const auto& pair : precomputedPathGrids)
		{
			pair.second->debugSaveImage(getType());
		}
	}
#endif // SL_DEBUG
	for (auto& activityType : groupAIs)
	{
		for (std::unique_ptr<IGroupAI>& groupAI : activityType.second)
		{
			groupAI->update();
			debugDrawCircle("draw_activities", groupAI->getPos(), groupAI->getRadius(), sf::Color(0, 128, 0));
		}
	}

	//	holy balls will this lag because of trees/fences/etc if you uncomment this
	/*for (std::map<Types::Type, EntityGrid*>::iterator it = grids.begin(); it != grids.end(); ++it)
	{
	it->second->update();
	}*/
	isUpdating = true;
	for (std::map<Types::Type, QuadTree<Entity>>::iterator it = quads.begin(); it != quads.end(); ++it)
	{
		if (pauseList.count(it->first) == 0)
		{
			it->second.update();
		}
	}
	//  Make sure we've always balanced at least the same percentage of entities as the percentage of
	//  the completion of one balance cycled. Hopefully this means spreading the workload out throughout
	//  the cycle, instead of hitting a giant slowdown if we balanced every single tree at once.
	while (balanceIterator != quads.end() && (float) entitiesBalanced / entitiesToBalance <= (float) balanceStep / balanceRate)
	{
		balanceIterator->second.balance();
		//std::cout << "There are " << balanceIterator->second->size() << " entities of type " << balanceIterator->first << " that were just balanced!" << std::endl;
		++balanceIterator;
		++quadsBalanced;
	}
	//  We've completed one balance cycle, so reset everything for the next cycle.
	if (balanceStep >= balanceRate)
	{
		balanceStep = 0;
		entitiesToBalance = 0;
		entitiesBalanced = 0;
		quadsBalanced = 0;
		for (std::map<Types::Type, QuadTree<Entity>>::iterator it = quads.begin(); it != quads.end(); ++it)
		{
			entitiesToBalance += it->second.size();
		}
		balanceIterator = quads.begin();
	}

	++balanceStep;
	for (auto mit = lists.begin(); mit != lists.end(); ++mit)
	{
		if (pauseList.count(mit->first) == 0)
		{
			for (Entity& entity : mit->second)
			{
				entity.update();
			}
		}
	}
	for (Entity *entity : deactivationQueue)
	{
		inactiveEntities.insert(entity->getLink());
	}
	deactivationQueue.clear();

	delayAddEntities();

	clearDeletedEntityList();

	isUpdating = false;
	for (std::function<void()>& f : callAtEndOfUpdate)
	{
		f();
	}
	callAtEndOfUpdate.clear();
}

void Scene::drawEntities(RenderTarget& target, const Rect<int>& bBox, const sf::Vector2f& dtp)
{
	for (std::map<Types::Type, EntityGrid>::iterator it = grids.begin(); it != grids.end(); ++it)
	{
		it->second.draw(target, bBox);
	}
	for (std::map<Types::Type, QuadTree<Entity>>::iterator it = quads.begin(); it != quads.end(); ++it)
	{
		it->second.draw(target, bBox);
#ifdef SL_DEBUG
	if (/*Input::keyHeld(Input::F9) && */it->first == Entity::makeType("Car"))
	{
		//->second->drawAsTree(target, dtp, 640 * 2, 64, bBox);
	}
#endif
	}
	for (auto mit = lists.begin(); mit != lists.end(); ++mit)
	{
		for (Entity& entity : mit->second)
		{
			if (entity.getMask().getBoundingBox().intersects(bBox))
			{
				entity.draw(target);
			}
		}
	}
	target.render();

#ifdef SL_DEBUG
	const int maps = heightMaps.size();
	for (const HeightMap& map : heightMaps)
	{
		map.debugDraw(*this, target);
	}
#endif
}

void Scene::createAndRegisterRealTimePathGrid(std::initializer_list<const char*> types, const std::string& manager)
{
	ASSERT(types.size() > 0);

	for (const char* type : types)
	{
		getWorld()->addGridToPathManager(manager, this, shared<pf::RealTimePathGrid>(*this, type));
	}
}

void Scene::createAndRegisterPrecomputedPathGrid(const Entity::Type& type, const std::string& manager)
{
	auto grid = shared<pf::PrecomputedPathGrid>(*this, type);
	getWorld()->addGridToPathManager(manager, this, grid);
	precomputedPathGrids.insert(std::make_pair(type, grid));
}


void Scene::setWorldCoords(const sf::Vector3i& worldCoords)
{
	this->worldCoords = worldCoords;
}

void Scene::setWorldScale(float scale)
{
	worldScale = scale;
}

void Scene::serializeScene(Serialize& out) const
{
	out.startObject("Scene");
	serializeAbstractScene(out);
	// We should only be issuing a save game at the end of a Game's update to ensure there's no fuckery involving delayedAddEntties, callAfterUpdate, etc ...
	ASSERT(deactivationQueue.empty());
	ASSERT(deleteList.empty());
	ASSERT(callAtEndOfUpdate.empty());
	for (const auto& p : delayedAddEntities)
	{
		ASSERT(p.second.empty());
	}

	out.i("width", width);
	out.i("height", height);
	std::vector<const Entity*> toSerialize;
	std::map<std::string, int> notSerialized;
	auto trySerializeEntity = [&](const Entity *entity) {
		if (entity->shouldSerialize())
		{
			toSerialize.push_back(entity);
		}
		else
		{
			++notSerialized[entity->getType()];
		}
	};
	for (const Entity *entity : updateableGridEntities)
	{
		trySerializeEntity(entity);
	}
	for (const Entity *entity : nonUpdateableGridEntities)
	{
		trySerializeEntity(entity);
	}
	for (const auto& quad : quads)
	{
		quad.second.forEach(trySerializeEntity);
	}
	for (const auto& list : lists)
	{
		for (const Entity& entity : list.second)
		{
			trySerializeEntity(&entity);
		}
	}
	out.startArray("activeEntities", toSerialize.size());
	for (const Entity *entity : toSerialize)
	{
		out.saveSerializable("activeEntity", entity);
	}
	out.endArray("activeEntities");
	out.f("viewPos.x", viewPos.x);
	out.f("viewPos.y", viewPos.y);
	toSerialize.clear();
	for (const Entity& entity : inactiveEntities)
	{
		trySerializeEntity(&entity);
	}
	out.startArray("inactiveEntities", toSerialize.size());
	for (const Entity *entity : toSerialize)
	{
		out.saveSerializable("inactiveEntity", entity);
	}
	out.endArray("inactiveEntities");
	for (const auto& ns : notSerialized)
	{
		Logger::log(Logger::Serialization, Logger::Debug) << "Did not serialize '" << ns.first << "' (x" << ns.second << ")" << std::endl;
	}
	//int delayedAddCount = 0;
	//for (const auto& p : delayedAddEntities)
	//{
	//	for (const Entity& entity : p.second)
	//	{
	//		if (entity.shouldSerialize())
	//		{
	//			++delayedAddCount;
	//		}
	//	}
	//}
	//out.startArray("delayedAddEntities", delayedAddCount);
	//for (const auto& p : delayedAddEntities)
	//{
	//	for (const Entity& entity : p.second)
	//	{
	//		if (entity.shouldSerialize())
	//		{
	//			out.u8("storageType", static_cast<uint8_t>(p.first));
	//			out.saveSerializable("delayedAddEntity", &entity);
	//		}
	//	}
	//}
	//out.endArray("delayedAddEntities");
	// @TODO-SAVE serialize heightmaps
	out.b("isUpdating", isUpdating);
	out.ptr("player", player);
	AutoSerialize::serialize(out, pauseList);
	out.s("type", type);
	out.i("worldCoords.x", worldCoords.x);
	out.i("worldCoords.y", worldCoords.y);
	out.i("worldCoords.z", worldCoords.z);
	out.f("worldScale", worldScale);
	// id ignored purposely
	out.b("mapDrawable?", mapDrawable != nullptr);
	if (mapDrawable)
	{
		auto image = mapDrawable->makeImage();
		out.u16("image.w", image.getSize().x);
		out.u16("image.h", image.getSize().y);
		out.raw("image.data", image.getPixelsPtr(), 4 * image.getSize().x * image.getSize().y);
	}
	// @TODO-SAVE mapDrawable, precomputedPathGrids, groupAIs = ???
	out.endObject("Scene");
}

void Scene::deserializeScene(Deserialize& in)
{
	in.startObject("Scene");
	deserializeAbstractScene(in);
	width = in.i("width");
	height = in.i("height");
	const int activeCount = in.startArray("activeEntities");
	for (int i = 0; i < activeCount; ++i)
	{
		auto entity = loadSerializable<Entity>(in, "activeEntity");
		const StorageType storage = entity->storageType;
		// to ensure that it will get added properly and bug-free
		entity->scene = nullptr;
		entity->storageType = StorageType::InNone;
		switch (storage)
		{
		case StorageType::InList:
			addEntityToList(entity.release());
			break;
		case StorageType::InQuad:
			addEntityToQuad(entity.release());
			break;
		case StorageType::InGrid:
			//addEntityToGrid(entity.release());
			addImmediately(entity.release(), StorageType::InGrid);
			break;
		case StorageType::InGridNoUpdate:
			//addEntityToGridNoUpdate(entity.release());
			addImmediately(entity.release(), StorageType::InGridNoUpdate);
			break;
		}
	}
	in.endArray("activeEntities");
	viewPos.x = in.f("viewPos.x");
	viewPos.y = in.f("viewPos.y");
	const int inactiveCount = in.startArray("inactiveEntities");
	for (int i = 0; i < inactiveCount; ++i)
	{
		auto entity = loadSerializable<Entity>(in, "inactiveEntity");
		inactiveEntities.insert(entity.release()->getLink());
	}
	in.endArray("inactiveEntities");
	//const int delayedAddCount = in.startArray("delayedAddEntities");
	//for (int i = 0; i < inactiveCount; ++i)
	//{
	//	const StorageType storageType = static_cast<StorageType>(in.u8("storageType"));
	//	auto entity = loadSerializable<Entity>(in, "delayedAddEntity");
	//	delayedAddEntities[storageType].insert(entity.release()->getLink());
	//}
	//in.endArray("delayedAddEntities");
	// @TODO-SAVE deserialize heightmaps
	isUpdating = in.b("isUpdating");
	player = static_cast<Player*>(in.ptr("player"));
	AutoDeserialize::deserialize(in, pauseList);
	type = in.s("type");
	worldCoords.x = in.i("worldCoords.x");
	worldCoords.y = in.i("worldCoords.y");
	worldCoords.z = in.i("worldCoords.z");
	worldScale = in.f("worldScale");
	if (in.b("mapDrawable?"))
	{
		const auto imageWidth = in.u16("image.w");
		const auto imageHeight = in.u16("image.h");
		std::vector<uint8_t> buffer(4 * imageWidth * imageHeight);
		in.raw("image.data", buffer.data(), buffer.size());
		sf::Image image;
		image.create(imageWidth, imageHeight, buffer.data());
		makeMap(image);
	}
	in.endObject("Scene");
}




/*  private	 */
void Scene::rectifyBalanceIterator()
{
	//  Reset the iterator to make it valid again in case the std::map has changed.
	balanceIterator = quads.begin();
	//  Then move the iterator back to its previous position.
	for (std::size_t i = 0; balanceIterator != quads.end() && i < quadsBalanced; ++balanceIterator, ++i);
}

void Scene::onHandleEvents(const EventQueue& events)
{
	const Event *ev = nullptr;
	auto handleEventFunc = [&ev](Entity *entity)
	{
		entity->handleEvent(*ev);
	};
	for (int i = 0; i < Event::Category::EventTypeNum; ++i)
	{
		EventIterator eventIt = events.iterator(static_cast<Event::Category>(i));
		while (eventIt.poll(ev))
		{
			for (auto& quad : quads)
			{
				quad.second.forEach(handleEventFunc);
			}
			for (Entity *entity : updateableGridEntities)
			{
				entity->handleEvent(*ev);
			}
			for (auto& list : lists)
			{
				for (Entity &entity : list.second)
				{
					entity.handleEvent(*ev);
				}
			}
		}
	}
}

void Scene::delayAddEntities()
{
	// To get around issues where one entity being created added adds other ones, we do this.
	// You probably shouldn't be adding too many entities during onSceneEnter() anyway... let's hope there's no long chains here.
	std::vector<std::pair<StorageType, Entity*>> toAdd;
	for (auto& storageType : delayedAddEntities)
	{
		for (Entity& entity : storageType.second)
		{
			toAdd.push_back(std::make_pair(storageType.first, &entity));
		}
	}
	for (std::pair<StorageType, Entity*>& entity : toAdd)
	{
		addImmediately(entity.second, entity.first);
	}
	for (auto& storageType : delayedAddEntities)
	{
		if (!storageType.second.empty())
		{
			Logger::log(Logger::Core, Logger::Debug) << "After delayAddEntities() [storageType = " << static_cast<int>(storageType.first) << "] we ended up adding more [" << storageType.second.size() << "] in the process (probably from onSceneEnter()?)";
		}
	}
}

void Scene::clearDeletedEntityList()
{
	std::vector<Entity*> skipped;
	for (Entity *entity : deleteList)
	{
		if (entity->getAliveRef().canDestroy())
		{
			if (entity->storageType == InGrid)
			{
				auto it = std::find(updateableGridEntities.begin(), updateableGridEntities.end(), entity);
				ASSERT(it != updateableGridEntities.end());
				*it = updateableGridEntities.back();
				updateableGridEntities.pop_back();
			}
			else if (entity->storageType == InGridNoUpdate)
			{
				if (entity->storageType == InGrid)
				{
					auto it = std::find(nonUpdateableGridEntities.begin(), nonUpdateableGridEntities.end(), entity);
					ASSERT(it != nonUpdateableGridEntities.end());
					*it = nonUpdateableGridEntities.back();
					nonUpdateableGridEntities.pop_back();
				}

			}
			auto it = precomputedPathGrids.find(entity->getType());
			ASSERT(it == precomputedPathGrids.end() || entity->isStatic());
			const Rect<int> entityBounds = entity->getMask().getBoundingBox();
			delete entity;
			// we must delete it first, as this removes it from all collision-related lists.
			if (it != precomputedPathGrids.end())
			{
				it->second->recompute(entityBounds);
			}
		}
		else
		{
			skipped.push_back(entity);
		}
	}
	deleteList = std::move(skipped);
}

void Scene::recomputeAreaAroundEntityIfNecessary(const Entity& entity)
{
	auto it = precomputedPathGrids.find(entity.getType());
	if (it != precomputedPathGrids.end())
	{
		ASSERT(entity.isStatic());
		it->second->recompute(entity.getMask().getBoundingBox());
	}
}

}// End of sl namespace
