/**
 * @section DESCRIPTION
 * Essentially a class for representing any physical space in the game, be it the town, a shop, etc.
 */
#ifndef SL_PHYSICSSCENE_HPP
#define SL_PHYSICSSCENE_HPP

#include <SFML/Graphics/Rect.hpp>

#include <initializer_list>
#include <functional>
#include <map>
#include <memory>
#include <set>
#include <string>
#include <type_traits>
#include <vector>

#include "Engine/Physics/EntityGrid.hpp"
#include "Engine/Physics/HeightMap.hpp"
#include "Engine/Physics/QuadTree.hpp"
#include "Engine/PoolAllocator.hpp"
#include "Engine/AbstractScene.hpp"
#include "Engine/ScenePos.hpp"
#include "Engine/Types.hpp"
#include "NPC/GroupAI/GroupAI.hpp"
#include "NPC/NPC.hpp"
#include "Pathfinding/PrecomputedPathGrid.hpp"
#include "Town/Map/Map.hpp"
#include "Utility/IntrusiveList.hpp"
#include "Utility/Trig.hpp"
#include "Utility/TypeToString.hpp"

namespace sl
{

enum StorageType : char
{
	InList,
	InQuad,
	InGrid,
	InGridNoUpdate,
	InNone
};

class Entity;
class Player;
class Deserialize;
class Serialize;

namespace map
{
class Icon;
class MapData;
} // map

class Scene : public AbstractScene
{
public:

	/**
	 * Constructs a Scene
	 * @param game The Game this Scene belongs to
	 * @param width How wide the place is
	 * @param height How high the place is
	 */
	Scene(Game *game, unsigned int width, unsigned int height, unsigned int viewWidth, unsigned int viewHeight, const std::string& type = "Scene");

	/**
	 * Destructs the Scene and all Entities created within it.
	 */
	virtual ~Scene();

	/**
	 * Checks for collisions at a certain Entity
	 * @param e The Entity to use for collisions
	 * @param type The Entity Type to check for collisions with
	 * @return The first Entity found in the region (nullptr if none found, so can be treated like a bool)
	 */
	Entity* checkCollision(const Entity& e, const Types::Type& type);

	/**
	 * Checks for collisions with a bounding box
	 * @param e The bounding box to use for collisions
	 * @param type The Entity Type to check for collisions with
	 * @return The first Entity found in the region (nullptr if none found, so can be treated like a bool)
	 */
	Entity* checkCollision(const Rect<int>& bBox, const Types::Type& type);

	template <typename EntityType, typename Mask>
	EntityType* checkCollision(const Mask& mask);

	/**
	 * Fills a list with all Entities that collide with a certain bounding box
	 * @param results The list to fill (does NOT clear beforehand - appends)
	 * @param bBox The bounding box to check for collisions against
	 * @param type The Entity Type to check for collisions with
	 */
	void cull(std::list<Entity*>& results, const Rect<int>& bBox, const Types::Type& type);

	/**
	 * Fills a list with all Entities that collide with a certain Entity
	 * @param results The list to fill (does NOT clear beforehand - appends)
	 * @param bBox The Entity to check for collisions against
	 * @param type The Entity Type to check for collisions with
	 */
	void cull(std::list<Entity*>& results, const Entity& e, const Types::Type& type);

	/**
	 * Fills a list with all Entities that collide with a line between p1 and p2
	 * @param results The list to fill (does NOT clear beforehand - appends)
	 * @param p1 Start point of line
	 * @param p2 End point of line
	 * @param type The Entity type fo check for collisions with
	 */
	void cullFiniteLine(std::list<Entity*>& results, const Vector2f& p1, const Vector2f& p2, const Types::Type& type);

	template <typename EntityType, typename Mask>
	void cull(std::list<EntityType*>& output, const Mask& mask);
	
	enum class RaycastResult
	{
		DidNotMove = 0,
		MovedPartially,
		MovedFully
	};
	/**
	 * Entity / CollisionMask-based raycasting
	 * @param e The Entity to move through space
	 * @param type The types to query against
	 * @param motion The displacement to attempt to move
	 * @param result The resulting vector (when moving mask as far as it could)
	 * @return Whether the object moved, stayed still, or partially moved (see enum RaycastResult)
	 */
	RaycastResult raycast(const Entity& e, const std::initializer_list<const char*>& types, const Vector2f& motion, Vector2f& result);

	/**
	* Line-based raycasting
	* @param pos The starting point to raycast from
	* @param type The types to query against
	* @param motion The displacement to attempt to move
	* @param result The resulting vector (when moving mask as far as it could)
	* @return Whether the object moved, stayed still, or partially moved (see enum RaycastResult)
	*/
	RaycastResult raycast(const Vector2f& pos, const std::initializer_list<const char*>& types, const Vector2f& motion, Vector2f& result);

	 /**
	 * Adds a static Entity into the Grid system. O(1) collision detection
	 * @param obj The Entity to add
	 */
	void addEntityToGrid(Entity *obj);

	/**
	 * Adds a static Entity into the Grid system. O(1) collision detection
	 * The Entity won't even have update() called on it for added efficiency
	 * only draw, so only use this for static geometry and things like that.
	 * @param obj The Entity to add
	 */
	void addEntityToGridNoUpdate(Entity *obj);

	/**
	 * Adds an Entity into the QuadTree (tree) system. O(log(n)) collision detection
	 * @param obj The Entity to add
	 */
	void addEntityToQuad(Entity *obj);

	/**
	 * Adds an Entity into the List system. O(n) collision detection
	 * @param obj The Entity to add
	 */
	void addEntityToList(Entity *obj);

	/**
	 * Adds an Entity instantly (instead of waiting for the next tick like the above few functions
	 * @param obj The Entity to add
	 * @param storageType The storage container type (list, quad, grid) to insert into
	 */
	void addImmediately(Entity *obj, StorageType storageType);

	/**
	 * Gets the width of the level
	 * @return The width
	 */
	int getWidth() const;

	/**
	 * Gets the height of the level
	 * @return The height
	 */
	int getHeight() const;

	/**
	 * @return tile size of scene in pixels (assumes same for vert/horiz)
	 */
	int getTileSize() const;

	/**
	 * Returns where the camera should look in the Town.
	 * @return The vector representing where the camera should be rendered.
	 */
	const sf::Vector2f getView() const override;
	/**
	 * Changes where the camera looks in the Town.
	 * @param x The x position
	 * @param y The y position
	 */
	void setView(int x, int y) override;

	//please only ever call this from Entity::destroy()
	void deleteEntity(Entity *entity);

	void activateEntity(Entity *entity);

	void deactivateEntity(Entity *entity);

	void transferEntity(Entity *entity);
	/**
	 * Adds a heightmap to the scene - the order maps are added into matters,
	 * so ones added later override earlier ones
	 * @param heightMap The HeightMap to add
	 */
	void addHeightMap(HeightMap *heightMap);

	/*
	 * @param x The x pos to check at (in pixels)
	 * @param y The y pos to check at (in pixels)
	 * @return The heightmap value at the given pos
	 */
	int getHeightAt(int x, int y) const;
	inline int getHeightAt(const Vector2f& pos) const
	{
		return getHeightAt(pos.x, pos.y);
	}
	/*
	 * @param area The area to check for
	 * @return The highest heightmap value inside the input rect
	 */
	int getHeightAt(const Rect<int>& area) const;


	// @todo figure out something better than this
	Player* getPlayer() const;
	// please for now only call this from Player
	void setPlayer(Player *player);

	void pause(const std::string& type);

	void unpause(const std::string& type);

	/**
	 * @return a non-owning reference to the minimap drawable, if available, otherwise nullptr
	 */
	map::Map* getMap() const;

	void makeMap(const map::MapData& mapData);

	void makeMap(const sf::Image& mapImage);

	const std::string& getType() const;

	template <typename Activity>
	void getActivities(NPC& npc, std::vector<Activity*>& results);

	template <typename Activity>
	void registerActivity(typename std::unique_ptr<Activity> activity);

	/** 
	 * @return The position of the top-left corner of the level in worldspace (see comment at top of World.hpp)
	 */
	const sf::Vector3i& getWorldCoords() const;

	float getWorldScale() const;

	int getSceneID() const;

	int instanceCount(const std::string& id) const;

protected:
	/**
	 * Updates all Entities in the level
	 * YOU MUST CALL THIS IF YOU WANT STUFF TO UPDATE IN SUBCLASSES
	 * The alternative was to define AbstractScene::onUpdate() in this class and have
	 * virtual Scene::onOnUpdate() which would be called in onUpdate()
	 * along with doing the things in this method but that just seemed kind of
	 * retarded having onOnOnOnOnOnWhatever()
	 * THIS ALSO UPDATE ACTIVITIES!
	 */
	void updateEntities();

	/**
	 * Draws all Entities in the level within the bounding box.
	 * YOU MUST CALL THIS IF YOU WANT STUFF TO DRAW IN SUBCLASSES.
	 * See the comment above this for updateEntities for an explanation.
	 * @param target The RenderTarget to render to
	 * @param bBox The bounding box to use for the frustrum
	 */
	void drawEntities(RenderTarget& target, const Rect<int>& bBox, const sf::Vector2f& dtp = sf::Vector2f(0, 0));

	/**
	 * Helper function to create and register path grids that query the level using collision detection
	 * instead of a pre-computed grid
	 * @param types A list of Entity types
	 * @param manager The name to register the PathManager to
	 */
	void createAndRegisterRealTimePathGrid(std::initializer_list<const char*> types, const std::string& manager);

	void createAndRegisterPrecomputedPathGrid(const Entity::Type& type, const std::string& manager);

	void setWorldCoords(const sf::Vector3i& worldCoords);

	void setWorldScale(float scale);

	void serializeScene(Serialize& out) const;

	void deserializeScene(Deserialize& in);


	//!How wide the level is
	int width;
	//!How high the level is
	int height;
	//!How big the cells used for the Grid system are
	int gridSize;
	//!Entities stored in grids that should still be updated
	std::vector<Entity*> updateableGridEntities;
	//!Entities stored in grids that should not be updated (this is here for serialization!)
	std::vector<Entity*> nonUpdateableGridEntities;
	//!The Entities stored in Grids
	std::map<Types::Type, EntityGrid> grids;
	//!The Entities stored in QuadTree (tree)s
	std::map<Types::Type, QuadTree<Entity>> quads;
	//!The Entities stored in Lists
	std::map<Types::Type, IntrusiveList<Entity, IntrusiveListOwnershipType::Owned>> lists;
	struct EntityStorageShortcut
	{
		EntityGrid *grid = nullptr;
		QuadTree<Entity> *quad = nullptr;
		IntrusiveList<Entity, IntrusiveListOwnershipType::Owned> *list = nullptr;
	};
	//! Quick lookup into the 3 containers to avoid 3 map lookups (TODO: why not store them directly here to avoid more indirection?)
	std::map<Types::Type, EntityStorageShortcut> entitiesLookup;
	//!The bucket size used in the QuadTree (tree) system
	std::size_t bucketSize;
	//!Frequency of QuadTree (tree) balancing
	std::size_t balanceRate;
	//!Represents where the camera should look in the town. Position is where the middle of the screen is, not the top left.
	sf::Vector2f viewPos;


private:
	//non-copyable
	Scene(const Scene&) = delete;
	Scene& operator=(const Scene&) = delete;
	//non-moveable
	Scene(Scene&&) = delete;
	Scene& operator=(Scene&&) = delete;

	void onHandleEvents(const EventQueue& events) override;


	inline void rectifyBalanceIterator();

	void delayAddEntities();

	void clearDeletedEntityList();

	void recomputeAreaAroundEntityIfNecessary(const Entity& entity);

	void createMapDrawable();


	//!Updates since last QuadTree (tree) rebalance
	std::size_t balanceStep;
	std::size_t entitiesToBalance;
	std::size_t entitiesBalanced;
	std::size_t quadsBalanced;
	std::map<Types::Type, QuadTree<Entity>>::iterator balanceIterator;
	IntrusiveList<Entity, IntrusiveListOwnershipType::Owned> inactiveEntities;
	//! To ensure that when we deactivate ourselves/etc we don't move the Entity to inactiveEntities until after we are done updating, since then the list fucks up if we're iteraitng it. This list retains no ownership.
	std::vector<Entity*> deactivationQueue;
	//! Entities to delete after we're done a scene update - maintains indirect ownership. Will unlink any entities before destroying.
	std::vector<Entity*> deleteList;
	IntrusiveList<HeightMap, IntrusiveListOwnershipType::Owned> heightMaps;
	std::vector<std::function<void()>> callAtEndOfUpdate;
	bool isUpdating;
	Player *player;
	//! Entities to add next step (to avoid weird behavior when things happen the frame they're added)
	std::map<StorageType, IntrusiveList<Entity, IntrusiveListOwnershipType::Owned>> delayedAddEntities;
	// why is this even here?
	//std::map<std::string, std::shared_ptr<pf::PathGrid>> pathGrids;
	std::set<std::string> pauseList;
	std::string type;
	sf::Vector3i worldCoords;
	float worldScale;
	int id;
	std::unique_ptr<map::Map> mapDrawable;
	std::map<Entity::Type, std::shared_ptr<pf::PrecomputedPathGrid>> precomputedPathGrids;

	static int idCount;

protected://@todo make access to this a method?
	std::map<std::string, std::vector<std::unique_ptr<IGroupAI>>> groupAIs;
};


/**
 * @return The position of {\code pos} within {\code scene} in WORLDSPACE
 */
static inline sf::Vector3f worldPosition(const Scene& scene, const Vector2f& pos)
{
	return sf::Vector3f(pos.x, pos.y, 0.f) * (scene.getWorldScale() / scene.getTileSize()) + sf::Vector3f(scene.getWorldCoords());
}

static inline Vector2f worldPosition2(const Scene& scene, const Vector2f& pos)
{
	const sf::Vector3f worldPos = worldPosition(scene, pos);
	return Vector2f(worldPos.x, worldPos.y);
}

static inline Rect<int> worldRegion(const Scene& scene)
{
	const float sceneToWorld = scene.getWorldScale() / scene.getTileSize();
	return Rect<int>(scene.getWorldCoords().x, scene.getWorldCoords().y, scene.getWidth() * sceneToWorld, scene.getHeight() * sceneToWorld);
}

static inline Vector2f worldPosToScenePos(const Scene& scene, const Vector2f& worldPos)
{
	return Vector2f(worldPos.x - scene.getWorldCoords().x, worldPos.y - scene.getWorldCoords().y) * (scene.getTileSize() / scene.getWorldScale());
}

static inline Rect<int> worldPosToScenePos(const Scene& scene, const Rect<int>& worldRegion)
{
	const float worldToSceneScale = (scene.getTileSize() / scene.getWorldScale());
	return worldPosToScenePos(scene, worldRegion.topLeft()).to<int>() + Rect<int>(worldToSceneScale * worldRegion.width, worldToSceneScale * worldRegion.height);
}

/**
 * @param t Some Entity-like object (getPos()/getScene() required)
 * @Param u Another Entity-like object
 * @return the distance IN WORLDSPACE (see comment at top of World.hpp) between t and u
 */
template <typename T, typename U>
float worldDistance(const T& t, const U& u)
{
	const sf::Vector3f tPos = worldPosition(*t.getScene(), t.getPos());
	const sf::Vector3f uPos = worldPosition(*u.getScene(), u.getPos());
	const sf::Vector3f dif = tPos - uPos;
	return sqrtf(dif.x*dif.x + dif.y*dif.y + dif.z*dif.z);
}

template <typename Activity>
void Scene::getActivities(NPC& npc, std::vector<Activity*>& results)
{
	//static_assert(std::is_base_of(IGroupAI, Activity), "Activity must be a GroupAI");

	auto activities = groupAIs.find(TypeToString<Activity>());

	if (activities != groupAIs.end())
	{
		for (std::unique_ptr<IGroupAI>& activity : activities->second)
		{
			// HACK TO COMPILE ON GCC
			if (pointDistance(npc.getPos(), activity->getPos()) <= activity->getRadius())
			{
				results.push_back(static_cast<Activity*>(activity.get()));
			}
		}
	}
}

template <typename Activity>
void Scene::registerActivity(typename std::unique_ptr<Activity> activity)
{
	//static_assert(std::is_base_of(IGroupAI, Activity), "Activity must be a GroupAI");
	ASSERT(activity != nullptr);
	groupAIs[TypeToString<Activity>()].push_back(std::move(activity));
}

template <typename EntityType, typename Mask>
EntityType* Scene::checkCollision(const Mask& mask)
{
	return static_cast<EntityType*>(checkCollision(mask, TypeToString<EntityType>()));
}

template <typename EntityType, typename Mask>
void Scene::cull(std::list<EntityType*>& output, const Mask& mask)
{
	static_assert(std::is_base_of<Entity, EntityType>::value, "Can only query subclasses of Entity");
	cull(reinterpret_cast<std::list<Entity*>&>(output), mask, TypeToString<EntityType>());
}

// helper for querying based on areas
static inline Rect<int> rectAround(const Vector2f& pos, int size)
{
	return Rect<int>(pos.x - size, pos.y - size, 2 * size, 2 * size);
}

}// End of sl namespace

#endif
