/**
 * Assumes you are calling these inside the sl namespace.
 */
#ifndef SL_AUTODEFINEOBJECTSERIALIZATION_HPP
#define SL_AUTODEFINEOBJECTSERIALIZATION_HPP

#include "Engine/Serialization/Serialization.hpp"
#include "Engine/Serialization/AutoDeserialize.hpp"
#include "Engine/Serialization/AutoSerialize.hpp"

// TOOD: get the macro mapping in MacroUtil.hpp working and use this to let us write (Foo, x, y) instead of (Foo, &Foo::x, &Foo::y)

namespace sl
{
namespace AutoDefineObjectSerialization
{
template <typename C, typename T>
inline void deserializeFields(Deserialize& in, C& obj, T C::*field)
{
	AutoDeserialize::deserialize(in, obj.*field);
}

template <typename C, typename T, typename... Args>
inline void deserializeFields(Deserialize& in, C& obj, const T& field, Args... args)
{
	deserializeFields(in, obj, field);
	deserializeFields(in, obj, args...);
}

template <typename C, typename T>
T deserializeByTypeFromField(Deserialize& in, T C::*field)
{
	return AutoDeserialize::create<T>(in);
}

template <typename C, typename T>
T deserializeByTypeFromField(Deserialize& in, T (C::*field)() const)
{
	return AutoDeserialize::create<T>(in);
}

template <typename C, typename... Args>
inline C deserializeByConstructor(Deserialize& in, Args... args)
{
	return C{[&in](const auto& field) {
		return deserializeByTypeFromField(in, field);
	}(args)...};
}

template <typename C, typename T>
inline void serializeFields(Serialize& out, const C& obj, T (C::*field)() const)
{
	AutoSerialize::serialize(out, (obj.*field)());
}
template <typename C, typename T>
inline void serializeFields(Serialize& out, const C& obj, T C::*field)
{
	AutoSerialize::serialize(out, obj.*field);
}
template <typename C, typename T, typename... Args>
inline void serializeFields(Serialize& out, const C& obj, const T& field, Args... args)
{
	serializeFields(out, obj, field);
	serializeFields(out, obj, args...);
}
} // AutoDefineObjectSerialization
} // sl

//#define DEF_SERIALIZE_OBJ_AS_FIELD(field) val.field
#define DEF_DESERIALIZE_OBJ_FIELDS_IMPL(T, ...) namespace AutoDeserialize { \
	void deserialize(Deserialize& in, T& val) { \
		in.startObject(#T); \
		::sl::AutoDefineObjectSerialization::deserializeFields(in, val, __VA_ARGS__); /*MU_VA_MAP(DEF_SERIALIZE_OBJ_AS_FIELD, __VA_ARGS__));*/ \
		in.endObject(#T); \
	} \
	}

#define DEF_SERIALIZE_OBJ_FIELDS_IMPL(T, ...) namespace AutoSerialize { \
	void serialize(Serialize& out, const T& val) { \
		out.startObject(#T); \
		::sl::AutoDefineObjectSerialization::serializeFields(out, val, __VA_ARGS__); /*MU_VA_MAP(DEF_SERIALIZE_OBJ_AS_FIELD, __VA_ARGS__));*/ \
		out.endObject(#T); \
	} \
	}

#define DEF_DESERIALIZE_OBJ_CTOR_IMPL(T, ...) namespace AutoDeserialize { \
	template <> \
	T create(Deserialize& in) { \
		in.startObject(#T); \
		T ret = ::sl::AutoDefineObjectSerialization::deserializeByConstructor<T>(in, __VA_ARGS__); \
		in.endObject(#T); \
		return ret; \
	} \
	void deserialize(Deserialize& in, T& val) { \
		val = create<T>(in); \
	} \
	}

#define DEF_SERIALIZE_OBJ_FIELDS(T, ...) \
	DEF_DESERIALIZE_OBJ_FIELDS_IMPL(T, __VA_ARGS__) \
	DEF_SERIALIZE_OBJ_FIELDS_IMPL(T, __VA_ARGS__)

#define DEF_SERIALIZE_OBJ_CTOR(T, ...) \
	DEF_DESERIALIZE_OBJ_CTOR_IMPL(T, __VA_ARGS__) \
	DEF_SERIALIZE_OBJ_FIELDS_IMPL(T, __VA_ARGS__)

#define DEF_SERIALIZE_OBJ_MANUAL(T) namespace AutoDeserialize { \
	void deserialize(Deserialize& in, T& val) { \
		val.deserialize(in); \
	} \
	} \
	namespace AutoSerialize { \
	void serialize(Serialize& in, const T& val) { \
		val.serialize(in); \
	} \
	}

#endif // SL_AUTODEFINEOBJECTSERIALIZATION_HPP