#ifndef SL_AUTODESERIALIZE_HPP
#define SL_AUTODESERIALIZE_HPP

#include "Engine/Serialization/Serialization.hpp"
#include "Utility/Assert.hpp"
#include "Utility/Grid.hpp"
#include "Utility/Rect.hpp"
#include "Utility/Vector.hpp"

#include <iterator>
#include <set>
#include <type_traits>

namespace sl
{
namespace AutoDeserialize
{

template <typename... Args>
static void fields(Deserialize& in, Args&... args)
{
	ASSERT(sizeof...(args) == in.u8("fieldcount"));
	impl::fields(in, args...);
}

namespace impl
{
template <typename T>
inline void fields(Deserialize& in, T& val)
{
	deserialize(in, val);
}

template <typename T, typename... Args>
inline void fields(Deserialize& in, T& val, Args&... args)
{
	fields(in, val);
	fields(in, args...);
}
} // impl

// specialize this if default ctor not available
template <typename T>
inline T create(Deserialize& in)
{
	T val;
	deserialize(in, val);
	return val;
}

inline void deserialize(Deserialize& in, bool& val)
{
	val = in.u8("");
}

inline void deserialize(Deserialize& in, int& val)
{
	val = in.i("");
}

inline void deserialize(Deserialize& in, float& val)
{
	val = in.f("");
}

inline void deserialize(Deserialize& in, std::string& val)
{
	val = in.s("");
}

inline void deserialize(Deserialize& in, uint8_t& val)
{
	val = in.u8("");
}

inline void deserialize(Deserialize& in, uint16_t& val)
{
	val = in.u16("");
}

inline void deserialize(Deserialize& in, uint32_t& val)
{
	val = in.u32("");
}

inline void deserialize(Deserialize& in, uint64_t& val)
{
	val = in.u64("");
}

// make sure this is non-owning only! (which it should be...)
template <typename T>
inline void deserialize(Deserialize& in, T*& val)
{
	// need to match on a template pointer instead of Serializable*& because of weird implicit conversion rules
	static_assert(std::is_base_of<Serializable, T>::value, "This is only for Serializable pointers");
	val = static_cast<T*>(in.ptr(""));
}

template <typename T>
inline void deserialize(Deserialize& in, Rect<T>& val)
{
	deserialize(in, val.left);
	deserialize(in, val.top);
	deserialize(in, val.width);
	deserialize(in, val.height);
}

template <typename T>
inline void deserialize(Deserialize& in, Vector2<T>& val)
{
	deserialize(in, val.x);
	deserialize(in, val.y);
}

template <typename T, size_t N>
inline void deserialize(Deserialize& in, std::array<T, N>& val)
{
	const int n = in.startArray("array<T,N>");
	for (auto& elem : val)
	{
		deserialize(in, elem);
	}
	in.endArray("array<T,N>");
}

template <typename T, template <typename, typename> class ContainerOfT>
void deserialize(Deserialize& in, ContainerOfT<T, std::allocator<T>>& val)
{
	val.clear();
	const int n = in.startArray("generic_container<T>");
	//val.reserve(n);
	auto inserter = std::inserter(val, std::end(val));
	for (int i = 0; i < n; ++i)
	{
		*inserter++ = create<T>(in);
	}
	in.endArray("generic_container<T>");
}

template <typename U, typename V>
void deserialize(Deserialize& in, std::pair<U, V>& val)
{
	deserialize(in, val.first);
	deserialize(in, val.second);
}

template<typename U, typename V>
std::pair<U, V> create(Deserialize& in)
{
	U u = create<U>(in);
	V v = create<V>(in);
	return std::make_pair(std::move(u), std::move(v));
}

template <typename K, typename V>
void deserialize(Deserialize& in, std::map<K, V>& val)
{
	val.clear();
	const int n = in.startArray("map<K,V>");
	for (int i = 0; i < n; ++i)
	{
		K k = create<K>(in);
		V v = create<V>(in);;
		val.insert(std::make_pair(std::move(k), std::move(v)));
	}
	in.endArray("map<K,V>");
}

template <typename T>
void deserialize(Deserialize& in, std::set<T>& val)
{
	val.clear();
	const int n = in.startArray("set<T>");
	for (int i = 0; i < n; ++i)
	{
		val.insert(create<T>(in));
	}
	in.endArray("set<T>");
}

template <typename T, int N>
void deserialize(Deserialize& in, T (&arr)[N])
{
	const int n = in.startArray("T arr[N]");
	ASSERT(n == N);
	for (int i = 0; i < N; ++i)
	{
		deserialize(in, arr[i]);
	}
	in.endArray("T arr[N]");
}

template <typename T>
void deserialize(Deserialize& in, std::unique_ptr<T>& val)
{
	static_assert(std::is_base_of<Serializable, T>::value, "Can only auto-deserialize Serializables, otherwise we don't know how to allocate them");
	val.reset(static_cast<T*>(in.loadSerializable("AutoSerialized_Serializable").release()));
}

// @TODO optimise to use RLE? have a Grid<bool> specializtion too?
template <typename T>
void deserialize(Deserialize& in, Grid<T>& grid)
{
	in.startObject("Grid<T>");
	const int w = in.i("w");
	const int h = in.i("h");
	grid = Grid<T>(w, h);
	for (int y = 0; y < h; ++y)
	{
		for (int x = 0; x < w; ++x)
		{
			deserialize(in, grid.at(x, y));
		}
	}
	in.endObject("Grid<T>");
}

} // AutoDeserialize
} // sl

#endif // SL_AUTODESERIALIZE_HPP
