#ifndef SL_AUTOSERIALIZE_HPP
#define SL_AUTOSERIALIZE_HPP

#include "Engine/Serialization/Serialization.hpp"

// @TODO decide where to put Rect/Vector serialization. here? in those files? elsewhere?
#include "Utility/Grid.hpp"
#include "Utility/Rect.hpp"
#include "Utility/Vector.hpp"

#include <set>

namespace sl
{
namespace AutoSerialize
{

template <typename... Args>
inline void fields(Serialize& out, const Args&... args)
{
	static_assert(sizeof...(args) < 256);
	out.u8("fieldcount", sizeof...(args));
	impl::fields(out, args...);
}

namespace impl
{
	template <typename T>
	inline void fields(Serialize& out, const T& val)
	{
		serialize(out, val);
	}

	template <typename T, typename... Args>
	inline void fields(Serialize& out, const T& val, const Args&... args)
	{
		fields(out, val);
		fields(out, args...);
	}
} // impl

inline void serialize(Serialize& out, bool val)
{
	out.u8("", val);
}

inline void serialize(Serialize& out, int val)
{
	out.i("", val);
}

inline void serialize(Serialize& out, float val)
{
	out.f("", val);
}

inline void serialize(Serialize& out, const std::string& val)
{
	out.s("", val);
}

inline void serialize(Serialize& out, uint8_t val)
{
	out.u8("", val);
}

inline void serialize(Serialize& out, uint16_t val)
{
	out.u16("", val);
}

inline void serialize(Serialize& out, uint32_t val)
{
	out.u32("", val);
}

inline void serialize(Serialize& out, uint64_t val)
{
	out.u64("", val);
}

// make sure this is non-owning only! (which it should be...)
inline void serialize(Serialize& out, const Serializable *val)
{
	out.ptr("", val);
}

template <typename T>
inline void serialize(Serialize& out, const Rect<T>& val)
{
	serialize(out, val.left);
	serialize(out, val.top);
	serialize(out, val.width);
	serialize(out, val.height);
}

template <typename T>
inline void serialize(Serialize& out, const Vector2<T>& val)
{
	serialize(out, val.x);
	serialize(out, val.y);
}

template <typename T, size_t N>
inline void serialize(Serialize& out, const std::array<T, N>& val)
{
	out.startArray("array<T,N>", N);
	for (const auto& elem : val)
	{
		serialize(out, elem);
	}
	out.endArray("array<T,N>");
}

template <typename T, template <typename, typename> class ContainerOfT>
void serialize(Serialize& out, const ContainerOfT<T, std::allocator<T>>& val)
{
	out.startArray("generic_container<T>", val.size());
	for (const auto& elem : val)
	{
		serialize(out, elem);
	}
	out.endArray("generic_container<T>");
}

template <typename U, typename V>
void serialize(Serialize& out, const std::pair<U, V>& val)
{
	serialize(out, val.first);
	serialize(out, val.second);
}

template <typename K, typename V>
void serialize(Serialize& out, const std::map<K, V>& val)
{
	out.startArray("map<K,V>", val.size());
	for (const auto& elem : val)
	{
		serialize(out, elem.first);
		serialize(out, elem.second);
	}
	out.endArray("map<K,V>");
}

template <typename T>
void serialize(Serialize& out, const std::set<T>& val)
{
	out.startArray("set<T>", val.size());
	for (const T& elem : val)
	{
		serialize(out, elem);
	}
	out.endArray("set<T>");
}

template <typename T, int N>
void serialize(Serialize& out, const T (&arr)[N])
{
	out.startArray("T arr[N]", N);
	for (int i = 0; i < N; ++i)
	{
		serialize(out, arr[i]);
	}
	out.endArray("T arr[N]");
}

template <typename T>
void serialize(Serialize& out, const std::unique_ptr<T>& val)
{
	static_assert(std::is_base_of<Serializable, T>::value, "Can only auto-serialize Serializables, otherwise we don't know how to allocate them");
	out.saveSerializable("AutoSerialized_Serializable", val.get());
}

// @TODO optimise to use RLE? have a Grid<bool> specializtion too?
template <typename T>
void serialize(Serialize& out, const Grid<T>& grid)
{
	out.startObject("Grid<T>");
	out.i("w", grid.getWidth());
	out.i("h", grid.getHeight());
	for (int y = 0; y < grid.getHeight(); ++y)
	{
		for (int x = 0; x < grid.getWidth(); ++x)
		{
			serialize(out, grid.at(x, y));
		}
	}
	out.endObject("Grid<T>");
}

} // AutoSerialize
} // sl

#endif // SL_AUTOSERIALIZE_HPP
