#include "Engine/Serialization/BinaryDeserialize.hpp"

#include "Engine/Serialization/AutoDeserialize.hpp"
#include "Engine/Serialization/Serializable.hpp"
#include "Utility/Assert.hpp"

namespace sl
{

BinaryDeserialize::BinaryDeserialize(const std::string& fname, std::map<std::string, std::function<std::unique_ptr<Serializable>()>> ptrAllocators)
	:in(fname.c_str(), std::ios::binary)
	,debug(true)
	,ptrAllocators(std::move(ptrAllocators))
	,world(nullptr)
#ifdef SL_DEBUG
	,debugFile(fname + "deserialize_tags.txt")
#endif // SL_DEBUG
{
	//uint32_t ptrCountOffset;
	//in.read(reinterpret_cast<char*>(&ptrCountOffset), sizeof(uint32_t));
	//in.seekg(ptrCountOffset);
	//std::map<std::string, int> ptrCounts;
	//AutoDeserialize::deserialize(*this, ptrCounts);
	//for (const std::pair<std::string, int>& typeCount : ptrCounts)
	//{
	//	auto& available = availablePtrs[typeCount.first];
	//	const std::function<std::unique_ptr<Serializable>()>& allocator = ptrAllocators.at(typeCount.first);
	//	for (int i = 0; i < typeCount.second; ++i)
	//	{
	//		available.push_back(allocator());

	//	}
	//}
}

uint8_t BinaryDeserialize::u8(const std::string& name)
{
	debugField(name, "u8");
	uint8_t ret;
	in.read(reinterpret_cast<char*>(&ret), sizeof(ret));
	return ret;
}

uint16_t BinaryDeserialize::u16(const std::string& name)
{
	debugField(name, "u16");
	uint16_t ret;
	in.read(reinterpret_cast<char*>(&ret), sizeof(ret));
	return ret;
}

uint32_t BinaryDeserialize::u32(const std::string& name)
{
	debugField(name, "u32");
	uint32_t ret;
	in.read(reinterpret_cast<char*>(&ret), sizeof(ret));
	return ret;
}

uint64_t BinaryDeserialize::u64(const std::string& name)
{
	debugField(name, "u64");
	uint64_t ret;
	in.read(reinterpret_cast<char*>(&ret), sizeof(ret));
	return ret;
}

bool BinaryDeserialize::b(const std::string& name)
{
	debugField(name, "b");
	uint8_t ret;
	in.read(reinterpret_cast<char*>(&ret), sizeof(ret));
	ASSERT(ret == 0 || ret == 1);
	return ret != 0;
}

int BinaryDeserialize::i(const std::string& name)
{
	debugField(name, "i");
	int32_t ret;
	in.read(reinterpret_cast<char*>(&ret), sizeof(ret));
	return ret;
}

float BinaryDeserialize::f(const std::string& name)
{
	debugField(name, "f");
	float ret;
	in.read(reinterpret_cast<char*>(&ret), sizeof(ret));
	return ret;
}

std::string BinaryDeserialize::s(const std::string& name)
{
	debugField(name, "s");
	uint16_t len;
	in.read(reinterpret_cast<char*>(&len), sizeof(len));
	std::string val(len, '?');
	in.read(const_cast<char*>(val.data()), len);
	return std::move(val);
}

uint32_t BinaryDeserialize::raw(const std::string& name, void *val, uint32_t len)
{
	debugField(name, "raw");
	uint32_t foundLen;
	in.read(reinterpret_cast<char*>(&foundLen), sizeof(foundLen));
	ASSERT(len == foundLen);
	const int readLen = len < foundLen ? len : foundLen;
	in.read(reinterpret_cast<char*>(val), readLen);
	return readLen;
}

Serializable* BinaryDeserialize::ptr(const std::string& name)
{
	debugField(name, "ptr");
	uint64_t serializedPtr;
	in.read(reinterpret_cast<char*>(&serializedPtr), sizeof(serializedPtr));
	// type is only stored for non-null ptrs
	if (serializedPtr == 0)
	{
		return nullptr;
	}
	const std::string type = s(name + "_Type");
	auto it = ptrs.find(serializedPtr);
	if (it == ptrs.end())
	{
		ASSERT(undeserializedPtrs.count(serializedPtr) == 0);
		std::unique_ptr<Serializable> ptr = ptrAllocators.at(type)();
		it = ptrs.insert(std::make_pair(serializedPtr, ptr.get())).first;
		undeserializedPtrs.insert(std::make_pair(serializedPtr, std::move(ptr)));
	}
	return it->second;
}

std::unique_ptr<Serializable> BinaryDeserialize::loadSerializable(const std::string& name)
{
	debugField(name, "Serializable");
	uint64_t serializedPtr;
	in.read(reinterpret_cast<char*>(&serializedPtr), sizeof(uint64_t));
	const std::string type = s(name + "_Type");
	std::unique_ptr<Serializable> ret;
	auto it = undeserializedPtrs.find(serializedPtr);
	if (it == undeserializedPtrs.end())
	{
		ASSERT(ptrs.count(serializedPtr) == 0);
		ret = ptrAllocators.at(type)();
		ptrs.insert(std::make_pair(serializedPtr, ret.get()));
	}
	else
	{
		ret = std::move(it->second);
		undeserializedPtrs.erase(it);
	}
	ret->deserialize(*this);
	return std::move(ret);
}

int BinaryDeserialize::startArray(const std::string& name)
{
	debugField(name, "<array>");
	debugStack.push(std::make_pair(StackType::Array, name));
	return i(name);
}

void BinaryDeserialize::endArray(const std::string& name)
{
	// check to see if deserialization query is well-formed
	ASSERT(debugStack.top().first == StackType::Array);
	ASSERT(debugStack.top().second == name);
	debugStack.pop();
	// check to see if the data you are reading from matches
	debugField(name, "</array>");
}

void BinaryDeserialize::startObject(const std::string& name)
{
	debugField(name, "<object>");
	debugStack.push(std::make_pair(StackType::Object, name));
}

void BinaryDeserialize::endObject(const std::string& name)
{
	// check to see if deserialization query is well-formed
	ASSERT(debugStack.top().first == StackType::Object);
	ASSERT(debugStack.top().second == name);
	debugStack.pop();
	// check to see if the data you are reading from matches
	debugField(name, "</object>");
}

void BinaryDeserialize::finish()
{
	// make sure we loaded all deserialized objects
	ASSERT(undeserializedPtrs.empty());
	// ensure query is well-formed
	ASSERT(debugStack.empty());
	in.close();
}

World* BinaryDeserialize::getWorld() const
{
	return world;
}

void BinaryDeserialize::setWorld(World *world)
{
	this->world = world;
}


// private
void BinaryDeserialize::debugField(const std::string& name, const std::string& type)
{
	if (debug)
	{
		const std::string tag = name + "|" + type;
#ifdef SL_DEBUG
		for (int i = 0; i < debugStack.size(); ++i)
		{
			debugFile << "\t";
		}
		debugFile << tag << std::endl;
#endif // SL_DEBUG
		uint16_t len;
		in.read(reinterpret_cast<char*>(&len), sizeof(len));
		std::string found(len, '?');
		in.read(const_cast<char*>(found.data()), len);
		// @TODO throw bad-save-file exception instead?
		ASSERT(tag == found);
	}
}

} // sl