#include "Engine/Serialization/BinarySerialize.hpp"

#include "Engine/Serialization/Serializable.hpp"
#include "Utility/Assert.hpp"

namespace sl
{

BinarySerialize::BinarySerialize(const std::string& fname)
	:out(fname.c_str(), std::ios::binary)
	,debug(true)
#ifdef SL_DEBUG
	,debugFile(fname + "serialize_tags.txt")
#endif // SL_DEBUG
{
}

BinarySerialize::~BinarySerialize()
{
	if (out.is_open())
	{
		finish();
	}
}

void BinarySerialize::u8(const std::string& name, uint8_t val)
{
	debugField(name, "u8");
	out.write(reinterpret_cast<const char*>(&val), sizeof(uint8_t));
}

void BinarySerialize::u16(const std::string& name, uint16_t val)
{
	debugField(name, "u16");
	out.write(reinterpret_cast<const char*>(&val), sizeof(uint16_t));
}

void BinarySerialize::u32(const std::string& name, uint32_t val)
{
	debugField(name, "u32");
	out.write(reinterpret_cast<const char*>(&val), sizeof(uint32_t));
}

void BinarySerialize::u64(const std::string& name, uint64_t val)
{
	debugField(name, "u64");
	out.write(reinterpret_cast<const char*>(&val), sizeof(uint64_t));
}

void BinarySerialize::b(const std::string& name, bool val)
{
	debugField(name, "b");
	uint8_t b = val ? 1 : 0;
	out.write(reinterpret_cast<const char*>(&b), sizeof(uint8_t));
}

void BinarySerialize::i(const std::string& name, int val)
{
	debugField(name, "i");
	const int32_t i = val;
	out.write(reinterpret_cast<const char*>(&i), sizeof(int32_t));
}

void BinarySerialize::f(const std::string& name, float val)
{
	debugField(name, "f");
	out.write(reinterpret_cast<const char*>(&val), sizeof(float));
}

void BinarySerialize::s(const std::string& name, const std::string& val)
{
	debugField(name, "s");
	ASSERT(val.size() <= UINT16_MAX);
	const uint16_t len = static_cast<uint16_t>(val.size());
	out.write(reinterpret_cast<const char*>(&len), sizeof(uint16_t));
	out.write(val.c_str(), len);
}

void BinarySerialize::raw(const std::string& name, const void *val, uint32_t len)
{
	debugField(name, "raw");
	out.write(reinterpret_cast<const char*>(&len), sizeof(len));
	out.write(reinterpret_cast<const char*>(val), len);
}

void BinarySerialize::ptr(const std::string& name, const Serializable *val)
{
	ASSERT(val == nullptr || val->shouldSerialize());
	debugField(name, "ptr");
	uint64_t serializedPtr = (uint64_t)val;
	out.write(reinterpret_cast<const char*>(&serializedPtr), sizeof(serializedPtr));
	if (val != nullptr)
	{	
		s(name + "_Type", val->serializationType());
		ASSERT(val->serializationType() != "Don't fucking serialize this");
#ifdef SL_DEBUG
		savedPtrs.insert(val);
#endif // SL_DEBUG
	}
}

void BinarySerialize::saveSerializable(const std::string& name, const Serializable *val)
{
	ASSERT(val->shouldSerialize());
	ASSERT(val->serializationType() != "Don't fucking serialize this");
	debugField(name, "Serializable");
	uint64_t serializedPtr = (uint64_t)val;
	out.write(reinterpret_cast<const char*>(&serializedPtr), sizeof(serializedPtr));
	s(name + "_Type", val->serializationType());
#ifdef SL_DEBUG
	auto inserted = registeredPtrs.insert(val);
	ASSERT(inserted.second);
#endif // SL_DEBUG
	val->serialize(*this);
}

void BinarySerialize::startArray(const std::string& name, int length)
{
	debugField(name, "<array>");
	debugStack.push(std::make_pair(StackType::Array, name));
	i(name, length);
}

void BinarySerialize::endArray(const std::string& name)
{
	// check to see if deserialization query is well-formed
	ASSERT(debugStack.top().first == StackType::Array);
	ASSERT(debugStack.top().second == name);
	debugStack.pop();
	debugField(name, "</array>");
}

void BinarySerialize::startObject(const std::string& name)
{
	debugField(name, "<object>");
	debugStack.push(std::make_pair(StackType::Object, name));
}

void BinarySerialize::endObject(const std::string& name)
{
	// check to see if deserialization query is well-formed
	ASSERT(debugStack.top().first == StackType::Object);
	ASSERT(debugStack.top().second == name);
	debugStack.pop();
	debugField(name, "</object>");
}

void BinarySerialize::finish()
{
#ifdef SL_DEBUG
	for (const Serializable *ptr : savedPtrs)
	{
		ASSERT(registeredPtrs.count(ptr));
	}
#endif // SL_DEBUG
	out.close();
}


// private
void BinarySerialize::debugField(const std::string& name, const std::string& type)
{
	if (debug)
	{
		const std::string tag = name + "|" + type;
#ifdef SL_DEBUG
		for (int i = 0; i < debugStack.size(); ++i)
		{
			debugFile << "\t";
		}
		debugFile << tag << std::endl;
#endif // SL_DEBUG
		ASSERT(tag.size() < 0xFF);
		uint16_t len = static_cast<uint16_t>(tag.size());
		out.write(reinterpret_cast<const char*>(&len), sizeof(len));
		out.write(tag.c_str(), len);
	}
}

} // sl