#ifndef SL_SERIALIZATION_HPP
#define SL_SERIALIZATION_HPP

#include <string>
#include <cstdint>
#include <map>
#include <memory>

namespace sl
{

class Serializable;
class World;

class Serialize
{
public:
	virtual ~Serialize() {}

	virtual void u8(const std::string& name, uint8_t val) = 0;

	virtual void u16(const std::string& name, uint16_t val) = 0;

	virtual void u32(const std::string& name, uint32_t val) = 0;

	virtual void u64(const std::string& name, uint64_t val) = 0;

	virtual void b(const std::string& name, bool val) = 0;

	virtual void i(const std::string& name, int val) = 0;

	virtual void f(const std::string& name, float val) = 0;

	virtual void s(const std::string& name, const std::string& val) = 0;

	virtual void raw(const std::string& name, const void *val, uint32_t len) = 0;

	virtual void ptr(const std::string& name, const Serializable *val) = 0;

	virtual void saveSerializable(const std::string& name, const Serializable *val) = 0;

	virtual void startArray(const std::string& name, int length) = 0;

	virtual void endArray(const std::string& name) = 0;

	virtual void startObject(const std::string& name) = 0;

	virtual void endObject(const std::string& name) = 0;

	virtual void finish() = 0;
};

class Deserialize
{
public:
	virtual ~Deserialize() {}

	virtual uint8_t u8(const std::string& name) = 0;

	virtual uint16_t u16(const std::string& name) = 0;

	virtual uint32_t u32(const std::string& name) = 0;

	virtual uint64_t u64(const std::string& name) = 0;

	virtual bool b(const std::string& name) = 0;

	virtual int i(const std::string& name) = 0;

	virtual float f(const std::string& name) = 0;

	virtual std::string s(const std::string& name) = 0;

	virtual uint32_t raw(const std::string& name, void *val, uint32_t len) = 0;

	virtual Serializable* ptr(const std::string& name) = 0;

	virtual std::unique_ptr<Serializable> loadSerializable(const std::string& name) = 0;

	virtual int startArray(const std::string& name) = 0;

	virtual void endArray(const std::string& name) = 0;

	virtual void startObject(const std::string& name) = 0;

	virtual void endObject(const std::string& name) = 0;

	virtual void finish() = 0;

	/**
	 * This is kind of ugly, but it's the only way certain things can be nicely deserialized.
	 * @return The world we are deserializing, if it exists yet.
	 */
	virtual World* getWorld() const = 0;

	virtual void setWorld(World *world) = 0;
};

template <typename T>
std::unique_ptr<T> loadSerializable(Deserialize& in, const std::string& name)
{
	return std::unique_ptr<T>(static_cast<T*>(in.loadSerializable(name).release()));
}

#define SAVEVAR(method, variable) out.method(#variable, variable)
#define LOADVAR(method, variable) variable = in.method(#variable)
// SAVEPTR is just for consistency
#define SAVEPTR(variable) SAVEVAR(ptr, variable)
#define LOADPTR(variable) variable = static_cast<decltype(variable)>(in.ptr(#variable))

#define SAVEREF(variable) out.ptr(#variable, variable.get())
#define LOADREF(variable) variable = static_cast<decltype(variable.get())>(in.ptr(#variable))

} // sl

#endif // SL_SERIALIZATION_HPP