#include "Engine/Serialization/TileSerialization.hpp"

#include "Engine/Serialization/Serialization.hpp"
#include "Utility/Assert.hpp"

#define READ_NEXT_BYTE 0x80
#define DATA 0x7F
#define SEGMENT_SHIFT 7
#define SEGMENT_SIZE 128

namespace sl
{

static void writeTile(Serialize& serialize, TileType tile, int count)
{
	//std::cout << "Writing (" << tile << ", " << count << ")\n";
	ASSERT(static_cast<TileType>(tile & DATA) == tile);
	ASSERT(count > 0);
	if (count == 1)
	{
		serialize.u8("tile", tile);
	}
	else
	{
		serialize.u8("tile", READ_NEXT_BYTE | tile);
		while (count >= SEGMENT_SIZE)
		{
			serialize.u8("len", READ_NEXT_BYTE | (count & DATA));
			count >>= SEGMENT_SHIFT;
		}
		serialize.u8("len", count);
	}
}

bool serializeTiles(Serialize& serialize, const Grid<TileType>& tiles)
{
	serialize.u16("width", tiles.getWidth());
	serialize.u16("height", tiles.getHeight());
	auto it = tiles.cbegin();
	TileType last = *it;
	int count = 1;
	++it;
	while (it != tiles.cend())
	{
		if (*it == last)
		{
			++count;
		}
		else
		{
			writeTile(serialize, last, count);
			last = *it;
			count = 1;
		}
		++it;
	}
	if (count > 0)
	{
		writeTile(serialize, last, count);
	}
	return true;
}

bool deserializeTiles(Deserialize& deserialize, Grid<TileType>& tiles)
{
	const uint16_t width = deserialize.u16("width");
	const uint16_t height = deserialize.u16("height");
	tiles.resize(width, height);
	auto it = tiles.begin();
	do
	{
		uint8_t u = deserialize.u8("tile");
		const TileType tile = static_cast<TileType>(u & DATA);
		if (u & READ_NEXT_BYTE)
		{
			int count = 0;
			int mult = 1;
			do
			{
				u = deserialize.u8("len");
				count += mult * (u & DATA);
				mult <<= SEGMENT_SHIFT;
			}
			while (u & READ_NEXT_BYTE);

			ASSERT(count > 0);

			//std::cout << "Read (" << tile << ", " << count << ")\n";

			while (--count >= 0)
			{
				*(it++) = tile;
			}
		}
		else
		{
			//std::cout << "Read (" << tile << ", " << 1 << ")\n";
			*it = tile;
			++it;
		}
	}
	while (it != tiles.end());

	return true;
}

} // sl
