/**
 * @section DESCRIPTION
 * A generic class to manage states. States must implement: update(), onEnter(), onExit(), handleEvent(const Event&)
 */
#ifndef SL_STATEMANAGER_HPP
#define SL_STATEMANAGER_HPP

#include <memory>

namespace sl
{

class Event;

class GenericStateManager
{
public:
	virtual ~GenericStateManager() {}

	virtual void update() = 0;

	virtual void handleEvent(const Event& event) = 0;
};

template <class T>
class StateManager : public GenericStateManager
{
public:
	typedef std::unique_ptr<T> StatePtr;
	/**
	 * Constructs a StateManager
	 * @param currentState The state that the StateManager will start in
	 * @param globalState The global state that the StateManager will start in
	 */
	StateManager(StatePtr currentState = nullptr, StatePtr globalState = nullptr);
	/**
	 * Destructor
	 */
	~StateManager();

	/**
	 * Initializes the current states.
	 * @param currentState The new current state
	 * @param globalState The new global state
	 */
	void initStates(StatePtr currentState, StatePtr globalState = nullptr);
	/**
	 * Updates the current state
	 */
	void update() override;
	/**
	 * Changes the current state
	 */
	void changeState(StatePtr newState);
	/**
	 * Changes the global state
	 * @param newState The new global state
	 */
	void setGlobalState(StatePtr newState);
	/**
	 * Calls code for when a AbstractScene is entered
	 */
	void onSceneEnter();
	/**
	 * Calls code for when a AbstractScene is exited
	 */
	void onSceneExit();
	/**
	 * Passes Events down to states
	 */
	void handleEvent(const Event& event) override;

protected:
	StateManager(const StateManager&);
	StateManager& operator=(const StateManager&);


	//!The global state
	StatePtr globalState;
	//!The current local state
	StatePtr currentState;

	StatePtr newState;

	StatePtr newGlobalState;
};

template <class T>
StateManager<T>::StateManager(StatePtr currentState, StatePtr globalState)
	:globalState(std::move(globalState))
	,currentState(std::move(currentState))
{
}

template <class T>
StateManager<T>::~StateManager()
{
	if (globalState)
	{
		globalState->onExit();
	}
	if (currentState)
	{
		currentState->onExit();
	}
}

template <class T>
void StateManager<T>::initStates(StatePtr currentState, StatePtr globalState)
{
	if (this->currentState)
	{
		this->currentState->onExit();
		this->currentState.reset();
	}
	if (this->globalState)
	{
		this->globalState->onExit();
		this->globalState.reset();
	}
	this->newState = std::move(currentState);
	this->newGlobalState = std::move(globalState);
}

template <class T>
void StateManager<T>::update()
{
	if (newState)
	{
		if (currentState)
		{
			currentState->onExit();
		}
		currentState = std::move(newState);
		currentState->onEnter();
	}
	if (newGlobalState)
	{	if (globalState)
		{
			globalState->onExit();
		}
		globalState = std::move(newGlobalState);
		globalState->onEnter();
	}
	if (globalState)
	{
		globalState->update();
	}
	if (currentState)
	{
		currentState->update();
	}
}

template <class T>
void StateManager<T>::changeState(StatePtr newState)
{
	this->newState = std::move(newState);
}

template <class T>
void StateManager<T>::setGlobalState(StatePtr newState)
{	
	globalState = std::move(newState);
}

template <class T>
void StateManager<T>::onSceneEnter()
{
	if (currentState)
	{
		currentState->onSceneEnter();
	}
	if (globalState)
	{
		globalState->onSceneEnter();
	}
}

template <class T>
void StateManager<T>::onSceneExit()
{
	if (currentState)
	{
		currentState->onSceneExit();
	}
	if (globalState)
	{
		globalState->onSceneExit();
	}
}

template <class T>
void StateManager<T>::handleEvent(const Event& event)
{
	if (currentState)
	{
		currentState->handleEvent(event);
	}
	if (globalState)
	{
		globalState->handleEvent(event);
	}
}

} // sl

#endif
