#include "Home/BuildingLevel.hpp"

#include "Engine/Serialization/AutoDeserialize.hpp"
#include "Engine/Serialization/AutoSerialize.hpp"
#include "Engine/Game.hpp"
#include "Town/World.hpp"
#include "Engine/Graphics/Sprite.hpp"
#include "Home/Dungeon.hpp"
#include "Home/Internet/Computer.hpp"
#include "Home/Furniture.hpp"
#include "NPC/States/IdleState.hpp"
#include "NPC/States/LoliRoutineState.hpp"
#include "NPC/States/AdultWanderState.hpp"
#include "NPC/Stats/SoulGenerationSettings.hpp"
#include "NPC/NPC.hpp"
#include "NPC/Bed.hpp"
#include "NPC/Locators/FixedTargetLocator.hpp"
#include "NPC/States/ShopkeeperState.hpp"
#include "Pathfinding/GridPathManager.hpp"
#include "Engine/Physics/HeightRect.hpp"
#include "Player/Player.hpp"
#include "Town/StaticGeometry.hpp"
#include "Town/Generation/Interiors/EditingUndo.hpp"
#include "Town/Town.hpp"
#include "Utility/Assert.hpp"

#include <cmath> // for ceil()

namespace sl
{

int worldCoordsAtLevel(int level)
{
	return 10 * level;
}

// @TODO REMOVE AND REPLACE WITH A NON-STUBBED THING THAT READS FROM THE BUILDING'S ACTUAL WALL TYPES
static const std::string& TEMP_THING(int x, int y)
{
	static std::string defaultWallType = "?";
	return defaultWallType;
}

BuildingLevel::BuildingLevel(Game *game, int width, int height, int worldXOffset, int worldYOffset, const Rect<int>& buildingArea, float buildingScale, bool createPathOnWorldAdd)
	:Scene(game, width, height, game->getWindowWidth(), game->getWindowHeight())
	,tileGrid(0, 0, width / getTileSize(), height / getTileSize(), getTileSize(), getTileSize())
	,wallGrid(0, 0, width / getTileSize(), height / getTileSize(), getTileSize(), getTileSize(), *this, TEMP_THING)
	,below(nullptr)
	,above(nullptr)
	,town(nullptr)
	,worldspaceRect()
	,dungeon(nullptr)
	,createPathOnWorldAdd(createPathOnWorldAdd)
{
	setWorldCoords(sf::Vector3i(worldXOffset, worldYOffset, 0));
	setWorldScale(buildingScale);
	const int left = worldXOffset + buildingArea.left * getWorldScale();
	const int top = worldYOffset + buildingArea.top * getWorldScale();
	const int right = std::ceil(worldXOffset + (buildingArea.left + buildingArea.width) * getWorldScale());
	const int bottom = std::ceil(worldYOffset + (buildingArea.top + buildingArea.height) * getWorldScale());
	worldspaceRect = Rect<int>(left, top, right - left, bottom - top);
}


void BuildingLevel::addObject(const Vector2i& pos, const std::string& objectType, const json11::Json::object *instanceAttributes, FamilyGen *family)
{
	const towngen::HomeObjectPrototype *prototype = towngen::HomeObjectDatabase::get().getPrototype(objectType);
	ASSERT(prototype);
	if (dungeon)
	{
		const Rect<int>& tileArea = prototype->getTileArea();
		const int postx = pos.x / getTileSize();
		const int posty = pos.y / getTileSize();
		for (int i = 0; i < tileArea.width; ++i)
		{
			for (int j = 0; j < tileArea.height; ++j)
			{
				dungeon->setHasObject(tileArea.left + i + postx, tileArea.top + j + posty, true);
			}
		}
	}
	prototype->generate(Random::get(), *this, pos, town, below, above, instanceAttributes, family);
}

void BuildingLevel::setTile(const Vector2i& pos, const std::string& tileType)
{
	const towngen::HomeTileDatabase::Prototype *prototype = towngen::HomeTileDatabase::get().getPrototype(tileType);
	ASSERT(prototype);
	if (prototype)
	{
		const std::string& fname = prototype->getFilename();
		// @REFACTOR
		auto it = tileTextures.find(fname);
		if (it == tileTextures.end())
		{
			it = tileTextures.insert(std::make_pair(fname, TextureManager::getTexture(fname))).first;
		}
		tileGrid.setTexture(pos.x, pos.y, it->second.get(), prototype->getTextureRect());
		wallGrid.remove(pos.x, pos.y);
		tilesForSerialization[pos] = tileType;
	}
}

void BuildingLevel::removeTile(const Vector2i& pos)
{
	// set it to any texture but a 0x0 subrect of it
	tileGrid.setTexture(pos.x, pos.y, tileTextures.begin()->second.get(), sf::IntRect(0, 0, 0, 0));
	tilesForSerialization.erase(pos);
}

void BuildingLevel::addWall(const Vector2i& pos, const std::string& wallType)
{
	//@todo handle walltype here later
	wallGrid.fill(pos.x, pos.y);
}

void BuildingLevel::removeWall(const Vector2i& pos)
{
	wallGrid.remove(pos.x, pos.y);
}

void BuildingLevel::setConnections(BuildingLevel *below, BuildingLevel *above, Town *town)
{
	this->below = below;
	this->above = above;
	this->town = town;
}

const Rect<int>& BuildingLevel::getWorldspaceRect() const
{
	return worldspaceRect;
}

Dungeon* BuildingLevel::getDungeon()
{
	return dungeon;
}

void BuildingLevel::createDungeon(int x, int y)
{
	ASSERT(dungeon == nullptr);
	dungeon = new Dungeon(Vector2i(x * getTileSize(), y * getTileSize()), wallGrid, *this);
	addEntityToList(dungeon);
}

const WallGrid& BuildingLevel::getWallGrid() const
{
	return wallGrid;
}

void BuildingLevel::createPrecomputedPathgrid()
{
	ASSERT(!createPathOnWorldAdd);
	createPrecomputedPathgridImpl();
}

void BuildingLevel::serialize(Serialize& out) const
{
	out.startObject("BuildingLevel");
	serializeScene(out);
	// tileGrid and wallGrid computed via tilesForSerialization
	// tileTextures ignored as it's just a cache
	SAVEPTR(below);
	SAVEPTR(above);
	SAVEPTR(town);
	AutoSerialize::serialize(out, worldspaceRect);
	AutoSerialize::serialize(out, tilesForSerialization);
	out.b("dungeon?", dungeon != nullptr);
	if (dungeon)
	{
		out.i("dungeon.x", dungeon->getPos().x / getTileSize());
		out.i("dungeon.y", dungeon->getPos().y / getTileSize());
		dungeon->serializeDungeon(out);
	}
	out.endObject("BuildingLevel");
}

void BuildingLevel::deserialize(Deserialize& in)
{
	in.startObject("BuildingLevel");
	deserializeScene(in);
	// tileGrid and wallGrid computed via tilesForSerialization
	tileGrid.set(0, 0, width / getTileSize(), height / getTileSize(), getTileSize(), getTileSize());
	wallGrid.reset(0, 0, width / getTileSize(), height / getTileSize(), getTileSize(), getTileSize());
	LOADPTR(below);
	LOADPTR(above);
	LOADPTR(town);
	AutoDeserialize::deserialize(in, worldspaceRect);
	AutoDeserialize::deserialize(in, tilesForSerialization);
	// does both tileGrid + wallGrid
	for (const std::pair<Vector2i, std::string>& tile : tilesForSerialization)
	{
		setTile(tile.first, tile.second);
	}
	// dungeon at end to make sure BuildingLevel fully deserialized first
	ASSERT(dungeon == nullptr);
	if (in.b("dungeon?"))
	{
		const int dungx = in.i("dungeon.x");
		const int dungy = in.i("dungeon.y");
		createDungeon(dungx, dungy);
		dungeon->deserializeDungeon(in);
	}
	in.endObject("BuildingLevel");

	createPrecomputedPathgrid();
}

std::string BuildingLevel::serializationType() const
{
	return "BuildingLevel";
}

bool BuildingLevel::shouldSerialize() const
{
	return true;
}


/*		private		*/

void BuildingLevel::onUpdate(Timestamp timestamp)
{
	this->updateEntities();

#ifdef SL_DEBUG
	if (getGame()->getSettings().get("static_geo_grid_draw") && getPlayer())
	{
		const pf::GridPathManager *pathMan = dynamic_cast<const pf::GridPathManager*>(getWorld()->getPathManager("StaticGeometry"));
		if (pathMan)
		{
			if (staticGeoPathGridAdjlist.empty())
			{
				staticGeoPathGridAdjlist = pathMan->exctractGraphsAsAdjlists(this);
			}
			const int ts = getTileSize();
			for (const std::map<Vector2i, std::set<Vector2i>>& adjlist : staticGeoPathGridAdjlist)
			{
				for (int y = getViewRect().top / ts; y < (getViewRect().top + getViewRect().height) / ts; ++y)
				{
					for (int x = getViewRect().left / ts; x < (getViewRect().left + getViewRect().width) / ts; ++x)
					{
						if (!wallGrid.isSolid(x, y))
						{
							auto it = adjlist.find(Vector2i(x, y));
							if (it != adjlist.end())
							{
								const Vector2i& u = it->first;
								for (const Vector2i& v : it->second)
								{
									debugDrawLine("static_geo_grid_draw", Vector2f((u.x + 0.5f) * ts, (u.y + 0.5f) * ts), Vector2f((v.x + 0.5f) * ts, (v.y + 0.5f) * ts), sf::Color::Magenta);
								}
								//debugDrawCircle("static_geo_grid_draw", Vector2f((u.x + 0.5f) * ts, (u.y + 0.5f) * ts), ts / 4.f, sf::Color::Cyan, sf::Color(32, 255, 96, 128));
							}
						}
					}
				}
			}
		}
	}
#endif // SL_DEBUG
}

void BuildingLevel::onDraw(RenderTarget& target)
{
	target.draw(tileGrid);
	this->drawEntities(target, Rect<int>(0, 0, width, height));
}

void BuildingLevel::onWorldAdd()
{
	if (createPathOnWorldAdd)
	{
		createPrecomputedPathgridImpl();
	}
}


void BuildingLevel::adjustWorldOffset(int xOffset, int yOffset, int level)
{
	const sf::Vector3i& coords = getWorldCoords();
	ASSERT(coords.z == 0);
	setWorldCoords(sf::Vector3i(coords.x - xOffset, coords.y - yOffset, worldCoordsAtLevel(level)));
}

void BuildingLevel::createPrecomputedPathgridImpl()
{
	createAndRegisterPrecomputedPathGrid("StaticGeometry", "StaticGeometry");
	//createAndRegisterRealTimePathGrid({"StaticGeometry"}, "StaticGeometry");
	//getWorld()->getPathManager("StaticGeometry");
}

} // sl
