#ifndef SL_HOME_HPP
#define SL_HOME_HPP

#include "Engine/Graphics/TileGrid.hpp"
#include "Home/WallGrid.hpp"
#include "Engine/Scene.hpp"
#include "Town/Generation/Interiors/HomeObjectDatabase.hpp"
#include "Town/Generation/Interiors/HomeTileDatabase.hpp"
#include "Utility/Rect.hpp"
#include "Utility/Vector.hpp"

namespace sl
{

class Dungeon;
class FamilyGen;
class Random;
class Town;

class BuildingLevel : public Scene
{
public:
	/**
	 * @param game The game the building is in
	 * @param width The width of this scene (not tiles)
	 * @param height The height of this scene (not tiles)
	 * @param worldXOffset The x position in worldspace tiles that the top-left part of the scene (walkable or not) is at
	 * @param worldYOffset The y position in worldspace tiles that the top-left part of the scene (walkable or not) is at
	 * @param buildingArea The area (in scene tiles) that the building occupies
	 * @param buildingScale The scale of which the the building is compared to the overworld. If it's 0.5 then for every 0.5 tiles in the overworld, you travel 1 tile in this level
	 * @param createPathOnWorldAdd Whether to call createPrecomputedPathgrid() in onWorldAdd() - this is for optimization purposes.
	 */
	BuildingLevel(Game *game, int width, int height, int worldXOffset, int worldYOffset, const Rect<int>& buildingArea, float buildingScale, bool createPathOnWorldAdd);


	void setTile(const Vector2i& pos, const std::string& tileType);

	void removeTile(const Vector2i& pos);

	void addObject(const Vector2i& pos, const std::string& objectType, const json11::Json::object *instanceAttributes = nullptr, FamilyGen *family = nullptr);

	void addWall(const Vector2i& pos, const std::string& wallType);

	void removeWall(const Vector2i& pos);

	void setConnections(BuildingLevel *below, BuildingLevel *above, Town *town);

	/**
	 * @return The region in worldspace (tiles!) that the BUILDING (not including the empty space) occupies
	 */
	const Rect<int>& getWorldspaceRect() const;

	Dungeon* getDungeon();

	void createDungeon(int x, int y);

	const WallGrid& getWallGrid() const;

	/**
	 * Avoid creating this until we've constructed the level, as otherwise in debug builds generation takes like a minute since every setTile() call
	 * involves a tonne of recomputation.
	 */
	void createPrecomputedPathgrid();


	void serialize(Serialize& out) const override;

	void deserialize(Deserialize& in) override;

	std::string serializationType() const override;

	bool shouldSerialize() const override;

private:

	/**
	 * @param timestamp The current time (in seconds since game start)
	 */
	void onUpdate(Timestamp timetsamp) override;

	void onDraw(RenderTarget& target) override;

	void onWorldAdd() override;

	void adjustWorldOffset(int xOffset, int yOffset, int level);

	void createPrecomputedPathgridImpl();



	//! The graphical representation of the tiles
	TileGrid tileGrid;
	//! A cache of all textures used for tiles to make sure they don't get freed
	std::map<std::string, Texture> tileTextures;
	WallGrid wallGrid;
	BuildingLevel *below;
	BuildingLevel *above;
	Town *town;
	//! The area (in worldspace tiles) this building occupies.
	Rect<int> worldspaceRect;
	Dungeon *dungeon;
#ifdef SL_DEBUG
	std::vector<std::map<Vector2i, std::set<Vector2i>>> staticGeoPathGridAdjlist;
#endif // SL_DEBUG
	//! Used only for serialization. TODO: if space is an issue, you could optimize this to just store the filename for the JSON file specifying the town type, unless this is the basement
	std::map<Vector2i, std::string> tilesForSerialization;
	//! 
	bool createPathOnWorldAdd;
};

} // sl

#endif // SL_HOME_HPP
