#include "Home/Dungeon.hpp"

#include <functional>

#include "Engine/Graphics/Sprite.hpp"
#include "Engine/GUI/DrawableContainer.hpp"
#include "Engine/GUI/NormalBackground.hpp"
#include "Engine/Serialization/AutoDeserialize.hpp"
#include "Engine/Serialization/AutoSerialize.hpp"
#include "Home/BuildingLevel.hpp"
#include "Player/Player.hpp"

namespace sl
{

const int TILES_NEEDED = 6;

Dungeon::Dungeon(const Vector2i& pos, WallGrid& wallGrid, BuildingLevel& building)
	:Entity(pos.x, pos.y, 32, 32, "Dungeon")
	,startingAreaBounds(26, 14, 10, 18)
	,rooms()
	,hallwayDoor(Vector2i(30, 22), RoomDoor::Orientation::South, nullptr)
	,roomGrid(wallGrid.getWidth(), wallGrid.getHeight(), nullptr)
	,wallGrid(wallGrid)
	,inUse(false)
	,building(building)
	,hasObjectGrid(building.getWidth() / building.getTileSize(), building.getHeight() / building.getTileSize(), false)
	,generateStartingArea(true)
{
	hallway = std::make_unique<Room>(hallwayDoor);

	addDrawable(std::unique_ptr<Drawable>(new Sprite("dungeonedit.png")));


	//gui->applyBackground(gui::NormalBackground(sf::Color::Black, sf::Color::White));
	//gui.enable();
}

void Dungeon::deserializeDungeon(Deserialize& in)
{
	generateStartingArea = false;
	AutoDeserialize::deserialize(in, startingAreaBounds);
	const int roomCount = in.startArray("rooms");
	for (int i = 0; i < roomCount; ++i)
	{
		const RoomDoor::Orientation orientation = static_cast<RoomDoor::Orientation>(in.u8("orientation"));
		const int rdx = in.i("rdx");
		const int rdy = in.i("rdy");
		//doors.push_front(RoomDoor(Vector2i(rdx, rdy), orientation, &wallGrid));
		const bool success = placeDoor(rdx, rdy, orientation);
		ASSERT(success);
		const auto rcar = addRoom(doors.front());
		ASSERT(rcar == RoomCreationAttemptResult::Success);
		rooms.back()->deserialize(in);
	}
	in.endArray("rooms");
	// hallway ignored since never changes
	// no reason to store roomGrid, it gets built along the way
	// doors are created via the rooms (TODO: if we allow multiple doors this won't be the case)
	// gui ignored
	// should not save while in use
	AutoDeserialize::deserialize(in, hasObjectGrid);

	ASSERT(!shouldSerialize());
}

void Dungeon::serializeDungeon(Serialize& out) const
{
	AutoSerialize::serialize(out, startingAreaBounds);
	out.startArray("rooms", rooms.size());
	for (const auto& room : rooms)
	{
		out.u8("orientation", static_cast<uint8_t>(room->getDoor().getOrientation()));
		out.i("rdx", room->getDoor().x());
		out.i("rdy", room->getDoor().y());
		room->serialize(out);
	}
	out.endArray("rooms");
	// hallway ignored since never changes
	// no reason to store roomGrid, just rebuild it at the end
	
	ASSERT(rooms.size() == doors.size());
	// gui ignored
	ASSERT(!inUse);
	AutoSerialize::serialize(out, hasObjectGrid);
}

bool Dungeon::isStatic() const
{
	return true;
}

const Entity::Type Dungeon::getType() const
{
	return Entity::makeType("Dungeon");
}

bool Dungeon::explore(std::vector<Vector2i>& tilesInRoom, int x, int y, Room *room)
{
	if (roomGrid.at(x, y) == hallway.get())
	{
		return false;
	}
	//ASSERT(roomGrid.at(x, y) == nullptr);
	tilesInRoom.push_back(Vector2i(x, y));
	roomGrid.at(x, y) = room;
	if (wallGrid.canGoRight(x, y))
	{
		if (roomGrid.at(x + 1, y) == hallway.get() ||
		   (roomGrid.at(x + 1, y) == nullptr && !explore(tilesInRoom, x + 1, y, room)))
		{
			return false;
		}
	}
	if (wallGrid.canGoLeft(x, y))
	{
		if (roomGrid.at(x - 1, y) == hallway.get() ||
		   (roomGrid.at(x - 1, y) == nullptr && !explore(tilesInRoom, x - 1, y, room)))
		{
			return false;
		}
	}
	if (wallGrid.canGoDown(x, y))
	{
		if (roomGrid.at(x, y + 1) == hallway.get() ||
		   (roomGrid.at(x, y + 1) == nullptr && !explore(tilesInRoom, x, y + 1, room)))
		{
			return false;
		}
	}
	if (wallGrid.canGoUp(x, y))
	{
		if (roomGrid.at(x, y - 1) == hallway.get() ||
		   (roomGrid.at(x, y - 1) == nullptr && !explore(tilesInRoom, x, y - 1, room)))
		{
			return false;
		}
	}
	return true;
}

Dungeon::RoomCreationAttemptResult Dungeon::addRoom(RoomDoor& door)
{
	wallGrid.closeAllDoors();
	//wallGrid.openDoorAt(door.x(), door.y());
	recalculateHallway();
	const int dx = door.x();
	const int dy = door.y();
	if (getRoom(dx, dy) == nullptr)
	{
		ASSERT(door.getOrientation() == RoomDoor::Orientation::South);// remove when other orientations are implemented
		if (/*(door.orientation == Room::Door::Orientation::North && (!wallGrid.canGoUp(dx, dy - 1) || getRoom(dx, dy - 1) != hallway.get())) ||*/
			(door.getOrientation() == RoomDoor::Orientation::South && (/*!wallGrid.canGoUp(dx, dy + 1) || */getRoom(dx, dy + 2) != hallway.get()))/* ||
		    (door.orientation == Room::Door::Orientation::West  && (!wallGrid.canGoUp(dx - 1, dy) || getRoom(dx - 1, dy) != hallway.get())) ||
		    (door.orientation == Room::Door::Orientation::East  && (!wallGrid.canGoUp(dx + 1, dy) || getRoom(dx + 1, dy) != hallway.get()))*/)
		{
			return RoomCreationAttemptResult::NotConnectedToHallway;
		}
		//// block off the door with a (temp?) wall and recalculate the hallway
		//switch (door.orientation)
		//{
		//case Room::Door::Orientation::North:
		//	wallGrid.addWall(dx, dy - 1, WallGrid::Barrier::Horizontal);
		//	break;
		//case Room::Door::Orientation::South:
		//	wallGrid.addWall(dx, dy, WallGrid::Barrier::Horizontal);
		//	break;
		//case Room::Door::Orientation::West:
		//	wallGrid.addWall(dx, dy, WallGrid::Barrier::Vertical);
		//	break;
		//case Room::Door::Orientation::East:
		//	wallGrid.addWall(dx - 1, dy, WallGrid::Barrier::Vertical);
		//	break;
		//}
		//wallGrid.closeDoorAt(door.x(), door.y());
		recalculateHallway();
		std::vector<Vector2i> tilesInRoom;
		std::unique_ptr<Room> room = std::make_unique<Room>(door);
		//std::function<bool(int, int)> explore = [&](int x, int y) -> bool
		//{
		//	if (roomGrid.at(x, y) == hallway.get())
		//	{
		//		return false;
		//	}
		//	if (wallGrid.canGoRight(x, y) && !explore(x + 1, y))
		//	{
		//		return false;
		//	}
		//	if (wallGrid.canGoLeft(x, y) && !explore(x - 1, y))
		//	{
		//		return false;
		//	}
		//	if (wallGrid.canGoDown(x, y) && !explore(x, y + 1))
		//	{
		//		return false;
		//	}
		//	if (wallGrid.canGoUp(x, y) && !explore(x, y - 1))
		//	{
		//		return false;
		//	}
		//	tilesInRoom.push_back(sf::Vector2i(x, y));
		//	roomGrid.at(x, y) = room.get();
		//	return true;
		//};

		RoomCreationAttemptResult result = RoomCreationAttemptResult::Success;
		if (!explore(tilesInRoom, dx, dy - 1, room.get()))
		{
			result = RoomCreationAttemptResult::MultipleExits;
		}
		else if (tilesInRoom.size() >= TILES_NEEDED)
		{

			result = RoomCreationAttemptResult::Success;
		}
		else
		{
			return RoomCreationAttemptResult::NotBigEnough;
		}
		if (result != RoomCreationAttemptResult::Success)
		{
			//	clean up all placed tiles
			for (const Vector2i& tile : tilesInRoom)
			{
				roomGrid.at(tile.x, tile.y) = nullptr;
			}
		}
		else
		{
			room->getPositions().swap(tilesInRoom);
			rooms.push_back(std::move(room));
		}
		return result;
	}
	return RoomCreationAttemptResult::InvalidLocation;
}

Room* Dungeon::getRoom(int x, int y)
{
	return roomGrid.at(x, y);
}

const Room* Dungeon::getRoom(int x, int y) const
{
	return roomGrid.at(x, y);
}

const std::vector<std::unique_ptr<Room>>& Dungeon::getRooms() const
{
	return rooms;
}

bool Dungeon::placeDoor(int x, int y, RoomDoor::Orientation orientation)
{
	ASSERT(x == 2 * (x / 2));
	ASSERT(y == 2 * (y / 2));
	if (getDoorAt(x, y) == nullptr)
	{
		RoomDoor door(Vector2i(x, y), orientation, &wallGrid);
		doors.push_front(std::move(door));
		wallGrid.addDoor(doors.front());
		return true;
	}
	return false;
}

RoomDoor* Dungeon::getDoorAt(int x, int y)
{
	ASSERT(x == 2 * (x / 2));
	ASSERT(y == 2 * (y / 2));
	for (RoomDoor& door : doors)
	{
		if (door.x() == x && door.y() == y)
		{
			return &door;
		}
	}
	return nullptr;
}

bool Dungeon::removeDoorAt(int x, int y)
{
	ASSERT(x == 2 * (x / 2));
	ASSERT(y == 2 * (y / 2));
	for (auto it = doors.begin(), end = doors.end(); it != end; ++it)
	{
		if (it->x() == x && it->y() == y)
		{
			wallGrid.removeDoor(*it);
			doors.erase(it);
			return true;
		}
	}
	return false;
}

WallGrid& Dungeon::getWallGrid()
{
	return wallGrid;
}

BuildingLevel& Dungeon::getHome()
{
	return building;
}

Room& Dungeon::getHallway()
{
	ASSERT(hallway);
	return *hallway;
}

const Room& Dungeon::getHallway() const
{
	ASSERT(hallway);
	return *hallway;
}

void Dungeon::close()
{
	inUse = false;
	// doesn't seem to work...
	scene->getPlayer()->setActive(true);

	scene->unpause("Player");
	scene->unpause("NPC");
}

bool Dungeon::isEditableTile(int x, int y) const
{
	return !startingAreaBounds.containsPoint(x, y);
}

void Dungeon::findInhabitants()
{
	for (auto& room : rooms)
	{
		room->findInhabitants(scene);
	}
}


bool Dungeon::hasObject(int tx, int ty) const
{
	return hasObjectGrid.at(tx, ty);
}

void Dungeon::setHasObject(int tx, int ty, bool val)
{
	hasObjectGrid.at(tx, ty) = val;
}


/*			private				*/
void Dungeon::onEnter()
{
	gui.setScene(scene);

	if (generateStartingArea)
	{
		initStartingArea();
	}
}

void Dungeon::onExit()
{
	gui.setScene(nullptr);
}

void Dungeon::onUpdate()
{
	if (scene->getInputHandle().isKeyPressed(Input::KeyboardKey::F5))
	{
		wallGrid.updateAllTiles();
	}
	if (inUse)
	{
		ASSERT(gui.isEnabled());
		gui->update();
	}
	else
	{
		// to avoid freeing it from within its own callback.
		if (gui.isEnabled())
		{
			gui.disable();
			gui.releaseHandle();
		}
		// TODO: replaec this with a contextual pop-up wtf why is this here wtf?
		Player *player = scene->checkCollision<Player>(*this);
		if (player)
		{
			player->addActionCallout(Player::CalloutType::Use, "Edit Dungeon", [this,player] {
				gui.setHandle(unique<DungeonConstructionGUI>(*this));
				gui.enable();
				gui->setViewPos(player->getPos());
				inUse = true;
				player->setActive(false);// doesnt seem to work
				scene->pause("NPC");
				scene->pause("Player");
			});
		}
	}
}



void Dungeon::initStartingArea()
{
	//wallGrid.setFloorType(FloorType::ID::Concrete);

	const int dx = hallway->getDoor().x();
	const int dy = hallway->getDoor().y() - 2;
	placeDoor(dx, dy, RoomDoor::Orientation::South);
	RoomDoor *door = getDoorAt(dx, dy);
	ASSERT(door);
	door->open();
	const RoomCreationAttemptResult result = addRoom(*door);
	ASSERT(result == RoomCreationAttemptResult::Success);
}

void Dungeon::fill(int x, int y)
{
	roomGrid.at(x, y) = hallway.get();
	hallway->getPositions().push_back(Vector2i(x, y));
	if (roomGrid.at(x + 1, y) == nullptr && wallGrid.canGoRight(x, y))
	{
		fill(x + 1, y);
	}
	if (roomGrid.at(x - 1, y) == nullptr && wallGrid.canGoLeft(x, y))
	{
		fill(x - 1, y);
	}
	if (roomGrid.at(x, y - 1) == nullptr && wallGrid.canGoUp(x, y))
	{
		fill(x, y - 1);
	}
	if (roomGrid.at(x, y + 1) == nullptr && wallGrid.canGoDown(x,  y))
	{
		fill(x, y + 1);
	}
}

void Dungeon::recalculateHallway()
{
	ASSERT(hallway);
	for (const Vector2i& tile : hallway->getPositions())
	{
		ASSERT(roomGrid.at(tile.x, tile.y) == hallway.get());
		roomGrid.at(tile.x, tile.y) = nullptr;
	}
	hallway->getPositions().clear();
	//std::function<void(int, int)> fill = [&](int x, int y)
	//{
	//	roomGrid.at(x, y) = hallway.get();
	//	if (roomGrid.at(x + 1, y) == nullptr && wallGrid.canGoRight(x, y))
	//	{
	//		fill(x + 1, y);
	//	}
	//	if (roomGrid.at(x - 1, y) == nullptr && wallGrid.canGoLeft(x, y))
	//	{
	//		fill(x - 1, y);
	//	}
	//	if (roomGrid.at(x, y - 1) == nullptr && wallGrid.canGoUp(x, y))
	//	{
	//		fill(x, y - 1);
	//	}
	//	if (roomGrid.at(x, y + 1) == nullptr && wallGrid.canGoDown(x, y))
	//	{
	//		fill(x, y + 1);
	//	}
	//};
	fill(hallway->getDoor().x(), hallway->getDoor().y());
}

} // sl
