#ifndef SL_DUNGEON_HPP
#define SL_DUNGEON_HPP

#include <memory>
#include <list>
#include <vector>

#include "Engine/GUI/WidgetHandle.hpp"
#include "Home/DungeonConstructionGUI.hpp"
#include "Home/FurnitureType.hpp"
#include "Home/DecorTypeDatabase.hpp"
#include "Home/Room.hpp"
#include "Home/WallGrid.hpp"
#include "Engine/Entity.hpp"
#include "Utility/Grid.hpp"

namespace sl
{

class BuildingLevel;
class Deserialize;
class Serialize;

class Dungeon : public Entity
{
public:
	enum class RoomCreationAttemptResult
	{
		Success,
		InvalidLocation,
		NotConnectedToHallway,
		MultipleExits,
		NotBigEnough
	};

	/**
	 * @param pos The position of the dungeon edit object (in pixels)
	 * @param wallGrid The WallGrid of the dungeon
	 * @param building The building floor that this dungeon exists in (same as wallGrid)
	 */
	Dungeon(const Vector2i& pos, WallGrid& wallGrid, BuildingLevel& building);

	// purposefully does not implement shouldSerialize() so we don't have to write serializers for this.
	// these are named differently to avoid them being called implicitly via Serializable -> Entity interface
	void deserializeDungeon(Deserialize& in);

	void serializeDungeon(Serialize& out) const;


	bool isStatic() const override;

	const Type getType() const override;



	/**
	 * Attempts to add a Room at the given location
	 * @param door The door to attempt to create the room starting at
	 * @return The result of the room being created (or failure)
	 */
	RoomCreationAttemptResult addRoom(RoomDoor& door);

	Room* getRoom(int x, int y);

	const Room* getRoom(int x, int y) const;

	const std::vector<std::unique_ptr<Room>>& getRooms() const;

	bool placeDoor(int x, int y, RoomDoor::Orientation orientation);

	RoomDoor* getDoorAt(int x, int y);

	bool removeDoorAt(int x, int y);

	WallGrid& getWallGrid();

	BuildingLevel& getHome();

	Room& getHallway();

	const Room& getHallway() const;

	void close();

	bool isEditableTile(int x, int y) const;

	void findInhabitants();

	void recalculateHallway();

	bool hasObject(int tx, int ty) const;

	void setHasObject(int tx, int ty, bool val);

private:

	void onEnter() override;

	void onExit() override;

	void onUpdate() override;

	void initStartingArea();

	void fill(int x, int y);

	bool explore(std::vector<Vector2i>& tilesInRoom, int x, int y, Room *room);



	Rect<int> startingAreaBounds;

	std::vector<std::unique_ptr<Room>> rooms;
	std::unique_ptr<Room> hallway;
	RoomDoor hallwayDoor;
	Grid<Room*> roomGrid;
	WallGrid& wallGrid;
	std::list<RoomDoor> doors;

	gui::WidgetHandle<DungeonConstructionGUI> gui;
	bool inUse;
	BuildingLevel& building;
	//! Which tiles have objects stored on them - objects cannot overlap
	Grid<bool> hasObjectGrid;
	//! initStartingArea is only for when you are starting a new game, so for loading we shouldn't generalize it (it's saved anyway)
	bool generateStartingArea;
};

} // sl

#endif
