#include "Home/DungeonConstructionGUI.hpp"

#include <sstream>
#include <iomanip>

#include "Engine/Game.hpp"
#include "Engine/Graphics/Animation.hpp"
#include "Engine/Graphics/SFDrawable.hpp"
#include "Engine/Graphics/Sprite.hpp"
#include "Engine/Graphics/TextureManager.hpp"
#include "Engine/GUI/Button.hpp"
#include "Engine/GUI/DrawableContainer.hpp"
#include "Engine/GUI/FloatingWindow.hpp"
#include "Engine/GUI/NormalBackground.hpp"
#include "Engine/GUI/Scrollbar.hpp"
#include "Engine/GUI/WidgetGrid.hpp"
#include "Home/Dungeon.hpp"
#include "Home/BuildingLevel.hpp"
#include "Home/ItemChooserGUI.hpp"
#include "NPC/LayeredSpriteLoader.hpp"
#include "NPC/NPC.hpp"
#include "NPC/NPCStatusGUI.hpp"
#include "NPC/SoulConjugator.hpp"
#include "Town/Generation/Interiors/EditingUndo.hpp"
#include "Utility/Assert.hpp"

#include "Utility/JsonUtil.hpp"
#include "Utility/Math.hpp"
#include "Utility/Random.hpp"
#include "Utility/ToString.hpp"

namespace sl
{

const int buttonWidth = 100;
const int buttonHeight = 32;
const int buttonGap = 8;
const int leftBorder = 24;
const int topBorder = 24;

const int editBarWidth = 192;
const int editBarHeight = 256;

DungeonConstructionGUI::DungeonConstructionGUI(Dungeon& dungeon)
	:gui::GUIWidget(Rect<int>(512, 384))
	,state(State::DesignatingRooms) // arbitrarily assigned to be NOT the starting one, call changeState later since it does additional logic
	,dungeon(dungeon)
	,gridOverlay()
	,popup(nullptr)
	,rotation(0)
	,tileSize(-1)
	,hackTimer(-1)
	,currentlySelectedItemId(-1)
	,leftRoomIndex(0)
	,rightRoomIndex(0)
	,viewBox(512, 384)
	,currentTile("concrete")
	,currentWall("test_wall")
	,cursorRectShape(nullptr)
{
	registerMouseInputEventWithMousePos(Key::GUIClick, GUIWidget::InputEventType::Held,
		std::bind(&DungeonConstructionGUI::onLeftHold, this, std::placeholders::_1, std::placeholders::_2)
	);
	registerMouseInputEventWithMousePos(Key::GUIRightClick, GUIWidget::InputEventType::Held,
		std::bind(&DungeonConstructionGUI::onRightHold, this, std::placeholders::_1, std::placeholders::_2)
	);
	registerMouseInputEventWithMousePos(Key::GUIClick, GUIWidget::InputEventType::Pressed,
		std::bind(&DungeonConstructionGUI::onLeftClick, this, std::placeholders::_1, std::placeholders::_2)
	);
	registerMouseInputEventWithMousePos(Key::GUIRightClick, GUIWidget::InputEventType::Pressed,
		std::bind(&DungeonConstructionGUI::onRightClick, this, std::placeholders::_1, std::placeholders::_2)
	);

	int index = 0;
	registerStateChangeGUI("Edit Layout", State::EditingLayout, index++);
	registerStateChangeGUI("Edit Style", State::EditingFloorStyle, index++);
	registerStateChangeGUI("Place Objects", State::PlacingStuff, index++);
	registerStateChangeGUI("Assign Rooms", State::DesignatingRooms, index++);

	const Rect<int> buttonBox(leftBorder + (buttonWidth + buttonGap) * index, topBorder, 32, buttonHeight);
	auto button = unique<gui::Button>(unique<gui::TextLabel>(buttonBox, "X", gui::TextLabel::Config().centered().pad(8, 6)),
	                                  sf::Color(192, 192, 192), sf::Color::White,
	                                  gui::NormalBackground(sf::Color::Black, sf::Color(192, 192, 192)),
	                                  gui::NormalBackground(sf::Color(32, 32, 32), sf::Color::White)
	);
	button->registerMouseInputEvent(Key::GUIClick, GUIWidget::InputEventType::Pressed, [&]() {
		dungeon.close();
	});
	addWidget(std::move(button));

	auto popupWidget = unique<gui::TextLabel>(Rect<int>(leftBorder, 384 - topBorder - buttonHeight, 512 - 2 * leftBorder, buttonHeight),
		"Some helpful messages should go here!", gui::TextLabel::Config().centered().pad(8, 6));
	popupWidget->applyBackground(gui::NormalBackground(sf::Color::Black, sf::Color::White));
	popup = popupWidget.get();
	addWidget(std::move(popupWidget));
	popup->update(leftBorder, 666);

	objectPreview.setHandle(std::make_unique<gui::DrawableContainer>(Rect<int>(32, 32)));

	cursor.setHandle(std::make_unique<gui::DrawableContainer>(Rect<int>(16, 16)));

	auto cursorRect = std::make_unique<SFDrawable<sf::RectangleShape>>(sf::RectangleShape(sf::Vector2f(16.f, 16.f)));
	cursorRectShape = &cursorRect->getData();
	cursorRectShape->setFillColor(sf::Color::Transparent);
	cursorRectShape->setOutlineColor(sf::Color::White);
	cursorRectShape->setOutlineThickness(1.f);
	
	cursor->setDrawable(std::move(cursorRect));
	
	changeState(State::EditingLayout);
}


void DungeonConstructionGUI::setViewPos(const Vector2f& pos)
{
	viewBox.left = snapToGrid(pos.x - viewBox.width / 2);
	viewBox.top = snapToGrid(pos.y - viewBox.height / 2);
}


// private
void DungeonConstructionGUI::initChangeStateBar()
{

}


void DungeonConstructionGUI::onLeftClick(int x, int y)
{
	const Vector2i pos = screenToLevel(x, y);
	const Vector2i tposHalf(2 * (pos.x / (2 * tileSize)), 2 * (pos.y / (2 * tileSize)));
	switch (state)
	{
	case State::EditingLayout:
	case State::EditingWallStyle:
	case State::EditingFloorStyle:
		break;
	case State::PlacingStuff:
		if (!currentObject.empty())
		{
			addObject(snapToGrid(pos.x), snapToGrid(pos.y), currentObject);
		}
		break;
	case State::DesignatingRooms:
	{
		const char *str = nullptr;
		if (!dungeon.placeDoor(tposHalf.x, tposHalf.y, RoomDoor::Orientation::South))
		{
			str = "You must click on somewhere there isn't alrady a door.";
		}
		else
		{
			RoomDoor *door = dungeon.getDoorAt(tposHalf.x, tposHalf.y);
			ASSERT(door);
			const Dungeon::RoomCreationAttemptResult result = dungeon.addRoom(*door);
			if (result == Dungeon::RoomCreationAttemptResult::Success)
			{
				updateGridOverlay();
				str = "Room created successfully.";
			}
			else
			{
				if (result == Dungeon::RoomCreationAttemptResult::InvalidLocation)
				{
					str = "Invalid location to place room. Rooms can't share doors or be inside other existing rooms.";
				}
				else if (result == Dungeon::RoomCreationAttemptResult::NotConnectedToHallway)
				{
					str = "Invalid location to place room. Must be connected to hallway.";
				}
				else if (result == Dungeon::RoomCreationAttemptResult::MultipleExits)
				{
					str = "Multiple exits to room found. Rooms must only contain a single exit.";
				}
				else if (result == Dungeon::RoomCreationAttemptResult::NotBigEnough)
				{
					str = "Room is not big enough. Make it bigger.";
				}
				dungeon.removeDoorAt(tposHalf.x, tposHalf.y);
			}
		}
		ASSERT(str != nullptr);
		showMessage(str);
	}
	break;
	case State::TransferingInhabitants:
		break;
	}
}

void DungeonConstructionGUI::onRightClick(int x, int y)
{
	switch (state)
	{
	case State::EditingLayout:
	case State::EditingWallStyle:
	case State::EditingFloorStyle:
	case State::PlacingStuff:
	case State::DesignatingRooms:
	case State::TransferingInhabitants:
		break;
	}
}

void DungeonConstructionGUI::onLeftHold(int x, int y)
{
	const Vector2i pos = screenToLevel(x, y);
	const int tx = pos.x / dungeon.getHome().getTileSize();
	const int ty = pos.y / dungeon.getHome().getTileSize();
	const int floortx = 2 * (tx / 2);
	const int floorty = 2 * (ty / 2);
	switch (state)
	{
	case State::EditingLayout:
		if (!dungeon.getHome().getWallGrid().isSolid(floortx, floorty))
		{
			for (int fx = 0; fx < 2; ++fx)
			{
				for (int fy = 0; fy < 2; ++fy)
				{
					const int editX = floortx + fx;
					const int editY = floorty + fy;
					dungeon.getHome().addWall(Vector2i(editX, editY), currentWall);
					//ASSERT(!dungeon.getHome().getWallGrid().isSolid(editX, editY));
					addWall(editX, editY, currentWall);
				}
			}
		}
		break;
	case State::EditingWallStyle:
	case State::EditingFloorStyle:
		if (!currentTile.empty())
		{
			for (int fx = 0; fx < 2; ++fx)
			{
				for (int fy = 0; fy < 2; ++fy)
				{
					//ASSERT(!dungeon.getHome().getWallGrid().isSolid(floortx + fx, floorty + fy));
					addTile(floortx + fx, floorty + fy, currentTile);
				}
			}
		}
		break;
	case State::PlacingStuff:
	case State::DesignatingRooms:
	case State::TransferingInhabitants:
		break;
	}
}

void DungeonConstructionGUI::onRightHold(int x, int y)
{
	const Vector2i pos = screenToLevel(x, y);
	switch (state)
	{
	case State::EditingLayout:
		{
			const int ts = dungeon.getHome().getTileSize();
			const int tx = pos.x / ts;
			const int ty = pos.y / ts;
			const int floortx = 2 * (tx / 2);
			const int floorty = 2 * (ty / 2);
			const WallGrid& wallGrid = dungeon.getHome().getWallGrid();
			if (wallGrid.isSolid(tx, ty))
			{
				for (int fx = 0; fx < 2; ++fx)
				{
					for (int fy = 0; fy < 2; ++fy)
					{
						const int editX = floortx + fx;
						const int editY = floorty + fy;
						ASSERT(wallGrid.isSolid(editX, editY));
						dungeon.getHome().removeWall(Vector2i(editX, editY));
						if (editX > 0 && wallGrid.isSolid(editX - 1, editY))
						{
							addWall(editX - 1, editY, currentWall);
						}
						if (editX < dungeon.getHome().getWidth() / ts - 1 && wallGrid.isSolid(editX + 1, editY))
						{
							addWall(editX + 1, editY, currentWall);
						}
						if (editY > 0 && wallGrid.isSolid(editX, editY - 1))
						{
							addWall(editX, editY - 1, currentWall);
						}
						if (editX < dungeon.getHome().getHeight() / ts - 1 && wallGrid.isSolid(editX, editY + 1))
						{
							addWall(editX, editY + 1, currentWall);
						}
						addTile(editX, editY, currentTile);
					}
				}
			}
		}
		break;
	case State::EditingWallStyle:
	case State::EditingFloorStyle:
		//{
		//	const Vector2i tpos(pos.x / getTileSize(), pos.y / getTileSize());
		//	auto it = tiles.find(tpos);
		//	if (it != tiles.end())
		//	{
		//		tileGrid.setTexture(tpos.x, tpos.y, tileTextures.begin()->second.get(), sf::IntRect(0, 0, 0, 0));
		//		tiles.erase(it);
		//	}
		//}
		break;
	case State::PlacingStuff:
		{
			std::list<Entity*> toDelete;
			const int deleteRadius = 4;
			dungeon.getHome().cull(toDelete, Rect<int>(pos.x - deleteRadius, pos.y - deleteRadius, 2 * deleteRadius, 2 * deleteRadius), "EditingUndo");
			for (Entity *entity : toDelete)
			{
				static_cast<towngen::EditingUndo&>(*entity).undo();
			}
		}
		break;
	case State::DesignatingRooms:
	case State::TransferingInhabitants:
		break;
	}
}

void DungeonConstructionGUI::onUpdate(int x, int y)
{
	const InputHandle& input = scene->getInputHandle();
	if (input.isKeyPressed(sf::Keyboard::R))
	{
		++rotation;
		updatePlacementPreview();
	}
	if (hackTimer > 0)
	{
		if (--hackTimer == 0)
		{
			popup->update(leftBorder, 666);
		}
	}
	const Vector2i cursorPos(snapToGrid(scene->getMousePos().x - viewBox.left), snapToGrid(scene->getMousePos().y - viewBox.top));

	if (objectPreview.isEnabled())
	{
		objectPreview->update(cursorPos.x, cursorPos.y);
	}
	cursor->update(cursorPos.x, cursorPos.y);
	//editArea->update(0, 0);
}

void DungeonConstructionGUI::onMouseHover(int x, int y)
{
	// move view

	const InputHandle& input = scene->getInputHandle();
	//// view moving
	//const float viewSpeed = 6.f;
	//const int scrollBorder = 16;
	//if (input.isKeyHeld(Input::KeyboardKey::Left) || x < scrollBorder)
	//{
	//	viewPos.x -= viewSpeed;
	//}
	//if (input.isKeyHeld(Input::KeyboardKey::Right) || x >= scene->getViewWidth() - scrollBorder)
	//{
	//	viewPos.x += viewSpeed;
	//}
	//if (input.isKeyHeld(Input::KeyboardKey::Up) || y < scrollBorder)
	//{
	//	viewPos.y -= viewSpeed;
	//}
	//if (input.isKeyHeld(Input::KeyboardKey::Down) || y >= scene->getViewHeight() - scrollBorder)
	//{
	//	viewPos.y += viewSpeed;
	//}
	const int ts = dungeon.getHome().getTileSize();
	if (input.isKeyPressed(Input::KeyboardKey::Left) && viewBox.left >= ts)
	{
		viewBox.left -= ts;
	}
	if (input.isKeyPressed(Input::KeyboardKey::Right) && viewBox.left < dungeon.getHome().getWidth() - viewBox.width - ts)
	{
		viewBox.left += ts;
	}
	if (input.isKeyPressed(Input::KeyboardKey::Up) && viewBox.top >= ts)
	{
		viewBox.top -= ts;
	}
	if (input.isKeyPressed(Input::KeyboardKey::Down) && viewBox.top < dungeon.getHome().getHeight() - viewBox.height - ts)
	{
		viewBox.top += ts;
	}
	dungeon.getHome().setView(viewBox.left + viewBox.width / 2, viewBox.top + viewBox.height / 2);

	for (int i = 0; i < viewBox.width / tileSize; ++i)
	{
		for (int j = 0; j < viewBox.height / tileSize; ++j)
		{

			gridOverlay.at(viewBox.left / tileSize + i, viewBox.top / tileSize + j).setPosition(i * tileSize, j * tileSize);
		}
	}

	if (state == State::PlacingStuff)
	{
		// check if you can place things at the cursor, and if not, colour it red
		const bool canPlace = canPlaceObject(x, y);
		cursorRectShape->setFillColor(canPlace ? sf::Color::Transparent : sf::Color(255, 0, 0, 128));
		cursorRectShape->setOutlineColor(canPlace ? sf::Color::White : sf::Color::Red);
	}
}

void DungeonConstructionGUI::beforeDraw(sf::RenderTarget& target) const
{
	if (state == State::DesignatingRooms)
	{
		for (int i = 0; i < viewBox.width / tileSize; ++i)
		{
			for (int j = 0; j < viewBox.height / tileSize; ++j)
			{
				target.draw(gridOverlay.at(viewBox.left / tileSize + i, viewBox.top / tileSize + j));
			}
		}
	}
}

void DungeonConstructionGUI::onSceneEnter()
{
	tileSize = dungeon.getScene()->getTileSize();
	sf::RectangleShape rect(sf::Vector2f(tileSize, tileSize));
	rect.setFillColor(sf::Color::Transparent);
	gridOverlay.resize(dungeon.getScene()->getWidth() / tileSize, dungeon.getScene()->getHeight() / tileSize, rect);
	for (auto it = gridOverlay.begin(), end = gridOverlay.end(); it != end; ++it)
	{
		it->setPosition(it.getX() * tileSize, it.getY() * tileSize);
	}

	cursor.setScene(scene);
	cursor.enable();
	objectPreview.setScene(scene);
	objectPreview.enable();
}




void DungeonConstructionGUI::registerStateChangeGUI(const char *str, State newState, int col)
{
	const Rect<int> buttonBox(leftBorder + (buttonWidth + buttonGap) * col, topBorder, buttonWidth, buttonHeight);
	auto button = gui::Button::createGenericButton(buttonBox, str, std::bind(&DungeonConstructionGUI::changeState, this, newState));
	addWidget(std::move(button));
}

void DungeonConstructionGUI::registerSubtateChangeGUI(const char *str, State newSubstate, int col, int row)
{
	const Rect<int> buttonBox(leftBorder + (buttonWidth + buttonGap) * col, topBorder + (buttonGap + buttonHeight) * row, buttonWidth, buttonHeight);
	auto button = gui::Button::createGenericButton(buttonBox, str, std::bind(&DungeonConstructionGUI::changeState, this, newSubstate));
	registerSubstateGUI(std::move(button));
}

void DungeonConstructionGUI::changeState(State newState)
{
	for (gui::GUIWidget *widget : substateWidgets)
	{
		// @todo this assert gets hit if a widget was deleted on its own (ie closed via X button). remove it entirely?
		// does anyone use substateWidgets? I hope not!
		//ASSERT(removeWidget(*widget));
		removeWidget(*widget);
	}
	substateWidgets.clear();
	state = newState;
	switch (state)
	{
	case State::TransferingInhabitants:
		{
			const int col = 3;
			int row = 0;
			registerSubtateChangeGUI("Add Room", State::DesignatingRooms, col, ++row);
			registerSubtateChangeGUI("Transfer Inhabitants", State::TransferingInhabitants, col, ++row);

			const int roomCount = dungeon.getRooms().size();
			ASSERT(roomCount > 0);
			leftRoomIndex %= roomCount;
			rightRoomIndex %= roomCount;

			dungeon.findInhabitants();

			createTransferGUI(TransferSide::Left);
			createTransferGUI(TransferSide::Right);
		}
		break;
	case State::DesignatingRooms:
		{
			dungeon.recalculateHallway();
			updateGridOverlay();
			const int col = 3;
			int row = 0;
			registerSubtateChangeGUI("Add Room", State::DesignatingRooms, col, ++row);
			registerSubtateChangeGUI("Transfer Inhabitants", State::TransferingInhabitants, col, ++row);
		}
		break;
	case State::PlacingStuff:
		{
			/*auto itemChooser = unique<ItemChooserGUI<FurnitureType>>(leftBorder, topBorder + buttonHeight + buttonGap + 218 + 64, 128, 0, 2, dungeon.getFurnitureTypeDB());
			itemChooser->applyBackground(gui::NormalBackground(sf::Color::Black, sf::Color::White));
			itemChooser->onItemChangeCallback([&](FurnitureType::ID id)
			{
				currentlySelectedItemId = static_cast<int>(id);
				updatePlacementPreview();
			});
			registerSubstateGUI(std::move(itemChooser));*/

			const int editPreviewHeight = 128;
			// x/y offset for edit bar
			const int ex = leftBorder;
			const int ey = scene->getGame()->getWindowHeight() - editBarHeight - topBorder;
			auto objectChooser = unique<gui::GUIWidget>(Rect<int>(ex, ey, editBarWidth, editBarHeight));
			const int colWidth = editBarWidth / 2;
#define OBJECT_GUI_FIELD(field, row, col, w) auto field##Widget = unique<gui::TextLabel>(Rect<int>(ex + col * colWidth, ey + 16 * row, w * colWidth, 64), gui::TextLabel::Config().pad(8, 6)); gui::TextLabel *field##WidgetPtr = field##Widget.get()
			OBJECT_GUI_FIELD(desc, 0, 0, 2);
			OBJECT_GUI_FIELD(cost, 4, 0, 1);
			OBJECT_GUI_FIELD(have, 5, 0, 1);
			OBJECT_GUI_FIELD(girly, 4, 1, 1);
			OBJECT_GUI_FIELD(comfy, 5, 1, 1);
			OBJECT_GUI_FIELD(fancy, 6, 1, 1);
#undef OBJECT_GUI_FIELD
			auto objectChooserListPtr = unique<gui::WidgetList<DungeonObjectDatabase::Data::value_type>>(
				Rect<int>(ex, ey + editPreviewHeight, editBarWidth, editBarHeight - editPreviewHeight),
				[](const DungeonObjectDatabase::Data::value_type& obj) -> std::string
			{
				return obj.first;
			},
				[&, descWidgetPtr, costWidgetPtr, haveWidgetPtr, girlyWidgetPtr, comfyWidgetPtr, fancyWidgetPtr](const DungeonObjectDatabase::Data::value_type& obj)
			{
				currentObject = obj.second.representations.front(); // @TODO allow rotations
				const towngen::HomeObjectPrototype *prototype = towngen::HomeObjectDatabase::get().getPrototype(currentObject);
				ASSERT(prototype);
				auto preview = prototype->createPreview(0.3f);
				objectPreview->setDrawable(std::move(preview));
				const Rect<int> undoRect = prototype->undoRect();
				cursorRectShape->setSize(sf::Vector2f(undoRect.width, undoRect.height));
				cursorRectShape->move(undoRect.left, undoRect.height);
				descWidgetPtr->setString(obj.second.desc);
				costWidgetPtr->setString("Cost: $" + std::to_string(obj.second.cost));
				haveWidgetPtr->setString("Have: 0");
				const float girly = obj.second.style[DungeonObjectDatabase::Entry::GIRLY];
				std::stringstream ss;
				ss << (girly >= 0.f ? "Girly: " : "Boyish: ") << std::setprecision(2) << std::fabs(girly);
				girlyWidgetPtr->setString(ss.str());
				const float comfy = obj.second.style[DungeonObjectDatabase::Entry::COMFY];
				ss.str("");
				ss << (comfy >= 0.f ? "Cozy: " : "Impersonal: ") << std::setprecision(2) << std::fabs(comfy);
				comfyWidgetPtr->setString(ss.str());
				const float fancy = obj.second.style[DungeonObjectDatabase::Entry::FANCY];
				ss.str("");
				ss << (fancy >= 0.f ? "Fancy: " : "Ghetto: ") << std::setprecision(2) << std::fabs(fancy);
				fancyWidgetPtr->setString(ss.str());
			}
			);
			objectChooserListPtr->setPreview([](const DungeonObjectDatabase::Data::value_type& obj) -> std::unique_ptr<gui::GUIWidget>
			{
				auto preview = towngen::HomeObjectDatabase::get().getPrototype(obj.second.representations.front())->createPreview();
				return unique<gui::DrawableContainer>(Rect<int>(32, 32), std::move(preview));
			});
			objectChooserListPtr->updateItems(DungeonObjectDatabase::get());
			objectChooser->addWidget(std::move(objectChooserListPtr));
			objectChooser->addWidget(std::move(descWidget));
			objectChooser->addWidget(std::move(costWidget));
			objectChooser->addWidget(std::move(haveWidget));
			objectChooser->addWidget(std::move(girlyWidget));
			objectChooser->addWidget(std::move(comfyWidget));
			objectChooser->addWidget(std::move(fancyWidget));
			objectChooser->applyBackground(gui::NormalBackground(sf::Color::Black, sf::Color::White));
			currentObject = towngen::HomeObjectDatabase::get().begin()->first;
			registerSubstateGUI(std::move(objectChooser));

			objectPreview.enable();
		}
		break;
	case State::EditingWallStyle:
		{
			ERROR("not implemented");
			const int col = 1;
			int row = 0;
			registerSubtateChangeGUI("Wall Style", State::EditingWallStyle, col, ++row);
			registerSubtateChangeGUI("Floor Style", State::EditingFloorStyle, col, ++row);
			/*auto itemChooser = unique<ItemChooserGUI<WallType>>(leftBorder, topBorder + buttonHeight + buttonGap + 218 + 64, 128, 0, 2, dungeon.getWallTypeDB());
			itemChooser->applyBackground(gui::NormalBackground(sf::Color::Black, sf::Color::White));
			itemChooser->onItemChangeCallback([&](WallType::ID id)
			{
				currentlySelectedItemId = static_cast<int>(id);
			});
			registerSubstateGUI(std::move(itemChooser));*/
			objectPreview.disable();
		}
		break;
	case State::EditingFloorStyle:
		{
			// TODO imeplement wall style changing again
			//const int col = 1;
			//int row = 0;
			//registerSubtateChangeGUI("Wall Style", State::EditingWallStyle, col, ++row);
			//registerSubtateChangeGUI("Floor Style", State::EditingFloorStyle, col, ++row);
			/*auto itemChooser = unique<ItemChooserGUI<FloorType>>(leftBorder, topBorder + buttonHeight + buttonGap + 218 + 64, 128, 0, 2, dungeon.getFloorTypeDB());
			itemChooser->applyBackground(gui::NormalBackground(sf::Color::Black, sf::Color::White));
			itemChooser->onItemChangeCallback([&](FloorType::ID id)
			{
				currentlySelectedItemId = static_cast<int>(id);
			});
			registerSubstateGUI(std::move(itemChooser));*/
			
			// 48 is to avoid overlapping the floor style thing
			auto tileChooser = unique<gui::WidgetList<towngen::HomeTileDatabase::Data::value_type>>(
				Rect<int>(leftBorder, scene->getGame()->getWindowHeight() - editBarHeight - topBorder, editBarWidth, editBarHeight),
				[](const towngen::HomeTileDatabase::Data::value_type& obj) -> std::string
			{
				return obj.first;
			},
				[&](const towngen::HomeTileDatabase::Data::value_type& obj)
			{
				currentTile = obj.first;
			}
			);
			tileChooser->applyBackground(gui::NormalBackground(sf::Color::Black, sf::Color::White));
			tileChooser->updateItems(towngen::HomeTileDatabase::get());
			registerSubstateGUI(std::move(tileChooser));

			objectPreview.disable();
		}
		break;
	case State::EditingLayout:
		{
			const int col = 0;
			int row = 0;
			registerSubtateChangeGUI("Fill/Excavate", State::EditingLayout, col, ++row);
		}
		break;
	default:
		break;
	}
	updatePlacementPreview();
}

void DungeonConstructionGUI::updateGridOverlay()
{
	const int colourCount = 5;//6;
	const sf::Color roomOverlayColours[colourCount] = {
	//	sf::Color(32, 128, 196, 96),
		sf::Color(32, 196, 128, 96),
		sf::Color(128, 32, 196, 96),
		sf::Color(128, 196, 32, 96),
		sf::Color(196, 32, 128, 96),
		sf::Color(196, 128, 32, 96)
	};
	std::map<int, sf::Color> idToCol;
	idToCol[dungeon.getHallway().getID()] = sf::Color(222, 255, 164, 32);
	for (unsigned int i = 0; i < dungeon.getRooms().size(); ++i)
	{
		idToCol[dungeon.getRooms()[i]->getID()] = roomOverlayColours[i % colourCount];
	}
	for (auto it = gridOverlay.begin(), end = gridOverlay.end(); it != end; ++it)
	{
		sf::RectangleShape& rect = *it;
		const Room *room = dungeon.getRoom(it.getX(), it.getY());
		if (room/* && room != &dungeon.getHallway()*/)
		{
			rect.setOutlineColor(sf::Color(255, 255, 255, 128));
			rect.setOutlineThickness(1.f);
			auto it = idToCol.find(room->getID());
			if (it != idToCol.end())
			{
				rect.setFillColor(it->second);
			}
			else // SOMETHING VERY BAD HAPPENED
			{
				rect.setFillColor(sf::Color(255, 32, 32, 128));
			}
		}
		else
		{
			rect.setOutlineColor(sf::Color::Transparent);
			rect.setOutlineThickness(0.f);
			rect.setFillColor(sf::Color::Transparent);
		}
	}
}

void DungeonConstructionGUI::registerSubstateGUI(std::unique_ptr<gui::GUIWidget> widget)
{
	substateWidgets.push_back(widget.get());
	addWidget(std::move(widget));
}

void DungeonConstructionGUI::updatePlacementPreview()
{
	objectPreview->setDrawable(nullptr);
	switch (state)
	{
	case State::EditingWallStyle:
	case State::EditingFloorStyle:
	case State::EditingLayout:
		{
			sf::RectangleShape rect(sf::Vector2f(tileSize, tileSize));
			rect.setFillColor(sf::Color(128, 128, 128, 128));
			rect.setOutlineColor(sf::Color::White);
			rect.setOutlineThickness(1.f);
			objectPreview->setDrawable(uniqueAs<SFDrawable<sf::RectangleShape>, Drawable>(rect));
		}
		break;
	case State::PlacingStuff:
		{
			//const FurnitureType::ID furnitureID = static_cast<FurnitureType::ID>(currentlySelectedItemId);
			//const FurnitureType::Definition& def = dungeon.getFurnitureTypeDB().get(furnitureID);
			//auto sprite = unique<Sprite>(def.previewFilename.c_str(), def.offset.x, def.offset.y);
			//sprite->setAlpha(0.5f);
			//placementPreview->setDrawable(std::move(sprite));
		}
		break;
	case State::DesignatingRooms:
		break;
	case State::TransferingInhabitants:
		break;
	}
}

void DungeonConstructionGUI::showMessage(const char *str)
{
	popup->setString(str);
	popup->update(leftBorder, 384 - topBorder - buttonHeight);
	hackTimer = 120;
}

void DungeonConstructionGUI::transferNPC(TransferSide side, int npcIndex)
{
	const int sourceIndex = side == TransferSide::Left ? leftRoomIndex : rightRoomIndex;
	const int destIndex = side == TransferSide::Left ? rightRoomIndex : leftRoomIndex;

	const Room& sourceRoom = *dungeon.getRooms()[sourceIndex];
	const Room& destRoom = *dungeon.getRooms()[destIndex];

	NPC *npc = sourceRoom.getInhabitants()[npcIndex];
	const Vector2i& newPos = Random::get().choose(destRoom.getPositions());
	const int offsetX = Random::get().randomFloat(0.25f, 0.75f) * tileSize;
	const int offsetY = Random::get().randomFloat(0.25f, 0.75f) * tileSize;
	npc->setPos(newPos.x * tileSize + offsetX, newPos.y * tileSize + offsetY);

	changeState(State::TransferingInhabitants);

	//dungeon.getHome().unpause("NPC");
	npc->update();
	//dungeon.getHome().pause("NPC");
}

void DungeonConstructionGUI::scrollRooms(TransferSide side, int offset)
{
	const int roomCount = dungeon.getRooms().size();
	int& index = side == TransferSide::Left ? leftRoomIndex : rightRoomIndex;
	index = (index + offset) % roomCount;
	if (index < 0)
	{
		index += roomCount;
	}
	ASSERT(index >= 0 && index < roomCount);

	changeState(State::TransferingInhabitants);
}

void DungeonConstructionGUI::createTransferGUI(TransferSide side)
{
	const int roomIndex = side == TransferSide::Left ? leftRoomIndex : rightRoomIndex;
	const Room& room = *dungeon.getRooms()[roomIndex];
	const auto& npcList = room.getInhabitants();
	const int transferGUIWidth = 164;
	const int transferbuttonWidth = 24;
	const int transferGUIGap = 16;
	const int baseOffsetX = side == TransferSide::Left ? leftBorder : leftBorder + transferGUIWidth + transferGUIGap;
	const int baseOffsetY = topBorder + buttonHeight + buttonGap;
	const int nameListOffsetX = side == TransferSide::Left ? baseOffsetX : baseOffsetX + transferbuttonWidth;
	const int nameListHeight = 32;
	const int transferButtonOffsetX = side == TransferSide::Left ? baseOffsetX + transferGUIWidth - transferbuttonWidth : baseOffsetX;
	const int topBarHeight = 24;
	const int roomSwitchbuttonWidth = 16;
	const int transferGUIHeight = topBarHeight + npcList.size() * nameListHeight;

	auto widget = unique<gui::GUIWidget>(Rect<int>(baseOffsetX, baseOffsetY, transferGUIWidth, transferGUIHeight));

	const gui::NormalBackground background(sf::Color::Black, sf::Color::White);

	std::stringstream ss;
	if (roomIndex == 0)
	{
		ss << "Holding Cell";
	}
	else
	{
		ss << "Room " << roomIndex;
	}
	auto nameList = unique<gui::TextLabel>(Rect<int>(baseOffsetX + roomSwitchbuttonWidth, baseOffsetY - 3, transferGUIWidth - roomSwitchbuttonWidth * 2, topBarHeight), ss.str(), gui::TextLabel::Config().centered());
	widget->addWidget(std::move(nameList));

	auto createRoomChanger = [&](int x, int offset, const char *str)
	{
		const Rect<int> butBox(x, baseOffsetY, roomSwitchbuttonWidth, topBarHeight);
		auto but = gui::Button::createGenericButton(butBox, str, std::bind(&DungeonConstructionGUI::scrollRooms, this, side, offset));
		widget->addWidget(std::move(but));
	};
	createRoomChanger(baseOffsetX, -1, "<" );
	createRoomChanger(baseOffsetX + transferGUIWidth - roomSwitchbuttonWidth, 1, ">");

	for (int i = 0; i < npcList.size(); ++i)
	{
		Soul& soul = *npcList[i]->getSoul();
		const int yOffset = i * nameListHeight + baseOffsetY + topBarHeight;

		// name
		auto nameSlot = unique<gui::GUIWidget>(Rect<int>(nameListOffsetX, yOffset, transferGUIWidth - transferbuttonWidth, nameListHeight));
		const Rect<int> nameBox(nameListOffsetX + 32, yOffset, transferGUIWidth - 32 - transferbuttonWidth, nameListHeight);
		std::unique_ptr<gui::Button> nameButton = gui::Button::createGenericButton(nameBox, SoulConjugator(soul).fullName(), [&](){
			auto statusGUI = unique<NPCStatusGUI>(380, 180, soul);
			auto window = unique<gui::FloatingWindow>(std::move(statusGUI));
			window->setTitle(SoulConjugator(soul).fullName());
			registerSubstateGUI(std::move(window));
			//FloatingWindow *popupWindow = new FloatingWindow(std::move(statusGUI));
			//scene->registergui::GUIWidget(popupWindow);
		});
		nameSlot->addWidget(std::move(nameButton));

		// preview image (part of name widget)
		auto npcPreview = unique<gui::DrawableContainer>(Rect<int>(nameListOffsetX, yOffset, 32, 32));
		npcPreview->setDrawable(LayeredSpriteLoader::create(LayeredSpriteLoader::Config(soul.appearance, "walk", 32, Vector2i(-4, -4)), "south"));
		nameSlot->addWidget(std::move(npcPreview));
		nameSlot->applyBackground(background);
		widget->addWidget(std::move(nameSlot));

		// transfer gui::Button
		const char *transferText = side == TransferSide::Left ? "->" : "<-";
		const Rect<int> transferbuttonBox(transferButtonOffsetX, yOffset, transferbuttonWidth, nameListHeight);
		auto transferButton = gui::Button::createGenericButton(transferbuttonBox, transferText, std::bind(&DungeonConstructionGUI::transferNPC, this, side, i));
		//transfergui::Button->applyBackground(background);
		widget->addWidget(std::move(transferButton));
	}
	widget->applyBackground(background);
	registerSubstateGUI(std::move(widget));
}

void DungeonConstructionGUI::addObject(int x, int y, const std::string& type, const json11::Json::object *instanceAttributes)
{
	const towngen::HomeObjectPrototype *prototype = towngen::HomeObjectDatabase::get().getPrototype(type);
	ASSERT(prototype);
	if (prototype)
	{
		auto undo = prototype->generate(Random::get(), dungeon.getHome(), Vector2i(x, y), nullptr, nullptr, nullptr, instanceAttributes, nullptr);
		objects.insert(undo.get());
		if (instanceAttributes)
		{
			this->instanceAttributes[undo.get()] = *instanceAttributes;
		}
		towngen::EditingUndo *undoPtr = undo.get();
		towngen::EditingUndo *instanceAttributeKey = instanceAttributes ? undoPtr : nullptr;
		undo->addUndoFunc([&, undoPtr, instanceAttributeKey]
		{
			objects.erase(undoPtr);
			if (instanceAttributeKey)
			{
				objects.erase(instanceAttributeKey);
			}
		});
		dungeon.getHome().addEntityToQuad(undo.release());
		const Rect<int>& tileArea = prototype->getTileArea();
		const int postx = x / tileSize;
		const int posty = y / tileSize;
		for (int i = 0; i < tileArea.width; ++i)
		{
			for (int j = 0; j < tileArea.height; ++j)
			{
				dungeon.setHasObject(tileArea.left + i + postx, tileArea.top + j + posty, true);
			}
		}
	}
}

void DungeonConstructionGUI::addTile(int tx, int ty, const std::string& id)
{
	const Vector2i tPos(tx, ty);
	dungeon.getHome().setTile(tPos, id);
	tiles[tPos] = id;
}

void DungeonConstructionGUI::addWall(int tx, int ty, const std::string& id)
{
	BuildingLevel& level = dungeon.getHome();
	const Vector2i tPos(tx, ty);
	walls[tPos] = id;
	level.removeTile(tPos);
	auto it = tiles.find(tPos);
	if (it != tiles.end())
	{
		tiles.erase(it);
	}
	const int txe = std::min(tx + 1, level.getWidth() / level.getTileSize());
	const int tye = std::min(ty + 1, level.getHeight() / level.getTileSize());
	for (int i = std::max(0, tx - 1); i < txe; ++i)
	{
		for (int j = std::max(0, ty - 1); j < tye; ++j)
		{
			if (i != 0 || j != 0)
			{
				removeWallIfRedundant(i, j);
			}
		}
	}
}

void DungeonConstructionGUI::removeWallIfRedundant(int tx, int ty)
{
	auto it = walls.find(Vector2i(tx, ty));
	if (it != walls.end())
	{
		const bool wallLeft = tx == 0 || walls.count(Vector2i(tx - 1, ty)) > 0;
		const bool wallRight = tx == dungeon.getHome().getWidth() / dungeon.getHome().getTileSize() || walls.count(Vector2i(tx + 1, ty)) > 0;
		const bool wallUp = ty == 0 || walls.count(Vector2i(tx, ty - 1)) > 0;
		const bool wallDown = ty == dungeon.getHome().getHeight() / dungeon.getHome().getTileSize() || walls.count(Vector2i(tx, ty + 1)) > 0;
		if (wallLeft && wallRight && wallUp && wallDown)
		{
			walls.erase(it);
		}
	}
}

int DungeonConstructionGUI::snapToGrid(int x) const
{
	return dungeon.getHome().getTileSize() * (x / dungeon.getHome().getTileSize());
}

int DungeonConstructionGUI::snapToHalfGrid(int x) const
{
	return 2 * snapToGrid(x) / 2;
}

Vector2i DungeonConstructionGUI::snapToHalfGrid(int x, int y)
{
	return Vector2i(snapToHalfGrid(x), snapToHalfGrid(y));
}

Vector2i DungeonConstructionGUI::screenToLevel(int x, int y) const
{
	return Vector2i(viewBox.left + x, viewBox.top + y);
}


bool DungeonConstructionGUI::canPlaceObject(int mx, int my) const
{
	const Vector2i pos = screenToLevel(mx, my);
	const int ts = dungeon.getHome().getTileSize();
	const int tx = pos.x / ts;
	const int ty = pos.y / ts;
	if (!dungeon.isEditableTile(tx, ty))
	{
		return false;
	}
	const towngen::HomeObjectPrototype *prototype = towngen::HomeObjectDatabase::get().getPrototype(currentObject);
	ASSERT(prototype);
	const Rect<int>& tileArea = prototype->getTileArea();
	for (int i = 0; i < tileArea.width; ++i)
	{
		for (int j = 0; j < tileArea.height; ++j)
		{
			if (dungeon.hasObject(tileArea.left + i + tx, tileArea.top + j + ty))
			{
				return false;
			}
		}
	}
	return true;
}

} // sl
