/**
 * @section DESCRIPTION
 * A scrollable box of tiles for selecting info about some kind of DecorTypeDatabase's contents
 */
#ifndef SL_ITEMGHOOSERGUI_HPP
#define SL_ITEMGHOOSERGUI_HPP

#include "Engine/Graphics/Sprite.hpp"
#include "Engine/GUI/DrawableContainer.hpp"
#include "Engine/GUI/GUIWidget.hpp"
#include "Engine/GUI/Scrollbar.hpp"
#include "Engine/GUI/WidgetGrid.hpp"
#include "Home/DecorTypeDatabase.hpp"
#include "Utility/Memory.hpp"


namespace sl
{

namespace gui
{
class WidgetGrid;
class Scrollbar;
} // gui

template <typename T>
class ItemChooserGUI : public gui::GUIWidget
{
public:
	typedef const DecorTypeDatabase<T> ItemDB;

	/**
	 * @param x The x position of the window
	 * @param y The y position of the window
	 * @param scrollHeight The height in pixels of the scrollable area
	 * @param topHeight The height in pixels of the top area reserved for whatever you want
	 * @param columns How many tiles are in each row
	 * @param db The DecorTypeDatabase to read from
	 */
	ItemChooserGUI(int x, int y, int scrollHeight, int topHeight, int columns, const ItemDB& db);

	/**
	 * Re-reads all info from the item database and recreates the scrollable tile area
	 */
	void updateItems();

	/**
	 * @return The currently selected item
	 */
	const T& getSelected() const;

	/**
	 * @return The currently selected item's ID
	 */
	typename T::ID getSelectedID() const;

	/**
	 * Registers a callback to be called when a new item is selected
	 * @param callback The function to call
	 */
	void onItemChangeCallback(std::function<void(typename T::ID)> callback);

	/**
	 * Change the currently selected item
	 * @param col The column to select
	 * @param row The row to select
	 */
	void select(int col, int row);

private:

	/**
	 * @param cols The number of columns
	 * @return The width in pixels an ItemChooserGUI should be if it has cols columns
	 */
	int computeWidth(int cols) const;

	/**
	 * @param scrollHeight The height of the scrollable area
	 * @return The height in pixels an ItemChooserGUI should be with a scroll area of scorllHeight
	 */
	int computeHeight(int scrollHeight) const;

	/**
	 * @param col Input column
	 * @param row Input row
	 * @return the index (ID) in the DB of the item at the input col/row
	 */
	int index(int col, int row) const;




	//! The database to read info from
	const ItemDB& db;
	//! The grid of tiles for the scroll area
	gui::WidgetGrid *grid;
	//! The scrollable area itself
	gui::Scrollbar *scrollBar;
	//! How many columns per row in scrollable area
	int columns;
	//! Height of scrollable area in pixels
	int scrollHeight;
	//! Height in pixels to reserve for top info area
	int topHeight;
	//! Column currently selected
	int colSelected;
	//! Row currently selected
	int rowSelected;
	//! Callback to call (if set) when item selection is changed
	std::function<void(typename T::ID)> callback;
	//! Width (and height... for now) of each item tile in the scroll area
	const int columnWidth;
	//! Width (in pixels) of the scrollbar (note: doesn't do shit - pls update Scrollbar to allow different sizes
	const int scrollbarWidth;
};

// public
template <typename T>
ItemChooserGUI<T>::ItemChooserGUI(int x, int y, int scrollHeight, int topHeight, int columns, const ItemDB& db)
	:GUIWidget(Rect<int>(x, y, 0, 0))
	,db(db)
	,grid(nullptr)
	,scrollBar(nullptr)
	,columns(columns)
	,scrollHeight(scrollHeight)
	,topHeight(topHeight)
	,colSelected(0)
	,rowSelected(0)
	,callback()
	,columnWidth(64)
	,scrollbarWidth(16)
{
	setSize(computeWidth(columns), computeHeight(scrollHeight));
	auto scroll = std::make_unique<gui::Scrollbar>(Rect<int>(x, y + topHeight, computeWidth(columns), scrollHeight), nullptr);
	scrollBar = scroll.get();
	addWidget(std::move(scroll));
	updateItems();
}

template <typename T>
void ItemChooserGUI<T>::updateItems()
{
	if (grid)
	{
		removeWidget(*grid);
	}
	grid = new gui::WidgetGrid(0, 0, columns, (T::IDCount + columns - 1) / columns);
	//addWidget(std::unique_ptr<GUIWidget>(grid));
	scrollBar->setScrollArea(std::unique_ptr<gui::GUIWidget>(grid));

	for (int i = 0; i < T::IDCount; ++i)
	{
		const typename T::ID id = static_cast<typename T::ID>(i);
		const typename T::Definition& def = db.get(id);
		const int col = i % columns;
		const int row = i / columns;
		const int x = col * columnWidth;
		const int y = topHeight + row * columnWidth;
		auto sprite = unique<Sprite>(def.previewFilename.c_str(), -16, -16);
		auto button = std::make_unique<gui::DrawableContainer>(Rect<int>(x, y, columnWidth, columnWidth));
		button->applyBackground(gui::NormalBackground(sf::Color::Transparent, sf::Color(128, 128, 128)));
		button->setDrawable(std::move(sprite));
		button->registerMouseInputEvent(Key::GUIClick, GUIWidget::InputEventType::Pressed, std::bind(&ItemChooserGUI<T>::select, this, col, row));
		grid->setWidget(col, row, std::move(button));
	}
}

template <typename T>
const T& ItemChooserGUI<T>::getSelected() const
{
	return db.get(getSelectedID());
}

template <typename T>
typename T::ID ItemChooserGUI<T>::getSelectedID() const
{
	return static_cast<typename T::ID>(index(colSelected, rowSelected));
}

template <typename T>
void ItemChooserGUI<T>::onItemChangeCallback(std::function<void(typename T::ID)> callback)
{
	this->callback = callback;
	callback(getSelectedID());
}

template <typename T>
void ItemChooserGUI<T>::select(int col, int row)
{
	if (col != colSelected || row != rowSelected)
	{
		colSelected = col;
		rowSelected = row;

		if (callback)
		{
			callback(getSelectedID());
		}
	}
}

// private
template <typename T>
int ItemChooserGUI<T>::computeWidth(int cols) const
{
	return columns * columnWidth + scrollbarWidth;
}

template <typename T>
int ItemChooserGUI<T>::computeHeight(int scrollHeight) const
{
	return scrollHeight + topHeight;
}

template <typename T>
int ItemChooserGUI<T>::index(int col, int row) const
{
	return col + row * columns;
}


} // sl

#endif // SL_ITEMGHOOSERGUI_HPP
