#include "Home/OpenableDoor.hpp"

#include "Engine/Graphics/Sprite.hpp"
#include "Engine/Input/Input.hpp"
#include "Engine/GUI/NormalBackground.hpp"
#include "Engine/Scene.hpp"
#include "Player/Player.hpp"
#include "Town/StaticGeometry.hpp"

namespace sl
{

OpenableDoor::OpenableDoor(int x, int y, DoorType type)
	:Entity(x, y - 8, 32, 24, "OpenableDoor")
	,opened(false)
	,type(type)
	,geo()
{
}

OpenableDoor::~OpenableDoor()
{
	for (Entity& g : geo)
	{
		g.destroy();
	}
}

bool OpenableDoor::isStatic() const
{
	return true;
}

const Entity::Type OpenableDoor::getType() const
{
	return Entity::makeType("OpenableDoor");
}

void OpenableDoor::onEnter()
{
	ASSERT(geo.empty());
	if (opened)
	{
		open();
	}
	else
	{
		close();
	}
}

void OpenableDoor::onExit()
{
	for (Entity& g : geo)
	{
		g.destroy();
	}
	geo.clear();
}

void OpenableDoor::onUpdate()
{
	Player *player = static_cast<Player*>(scene->checkCollision(*this, "Player"));
	if (player)
	{
		player->addActionCallout(Player::CalloutType::Use, opened ? "Close" : "Open", [this] {
			if (opened)
			{
				close();
			}
			else
			{
				open();
			}
		});
	}
}

void OpenableDoor::open()
{
	opened = true;

	switch (type)
	{
	case DoorType::Regular:
		break;
	case DoorType::SlidingBookcase:
		for (Entity& g : geo)
		{
			g.destroy();
		}
		geo.clear();
		StaticGeometry *g = new StaticGeometry(pos.x + 32, pos.y + 8, 32, 8, unique<Sprite>("furniture/sliding_bookcase.png", 0, 40));
		g->doNotSerialize();
		scene->addEntityToList(g);
		geo.insert(g->getMultiLink());
		break;
	}
}

void OpenableDoor::close()
{
	opened = false;

	switch (type)
	{
	case DoorType::Regular:
		break;
	case DoorType::SlidingBookcase:
		for (Entity& g : geo)
		{
			g.destroy();
		}
		geo.clear();
		StaticGeometry *g = new StaticGeometry(pos.x, pos.y + 8, 32, 8, unique<Sprite>("furniture/sliding_bookcase.png", 0, 40));
		g->doNotSerialize();
		scene->addEntityToList(g);
		geo.insert(g->getMultiLink());
		break;
	}
}

void OpenableDoor::deserialize(Deserialize& in)
{
	deserializeEntity(in);
	LOADVAR(b, opened);
	type = static_cast<DoorType>(in.u8("type"));
	// geo created upon onEnter()
}

void OpenableDoor::serialize(Serialize& out) const
{
	serializeEntity(out);
	SAVEVAR(b, opened);
	out.u8("type", static_cast<uint8_t>(type));
	// ignore geo, can just recreate
}

} // sl
