#include "Home/Room.hpp"

#include <algorithm>
#include <list>
#include <set>

#include "Engine/Serialization/Impl/SerializeVector.hpp"
#include "Engine/Serialization/AutoDeserialize.hpp"
#include "Engine/Serialization/AutoSerialize.hpp"
#include "Engine/Scene.hpp"
#include "NPC/NPC.hpp"

namespace sl
{

Room::Room(RoomDoor& door)
	:positions()
	,inhabitants()
	,door(door)
{
}

void Room::addInhabitant(NPC& npc)
{
	for (const NPC* n : inhabitants)
	{
		if (n == &npc)
		{
			return;
		}
	}
	inhabitants.push_back(&npc);
	npc.setRoom(this);
}

void Room::removeInhabitant(NPC& npc)
{
	for (NPC*& n : inhabitants)
	{
		if (n == &npc)
		{
			npc.setRoom(nullptr);
			n = inhabitants.back();
			inhabitants.pop_back();
			break;
		}
	}
}

const std::vector<NPC*>& Room::getInhabitants() const
{
	return inhabitants;
}

int Room::getID() const
{
	return computeID(door.x(), door.y());
}

std::vector<Vector2i>& Room::getPositions()
{
	return positions;
}

const std::vector<Vector2i>& Room::getPositions() const
{
	return positions;
}

void Room::findInhabitants(Scene *scene, bool removeOldOnes)
{
	std::set<NPC*> foundNPCs;
	if (!removeOldOnes)
	{
		for (NPC *npc : inhabitants)
		{
			foundNPCs.insert(npc);
		}
	}
	
	std::list<NPC*> queryList;
	const int ts = scene->getTileSize();
	Rect<int> queryBox(ts, ts);
	for (const Vector2i& pos : positions)
	{
		queryList.clear();
		queryBox.left = pos.x * ts;
		queryBox.top = pos.y * ts;
		scene->cull(queryList, queryBox);
		for (NPC *npc : queryList)
		{
			foundNPCs.insert(npc);
		}
	}
	
	inhabitants.resize(foundNPCs.size());
	std::move(foundNPCs.begin(), foundNPCs.end(), inhabitants.begin());
}

void Room::deserialize(Deserialize& in)
{
	AutoDeserialize::fields(in, positions, inhabitants);
}

void Room::serialize(Serialize& out) const
{
	AutoSerialize::fields(out, positions, inhabitants);
}

/*static*/ int Room::computeID(int x, int y)
{
	return x + (y << 16);
}

} // sl