#include "Home/WallGrid.hpp"

#include <SFML/Graphics/RectangleShape.hpp>

#include "Engine/Graphics/Sprite.hpp"
#include "Engine/Graphics/SFDrawable.hpp"
#include "Engine/Graphics/TextureManager.hpp"
#include "Engine/Graphics/TileGrid.hpp"
#include "Engine/Scene.hpp"
#include "Home/RoomDoor.hpp"
#include "Town/Generation/Interiors/HomeTileDatabase.hpp"
#include "Town/StaticGeometry.hpp"
#include "Utility/Assert.hpp"
#include "Utility/Logger.hpp"
#include "Utility/Memory.hpp"

namespace sl
{



WallGrid::WallGrid(int x, int y, int width, int height, int tileWidth, int tileHeight, Scene& scene, std::function<const std::string&(int, int)> wallType)
	:wallEntities(width, height)
	,walls(width, height, true)
	,wallType(std::move(wallType))
	,doorGrid(width, height, nullptr)
	,scene(scene)
	,x(x)
	,y(y)
	,width(width)
	,height(height)
	,tileWidth(tileWidth)
	,tileHeight(tileHeight)
{
}

bool WallGrid::isSolid(int x, int y) const
{
	return walls.at(x, y);
}

bool WallGrid::canGoRight(int x, int y) const
{
	const bool one = !isSolid(x, y);
	const bool two = x < walls.getWidth() - 1;
	const bool three = !(walls.at(x + 1, y));
	return one && two && three;
}

bool WallGrid::canGoLeft(int x, int y) const
{
	const bool one = !isSolid(x, y);
	const bool two = x > 0;
	const bool three = !(walls.at(x - 1, y));
	return one && two && three;
}

bool WallGrid::canGoUp(int x, int y) const
{
	if (y > 0 && doorGrid.at(2 * (x / 2), 2 * ((y - 2) / 2)))
	{
		if (!doorGrid.at(2 * (x / 2), 2 * ((y - 2) / 2))->isOpen())
		{
			return false;
		}
	}
	return !isSolid(x, y) && y > 0 && !(walls.at(x, y - 1));
}

bool WallGrid::canGoDown(int x, int y) const
{
	if (doorGrid.at(2 * (x / 2), 2 * (y / 2)))
	{
		if (!doorGrid.at(2 * (x / 2), 2 * (y / 2))->isOpen())
		{
			return false;
		}
	}
	const bool one = !isSolid(x, y);
	const bool two = y < walls.getHeight() - 1;
	const bool three = !(walls.at(x, y + 1));
	return one && two && three;
}

int WallGrid::getWidth() const
{
	return width;
}

int WallGrid::getHeight() const
{
	return height;
}

int WallGrid::getTileSize() const
{
	// @todo support both EVERYWHERE (must change in a buttload of places)
	ASSERT(tileWidth == tileHeight);
	return tileWidth;
}



void WallGrid::print() const
{
	Logger::log(Logger::Temp, Logger::Debug) << "\n\n\n\n\nGRID:\n";
	for (int j = 0; j < height; ++j)
	{
		for (int i = 0; i < width; ++i)
		{
			Logger::log(Logger::Temp, Logger::Debug) << (walls.at(i, j) ? 'X' : ' ');
		}
		Logger::log(Logger::Temp, Logger::Debug) << "\n";
	}
	Logger::log(Logger::Temp, Logger::Debug) << "\n";
}

void WallGrid::fill(int x, int y)
{
	bool& wall = walls.at(x, y);
	if (!wall)
	{
		wall = true;
		updateTilesSurrounding(x, y);
	}
}

void WallGrid::remove(int x, int y)
{
	bool& wall = walls.at(x, y);
	if (wall)
	{
		wall = false;
		updateTilesSurrounding(x, y);
	}
}

void WallGrid::closeAllDoors()
{
	for (auto it = doorGrid.begin(), end = doorGrid.end(); it != end; ++it)
	{
		RoomDoor *door = *it;
		if (door && door->isOpen())
		{
			door->close();
			updateTile(it.getX(), it.getY());
		}
	}
}

void WallGrid::openAllDoors()
{
	for (auto it = doorGrid.begin(), end = doorGrid.end(); it != end; ++it)
	{
		if (*it && !(*it)->isOpen())
		{
			(*it)->open();
			updateTile(it.getX(), it.getY());
		}
	}
}

bool WallGrid::closeDoorAt(int x, int y)
{
	RoomDoor *door = doorGrid.at(x, y);
	if (door)
	{
		if (door->isOpen())
		{
			door->close();
			updateTile(x, y);
		}
		return true;
	}
	return false;
}

bool WallGrid::openDoorAt(int x, int y)
{
	RoomDoor *door = doorGrid.at(x, y);
	if (door)
	{
		if (!door->isOpen())
		{
			door->open();
			updateTile(x, y);
		}
		return true;
	}
	return false;
}

const RoomDoor* WallGrid::getDoorAt(int x, int y) const
{
	return doorGrid.at(x, y);
}

void WallGrid::addDoor(RoomDoor& door)
{
	doorGrid.at(door.x(), door.y()) = &door;
	updateTile(door.x(), door.y());
}

void WallGrid::removeDoor(RoomDoor& door)
{
	doorGrid.at(door.x(), door.y()) = nullptr;
	updateTile(door.x(), door.y());
}

void WallGrid::updateAllTiles()
{
	for (int i = 0; i < width; ++i)
	{
		for (int j = 0; j < height; ++j)
		{
			updateTile(i, j);
		}
	}
}

void WallGrid::reset(int x, int y, int width, int height, int tileWidth, int tileHeight)
{
	this->x = x;
	this->y = y;
	this->width = width;
	this->height = height;
	this->tileWidth = tileWidth;
	this->tileHeight = tileHeight;
	wallEntities.resize(width, height);
	walls.resize(width, height, true);
	doorGrid.resize(width, height, nullptr);
}

Scene& WallGrid::getScene()
{
	return scene;
}

/*			private			*/
void WallGrid::updateTilesSurrounding(int x, int y)
{
	for (int i = std::max(0, x - 1); i < std::min(width, x + 2); ++i)
	{
		for (int j = std::max(0, y - 1); j < std::min(height, y + 2); ++j)
		{
			updateTile(i, j);
		}
	}
}

void WallGrid::updateTile(int x, int y)
{
	for (Entity *entity : wallEntities.at(x, y))
	{
		entity->destroy();
	}
	wallEntities.at(x, y).clear();
	if (walls.at(x, y))
	{
		const bool wallRight = x < width - 1 && walls.at(x + 1, y);
		const bool wallLeft = x > 0 && walls.at(x - 1, y);
		const bool wallUp = y > 0 && walls.at(x, y - 1);
		const bool wallDown = y < height - 1 && walls.at(x, y + 1);
		// only create walls on borders to non-wall areas
		if (!wallRight || !wallLeft || !wallUp || !wallDown)
		{
			StaticGeometry *wall = new StaticGeometry(x * tileWidth, y * tileHeight, 16, 16, nullptr);
			wall->doNotSerialize();

			auto addGreyArea = [&](int gx, int gy, int gw, int gh)
			{
				sf::RectangleShape rect(sf::Vector2f(gw, gh));
				rect.setFillColor(sf::Color(43, 43, 33));
				rect.setOrigin(-gx, -gy);
				auto greyArea = unique<SFDrawable<sf::RectangleShape>>(rect);
				wall->addDrawable(std::move(greyArea));
			};
			// @todo fix the little bugs with wall placement (ie diagonal walls issue)
			if (!wallDown)
			{
				// @todo make some kind of wall database or something to keep track of offsets and distinguish between types
				// and use wallType(x, y) to do that
				auto top = unique<Sprite>("tiles/inside_tiles.png", 0, tileHeight);
				auto bottom = unique<Sprite>("tiles/inside_tiles.png", 0);
				top->setRenderedSubrect(Rect<int>(1 * tileWidth, 3 * tileHeight, tileWidth, tileHeight));
				bottom->setRenderedSubrect(Rect<int>(0, 2 * tileHeight, tileWidth, tileHeight));
				wall->addDrawable(std::move(top));
				wall->addDrawable(std::move(bottom));
				if (wallLeft && x > 0 && y < height - 1 && walls.at(x - 1, y + 1))
				{
					const float ghScale = (y < height - 2 && walls.at(x - 1, y + 2)) ? 2.5f : 1.5f;
					addGreyArea(-tileWidth / 2, -1.5f * tileHeight, tileWidth / 2, ghScale * tileHeight);
				}
				if (wallRight && x < width - 1 && y < height - 1 && walls.at(x + 1, y + 1))
				{
					const float ghScale = (y < height - 2 && walls.at(x - 1, y + 2)) ? 2.5f : 1.5f;
					addGreyArea(tileWidth, -1.5f * tileHeight, tileWidth / 2, ghScale * tileHeight);
				}
				addGreyArea(0, -1.5f * tileHeight, tileWidth, tileHeight / 2);
			}
			if (!wallUp)
			{
				int gw = tileWidth;
				int gx = 0;
				if (wallLeft)
				{
					gw += tileWidth / 2;
					gx -= tileWidth / 2;
				}
				if (wallRight)
				{
					gw += tileWidth / 2;
				}
				addGreyArea(gx, 0, gw, tileHeight / 2);
			}
			if (!wallLeft)
			{
				addGreyArea(0, 0, tileWidth / 2, tileHeight);
			}
			if (!wallRight)
			{
				addGreyArea(tileWidth / 2, 0, tileWidth / 2, tileHeight);
			}
			wallEntities.at(x, y).push_back(wall);
			scene.addImmediately(wall, StorageType::InGrid);
		}
	}
	else
	{

		// doors
		if (doorGrid.at(x, y))
		{
			Entity *entity = nullptr;
			if (doorGrid.at(x, y)->isOpen())
			{
				entity = new StaticGeometry(x * tileWidth, y * tileHeight, 0, 0, unique<Sprite>("cell_door_open.png"));
			}
			else
			{
				entity = new StaticGeometry(x * tileWidth, y * tileHeight + 30, 32, 4, unique<Sprite>("cell_door.png", 0, 30));
			}
			ASSERT(entity);
			wallEntities.at(x, y).push_back(entity);
			scene.addImmediately(entity, StorageType::InList);
		}
	}
}

} // sl
