#include "Items/Item.hpp"

#include <map>
#include <sstream>

#include "Engine/Serialization/Serialization.hpp"
#include "Engine/Serialization/AutoDefineObjectSerialization.hpp"
#include "Utility/Assert.hpp"
#include "Utility/JsonUtil.hpp"
#include "Utility/Memory.hpp"
#include "Utility/ToString.hpp"

namespace sl
{

int UNLIMITED_USES = -1;

std::vector<Item::ItemData> Item::staticItemData;

std::map<std::string, Item::Type> Item::typeMap;

Item::StrCat Item::categoryMap[CATEGORY_COUNT] = {
	{"misc",       Item::Misc      },
	{"consumable", Item::Consumable},
	{"equipable",  Item::Equipable },
	{"edible",     Item::Edible    },
	{"useable",    Item::Usable    },
	{"toy",        Item::Toy       },
	{"candy",      Item::Candy     }
};

/*static*/void Item::initStaticData()
{
	static bool loaded = false;
	if (!loaded)
	{
		const std::string itemsFile("data/items.json");
		std::string err;
		json11::Json data = readJsonFile(itemsFile, err);
		if (err.empty())
		{
			for (const json11::Json& object : data.array_items())
			{
				const json11::Json::object& vars = object.object_items();
				const std::string& type = vars.at("type").string_value();

				auto ret = typeMap.insert(std::make_pair(type, typeMap.size()));
				if (!ret.second)
				{
					ERROR("duplicate definition of \'" + type + "\' in " + itemsFile);
				}
				staticItemData.push_back(ItemData());
				ItemData& data = staticItemData.back();

				data.typeName = type;
				auto weight = vars.find("weight");
				if (weight != vars.end())
				{
					data.weight = unique<int>(weight->second.int_value());
				}
				auto value = vars.find("value");
				if (value != vars.end())
				{
					data.value = unique<int>(value->second.int_value());
				}
				auto name = vars.find("name");
				if (name != vars.end())
				{
					data.name = unique<std::string>(name->second.string_value());
				}
				auto desc = vars.find("desc");
				if (desc != vars.end())
				{
					data.desc = unique<std::string>(desc->second.string_value());
				}
				auto maxUses = vars.find("maxUses");
				if (maxUses != vars.end())
				{
					data.maxUses = unique<int>(maxUses->second.int_value());
				}
				auto equip = vars.find("equip");
				if (equip != vars.end())
				{
					const json11::Json::object& equipObj = equip->second.object_items();
					data.equipModifiers = unique<EquipModifiers>();
					data.equipModifiers->dex = equipObj.at("dex").int_value();
					data.equipModifiers->def = equipObj.at("def").int_value();
					data.equipModifiers->style = equipObj.at("style").int_value();
				}
				auto nutrition = vars.find("nutrition");
				if (nutrition != vars.end())
				{
					const json11::Json::object& nutObj = nutrition->second.object_items();
					data.nutrition = unique<Nutrition>();
#define LOAD_NUTRTITION_MACRO(macro) if (nutObj.count(#macro)) data.nutrition->macro = nutObj.at(#macro).number_value()
					LOAD_NUTRTITION_MACRO(starch);
					LOAD_NUTRTITION_MACRO(fiber);
					LOAD_NUTRTITION_MACRO(sugar);
					LOAD_NUTRTITION_MACRO(omega3);
					LOAD_NUTRTITION_MACRO(polyunsaturated);
					LOAD_NUTRTITION_MACRO(monounsaturated);
					LOAD_NUTRTITION_MACRO(saturated);
					LOAD_NUTRTITION_MACRO(trans);
					LOAD_NUTRTITION_MACRO(protein);
#undef LOAD_NUTRITION_MACRO
				}
				auto flavor = vars.find("flavor");
				if (flavor != vars.end())
				{
					const json11::Json::object& flavObj = flavor->second.object_items();
					data.flavor = unique<FlavorProfile>();
#define LOAD_FLAVOR(axis) if (flavObj.count(#axis)) data.flavor->axis() = flavObj.at(#axis).number_value()
					LOAD_FLAVOR(sweet);
					LOAD_FLAVOR(sour);
					LOAD_FLAVOR(salty);
					LOAD_FLAVOR(bitter);
					LOAD_FLAVOR(umami);
					LOAD_FLAVOR(spicy);
#undef LOAD_FLAVOR
				}
				auto subdue = vars.find("subdue");
				if (subdue != vars.end())
				{
					const json11::Json::object& subdueObj = subdue->second.object_items();
					data.subdueItem = unique<SubdueItem>();
					const std::map<std::string, SubdueItem::ResultingState> strToState = {
						{"tied", SubdueItem::ResultingState::Tied},
						{"chloro", SubdueItem::ResultingState::Chloroformed}
					};
					data.subdueItem->successState = strToState.at(subdueObj.at("success_state").string_value());
					data.subdueItem->timer = subdueObj.at("timer").int_value();
				}
				auto stalk = vars.find("stalk");
				if (stalk != vars.end())
				{
					const json11::Json::object& stalkObj = stalk->second.object_items();
					data.stalkItem = unique<StalkItem>();
					data.stalkItem->photo = stalkObj.at("photo").bool_value();
					data.stalkItem->timer = stalkObj.at("timer").int_value();
					data.stalkItem->radius = stalkObj.at("radius").number_value();
				}
				auto categories = vars.find("categories");
				if (categories != vars.end())
				{
					data.categories = Misc;
					for (const json11::Json& cat : categories->second.array_items())
					{
						data.categories |= strToCategory(cat.string_value());
					}
				}
			}
		}
		else
		{
			ERROR("error opening or parsing " + itemsFile + ":  " + err);
		}
		loaded = true;
	}
}

Item::Item()
	:type(-1)
	,uses(-2)
{
	initStaticData();
}

Item::Item(Type type)
	:type(std::move(type))
	,uses(data().maxUses ? *data().maxUses : UNLIMITED_USES)
{
	initStaticData();
}

bool Item::hasComponent(Category category) const
{
	return static_cast<uint32_t>(category) & data().categories;
}

bool Item::hasComponents(Categories categories) const
{
	return (categories & data().categories) == categories;
}

Item::Type Item::getType() const
{
	return type;
}

int Item::getWeight() const
{
	ASSERT(data().weight);
	return *data().weight;
}

int Item::getValue() const
{
	ASSERT(data().value);
	return *data().value;
}

std::string Item::getName() const
{
	ASSERT(data().name);
	const std::string& baseName = *data().name;
	std::stringstream buffer;
	// @TODO make items have owners and add that here
	//switch (type)
	{
	// item-specific overrides go here
	//default:
		buffer << baseName;
	}
	// TODO: add this back?
	//if (hasComponent(Category::Consumable) && getMaxUses() > 1)
	//{
	//	buffer << " (" << uses << "/" << getMaxUses() << ")";
	//}
	return std::move(buffer.str());
}

std::string Item::getDesc() const
{
	ASSERT(data().desc);
	return *data().desc;
}

int Item::getMaxUses() const
{
	ASSERT(data().maxUses);
	return *data().maxUses;
}

bool Item::hasUses() const
{
	return uses != -1;
}

int Item::getUses() const
{
	return uses;
}

void Item::setUses(int uses)
{
	ASSERT(uses >= 0 && uses <= *data().maxUses);
	this->uses = uses;
}

const Item::EquipModifiers& Item::getEquipModifiers() const
{
	ASSERT(data().equipModifiers);
	return *data().equipModifiers;
}

const Nutrition& Item::getNutrition() const
{
	ASSERT(data().nutrition);
	return *data().nutrition;
}

const FlavorProfile& Item::getFlavor() const
{
	ASSERT(data().flavor);
	return *data().flavor;
}

float Item::getGirth() const
{
	ASSERT(data().girth);
	return *data().girth;
}

float Item::getLength() const
{
	ASSERT(data().length);
	return *data().length;
}

const SubdueItem* Item::getSubdueStats() const
{
	return data().subdueItem.get();
}

const StalkItem* Item::getStalkStats() const
{
	return data().stalkItem.get();
}

bool Item::operator<(const Item& rhs) const
{
	if (type == rhs.type)
	{
		return uses < rhs.uses;
	}
	else
	{
		return type < rhs.type;
	}
}


void Item::serialize(Serialize& out) const
{
	out.startObject("Item");
	// don't bother serializing static data in staticItemData
	out.i("type", type);
	out.i("uses", uses);
	out.endObject("Item");
}

void Item::deserialize(Deserialize& in)
{
	in.startObject("Item");
	type = in.i("type");
	uses = in.i("uses");
	in.endObject("Item");
}



/*static*/ Item::Categories Item::categories(Type type)
{
	return staticItemData.at(type).categories;
}

/*static*/ Item::Type Item::strToType(const std::string& name)
{
	initStaticData();
	auto it = typeMap.find(name);
	if (it != typeMap.end())
	{
		return it->second;
	}
	ERROR("type not defined: " + name);
	return -1;
}

/*static*/ const std::string& Item::typeToStr(Type type)
{
	return staticItemData[static_cast<int>(type)].typeName;
}

/*static*/ Item::Category Item::strToCategory(const std::string& name)
{
	for (int i = 0; i < CATEGORY_COUNT; ++i)
	{
		if (categoryMap[i].str == name)
		{
			return categoryMap[i].cat;
		}
	}
	ERROR("unknown category: " + name);
	return None;
}

/*static*/ const std::string& Item::categoryToStr(Category category)
{
	static std::string invalid("invalid");
	for (int i = 0; i < CATEGORY_COUNT; ++i)
	{
		if (categoryMap[i].cat == category)
		{
			return categoryMap[i].str;
		}
	}
	ERROR("invalid category: " + std::to_string(category) + "; cannot use compound categories");
	return invalid;
}

// private
//Item::ItemData& Item::data()
//{
//	return staticItemData.at(type);
//}

const Item::ItemData& Item::data() const
{
	return staticItemData.at(type);
}

DEF_SERIALIZE_OBJ_MANUAL(Item)

} // sl
