#ifndef SL_ITEM_HPP
#define SL_ITEM_HPP

#include <cstdint>
#include <map>
#include <memory>
#include <string>
#include <utility>
#include <vector>

#include "Engine/Graphics/Drawable.hpp"
#include "Engine/Serialization/AutoDeclareObjectSerialization.hpp"
#include "Items/StalkItem.hpp"
#include "Items/SubdueItem.hpp"
#include "NPC/Stats/FlavorProfile.hpp"
#include "NPC/Stats/Nutrition.hpp"

namespace sl
{

class Deserialize;
class Serialize;

class Item
{
public:
	typedef int Type;
	enum Category : uint32_t
	{
		None       = 0,
		Misc       = 1,
		Consumable = 1<<1,
		Equipable  = 1<<2,
		Edible     = 1<<3,
		Usable     = 1<<4,
		Toy        = 1<<5,
		Insertible = 1<<6,
		Candy      = 1<<7
	};
	typedef uint32_t Categories;
	static const int CATEGORY_COUNT = 8; // don't count None
	static const Categories AllCategories = (1 << CATEGORY_COUNT) - 1;

	struct EquipModifiers
	{
		int dex;
		int def;
		int style;
	};

	Item(); // for deserializaiton
	
	//Item(const std::string& type);

	Item(Type type);

	bool hasComponent(Category category) const;

	bool hasComponents(Categories categories) const;

	Type getType() const;
	int getWeight() const;
	int getValue() const;
	std::string getName() const;
	std::string getDesc() const;
	bool hasUses() const;
	int getMaxUses() const;
	int getUses() const;
	void setUses(int uses);
	const EquipModifiers& getEquipModifiers() const;
	const Nutrition& getNutrition() const;
	const FlavorProfile& getFlavor() const;
	float getGirth() const; // circ in cm
	float getLength() const; // in cm
	const SubdueItem* getSubdueStats() const;
	const StalkItem* getStalkStats() const;

	bool operator<(const Item& rhs) const;



	void serialize(Serialize& out) const;

	void deserialize(Deserialize& in);



	static Categories categories(Type type);

	static Type strToType(const std::string& name);

	static const std::string& typeToStr(Type type);

	static Category strToCategory(const std::string& name);

	static const std::string& categoryToStr(Category category);

	static std::unique_ptr<Drawable> typeToDrawable(Type type);

private:
	struct ItemData
	{
		std::string typeName; // for inverse-lookup on type
		std::unique_ptr<int> weight;
		std::unique_ptr<int> value;
		std::unique_ptr<std::string> name;
		std::unique_ptr<std::string> desc;
		std::unique_ptr<int> maxUses;
		std::unique_ptr<Item::EquipModifiers> equipModifiers;
		std::unique_ptr<Nutrition> nutrition;
		std::unique_ptr<FlavorProfile> flavor;
		std::unique_ptr<float> girth;
		std::unique_ptr<float> length;
		std::unique_ptr<SubdueItem> subdueItem;
		std::unique_ptr<StalkItem> stalkItem;
		Item::Categories categories;
	};



	//ItemData& data();

	const ItemData& data() const;

	static void initStaticData();


	Type type;
	int uses;
	
	static std::vector<ItemData> staticItemData;

	static std::map<std::string, Type> typeMap;

	struct StrCat
	{
		std::string str;
		Category cat;
	};
	static StrCat categoryMap[CATEGORY_COUNT];
};

DEC_SERIALIZE_OBJ(Item)

} // namespace sl

#endif // SL_ITEM_HPP
