#include "Items/ItemUseTarget.hpp"

#include "Engine/Scene.hpp"
#include "NPC/NPC.hpp"
#include "NPC/Locators/Target.hpp"
#include "Utility/Assert.hpp"
#include "Utility/Trig.hpp"

#include <algorithm>

namespace sl
{

static float UNSET_DIRECTION = -1.f;

/*static*/ ItemUseTarget ItemUseTarget::fromRadius(float radius)
{
	ItemUseTarget instance(Mode::Circle);
	instance.radius = radius;
	return instance;
}

/*static*/ ItemUseTarget ItemUseTarget::fromRadiusWithOffset(float radius, float direction, float offset)
{
	ASSERT(direction >= 0.f && direction < 360.f);
	ItemUseTarget instance = fromRadius(radius);
	instance.direction = direction;
	instance.offset = offset;
	return instance;
}

/*static*/ ItemUseTarget ItemUseTarget::fromLine(float direction, float radius)
{
	ASSERT(direction >= 0.f && direction < 360.f);
	ItemUseTarget instance(Mode::Line);
	instance.direction = direction;
	instance.radius = radius;
	return instance;
}

bool ItemUseTarget::isTargetViable(const Target& target, const NPC& npc) const
{
	const Vector2f tpos = target.getPos();
	const Vector2f npos = npc.getPos();
	switch (mode)
	{
	case Mode::Circle:
		return pointDistance(npos, tpos) <= radius;
	case Mode::Line:
		return pointDistance(npos, tpos) <= radius && npc.getMask().intersects(tpos, npos);
	}
	if (!types.empty() && std::find(types.begin(), types.end(), npc.getSoul()->getType()) == types.end())
	{
		return false;
	}
	if (predicate && !predicate(npc))
	{
		return false;
	}
	return true;
}

bool ItemUseTarget::query(std::vector<NPC*>& output, const Target& target) const
{
	std::list<Entity*> entitiesFound;
	static const auto npcType = TypeToString<NPC>();
	switch (mode)
	{
	case Mode::Circle:
		target.getScene()->cull(entitiesFound, Rect<int>(target.getX() - radius + lengthdirX(offset, direction), target.getY() - radius + lengthdirY(offset, direction), 2 * radius, 2 * radius), npcType);
		break;
	case Mode::Line:
		target.getScene()->cullFiniteLine(entitiesFound, target.getPos(), target.getPos() + lengthdir(radius, direction), npcType);
		break;
	}
	bool found = false;
	for (Entity *e : entitiesFound)
	{
		NPC *npc = static_cast<NPC*>(e);
		if (isTargetViable(target, *npc))
		{
			output.push_back(npc);
			found = true;
		}
	}
	return found;
}

ItemUseTarget& ItemUseTarget::setPredicate(std::function<bool(const NPC&)> pred)
{
	ASSERT(!predicate);
	predicate = std::move(pred);
	return *this;
}

ItemUseTarget& ItemUseTarget::addNPCType(std::string type)
{
	types.push_back(std::move(type));
	return *this;
}

// private
ItemUseTarget::ItemUseTarget(Mode mode)
	:mode(mode)
	,direction(UNSET_DIRECTION)
	,offset(0.f)
{
}

} // sl
