/**
 * @section DESCRIPTION
 * Information about where to target the usage of a {\ref ItemUseAction}.
 * Each individual {\ref ItemUseAction} might have further criteria
 * for rejecting {\ref NPC} targets, if any are required for the usage.
 */
#ifndef SL_ITEMUSETARGET_HPP
#define SL_ITEMUSETARGET_HPP

#include <functional>
#include <limits>
#include <string>
#include <vector>

namespace sl
{

class NPC;
class Target;

class ItemUseTarget
{
public:
	static ItemUseTarget fromRadius(float radius);

	static ItemUseTarget fromRadiusWithOffset(float radius, float direction, float offset);

	static ItemUseTarget fromLine(float direction, float radius = std::numeric_limits<float>::max());

	/**
	 * @param npc The NPC to check if it's viable according to the target info supplied.
	 * @return If {\ref npc} is a viable target at {\ref target}
	 */
	bool isTargetViable(const Target& target, const NPC& npc) const;

	/**
	* @param output Viable targets for this item to be used on. Is only added to, not cleared. [output variable]
	* @param target The location the usage is targeted at. For undirected, this will be the center.
	* @return True if any NPC targets were added to output, False otherwise	 */
	bool query(std::vector<NPC*>& output, const Target& target) const;

	ItemUseTarget& setPredicate(std::function<bool(const NPC&)> pred);

	// works as OR, obviously... applied on top of predicate.
	ItemUseTarget& addNPCType(std::string type);

private:
	enum class Mode
	{
		Line,
		Circle
	};
	ItemUseTarget(Mode mode);

	
	Mode mode;
	float radius;
	//! Direction to use it in. Used for both offset-radius and line modes.
	float direction;
	float offset;
	std::vector<std::string> types;
	std::function<bool(const NPC&)> predicate;
};

} // sl

#endif // SL_ITEMUSETARGET_HPP