#include "Menus/HUD.hpp"

#include <sstream>

#include "Town/World.hpp"
#include "Engine/Graphics/SFDrawable.hpp"
#include "Engine/GUI/Bar.hpp"
#include "Engine/GUI/DrawableContainer.hpp"
#include "Engine/GUI/NormalBackground.hpp"
#include "Engine/GUI/TextLabel.hpp"
#include "Player/PlayerStats.hpp"
#include "Player/Actions/PlayerActionFactory.hpp"
#include "Town/Map/Map.hpp"
#include "Utility/DataToString.hpp"


namespace sl
{
namespace gui
{

static constexpr int mapWidth = 512 - 32;
static constexpr int mapHeight = 384 - 32;
static constexpr int minimapWidth = 128;
static constexpr int minimapHeight = 128;

HUD::HUD(const PlayerStats& stats)
	:GUIWidget(Rect<int>(512, 384))// ???
	,stats(stats)
	,hpBar(nullptr)
	,staminaBar(nullptr)
	,hotbar(nullptr)
	,mapWidget(nullptr)
	,map(nullptr)
	,mapMode(MapMode::Minimap)
	,lastAction(NO_CUSTOM_ACTION)
	,actionGUI(nullptr)
{
	auto timeLabel = unique<TextLabel>(Rect<int>(8, 4, 512, 16), "insert time here!", TextLabel::Config().modifedOften());
	time = timeLabel.get();
	addWidget(std::move(timeLabel));

	auto hpBarWidget = unique<Bar>(Rect<int>(8, 20, 100, 6), sf::Color(24, 96, 48), sf::Color(32, 177, 64), sf::Color::White, sf::Color::Red);
	hpBar = hpBarWidget.get();
	addWidget(std::move(hpBarWidget));

	auto staminarBarWidget = unique<Bar>(Rect<int>(8, 32, 96, 6), sf::Color(164, 48, 32), sf::Color(0, 164, 32), sf::Color::White, sf::Color::Black);
	staminaBar = staminarBarWidget.get();
	addWidget(std::move(staminarBarWidget));

	//makeHotbar();
}

void HUD::toggleMapMode()
{
	switch (mapMode)
	{
	case MapMode::Disabled:
		mapMode = MapMode::Minimap;
		break;
	case MapMode::Minimap:
		mapMode = MapMode::Map;
		break;
	case MapMode::Map:
		mapMode = MapMode::Disabled;
		break;
	}
	refreshMapGUI();
}

// private
void HUD::refreshMapGUI()
{
	if (mapWidget)
	{
		removeWidget(*mapWidget);
		mapWidget = nullptr;
	}
	if (map)
	{
		switch (mapMode)
		{
		case MapMode::Disabled:
			break;
		case MapMode::Minimap:
			map->setLOD(map::LOD::Minimap);
			mapWidget = new DrawableContainer(
				Rect<int>(
					512 - 8 - minimapWidth,
					8,
					minimapWidth,
					minimapHeight),
				map);
			mapWidget->applyBackground(gui::NormalBackground(sf::Color::Transparent, sf::Color::White));
			addWidget(std::unique_ptr<GUIWidget>(mapWidget));
			break;
		case MapMode::Map:
			map->setLOD(map::LOD::Map);
			mapWidget = new DrawableContainer(
				Rect<int>(
					(scene->getViewWidth() - mapWidth) / 2,
					(scene->getViewHeight() - mapHeight) / 2,
					mapWidth,
					mapHeight),
				map);
			mapWidget->applyBackground(gui::NormalBackground(sf::Color::Transparent, sf::Color::White));
			addWidget(std::unique_ptr<GUIWidget>(mapWidget));
			break;
		}
	}
}

void HUD::onUpdate(int x, int y)
{
	hpBar->setValue(std::clamp((float)stats.soul->getStatus().hp / stats.soul->getStatus().maxhp, 0.f, 1.f));
	staminaBar->setValue(stats.soul->getStatus().stamina);
	if (map)
	{
		const Scene *s = dynamic_cast<const Scene*>(scene);
		ASSERT(s);
		map->setPositionInMap(s->getView().x / s->getWidth(), s->getView().y / s->getHeight());
	}

	const GameTime& date = scene->getWorld()->getGameTime();

	// @todo refactor this somewhere common
	std::stringstream ss;
	ss << formatTime(date.time) << "   -   " << formatDateLong(date.date) << "       Cash: $" << stats.money;
	time->setString(ss.str());

	if (lastAction != stats.currentActionId)
	{
		lastAction = stats.currentActionId;
		if (actionGUI)
		{
			removeWidget(*actionGUI);
		}
		if (lastAction != NO_CUSTOM_ACTION)
		{
			constexpr int gap = 8;
			constexpr int w = 128;
			constexpr int h = 24;
			constexpr int icon_w = 24;
			auto actionWidget = unique<GUIWidget>(Rect<int>(gap, 384 - gap - h, w, h));
			auto action = playerActionFromId(lastAction);
			actionWidget->addWidget(unique<TextLabel>(Rect<int>(gap + icon_w + 4, 384 - gap - h, w - icon_w, h), action->getName(), TextLabel::Config().centered()));
			actionWidget->addWidget(unique<DrawableContainer>(Rect<int>(gap + 4, 384 - gap - h + 4, 16, 16), action->getIcon()));
			NormalBackground bg(sf::Color(0, 0, 0, 128), sf::Color::White);
			actionWidget->applyBackground(bg);
			actionGUI = actionWidget.get();
			addWidget(std::move(actionWidget));
		}
	}
}

void HUD::onSceneEnter()
{
	const Scene *s = dynamic_cast<const Scene*>(scene);
	ASSERT(s);
	map = s->getMap();
	refreshMapGUI();
}

void HUD::onSceneExit()
{
	map = nullptr;
}

// deprecated?
void HUD::makeHotbar()
{
	//if (hotbar)
	//{
	//	removeWidget(*hotbar);
	//}
	//const int width = 80;
	//const int height = 80;
	//const int gap = 8;
	//const int smallgap = 2;
	//const int textheight = 12;
	//auto hotbarWidget = unique<DrawableContainer>(Rect<int>(gap, 384 - gap - textheight - height, width, height));
	//const Vector2i hbp = hotbarWidget->getBox().topLeft();
	//NormalBackground bg(sf::Color(0, 0, 0, 128), sf::Color::White);
	//if (stats.hotbar.getSize() > 0)
	//{
	//	//hotbarWidget->setDrawable(unique<);
	//	auto selectedText = unique<TextLabel>(hbp + Rect<int>(0, height - textheight, width, textheight), (**stats.hotbar.getSelected()).getName(), TextLabel::HorizontalPolicy::Center);
	//	hotbarWidget->addWidget(std::move(selectedText));
	//	if (stats.hotbar.getSize() > 1)
	//	{
	//		auto upText = unique<TextLabel>(hbp + Rect<int>(0, height + smallgap, width, textheight), (**stats.hotbar.getSelected(1)).getName(), TextLabel::HorizontalPolicy::Center);
	//		upText->applyBackground(bg);
	//		hotbarWidget->addWidget(std::move(upText));
	//	}
	//	if (stats.hotbar.getSize() > 2)
	//	{
	//		auto downText = unique<TextLabel>(hbp + Rect<int>(0, -textheight, width, textheight), (**stats.hotbar.getSelected(-1)).getName(), TextLabel::HorizontalPolicy::Center);
	//		downText->applyBackground(bg);
	//		hotbarWidget->addWidget(std::move(downText));
	//	}
	//}
	//hotbarWidget->applyBackground(bg);
	//hotbar = hotbarWidget.get();
	//addWidget(std::move(hotbarWidget));
}

} // gui
} // sl
