#include "Menus/HomeEditor/HomeEditor.hpp"

#include <algorithm>
#include <iostream>

#include "SFML/Graphics/RectangleShape.hpp"

#include "Engine/Game.hpp"
#include "Engine/Graphics/TextureManager.hpp"
#include "Engine/Graphics/Sprite.hpp"
#include "Engine/Graphics/SFDrawable.hpp"
#include "Engine/GUI/NormalBackground.hpp"
#include "Engine/GUI/DrawableContainer.hpp"
#include "NPC/Region.hpp"
#include "Town/Generation/Interiors/EditingUndo.hpp"
#include "Utility/Assert.hpp"
#include "Utility/JsonUtil.hpp"
#include "Utility/Random.hpp"

namespace sl
{
namespace hed
{

const int editBarWidth = 192;
const int editBarHeight = 256;

HomeEditor::HomeEditor(Game *game)
	:Scene(game, 512 * 5, 384 * 4, game->getWindowWidth(), game->getWindowHeight())
	,state(State::Objects)
	,currentFile("test")
	,viewBox(0, 0, /*getWidth() / 2, getHeight() / 2, */getViewWidth(), getViewHeight())
	,viewMoveCoolDown(0)
	,wallGrid(0, 0, getWidth() / getTileSize(), getHeight() / getTileSize(), getTileSize(), getTileSize(), *this, std::bind(&HomeEditor::getWallType, this, std::placeholders::_1, std::placeholders::_2))
	,tileGrid(0, 0, getWidth() / getTileSize(), getHeight() / getTileSize(), getTileSize(), getTileSize())
	,currentTile("red_carpet")
	,currentWall("test_wall")
	,tileEditSize(1)
	,currentlyPlacingRegion(nullptr)
	,currentlyPlacingRegionShape(nullptr)
{
	editArea.setHandle(unique<gui::GUIWidget>(Rect<int>(0, 0, game->getWindowWidth()/* - editBarWidth*/, game->getWindowHeight())));
	editArea->registerMouseInputEventWithMousePos(Key::GUIClick, gui::GUIWidget::InputEventType::Pressed, std::bind(&HomeEditor::onLeftClick, this, std::placeholders::_1, std::placeholders::_2));
	editArea->registerMouseInputEventWithMousePos(Key::GUIRightClick, gui::GUIWidget::InputEventType::Pressed, std::bind(&HomeEditor::onRightClick, this, std::placeholders::_1, std::placeholders::_2));
	editArea->registerMouseInputEventWithMousePos(Key::GUIClick, gui::GUIWidget::InputEventType::Held, std::bind(&HomeEditor::onLeftHold, this, std::placeholders::_1, std::placeholders::_2));
	editArea->registerMouseInputEventWithMousePos(Key::GUIRightClick, gui::GUIWidget::InputEventType::Held, std::bind(&HomeEditor::onRightHold, this, std::placeholders::_1, std::placeholders::_2));
	editArea.setScene(this);
	editArea.enable();

	objectPreview.setHandle(std::make_unique<gui::DrawableContainer>(Rect<int>(32, 32)));
	objectPreview.setScene(this);
	objectPreview.enable();

	cursor.setHandle(std::make_unique<gui::DrawableContainer>(Rect<int>(16, 16)));
	cursor.setScene(this);
	cursor.enable();

	updateCursorSize();

	auto objectChooserPtr = std::make_unique<gui::WidgetList<towngen::HomeObjectDatabase::Data::value_type>>(
		Rect<int>(0, 0, editBarWidth, editBarHeight),
		[](const towngen::HomeObjectDatabase::Data::value_type& obj) -> std::string
		{
			return obj.first;
		},
		[&](const towngen::HomeObjectDatabase::Data::value_type& obj)
		{
			currentObject = obj.first;
			auto preview = towngen::HomeObjectDatabase::get().getPrototype(currentObject)->createPreview(0.3f);
			objectPreview->setDrawable(std::move(preview));
		}
	);
	objectChooserPtr->setPreview([](const towngen::HomeObjectDatabase::Data::value_type& obj) -> std::unique_ptr<gui::GUIWidget>
	{
		return std::make_unique<gui::DrawableContainer>(Rect<int>(32, 32), obj.second.createPreview());
	});

	objectChooser.setHandle(std::move(objectChooserPtr));
	objectChooser.setScene(this);
	objectChooser.enable();
	objectChooser->applyBackground(gui::NormalBackground(sf::Color::Black, sf::Color::White));

	objectChooser->updateItems(towngen::HomeObjectDatabase::get());




	
	auto tileChooserPtr = std::make_unique<gui::WidgetList<towngen::HomeTileDatabase::Data::value_type>>(
		Rect<int>(0, 0, editBarWidth, editBarHeight),
		[](const towngen::HomeTileDatabase::Data::value_type& obj) -> std::string
		{
			return obj.first;
		},
			[&](const towngen::HomeTileDatabase::Data::value_type& obj)
		{
			currentTile = obj.first;
		}
	);

	tileChooser.setHandle(std::move(tileChooserPtr));
	tileChooser.setScene(this);
	tileChooser->applyBackground(gui::NormalBackground(sf::Color::Black, sf::Color::White));

	tileChooser->updateItems(towngen::HomeTileDatabase::get());
}

const sf::Vector2f HomeEditor::getView() const
{
	return sf::Vector2f(viewBox.left + viewBox.width / 2, viewBox.top + viewBox.height / 2);
}

void HomeEditor::onUpdate(Timestamp timestamp)
{
	updateEntities();
	const int ts = getTileSize();
	if (viewMoveCoolDown > 0)
	{
		--viewMoveCoolDown;
	}
	auto pressedWithCooldown = [this](Input::KeyboardKey key) {
		if (input.isKeyPressed(key))
		{
			viewMoveCoolDown = 32;
			return true;
		}
		if (input.isKeyHeld(key) && viewMoveCoolDown < 16)
		{
			viewMoveCoolDown = 17;
			return true;
		}
		return false;
	};
	if (viewBox.left >= ts && pressedWithCooldown(Input::KeyboardKey::Left))
	{
		viewBox.left -= ts;
	}
	if (viewBox.left < getWidth() - ts && pressedWithCooldown(Input::KeyboardKey::Right))
	{
		viewBox.left += ts;
	}
	if (viewBox.top >= ts && pressedWithCooldown(Input::KeyboardKey::Up))
	{
		viewBox.top -= ts;
	}
	if (viewBox.top < getHeight() - ts && pressedWithCooldown(Input::KeyboardKey::Down))
	{
		viewBox.top += ts;
	}

	if (input.isKeyPressed(Input::KeyboardKey::T))
	{
		state = State::Tiles;
		tileChooser.enable();
		objectChooser.disable();
		objectPreview.disable();
		editArea->setSize(getGame()->getWindowWidth() - editBarWidth, getGame()->getWindowHeight());
		if (currentlyPlacingRegion)
		{
			currentlyPlacingRegion->destroy();
			currentlyPlacingRegion = nullptr;
		}
	}
	else if (input.isKeyPressed(Input::KeyboardKey::O))
	{
		state = State::Objects;
		tileChooser.disable();
		objectChooser.enable();
		objectPreview.enable();
		editArea->setSize(getGame()->getWindowWidth() - editBarWidth, getGame()->getWindowHeight());
		if (currentlyPlacingRegion)
		{
			currentlyPlacingRegion->destroy();
			currentlyPlacingRegion = nullptr;
		}
	}
	else if (input.isKeyPressed(Input::KeyboardKey::W))
	{
		state = State::Walls;
		tileChooser.disable();
		objectChooser.disable();
		objectPreview.disable();
		editArea->setSize(getGame()->getWindowWidth(), getGame()->getWindowHeight());
		if (currentlyPlacingRegion)
		{
			currentlyPlacingRegion->destroy();
			currentlyPlacingRegion = nullptr;
		}
	}
	else if (input.isKeyPressed(Input::KeyboardKey::R))
	{
		if (state == State::Regions || currentRegion.empty())
		{
			std::cout << "Please enter the region name";
			std::cin >> currentRegion;
		}
		state = State::Regions;
		tileChooser.disable();
		objectChooser.disable();
		objectPreview.disable();
		editArea->setSize(getGame()->getWindowWidth(), getGame()->getWindowHeight());
	}
	else if (input.isKeyPressed(Input::KeyboardKey::S))
	{
		save();
	}
	else if (input.isKeyPressed(Input::KeyboardKey::L))
	{
		load();
	}
	else if (input.isKeyPressed(Input::KeyboardKey::C))
	{
		clear();
	}
	else if (input.isKeyPressed(Input::KeyboardKey::F))
	{
		std::cout << "Please enter the new filename: ";
		std::cin >> currentFile;
		std::cout << std::endl << "Now saving to file '" << filename() << "'" << std::endl;
	}
	else if (input.isKeyPressed(Input::KeyboardKey::A))
	{
		if (tileEditSize < 8)
		{
			tileEditSize *= 2;
		}
		else
		{
			tileEditSize = 1;
		}
		updateCursorSize();
	}
	
	
	const Vector2i cursorPos(snapToGrid(getMousePos().x - viewBox.left), snapToGrid(getMousePos().y - viewBox.top));
	switch (state)
	{
	case State::Objects:
		objectChooser->update(getGame()->getWindowWidth() - editBarWidth, 0);
		objectPreview->update(cursorPos.x, cursorPos.y);
		break;
	case State::Tiles:
		tileChooser->update(getGame()->getWindowWidth() - editBarWidth, 0);
		break;
	}
	cursor->update(cursorPos.x, cursorPos.y);
	editArea->update(0, 0);
}

void HomeEditor::onDraw(RenderTarget& target)
{
	target.draw(tileGrid);

	drawEntities(target, viewBox);
}


void HomeEditor::load()
{
	std::string err;
	json11::Json data = readJsonFile(filename(), err);
	if (err.empty())
	{
		clear();
		for (const auto& rootdata : data.object_items())
		{
			if (rootdata.first == "objects")
			{
				for (const json11::Json& object : rootdata.second.array_items())
				{
					const json11::Json::object& vars = object.object_items();
					const int x = vars.at("x").int_value();
					const int y = vars.at("y").int_value();
					const std::string& type = vars.at("type").string_value();
					addObject(x, y, type, vars.count("attributes") ? &vars.at("attributes").object_items() : nullptr);
				}
			}
			else if (rootdata.first == "tiles")
			{
				for (const json11::Json& tile : rootdata.second.array_items())
				{
					const json11::Json::object& vars = tile.object_items();
					const int x = vars.at("x").int_value();
					const int y = vars.at("y").int_value();
					const std::string& id = vars.at("id").string_value();
					addTile(x, y, id);
					wallGrid.remove(x, y);
				}
			}
			else if (rootdata.first == "walls")
			{
				for (const json11::Json& wall : rootdata.second.array_items())
				{
					const json11::Json::object& vars = wall.object_items();
					const int x = vars.at("x").int_value();
					const int y = vars.at("y").int_value();
					const std::string& id = vars.at("id").string_value();
					walls.insert(std::make_pair(Vector2i(x, y), id));
				}
			}
			else if (rootdata.first == "regions")
			{
				for (const json11::Json& region : rootdata.second.array_items())
				{
					const json11::Json::object& vars = region.object_items();
					const Rect<int> rect = parseRect(vars);
					const std::string& type = vars.at("type").string_value();
					addRegion(createRegion(rect, type));
				}
			}
			else if (rootdata.first == "width")
			{
				// do not manually edit these (at least not now)
			}
			else if (rootdata.first == "height")
			{
				// do not manually edit these (at least not now)
			}
			else if (rootdata.first == "xoff")
			{
				// do not manually edit these (at least not now)
			}
			else if (rootdata.first == "yoff")
			{
				// do not manually edit these (at least not now)
			}
			else if (rootdata.first == "walkableWidth" || rootdata.first == "walkableHeight")
			{
				// do not manually edit these
			}
			else
			{
				ERROR("undefined data");
			}
		}
	}
	else
	{
		ERROR(err);
	}
}

void HomeEditor::save() const
{
	json11::Json::object root;

	const Rect<int> bounds = computeBoundingBoxInTileSpace();
	// we want to ensure that there is at least a certain amount on all sides of the house so the view doesn't fuck up
	const int minHorizGap = 20;
	const int minVertGap = 16;
	// so shift everything by (xOffset, yOffset)
	const int xOffset = minHorizGap - bounds.left;
	const int yOffset = minVertGap - bounds.top;
	// and compute the width/height accordingly
	const int widthInTiles = minHorizGap * 2 + bounds.width;
	const int heightInTiles = minVertGap * 2 + bounds.height;

	//@ todo have custom offsets instead of computing them automatically (use HomeEditor::homeOrigin)
	root["width"] = widthInTiles;
	root["height"] = heightInTiles;
	root["xoff"] = minHorizGap;
	root["yoff"] = minVertGap;
	root["walkableWidth"] = bounds.width;
	root["walkableHeight"] = bounds.height;

	json11::Json::array objectsOut;
	for (const towngen::EditingUndo *object : objects)
	{
		json11::Json::object objectOut;
		objectOut["x"] = object->getInstance().x + xOffset * getTileSize();
		objectOut["y"] = object->getInstance().y + yOffset * getTileSize();
		objectOut["type"] = object->getInstance().type;
		auto att = instanceAttributes.find(object);
		if (att != instanceAttributes.end())
		{
			objectOut["attributes"] = att->second;
		}
		objectsOut.push_back(std::move(objectOut));
	}
	root["objects"] = std::move(objectsOut);

	json11::Json::array tilesOut;
	for (const auto& tile : tiles)
	{
		json11::Json::object tileOut;
		tileOut["x"] = tile.first.x + xOffset;
		tileOut["y"] = tile.first.y + yOffset;
		tileOut["id"] = tile.second;
		tilesOut.push_back(std::move(tileOut));
	}
	root["tiles"] = std::move(tilesOut);

	json11::Json::array wallsOut;
	for (const auto& wall : walls)
	{
		const int x = wall.first.x;
		const int y = wall.first.y;
		if (wallGrid.isSolid(x, y))
		{
			const bool wallRight = x < getWidth() / getTileSize() - 1 && wallGrid.isSolid(x + 1, y);
			const bool wallLeft = x > 0 && wallGrid.isSolid(x - 1, y);
			const bool wallUp = y > 0 && wallGrid.isSolid(x, y - 1);
			const bool wallDown = y < getHeight() / getTileSize() - 1 && wallGrid.isSolid(x, y + 1);
			if (!wallRight || !wallLeft || !wallUp || !wallDown)
			{
				json11::Json::object wallOut;
				wallOut["x"] = x + xOffset;
				wallOut["y"] = y + yOffset;
				wallOut["id"] = wall.second;
				wallsOut.push_back(std::move(wallOut));
			}
		}
	}
	root["walls"] = std::move(wallsOut);

	json11::Json::array regionsOut;
	for (const Region *region : regions)
	{
		auto regionOut = writeRect(region->getMask().getBoundingBox());
		regionOut["type"] = region->getRegionType();
		regionsOut.push_back(std::move(regionOut));
	}
	root["regions"] = std::move(regionsOut);
	
	std::string err = saveJsonFile(filename(), json11::Json(std::move(root)));

	if (!err.empty())
	{
		ERROR(err);
	}
}

void HomeEditor::clear()
{
	// undo() removes objects from the set, which invalidates iterators, hence why we increment first
	for (auto it = objects.begin(), end = objects.end(); it != end; )
	{
		towngen::EditingUndo *undo = *it;
		++it;
		undo->undo();
	}
	ASSERT(objects.empty());
	tiles.clear();
	walls.clear();
	tileGrid.resetTextures();
	const int tw = getWidth() / getTileSize();
	const int th = getHeight() / getTileSize();
	for (int i = 0; i < tw; ++i)
	{
		for (int j = 0; j < th; ++j)
		{
			wallGrid.fill(i, j);
		}
	}
	for (Region *region : regions)
	{
		region->destroy();
	}
	regions.clear();
	if (currentlyPlacingRegion)
	{
		currentlyPlacingRegion->destroy();
		currentlyPlacingRegion = nullptr;
	}
}

const std::string HomeEditor::filename() const
{
	return "data/gendata/homes/" + currentFile + ".json";
}

void HomeEditor::onLeftClick(int x, int y)
{
	const Vector2i pos = screenToLevel(x, y);
	switch (state)
	{
	case State::Tiles:
		break;
	case State::Objects:
		if (!currentObject.empty())
		{
			addObject(snapToGrid(pos.x), snapToGrid(pos.y), currentObject);
		}
		break;
	case State::Walls:
		break;
	case State::Regions:
		{
			regionSelectedTopLeft.x = snapToGrid(pos.x);
			regionSelectedTopLeft.y = snapToGrid(pos.y);
			currentlyPlacingRegion = createRegion(computeRegionRect(pos.x, pos.y), currentRegion, &currentlyPlacingRegionShape).release();
			addEntityToList(currentlyPlacingRegion);
		}
		break;
	}
}

void HomeEditor::onRightClick(int x, int y)
{
	const Vector2i pos = screenToLevel(x, y);
	switch (state)
	{
	case State::Tiles:
		break;
	case State::Objects:
		break;
	case State::Walls:
		break;
	case State::Regions:
	{
		Region *regionUnderMouse = static_cast<Region*>(checkCollision(Rect<int>(pos.x, pos.y, 1, 1), "Region"));
		if (regionUnderMouse)
		{
			regions.erase(regionUnderMouse);
			regionUnderMouse->destroy();
		}
	}
		break;
	}
}

void HomeEditor::onLeftHold(int x, int y)
{
	const Vector2i pos = screenToLevel(x, y);
	const int tx = pos.x / getTileSize();
	const int ty = pos.y / getTileSize();
	switch (state)
	{
	case State::Tiles:
		if (!currentTile.empty())
		{
			for (int dx = 0; dx < tileEditSize; ++dx)
			{
				for (int dy = 0; dy < tileEditSize; ++dy)
				{
					addTile(tx + dx, ty + dy, currentTile);
				}
			}
		}
		break;
	case State::Objects:
		break;
	case State::Walls:
		if (!wallGrid.isSolid(tx, ty))
		{
			for (int dx = 0; dx < tileEditSize; ++dx)
			{
				for (int dy = 0; dy < tileEditSize; ++dy)
				{
					addWall(tx + dx, ty + dy, currentWall);
				}
			}
		}
		break;
	case State::Regions:
		ASSERT(currentlyPlacingRegion && currentlyPlacingRegionShape);
		const Rect<int> newRect = computeRegionRect(pos.x, pos.y);
		currentlyPlacingRegion->setRegionBounds(newRect);
		currentlyPlacingRegionShape->setSize(sf::Vector2f(newRect.width, newRect.height));
		break;
	}
}

void HomeEditor::onRightHold(int x, int y)
{
	const Vector2i pos = screenToLevel(x, y);
	switch (state)
	{
	case State::Tiles:
		{
			for (int dx = 0; dx < tileEditSize; ++dx)
			{
				for (int dy = 0; dy < tileEditSize; ++dy)
				{
					const Vector2i tpos(pos.x / getTileSize() + dx, pos.y / getTileSize() + dy);
					auto it = tiles.find(tpos);
					if (it != tiles.end())
					{
						tileGrid.setTexture(tpos.x, tpos.y, tileTextures.begin()->second.get(), sf::IntRect(0, 0, 0, 0));
						tiles.erase(it);
					}
				}
			}
		}
		break;
	case State::Objects:
		{
			std::list<Entity*> toDelete;
			const int deleteRadius = 4;
			cull(toDelete, Rect<int>(pos.x - deleteRadius, pos.y - deleteRadius, 2 * deleteRadius, 2 * deleteRadius), "EditingUndo");
			for (Entity *entity : toDelete)
			{
				static_cast<towngen::EditingUndo&>(*entity).undo();
			}
		}
		break;
	case State::Walls:
		{
			for (int dx = 0; dx < tileEditSize; ++dx)
			{
				for (int dy = 0; dy < tileEditSize; ++dy)
				{
					const int tx = pos.x / getTileSize() + dx;
					const int ty = pos.y / getTileSize() + dy;
					if (wallGrid.isSolid(tx, ty))
					{
						wallGrid.remove(tx, ty);
						if (tx > 0 && wallGrid.isSolid(tx - 1, ty))
						{
							addWall(tx - 1, ty, currentWall);
						}
						if (tx < getWidth() / getTileSize() - 1 && wallGrid.isSolid(tx + 1, ty))
						{
							addWall(tx + 1, ty, currentWall);
						}
						if (ty > 0 && wallGrid.isSolid(tx, ty - 1))
						{
							addWall(tx, ty - 1, currentWall);
						}
						if (tx < getHeight() / getTileSize() - 1 && wallGrid.isSolid(tx, ty + 1))
						{
							addWall(tx, ty + 1, currentWall);
						}
						addTile(tx, ty, currentTile);
					}
				}
			}
		}
		break;
	case State::Regions:
		break;
	}
}


void HomeEditor::addObject(int x, int y, const std::string& type, const json11::Json::object *instanceAttributes)
{
	const towngen::HomeObjectPrototype *prototype = towngen::HomeObjectDatabase::get().getPrototype(type);
	ASSERT(prototype);
	
	auto undo = prototype->generate(Random::get(), *this, Vector2i(x, y), nullptr, nullptr, nullptr, instanceAttributes, nullptr);
	objects.insert(undo.get());
	if (instanceAttributes)
	{
		this->instanceAttributes[undo.get()] = *instanceAttributes;
	}
	towngen::EditingUndo *undoPtr = undo.get();
	towngen::EditingUndo *instanceAttributeKey = instanceAttributes ? undoPtr : nullptr;
	undo->addUndoFunc([&, undoPtr, instanceAttributeKey]
	{
		objects.erase(undoPtr);
		if (instanceAttributeKey)
		{
			objects.erase(instanceAttributeKey);
		}
	});
	addEntityToQuad(undo.release());
}

void HomeEditor::addTile(int tx, int ty, const std::string& id)
{
	const towngen::HomeTileDatabase::Prototype *prototype = towngen::HomeTileDatabase::get().getPrototype(id);
	ASSERT(prototype);
	if (prototype)
	{
		const std::string& fname = prototype->getFilename();
		auto it = tileTextures.find(fname);
		if (it == tileTextures.end())
		{
			it = tileTextures.insert(std::make_pair(fname, TextureManager::getTexture(fname))).first;
		}
		tileGrid.setTexture(tx, ty, it->second.get(), prototype->getTextureRect());
		tiles[Vector2i(tx, ty)] = id;
	}
}

void HomeEditor::addWall(int tx, int ty, const std::string& id)
{
	const Vector2i tPos(tx, ty);
	walls[tPos] = id;
	wallGrid.fill(tx, ty);
	auto it = tiles.find(tPos);
	if (it != tiles.end())
	{
		tiles.erase(it);
	}
	const int txe = std::min(tx + 1, getWidth() / getTileSize());
	const int tye = std::min(ty + 1, getHeight() / getTileSize());
	for (int i = std::max(0, tx - 1); i < txe; ++i)
	{
		for (int j = std::max(0, ty - 1); j < tye; ++j)
		{
			if (i != 0 || j != 0)
			{
				removeWallIfRedundant(i, j);
			}
		}
	}
}

void HomeEditor::addRegion(std::unique_ptr<Region> region)
{
	if (region->getScene() != this)
	{
		addEntityToQuad(region.get());
	}
	regions.insert(region.release());
}

std::unique_ptr<Region> HomeEditor::createRegion(const Rect<int> area, const std::string& type, sf::RectangleShape **drawableCapture) const
{
	std::unique_ptr<Region> region = unique<Region>(area, type);
	sf::RectangleShape rect;
	rect.setPosition(area.left, area.top);
	rect.setSize(sf::Vector2f(area.width, area.height));
	const sf::Color color = getRandomColor();
	rect.setOutlineColor(color);
	rect.setOutlineThickness(1.f);
	rect.setFillColor(sf::Color(color.r, color.g, color.b, 32));
	auto drawable = unique<SFDrawable<sf::RectangleShape>>(std::move(rect));
	if (drawableCapture)
	{
		*drawableCapture = &drawable->getData();
	}
	region->addDrawable(std::move(drawable));
	return region;
}

void HomeEditor::removeWallIfRedundant(int tx, int ty)
{
	auto it = walls.find(Vector2i(tx, ty));
	if (it != walls.end())
	{
		const bool wallLeft = tx == 0 || walls.count(Vector2i(tx - 1, ty)) > 0;
		const bool wallRight = tx == getWidth() / getTileSize() || walls.count(Vector2i(tx + 1, ty)) > 0;
		const bool wallUp = ty == 0 || walls.count(Vector2i(tx, ty - 1)) > 0;
		const bool wallDown = ty == getHeight() / getTileSize() || walls.count(Vector2i(tx, ty + 1)) > 0;
		if (wallLeft && wallRight && wallUp && wallDown)
		{
			walls.erase(it);
		}
	}
}

int HomeEditor::snapToGrid(int x) const
{
	return getTileSize() * (x / getTileSize());
}

const std::string& HomeEditor::getWallType(int x, int y) const
{
	static std::string defaultWall = "?";
	auto it = walls.find(Vector2i(x, y));
	if (it != walls.end())
	{
		return it->second;
	}
	return defaultWall;
}

Rect<int> HomeEditor::computeBoundingBoxInTileSpace() const
{
	auto it = tiles.cbegin();
	int left = it->first.x;
	int right = it->first.x;
	int top = it->first.y;
	int bottom = it->first.y;
	++it;
	for ( ; it != tiles.cend(); ++it)
	{
		left = std::min(left, it->first.x);
		right = std::max(right, it->first.x);
		top = std::min(top, it->first.y);
		bottom = std::max(bottom, it->first.y);
	}
	return Rect<int>(left, top, right - left + 1, bottom - top + 1);
}

Vector2i HomeEditor::screenToLevel(int x, int y) const
{
	return Vector2i(viewBox.left + x, viewBox.top + y);
}

void HomeEditor::updateCursorSize()
{
	auto cursorRect = unique<SFDrawable<sf::RectangleShape>>(sf::RectangleShape(sf::Vector2f(tileEditSize * getTileSize(), tileEditSize * getTileSize())));
	cursorRect->getData().setFillColor(sf::Color::Transparent);
	cursorRect->getData().setOutlineColor(sf::Color::White);
	cursorRect->getData().setOutlineThickness(1.f);
	cursor->setDrawable(std::move(cursorRect));
}

Rect<int> HomeEditor::computeRegionRect(int x, int y) const
{
	const int x1 = regionSelectedTopLeft.x;
	const int y1 = regionSelectedTopLeft.y;
	const int x2 = snapToGrid(x);
	const int y2 = snapToGrid(y);
	return Rect<int>(std::min(x1, x2), std::min(y1, y2), std::abs(x1 - x2) + getTileSize(), std::abs(y1 - y2) + getTileSize());
}

sf::Color HomeEditor::getRandomColor() const
{
	return Random::get().choose({ sf::Color::Red, sf::Color::Green, sf::Color::Blue, sf::Color::Yellow, sf::Color::Magenta, sf::Color::Cyan });
}

} // hed
} // sl