#ifndef SL_HED_MAINMENU_HPP
#define SL_HED_MAINMENU_HPP

#include <set>

#include "Engine/Graphics/Texture.hpp"
#include "Engine/Graphics/TileGrid.hpp"
#include "Engine/GUI/WidgetHandle.hpp"
#include "Engine/GUI/WidgetList.hpp"
#include "Engine/Scene.hpp"
#include "Home/WallGrid.hpp"
#include "Town/Generation/Interiors/HomeObjectDatabase.hpp"
#include "Town/Generation/Interiors/HomeTileDatabase.hpp"
#include "Utility/Grid.hpp"

namespace sl
{

namespace towngen
{
class EditingUndo;
} // towngen

namespace gui
{
class DrawableContainer;
} // gui

class Region;

namespace hed
{

class HomeEditor : public Scene
{
public:
	HomeEditor(Game *game);

private:
	const sf::Vector2f getView() const override;

	void onUpdate(Timestamp timestamp) override;

	void onDraw(RenderTarget& target) override;




	void load();

	void save() const;

	void clear();

	const std::string filename() const;

	void onLeftClick(int x, int y);

	void onRightClick(int x, int y);

	void onLeftHold(int x, int y);

	void onRightHold(int x, int y);

	void addObject(int x, int y, const std::string& type, const json11::Json::object *instanceAttributes = nullptr);

	void addTile(int tx, int ty, const std::string& id);

	void addWall(int tx, int ty, const std::string& id);

	void addRegion(std::unique_ptr<Region> region);

	/**
	 * @param drawableCapture [output] If present, a pointer somewhere to store the address of the created region's drawable
	 */
	std::unique_ptr<Region> createRegion(const Rect<int> area, const std::string& type, sf::RectangleShape **drawableCapture = nullptr) const;

	void removeWallIfRedundant(int tx, int ty);
	 
	int snapToGrid(int x) const;

	const std::string& getWallType(int x, int y) const;

	Rect<int> computeBoundingBoxInTileSpace() const;

	Vector2i screenToLevel(int x, int y) const;

	void updateCursorSize();

	Rect<int> computeRegionRect(int x, int y) const;

	sf::Color getRandomColor() const;

	




	enum class State
	{
		Tiles,
		Objects,
		Walls,
		Regions
	};
	/// Data to keep
	State state;
	std::string currentFile;
	Rect<int> viewBox;
	int viewMoveCoolDown;


	/// ----------------------- Data used for exporting -----------------------
	//! A collection of Undo objects corresponding to all placed objects for the purpose of undoing them
	std::set<towngen::EditingUndo*> objects;
	//! Attributes that correspond to things in the objects member. Stuff is automatically removed from here by it too via undo()
	std::map<const towngen::EditingUndo*, json11::Json::object> instanceAttributes;
	//! All tiles placed indexed by their position (in tiles)
	std::map<Vector2i, std::string> tiles;
	//! All walls placed indexed by their position (in tiles)
	std::map<Vector2i, std::string> walls;
	//! The tile in the home that corresponds to the top left corner of the building exterior
	Vector2i homeOrigin;
	//! All regions
	std::set<Region*> regions;
	

	/// ------------- Data used for graphically showing the level -------------
	//! The physical representation of all walls
	WallGrid wallGrid;
	//! The graphical representation of the tiles
	TileGrid tileGrid;
	//! A cache of all textures used for tiles to make sure they don't get freed
	std::map<std::string, Texture> tileTextures;


	/// -------------- Data used for the GUI portion of the editor ---------------
	//! The currently selected object's type id
	std::string currentObject;
	//! The currently selected tile's type id
	std::string currentTile;
	//! The currently selected wall's type id
	std::string currentWall;
	//! A widget covering the entire screen that exists just to capture mouse actions
	gui::WidgetHandle<> editArea;
	//! The GUI list showing all objects to choose from
	gui::WidgetHandle<gui::WidgetList<towngen::HomeObjectDatabase::Data::value_type>> objectChooser;
	//! The GUI list showing all tiles to choose from
	gui::WidgetHandle<gui::WidgetList<towngen::HomeTileDatabase::Data::value_type>> tileChooser;
	//! The cursor which shows a transparent preview of what you're about to place
	gui::WidgetHandle<gui::DrawableContainer> objectPreview;
	//! Rectangular cursor to show where we're placing things
	gui::WidgetHandle<gui::DrawableContainer> cursor;
	//! How many squares across to use for tile/wall editing (in a tileEditSize by tileEditSize square)
	int tileEditSize;
	Vector2i regionSelectedTopLeft;
	std::string currentRegion;
	Region *currentlyPlacingRegion;
	//! Read-only access to the drawable of currentlyPlacingRegion
	sf::RectangleShape *currentlyPlacingRegionShape;
	
};

} // hed
} // sl

#endif // SL_HED_MAINMENU_HPP