#include "Menus/InputMenu.hpp"

#include "Engine/GUI/BasicMenuBuilder.hpp"
#include "Engine/GUI/NormalBackground.hpp"
#include "Engine/GUI/WidgetList.hpp"
#include "Engine/Input/Key.hpp"
#include "Engine/Game.hpp"
#include "Menus/PauseScene.hpp"

#include <iterator>
#include <string>
#include <utility>
#include <vector>

namespace sl
{
namespace gui
{

std::unique_ptr<GUIWidget> makeKeybindMenu(ControllerBinds& controls, const std::function<void()>& onBack)
{
	class BindInputCapture : public GUIWidget
	{
	public:
		BindInputCapture(Key key, ControllerBinds& controls, WidgetList<std::pair<Key, std::string>> *rootList, GUIWidget *bindList)
			:GUIWidget(Rect<int>(0, 0))
			,key(key)
			,controls(controls)
			,rootList(rootList)
			,bindList(bindList)
		{
			constexpr int x = 32;
			constexpr int w = 512 - 2 * x;
			constexpr int h = 160;
			constexpr int y = (384 - h) / 2;
			auto tl = std::make_unique<TextLabel>(Rect<int>(x, y, w, h), /*"Press the key/button/etc to bind"*/"Press key to bind", TextLabel::Config().centered().fontSize(3).pad(32, 32));
			tl->applyBackground(NormalBackground(sf::Color::Black, sf::Color::White));
			addWidget(std::move(tl));
		}
	private:
		void onUpdate(int x, int y) override
		{
			auto justPressed = scene->getInputHandle().getJustPressed(0);
			for (const KeyBind& bind : justPressed)
			{
				controls.addKeybind(key, bind);
			}
			if (!justPressed.empty())
			{
				scene->destroy();
				bindList->destroy();
				rootList->refresh();
			}
		}

		Key key;
		ControllerBinds& controls;
		WidgetList<std::pair<Key, std::string>> *rootList;
		GUIWidget *bindList;
	};

	static const std::pair<Key, std::string> binds[] = {
		{ Key::MoveLeft, "Move left" },
		{ Key::MoveRight, "Move right" },
		{ Key::MoveUp, "Move up" },
		{ Key::MoveDown, "Move down" },
		{ Key::Sprint, "Sprint" },
		{ Key::Interact, "Interact" },
		{ Key::UseItem, "Use current item" },
		{ Key::CycleCallout, "Cycle interact" },
		{ Key::ActionChooser, "Open action menu" },
		{ Key::ExitCar, "Exit (vehicle)" },
		{ Key::Brake, "Brake (vehicle)" },
		{ Key::Accelerate, "Accelerate (vehicle)" },
		{ Key::Reverse, "Reverse (vehicle)" },
		{ Key::SteerLeft, "Turn left (Vehicle)" },
		{ Key::SteerRight, "Turn right (Vehicle)" },
		{ Key::Cruise, "Toggle cruise control (vehicle)" },
		{ Key::Pause, "Pause" },
		{ Key::OpenInventory, "Inventory" },
		{ Key::ToggleHud, "Toggle HUD" },
		{ Key::ToggleMap, "Toggle map" },
		{ Key::GUIUp, "Up (Menus)" },
		{ Key::GUIDown, "Down (Menus)" },
		{ Key::GUILeft, "Left (Menus)" },
		{ Key::GUIRight, "Right (Menus)" },
		{ Key::GUIAccept, "Accept (Menus)" },
		{ Key::GUIBack, "Back (Menus)" }
	};
	static constexpr int viewHeight = 384;
	static constexpr int listWidth = 256;
	static constexpr int wgap = 32;
	static constexpr int hgap = 16;
	static constexpr int backHeight = 24;
	auto base = std::make_unique<GUIWidget>(Rect<int>(512, 384));
	std::unique_ptr<WidgetList<std::pair<Key, std::string>>> list = std::make_unique<WidgetList<std::pair<Key, std::string>>>(
		Rect<int>(wgap, hgap, listWidth, viewHeight - 3*hgap - backHeight),
		[&controls](const std::pair<Key, std::string>& opt) {
			return opt.second + ": " + controls.getBindsString(opt.first, false);
		}
	);
	// this is separate so we can capture the list
	list->setOnSelect([&controls, rootList = list.get()](const std::pair<Key, std::string>& opt) {
		const Key key = opt.first;
		enum class BindOptions
		{
			Add,
			Clear
		};
		static constexpr BindOptions bindOptions[2] = { BindOptions::Add, BindOptions::Clear };
		std::unique_ptr<WidgetList<BindOptions>> bindList = std::make_unique<WidgetList<BindOptions>>(
			Rect<int>(wgap + listWidth, hgap, 212, 48),
			[](BindOptions bopt) {
				return bopt == BindOptions::Add ? "Add" : "Clear";
			}
		);
		// this is separate so we can capture the list
		bindList->setOnSelect([thisWidget = bindList.get(), key, &controls, rootList](BindOptions bopt) {
			switch (bopt)
			{
			case BindOptions::Add:
				rootList->getScene()->getGame()->addScene(unique<PauseScene>(rootList->getScene(), std::make_unique<BindInputCapture>(key, controls, rootList, thisWidget)));
				break;
			case BindOptions::Clear:
				controls.clearKeybinds(key);
				thisWidget->destroy();
				rootList->refresh();
				break;
			}
		});
		bindList->registerInputEvent(Key::GUIBack, GUIWidget::InputEventType::Released, [thisWidget = bindList.get()] {
			thisWidget->destroy();
		});
		bindList->setButtonHeight(24);
		bindList->updateItems(bindOptions);
		rootList->addWidget(std::move(bindList));
	});
	list->setButtonHeight(24);
	list->updateItems(binds);
	list->registerInputEvent(Key::GUIBack, GUIWidget::InputEventType::Released, onBack);
	list->applyBackground(NormalBackground(sf::Color::Black, sf::Color::White));
	base->addWidget(std::move(list));
	base->addWidget(Button::createGenericButton(Rect<int>(wgap, viewHeight - hgap - backHeight, listWidth, backHeight), "Back", onBack, 2));
	return base;
}

static std::string codepoint(int i)
{
	return std::string(1, static_cast<char>(i));
}

std::unique_ptr<GUIWidget> makeGamepadConfigurationMenu(const std::function<void()>& onBack)
{
	static const std::vector<std::pair<KeyBind, std::string>> xboxConfig = {
		{ XBOX360::A, codepoint(136) },
		{ XBOX360::B, codepoint(139) },
		{ XBOX360::X, codepoint(137) },
		{ XBOX360::Y, codepoint(138) },
		{ XBOX360::LeftBumper, "LB" },
		{ XBOX360::RightBumper, "RB" },
		{ XBOX360::Back, "Back" },
		{ XBOX360::Start, "Start" },
		{ XBOX360::LeftTrigger, "LT" },
		{ XBOX360::RightTrigger, "RT" },
		{ XBOX360::LeftStickPressed, "LS In" },
		{ XBOX360::RightStickPressed, "RS In" },
		{ XBOX360::LeftStickLeft, "LS Left" },
		{ XBOX360::LeftStickRight, "LS Right" },
		{ XBOX360::LeftStickDown, "LS Down" },
		{ XBOX360::LeftStickUp, "LS Up" },
		{ XBOX360::RightStickLeft, "RS Left" },
		{ XBOX360::RightStickRight, "RS Right" },
		{ XBOX360::RightStickDown, "RS Down" },
		{ XBOX360::RightStickUp, "RS Up" },
		{ XBOX360::DPadUp, codepoint(130) },
		{ XBOX360::DPadDown, codepoint(131) },
		{ XBOX360::DPadLeft, codepoint(128) },
		{ XBOX360::DPadRight, codepoint(129) }
	};
	static const std::vector<std::pair<KeyBind, std::string>> playstationConfig = {
		{ XBOX360::A, codepoint(135) },
		{ XBOX360::B, codepoint(134) },
		{ XBOX360::X, codepoint(132) },
		{ XBOX360::Y, codepoint(133) },
		{ XBOX360::LeftBumper, "L1" },
		{ XBOX360::RightBumper, "R1" },
		{ XBOX360::Back, "Select" },
		{ XBOX360::Start, "Start" },
		{ XBOX360::LeftTrigger, "L2" },
		{ XBOX360::RightTrigger, "R2" },
		{ XBOX360::LeftStickPressed, "LS In" },
		{ XBOX360::RightStickPressed, "RS In" },
		{ XBOX360::LeftStickLeft, "LS Left" },
		{ XBOX360::LeftStickRight, "LS Right" },
		{ XBOX360::LeftStickDown, "LS Down" },
		{ XBOX360::LeftStickUp, "LS Up" },
		{ XBOX360::RightStickLeft, "RS Left" },
		{ XBOX360::RightStickRight, "RS Right" },
		{ XBOX360::RightStickDown, "RS Down" },
		{ XBOX360::RightStickUp, "RS Up" },
		{ XBOX360::DPadUp, codepoint(130) },
		{ XBOX360::DPadDown, codepoint(131) },
		{ XBOX360::DPadLeft, codepoint(128) },
		{ XBOX360::DPadRight, codepoint(129) }
	};

	static auto runConfigurationSequence = [](GUIWidget *rootWidget, const std::vector<std::pair<KeyBind, std::string>>& config) {
		class BindListener : public GUIWidget
		{
		public:
			BindListener(const std::vector<std::pair<KeyBind, std::string>>& config)
				:GUIWidget(Rect<int>(512, 384))
				,config(&config)
				,it(config.cbegin())
				,alertTimer(0)
			{
				addWidget(Button::createGenericButton(Rect<int>(32, 384 - 64 - 32 - 16, 512 - 64, 32), "Skip", [this] {
					ASSERT(it != this->config->cend());
					++it;
					refresh();
				}, 2));
				addWidget(Button::createGenericButton(Rect<int>(32, 384 - 64, 512 - 64, 32), "Cancel", [this] {
					scene->destroy();
				}, 2));
				auto alertWidget = std::make_unique<TextLabel>(Rect<int>(32, 200, 512 - 64, 32), TextLabel::Config().centered().fontSize(2));
				alertWidget->applyBackground(NormalBackground(sf::Color::Black, sf::Color::White));
				alert = alertWidget.get();
				addWidget(std::move(alertWidget));
				refresh();
			}
		private:
			void onUpdate(int x, int y) override
			{
				if (it == config->cend())
				{
					Input::saveGamepadCustomNames(translations);
					scene->destroy();
					return;
				}
				auto justPressed = scene->getInputHandle().getJustPressed(0);
				for (const KeyBind& bind : justPressed)
				{
					if (bind.type() == KeyBind::DeviceType::GamepadAxis || bind.type() == KeyBind::DeviceType::GamepadButton) {
						const auto inserted = translations.insert(std::make_pair(bind, it->second));
						if (inserted.second)
						{
							++it;
							refresh();
							return;
						}
						else
						{
							alertTimer = 45;
							alert->setString(it->second + " already assigned to " + inserted.first->second);
							return;
						}
					}
				}
				if (alertTimer > 0 && --alertTimer == 0)
				{
					refresh();
				}
			}

			void refresh()
			{
				if (it != config->end())
				{
					alert->setString("Press " + it->second);
				}
			}

			const std::vector<std::pair<KeyBind, std::string>> *config;
			std::vector<std::pair<KeyBind, std::string>>::const_iterator it;
			std::map<KeyBind, std::string> translations;
			GUIWidget *rootWidget;
			TextLabel *alert;
			int alertTimer;
		};

		rootWidget->getScene()->getGame()->addScene(unique<PauseScene>(rootWidget->getScene(), unique<BindListener>(config)));
	};
	// wrap it in a base widget just so we can capture it to get the scene... gross
	auto base = std::make_unique<GUIWidget>(Rect<int>(512, 384));
	gui::BasicMenuBuilder builder(320, 384, 16, 32);
	builder.addButton("XBOX", std::bind(runConfigurationSequence, base.get(), std::ref(xboxConfig)));
	builder.addButton("Playstation", std::bind(runConfigurationSequence, base.get(), std::ref(playstationConfig)));
	builder.addBackButton(onBack);
	auto menu = builder.create();
	menu->move((512 - 320) / 2, (384 - menu->getBox().height) / 2);
	menu->update();
	base->addWidget(std::move(menu));
	return base;
}

} // gui
} // sl