#include "Menus/Pause/InventoryMenu.hpp"

#include <functional>
#include <sstream>

#include "Engine/Game.hpp"
#include "Town/World.hpp"
#include "Engine/GUI/NormalBackground.hpp"
#include "Engine/GUI/WidgetList.hpp"
#include "Player/PlayerStats.hpp"
#include "Utility/Assert.hpp"
#include "Utility/Memory.hpp"
#include "Utility/ToString.hpp"

namespace sl
{

InventoryMenu::InventoryMenu(Game *game, World *world, PlayerStats& stats)
	:AbstractScene(game, 512, 384, -2)
	,world(world)
	,stats(stats)
	,subMenu(SubMenu::Items)
{
	widget.setScene(this);

	changeSubMenu(subMenu);
}


/*			private			*/
const sf::Vector2f InventoryMenu::getView() const
{
	return sf::Vector2f(getViewWidth() / 2, getViewHeight() / 2);
}

void InventoryMenu::onUpdate(int64_t timestamp)
{
	if (input.isKeyPressed(Input::KeyboardKey::Escape))
	{
		close();
	}

	widget->update();

	deleteNextStep.reset(nullptr);
}

void InventoryMenu::onDraw(RenderTarget& target)
{
}

void InventoryMenu::changeSubMenu(SubMenu subMenu)
{
	this->subMenu = subMenu;
	widget.disable();
	deleteNextStep = widget.take();
	switch (subMenu)
	{
	case SubMenu::Items:
		{
			const int xOffset = 64;
			const int yOffset = 48;
			const int itemListWidth = 160;
			const int headerHeight = 32;
			
			gui::NormalBackground blackWhiteBG(sf::Color::Black, sf::Color::White);

			auto menu = unique<gui::GUIWidget>(Rect<int>(xOffset, yOffset, getViewWidth() - 2 * xOffset, getViewHeight() - 2 * yOffset));
			menu->applyBackground(gui::NormalBackground(sf::Color(14, 14, 14), sf::Color::White));


			// --------- item info box ---------
			const int itemBoxHGap = 24;
			const int itemBoxVGap = 24;
			const int itemGap = 24;
			const int itemBoxX = itemListWidth + xOffset + itemBoxHGap;
			const int itemBoxWidth = menu->getBox().width - (itemBoxX - xOffset) - itemBoxHGap;
			const int itemBoxY = yOffset + headerHeight;
			//const int itemBoxHeight = menu->getBox().height - (itemBoxY - yOffset) - itemBoxVGap;
			int ySoFar = itemBoxY;
			const int nameBoxHeight = 32;
			auto nameBox = std::make_unique<gui::TextLabel>(Rect<int>(itemBoxX, ySoFar, itemBoxWidth, nameBoxHeight),
				gui::TextLabel::Config().valign(gui::TextLabel::Middle).pad(32, 0));
			gui::TextLabel *nameBoxPtr = nameBox.get();
			nameBox->applyBackground(blackWhiteBG);
			menu->addWidget(std::move(nameBox));
			ySoFar += nameBoxHeight + itemGap;

			/*const int previewBoxHeight = 96;
			auto previewBox = unique<gui::GUIWidget>(Rect<int>(itemBoxX, ySoFar, itemBoxWidth, previewBoxHeight));
			previewBox->applyBackground(blackWhiteBG);
			menu->addWidget(std::move(previewBox));
			ySoFar += previewBoxHeight + itemGap;*/

			const int descBoxHeight = 160;
			auto descBox = std::make_unique<gui::TextLabel>(Rect<int>(itemBoxX, ySoFar, itemBoxWidth, descBoxHeight),
				gui::TextLabel::Config().pad(16, 16));
			gui::TextLabel *descBoxPtr = descBox.get();
			descBox->applyBackground(blackWhiteBG);
			menu->addWidget(std::move(descBox));
			ySoFar += nameBoxHeight + itemGap;


			// ---------   item list   ---------
			std::vector<Inventory::ItemRef> items;
			stats.soul->items.queryItems(items);
			typedef std::map<Inventory::ItemRef, int, Inventory::ItemComp> ItemsByType;
			ItemsByType itemsByType;
			for (const Inventory::ItemRef& item : items)
			{
				++itemsByType[item];
			}
			const int itemListY = menu->getBox().top + headerHeight;
			const Rect<int> itemListBox(menu->getBox().left, itemListY, itemListWidth, menu->getBox().height - headerHeight);
			auto itemList = std::make_unique<gui::WidgetList<ItemsByType::value_type>>(
				itemListBox,
				[](const std::pair<Inventory::ItemRef, int>& item) -> std::string
				{
					return item.first->getName() + " x " + std::to_string(item.second);
				},
				[nameBoxPtr, descBoxPtr](const std::pair<Inventory::ItemRef, int>& item)
				{
					nameBoxPtr->setString(item.first->getName());
					descBoxPtr->setString(item.first->getDesc());
					// @todo update preview later
				}
			);
			itemList->applyBackground(blackWhiteBG);
			itemList->updateItems(itemsByType);
			menu->addWidget(std::move(itemList));


			widget.setHandle(std::move(menu));
				
			const int horizBorder = (getViewWidth() - widget->getBox().width) / 2;
			const int vertBorder = (getViewHeight() - widget->getBox().height) / 2;
				
			widget->move(horizBorder, vertBorder);
			widget->update();
		}
		break;
	default:
		ERROR("undefined state transition");
		break;
	}
	widget.enable();
}

void InventoryMenu::close()
{
	world->start();
	getGame()->removeScene(this);
}

} // sl
