#include "NPC/AnimationDatabase.hpp"

#include "Utility/Assert.hpp"
#include "Utility/JsonUtil.hpp"
#include "Utility/Mapping.hpp"

#include <functional>
#include <memory>

namespace sl
{

const AnimationDatabase::Entry* AnimationDatabase::get(const std::string& id) const
{
	auto it = entries.find(id);
	return it != entries.end() ? &it->second : nullptr;
}

const std::map<std::string, AnimationDatabase::Entry>& AnimationDatabase::getEntries() const
{
	return entries;
}

/*static*/ const AnimationDatabase& AnimationDatabase::get()
{
	static const AnimationDatabase db;
	return db;
}

// private:
AnimationDatabase::AnimationDatabase()
{
	std::string err;
	const json11::Json data = readJsonFile("data/animations.json", err);
	if (!err.empty())
	{
		ERROR(err);
		return;
	}
	for (const std::pair<std::string, json11::Json>& animDef : data.object_items())
	{
		auto r = entries.insert(std::make_pair(animDef.first, Entry()));
		ASSERT(r.second);
		Entry& entry = r.first->second;
		for (const json11::Json& layerDef : animDef.second.array_items())
		{
			const auto& obj = layerDef.object_items();
			std::unique_ptr<Framerate> framerate;
			auto it = obj.find("fixed");
			if (it != obj.end())
			{
				framerate = std::make_unique<Framerate>(it->second.int_value());
			}
			it = obj.find("varied");
			if (it != obj.end())
			{
				ASSERT(framerate == nullptr);
				framerate = std::make_unique<Framerate>(
					mapVec(
						it->second.array_items(),
						std::mem_fn(&json11::Json::int_value)));
			}
			it = obj.find("mixed");
			if (it != obj.end())
			{
				ASSERT(framerate == nullptr);
				framerate = std::make_unique<Framerate>(
					mapVec(
						it->second.array_items(),
						[](const json11::Json& js) -> Framerate::Frame {
							ASSERT(js.array_items().size() == 2);
							return {
								js.array_items()[0].int_value(),
								js.array_items()[1].int_value()
							};
						}));
			}
			ASSERT(framerate != nullptr);
			it = obj.find("layers");
			if (it != obj.end())
			{
				for (const json11::Json& layerJson : it->second.array_items())
				{
					const NPCSpriteLayer layer = strToNPCSpriteLayer(layerJson.string_value());
					auto r = entry.overrides.insert(std::make_pair(layer, *framerate));
					ASSERT(r.second);
				}
			}
			else
			{
				entry.base = *framerate;
			}
		}
	}
}


} // sl
