#include "NPC/Behaviors/AcceptTalkWithPlayer.hpp"

#include "Player/Player.hpp"
#include "Player/States/PlayerTalkState.hpp"
#include "Police/CrimeUtil.hpp"
#include "NPC/Behaviors/Dialogue/DialogueTransitions.hpp"
#include "NPC/Behaviors/BehaviorContext.hpp"
#include "NPC/Behaviors/WalkTo.hpp"
#include "NPC/Locators/EntityLocator.hpp"
#include "NPC/NPC.hpp"
#include "Utility/Assert.hpp"

namespace sl
{
namespace ai
{

static const int defaultTimesToWait = 4;

AcceptTalkWithPlayer::AcceptTalkWithPlayer(std::string startNode, Type type)
	:BehaviorTreeNode("AcceptTalkWithPlayer")
	,talking(false)
	,startNode(std::move(startNode))
	,action()
	,type(type)
{
}

BehaviorTreeNode::Result AcceptTalkWithPlayer::execute(BehaviorContext& context)
{
	if (talking)
	{
		action = getDialogueAction(context);
		if (action)
		{
			talking = false;
		}
		// both if it ended or not should still result in Running. It can execute child the next tick if it wants to.
		return Result::Running;
	}
	else if (action)
	{
#ifdef SL_DEBUG
		const Result actionResult = action->debugExecute(context);
#else
		const Result actionResult = action->execute(context);
#endif // SL_DEBUG
		if (actionResult != Result::Running)
		{
			// we don't care if it failed. We're treating "child is done executing" as SUCCESS,
			// as we already have "can't path to player" as FAILURE in the InitiateTalkWithPlayer case.
			// we can always change this later.
			abort(context);
			return Result::Success;
		}
		return Result::Running;
	}
	else // node just starting
	{
		Event *ev;
		while (ev = context.pollEvents())
		{
			if (ev->getCategory() == Event::Category::NPCInteraction)
			{
				if (ev->getName() == "talk")
				{
					std::string startNodeID;
					switch (type)
					{
					case Type::Absolute:
						startNodeID = startNode;
						break;
					case Type::ContextLookup:
						startNodeID = context.getVariable(startNode).asString();
						break;
					}
					ev->respond(startNodeID);
					NPC& npc = context.getNPC();
					Player *player = npc.getScene()->getPlayer();
					ASSERT(player);
					player->getState().changeState(unique<PlayerTalkState>(*player, context.getNPC(), startNodeID));
					npc.face(player->getPos());
					talking = true;
					return Result::Running;
				}
			}
		}
	}
	abort(context);
	return Result::Failure;
}

void AcceptTalkWithPlayer::abort(BehaviorContext& context)
{
	talking = false;
	if (action)
	{
		action->abort(context);
		action = nullptr;
	}
}

} // ai
} // sl