#include "NPC/Behaviors/BehaviorContext.hpp"

#include "Engine/Scene.hpp"
#include "Town/World.hpp"
#include "NPC/NPC.hpp"
#include "Utility/Memory.hpp"

namespace sl
{
namespace ai
{
BehaviorContext::BehaviorContext(NPC& npc)
	:npc(&npc)
	,cachedEvents()
	,eventsIt(cachedEvents.end())
{
}

void BehaviorContext::setNPCTarget(const Target& target)
{
	getNPC().setTarget(target);
}

NPCMoveResult BehaviorContext::npcMoveTowardsTarget()
{
	return getNPC().moveTowardsTarget();
}

const GameTime& BehaviorContext::getGameTime() const
{
	return npc->getScene()->getWorld()->getGameTime();
}

NPC& BehaviorContext::getNPC()
{
	return *npc;
}

Soul& BehaviorContext::getSoul()
{
	return *getNPC().getSoul();
}

Scene* BehaviorContext::getScene()
{
	return getNPC().getScene();
}

World* BehaviorContext::getWorld()
{
	return getNPC().getWorld();
}

Player* BehaviorContext::getPlayer()
{
	return getNPC().getWorld()->getPlayer();
}

Ref<NPC, true> BehaviorContext::lockNPC()
{
	return npc.lock();
}

void BehaviorContext::pushVariable(const std::string& var, ls::Value val)
{
	variables[var].push_back(std::move(val));
}

void BehaviorContext::popVariable(const std::string& var)
{
	auto it = variables.find(var);
	ASSERT(it != variables.end());
	ASSERT(!it->second.empty());
	it->second.pop_back();
}

void BehaviorContext::pushPODStore(const std::string& var, std::vector<uint8_t>& location)
{
	locations[var].push_back(&location);
}

void BehaviorContext::popPODStore(const std::string& var)
{
	auto it = locations.find(var);
	ASSERT(it != locations.end());
	it->second.pop_back();
}

const ls::Value& BehaviorContext::getVariable(const std::string& var)
{
	auto it = variables.find(var);
	ASSERT(it != variables.end());
	return it->second.back();
}

void BehaviorContext::setVariable(const std::string& var, ls::Value val)
{
	auto it = variables.find(var);
	ASSERT(it != variables.end());
	ASSERT(!it->second.empty());
	it->second.back() = std::move(val);
}

Random& BehaviorContext::rng()
{
	return Random::get();
}

void BehaviorContext::addEvent(Event ev)
{
	cachedEvents.push_back(std::move(ev));
	eventsIt = cachedEvents.begin();
}

void BehaviorContext::registerSceneEvent(Event ev)
{
	getScene()->registerLocalEvent(unique<Event>(std::move(ev)));
}

void BehaviorContext::registerSceneEvent(std::string name)
{
	getScene()->registerLocalEvent(unique<Event>(Event::Category::NPCInteraction, std::move(name)));
}

void BehaviorContext::clearEvents()
{
	cachedEvents.clear();
	eventsIt = cachedEvents.end();
}

Event* BehaviorContext::pollEvents()
{
	if (eventsIt != cachedEvents.end())
	{
		return &*(eventsIt++);
	}
	eventsIt = cachedEvents.begin();
	return nullptr;
}

} // ai
} // sl
