#ifndef SL_AI_BEHAVIORCONTEXT_HPP
#define SL_AI_BEHAVIORCONTEXT_HPP

#include <cstdint>
#include <cstring>
#include <map>
#include <string>
#include <vector>

#include "Engine/Events/EventQueue.hpp"
#include "Engine/Ref.hpp"
#include "Utility/Assert.hpp"
#include "Utility/DynamicStore.hpp"
#include "Utility/Random.hpp"
#include "Utility/TypeTraits.hpp"

namespace sl
{
class GameTime;
class NPC;
class Player;
class Scene;
class Soul;
class Target;
class World;

enum class NPCMoveResult : char;
} // sl

namespace sl
{
namespace ai
{

class BehaviorContext
{
public:
	BehaviorContext(NPC& npc);


	void setNPCTarget(const Target& target);

	NPCMoveResult npcMoveTowardsTarget();

	const GameTime& getGameTime() const;

	NPC& getNPC();
	
	Soul& getSoul();

	Scene* getScene();

	World* getWorld();

	Player* getPlayer();

	Ref<NPC, true> lockNPC();

	// call from MemoryNode::execute()
	void pushVariable(const std::string& var, ls::Value val);
	// call from MemoryNode::execute()
	void popVariable(const std::string& var);
	// call from MemoryNode::execute()
	void pushPODStore(const std::string& var, std::vector<uint8_t>& location);
	// call from MemoryNode::execute()
	void popPODStore(const std::string& var);

	const ls::Value& getVariable(const std::string& var);

	void setVariable(const std::string& var, ls::Value val);

	/**
	 * Simply a wrapper around DynamicStore in a ls::Value
	 */
	template <typename T>
	const T* getData(const std::string& var);

	/**
	 * Simply a wrapper around DynamicStore in a ls::Value
	 */
	template <typename T>
	void setData(const std::string& var, const T& data);

	template <typename T>
	void readPOD(const std::string& var, T& value) const;

	template <typename T>
	void writePOD(const std::string& var, const T& value);

	template <typename T>
	T getPOD(const std::string& var) const;

	Random& rng();

	/**
	 * Adds an event from an already-spawned place (ie call from Scene, etc)
	 */
	void addEvent(Event ev);

	/**
	 * Create an event and spawn it into the world
	 */
	void registerSceneEvent(Event ev);

	void registerSceneEvent(std::string name);

	void clearEvents();

	Event* pollEvents();


private:
	Ref<NPC> npc;
	//! Raw data storage for POD storage, as a stack
	std::map<std::string, std::vector<std::vector<uint8_t>*>> locations;
	//! Stored variables, as a stack
	std::map<std::string, std::vector<ls::Value>> variables;
	std::vector<Event> cachedEvents;
	std::vector<Event>::iterator eventsIt;
};

template <typename T>
const T* BehaviorContext::getData(const std::string& var)
{
	return getVariable(var).asData().as<T>();
}

template <typename T>
void BehaviorContext::setData(const std::string& var, const T& data)
{
	setVariable(var, ls::Value(DynamicStore::make<T>(data)));
}

template <typename T>
void BehaviorContext::readPOD(const std::string& var, T& value) const
{
	static_assert(IsMemcopyable<T>::value, IsMemcopyable_Error);

	auto it = locations.find(var);
	ASSERT(it != locations.end());
	const std::vector<uint8_t> *buffer = it->second.back();
	ASSERT(sizeof(T) == buffer->size());
	std::memcpy(reinterpret_cast<uint8_t*>(&value), buffer->data(), buffer->size());
}

template <typename T>
void BehaviorContext::writePOD(const std::string& var, const T& value)
{
	static_assert(IsMemcopyable<T>::value, IsMemcopyable_Error);

	auto it = locations.find(var);
	ASSERT(it != locations.end());
	std::vector<uint8_t> *buffer = it->second.back();
	buffer->resize(sizeof(T));
	std::memcpy(buffer->data(), reinterpret_cast<const uint8_t*>(&value), sizeof(T));
}

template <typename T>
T BehaviorContext::getPOD(const std::string& var) const
{
	T t;
	readPOD(var, t);
	return t;
}

} // ai
} // sl

#endif // SL_AI_BEHAVIORCONTEXT_HPP