#include "NPC/Behaviors/Dialogue/DialogueTransitions.hpp"

#include "NPC/Behaviors/AcceptTalkWithPlayer.hpp"
#include "NPC/Behaviors/Counter.hpp"
#include "NPC/Behaviors/CustomCondition.hpp"
#include "NPC/Behaviors/FixedResult.hpp"
#include "NPC/Behaviors/MemoryNode.hpp"
#include "NPC/Behaviors/Not.hpp"
#include "NPC/Behaviors/Parallel.hpp"
#include "NPC/Behaviors/Selector.hpp"
#include "NPC/Behaviors/WalkTo.hpp"
#include "NPC/Behaviors/While.hpp"
#include "NPC/Behaviors/Util.hpp"
#include "NPC/Locators/EntityLocator.hpp"
#include "NPC/States/DraggedState.hpp"
#include "NPC/NPC.hpp"
#include "Player/Player.hpp"
#include "Player/States/PlayerTalkState.hpp"
#include "Police/CrimeUtil.hpp"

#include <functional>
#include <map>
#include <string>

namespace sl
{
namespace ai
{

// -1 is no boredom
static std::unique_ptr<BehaviorTreeNode> doInteruptableByPlayer(std::unique_ptr<BehaviorTreeNode> action, std::vector<std::string> boredomMessages, int boredomMessageTimer, int boredom = -1)
{
	std::unique_ptr<BehaviorTreeNode> conversationFinished = unique<AcceptTalkWithPlayer>("resumeNode", AcceptTalkWithPlayer::Type::ContextLookup);
	if (boredom != -1)
	{
		conversationFinished = makebt<Selector>(
			std::move(conversationFinished),
			unique<CustomCondition>([](BehaviorContext& context) {
			//std::cout << "boredom = " << context.getVariable("boredom").asInt() << std::endl;
			if (context.getVariable("boredom").asInt() > 45 * 30)
			{
				context.getNPC().makeTextBubble("I'm bored, I'm gonna go do something else, bye!");
				return true; // break the while loop
			}
			return false; // keep doing the while loop, following
		}));
	}
	auto active = unique<Parallel>();
	active->addChild(std::move(action));
	active->addIgnoredChild(unique<CustomCondition>(
	[msgs = std::move(boredomMessages), boredomMessageTimer](BehaviorContext& context) {
		if ((context.getVariable("boredom").asInt() + 1) % boredomMessageTimer == 0)
		{
			// TODO: parameterize for when this won't make sense? (ie waiting)
			context.getNPC().makeTextBubble(context.rng().choose(msgs));
		}
		return true;
	}), true);
	// The reason this is is a while loop, is because AcceptTalkWithPlayer will return SUCCESS
	// (and thus the condition FAILURE) only when dialogue has been established and it has ended completely.
	// that way she will keep following you until you respond to her.
	return unique<Counter>(While::makeActive(std::move(unique<Not>(std::move(conversationFinished))), std::move(active)), "boredom");
}

// @TODO move?
// TODO: parameterize based on her personality/relationship to you/emotions
static std::unique_ptr<BehaviorTreeNode> followPlayer()
{
	return doInteruptableByPlayer(
		unique<WalkTo>(unique<EntityLocator>("Player")),
		{
			"Hurry up, I'm getting bored...",
			"When are we gonna get there?",
			"Where are we going?"
		},
		10 * 30,
		45 * 30);
}

static std::unique_ptr<BehaviorTreeNode> wait()
{
	return doInteruptableByPlayer(
		unique<FixedResult>(BehaviorTreeNode::Result::Success),
		{
			"How long do I have to wait here?",
			"I'm betting bored...",
			"I want ice cream!",
			"What are we gonna do after?",
			"ICE CREAM! ICE CREAM! ICE CREAM!",
			"Why do I have to wait? Let's have fun already!",
			"Waiting is boring..."
		},
		23 * 30,
		128 * 30);
}

std::unique_ptr<BehaviorTreeNode> getDialogueAction(BehaviorContext& context)
{
	static const char *dst_prefix = "dialogue_subtree_";
	static const int dst_prefix_len = strlen(dst_prefix);
	// <subtreename, <creationfunction, [paramname1, parammname2, ...]>
	// where the param names are the varialbe names associated with extra action() parameters
#define DLGSUBTREE(name, f, ...) std::make_pair(name, std::make_pair(f, std::vector<std::string>({ __VA_ARGS__ })))
	static std::map<std::string, std::pair<std::function<std::unique_ptr<BehaviorTreeNode>()>, std::vector<std::string>>> subtrees = {
		DLGSUBTREE("follow_to_secluded", followPlayer, "resumeNode"),
		DLGSUBTREE("follow_me", followPlayer, "resumeNode"),
		DLGSUBTREE("wait_here", wait)
	};
	Event *ev;
	while (ev = context.pollEvents())
	{
		if (ev->getCategory() == Event::Category::NPCInteraction)
		{
			if (ev->getName() == "changeState_kidnapped")
			{
				NPC& npc = context.getNPC();
				Player *player = npc.getScene()->getPlayer();
				ASSERT(player);
				const int crimeId = commitCrime(npc, Crime::Type::Kidnapping);
				npc.getSoul()->setCrimeReported(crimeId, false);
				// TODO: non-chloroform kidnapping
				npc.getSoul()->getStatus().chloroformTimer = 30 * 60; // 1 minute in-game
				npc.changeState(unique<DraggedState>(npc, *player));
				return unique<FixedResult>(BehaviorTreeNode::Result::Success);
			}
			else if (ev->getName() == "changeState_idle" || ev->getName() == "changeState_shopping")
			{
				// TODO: combine these into one changeState_finish event?
				return unique<FixedResult>(BehaviorTreeNode::Result::Success);
			}
			else if (strncmp(ev->getName().c_str(), dst_prefix, dst_prefix_len) == 0)
			{
				const char *treeName = ev->getName().c_str() + dst_prefix_len;
				auto it = subtrees.find(treeName);
				if (it != subtrees.end())
				{
					auto subtree = unique<MemoryNode>(it->second.first());
					const int paramCount = ev->getFlag("paramCount").asInt();
					ASSERT(paramCount == it->second.second.size());
					for (int i = 0; i < paramCount; ++i)
					{
						// additional parameters to action() are encoded as "0", "1", etc...
						//      (see PlayerTalkState for local:action() definition)
						subtree->addDeclaration(MemoryNode::var(it->second.second[i], ev->getFlag(std::to_string(i))));
					}
					return std::move(subtree);
				}
				else
				{
					ERROR(std::string("could not find action: ") + treeName);
				}
			}
		}
	}
	return nullptr;
}


} // ai
} // sl
