#include "NPC/Behaviors/FollowSchedule.hpp"

#include "Engine/Scene.hpp"
#include "NPC/Behaviors/BehaviorContext.hpp"
#include "NPC/Behaviors/WalkTo.hpp"
#include "NPC/Locators/PathfindingTargetLocator.hpp"
#include "NPC/NPC.hpp"
#include "Pathfinding/NodePredicates/InCachedPath.hpp"
#include "Pathfinding/Path.hpp"
#include "Utility/FlatMap.hpp"

namespace sl
{

// TODO: move somewhere common?
static std::unique_ptr<TargetLocator> pathfindingTargetLocatorToCached(const std::shared_ptr<const pf::Path>& cachedPath, const pf::WeightFunc& heuristic, const std::string& weightMap)
{
	return std::make_unique<PathfindingTargetLocator>(
		pf::InCachedPath(*cachedPath),
		heuristic,
		weightMap,
		[cachedPath](std::deque<pf::Node>& path) {
		const auto& lookup = cachedPath->getLookup();
		auto it = lookup.find(path.back().getID());
		// since we should only be here on success
		ASSERT(it != lookup.end());
		//std::cout << "Extended path by postProcess() from " << path.size();
		path.insert(path.end(), it->second, cachedPath->getNodes().cend());
		//std::cout << " to " << path.size() << std::endl;
	});
}

namespace ai
{

FollowSchedule::FollowSchedule()
	:BehaviorTreeNode("Schedule")
	,toAbort(children.end())
{
}

BehaviorTreeNode::Result FollowSchedule::execute(BehaviorContext& context)
{
	// DEMO HACK REMOVE?
	if (context.getSoul().scheduleValid != 69)
	{
		return Result::Success;
	}
	const auto s = schedule(context);
	ASSERT(!s.startId.empty() && !s.endId.empty());
	if (s.inBetween())
	{
		//const std::string id = s.inBetween() ? s.startId : s.startId + "_to_" + s.endId;
		// TODO: how to handle interuptable vs non-interuptable stuff?
		if (toAbort != children.end() && toAbort->first != s.startId)
		{
			abort(context);
		}
		// only if we just started the activity, and only off-screen/far away, should we teleport to it.
		if (toAbort == children.end() || toAbort->first != s.startId)
		{
			teleportIfOffsceenAndFar(context, s);
		}
		toAbort = children.find(s.startId);
		ASSERT(toAbort != children.end());
		const Result ret =
#ifdef SL_DEBUG
		toAbort->second->debugExecute(context);
#else // SL_DEBUG
		toAbort->second->execute(context);
#endif // SL_DEBUG
		if (ret != Result::Running)
		{
			toAbort = children.end();
		}
		// TOOD: figure out how to handle this - should it repeat activities? should it be configurable?
		if (ret == Result::Success)
		{
			return Result::Running;
		}
		return ret;
	}
	else
	{
		if (!teleportIfOffsceenAndFar(context, s))
		{
			// follow the path
			if (!walkTo)
			{
				// TODO: give areas regions not just points, then use the Or predicate to optionally enter other ways
				const std::shared_ptr<const pf::Path> cachedPath = s.path(context.getNPC(), *context.getWorld());
				walkTo = std::make_unique<WalkTo>(pathfindingTargetLocatorToCached(cachedPath, pf::noHeuristic, "npc_routine"));
			}
			auto walkRet = walkTo->execute(context);
			if (walkRet == Result::Success)
			{
				// TODO: make them start the activity early? (could be possible if they were somehow moved faster)
				return Result::Running;
			}
			return walkRet;
		}
		return Result::Success;
	}
}

void FollowSchedule::abort(BehaviorContext& context)
{
	if (toAbort != children.end())
	{
		toAbort->second->abort(context);
		toAbort = children.end();
	}
	if (walkTo)
	{
		walkTo->abort(context);
		walkTo.reset();
	}
}

void FollowSchedule::addChild(const std::string& id, std::unique_ptr<BehaviorTreeNode> behavior)
{
#ifdef SL_DEBUG
	behavior->setParent(this);
#endif // SL_DEBUG
	children.insert(std::make_pair(id, std::move(behavior)));

}

/*static*/float FollowSchedule::getPriority(BehaviorContext& context)
{
	return schedule(context).priority(context.getSoul());
}

/*static*/bool FollowSchedule::teleportIfOffsceenAndFar(BehaviorContext& context, const Schedule::InterpolatedSchedule::Interpolation& interpolation)
{
	// TODO: re-enable and debug why it causes weird pathing
	return false;
	// teleport if you're off screen AND 24 worldspace (ie ~tiles in overworld) off target
	const ScenePos teleportTo = interpolation.location(context.getNPC(), *context.getWorld());
	const bool teleport =
		!context.getScene()->getViewRect().containsPoint(context.getNPC().getPos()) &&
		worldDistance(teleportTo, context.getNPC()) > 24;
	if (teleport)
	{
		// just go to where you should be, without any checks, etc.
		if (context.getScene() != teleportTo.getScene())
		{
			teleportTo.getScene()->transferEntity(&context.getNPC());
		}
		context.getNPC().setPos(teleportTo.getPos());
		context.getNPC().face(teleportTo.getPos());
	}
	return teleport;
}


// private
/*static*/Schedule::InterpolatedSchedule::Interpolation FollowSchedule::schedule(BehaviorContext& context)
{
	return context.getNPC().getSoul()->getInterpolatedSchedule(context.getGameTime()).interpolate(context.getGameTime());
}

} // ai
} // sl