#include "NPC/Behaviors/IfElse.hpp"

#include "NPC/Behaviors/AlwaysSucceed.hpp"
#include "NPC/Behaviors/Sequence.hpp"
#include "Utility/Assert.hpp"
#include "Utility/Memory.hpp"

namespace sl
{
namespace ai
{

IfElse::IfElse(std::unique_ptr<BehaviorTreeNode> condition, std::unique_ptr<BehaviorTreeNode> onSuccess, std::unique_ptr<BehaviorTreeNode> onFail, bool isActive)
	:BehaviorTreeNode("IfElse")
	,condition(std::move(condition))
	,onSuccess(std::move(onSuccess))
	,onFail(std::move(onFail))
	,conditionResult(Result::Running) // conditionResult == Running -> we need to run condition
	,isActive(isActive)
{
#ifdef SL_DEBUG
	this->condition->setParent(this);
	this->onSuccess->setParent(this);
	this->onFail->setParent(this);
#endif // SL_DEBUG
}

BehaviorTreeNode::Result IfElse::execute(BehaviorContext& context)
{
	// conditionResult != Running --> onSuccess or onFail still running
	if (conditionResult == Result::Running || isActive)
	{
#ifdef SL_DEBUG
		conditionResult = condition->debugExecute(context);
#else // SL_DEBUG
		conditionResult = condition->execute(context);
#endif // SL_DEBUG
	}

	if (conditionResult == Result::Success)
	{
#ifdef SL_DEBUG
		const Result result = onSuccess->debugExecute(context);
#else // SL_DEBUG
		const Result result = onSuccess->execute(context);
#endif // SL_DEBUG
		if (result != Result::Running)
		{
			conditionResult = Result::Running;
			return result;
		}
	}
	else if (conditionResult == Result::Failure)
	{
#ifdef SL_DEBUG
		const Result result = onFail->debugExecute(context);
#else // SL_DEBUG
		const Result result = onFail->execute(context);
#endif // SL_DEBUG
		if (result != Result::Running)
		{
			conditionResult = Result::Running;
			return result;
		}
	}
	return Result::Running;
}

void IfElse::abort(BehaviorContext& context)
{
	switch (conditionResult)
	{
	case Result::Running:
		condition->abort(context);
		break;
	case Result::Success:
		onSuccess->abort(context);
		break;
	case Result::Failure:
		onFail->abort(context);
	}
	conditionResult = Result::Running;
}


} // ai
} // sl