#include "NPC/Behaviors/InitiateTalkWithPlayer.hpp"

#include "Player/Player.hpp"
#include "Player/States/PlayerTalkState.hpp"
#include "NPC/Behaviors/Dialogue/DialogueTransitions.hpp"
#include "NPC/Behaviors/BehaviorContext.hpp"
#include "NPC/Behaviors/WalkTo.hpp"
#include "NPC/Locators/EntityLocator.hpp"
#include "NPC/NPC.hpp"
#include "Utility/Assert.hpp"

namespace sl
{
namespace ai
{

static const int defaultTimesToWait = 4;

InitiateTalkWithPlayer::InitiateTalkWithPlayer(std::string startNode)
	:BehaviorTreeNode("InitiateTalkWithPlayer")
	,startNode(std::move(startNode))
	,walkResult(Result::Running)
	,waitTimer(0)
	,timesToWait(defaultTimesToWait)
	,walkToPlayer(unique<WalkTo>(unique<EntityLocator>("Player")))
	,talking(false)
	,action()
{
}

BehaviorTreeNode::Result InitiateTalkWithPlayer::execute(BehaviorContext& context)
{
	if (talking)
	{
		action = getDialogueAction(context);
		if (action)
		{
			talking = false;
		}
		// both if it ended or not should still result in Running. It can execute child the next tick if it wants to.
		return Result::Running;
	}
	else if (action)
	{
#ifdef SL_DEBUG
		const Result actionResult = action->debugExecute(context);
#else
		const Result actionResult = action->execute(context);
#endif // SL_DEBUG
		if (actionResult != Result::Running)
		{
			// we don't care if it failed. We're treating "child is done executing" as SUCCESS,
			// as we already have "can't path to player" as FAILURE in the InitiateTalkWithPlayer case.
			// we can always change this later.
			abort(context);
			return Result::Success;
		}
		return Result::Running;
	}
	else // node just starting, or still running walkToPlayer
	{
		if (walkResult == Result::Running)
		{
#ifdef SL_DEBUG
			walkResult = walkToPlayer->debugExecute(context);
#else // SL_DEBUG
			walkResult = walkToPlayer->execute(context);
#endif // SL_DEBUG
		}
		if (walkResult == Result::Success)
		{
			context.getNPC().setDialogueStartOverride(startNode);
			if (timesToWait > 0)
			{
				if (--waitTimer <= 0)
				{
					if (timesToWait >= 4)
					{
						context.getNPC().makeTextBubble("...");
					}
					else if (timesToWait == 3)
					{
						context.getNPC().makeTextBubble("Hey, Mister...");
					}
					else if (timesToWait == 2)
					{
						context.getNPC().makeTextBubble("Hey, talk to me! Stop ignoring me!");
					}
					waitTimer = 93;//326;
					--timesToWait;
				}
				walkResult = Result::Running;
				return Result::Running;
			}
			// he's ignored her enough... force it!
			Player *player = context.getNPC().getScene()->getPlayer();
			ASSERT(player);
			player->getState().changeState(unique<PlayerTalkState>(*player, context.getNPC(), startNode));
			timesToWait = defaultTimesToWait;
			waitTimer = 0;
			walkResult = Result::Running; // generic reset state
			talking = true;
			// fall through to Running return
		}
		else if (walkResult == Result::Failure)
		{
			abort(context);
			return Result::Failure;
		}
		// fall through to Running return
	}
	return Result::Running;
}

void InitiateTalkWithPlayer::abort(BehaviorContext& context)
{
	walkResult = Result::Running;
	walkToPlayer->abort(context);
	if (action)
	{
		action->abort(context);
		action = nullptr;
	}
	talking = false;
	waitTimer = 0;
	timesToWait = defaultTimesToWait;
}

} // ai
} // sl