#ifndef SL_AI_JOINACTIVITY_HPP
#define SL_AI_JOINACTIVITY_HPP

#include <functional>
#include <memory>
#include <set>
#include <string>

#include "NPC/Behaviors/BehaviorTreeNode.hpp"
#include "NPC/Behaviors/MemoryNode.hpp"
#include "NPC/GroupAI/GroupAI.hpp"
#include "Engine/Entity.hpp"
#include "Engine/Scene.hpp"

namespace sl
{
namespace ai
{

template <typename Activity>
class JoinActivity : public BehaviorTreeNode
{
public:
	JoinActivity(std::unique_ptr<BehaviorTreeNode> activity);

	Result execute(BehaviorContext& context) override;

	void abort(BehaviorContext& context) override;

private:
	std::function<std::string()> message;
	enum class State
	{
		HaventAsked,
		Asked,
		Joined
	};
	State state;
	std::vector<std::pair<NPC*, int>> toAsk;
	std::set<NPC*> aboutToRespond;
	int yesResponses;
	int noResponses;
	std::unique_ptr<BehaviorTreeNode> activity;
	typename GroupAI<Activity>::Membership membership;
	Activity *activityToJoin;
};

template <typename Activity>
JoinActivity<Activity>::JoinActivity(std::unique_ptr<BehaviorTreeNode> activity)
	:BehaviorTreeNode("JoinActivity")
	,state(State::HaventAsked)
	,toAsk()
	,yesResponses(0)
	,noResponses(0)
	,activity(std::move(activity))
	,activityToJoin(nullptr)
{
#ifdef SL_DEBUG
	this->activity->setParent(this);
#endif // SL_DEBUG
}

template <typename Activity>
BehaviorTreeNode::Result JoinActivity<Activity>::execute(BehaviorContext& context)
{
	switch (state)
	{
	case State::HaventAsked:
		{
			std::vector<Activity*> activitiesNearby;
			context.getNPC().getScene()->getActivities(context.getNPC(), activitiesNearby);
			if (activitiesNearby.empty())
			{
				context.getNPC().makeTextBubble("Anyone want to play tag?");
				auto act = unique<Activity>();
				activityToJoin = act.get();
				context.getNPC().getScene()->registerActivity(std::move(act));
			}
			else
			{
				// @TODO try multiple activities
				activityToJoin = context.rng().choose(activitiesNearby);
				context.getNPC().makeTextBubble("Can I join?");
				for (NPC *npcToAsk : activityToJoin->getParticipants())
				{
					toAsk.push_back(std::make_pair(npcToAsk, context.rng().randomInclusive(34, 69)));
				}
			}
			context.pushVariable("activity", ls::Value(activityToJoin));
			state = State::Asked;
			break;
		}
	case State::Asked:
		{
			for (int i = 0; i < toAsk.size(); ++i)
			{
				// ask them for their response
				if (--toAsk[i].second <= 0)
				{
					NPC *npc = toAsk[i].first;
					toAsk[i] = std::move(toAsk.back());
					toAsk.pop_back();
					auto res = aboutToRespond.insert(npc);
					ASSERT(res.second);
					auto respondIt = res.first;
					npc->handleEvent(Event(Event::Category::NPCInteraction, std::string(TypeToString<Activity>()) + "_join",
						[&, npc, respondIt](const std::string& response)
					{
						if (response == "yes")
						{
							npc->makeTextBubble("Sure, you can join!");
							++yesResponses;
						}
						else
						{
							ASSERT(response == "no");
							npc->makeTextBubble("No! I don't like you!");
							++noResponses;
						}
						aboutToRespond.erase(respondIt);
					}));
				}
			}
			if (toAsk.empty() && aboutToRespond.empty())
			{
				if (yesResponses >= noResponses)
				{
					state = State::Joined;
					membership = activityToJoin->addParticipant(context.getNPC());
				}
				else
				{
					yesResponses = 0;
					noResponses = 0;
					return Result::Failure;
				}
			}
		}
		break;
	case State::Joined:
#ifdef SL_DEBUG
		Result res = activity->debugExecute(context);
#else // SL_DEBUG
		Result res = activity->execute(context);
#endif // SL_DEBUG
		if (res != Result::Running)
		{
			context.popVariable("activity");
			state = State::HaventAsked;
			membership.leave();
		}
		return res;
	}
	return Result::Running;
}

template <typename Activity>
void JoinActivity<Activity>::abort(BehaviorContext& context)
{
	toAsk.clear();
	yesResponses = 0;
	noResponses = 0;
	membership.leave();
	activityToJoin = nullptr;
	if (state == State::Joined)
	{
		activity->abort(context);
	}
	if (state == State::Asked || state == State::Joined)
	{
		context.popVariable("activity");
	}
	state = State::HaventAsked;
}

} // ai
} // sl

#endif // SL_AI_JOINACTIVITY_HPP
