#include "NPC/Behaviors/MemoryNode.hpp"

#include "NPC/Behaviors/BehaviorContext.hpp"

namespace sl
{
namespace ai
{

MemoryNode::MemoryNode(std::unique_ptr<BehaviorTreeNode> node)
	:BehaviorTreeNode("MemoryNode")
	,node(std::move(node))
	,initialized(false)
{
#ifdef SL_DEBUG
	this->node->setParent(this);
#endif // SL_DEBUG
}

BehaviorTreeNode::Result MemoryNode::execute(BehaviorContext& context)
{
	if (!initialized)
	{
		for (const auto& var : vars)
		{
			context.pushVariable(var.first, var.second);
		}
		for (auto& pod : POD)
		{
			context.pushPODStore(pod.first, pod.second);
		}
		initialized = true;
	}
#ifdef SL_DEBUG
	BehaviorTreeNode::Result ret = node->debugExecute(context);
#else // SL_DEBUG
	BehaviorTreeNode::Result ret = node->execute(context);
#endif // SL_DEBUG
	if (ret != BehaviorTreeNode::Result::Running)
	{
		abort(context);
	}
	return ret;
}

void MemoryNode::abort(BehaviorContext& context)
{
	if (initialized)
	{
		initialized = false;
		for (const auto& var : vars)
		{
			context.popVariable(var.first);
		}
		for (const auto& pod : POD)
		{
			context.popPODStore(pod.first);
		}
		node->abort(context);
	}
}

void MemoryNode::addDeclaration(MemoryDeclaration decl)
{
	switch (decl.type)
	{
	case MemoryDeclaration::Type::Var:
		vars.push_back(std::make_pair(decl.name, std::move(decl.var)));
		break;
	case MemoryDeclaration::Type::POD:
		POD.push_back(std::make_pair(decl.name, std::move(decl.data)));
		break;
	}
}

/*static*/ MemoryNode::MemoryDeclaration MemoryNode::var(const std::string& name, ls::Value defaultValue)
{
	MemoryDeclaration decl;
	decl.type = MemoryDeclaration::Type::Var;
	decl.name = name;
	decl.var = std::move(defaultValue);
	return decl;
}

/*static*/MemoryNode::MemoryDeclaration MemoryNode::pod(const std::string& name)
{
	MemoryDeclaration decl;
	decl.type = MemoryDeclaration::Type::POD;
	decl.name = name;
	return decl;
}

} // ai
} // sl