#ifndef SL_AI_MEMORYNODE_HPP
#define SL_AI_MEMORYNODE_HPP

#include "NPC/Behaviors/BehaviorTreeNode.hpp"
#include "Parsing/Binding/Value.hpp"
#include "Utility/TypeTraits.hpp"

#include <cstdint>
#include <cstring>
#include <map>
#include <memory>
#include <vector>

namespace sl
{
namespace ai
{

class MemoryNode : public BehaviorTreeNode
{
public:
	class MemoryDeclaration
	{
		enum class Type
		{
			POD,
			Var
		};
		Type type;
		std::vector<uint8_t> data;
		ls::Value var;
		std::string name;

		friend class MemoryNode;
	};
	static MemoryDeclaration var(const std::string& name, ls::Value val = ls::Value());

	static MemoryDeclaration pod(const std::string& name);

	template <typename T>
	static MemoryDeclaration pod(const std::string& name, const T& data);

	MemoryNode(std::unique_ptr<BehaviorTreeNode> node);

	// not using initializer_list<MemoryDeclaration> because you can't just do {...} within make_unique
	// as it can't deduce the type, so instead, to make it easier on users to instantitate this class
	// we'll use variadic templates here.
	template <typename... Decls>
	MemoryNode(std::unique_ptr<BehaviorTreeNode> node, Decls... declarations);

	Result execute(BehaviorContext& context) override;

	void abort(BehaviorContext& context) override;

	template <typename... Decls>
	void addDeclaration(MemoryDeclaration decl, Decls... declarations)
	{
		addDeclaration(std::forward(decl));
		addDeclaration(std::forward<Decls...>(declarations)...);
	}

	void addDeclaration(MemoryDeclaration decl);

private:

	// <VarName, DefaultValue>
	std::vector<std::pair<std::string, ls::Value>> vars;
	std::vector<std::pair<std::string, std::vector<uint8_t>>> POD;
	std::unique_ptr<BehaviorTreeNode> node;
	bool initialized;
};

template <typename... Decls>
MemoryNode::MemoryNode(std::unique_ptr<BehaviorTreeNode> node, Decls... declarations)
	:MemoryNode(std::move(node))
{
	addDeclaration(std::forward<Decls>(declarations)...);
}

template <typename T>
/*static*/ MemoryNode::MemoryDeclaration MemoryNode::pod(const std::string& name, const T& defaultValue)
{
	static_assert(IsMemcopyable<T>::value, IsMemcopyable_Error);

	MemoryDeclaration decl;
	decl.type = MemoryDeclaration::Type::POD;
	decl.name = name;
	decl.data.resize(sizeof(T));
	std::memcpy(decl.data.data(), reinterpret_cast<const uint8_t*>(&defaultValue), sizeof(T));
	return decl;
}

} // ai
} // sl

#endif // SL_AI_MEMORYNODE_HPP