#include "NPC/Behaviors/Parallel.hpp"

#include "Utility/Assert.hpp"

namespace sl
{
namespace ai
{

Parallel::Parallel(ReturnStrategy strat, const char *customName)
	:BehaviorTreeNode(customName)
	,children()
	,strat(strat)
	,running(false)
{
}

BehaviorTreeNode::Result Parallel::execute(BehaviorContext& context)
{
	ASSERT(!children.empty());
	running = true;
	for (Child& child : children)
	{
		if (child.terminalResult == Result::Running || child.repeating)
		{
#ifdef SL_DEBUG
			child.terminalResult = child.node->debugExecute(context);
#else // SL_DEBUG
			child.terminalResult = child.node->execute(context);
#endif // SL_DEBUG
		}
	}
	int total = 0;
	int succeeded = 0;
	int failed = 0;
	for (const Child& child : children)
	{
		if (!child.ignored)
		{
			++total;
			switch (child.terminalResult)
			{
			case Result::Success:
				++succeeded;
				break;
			case Result::Failure:
				++failed;
				break;
			case Result::Running:
				break;
			}
		}
	}
	ASSERT(total > 0);
	Result result = Result::Running;
	switch (strat)
	{
	case ReturnStrategy::AnySucceed:
		if (succeeded > 0)
		{
			result = Result::Success;
		}
		else if (total == failed)
		{
			result = Result::Failure;
		}
		break;
	case ReturnStrategy::AllMustSucceed:
		if (succeeded == total)
		{
			result = Result::Success;
		}
		else if (failed > 0)
		{
			result = Result::Failure;
		}
		break;
	}
	if (result != Result::Running)
	{
		abort(context);
	}
	return result;
}

void Parallel::addChild(std::unique_ptr<BehaviorTreeNode> child)
{
#ifdef SL_DEBUG
	child->setParent(this);
#endif // SL_DEBUG
	children.push_back({ std::move(child), false, false });
}

void Parallel::addIgnoredChild(std::unique_ptr<BehaviorTreeNode> child, bool repeating)
{
#ifdef SL_DEBUG
	child->setParent(this);
#endif // SL_DEBUG
	children.push_back({ std::move(child), true, repeating });
}

void Parallel::abort(BehaviorContext& context)
{
	if (running)
	{
		// is this the correct behaviour?
		for (Child& child : children)
		{
			child.node->abort(context);
			child.terminalResult = Result::Running;
		}
		running = false;
	}
}

} // ai
} // sl