#include "NPC/Dialogue/DialogueGUI.hpp"

#include "Engine/GUI/Button.hpp"
#include "Engine/GUI/NormalBackground.hpp"
#include "Engine/GUI/Scrollbar.hpp"
#include "Engine/GUI/WidgetList.hpp"
#include "Engine/Scene.hpp"
#include "NPC/ClothesDatabase.hpp"
#include "NPC/Stats/Soul.hpp"
#include "NPC/Stealth/Suspicion.hpp"
#include "NPC/Dialogue/AttributeLoading.hpp"
#include "NPC/Dialogue/DialogueBindings.hpp"
#include "NPC/Dialogue/Popup.hpp"
#include "NPC/NPC.hpp"
#include "Player/Fetishes.hpp"
#include "Player/PlayerStats.hpp"
#include "Parsing/Binding/BoundNumber.hpp"
#include "Parsing/Binding/Bindings.hpp"
#include "Parsing/Parsers/FileParser.hpp"
#include "Town/World.hpp"
#include "Utility/Assert.hpp"
#include "Utility/ContainerAlgorithms.hpp"
#include "Utility/Logger.hpp"
#include "Utility/Memory.hpp"
#include "Utility/Random.hpp"
#include "Utility/Ranges.hpp"

// temp?
#include "Parsing/Errors/SyntaxError.hpp"
#include "Police/CrimeEntity.hpp"
#include "Police/CrimeUtil.hpp"

namespace sl
{
// TODO: remove once you're confident in the threading BS
static void blockingPrint(const char *c) {
	static std::mutex m;
	std::unique_lock<std::mutex> l(m);
	std::cout << c << std::endl;
}
static const int barwidth = 16;

static const int headerheight = 128;

static const int buttonheight = 32;

static const int HPBAR_WIDTH = 160;
static const int HPBAR_HEIGHT = 16;
static const int HPBAR_X = 8;
static const int HPBAR_Y = 8;


DialogueGUI::DialogueGUI(ls::Bindings& context, const Rect<int>& box, NPC& npcEntity, PlayerStats& player)
	:GUIWidget(box)
	,context(context)
	,buttonWidth(box.width - barwidth)
	,message(new gui::TextLabel(Rect<int>(box.left + 1, box.top + 1, box.width - 3 - 100, headerheight - 2), gui::TextLabel::Config().pad(16, 8).setWritePolicy(gui::TextLabel::Letter, 5.f)))
	,optionsRect(box.left + 1, box.top + headerheight + 1, box.width - 1, box.height - headerheight - 3)
	,optionsList(nullptr)
	,scenarios()
	,lines()
	,playerHPBar(*player.soul)
	,enemyHPBar(*npcEntity.getSoul())
	,player(player)
	,npcEntity(npcEntity)
	,npc(*npcEntity.getSoul())
	,popup(nullptr)
	,threadRunning(ThreadRunning::Main)
	,messageThreadDone(true)
	,currentChoiceType(ChoiceType::ClickThrough)
	,choiceAreaHeightSofar(0)
	,choiceIndex(0)
	,forkChoicesUsed(true)
	,currentOutcome(nullptr)
	,currentScenario(nullptr)
{
	auto popupWidget = unique<Popup>(Vector2i(box.width - Popup::WIDTH, -Popup::HEIGHT + headerheight), npc.appearance);
	popup = popupWidget.get();
	addWidget(std::move(popupWidget));
	// make sure to add this after the pop-up so it renders afterwards.
	this->addWidget(std::unique_ptr<gui::TextLabel>(message));

	//message->applyBackground(gui::NormalBackground(sf::Color::Black, sf::Color(192, 192, 192)));

	lines.push_back(Line(box.left, box.top + headerheight, box.left + box.width, box.top + headerheight, sf::Color::White));
	lines.push_back(Line(box.left + 1, box.top + headerheight - 1, box.left + box.width - 1, box.top + headerheight - 1, sf::Color(192, 192, 192)));
	lines.push_back(Line(box.left + 1, box.top + box.height - 1, box.left + box.width - 1, box.top + box.height - 1, sf::Color(192, 192, 192)));

	context.registerFunction("local", "attack", *this, &DialogueGUI::attack);

	context.registerFunction("local", "takePicture", [&] {
		player.stalkingData.stalk(
			npc.getID(),
			npcEntity.getWorld()->getGameTime(),
			NPCPhoto{
				NPCPhoto::getColorFromSurroundings(&npcEntity),
				npc.appearance,
				popup->getCategory(),
				popup->getAction(),
				popup->getFace()});
	});
	// TODO: fix sepplesToLoliscript() for std::bind things?
	context.registerFunction("local", "viewerCount", [&npcEntity]() -> int {
		return computeViewers(npcEntity).size();
	});
	context.registerFunction("local", "privacy", std::function<float()>(std::bind(computePrivacy, ScenePos(npcEntity), nullptr, std::vector<NPC*>{&npcEntity})));


	// actually maybe these shouldn't be directly exposed? uncomment if needed
	//context.registerVar("player", "hp", unique<ls::BoundNumber<decltype(player.soul->getStatus().hp)>>(player.soul->getStatus().hp));
	//context.registerVar("player", "maxhp", unique<ls::BoundNumber<decltype(player.soul->getStatus().maxhp)>>(player.soul->getStatus().maxhp));
	//context.registerVar("loli", "hp", unique<ls::BoundNumber<decltype(npc.getStatus().hp)>>(npc.getStatus().hp));
	//context.registerVar("loli", "maxhp", unique<ls::BoundNumber<decltype(npc.getStatus().maxhp)>>(npc.getStatus().maxhp));

	registerDialogueBindings(*player.soul, npc, context);

	// needs access to currentSexActs, and doesn't benefit from registerDialogueBinding's loli+player dual registration since we only keep info for currentSexActs on the loli
	context.registerFunction("loli", "sexCountIncludingCurrent", [&](const ls::Arguments& args) -> int {
		int count = 0;
		bool query[Soul::SexType_Count];
		if (args.count() == 0)
		{
			std::fill(std::begin(query), std::end(query), true);
		}
		else
		{
			std::fill(std::begin(query), std::end(query), false);
			for (int i = 0; i < args.count(); ++i)
			{
				query[args[i].asInt()] = true;
			}
		}
		for (int i = 0; i < Soul::SexType_Count; ++i)
		{
			if (query[i])
			{
				count += npc.getSexCount(static_cast<Soul::SexType>(i));
			}
		}
		for (int act : currentSexActs)
		{
			if (query[act])
			{
				++count;
			}
		}
		return count;
	});

	itemQuery.registerInventory("player", &player.soul->items, player.soul.get());
	itemQuery.registerInventory("loli", &npc.items, &npc);

	ls::registerFlags(context, player.soul->getFlags(), "player");
	ls::registerFlags(context, npc.getFlags(), "loli");
	ls::registerFlags(context, tempFlags, "temp");
	ls::registerFlags(context, player.permFlags, "perm");

	initAttributes("", npc, attributes);
	initAttributes("player_", *player.soul, attributes);

	context.registerFunction("fetish", "enabled", [](std::string id) {
		return Fetishes::get().isEnabled(id);
	});

	context.registerFunction("local", "markSexAct", [this](int typeIndex, ls::OptionalParam<bool> isConsensual) {
		LSASSERT(typeIndex >= 0 && typeIndex < static_cast<int>(Soul::SexType::COUNT));
		if (currentSexActs.empty())
		{
			currentCrimeId = commitCrime(this->npcEntity, Crime::Type::SexualAssault, isConsensual.get(true));
		}
		currentSexActs.insert(typeIndex);
	});
	context.registerFunction("local", "finalizeSex", [this] {
		if (currentSexActs.empty())
		{
			currentSexActs.insert(static_cast<int>(Soul::SexType::Molestation));
		}
		for (int typeIndex : currentSexActs)
		{
			const Soul::SexType type = static_cast<Soul::SexType>(typeIndex);
			this->player.soul->incrementSex(type);
			this->npc.incrementSex(type);
		}
		currentSexActs.clear();
	});

	context.registerFunction("local", "nakadashi", [&npcEntity, &player] {
		npcEntity.getSoul()->mate(*npcEntity.getWorld(), *player.soul, Random::get());
	});

	// ---------------------------------------------------------------------------
	// -------------------------------- Clothing / Popups ------------------------
	context.registerFunction("loli", "setClothesOn", [&](ClothingType type, bool val)
	{
		npc.appearance.clothingLayers.fromEnum(type) = val;
		popup->reload(true);
		npcEntity.refreshDrawables();
	});

	context.registerFunction("loli", "clothesOn", [&](ClothingType type)
	{
		// @HACK @TODO right now we're representing no hat with "". what should we do instead?
		return npc.appearance.clothingLayers.fromEnum(type) && !npc.appearance.getClothing(type).empty();
	});

	context.registerFunction("loli", "hasClothes", [&](ClothingType type) -> bool
	{
		const std::string& clothingId = npc.appearance.getClothing(type);
		if (clothingId == "")
		{
			return false;
		}
		const auto& info = ClothesDatabase::get().getInfo(clothingId);
		return info.exactName != "";
	});

	context.registerFunction("loli", "genericClothesName", [&](ClothingType type) -> std::string
	{
		const auto& info = ClothesDatabase::get().getInfo(npc.appearance.getClothing(type));
		return info.genericName;
	});

	context.registerFunction("loli", "exactClothesName", [&](ClothingType type) -> std::string
	{
		const auto& info = ClothesDatabase::get().getInfo(npc.appearance.getClothing(type));
		return info.exactName;
	});

	context.registerFunction("loli", "genericOutfitTopName", [&]
	{
		const auto& info = ClothesDatabase::get().getInfo(npc.appearance.getClothing(ClothingType::Outfit));
		return info.genericTopName;
	});

	context.registerFunction("loli", "genericOutfitBottomName", [&]
	{
		const auto& info = ClothesDatabase::get().getInfo(npc.appearance.getClothing(ClothingType::Outfit));
		return info.genericBottomName;
	});

	context.registerFunction("loli", "hasCumOn", [&](std::string layer) -> bool
	{
		return npc.appearance.getCumLayers().count(layer) != 0;
	});

	context.registerFunction("loli", "addCumOn", [&](std::string layer)
	{
		npc.appearance.addCumLayer(layer);
		popup->reload(true);
	});

	context.registerFunction("loli", "removeCumOn", [&](std::string layer)
	{
		npc.appearance.removeCumLayer(layer);
		popup->reload(true);
	});

	popup->registerFunctions(context, "loli");

	// ---------------------------------------------------------------------------
	// --------------------------------- Item / Inventory ------------------------
	// TODO: loli inventory too? separate loli + player-specific stuff into some otherr place??
	context.registerFunction("player", "hasItemFromCategories", [&](const ls::Arguments& args) -> bool
	{
		std::vector<Inventory::ItemRef> results;
		for (int i = 0; i < args.count(); ++i)
		{
			player.soul->items.queryItems(results, Item::strToCategory(args[i].asString()));
		}
		return !results.empty();
	});

	context.registerFunction("player", "hasItemFromByTypes", [&](const ls::Arguments& args) -> bool
	{
		std::vector<Inventory::ItemRef> results;
		for (int i = 0; i < args.count(); ++i)
		{
			player.soul->items.queryItems(results, Item::strToType(args[i].asString()));
		}
		return !results.empty();
	});

	context.registerFunction("item", "giftTo", [&](std::string target)
	{
		itemQuery.giftItem(target);
	});

	context.registerFunction("item", "consumeTo", [&](std::string target)
	{
		itemQuery.consumeItem(target);
	});

	context.registerFunction("item", "gift", [&]
	{
		itemQuery.giftItem("loli");
	});

	context.registerFunction("item", "consume", [&]
	{
		itemQuery.consumeItem("loli");
	});

	context.registerFunction("item", "getType", [&]() -> std::string
	{
		return Item::typeToStr(itemQuery.getSelectedItem()->getType());
	});
#define REGISTER_ITEM_GETTER(getter, type) \
	context.registerFunction("item", #getter, [&]() -> type \
	{ \
		return itemQuery.getSelectedItem()->getter(); \
	})
	REGISTER_ITEM_GETTER(getWeight, int);
	REGISTER_ITEM_GETTER(getValue, int);
	REGISTER_ITEM_GETTER(getName, std::string);
	REGISTER_ITEM_GETTER(getDesc, std::string);
	REGISTER_ITEM_GETTER(getMaxUses, int);
	REGISTER_ITEM_GETTER(getUses, int);
	REGISTER_ITEM_GETTER(getGirth, float);
	REGISTER_ITEM_GETTER(getLength, float);
#undef REGISTER_ITEM_GETTER
	context.registerFunction("item", "hasCategories", [&](const ls::Arguments& args) -> ls::Value
	{
		Item::Categories categories = Item::None;
		for (int i = 0; i < args.count(); ++i)
		{
			categories |= Item::strToCategory(args[i].asString());
		}
		return ls::Value(itemQuery.getSelectedItem()->hasComponents(categories));
	});
	context.registerFunction("item", "flavorResponse", [&]() -> float
	{
		return npc.preferences.calculateResponse(itemQuery.getSelectedItem()->getFlavor());
	});




	// ---------------------------------------------------------------------------
	// ------------------------------ Message Block ------------------------------
	context.registerFunction("scenario", "id", [&] { return scenarioId; });

	//context.registerFunction("local", "msg", *this, &DialogueGUI::addMessage);
	context.registerFunction("local", "msg", [this](const ls::Arguments& args) {
		std::stringstream ss;
		ss << args[0].asString();
		for (int i = 1; i < args.count(); ++i)
		{
			ss << ' ' << args[i].asString();
		}
		addMessage(ss.str());
	});

	context.registerFunction("local", "unusedFork", [this]
	{
		// only call this if you know the message block will end after the user does a fork.
		this->addMessage("You should not see this. This to work around a bug in NPC / DialogueGUI's fork() method");
	});

	context.registerFunction("local", "addForkChoice", [this](const ls::Arguments& args)
	{
		VERIFY_ARG_COUNT(args, 1, 2);
		if (args.count() == 1)
		{
			const std::string message = args[0].asString();
			this->forkChoices.push_back(std::make_pair(message, message));
		}
		else
		{
			const std::string key = args[0].asString();
			const std::string message = args[1].asString();
			this->forkChoices.push_back(std::make_pair(key, message));
		}
		forkChoicesUsed = false;
	});

	
	context.registerFunction("local", "fork", [this](const ls::Arguments& args) -> int
	{
		for (int i = 0; i < args.count(); ++i)
		{
			const std::string message = args[i].asString();
			this->forkChoices.push_back(std::make_pair(message, message));
		}
		return fork();
	});

	// A boolean condition will change whether the following argument will be allowed or not.
	context.registerFunction("local", "forkWithConds", [this](const ls::Arguments& args) -> int
	{
		ASSERT(forkChoices.empty());
		ASSERT(args.count() % 2 == 0);
		for (int i = 0; i < args.count(); i += 2)
		{
			if (!args[i].asBool())
			{
				continue;
			}
			const std::string message = args[i + 1].asString();
			this->forkChoices.push_back(std::make_pair(message, message));
		}
		forkChoicesUsed = true;
		return fork();
	});

	context.registerFunction("local", "forkAsStr", [this](const ls::Arguments& args) -> std::string
	{
		for (int i = 0; i < args.count(); ++i)
		{
			const std::string message = args[i].asString();
			this->forkChoices.push_back(std::make_pair(message, message));
		}
		forkChoicesUsed = true;
		return forkChoices[fork()].first;
	});
}

DialogueGUI::~DialogueGUI()
{
	ASSERT(threadRunning == ThreadRunning::Main);
	// this shouldn't be necessary, but just in case...
	while (!messageThreadDone)
	{
		messageCond.notify_all();
	}
	if (messageThread.joinable())
	{
		messageThread.join();
	}
	runAfterMessageThread();
}

void DialogueGUI::setScenario(const std::string& scenarioName)
{
	scenarios.clear();
	const Scenario& scenario = ls::FileParser::get().readDialogueFile(scenarioName);
	scenarioId = scenario.getIdentifier();
	scenarioStack.clear();
	scenarioStack.push_back(scenarioId);
	runScenario(scenario);
}

void DialogueGUI::runAfterMessageDone(MoveableFunction<void> f)
{
	if (threadRunning == ThreadRunning::Message)
	{
		toRunAfterMessageThread.push_back(std::move(f));
	}
	else
	{
		f();
	}
}



/*		GUIWidget private methods		*/
void DialogueGUI::onDraw(sf::RenderTarget& target) const
{
	for (const Line& line : lines)
	{
		line.draw(target);
	}
}

void DialogueGUI::beforeDraw(sf::RenderTarget& target) const
{
	playerHPBar.draw(target);
	enemyHPBar.draw(target);
}

void DialogueGUI::onUpdate(int x, int y)
{
	const int xdif = x - box.left;
	const int ydif = y - box.top;

	if (xdif || ydif)
	{
		for (Line& line : lines)
		{
			line.update(line.getCurrentBounds().left + xdif, line.getCurrentBounds().top + ydif);
		}
	}
	playerHPBar.update(x, y - HPBAR_HEIGHT);
	enemyHPBar.update(x + box.width - HPBAR_WIDTH, y - HPBAR_HEIGHT);
	if (message->isDoneWriting())
	{
		//Scene& physScene = *static_cast<Scene*>(const_cast<AbstractScene*>(scene));
		//physScene.pause("NPC");
		//physScene.pause("Car");
	}

	// DEMO HACK
	if (!currentSexActs.empty())
	{
		CrimeEntity *kidnappingCrime = new CrimeEntity(npcEntity.getPos(), currentCrimeId, true, 3);
		npcEntity.getScene()->addEntityToList(kidnappingCrime);
	}
}
/*		DialogueGUI private methods		*/
void DialogueGUI::chooseOption(int index)
{
	if (isUpdating())
	{
		callThisLater(std::bind(&DialogueGUI::chooseOption, this, index));
	}
	else if (message->isDoneWriting())
	{
		Logger::log(Logger::GUI, Logger::Debug) << "Clicked on [" << index << "]\n";
		switch (currentChoiceType)
		{
		case ChoiceType::ItemSelect:
		case ChoiceType::Scenario:
		{
			int scenarioIndex = index;
			if (!itemQuery.getItems().empty())
			{
				ASSERT(index >= 0 && index < itemQuery.getItems().size());
				itemQuery.selectItem(index);
				// when invoking an item selection, there is only one thing that can be linked,
				// and that outcome wil have scenarioIndex 0.
				scenarioIndex = 0;
			}
			else
			{
				ASSERT(index >= 0 && index < scenarios.size());
			}

			auto& choice = scenarios[scenarioIndex];
			switch (choice.visitType)
			{
			case ScenarioVisitType::Normal:
				scenarioId = choice.scenario->getIdentifier();
				// todo: should loops contribute? ie do a check if (scenarioStack.back() != scenarioId)
				scenarioStack.push_back(scenarioId);
				break;
			case ScenarioVisitType::Undo:
				ASSERT(scenarioStack.size() >= 2);
				scenarioStack.pop_back();
				scenarioId = scenarioStack.back();
				break;
			}
			runScenario(*choice.scenario);
		}
		break;
		case ChoiceType::ForkCall:
		{
			forkChoice = index;

			{
				std::unique_lock<std::mutex> lock(communicationMutex);
				threadRunning = ThreadRunning::Message;
			}
			messageCond.notify_all();
			{
				std::unique_lock<std::mutex> lock(mainMutex);
				// to eliminate threading issues, we make sure that the two threads are actually just being executed
				// in lock-step. so keep running the other one until it waits - then resume with this thread.
				// the reason we're even using threads is to avoid manually coding loliscript-C++ coroutines into loliscript.
				mainCond.wait(lock, [this] {return threadRunning == ThreadRunning::Main;});
			}

			forkChoices.clear();

			consumeMessage();

			createChoiceButtons();
		}
		break;
		default:
			ERROR(":(");
		}
	}
}

void DialogueGUI::advanceDialogue()
{
	if (isUpdating())
	{
		callThisLater(std::bind(&DialogueGUI::advanceDialogue, this));
	}
	else
	{
		consumeMessage();

		createChoiceButtons();
	}
}

// this used to be in updateEverything(). maybe we'll need it again?

//	NOTE: these can no longer be assumed to be true since WidgetGrid::setRows might get
//	called later since it's updating right now
//ASSERT(!scenarios.empty());
//ASSERT(optionList->getColumns() == 1);
//ASSERT(optionList->getRows() == scenarios.size());
//Scene& physScene = *static_cast<Scene*>(const_cast<AbstractScene*>(scene));
//physScene.unpause("NPC");
//physScene.unpause("Car");

void DialogueGUI::addScenario(const std::pair<std::string, std::string>& link, ScenarioVisitType type)
{
	const Scenario *scenario = nullptr;
	try
	{
		scenario = &ls::FileParser::get().readDialogueFile(link.first);
	}
	catch (ls::SyntaxError& see)
	{
		Logger::log(Logger::Dialogue, Logger::Serious) << "Syntax error: " << see.what() << std::endl;
		return;
	}
	if (scenario->isAvailable(context))
	{
		const std::string title = link.second.empty() ? scenario->getTitle(context) : link.second;
		scenarios.push_back({
			std::move(title),
			scenario,
			type
		});
	}
}

DialogueGUI::HPBar::HPBar(Soul& soul)
	:soul(soul)
	,back(sf::Vector2f(HPBAR_WIDTH - 2, HPBAR_HEIGHT - 2))
	,red(sf::Vector2f(HPBAR_WIDTH - HPBAR_X * 2 - 2, HPBAR_HEIGHT - HPBAR_Y * 2 - 2))
	,bar(sf::Vector2f(HPBAR_WIDTH - HPBAR_X * 2 - 2, HPBAR_HEIGHT - HPBAR_Y * 2 - 2))
{
	back.setFillColor(sf::Color::Black);
	back.setOutlineColor(sf::Color::White);
	back.setOutlineThickness(1);

	red.setFillColor(sf::Color::Red);
	red.setOutlineColor(sf::Color::White);
	red.setOutlineThickness(1);

	bar.setFillColor(sf::Color::Green);
	
}

void DialogueGUI::HPBar::draw(sf::RenderTarget& target) const
{
	target.draw(back);
	target.draw(red);
	target.draw(bar);
}

void DialogueGUI::HPBar::update(int x, int y)
{
	++x;
	back.setPosition(x, y);
	red.setPosition(x + HPBAR_X, y + HPBAR_Y);
	bar.setPosition(x + HPBAR_X, y + HPBAR_Y);
	const int barWidth = (HPBAR_WIDTH - 2 * HPBAR_X) * (float) soul.getStatus().hp / soul.getStatus().maxhp;
	bar.setSize(sf::Vector2f(barWidth, bar.getSize().y));
}

void DialogueGUI::attack(float strScale, float dexScale, float baseHitChance, float staminaCost, float staminaDamage, std::string successString, std::string missString)
{
	const float strBaseDamage = 0.5f;
	const float dexBaseDamage = 0.2f;
	const float critChance = 0.15f;
	Random& rng = Random::get();

	float damage = -1.f;

	const bool success = rng.chance(std::max(0.f, player.soul->dexterity() - npc.dexterity() * baseHitChance));
	if (success)
	{
		damage = strScale * player.soul->strength() * strBaseDamage + dexScale * player.soul->dexterity() * dexBaseDamage;
		//messages.push(successString);
	}
	else
	{
		damage = strScale * npc.strength() * strBaseDamage + dexScale * npc.dexterity() * dexBaseDamage;
		//messages.push(missString);
	}

	if (rng.chance(critChance))
	{
		//messages.push("Critical hit!");
		damage *= 2.f;
	}

	std::stringstream ss;
	ss << damage << " is inflicted!";
	//messages.push(ss.str());

	if (success)
	{
		if (npc.getStatus().hp <= damage)
		{
			npc.getStatus().hp = 0;
			// kill !
		}
		else
		{
			npc.getStatus().hp -= damage;
			if (npc.getStatus().stamina <= staminaDamage)
			{
				// fall unconscious
				npc.getStatus().stamina = 0.f;
			}
			else
			{
				npc.getStatus().stamina -= staminaDamage;
			}
		}
	}
	else
	{
		if (player.soul->getStatus().hp <= damage)
		{
			player.soul->getStatus().hp = 0;
			// kill !
		}
		else
		{
			player.soul->getStatus().hp -= damage;
		}
	}

	player.soul->getStatus().stamina -= staminaCost;
}

void DialogueGUI::addMessage(std::string str)
{
	{
		std::unique_lock<std::mutex> lock(communicationMutex);
		threadRunning = ThreadRunning::Main;
	}
	blockingPrint("message waiting");
	mainCond.notify_all(); // in case the main thread is also waiting (which is the case for the first message)
	{
		std::unique_lock<std::mutex> lock(messageMutex);
		messageCond.wait(lock, [this] { return threadRunning == ThreadRunning::Message; });
	}
	// SFML runs into opengl errors when ran from another thread, and setString()/etc will involve opengl calls
	// since they write to a RenderTexter. To fix this, we introduced newMessage(Mutex).
	// In the update() call it will be polled once per frame (Ew! - but there's only 1 doing this per-frame so this should be okay? Let's see!).
	{
		std::unique_lock<std::mutex> newMessageLock(communicationMutex);
		ASSERT(newMessage.empty());
		newMessage = std::move(str);
	}
	blockingPrint("message wake up");
}

void DialogueGUI::consumeMessage()
{
	{
		std::unique_lock<std::mutex> lock(communicationMutex);
		threadRunning = ThreadRunning::Message;
	}
	blockingPrint("main waiting");
	messageCond.notify_all();
	std::unique_lock<std::mutex> lock(mainMutex);
	// to eliminate threadding issues, we make sure that the two threads are actually just being executed
	// in lock-step. so keep running the other one until it waits - then resume with this thread.
	// the reason we're even using threads is to avoid manually coding loliscript-C++ coroutines into loliscript.
	mainCond.wait(lock, [this] {return threadRunning == ThreadRunning::Main;});
	blockingPrint("main done waiting");
	// this is because this must be done on the main thread, not messageThread
	// or else SFML/opengl error because of threading shit.
	if (messageThreadDone && messageThread.joinable())
	{
		messageThread.join();
		blockingPrint("message thread joined");
		runAfterMessageThread();
	}
	std::unique_lock<std::mutex> newMessageLock(communicationMutex);
	if (!newMessage.empty())
	{
		message->setString(newMessage);
		//message->setWritePolicy(gui::TextLabel::Letter, 5.f); (possible redundant since in ctor?)
		newMessage.clear();
	}
}

void DialogueGUI::runScenario(const Scenario& scenario)
{
	currentScenario = &scenario;
	currentOutcome = &scenario.chooseOutcome(context, attributes);
	auto messageBlock = currentOutcome->getMessages();

	ASSERT(messageBlock != nullptr);
	if (messageThread.joinable())
	{
		messageThread.join();
	}
	runAfterMessageThread();
	messageThreadDone = false;
	threadRunning = ThreadRunning::Main;
	messageThread = std::thread([this, messageBlock] {
		// We must make sure the main thread is waiting, or else our notify_all() at the end is useless
		// This could probably be done before the notify_all(), but it's safer/more deterministic up here
		// since then we're in total lock-step.
		// This issue presented itself mostly when you had a message block with no messages (ie when the msg()'s were conditional)
		{
			blockingPrint("message thread launched");
			std::unique_lock<std::mutex> lock(messageMutex);
			messageCond.wait(lock, [this] { return threadRunning == ThreadRunning::Message; });
		}	
		blockingPrint("message thread running in proper");
		try
		{
			messageBlock->execute(context);
		}
		catch (const ls::RuntimeError& riee)
		{
			ERROR(riee.what());
		}
		{
			std::unique_lock<std::mutex> lock(communicationMutex);
			messageThreadDone = true;
			threadRunning = ThreadRunning::Main;
		}
		if (!forkChoices.empty())
		{
			if (forkChoicesUsed)
			{
				// If the dialogue file has a fork() action without any msg() actions following it, this causes a deadlock.
				// It looks nasty so let's just put this call in here instead.
				ERROR("please put an empty msg(\"\") call after unused forks for now until this is fixed.");
			}
			else
			{
				auto log = Logger::log(Logger::Dialogue, Logger::Debug);
				log << "message block ended with " << forkChoices.size() << " unused fork choices:" << std::endl;
				for (const auto& choice : forkChoices)
				{
					if (choice.first == choice.second)
					{
						log << "<\"" << choice.first << "\">" << std::endl;
					}
					else
					{
						log << "<\"" << choice.first << "\", \"" << choice.second << ">" << std::endl;
					}
				}
				log << "Please ensure that all addForkChoice() calls are followed by a fork() or a forkAsStr() call!" << std::endl;
				forkChoices.clear();
			}
		}
		mainCond.notify_all();
		blockingPrint("message thread ended");
	});
	blockingPrint("setting thread to msg");
	threadRunning = ThreadRunning::Message;
	messageCond.notify_all();
	blockingPrint("notified message thread");
	{
		std::unique_lock<std::mutex> lock(mainMutex);
		// to eliminate threadding issues, we make sure that the two threads are actually just being executed
		// in lock-step. so keep running the other one until it waits - then resume with this thread.
		// the reason we're even using threads is to avoid manually coding loliscript-C++ coroutines into loliscript.
		mainCond.wait(lock, [this] {return threadRunning == ThreadRunning::Main;});
		{
			std::unique_lock<std::mutex> lock(communicationMutex);
			if (!newMessage.empty())
			{
				message->setString(newMessage);
				//message->setWritePolicy(gui::TextLabel::Letter, 5.f); (possibly redundant since in ctor?)
				newMessage.clear();
			}
		}
	}
	// The thread might have instantly ended without any messages! (totally valid, see: exit.dlg)
	if (messageThreadDone)
	{
		runAfterMessageThread();
	}
	else
	{
		consumeMessage();
	}
#ifdef SL_DEBUG
	// print out the scenarioStack
	auto log = Logger::log(Logger::Dialogue, Logger::Debug);
	log << "scenarioStack:\n";
	for (const std::string& s : scenarioStack)
	{
		log << "\t" << s << std::endl;
	}
#endif // SL_DEBUG

	createChoiceButtons();
}

void DialogueGUI::createOptionList()
{
	auto scrollList = unique<gui::WidgetList<int>>(optionsRect + box.topLeft(),
		[this](int index) {
			return optionStrings[index];
		},
		std::bind(&DialogueGUI::chooseOption, this, std::placeholders::_1)
	);
	scrollList->setHorizAlignment(gui::TextLabel::Left);
	scrollList->updateItems(range<int>(0, optionStrings.size()));
	optionsList = scrollList.get();
	addWidget(std::move(scrollList));
}

void DialogueGUI::createChoiceButtons()
{
	if (optionsList)
	{
		removeWidget(*optionsList);
		optionsList = nullptr;
	}
	if (messageThreadDone)
	{
		computeLinks();
		// now add all the choices to the option menu, since we've finished the messages and need to do links now.
		if (!itemQuery.getItems().empty())
		{
			ASSERT(scenarios.size() == 1);
			optionStrings.resize(itemQuery.getItems().size());
			for (std::size_t i = 0; i < itemQuery.getItems().size(); ++i)
			{
				optionStrings[i] = itemQuery.getItems()[i]->getName();
			}
			createOptionList();
			currentChoiceType = ChoiceType::ItemSelect;
		}
		else
		{
			if (scenarios.empty())
			{
				addScenario(std::make_pair("exit", "SOMEONE FUCKED UP THEIR .DLG FILE! There's no scenarios here.\nFile : " + scenarioId + " - Click here to exit."));
			}
			optionStrings.resize(scenarios.size());
			for (std::size_t i = 0; i < scenarios.size(); ++i)
			{
				optionStrings[i] = scenarios[i].title;
			}
			createOptionList();
			currentChoiceType = ChoiceType::Scenario;
		}
	}
	else
	{
		if (!forkChoices.empty())
		{
			optionStrings.resize(forkChoices.size());
			for (std::size_t i = 0; i < forkChoices.size(); ++i)
			{
				optionStrings[i] = forkChoices[i].second;
			}
			createOptionList();
			currentChoiceType = ChoiceType::ForkCall;
		}
		else
		{
			// the x/y here are IGNORED because of a redesign in WidgetGrid -- kept here in case it's reverted
			//Rect<int> buttonRect(0, 1, box.width - barwidth - 2, scrollWindow->getBox();
			auto button = std::make_unique<gui::Button>(unique<gui::TextLabel>(optionsRect + box.topLeft(), ".\t.\t.", gui::TextLabel::Config().centered().pad(8, 6)),
				sf::Color(192, 192, 192), sf::Color::White,
				gui::NormalBackground(sf::Color::Black, sf::Color(192, 192, 192)),
				gui::NormalBackground(sf::Color(32, 32, 32), sf::Color::White)
			);
			button->registerMouseInputEvent(Key::GUIClick, GUIWidget::InputEventType::Released, std::bind(&DialogueGUI::advanceDialogue, this));
			button->registerInputEvent(Key::GUIAccept, GUIWidget::InputEventType::Released, std::bind(&DialogueGUI::advanceDialogue, this));
			optionsList = button.get();
			addWidget(std::move(button));
			currentChoiceType = ChoiceType::ClickThrough;
		}
	}
}

int DialogueGUI::fork()
{
	{
		std::unique_lock<std::mutex> lock(communicationMutex);
		threadRunning = ThreadRunning::Main;
	}
	mainCond.notify_all(); // in case the main thread is also waiting (which is the case for the first message)
	{
		std::unique_lock<std::mutex> lock(messageMutex);
		messageCond.wait(lock, [this] { return threadRunning == ThreadRunning::Message; });
	}
	return forkChoice;
}

void DialogueGUI::computeLinks()
{
	scenarios.clear();
	bool usePrevLinks = false;
	std::vector<std::pair<std::string, std::string>> links;
	std::vector<ScenarioConfig> backOptions;
	// TODO: decide if it should be execute() -> links(), or links() -> execute()
	currentOutcome->execute(context);
	currentScenario->getLinks(*currentOutcome, context, itemQuery, links);
	for (const std::pair<std::string, std::string>& link : links)
	{
		if (link.first == "?BACK?")
		{
			if (scenarioStack.size() < 2)
			{
				Logger::log(Logger::Dialogue, Logger::Debug) << "Unable to follow back() due to insufficient dialogue visit stack" << std::endl;
			}
			else
			{
				const Scenario& scenario = ls::FileParser::get().readDialogueFile(scenarioStack[scenarioStack.size() - 2]);
				if (scenario.isAvailable(context))
				{
					const std::string title = link.second.empty() ? scenario.getTitle(context) : std::move(link.second);
					backOptions.push_back({
						std::move(title),
						&scenario,
						ScenarioVisitType::Undo
					});
				}
			}
		}
		else if (link.first == "?PREV?")
		{
			usePrevLinks = true;
		}
		else
		{
			addScenario(link);
		}
	}

	if (usePrevLinks)
	{
		ASSERT(scenarioStack.size() >= 2);
		const Scenario& prevScenario = ls::FileParser::get().readDialogueFile(scenarioStack[scenarioStack.size() - 2]);
		if (prevScenario.isAvailable(context))
		{
			std::vector<std::pair<std::string, std::string>> prevLinks;
			const Outcome& prevScenarioOutcome = prevScenario.chooseOutcome(context, attributes);
			prevScenario.getLinks(prevScenarioOutcome, context, itemQuery, prevLinks);
			for (const std::pair<std::string, std::string>& prevLink : prevLinks)
			{
				if (prevLink.first == "?BACK?" || prevLink.first == "?PREV?")
				{
					// TODO: if this is a feature we would want, make it recursively apply this shit?
					// I think a back() on a prev() makes sense, but no recursive prev()'s...
					Logger::log(Logger::Area::NPC, Logger::Priority::Debug) << "Ignoring prev() or back() on node previous to a prev()" << std::endl;
				}
				else
				{
					addScenario(prevLink, ScenarioVisitType::Undo);
				}
			}
		}
	}

	//scenarios.clear(); why was this here?!?!?

	mergeInto(scenarios, backOptions);
}

void DialogueGUI::runAfterMessageThread()
{
	for (const auto& f : toRunAfterMessageThread)
	{
		f();
	}
	toRunAfterMessageThread.clear();
}

} // sl
