#ifndef SL_DIALOGUEGUI_HPP
#define SL_DIALOGUEGUI_HPP

#include <SFML/Graphics.hpp>

#include <condition_variable>
#include <map>
#include <memory>
#include <mutex>
#include <set>
#include <string>
#include <thread>
#include <vector>

#include "Engine/Graphics/Line.hpp"
#include "Engine/GUI/GUIWidget.hpp"
#include "Engine/GUI/TextLabel.hpp"
#include "NPC/Dialogue/ItemQuery.hpp"
#include "NPC/Dialogue/Scenario.hpp"
#include "NPC/Dialogue/Outcome.hpp"
#include "Parsing/Binding/Bindings.hpp"
#include "Parsing/Binding/Flags.hpp"
#include "Parsing/Binding/Arguments.hpp"
#include "Utility/MoveableFunction.hpp"

namespace sl
{

class NPC;
class Soul;
class PlayerStats;
class Popup;

class DialogueGUI : public gui::GUIWidget
{
public: 
	DialogueGUI(ls::Bindings& context, const Rect<int>& box, NPC& npcEntity, PlayerStats& player);

	~DialogueGUI();

	void setScenario(const std::string& filename);

	/**
	 * For when the message thread needs to run some code that would result in this DialogueGUI having destructive behaviour
	 * done to it (ie changeState()), so instead execute that function after the thread finishes.
	 */
	void runAfterMessageDone(MoveableFunction<void> f);

private:
	class HPBar
	{
	public:
		HPBar(Soul& soul);

		void draw(sf::RenderTarget& target) const;
		void update(int x, int y);
	
	private:
		Soul& soul;
		sf::RectangleShape back, red, bar;
		sf::Text nameText, hpText;
	};
	
	enum class ScenarioVisitType
	{
		Normal, // adds the vertex to the scenarioStack
		Undo    // does not add, and also undoes the previous one (ie a back() or prev() scenario)
	};
	struct ScenarioConfig
	{
		std::string title;
		const Scenario *scenario;
		ScenarioVisitType visitType;
	};
	enum class ChoiceType
	{
		//! This choice corresponds to an external scenario file and clicking the ith one will execute the ith outcome
		Scenario,
		//! This choice is from a loliscript fork() call, and no scenario/outcome bullshit will run.
		ForkCall,
		//! This choice is from a item:query() call
		ItemSelect,
		//! Just a normal . . . click-through
		ClickThrough
	};
	
	/*		GUIWidget private methods		*/
	void onDraw(sf::RenderTarget& target) const override;

	void beforeDraw(sf::RenderTarget& target) const override;

	void onUpdate(int x, int y) override;
	/*		DialogueGUI private methods		*/
	void chooseOption(int index);

	void advanceDialogue();

	void addScenario(const std::pair<std::string, std::string>& link, ScenarioVisitType type = ScenarioVisitType::Normal);

	void attack(float strScale, float dexScale, float baseHitChance, float staminaCost, float staminaDamage, std::string successString, std::string missString);

	void addMessage(std::string str);

	void consumeMessage();

	void runScenario(const Scenario& scenario);

	void createOptionList();

	void createChoiceButtons();

	/**
	 * Halts message thread (only call this from message thread!) until the user chooses a fork choice.
	 * It uses forkChoices to fork on.
	 * @return The index of the choice the user clicked on
	 */
	int fork();

	/**
	 * Computes all links from the scenario/outcome and creates scenarios based on it.
	 */
	void computeLinks();

	void runAfterMessageThread();

	


	//! 
	ls::Bindings& context;
	//!
	const int buttonWidth;
	//! The current message being displayed on screen
	gui::TextLabel * const message;
	//! The buttons within the scroll area
	//gui::WidgetGrid * const optionList;
	//! The scrollable option area
	//! Area the options list should take up.
	const Rect<int> optionsRect;
	//gui::Scrollbar * scrollWindow;
	gui::GUIWidget *optionsList;
	std::vector<std::string> optionStrings;
	//! All the possible scenarios that the player could fork on at his next choice
	std::vector<ScenarioConfig> scenarios;
	//!
	std::vector<Line> lines;
	//!
	HPBar playerHPBar, enemyHPBar;
	//!
	PlayerStats& player;
	//!
	NPC& npcEntity;
	//!
	Soul& npc;
	//! Attributes of the NPC to match against for dialogue outcome selection
	std::map<std::string, Attribute> attributes;
	//! The loli's popup drawable widget.
	Popup *popup;
	ItemQuery itemQuery;
	//! The identifier of the active scenario
	std::string scenarioId;
	//! Current dialogue path someone has done (needed for back() link call)
	std::vector<std::string> scenarioStack;
	//! Temporary flag store. anything in temp:flag() is stored here.
	ls::Flags tempFlags;
	//! CV that the message thread waits on.
	std::condition_variable messageCond;
	//! CV that the main thread waits on.
	std::condition_variable mainCond;
	//! Mutex for messageCond's lock
	std::mutex messageMutex;
	//! Mutex for mainCond's lock
	std::mutex mainMutex;
	//! Thread for launching message block parsing
	std::thread messageThread;
	enum class ThreadRunning
	{
		Main,
		Message
	};
	//! Which thread should be executing (the two threads are executed in lock-step)
	volatile ThreadRunning threadRunning;
	volatile bool messageThreadDone;
	//! Storage for messages added via addMessage(). see comment in addMessage() for details on why it's needed
	std::string newMessage;
	//! Mutex for access to newMessage/messageThreadDone/threadRunning any other shared data.
	std::mutex communicationMutex;
	
	//! The current choice type, which impacts what happens when you select the ith choice
	ChoiceType currentChoiceType;
	//! Only use this from within startNewChoice/addChoice. it keeps track of how much height has been used so far for the choice area
	int choiceAreaHeightSofar;
	//! How many choices have been added so far
	int choiceIndex;
	//! <key, message>
	std::vector<std::pair<std::string, std::string>> forkChoices;
	volatile int forkChoice;
	//! A check to make sure people don't addForkChoice() without a following fork() call (forkChoices.clear() in fork() breaks stuff - too lazy to debug right now)
	bool forkChoicesUsed;
	//! The currently executing Outcome. can be null if we're not executing one.
	const Outcome *currentOutcome;
	//! The currently executing Scenario
	const Scenario *currentScenario;
	//! This is to work around std::function's move opertors not moving their wrapped lambdas.
	std::vector<MoveableFunction<void>> toRunAfterMessageThread;
	//! Sex acts that occured in dialogue before a sex-stop-point
	std::set<int> currentSexActs;
	int currentCrimeId; // HACK DEMO ?!?!?
};

} // sl

#endif // SL_DIALOGUEGUI_HPP
