#include "NPC/Dialogue/Outcome.hpp"

#include <algorithm>
#include <cmath>

#include "Parsing/Binding/BoundNumber.hpp"
#include "Utility/Assert.hpp"
#include "Utility/Memory.hpp"


namespace sl
{

Outcome::Outcome(std::string id, std::unique_ptr<const ls::Statement> message, std::unique_ptr<const ls::Statement> result, std::unique_ptr<const ls::Statement> links, std::vector<std::pair<std::string, AttributeRange>>&& attributes, std::unique_ptr<const ls::Statement> availability, std::vector<std::unique_ptr<const Outcome>> outcomes, std::unique_ptr<const ls::Expression> priority)
	:id(std::move(id))
	,message(std::move(message))
	,result(std::move(result))
	,links(std::move(links))
	,availability(std::move(availability))
	,outcomes(std::move(outcomes))
	,priority(std::move(priority))
{
	for (std::pair<std::string, AttributeRange>& attribute : attributes)
	{
		this->attributes.insert(std::make_pair(std::move(attribute.first), std::move(attribute.second)));
	}
}

void Outcome::execute(ls::Bindings& context) const
{
	if (result)
	{
		try
		{
			result->execute(context);
		}
		catch (const ls::RuntimeError& riee)
		{
			ERROR(riee.what());
		}
	}
}

std::shared_ptr<const ls::Statement> Outcome::getMessages() const
{
	return message;
}

void Outcome::getLinks(ls::Bindings& context) const
{
	// this is now optional because we can have scenario-wide links
	if (links)
	{
		try
		{
			links->execute(context);
		}
		catch (const ls::RuntimeError& riee)
		{
			ERROR(riee.what());
		}
	}
}

void Outcome::getAttributeNames(std::vector<std::string>& output) const
{
	for (const auto& p : attributes)
	{
		output.push_back(p.first);
	}
}

bool Outcome::isAvailable(ls::Bindings& context) const
{
	if (availability)
	{
		bool allow = false;
		context.registerVar("outcome", "allow", unique<ls::BoundNumber<bool>>(allow));
		availability->execute(context);
		context.unregisterVar("outcome", "allow");
		return allow;
	}
	return true;
}

int Outcome::getPriority(ls::Bindings& context) const
{
	if (priority)
	{
		return priority->evaluate(context).asInt();
	}
	return 0;
}

float Outcome::calculateDistance(const ls::Bindings& context, const std::map<std::string, Attribute>& npcAttributes, const std::vector<std::string>& selectedAttributes) const
{
	if (selectedAttributes.empty())
	{
		return 0.f;
	}

	float distance = 0.f;
	float scalar = 0.f;

	for (const std::string& attribute : selectedAttributes)
	{
		const auto it = attributes.find(attribute);
		const float diff = it != attributes.end() ?
			it->second.calculateDifference(context, npcAttributes.at(attribute).getValue()) :
			npcAttributes.at(attribute).getExpectedDifference();
		distance += diff * diff;
		scalar += it != attributes.end() ? (it->second.getScalar() * it->second.getScalar()) : 1.f;
	}
	// compute the scalar that we need to multiply the result by to make up for the custom scalar modifiers
	scalar /= selectedAttributes.size();
	scalar = 1.f / scalar;
	return sqrt(scalar * distance);
}

const std::vector<std::unique_ptr<const Outcome>>& Outcome::getOutcomes() const
{
	return outcomes;
}

const std::string& Outcome::getIdentifier() const
{
	return id;
}

void Outcome::lint(ls::LintBindings& context) const
{
	context.addVariable("outcome", "allow");
	if (message)
	{
		message->lint(context);
	}
	if (result)
	{
		result->lint(context);
	}
	if (links)
	{
		links->lint(context);
	}
	if (availability)
	{
		availability->lint(context);
	}
	for (const auto& outcome : outcomes)
	{
		outcome->lint(context);
	}
	if (priority)
	{
		priority->lint(context);
	}
}

} // sl
