#include "NPC/Dialogue/Popup.hpp"

#include "Engine/Settings.hpp"
#include "Engine/GUI/DrawableContainer.hpp"
#include "NPC/LayeredSpriteLoader.hpp"

#include "Engine/Graphics/SFDrawable.hpp"
namespace sl
{

Popup::Popup(const Vector2i& topLeft, const NPCAppearance& npc)
	:gui::GUIWidget(Rect<int>(topLeft.x, topLeft.y, WIDTH, HEIGHT))
	,npc(npc)
	,framerateSpec()
	,repeats(true)
	,timer(-1)
	,isAnimating(false)
	,fadeProgress(1.f)
	,fadeSpeed(0.1f)
{
	initRenderingSetup();
}

void Popup::registerFunctions(ls::Bindings& context, const std::string& target)
{
	context.registerFunction(target, "setPopup", [&](std::string category, ls::OptionalParam<std::string> action, ls::OptionalParam<std::string> face, ls::OptionalParam<bool> preservesProgress)
	{
		set(std::move(category), action.get("default"), face.get("default"), FramerateSpec(), true, preservesProgress.get(false));
	});

	context.registerFunction(target, "setPopupAction", [&](std::string newAction, ls::OptionalParam<bool> preservesProgress)
	{
		action = std::move(newAction);
		reload(preservesProgress.get(false));
	});

	context.registerFunction(target, "setPopupFace", [&](std::string newFace, ls::OptionalParam<bool> preservesProgress)
	{
		face = std::move(newFace);
		reload(preservesProgress.get(false));
	});

	context.registerFunction(target, "playPopup", [&](std::string category, ls::OptionalParam<std::string> action, ls::OptionalParam<std::string> face, ls::OptionalParam<int> duration, ls::OptionalParam<bool> preservesProgress)
	{
		load(category, action.get("default"), face.get("default"), FramerateSpec(), true, preservesProgress.get(false));
		timer = duration.get(60);
	});

	context.registerFunction(target, "playPopupFace", [&](std::string face, ls::OptionalParam<int> duration, ls::OptionalParam<bool> preservesProgress)
	{
		load(category, action, face, FramerateSpec(), true, preservesProgress.get(false));
		timer = duration.get(60);
	});

	context.registerFunction(target, "getPopupCategory", [&]
	{
		return category;
	});

	context.registerFunction(target, "getPopupAction", [&]
	{
		return action;
	});

	context.registerFunction(target, "getPopupFace", [&]
	{
		return face;
	});

	context.registerFunction(target, "animatePopup", [&](ls::Value spec, bool repeats, ls::OptionalParam<float> speed) {
		set(
			category,
			action,
			face,
			spec.getType() == ls::Value::Type::String
				? FramerateSpec(spec.asString())
				: FramerateSpec(Framerate(spec.asInt())),
			repeats,
			true);
		if (showing())
		{
			container->getDrawable()->setAnimationSpeed(speed.get(1.f));
		}
	});
	context.registerFunction(target, "playAnimation", [&](std::string category, std::string action, std::string face, ls::Value spec, ls::OptionalParam<float> speed) {
		load(
			category,
			action,
			face,
			spec.getType() == ls::Value::Type::String
			? FramerateSpec(spec.asString())
			: FramerateSpec(Framerate(spec.asInt())),
			false,
			false);
		if (showing())
		{
			container->getDrawable()->setAnimationSpeed(speed.get(1.f));
			isAnimating = true;
		}
	});
}

void Popup::load(const std::string& category, const std::string& action, const std::string& face, FramerateSpec framerateSpec, bool repeats, bool preservesProgress)
{
	if (showing())
	{
		const float progress = preservesProgress && current ? current->getProgress() : 0.f;
		const bool shouldFadeTransition = fadeProgress != 0.f;
		if (shouldFadeTransition)
		{
			std::swap(current, fadeout);
			fadeProgress = preservesProgress ? 1.f : 0.f;
		}
		current = LayeredSpriteLoader::createPortrait(
			LayeredSpriteLoader::Config(npc, category, WIDTH).animate(std::move(framerateSpec), repeats),
			action.c_str(),
			face.c_str());
		current->setProgress(progress);
	}
	else
	{
		container->setDrawable(nullptr);
		current.reset();
		fadeout.reset();
	}
}

void Popup::set(std::string category, std::string action, std::string face, FramerateSpec framerateSpec, bool repeats, bool preservesProgress)
{
	this->category = std::move(category);
	this->action = std::move(action);
	this->face = std::move(face);
	this->framerateSpec = std::move(framerateSpec);
	this->repeats = repeats;
	reload(preservesProgress);
}

void Popup::reload(bool preservesProgress)
{
	load(category, action, face, framerateSpec, repeats, preservesProgress);
}


// private
void Popup::onUpdate(int x, int y)
{
	const int xdif = x - box.left;
	const int ydif = y - box.top;

	if (current)
	{
		// not necessary I think?
		// drawable->update(xdif, ydif);
		finalTexture.clear(sf::Color::Transparent);
		const bool handleFading = fadeProgress < 1.f;
		if (handleFading)
		{
			fadeProgress = std::min(fadeProgress + 0.05f, 1.f);
			const float drawAlpha = fadeProgress;//std::max(0.f, fadeProgress - 0.5f) * 2;
			const float fadeAlpha = 1.f - fadeProgress;//std::max(0.f, 0.5f - fadeProgress)
			// could not exist if it's the first time the popup came into being.
			// although, do we want to fade in in that case?
			if (fadeout)
			{
				bufferTexture.clear(sf::Color::Transparent);
				fadeout->draw(bufferTexture);
				bufferTexture.display();
				bufferSprite.setColor(sf::Color(255 * fadeAlpha, 255 * fadeAlpha, 255 * fadeAlpha, 255));
				finalTexture.draw(bufferSprite);
			}
			bufferSprite.setColor(sf::Color(255, 255, 255, drawAlpha * 255));
		}

		bufferTexture.clear(sf::Color::Transparent);
		current->draw(bufferTexture);
		bufferTexture.display();
		if (handleFading)
		{
			finalTexture.draw(bufferSprite, sf::BlendAdd);
		}
		else
		{
			current->update(0, 0);
			finalTexture.draw(bufferSprite);
		}
		finalTexture.display();
	}

	if (timer > 0 && --timer == 0)
	{
		reload(false);
	}

	if (isAnimating && (!container || container->getDrawable()->isDone()))
	{
		reload(false);
		isAnimating = false;
	}
}

void Popup::initRenderingSetup()
{
	auto widget = unique<gui::DrawableContainer>(box);
	container = widget.get();
	addWidget(std::move(widget));
	ASSERT(finalTexture.create(WIDTH, HEIGHT));
	auto spr = std::make_unique<SFDrawable<sf::Sprite>>();
	spr->getData().setTexture(finalTexture.getTexture());
	container->setDrawable(std::move(spr));
	ASSERT(bufferTexture.create(WIDTH, HEIGHT));
	bufferSprite.setTexture(bufferTexture.getTexture());
}

bool Popup::showing() const
{
	return Settings::fuckYouSingletonHack()->get("portraits_visible");
}

} // sl
