#include "NPC/Dialogue/Scenario.hpp"

#include <algorithm>

#include "NPC/Dialogue/ItemQuery.hpp"
#include "Parsing/Binding/BoundNumber.hpp"
#include "Parsing/Binding/BoundString.hpp"
#include "Utility/Assert.hpp"
#include "Utility/Logger.hpp"
#include "Utility/Memory.hpp"
#include "Utility/Random.hpp"

namespace sl
{

Scenario::Scenario(const std::string& identifier, std::unique_ptr<const ls::Statement> title, std::unique_ptr<const ls::Statement> availability, std::unique_ptr<const ls::Statement> links, std::vector<std::unique_ptr<const Outcome>>& outcomes)
	:title(std::move(title))
	,availability(std::move(availability))
	,links(std::move(links))
	,outcomes(std::move(outcomes))
	,identifier(identifier)
{
	ASSERT(!this->outcomes.empty());
}

const Outcome& Scenario::chooseOutcome(ls::Bindings& context, const std::map<std::string, Attribute>& npcAttributes) const
{
	return chooseOutcomeImpl(context, npcAttributes, outcomes);
}

std::string Scenario::getTitle(ls::Bindings& context) const
{
	std::string buffer("<unset title>");
	if (title)
	{
		context.registerVar("scenario", "title", unique<ls::BoundString>(buffer));
		title->execute(context);
	}
	return buffer;
}

bool Scenario::isAvailable(ls::Bindings& context) const
{
	if (availability)
	{
		bool allow = false;
		context.registerVar("scenario", "allow", unique<ls::BoundNumber<bool>>(allow));
		availability->execute(context);
		context.unregisterVar("scenario", "allow");
		return allow;
	}
	return true;
}

const std::string& Scenario::getIdentifier() const
{
	return identifier;
}


void Scenario::getLinks(const Outcome& outcome, ls::Bindings& context, ItemQuery& itemQuery, std::vector<std::pair<std::string, std::string>>& links) const
{
	bool outcomeFound = false;
	for (const auto& o : outcomes)
	{
		if (o.get() == &outcome)
		{
			outcomeFound = true;
			break;
		}
	}
	ASSERT(outcomeFound);
	links.clear();
	// make sure you keep these up to date with the linter
	context.registerFunction("local", "link", [&links](const ls::Arguments& args)
	{
		VERIFY_ARG_COUNT(args, 1, 2);
		const std::string fname = args[0].asString();//"resources/dlg/" + args[0].asString() + ".dlg";
		links.push_back(std::make_pair(fname, args.count() == 1 ? "" : args[1].asString()));
		Logger::log(Logger::Parsing, Logger::Debug) << "\n\n\nAdding in " << fname << "\n\n\n";
	});
	context.registerFunction("local", "back", [&context, &links](const ls::Arguments& args)
	{
		VERIFY_ARG_COUNT(args, 0, 1);
		links.push_back(std::make_pair("?BACK?", args.count() == 0 ? "Back." : args[0].asString()));
		Logger::log(Logger::Parsing, Logger::Debug) << "\n\nAdding in back()\n";
	});
	context.registerFunction("local", "prev", [&links]
	{
		links.push_back(std::make_pair("?PREV?", ""));
		Logger::log(Logger::Parsing, Logger::Debug) << "\n\nAdding in prev()\n";
	});
	context.registerFunction("query", "orFilter", [&itemQuery](const ls::Arguments& args)
	{
		for (int i = 0; i < args.count(); ++i)
		{
			itemQuery.addOrFilter(Item::strToCategory(args[i].asString()));
		}
	});
	context.registerFunction("query", "andFilter", [&itemQuery](const ls::Arguments& args)
	{
		for (int i = 0; i < args.count(); ++i)
		{
			itemQuery.addAndFilter(Item::strToCategory(args[i].asString()));
		}
	});
	context.registerFunction("query", "finish", [&itemQuery, &links](std::string inventoryName, std::string fname)
	{
		itemQuery.queryInventory(inventoryName);
		links.push_back(std::make_pair(fname, ""));
	});
	if (this->links)
	{
		this->links->execute(context);
	}
	outcome.getLinks(context);
}

void Scenario::lint(ls::LintBindings& context) const
{
	// TODO: add in temporary binding to ls::Bindings so these aren't available everywhere - just in links blocks.
	context.addFunction("local", "link");
	context.addFunction("local", "back");
	context.addFunction("local", "prev");
	context.addFunction("query", "orFilter");
	context.addFunction("query", "andFilter");
	context.addFunction("query", "finish");
	context.addVariable("scenario", "allow");
	context.addVariable("scenario", "title");
	if (title)
	{
		title->lint(context);
	}
	if (availability)
	{
		availability->lint(context);
	}
	if (links)
	{
		links->lint(context);
	}
	for (const auto& outcome : outcomes)
	{
		outcome->lint(context);
	}
}

// private

/*static*/ const Outcome& Scenario::chooseOutcomeImpl(ls::Bindings& context, const std::map<std::string, Attribute>& npcAttributes, const std::vector<std::unique_ptr<const Outcome>>& outcomes)
{
	std::vector<std::string> selectedAttributes;
	for (const auto& outcome : outcomes)
	{
		outcome->getAttributeNames(selectedAttributes);
	}

	// remove duplicates since some outcomes could be matching on the same attribute(s)
	std::sort(selectedAttributes.begin(), selectedAttributes.end());
	const auto end = std::unique(selectedAttributes.begin(), selectedAttributes.end());
	selectedAttributes.resize(end - selectedAttributes.begin());


	float lowestDistance;
	int highestPriority;
	const Outcome *lowest = nullptr;
	for (const auto& outcome : outcomes)
	{
		if (outcome->isAvailable(context))
		{
			// add in some random variance to break ties (additive too in case of 0-distance)
			const float distance = Random::get().randomFloat(0.1f) + outcome->calculateDistance(context, npcAttributes, selectedAttributes) * Random::get().randomFloat(0.9f, 1.1f);
			const int priority = outcome->getPriority(context);
			if (lowest == nullptr ||
				priority > highestPriority ||
				(priority == highestPriority && distance < lowestDistance))
			{
				lowestDistance = distance;
				highestPriority = priority;
				lowest = outcome.get();
			}
		}
	}
	ASSERT(lowest);
	if (!lowest->getOutcomes().empty())
	{
		return chooseOutcomeImpl(context, npcAttributes, lowest->getOutcomes());
	}
	return *lowest;
}

} // sl