#ifndef SL_GROUPAI_HPP
#define SL_GROUPAI_HPP

#include <set>
#include <vector>

#include "Engine/Events/Event.hpp"
#include "Utility/Vector.hpp"

namespace sl
{

class NPC;

class IGroupAI
{
public:
	virtual ~IGroupAI() {}

	virtual void update() = 0;

	virtual bool isMember(NPC& npc) const = 0;
	
	virtual const Vector2f& getPos() const = 0;
	
	virtual float getRadius() const = 0;

	virtual void broadcastEvent(Event&& ev) = 0;
};

template <typename Activity>
class GroupAI : public IGroupAI
{
public:
	class Membership
	{
	public:
		Membership();

		Membership(NPC& npc);

		Membership(Activity& activity, NPC& npc);

		Membership(Membership&& other);

		Membership& operator=(Membership&& rhs);


		
		~Membership();


		void leave();

		Activity* getActivity();

		const Activity* getActivity() const;

		Event* getNewEvents();

	private:
		Membership(const Membership&) = delete;
		Membership& operator=(const Membership&) = delete;

		Activity *activity;
		NPC *npc;
	};
	
	virtual typename GroupAI<Activity>::Membership addParticipant(NPC& participant) = 0;

	virtual bool isPersistent() const = 0;
	
	bool isMember(NPC& npc) const override;
	
	const Vector2f& getPos() const override;
	
	float getRadius() const override;

	int getParticipantCount() const;

	const std::vector<NPC*>& getParticipants() const;

	void broadcastEvent(Event&& ev);

	void update();

protected:
	GroupAI() {}

	virtual void onUpdate() {}
	
	Vector2f pos;
	float radius;
	std::vector<NPC*> participants;

private:
	GroupAI(const GroupAI<Activity>&);
	GroupAI<Activity>& operator=(const GroupAI<Activity>&);

	void removeParticipant(NPC *npc);

	/* Originally std::pair<std::pair<int,Event>, std::set<NPC*>>
	<<TimeLeftInFrames,Event>, set<ParticipantHasAcknowledged>>
	*/
	struct ActiveEvent
	{
		ActiveEvent(int timeleft, Event &&event, std::set<NPC*> participants)
			:timeleft(timeleft), event(std::move(event)), participants(std::move(participants))
		{
		}
		int timeleft;
		Event event;
		std::set<NPC*> participants;
	};
	std::vector<ActiveEvent> events;

	friend class Membership;
};

template <typename Activity>
bool GroupAI<Activity>::isMember(NPC& npc) const
{
	for (NPC *participant : participants)
	{
		if (participant == &npc)
		{
			return true;
		}
	}
	return false;
}

template <typename Activity>
const Vector2f& GroupAI<Activity>::getPos() const
{
	return pos;
}

template <typename Activity>
float GroupAI<Activity>::getRadius() const
{
	return radius;
}

template <typename Activity>
int GroupAI<Activity>::getParticipantCount() const
{
	return participants.size();
}

template <typename Activity>
const std::vector<NPC*>& GroupAI<Activity>::getParticipants() const
{
	return participants;
}

template <typename Activity>
void GroupAI<Activity>::removeParticipant(NPC* npc)
{
	for (NPC*& participant : participants)
	{
		if (npc == participant)
		{
			//participant = participants.back();
			participants.pop_back();
			return;
		}
	}
}

template <typename Activity>
void GroupAI<Activity>::broadcastEvent(Event &&ev)
{
	//std::pair<int, Event> why(123, std::move(ev));
	//events.push_back(std::pair<std::pair<int, Event>>(), std::set<NPC*>());
	events.emplace_back(123, std::move(ev), std::set<NPC*>());
	//events.push_back({std::pair<int,Event>(123, std::move(ev)), std::set<NPC*>()});
}

template <typename Activity>
void GroupAI<Activity>::update()
{
	for (unsigned int i = 0; i < events.size(); ++i)
	{
		if (--events[i].timeleft <= 0)
		{
			events[i] = std::move(events.back());
			events.pop_back();
			--i;
		}
	}
	onUpdate();
}

//-----------------------------------------------------------------------------
//---------------------------- GroupAI::Membership ----------------------------

template <typename Activity>
GroupAI<Activity>::Membership::Membership()
	:activity(nullptr)
	,npc(nullptr)
{
}

template <typename Activity>
GroupAI<Activity>::Membership::Membership(NPC& npc)
	:activity(nullptr)
	,npc(&npc)
{
}

template <typename Activity>
GroupAI<Activity>::Membership::Membership(Activity& activity, NPC& npc)
	:activity(&activity)
	,npc(&npc)
{
}

template <typename Activity>
GroupAI<Activity>::Membership::Membership(Membership&& other)
	:activity(other.activity)
	,npc(other.npc)
{
	other.activity = nullptr;
	other.npc = nullptr;
}

template <typename Activity>
typename GroupAI<Activity>::Membership& GroupAI<Activity>::Membership::operator=(Membership&& rhs)
{
	leave();
	activity = rhs.activity;
	npc = rhs.npc;
	rhs.activity = nullptr;
	rhs.npc = nullptr;
	return *this;
}

template <typename Activity>
GroupAI<Activity>::Membership::~Membership()
{
	leave();
}


template <typename Activity>
void GroupAI<Activity>::Membership::leave()
{
	if (activity)
	{
		activity->removeParticipant(npc);
	}
}

template <typename Activity>
Activity* GroupAI<Activity>::Membership::getActivity()
{
	return activity;
}

template <typename Activity>
const Activity* GroupAI<Activity>::Membership::getActivity() const
{
	return activity;
}

template <typename Activity>
Event* GroupAI<Activity>::Membership::getNewEvents()
{
	for (auto& ev : events)
	{
		if (ev.participants.insert(npc).second)
		{
			return &ev.event;
		}
	}
	return nullptr;
}



} // sl

#endif // SL_GROUPAI_HPP
