#include "NPC/LayeredSpriteLoader.hpp"

#include "Engine/Graphics/SpriteAnimation.hpp"
#include "Engine/Graphics/ColorSwapID.hpp"
#include "Engine/Graphics/DrawableCollection.hpp"
#include "Engine/Graphics/Sprite.hpp"
#include "NPC/ClothesDatabase.hpp"
#include "NPC/NPCAppearance.hpp"
#include "Engine/Entity.hpp"
#include "Utility/Assert.hpp"
#include "Utility/JsonUtil.hpp"
#include "Utility/Memory.hpp"

#include <filesystem>

namespace sl
{

namespace LayeredSpriteLoader
{

// Config
Config::Config(const NPCAppearance& npc, const std::string& animName, int frameWidth, Vector2i offset)
	:npc(npc)
	,animName(animName)
	,frameWidth(frameWidth)
	,offset(offset)
	,layerMask(npc.clothingLayers.toFlag()/* | ClothingLayers::Body*/)
	,angle(0.f)
	,framerateSpec()
	,repeats(repeats)
{
}

Config& Config::animate(Framerate framerate, bool repeats)
{
	return animate(FramerateSpec(std::move(framerate)), repeats);
}

Config& Config::animate(FramerateSpec framerateSpec, bool repeats)
{
	this->framerateSpec = std::move(framerateSpec);
	this->repeats = repeats;
	return *this;
}

struct DirMaskPair
{
	DirMask mask;
	const char *actionName;
	const char *fileName;
	bool mirror;
};

static const DirMaskPair directions[4] = {
	{East,  "east",  "side", false},
	{North, "north", "up",   false},
	{West,  "west",  "side", true },
	{South, "south", "down", false}
};

struct SwapLayer
{
	NPCSpriteLayer drawnLayer;
	std::string path;
	ColorSwapID swap;
	uint8_t clothingLayers;
	const char *face = nullptr;
};

void load(Entity& npc, const Config& config, int dirMask)
{
	for (const DirMaskPair& dir : directions)
	{
		if (dir.mask & dirMask)
		{
			npc.clearDrawablesInSet(dir.actionName);
			npc.addDrawableToSet(dir.actionName, create(config, dir.fileName, dir.mirror));
		}
	}
}

void load(Entity& npc, const std::initializer_list<Config>& configs, int dirMask)
{
	for (const DirMaskPair& dir : directions)
	{
		if (dir.mask & dirMask)
		{
			npc.clearDrawablesInSet(dir.actionName);
			auto sprite = unique<DrawableCollection>();
			for (const Config& config : configs)
			{
				sprite->addDrawable(create(config, dir.fileName, dir.mirror));
			}
			npc.addDrawableToSet(dir.actionName, std::move(sprite));
		}
	}
}

std::unique_ptr<Drawable> create(const Config& config, const char *action, bool mirrored)
{
	const SwapLayer layers[] = {
		{ NPCSpriteLayer::HairBack, "hair/" + config.npc.getHair() + "/back", config.npc.getHairSwapID(), ClothingLayers::Body },
		{ NPCSpriteLayer::Body, "body", config.npc.getBodySwapID(), ClothingLayers::Body, }, // nude body
		{ NPCSpriteLayer::Panties, "panties/" + config.npc.getClothing(ClothingType::Panties), config.npc.getClothingSwapID(ClothingType::Panties), ClothingLayers::Panties },
		{ NPCSpriteLayer::Bra, "bra/" + config.npc.getClothing(ClothingType::Bra), config.npc.getClothingSwapID(ClothingType::Bra), ClothingLayers::Bra },
		{ NPCSpriteLayer::Outfit, "outfit/" + config.npc.getClothing(ClothingType::Outfit), config.npc.getClothingSwapID(ClothingType::Outfit), ClothingLayers::Outfit },
		{ NPCSpriteLayer::HairFront, "hair/" + config.npc.getHair() + "/front", config.npc.getHairSwapID(), ClothingLayers::Body },
		{ NPCSpriteLayer::Accessory, "accessory/" + config.npc.getClothing(ClothingType::Accessory), config.npc.getClothingSwapID(ClothingType::Outfit), ClothingLayers::Accessory }
	};
	std::unique_ptr<DrawableCollection> sprite = unique<DrawableCollection>();
	for (const SwapLayer& layer : layers)
	{
		if (layer.clothingLayers & config.layerMask)
		{
			std::stringstream fname;
			fname << config.npc.getResourcePath() << "/" << config.animName << "/" << layer.path << "/" << action << ".png";
			if (std::filesystem::exists(TextureManager::getDirectory() + fname.str()))
			{
				auto anim = unique<SpriteAnimation>(fname.str().c_str(), config.frameWidth, config.framerateSpec.get(layer.drawnLayer), config.repeats, layer.swap, config.offset.x, config.offset.y, mirrored);
				if (config.angle)
				{
					anim->rotate(config.angle);
				}
				sprite->addDrawable(std::move(anim));
			}
		}
	}
	return std::move(sprite);
}

// TODO: move or refactor and pass into createPortrait as a parameter?
static const std::string* baseActionOverride(const std::string& baseAction, NPCSpriteLayer layer)
{
	// This isn't used right now, but will likely be in the future.
	// @HACK #DEMO externalize or even remove this in facor of a better system.
	return nullptr;
}

static const std::string* clothingActionOverride(const char *baseAction, NPCSpriteLayer drawnLayer, const NPCAppearance& npc)
{
	const std::string *ret = nullptr;
	//for (int i = 0; i < CLOTHING_TYPE_COUNT; ++i)
	static const ClothingType supported[] = { ClothingType::Outfit, ClothingType::Panties };
	for (const ClothingType type : supported)
	{
		if (npc.clothingLayers.fromEnum(type))
		{
			//const ClothingType type = static_cast<ClothingType>(i);
			const auto& animOverrides = ClothesDatabase::get().getInfo(npc.getClothing(type)).animOverrides;
			auto it = animOverrides.find(std::make_pair(baseAction, drawnLayer));
			if (it != animOverrides.end())
			{
				if (ret)
				{
					ERROR("conflicting action overrides");
				}
				ret = &it->second;
			}
		}
	}
	return ret;
}

static std::unique_ptr<SpriteAnimation> createLayer(const Config& config, const char *action, bool mirrored, const SwapLayer& layer, const std::string& resourcePath, const Vector2i& origin)
{
	// try <action>.png and if it fails, try default.png, and if it fails, show nothing
	const char *actionsToTry[2] = {
		action,
		"default",
	};
	const std::vector<std::string> facesToTry =
		layer.face == nullptr  ?
	std::vector<std::string>({
		""
	}) : std::vector<std::string>({
		std::string("/")  + layer.face,
		"/default"
	});
	for (const char *tryAction : actionsToTry)
	{
		for (const std::string& face : facesToTry)
		{
			std::stringstream fname;
			fname << resourcePath << "/popups/" << config.animName << "/" << layer.path << "/" << tryAction << face << ".png";
			if (std::filesystem::exists(TextureManager::getDirectory() + fname.str()))
			{
				auto anim = unique<SpriteAnimation>(fname.str().c_str(), config.frameWidth, config.framerateSpec.get(layer.drawnLayer), config.repeats, layer.swap, config.offset.x, config.offset.y, mirrored);
				anim->setOrigin(origin);
				if (config.angle)
				{
					anim->rotate(config.angle);
				}
				return anim;
			}
		}
	}
	return nullptr;
}

std::unique_ptr<Drawable> createPortrait(const Config& config, const char *action, const char *face, bool mirrored)
{
	std::vector<SwapLayer> layers = {
		{ NPCSpriteLayer::HairBack, "hair/" + config.npc.getHair() + "/back", config.npc.getHairSwapID(), ClothingLayers::Body },
		{ NPCSpriteLayer::Body, "body", config.npc.getBodySwapID(), ClothingLayers::Body, }, // nude body
		{ NPCSpriteLayer::Face, "face"/* + config.npc.getFace()*/, config.npc.getBodySwapID() | config.npc.getHairSwapID(), ClothingLayers::Body, face },
		{ NPCSpriteLayer::Panties, "panties/" + config.npc.getClothing(ClothingType::Panties), config.npc.getClothingSwapID(ClothingType::Panties), ClothingLayers::Panties },
		{ NPCSpriteLayer::Bra, "bra/" + config.npc.getClothing(ClothingType::Bra), config.npc.getClothingSwapID(ClothingType::Bra), ClothingLayers::Bra },
		{ NPCSpriteLayer::Socks, "socks/" + config.npc.getClothing(ClothingType::Socks), config.npc.getClothingSwapID(ClothingType::Socks), ClothingLayers::Socks },
		// shoes ?
		{ NPCSpriteLayer::Outfit, "outfit/" + config.npc.getClothing(ClothingType::Outfit), config.npc.getClothingSwapID(ClothingType::Outfit), ClothingLayers::Outfit },
		{ NPCSpriteLayer::HairFront, "hair/" + config.npc.getHair() + "/front", config.npc.getHairSwapID(), ClothingLayers::Body },
		{ NPCSpriteLayer::Accessory, "accessory/" + config.npc.getClothing(ClothingType::Accessory), config.npc.getClothingSwapID(ClothingType::Accessory), ClothingLayers::Accessory }
	};
	for (const std::string& cumLayer : config.npc.getCumLayers())
	{
		layers.push_back({ NPCSpriteLayer::Unsupported, "cum/" + cumLayer, ColorSwapID(), ClothingLayers::Body});
	}
	// To reduce duplicated sprites that are just offsets of other ones between kinder/elem/middle (ie face, hair)
	// we let them store a JSON file of format { "path" : "middle/spread", xoff : 0, yoff : 16 } for example
	auto offsetInfo = tryReadJsonFile("resources/img/" + config.npc.getResourcePath() + "/popups/" + config.animName + "/offset.json");
	std::vector<std::pair<Vector2i, std::string>> offsetAndResources = {
		std::pair(Vector2i(0, 0), config.npc.getResourcePath())
	};
	if (!offsetInfo.is_null())
	{
		offsetAndResources.push_back(std::make_pair(
			Vector2i(
				-offsetInfo.object_items().at("xoff").int_value(),
				-offsetInfo.object_items().at("yoff").int_value()),
			offsetInfo.object_items().at("path").string_value()));
	}
	std::unique_ptr<DrawableCollection> sprite = unique<DrawableCollection>();
	for (const SwapLayer& layer : layers)
	{
		if (layer.clothingLayers & config.layerMask)
		{
			std::unique_ptr<Drawable> layerAnim;
			// allows overrides
			const std::string *over = clothingActionOverride(action, layer.drawnLayer, config.npc);
			// if it fails (sprites not there), try the offset for that action if it exists (added earlier)
			for (const std::pair<Vector2i, std::string>& offsetAndResource : offsetAndResources)
			{
				// try override
				if (!layerAnim && over)
				{
					layerAnim = createLayer(config, over->c_str(), mirrored, layer, offsetAndResource.second, offsetAndResource.first);
				}
				// try normal layer
				if (!layerAnim)
				{
					layerAnim = createLayer(config, action, mirrored, layer, offsetAndResource.second, offsetAndResource.first);
				}
			}
			if (layerAnim)
			{
				sprite->addDrawable(std::move(layerAnim));
			}
		}
	}
	return sprite;
}

} // LayeredSpriteLoader

} // sl
