#ifndef SL_LAYEREDPSRITELOADER_HPP
#define SL_LAYEREDPSRITELOADER_HPP

#include <memory>
#include <string>

#include "Engine/Graphics/Drawable.hpp"
#include "Engine/Graphics/Framerate.hpp"
#include "NPC/ClothingLayers.hpp"
#include "NPC/FramerateSpec.hpp"
#include "NPC/NPCSpriteLayers.hpp"

namespace sl
{

class Entity;
class NPCAppearance;

namespace LayeredSpriteLoader
{

class Config
{
public:
	/**
	 * @param appearance The NPCAppearance to construct sprites for
	 * @param animName The animation folder to use (ie 'walking', 'talking', 'spread')
	 * @param frameWidth How many pixels wide the frame is. This should be a multiple of the file's width. If it's >1 then this is an animation
	 * @param offset The offsets to shift the image by. (primarily used for overworld sprites)
	 */
	Config(const NPCAppearance& npc, const std::string& animName, int frameWidth, Vector2i offset = Vector2i(0, 0));

	Config& animate(Framerate framerate, bool repeats = true);

	Config& animate(FramerateSpec framerateSpec, bool repeats = true);

	const NPCAppearance& npc;
	std::string animName;
	int frameWidth;
	Vector2i offset;
	//! This is initalized based on Soul.clothingLayers, but you can override it after construction by changing this field.
	uint8_t layerMask;
	//! Defaults to 0. allows overriding for rotation.
	float angle;
	FramerateSpec framerateSpec;
	//! Defaults to true.
	bool repeats;
};

enum DirMask
{
	North = 1,
	East = 2,
	West = 4,
	South = 8,
	All = North | East | West | South
};

/**
 * Invokes create() on the directions given to load overworld sprites for an NPC
 * @param npc The entity to add the created drawables to (using "north", "south", "west", "east" as the keys)
 * @param config The configuration to use in create()
 * @param dirMask Which directions to use. Defaults to all 4
 */
void load(Entity& npc, const Config& config, int dirMask = All);

/**
 * Invokes create() on the directions given to load overworld sprites for an NPC
 * @param npc The entity to add the created drawables to (using "north", "south", "west", "east" as the keys)
 * @param config The configuration to use in create()
 * @param dirMask Which directions to use. Defaults to all 4
 */
void load(Entity& npc, const std::initializer_list<Config>& configs, int dirMask = All);

/**
 * Creates an overworld sprite for an NPC.
 * @param config config info for offsets/animation speeds/etc.
 * @param action The 'action' - for walking this is the direction (terminal level in resources) ie "west" for "west.png"
 * @param mirrored Whether to horizontally mirror the sprite (ie to re-use side.png for west/east)
 */
std::unique_ptr<Drawable> create(const Config& config, const char *action, bool mirrored = false);

/**
 * Creates a dialogue-appropriate big portrait sprite of an NPC
 * @config Config info for offsets/setc
 * @param action The terminal action (ie bj, default, etc) in the filename
 * @param face The face filename to use
 * @param mirrored Whether to horizontally mirror the sprite (ie to re-use side.png for west/east)
 */
std::unique_ptr<Drawable> createPortrait(const Config& config, const char *action, const char *face, bool mirrored = false);

} // LayeredSpriteLoader

} // sl

#endif // SL_LAYEREDPSRITELOADER_HPP