#include "NPC/LocationPathCache.hpp"

#include "Engine/Serialization/AutoDeserialize.hpp"
#include "Engine/Serialization/AutoSerialize.hpp"
#include "NPC/NPC.hpp"
#include "NPC/Stats/Soul.hpp"
#include "Pathfinding/NodePredicates/IsNode.hpp"
#include "Pathfinding/Heuristics.hpp"
#include "Pathfinding/PathManager.hpp"
#include "Utility/Logger.hpp"

namespace sl
{

std::shared_ptr<const pf::Path> LocationPathCache::getPath(const std::string& from, const std::string to, NPC& npc, pf::PathManager& pathManager, const pf::WeightMap& weights, float heuristicStrength) const
{
	const std::pair<std::string, std::string> key = std::make_pair(from, to);
	auto it = paths.find(key);

	if (it == paths.end())
	{
		// maybe return null here in the future? for now let's assert since I can't imagine why we would accept invalid locations
		const TargetLocator *fromTarget = npc.getSoul()->getLocation(from);
		ASSERT(fromTarget);
		const TargetLocator *toTarget = npc.getSoul()->getLocation(to);
		ASSERT(toTarget);
		const pf::NodeID fromNode = pathManager.getClosestNodeID(fromTarget->findTarget(npc.getBehaviorContext()));
		const pf::NodeID toNode = pathManager.getClosestNodeID(toTarget->findTarget(npc.getBehaviorContext()));
		std::shared_ptr<pf::Path> path = pathManager.getPath(pf::PathManager::PathSpec{ fromNode, pf::IsNode(toNode), &weights, pf::ManhattanDistance(toNode, heuristicStrength), nullptr }, pf::PathManager::Priority::Realtime);
		std::deque<pf::Node> reverseNodes(path->getNodes());
		std::reverse(reverseNodes.begin(), reverseNodes.end());
		it = paths.insert(std::make_pair(key, std::move(path))).first;
		paths.insert(std::make_pair(std::make_pair(to, from), std::make_shared<pf::Path>(std::move(reverseNodes))));
	}

	return it->second;
}

void LocationPathCache::deserialize(Deserialize& in)
{
	const int loaded = in.startArray("LocationPathCache");
	for (int i = 0; i < loaded; ++i)
	{
		auto key = AutoDeserialize::create<std::pair<std::string, std::string>>(in);
		auto nodes = AutoDeserialize::create<std::deque<pf::Node>>(in);
		auto it = paths.insert(
			std::make_pair(
				std::move(key),
				std::make_shared<pf::Path>(std::move(nodes))));
		Logger::log(Logger::Serialization, Logger::Useless) << "pathlen = " <<it.first->second->totalSceneDistance() << std::endl;
	}
	in.endArray("LocationPathCache");
}

void LocationPathCache::serialize(Serialize& out) const
{
	// since it's just a cache, we can be picky about what we save and only do ones we're 100% sure about.
	const int saved = std::count_if(paths.begin(), paths.end(), [](const auto& path) { return path.second->isCalculated(); });
	if (saved != paths.size())
	{
		Logger::log(Logger::Serialization, Logger::Debug) << "Could only save " << saved << " / " << paths.size() << std::endl;
	}
	out.startArray("LocationPathCache", saved);
	for (const auto& path : paths)
	{
		if (path.second->isCalculated())
		{
			AutoSerialize::serialize(out, path.first);
			AutoSerialize::serialize(out, path.second->getNodes());
			Logger::log(Logger::Serialization, Logger::Debug) << "pathlen = " << path.second->totalSceneDistance() << std::endl;
		}
	}
	out.endArray("LocationPathCache");
}

} // sl