#include "NPC/Locators/EntityLocatorCommon.hpp"

#include "Engine/Scene.hpp"
#include "Town/World.hpp"
#include "Utility/Assert.hpp"

#include <algorithm>
#include <queue>

namespace sl
{

std::vector<Target> findTargetsImpl(const Entity& sourceEntity, Entity::Type queryType, LocatorHeuristic heuristic, int radius, const std::function<bool(Entity*)>& predicate, int maxTargets)
{	
	// @todo implement the other Heuristics
	ASSERT(heuristic == LocatorHeuristic::ExactlyClosest);
	// this const-cast is mostly fine since we only need it for collision querying.
	Scene& rootScene = const_cast<Scene&>(*sourceEntity.getScene());
	const Vector2f sourceEntityInWorld = worldPosition2(rootScene, sourceEntity.getPos());
	const Rect<int> worldspaceRegion = radius == LOCATOR_UNBOUNDED_DISTANCE
	                                 ? worldRegion(rootScene)
	                                 : Rect<int>(
									       sourceEntityInWorld.x - radius,
								           sourceEntityInWorld.y - radius,
	                                       radius * 2,
	                                       radius * 2);

	std::vector<Scene*> scenesToCheck = { &rootScene };
	rootScene.getWorld()->getScenesWithin(scenesToCheck, worldspaceRegion);

	std::list<Entity*> entities;
	// @todo optimize by making it query in slowly increasing rect sizes until it reaches the max
	for (Scene *scene : scenesToCheck)
	{
		const Rect<int> worldspaceRegionInScene = radius == LOCATOR_UNBOUNDED_DISTANCE
		                                        ? Rect<int>(scene->getWidth(), scene->getHeight())
		                                        : worldPosToScenePos(*scene, worldspaceRegion);
		scene->cull(entities, worldspaceRegionInScene, queryType);
	}

	typedef std::pair<Entity*, float> CachedDist;
	auto cachedDistCompare = [](const CachedDist& a, const CachedDist& b) -> bool {
		return a.second < b.second;
	};
	std::priority_queue<CachedDist, std::vector<CachedDist>, decltype(cachedDistCompare)> heap(cachedDistCompare);
	for (Entity *entity : entities)
	{
		if (entity != &sourceEntity && (!predicate || predicate(entity)))
		{
			heap.push(std::make_pair(entity, pointDistance(sourceEntity.getPos(), entity->getPos())));
		}
	}
	const int maxPossibleTargets = maxTargets == LOCATOR_NO_MAX_TARGETS
	                             ? heap.size()
	                             : std::min<int>(maxTargets, heap.size());
	std::vector<Target> targets(maxPossibleTargets);
	for (int i = 0; i < maxPossibleTargets; ++i)
	{
		targets[i] = heap.top().first;
		heap.pop();
	}
	return targets;
}

} // sl
