#include "NPC/Locators/Target.hpp"

#include "Pathfinding/Path.hpp"
#include "Engine/Entity.hpp"
#include "Engine/Scene.hpp"
#include "Utility/Assert.hpp"

namespace sl
{

const int invalid = -1;

Target::Target()
	:mode(None)
	,other(nullptr)
	,scenePos()
	,path(nullptr)
{
}

Target::Target(const Target& other)
	:mode(other.mode)
	,other(other.other)
	,scenePos(other.scenePos)
	,path(other.path)
{
}

Target::Target(const ScenePos& scenePos)
	:mode(Coordinates)
	,other(nullptr)
	,scenePos(scenePos)
	,path(nullptr)
{
}

Target::Target(const Vector2f& pos, Scene *scene)
	:Target(ScenePos(*scene, pos))
{
}

Target::Target(int x, int y, Scene *scene)
	:Target(Vector2f(x, y), scene)
{
}

Target::Target(Entity *other)
	:mode(EntityFollow)
	,other(other)
	,scenePos()
	,path(nullptr)
{
}

Target::Target(const pf::Path *path)
	:mode(PathingResultFollow)
	,other(nullptr)
	,scenePos()
	,path(path)
{
}


void Target::set(const Vector2f& pos, Scene *scene)
{
	set(pos.x, pos.y, scene);
}

void Target::set(int x, int y, Scene *scene)
{
	mode = Coordinates;
	other = nullptr;
	scenePos = ScenePos(*scene, Vector2f(x, y));
	path = nullptr; 
}

void Target::set(Entity *other)
{
	mode = EntityFollow;
	this->other = other;
	scenePos = ScenePos(); // dynamically checks other instead
	path = nullptr;
}

void Target::set(const pf::Path *path)
{
	mode = PathingResultFollow;
	other = nullptr;
	scenePos = ScenePos(); // checks scene of last path node
	this->path = path;
}

void Target::unset()
{
	mode = None;
	other = nullptr;
	scenePos = ScenePos();
	path = nullptr;
}

bool Target::absorbPathData()
{
	if (mode == PathingResultFollow)
	{
		if (path->didPathingFail() || path->getNodes().empty())
		{
			unset();
			return false;
		}
		if (!isCalculating())
		{
			const pf::Node& lastNode = path->lastNode();
			set(lastNode.getX(), lastNode.getY(), lastNode.getScene());
			return true;
		}
	}
	return false;
}

int Target::getX() const
{
	switch (mode)
	{
		case None:
			return invalid;
		case Coordinates:
			return scenePos.getPos().x;
		case EntityFollow:
			return other->getPos().x;
		case PathingResultFollow:
			return path->lastNode().getX();
	}
	ERROR("stop complaining");
	return invalid;
}

int Target::getY() const
{
	switch (mode)
	{
		case None:
			return invalid;
		case Coordinates:
			return scenePos.getPos().y;
		case EntityFollow:
			return other->getPos().y;
		case PathingResultFollow:
			return path->lastNode().getY();
	}
	ERROR("stop complaining");
	return invalid;
}

Vector2<float> Target::getPos() const
{
	switch (mode)
	{
		case None:
			return Vector2<float>(invalid, invalid);
		case Coordinates:
			return scenePos.getPos();
		case EntityFollow:
			return other->getPos();
		case PathingResultFollow:
			return path->lastNode().getPos();
	}
	ERROR("stop complaining");
	return Vector2<float>(invalid, invalid);
}

Scene* Target::getScene() const
{
	switch (mode)
	{
	case Coordinates:
		return scenePos.getScene();
	case EntityFollow:
		return other->getScene();
	case PathingResultFollow:
		return path->lastNode().getScene();
	}
	return nullptr;
}

ScenePos Target::getScenePos() const
{
	return ScenePos(*getScene(), getPos());
}

Entity* Target::getEntity() const
{
	if (mode == EntityFollow)
	{
		return other;
	}
	return nullptr;
}

bool Target::isCalculating() const
{
	return mode == PathingResultFollow && !path->isCalculated() && !path->didPathingFail();
}

Target::operator bool() const
{
	switch (mode)
	{
	case Target::Mode::None:
		return false;
	case Target::Mode::Coordinates:
		return true;
	case Target::Mode::EntityFollow:
		return other;
	case Target::Mode::PathingResultFollow:
		return !isCalculating();
	}
	return false;
}

namespace AutoDeserialize
{
void deserialize(Deserialize& in, Target& target)
{
	switch (static_cast<Target::Mode>(in.u8("target.mode")))
	{
	case Target::Mode::None:
		target = Target();
		break;
	case Target::Mode::Coordinates:
		target = Target(AutoDeserialize::create<ScenePos>(in));
		break;
	case Target::Mode::EntityFollow:
		target = Target(static_cast<Entity*>(in.ptr("entity")));
		break;
	case Target::Mode::PathingResultFollow:
		ERROR("not supported");
	}
}
} // AutoDeserialize

namespace AutoSerialize
{
void serialize(Serialize& out, const Target& target)
{
	// TODO: I'm not sure how the fuck we're supposed to serialize paths.
	// It's possible we could stop the path thread mid-calculation,
	// then serialize all queued threads.
	// If it's already-computed then we can directly serialize I guess.
	switch (target.mode)
	{
	case Target::Mode::None:
		out.u8("target.mode", static_cast<uint8_t>(Target::Mode::None));
		break;
	case Target::Mode::Coordinates:
		out.u8("target.mode", static_cast<uint8_t>(Target::Mode::Coordinates));
		AutoSerialize::serialize(out, target.scenePos);
		break;
	case Target::Mode::EntityFollow:
		if (target.other)
		{
			out.u8("target.mode", static_cast<uint8_t>(Target::Mode::EntityFollow));
			out.ptr("entity", target.other.get());
		}
		else
		{
			out.u8("target.mode", static_cast<uint8_t>(Target::Mode::None));
		}
		break;
	case Target::Mode::PathingResultFollow:
		out.u8("target.mode", static_cast<uint8_t>(Target::Mode::None));
		break;
	}
}
} // AutoSerialize

} // sl
