#include "NPC/NPC.hpp"

#include <initializer_list>

#include "Engine/Scene.hpp"

#include "Engine/Graphics/SpriteAnimation.hpp"
#include "Engine/Graphics/Sprite.hpp"
#include "Engine/Settings.hpp"
#include "Engine/Serialization/Impl/SerializeSoul.hpp"
#include "Engine/Serialization/AutoDeserialize.hpp"
#include "Engine/Serialization/AutoSerialize.hpp"
#include "NPC/NPCState.hpp"
#include "NPC/States/IdleState.hpp"
#include "NPC/LayeredSpriteLoader.hpp"
#include "NPC/States/AdultWanderState.hpp"
#include "NPC/States/KidnappedState.hpp"
#include "NPC/States/LoliRoutineState.hpp"
#include "NPC/States/ToddlerState.hpp"
#include "NPC/States/Police/CopPatrolState.hpp"
#include "NPC/Stats/NationalityDatabase.hpp"
#include "NPC/NPCTextBox.hpp"
#include "Pathfinding/PathManager.hpp"
#include "Pathfinding/Heuristics.hpp"
#include "Pathfinding/NodePredicates/IsNode.hpp"
#include "Player/Player.hpp"
#include "Town/Cars/Car.hpp"
#include "Town/Cars/CarControls.hpp"
#include "Town/Town.hpp"
#include "Town/TownTiles.hpp"
#include "Town/World.hpp"
#include "Utility/Assert.hpp"
#include "Utility/Trig.hpp"
#include "Utility/Memory.hpp"
#include "Utility/Random.hpp"
#include "Utility/TypeToString.hpp"

namespace sl
{

DECLARE_TYPE_TO_STRING(NPC)

const int recalcThreshold = 128;

NPC::NPC(DeserializeConstructor)
	:Entity(0, 0, 16, 16, "NPC")
	,minimapIcon(Vector2f(-1.f, -1.f), sf::Color(255, 255, 128))
	,behaviorContext(*this)
	,currentWeightMap("uniform_weight")
	,walkSpeed(1.5f + Random::get().randomFloat(1.f))
	,runSpeed(walkSpeed * 2.f)
	,walkSpeedMultiplier(1.f)
#ifdef SL_DEBUG
	,debugDrawBehaviorTree(false)
#endif // SL_DEBUG
{
	mask.setOrigin(8, 8);
}

NPC::NPC(int x, int y, std::shared_ptr<Soul> soul)
	:Entity(x, y, 16, 16, "NPC")
	,soul(soul)
	,pathManager(nullptr)
	,mobilityState(State::OnFoot)
	,car(nullptr)
	,carControls(nullptr)
	,minimapIcon(pos, sf::Color(255, 255, 128))
	,behaviorContext(*this)
	,targetIsPathing(false)
	,currentWeightMap("uniform_weight")
	,walkSpeed(1.5f + Random::get().randomFloat(1.f))
	,runSpeed(walkSpeed * 2.f)
	,walkSpeedMultiplier(1.f)
	,facingDir(South)
#ifdef SL_DEBUG
	,debugDrawBehaviorTree(false)
#endif // SL_DEBUG
{
	mask.setOrigin(8, 8);


	{
		minimapIcon.setDetailedIcon(map::LOD::Minimap, std::make_unique<Sprite>("map/loli_icon_mm.png", 4, 4));
		minimapIcon.setDetailedIcon(map::LOD::Map, std::make_unique<Sprite>("map/loli_icon.png", 8, 8));
	}
}


const Types::Type NPC::getType() const
{
	return Entity::makeType("NPC");
}

void NPC::initStates(std::unique_ptr<NPCState> startingState, std::unique_ptr<NPCState> globalState)
{
	this->state.initStates(std::move(startingState), std::move(globalState));
}

void NPC::initDefaultState()
{
	if (soul->getType() == "loli")
	{
		if (soul->getAgeYears() <= 3.5f)
		{
			initStates(unique<ToddlerState>(*this));
		}
		else
		{
			if (soul->getFlags().get("captive").asBool())
			{
				initStates(unique<KidnappedState>(*this));
			}
			else
			{
				initStates(unique<LoliRoutineState>(*this));
			}
		}
	}
	else if (soul->getType() == "adultmale" || soul->getType() == "adultfemale")
	{
		initStates(unique<AdultWanderState>(*this));
	}
	else if (soul->getType() == "cop")
	{
		initStates(unique<CopPatrolState>(*this, car));
	}
	else
	{
		ERROR("unknown type: " + soul->getType());
	}
}

void NPC::setRoom(Room *room)
{
	this->room = room;
}

Room* NPC::getRoom() const
{
	return room;
}

void NPC::handleEvent(const Event& event)
{
	if (event.getCategory() == Event::Category::Debug)
	{
#ifdef SL_DEBUG
		if (event.getName() == "toggle debug btree" && pointDistance(getPos(), *event.getLocation()) <= 128.f)
		{
			debugDrawBehaviorTree = !debugDrawBehaviorTree;
		}
#endif // SL_DEBUG
	}
	state.handleEvent(event);
	behaviorContext.addEvent(Event::copy(event));
}

std::shared_ptr<Soul> NPC::getSoul()
{
	return soul;
}

const std::shared_ptr<const Soul> NPC::getSoul() const
{
	return soul;
}

void NPC::setDialogueStartOverride(std::string newStart)
{
	// make sure we're not erasing another override!
	ASSERT(dialogueStartOverride == newStart || dialogueStartOverride.empty());
	dialogueStartOverride = std::move(newStart);
}

const std::string* NPC::getDialogueStartOverride() const
{
	return dialogueStartOverride.empty() ? nullptr : &dialogueStartOverride;
}

void NPC::resetDialogueStartOverride()
{
	dialogueStartOverride.clear();
}

void NPC::setWalkSpeedMultiplier(float mult)
{
	ASSERT(mult >= 0.f);
	walkSpeedMultiplier = mult;
}

float NPC::getWalkSpeedMultiplier() const
{
	return walkSpeedMultiplier;
}

void NPC::setPathfindingWeightMap(std::string weightMap)
{
	currentWeightMap = std::move(weightMap);
}

const std::string& NPC::getCurrentPathfindingWegithMap() const
{
	return currentWeightMap;
}

bool NPC::canHear(const Vector2f& noisePos, float noiseRadius) const
{
	// @todo factor in background noise, NPC's hearing, etc?
	return (pointDistance(pos, noisePos) <= noiseRadius);
}

bool NPC::canSee(const Vector2f& sightPos) const
{
	static const Vector2f facingToUnit[4] = {
		Vector2f(1.f, 0.f),
		Vector2f(0.f, -1.f),
		Vector2f(-1.f, 0.f),
		Vector2f(0.f, 1.f)
	};
	const Vector2f toSight = (sightPos - pos).normalized();
	// since unit facing dotted with unit vector to the position gives cos(theta)
	// then we can use that to check if it's behind us easily, and also get a nice vision cone out
	// of it.
	const float dot = toSight.dot(facingToUnit[getFacingDir()]);
	if (dot < 0.f)
	{
		// object is behind us
		return false;
	}
	const float d1 = 256.f * dot;
	const float d2 = (256.f * 0.4f) / (0.4f + atan(acosf(dot) / 2.f));
	const float sightDistance = 0.6f * d2 + 0.4f * d1;
	scene->debugDrawLine("sight_test", pos, pos + lengthdir(sightDistance, pointDirection(pos, sightPos)), sf::Color::Yellow);
	//const float dirToSight = pointDirection(pos, sightPos);
	//const int dirInt = static_cast<int>((dirToSight + 45) / 90) % 4;
	if (pointDistance(pos, sightPos) > sightDistance)
	{
		return false;
	}
	std::list<Entity*> results;
	scene->cullFiniteLine(results, pos, sightPos, "StaticGeometry");
	return results.empty();
}

void NPC::enterCar(Car& car)
{
	this->car = &car;
	ASSERT(this->car != nullptr);
	car.setDriver(this, &state, carControls);
	mobilityState = State::InCar;
	Town *town = static_cast<Town*>(scene);
	ASSERT(town == car.getScene());
	pathManager = &town->getRoadPathManager();
}

void NPC::exitCar()
{
	car->removeDriver();
	car = nullptr;
	carControls = nullptr;
	mobilityState = State::OnFoot;
	pathManager = getScene()->getWorld()->getPathManager("StaticGeometry");
}

void NPC::face(const Vector2f& posToFace)
{
	const float dir = pointDirection(pos, posToFace);
	const int dirInt = static_cast<int>((dir + 45) / 90) % 4;
	face(static_cast<FacingDirection>(dirInt));
}

void NPC::face(FacingDirection dir)
{
	facingDir = dir;
	setDrawableSet(facingDirToStr(facingDir));
}

static inline int getNearTextboxes(const ScenePos& location, const std::string& filter)
{
	std::list<NPCTextBox*> found;
	location.getScene()->cull(found, rectAround(location.getPos(), 96));
	return std::count_if(found.cbegin(), found.cend(), [&filter](const NPCTextBox *textbox) {
		return textbox->getMessage() == filter;
	});
}

void NPC::makeTextBubble(const std::string& message)
{
	if (scene->getViewRect().containsPoint(getPos().to<int>()) && getNearTextboxes(ScenePos(*this), message) < 5)
	{
		NPCTextBox *bubble = new NPCTextBox(*this, message);
		textBoxes.insert(bubble->link);
		scene->addEntityToList(bubble);
	}
}

void NPC::changeState(std::unique_ptr<NPCState> newState)
{
	state.changeState(std::move(newState));
}

void NPC::refreshDrawables()
{
	// TOOD: handle other animations?
	loadWalkAnims();
}


void NPC::serialize(Serialize& out) const
{
	out.startObject("NPC");
	serializeEntity(out);
	AutoSerialize::serialize(out, soul);
	AutoSerialize::serialize(out, target);
	// TODO state?
	// room purposefully ignored here - it'll get set when the room you're in is serialized
	// path?
	// pathman?
	ASSERT(mobilityState == State::OnFoot);
	ASSERT(car == nullptr);
	ASSERT(carControls == nullptr);
	// minimapIcon contains no persistent info
	// behaviorContext ignored purposefully
	// targetIsPathing?
	// dialogueStartOverride, currentWeightMap, walkSpeedMultiplier is set by the behavior tree/states anyway, no need to keep this.
	// walkOffset, walkSpeed, runSpeed are not needed to be serialized, they're random and inconsequential.
	SAVEVAR(u8, facingDir);
	// textBoxes ignored on purpose - these go away quickly enough that I don't care.
	out.endObject("NPC");
}

void NPC::deserialize(Deserialize& in)
{
	in.startObject("NPC");
	deserializeEntity(in);
	AutoDeserialize::deserialize(in, soul);
	AutoDeserialize::deserialize(in, target);
	// state?
	// room ignored, see serialize()'s comment
	// path?
	// pathman?
	mobilityState = State::OnFoot;
	car = nullptr;
	carControls = nullptr;
	facingDir = static_cast<FacingDirection>(in.u8("facingDir"));
	in.endObject("NPC");

	initDefaultState();
}

/*		private			*/
// NPCState interface
NPCMoveResult NPC::moveTowardsTarget()
{	
	if (target.isCalculating() || (path && !path->isCalculated()))
	{
		return NPCMoveResult::CalculatingPath;
	}
	else if (path && path->didPathingFail())
	{
		return NPCMoveResult::SomethingInWay;
	}

	if (targetIsPathing)
	{
		if (target.absorbPathData())
		{
			targetIsPathing = false;
		}
	}

	if (target)
	{
		// check if they're close and in vision
		std::list<Entity*> results;
		const bool targetIsInThisScene = target.getScene() == getScene();
		//const bool inRange = (targetIsInThisScene && pointDistance(target.getPos(), pos) < (mobilityState == State::InCar ? 1024 : 222));
		//if (inRange)
		//{
		//	scene->cullFiniteLine(results, pos, target.getPos(), "StaticGeometry");
		//}

		//// close and nothing in way - walk directly towards
		//if (inRange && results.empty())
		//{
		//	moveTowards(target.getPos());
		//	scene->debugDrawLine("draw_npc_path_graph", pos, target.getPos(), sf::Color::Yellow);
		//}
		//else // far away - use pathfinding
		{
			if (!path)
			{
				recalculatePath();
			}

			if (path)
			{
				if (path->isFinished())
				{
					recalculatePath();
				}
				else
				{
					//const int d = scene->getWorld().getDistance();
					const int d = pointDistance(path->lastNode().getPos(), target.getPos());
					const bool targetInSameSceneAsLastPoint = (path->lastNode().getScene() == target.getScene());
					if (path->isFinished()  || // we're done   OR   the target has moved too far away OR target isn't even in the same level as the end
						(path->isCalculated() && ((d > recalcThreshold && targetInSameSceneAsLastPoint) || !targetInSameSceneAsLastPoint)))
					{
						recalculatePath();
					}
				}
			}

			if (!path->isCalculated())
			{
				return NPCMoveResult::CalculatingPath;
			}
			else if (path->didPathingFail())
			{
				return NPCMoveResult::SomethingInWay;
			}

			// path is valid and followable
			if (path && path->isCalculated() && !path->isFinished())
			{
#ifdef SL_DEBUG
				if (Settings::fuckYouSingletonHack()->get("draw_npc_path_graph"))
				{
					std::map<Scene*, std::vector<Vector2f>> drawPoints;
					drawPoints[scene].push_back(pos);
					for (const pf::Node& node : path->getNodes())
					{
						drawPoints[node.getScene()].push_back(node.getPos());
					}
					for (auto& pointList : drawPoints)
					{
						for (const Vector2f& p : pointList.second)
						{
							pointList.first->debugDrawCircle("draw_npc_path_graph", p, 3.f, sf::Color(255, 0, 255, 64));
						}
						pointList.first->debugDrawLines("draw_npc_path_graph", pointList.second, sf::Color::White);
					}
				}
#endif // SL_DEBUG

				const pf::Node& node = path->nextNode();
				if (node.getScene() == getScene())
				{
					const Vector2f& nextTarget = node.getPos() + walkOffset;
					const float dist = pointDistance(pos, nextTarget);
					const float distThresholdFor = mobilityState == State::OnFoot ? getWalkSpeed() * 1.5f : 64.f;
					// check to see if we drove past our node and should instead follow the next node
					if (mobilityState == State::InCar && path->getNodes().size() > 1)
					{
						const Vector2f& nextNextTarget = path->getNodes()[1].getPos();
						if (pointDistance(pos, nextNextTarget) <= pointDistance(nextTarget, nextNextTarget))
						{
							path->pop();
						}
					}
					if (dist > distThresholdFor)
					{
						// TODO: when following a path, we don't want it to repeatedly recalculate.
						// and when we compute to the nearest path we need to search for nodes PAST
						// the one we're currenty stuc at!
						if (!moveTowards(nextTarget))
						{
							return NPCMoveResult::SomethingInWay;
						}
					}
					else
					{
						path->pop();
						walkOffset.x = Random::get().randomFloat(-walkSpeed*1.75f, walkSpeed*1.75f);
						walkOffset.y = Random::get().randomFloat(-walkSpeed*1.75f, walkSpeed*1.75f);
					}
				}
				else
				{
					Scene *newScene = const_cast<Scene*>(node.getScene());
					pos = node.getPos();
					newScene->transferEntity(this);
					path->pop();
				}
			}
		}
		if (target.getScene() == scene)
		{
			const float dist = pointDistance(pos, target.getPos());//worldDistance(*this, target);
			const float distThreshold = mobilityState == State::OnFoot ? 24.f /*speed * 2*/ : 64;
			if (dist <= distThreshold)
			{
				return NPCMoveResult::TargetReached;
			}
		}
		return NPCMoveResult::Moved;
	}
	return NPCMoveResult::NoTarget;
}

void NPC::loadWalkAnims()
{
	LayeredSpriteLoader::load(*this, LayeredSpriteLoader::Config(getSoul()->appearance, "walk", 32, Vector2i(16, 24)).animate(5));
}

void NPC::loadPregnantAnims()
{
	/*
	const int offsetX = 12;
	const int offsetY = 21;
	const ColorSwapID swapID = soul->getColorSwapID();
	// @todo make pregnant animations for hula / dress / maid
	std::string clothing = (soul->getClothing() != "hula" && soul->getClothing() != "dress" && soul->getClothing() != "maid") ? soul->getClothing() : "nude";
	const std::string dir = soul->getType() + "/pregnant/" + clothing + "/";
	// for some reason I can't std::forward these if I use them in the unique<>() ctor....
	const std::string filename = dir + "west.png";
	clearDrawablesInSet("north");
	addDrawableToSet("north", unique<SpriteAnimation>(filename, 24, 1, true, swapID, offsetX, offsetY));
	setAnimationSpeed("north", 0.f);
	clearDrawablesInSet("east");
	addDrawableToSet("east", unique<SpriteAnimation>(filename, 24, 1, true, swapID, offsetX, offsetY));
	setAnimationSpeed("east", 0.f);
	clearDrawablesInSet("west");
	addDrawableToSet("west", unique<SpriteAnimation>(filename, 24, 1, true, swapID, offsetX, offsetY));
	setAnimationSpeed("west", 0.f);
	clearDrawablesInSet("south");
	addDrawableToSet("south", unique<SpriteAnimation>(filename, 24, 1, true, swapID, offsetX, offsetY));
	setAnimationSpeed("south", 0.f);
	*/
}

void NPC::setTarget(const Target& target)
{
	this->target = target;
}

Target NPC::setTargetViaPathing(const pf::NodePredicate& targetPred, const pf::WeightFunc& heuristic, const std::string& weightMap, const std::function<void(std::deque<pf::Node>&)>& postProcess)
{
	if (path && !path->isCalculated())
	{
		pathManager->cancelCalculation(path);
	}
	ASSERT(!car); // @TODO implement for cars too
	const pf::NodeID start = pathManager->getClosestNodeID(Target(pos.x, pos.y, scene));


	path = pathManager->getPath(
		pf::PathManager::PathSpec{
			start,
			targetPred,
			&weightsMap.at(weightMap.empty() ? currentWeightMap : weightMap),
			heuristic,
			&behaviorContext,
			postProcess});
	targetIsPathing = true;
	return Target(path.get());
}





//------------------------------------------------
//             actual private
//------------------------------------------------
void NPC::onUpdate()
{
	if (carControls)
	{
		carControls->reset();
	}
	state.update();
	// @TODO centralize btree updating so it doesn't miss events?
	// but unless this NPC is sending an event to itself, then idk
	// unless it sends an event to another NPC, then that NPC sends
	// an event back???

	// btree done updating, so remove cached events it just handled
	behaviorContext.clearEvents();

	std::shared_ptr<Soul> baby = soul->giveBirth(*getScene()->getWorld());
	if (baby)
	{
		NPC *babyNPC = new NPC(pos.x, pos.y, baby);
		// born into slavery
		babyNPC->getSoul()->getFlags().set("captive", soul->getFlags().get("captive"));
		babyNPC->initDefaultState();
		scene->addEntityToQuad(babyNPC);
		loadWalkAnims();
	}

	calculateHeight();
}

void NPC::onEnter()
{
	state.onSceneEnter();
	minimapIcon.setScene(scene);
	minimapIcon.show();

	ASSERT(soul->getNPC() == nullptr);
	soul->setNPC(this);
}

void NPC::onExit()
{
	ASSERT(soul->getNPC());
	soul->setNPC(nullptr);

	minimapIcon.hide();
	minimapIcon.setScene(nullptr);
	state.onSceneExit();
}

void NPC::recalculatePath()
{
	// if no target, we just crash (plus no sane way to do anything)
	if (pathManager && target)
	{
		if (path && !path->isCalculated())
		{
			pathManager->cancelCalculation(path);
		}
		Vector2f startPos = pos;
		if (mobilityState == State::InCar && path && !path->isFinished())
		{
			startPos = path->nextNode().getPos();
		}
		const pf::NodeID start = pathManager->getClosestNodeID(Target(startPos.x, startPos.y, scene));
		const pf::NodeID end = pathManager->getClosestNodeID(target);
		if (!car)
		{
			path = pathManager->getPath(
				pf::PathManager::PathSpec{
					start,
					pf::IsNode(end),
					&weightsMap.at(currentWeightMap),
					pf::ManhattanDistance(end, 5.f),
					&behaviorContext});
		}
		else
		{
			path = pathManager->getPath(start, end);
		}
	}
	ASSERT(pathManager);
}

bool NPC::moveTowards(const Vector2f& targetPos)
{
	switch (mobilityState)
	{
		case State::OnFoot:
		{
			const float dir = pointDirection(pos, targetPos);
			const Vector2f move = lengthdir(getWalkSpeed(), dir);
			//std::list<NPC*> npcsInWay;
			//scene->cull(npcsInWay, mask.getBoundingBox() + Vector2i(move));
			std::list<Entity*> stuffInWay;
			scene->cullFiniteLine(stuffInWay, pos, pos + move, "StaticGeometry");
			const bool canMove = stuffInWay.empty();
			//mask.getBoundingBox() + move.to<int>(), "StaticGeometry") == nullptr);
			//for (NPC *npc : npcsInWay)
			//{
			//	// this is just an arbitrary unique comparison so that it
			//	// gives the right-of-way symmetrically between two NPCs
			//	if (npc < this)
			//	{
			//		canMove = false;
			//		break;
			//	}
			//}
			if (canMove)
			{
				pos += move;
				face(targetPos);
				setAnimationSpeed(1.f);
			}
			else if (textBoxes.empty() && Random::get().chance(0.01f))
			{
				//const int r = Random::get().random(5);
				//switch (r)
				//{
				//case 0:
				//case 1:
				//	makeTextBubble("Get out of my way!");
				//	break;
				//case 2:
				//	makeTextBubble("U-Um... Excuse me..");
				//	break;
				//case 3:
				//	makeTextBubble("Can I get by?");
				//	break;
				//case 4:
				//	makeTextBubble("...");
				//	break;
				//}
			}
			return canMove;
		}
		break;

		case State::InCar:
		{
			const Vector2f& carPos = car->getPos();
			const int destX = targetPos.x;// @hack for demo so cars drive in middle of street
			const int destY = targetPos.y;
			
			car->driftTowardsCenter(0.1f, 0.f);

			
			const float desiredAngle = pointDirection(carPos.x, carPos.y, destX, destY);
			const float difference = desiredAngle - car->getAngle();
			const int distance = pointDistance(carPos.x, carPos.y, destX, destY);
			if (distance < 256)
			{
				if (car->getSpeed() > 6)
				{
					carControls->brake();
				}
				else if (car->getSpeed() < 4)
				{
					carControls->accelerate();
				}
			}
			else if (car->getSpeed() < 8)
			{
				carControls->accelerate();
			}

			if (difference > 5 || difference < -5)
			{
				if ((difference >= 0 && difference < 180) || difference < -180)
				{
					carControls->turnLeft();
				}
				else
				{
					carControls->turnRight();
				}

				if ((difference * 2 > distance || -difference * 2 > distance) && car->getSpeed() > 4)
				{
					carControls->brake();
				}
			}
			else
			{
//				const int diffX = pos.x - destX;
//				if (diffX < 10 && diffX > -10)
//				{
//					if (diffX > 0)
//					{
//						pos.x -= 0.5;
//					}
//					else
//					{
//						pos.x += 0.5;
//					}
//				}
//				const int diffY = pos.y - destY;
//				if (diffY < 10 && diffY > -10)
//				{
//					if (diffY > 0)
//					{
//						pos.y -= 0.5;
//					}
//					else
//					{
//						pos.y += 0.5;
//					}
//				}
			}

			std::list<Entity*> carsInWay;
			Town *town = static_cast<Town*>(car->getScene());
			if (town->getViewRect().intersects(car->getMask().getBoundingBox()))
			{
				const int oldx = carPos.x, oldy = carPos.y;
				car->setPos(carPos.x + lengthdirX(48, car->getAngle()), carPos.y + lengthdirY(48, car->getAngle()));
				scene->cull(carsInWay, *this, Entity::makeType("Car"));
				car->setPos(oldx, oldy);
			}
			if (!carsInWay.empty())// || (!path->nextNode().isAvailable() && pointDistance(prevNode.getPos(), pos) < 64))
			{
				carControls->brake();
				//	because sometimes people don't FUCKING STOP (and move sideways.... I don't get it)
				if (car->getSpeed() < 4)
				{
					car->stopForReal();
				}
			}
		}
		break;
	}
	return true;
}

float NPC::getWalkSpeed() const
{
	return walkSpeed * walkSpeedMultiplier;
}

// TODO: move this somewhere else?
static inline float avoidScaryPlayerHeuristic(const pf::NodeID& node, ai::BehaviorContext *context)
{
	auto lockedNPC = context->lockNPC();
	if (!lockedNPC || context->getPlayer() == nullptr)
	{
		return 1.f;
	}
	constexpr float minDesiredDist = 4.f; // in worldspace (ie ~tiles)
	constexpr float scalar = 8.f; // so that they stop caring much about the road, etc
	return 1.f + scalar * std::fmaxf(minDesiredDist - worldDistance(*context->getPlayer(), ScenePos(*node.getScene(), node.scenePos())), 0);
}


std::map<std::string, pf::WeightMap> initWeightsMap()
{
	std::map<std::string, std::map<std::string, pf::WeightFunc>> weightsMap;


	{
		auto& weights = weightsMap["npc_routine"];
		weights.insert(std::make_pair("Town", pf::WeightFunc([](const pf::NodeID& node, ai::BehaviorContext*) -> float
		{
			const TileType tile = static_cast<Town*>(node.getScene())->getTileAtInTileCoords(node.getX(), node.getY());
			if (isRoad(tile))
			{
				return 10000.f;
			}
			else if (isSidewalk(tile) || tile == CrosswalkH || tile == CrosswalkV)
			{
				return 1.f;
			}
			else if (isParkPath(tile))
			{
				return 2.f;
			}
			return 10.f;
		})));
		weights.insert(std::make_pair("Scene", pf::uniformWeight));
	}

	{
		auto& weights = weightsMap["park_play"];
		weights.insert(std::make_pair("Town", pf::WeightFunc([](const pf::NodeID& node, ai::BehaviorContext*) -> float
		{
			const TileType tile = static_cast<Town*>(node.getScene())->getTileAtInTileCoords(node.getX(), node.getY());
			if (isRoad(tile))
			{
				return 10000.f;
			}
			else if (isSidewalk(tile) || tile == CrosswalkH || tile == CrosswalkV)
			{
				return 1.f;
			}
			else if (isParkPath(tile))
			{
				return 2.f;
			}
			return 3;
		})));
		weights.insert(std::make_pair("Scene", pf::uniformWeight));
	}

	{
		auto& weights = weightsMap["flee_player"];
		weights.insert(std::make_pair("Town", avoidScaryPlayerHeuristic));
		weights.insert(std::make_pair("Scene", avoidScaryPlayerHeuristic));
	}

	{
		auto& weights = weightsMap["uniform_weight"];
		weights.insert(std::make_pair("Town", pf::uniformWeight));
		weights.insert(std::make_pair("Scene", pf::uniformWeight));
	}

	return std::move(weightsMap);
}

/*static*/const std::map<std::string, std::map<std::string, pf::WeightFunc>> NPC::weightsMap = initWeightsMap();

}// End of sl namespace
