/**
 * @section DESCRIPTION
 * A class for representing NPCs in PhysicsScenes. Do not subclass, instead use NPCStates to extend
 * functionality. Or do whatever, you probably know better than I do anyways.
 */
#ifndef SL_NPC_HPP
#define SL_NPC_HPP

#include <SFML/Graphics/RenderTarget.hpp>
#include <SFML/System/Vector2.hpp>

#include <memory>

#include "Engine/Entity.hpp"
#include "Engine/StateManager.hpp"
#include "Engine/Serialization/DeserializeConstructor.hpp"
#include "NPC/Behaviors/BehaviorContext.hpp"
#include "NPC/Locators/Target.hpp"
#include "NPC/Locators/TargetLocator.hpp"
#include "NPC/Stats/Soul.hpp"
#include "NPC/FacingDirection.hpp"
#include "NPC/NPCState.hpp"
#include "Pathfinding/Path.hpp"
#include "Pathfinding/PathfindingCommon.hpp"
#include "Town/Map/Icon.hpp"

namespace sl
{

class Car;
class CarControls;
class NPCState;
class Room;

namespace pf
{
class PathManager;
} // pf

// see NPC::moveTowardsTarget
enum class NPCMoveResult : char
{
	Moved,
	SomethingInWay,
	TargetReached,
	NoTarget,
	CalculatingPath
};

class NPC : public Entity
{
public:
	DECLARE_SERIALIZABLE_METHODS(NPC)

	NPC(DeserializeConstructor);

	NPC(int x, int y, std::shared_ptr<Soul> soul);
	
	/**
	 * Sets the starting state for the NPC. You must call this before the NPC gets updated.
	 * This was in here to get around the fact that you can't construct a NPCState without
	 * an NPC, so I couldn't just define the starting state in the constructor.
	 * @param startingState The starting state to set this NPC to
	 * @param globalState The starting global state this NPC is set to
	 */
	void initStates(std::unique_ptr<NPCState> startingState, std::unique_ptr<NPCState> globalState = nullptr);

	void initDefaultState();

	/**
	 * Returns the Entity type
	 * @return The concrete Type NPC
	 */
	const Types::Type getType() const;

	void setRoom(Room *room);

	Room* getRoom() const;

	/**
	 * Responds to events and passes them down to the current states
	 * @param event The event to handle
	 */
	void handleEvent(const Event& event) override;

	/**
	 * Acess the NPC's Soul, used for accessing statistics about the NPC
	 * @return The Soul
	 */
	std::shared_ptr<Soul> getSoul();

	const std::shared_ptr<const Soul> getSoul() const;

	/**
	 * Changes which Dialogue node a non-forced conversation will start at.
	 * Only call this is there is not an override already!
	 * @param newStart The name of the new start node
	 */
	void setDialogueStartOverride(std::string newStart);

	/**
	 * @return The Dialogue node to start at. If null, then start at the default for the NPCState
	 */
	const std::string* getDialogueStartOverride() const;

	void resetDialogueStartOverride();

	void setWalkSpeedMultiplier(float mult);

	float getWalkSpeedMultiplier() const;

	void setPathfindingWeightMap(std::string weightMap);

	const std::string& getCurrentPathfindingWegithMap() const;

	FacingDirection getFacingDir() const { return facingDir; }


	// TODO: make the following few functions methods of some NPC::Controller friend class
	// rather than having them public? then we could pass tht to BehaviorContext...

	bool canHear(const Vector2f& noisePos, float noiseRadius) const;

	bool canSee(const Vector2f& sightPos) const;

	ai::BehaviorContext& getBehaviorContext() { return behaviorContext; }

	Car* getCar() { return car; }

	void enterCar(Car& car);

	void exitCar();

	void face(const Vector2f& posToFace);

	void face(FacingDirection dir);

	void makeTextBubble(const std::string& message);

	void changeState(std::unique_ptr<NPCState> newState);

	void refreshDrawables();

private:
	enum class State
	{
		OnFoot,
		InCar
	};

	// Interface available to NPCState
	/**
	 * Attempts to move towards the NPC's target
	 * @return The result of the move attempt
	 */
	NPCMoveResult moveTowardsTarget();

	void loadWalkAnims();

	void setTarget(const Target& target);

	/**
	 * @param weightMap If empty, uses the NPC's current weightMap. Otherwise uses the provided one
	 */
	Target setTargetViaPathing(const pf::NodePredicate& targetPred, const pf::WeightFunc& heuristic, const std::string& weightMap = std::string(), const std::function<void(std::deque<pf::Node>&)>& postProcess = std::function<void(std::deque<pf::Node>&)>());

	

	// Actual private stuff (DO NOT CALL FROM NPCSTATE)
	/**
	 * Updates the NPC and its state.
	 */
	void onUpdate() override;

	void onEnter() override;

	void onExit() override;

	void recalculatePath();

	/** 
	 * Moves directly towards the input location without doing any kind of pathfinding
	 * @param target The location to move towards
	 * @return true if moved, false if something in the way
	 */
	bool moveTowards(const Vector2f& target);

	void loadPregnantAnims();

	float getWalkSpeed() const;




	//!The unique information for the NPC
	std::shared_ptr<Soul> soul;
	//!The target position the NPC might move towards
	Target target;
	//!The current state
	StateManager<NPCState> state;
	//!The Room the NPC lives in
	Room *room;
	//!The path the NPC is following (if they are going towards one)
	std::shared_ptr<pf::Path> path;
	//!The path manager to use for calculating pathing
	pf::PathManager *pathManager;
	//!The player's mobility state (in car, on foot, etc) which affects pathfinding/movement
	State mobilityState;
	//!The Car the ai is driving. null if not in car. Not a Ref<Car> since the car shouldn't be destroyed while you're in it...
	Car *car;
	//!Controls for the car. null if not in car
	CarControls *carControls;

	map::Icon minimapIcon;

	ai::BehaviorContext behaviorContext;

	bool targetIsPathing;

	std::string currentWeightMap;

	std::string dialogueStartOverride;

	float walkSpeed;
	float runSpeed;
	float walkSpeedMultiplier;

	Vector2f walkOffset;

	FacingDirection facingDir;

	IntrusiveList<Entity> textBoxes;

	friend class NPCState;
	friend class PathfindingTargetLocator;
	friend class ai::BehaviorContext;

#ifdef SL_DEBUG
	public: bool debugDrawBehaviorTree; private:
#endif // SL_DEBUG

public:
	static const std::map<std::string, std::map<std::string, pf::WeightFunc>> weightsMap;
};

}// End of sl namespace

#endif
