#include "NPC/NPCAppearance.hpp"

#include "Engine/Serialization/AutoDeserialize.hpp"
#include "Engine/Serialization/AutoSerialize.hpp"
#include "Engine/Serialization/Serialization.hpp"
#include "NPC/Stats/NPCColorSwaps.hpp"
#include "NPC/Stats/Soul.hpp"
#include "NPC/ClothesDatabase.hpp"

namespace sl
{

NPCAppearance::NPCAppearance()
	:type("uninitialized")
	,ageDays(-1)
{
	resetSwaps();
}

NPCAppearance::NPCAppearance(std::string type, int ageDays, const DNA& dna)
	:type(std::move(type))
	,ageDays(ageDays)
	,expressedDNA(dna)
{
}

NPCAppearance& NPCAppearance::init(const Soul& soul)
{
	this->type = soul.getType();
	this->ageDays = soul.getAgeDays();
	expressedDNA = soul.getDNA();
	resetSwaps();
	return *this;
}

void NPCAppearance::serialize(Serialize& out) const
{
	out.startObject("NPCAppearance");
	out.u8("clothingLayers", clothingLayers.toFlag());
	SAVEVAR(s, type);
	out.s("expressedDNA", expressedDNA.genes.to_string());
	SAVEVAR(u32, ageDays);
	// swapID stuff ignored
	AutoSerialize::serialize(out, cumLayers);
	AutoSerialize::serialize(out, clothing);
	AutoSerialize::serialize(out, clothingColour);
	SAVEVAR(s, hair);
	out.endObject("NPCAppearance");
}

void NPCAppearance::deserialize(Deserialize& in)
{
	in.startObject("NPCAppearance");
	clothingLayers.fromFlag(in.u8("clothingLayers"));
	LOADVAR(s, type);
	expressedDNA.genes = decltype(expressedDNA.genes)(in.s("expressedDNA"));
	LOADVAR(u32, ageDays);
	// swapID stuff ignored
	AutoDeserialize::deserialize(in, cumLayers);
	AutoDeserialize::deserialize(in, clothing);
	AutoDeserialize::deserialize(in, clothingColour);
	LOADVAR(s, hair);
	in.endObject("NPCAppearance");

	resetSwaps();
}

const std::string& NPCAppearance::getType() const
{
	return type;
}

void NPCAppearance::setType(std::string type)
{
	this->type = std::move(type);
}

float NPCAppearance::getAgeYears() const
{
	return ageDays / 365.f;
}

void NPCAppearance::setAge(int ageInDays)
{
	this->ageDays = ageInDays;
}

const ColorSwapID& NPCAppearance::getBodySwapID() const
{
	return bodySwapID;
}

const ColorSwapID& NPCAppearance::getClothingSwapID(ClothingType clothingType) const
{
	const int typeIndex = static_cast<int>(clothingType);
	if (!isClothingSwapIDValid[typeIndex] && type == "loli") // TODO hack fix this shit later
	{
		if (getClothing(clothingType).empty())
		{
			static ColorSwapID empty;
			return empty;
		}
		cachedClothingSwapID[typeIndex] = ClothesDatabase::get().getInfo(getClothing(clothingType)).swaps.at(getClothingColour(clothingType)).swapID;
		isClothingSwapIDValid[typeIndex] = true;
	}
	return cachedClothingSwapID[typeIndex];
}

const ColorSwapID& NPCAppearance::getHairSwapID() const
{
	if (!isHairSwapIDValid)
	{
		// @todo change this when adding hair dye
		cachedHairSwapID = toColorSwapID(expressedDNA.getHairColor()) | bodySwapID;
		isHairSwapIDValid = true;
	}
	return cachedHairSwapID;
}

const std::string& NPCAppearance::getHair() const
{
	return hair;
}

void NPCAppearance::setHair(std::string newHair)
{
	hair = std::move(newHair);
}

const std::string& NPCAppearance::getClothing(ClothingType clothingType) const
{
	return clothing[static_cast<int>(clothingType)];
}

const std::string& NPCAppearance::getClothingColour(ClothingType clothingType) const
{
	return clothingColour[static_cast<int>(clothingType)];
}

void NPCAppearance::setClothing(ClothingType clothingType, const std::string& newClothing)
{
	clothing[static_cast<int>(clothingType)] = newClothing;
	clothingLayers.fromEnum(clothingType) = true;
	//@hack please fix todo
	if (type != "cop" && type != "adultmale" && type != "adultfemale")
		clothingColour[static_cast<int>(clothingType)] = ClothesDatabase::get().getRandomSwap(Random::get(), getClothing(clothingType));
	isClothingSwapIDValid[static_cast<int>(clothingType)] = false;
}

void NPCAppearance::setClothingColour(ClothingType clothingType, const std::string& newColour)
{
	clothingColour[static_cast<int>(clothingType)] = newColour;
	isClothingSwapIDValid[static_cast<int>(clothingType)] = false;
}

std::string NPCAppearance::getResourcePath() const
{
	std::string path = getType();
	if (path == "loli")
	{
		if (getAgeYears() <= 3.5f)
		{
			path += "/tod";
		}
		else if (getAgeYears() <= 7.5f)
		{
			path += "/kinder";
		}
		else if (getAgeYears() <= 11.5f)
		{
			path += "/elem";
		}
		else// if (getAgeYears() <= 14)
		{
			path += "/middle";
		}
	}
	else if (path == "cop")
	{
		return "adultmale";
	}
	// In the other cases there's no need to append
	return std::move(path);
}

const std::set<std::string>& NPCAppearance::getCumLayers() const
{
	return cumLayers;
}

void NPCAppearance::addCumLayer(const std::string& layer)
{
	cumLayers.insert(layer);
}

bool NPCAppearance::removeCumLayer(const std::string& layer)
{
	return cumLayers.erase(layer) > 0;
}

// private
void NPCAppearance::resetSwaps()
{
	std::fill(std::begin(isClothingSwapIDValid), std::end(isClothingSwapIDValid), false);
	isHairSwapIDValid = false;
	bodySwapID = toColorSwapID(expressedDNA.getSkinColor()) | toColorSwapID(expressedDNA.getEyeColor());
}

} // sl