/**
 * @section DESCRIPTION
 * The base class for NPC states.
 */
#ifndef SL_NPCSTATE_HPP
#define SL_NPCSTATE_HPP

#include <memory>
#include <string>

#include "Pathfinding/PathfindingCommon.hpp"
#include "Utility/Vector.hpp"

namespace sl
{

class Car;
class NPC;
class Target;
class Event;
class Soul;

enum class NPCMoveResult : char;

class NPCState
{
public:
	/**
	 * Destructor for NPCState
	 */
	virtual ~NPCState() {};
	/**
	 * Code to be executed when the state is entered.
	 */
	virtual void onEnter() {}
	/**
	 * Code to be executed when the state is exited
	 */
	virtual void onExit() {}
	/**
	 * To be executed when the NPC'S SCENE is entered
	 */
	virtual void onSceneEnter() {}
	/**
	 * To be executed when the NPC'S SCENE exits
	 */
	virtual void onSceneExit() {}
	/**
	 * Updates the state.
	 */
	virtual void update() {}
	/**
	 * Handles events
	 * @param event The event to handle in the substate
	 */
	virtual void handleEvent(const Event& event) {}

protected:
	/**
	 * Initializes the state
	 * @param owner The NPC this state belongs to
	 */
	NPCState(NPC& owner);
	/**
	 * Access the owner's position in subclasses.
	 * @return Returns a reference to the owner's position vector.
	 */
	Vector2<float>& getPos();
	/**
	 * Changes state
	 * @param newState The new state
	 */
	void changeState(std::unique_ptr<NPCState> newState);

	/**
	 * See NPC::moveTowardsTarget
	 * @return The result of the move attempt
	 */
	NPCMoveResult moveTowardsTarget();

	void setPathManager(const std::string& manager);

	void loadWalkAnims();

	void enterCar(Car& car);

	void exitCar();

	void face(const Vector2f& posToFace);

	void setTarget(const Target& target);

	Target setTargetViaPathing(const pf::NodePredicate& targetPred, const pf::WeightFunc& heuristic, const std::string& weightMap);

	bool reportCrime(int crimeId);





	//!The NPC this state belongs to's target
	Target& target;
	//!The NPC this state belong to's position
	Vector2<float>& pos;
	//!The NPC this state belongs to
	NPC& owner;
};

} // sl

#endif // SL_NPCSTATE_HPP
