#include "NPC/Schedule.hpp"

#include <algorithm>

#include "NPC/NPC.hpp"
#include "NPC/Stats/Soul.hpp"
#include "Town/World.hpp"
#include "Utility/Assert.hpp"
#include "Utility/DataToString.hpp"

namespace sl
{

const std::pair<std::string, float> Schedule::idle("Idle", unavailable);

std::pair<std::string, float> Schedule::current(const Soul& soul, const GameTime& time) const
{
	std::pair<std::string, float> best = idle;
	for (const Activity& activity : activities)
	{
		if (activity.time->isWithin(time))
		{
			const float priority = activity.priority(soul);
			if (priority > best.second)
			{
				best.first = activity.id;
				best.second = priority;
			}
		}
	}
	return idle;
}

std::vector<std::pair<std::string, float>> Schedule::allCurrent(const Soul& soul, const GameTime& time) const
{
	std::vector<std::pair<std::string, float>> available;
	for (const Activity& activity : activities)
	{
		if (activity.time->isWithin(time))
		{
			const float priority = activity.priority(soul);
			if (priority > unavailable)
			{
				available.push_back(std::make_pair(activity.id, priority));
			}
		}
	}
	std::sort(
		available.begin(),
		available.end(),
		[](const std::pair<std::string, float>& lhs, const std::pair<std::string, float>& rhs) {
			return lhs.second > rhs.second;
		});
	return available;
}

Schedule::InterpolatedSchedule Schedule::computeSchedule(const Soul& soul, const GameTime& start, const GameTime& end) const
{
	InterpolatedSchedule ret;
	// TODO: handle conflicts and find the optimial schedule
	std::vector<InterpolatedSchedule::Slot> allTimes;
	for (const Activity& activity : activities)
	{
		std::vector<TimeInterval> times = activity.time->generate(start, end);
		for (const TimeInterval& time : times)
		{
			allTimes.push_back(InterpolatedSchedule::Slot{activity.id, time.start, time.end});
		}
	}
	std::sort(
		allTimes.begin(),
		allTimes.end(),
		[](const InterpolatedSchedule::Slot& lhs, const InterpolatedSchedule::Slot& rhs) {
			return lhs.start < rhs.start;
		});
	//std::cout << "\n\n\n\n\n\n\n\n\nComputed schedule:\n";
	for (int i = 0; i < allTimes.size(); ++i)
	{
		//std::cout << allTimes[i].id << " : ";
		//auto date = [](const GameTime& time) {
		//	std::cout << formatTime(time.time) << " (" << formatDateShort(time.date) << ")";
		//};
		//date(allTimes[i].start);
		//std::cout << "   -   ";
		//date(allTimes[i].end);
		//std::cout << std::endl;
		if (i > 0)
		{
			// HACK DEMO figoure out why sometimes they overlap by 2
			//ASSERT(allTimes[i - 1].end <= allTimes[i].start);
		}
		ret.add(allTimes[i]);
	}
	return ret;
}

void Schedule::registerActivity(std::string id, std::string location, std::unique_ptr<const TimeIntervalGenerator>&& time, std::function<float(const Soul&)> priority)
{
	activities.push_back(Activity{ std::move(id), std::move(location), std::move(time), std::move(priority) });
}

void Schedule::registerActivity(std::string id, std::string location, std::unique_ptr<const TimeIntervalGenerator>&& time, float fixedPriority)
{
	registerActivity(std::move(id), std::move(location), std::move(time), [fixedPriority](const Soul&) { return fixedPriority; });
}

float Schedule::getPriority(const std::string& id, const Soul& soul) const
{
	auto it = std::find_if(activities.begin(), activities.end(), [&id](const Activity& activity) { return activity.id == id; });
	return it != activities.end() ? it->priority(soul) : -1.f;
}

const std::string& Schedule::getLocationId(const std::string& id) const
{
	static const std::string invalid;
	auto it = std::find_if(activities.begin(), activities.end(), [&id](const Activity& activity) { return activity.id == id; });
	return it != activities.end() ? it->location : invalid;
}


// -------------------------------- InterpolatedSchedule ----------------------
void Schedule::InterpolatedSchedule::add(Slot slot)
{
	slots.push_back(std::move(slot));
}

Schedule::InterpolatedSchedule::Interpolation Schedule::InterpolatedSchedule::interpolate(const GameTime& time) const
{
	Interpolation ret;
	if (!slots.empty())
	{
		if (time < slots.front().start)
		{
			ret.endId = slots.front().id;
			ret.progress = 0.f;
		}
		else if (slots.back().end < time)
		{
			ret.startId = slots.back().id;
			ret.progress = 1.f;
		}
		else
		{
			for (int i = 0; i < slots.size(); ++i)
			{
				if (slots[i].start <= time && time <= slots[i].end)
				{
					ret.startId = ret.endId = slots[i].id;
					const uint64_t a = slots[i].start.u64();
					const uint64_t b = slots[i].end.u64();
					const uint64_t x = time.u64();
					ret.progress = (double)(x - a) / (b - a);
					break;
				}
				if (time < slots[i].start)
				{
					ASSERT(i > 0);
					ASSERT(slots[i - 1].end <= time);
					ret.startId = slots[i - 1].id;
					ret.endId = slots[i].id;
					const uint64_t a = slots[i - 1].end.u64();
					const uint64_t b = slots[i].start.u64();
					const uint64_t x = time.u64();
					ret.progress = (double)(x - a) / (b - a);
					break;
				}
			}
		}
		ASSERT(ret.progress >= 0.f && ret.progress <= 1.f);
	}
	else
	{
		ret.startId = ret.endId = "invalid";
		ret.progress = -1.f;
	}
	return ret;
}

// ---------------------- InterpolatedSchedule::Interpolation -----------------
float Schedule::InterpolatedSchedule::Interpolation::priority(const Soul& soul) const
{
	return startContrib() * soul.getSchedule().getPriority(startId, soul)
	     + endContrib() * soul.getSchedule().getPriority(endId, soul);
}

ScenePos Schedule::InterpolatedSchedule::Interpolation::location(NPC& npc, const World& world) const
{
	if (inBetween())
	{
		const std::string& locationId = npc.getSoul()->getSchedule().getLocationId(startId);
		return npc.getSoul()->getLocation(locationId)->findTarget(npc.getBehaviorContext()).getScenePos();
	}
	else
	{
		const std::shared_ptr<const pf::Path> p = path(npc, world);
		return p->interpolate(progress * p->totalSceneDistance()).position();
	}
}

std::shared_ptr<const pf::Path> Schedule::InterpolatedSchedule::Interpolation::path(NPC& npc, const World& world) const
{
	// TODO: allow multiple schedules rather than manually referencing the normal schedule here - perhaps pass the original Schedule as a refernece, or store both the locationid + id?
	return npc.getSoul()->locationPathCache.getPath(
		npc.getSoul()->getSchedule().getLocationId(startId),
		npc.getSoul()->getSchedule().getLocationId(endId),
		npc,
		*world.getPathManager("StaticGeometry"),
		NPC::weightsMap.at("npc_routine"),
		5.f);
}

bool Schedule::InterpolatedSchedule::Interpolation::inBetween() const
{
	return startId == endId;
}

// private
float Schedule::InterpolatedSchedule::Interpolation::startContrib() const
{
	return 0.5f * (1.f - progress);
}

float Schedule::InterpolatedSchedule::Interpolation::endContrib() const
{
	return 0.5f * progress;
}

} // sl