#include "NPC/States/AdultWanderState.hpp"

#include "Engine/Events/Event.hpp"
#include "Town/World.hpp"
#include "NPC/NPC.hpp"
#include "Player/Player.hpp"
#include "Police/CrimeEntity.hpp"
#include "Town/Town.hpp"
#include "Utility/Random.hpp"
#include "Utility/Trig.hpp"
#include "Utility/Time/RecurringTimeInterval.hpp"

#include "Utility/ProportionList.hpp"

namespace sl
{

const sf::Color shirtOutline(51, 51, 51);
const sf::Color shirtShadow(192, 192, 192);
const sf::Color shirtLight(254, 254, 254);

const ColorSwapID blackShirt = ColorSwapID({
	std::make_pair(shirtOutline, sf::Color(32, 32, 32)),
	std::make_pair(shirtShadow, sf::Color(48, 48, 48)),
	std::make_pair(shirtLight, sf::Color(64, 64, 64))
});

const ColorSwapID redShirt = ColorSwapID({
	std::make_pair(shirtOutline, sf::Color(90, 18, 18)),
	std::make_pair(shirtShadow, sf::Color(133, 33, 33)),
	std::make_pair(shirtLight, sf::Color(159, 45, 45))
});

AdultWanderState::AdultWanderState(NPC& owner)
	:NPCState(owner)
	,off(Random::get().random(32) - 16, Random::get().random(32) - 16)
{
	const bool isMale = owner.getSoul()->getDNA().isMale();
	owner.getSoul()->setType(isMale ? "adultmale" : "adultfemale");
	owner.getSoul()->appearance.setClothing(ClothingType::Outfit, isMale ? "tshirtjeans" : "tanktopjeans");
	
	const int offsetX = 12;
	const int offsetY = 21;
	ProportionList<ColorSwapID> clothingSwaps;
	clothingSwaps.insert(ColorSwapID(), 10);
	clothingSwaps.insert(blackShirt, 3);
	clothingSwaps.insert(redShirt, 4);

	const DNA& dna = owner.getSoul()->getDNA();

	loadWalkAnims();

	owner.setDrawableSet(Random::get().choose({"north", "east", "west", "south", "south", "south"}));

	owner.setPathfindingWeightMap("npc_routine");
}

void AdultWanderState::update()
{
	const Player *player = owner.getScene()->getPlayer();
	if (player)
	{
		const float distToPlayer = pointDistance(player->getPos(), owner.getPos());
		if (distToPlayer < 64)
		{
			face(player->getPos() + (off * 10.f / distToPlayer));
		}
	}

	std::list<CrimeEntity*> crimes;
	owner.getScene()->cull(crimes, rectAround(owner.getPos(), 256));
	for (CrimeEntity *crime : crimes)
	{
		if (owner.canSee(crime->getPos()) && reportCrime(crime->getCrimeId()))
		{
			owner.makeTextBubble("Hello, 911?");
		}
	}

	owner.setWalkSpeedMultiplier(0.3f); // so lolis can catch up for them - TODO: change this when you let lolis call to them to get their attention

	if (target)
	{
		const NPCMoveResult result = moveTowardsTarget();
		if (result == NPCMoveResult::TargetReached || result == NPCMoveResult::SomethingInWay)
		{
			findNewWaypoint();
			owner.getScene()->debugDrawText("draw_npc_path_graph", owner.getPos() + Vector2f(0, -64), result == NPCMoveResult::TargetReached ? "TargetReached" : "SomethingInWay");
		}
		else if (result == NPCMoveResult::CalculatingPath)
		{
			// animate something over their head?
			std::stringstream ss;
			ss << "calculating path";
			owner.getScene()->debugDrawText("draw_npc_path_graph", owner.getPos() + Vector2f(12, -32), ss.str());
		}
	}
	else if (!target.isCalculating())
	{
		findNewWaypoint();
		owner.getScene()->debugDrawText("draw_npc_path_graph", owner.getPos() + Vector2f(0, -96), "No Target");
	}
}

void AdultWanderState::handleEvent(const Event& event)
{
	if (event.getCategory() == Event::Category::NPCInteraction)
	{
		static int timer = 0;
		if (event.getName() == "talk" && --timer <= 0)
		{
			timer = 90;
			owner.makeTextBubble(Random::get().choose({ "Sorry, I'm busy" }));
		}
	}
}

void AdultWanderState::onEnter()
{
	setPathManager("StaticGeometry");
}

void AdultWanderState::findNewWaypoint()
{
	const GameTime& time = owner.getScene()->getWorld()->getGameTime();
	static const RecurringTimeInterval daytime(TimeOfDay(7, 0), 60*60*14, RecurringTimeInterval::AnyDayOfWeek);
	if (daytime.isWithin(time))
	{
		// wander town during daytime
		if (owner.getScene()->getType() == "Town")
		{
			const int townWidth = owner.getScene()->getWidth();
			const int townHeight = owner.getScene()->getHeight();
			do
			{
				//pf::NodeID randomNode = owner.getScene()->getWorld()->getPathManager("StaticGeometry")->getRandomID();
				//target.set(randomNode.getX() * randomNode.getScene()->getTileSize(), randomNode.getY() * randomNode.getScene()->getTileSize(), randomNode.getScene());
				target.set(townWidth / 4 + Random::get().random(townWidth / 2), townHeight / 4 + Random::get().random(townHeight / 2), owner.getScene());
			}
			while (!isSidewalk(static_cast<Town*>(target.getScene())->getTileAt(target.getX(), target.getY())) ||
				target.getScene()->checkCollision(Rect<int>(target.getX() - 32, target.getY() - 32, 64, 64), "StaticGeometry") != nullptr);
			//owner.getScene()->debugDrawRect("draw_npc_path_graph", Rect<int>(target.getX() - 32, target.getY() - 32, 64, 64), sf::Color::Red);
		}
		else // walk towards town
		{
			if (!target && !target.isCalculating())
			{
				setTarget(setTargetViaPathing(
					[](const pf::NodeID& node, ai::BehaviorContext*) -> bool
				{
					return node.getScene()->getType() == "Town";
				}, pf::uniformWeight, owner.getCurrentPathfindingWegithMap()));
			}
		}
	}
	else
	{
		// go home
		const Target bedroom = owner.getSoul()->getLocation("bedroom")->findTarget();
		if (bedroom.getScene() != owner.getScene() || pointDistance(bedroom.getPos(), owner.getPos()) > 32.f)
		{
			target = owner.getSoul()->getLocation("bedroom")->findTarget();
		}
		else
		{
			// sleep? nah, adults don't need sleep...
			target.unset();
		}
	}
}

} // sl
