#include "NPC/States/Police/CopPatrolState.hpp"

#include "Town/World.hpp"
#include "Engine/Graphics/SpriteAnimation.hpp"
#include "NPC/NPC.hpp"
#include "NPC/Locators/Target.hpp"
#include "NPC/States/IdleState.hpp"
#include "NPC/LayeredSpriteLoader.hpp"
#include "Player/Player.hpp"
#include "Police/CrimeEntity.hpp"
#include "Town/Town.hpp"
#include "Town/Cars/Car.hpp"
#include "Utility/Trig.hpp"

#include "NPC/Blood.hpp"
#include "Utility/Random.hpp"
#include "Utility/Trig.hpp"

#include "NPC/Behaviors/CustomCondition.hpp"
#include "NPC/Behaviors/ForgetfulSelector.hpp"
#include "NPC/Behaviors/IfElse.hpp"
#include "NPC/Behaviors/MovementBehaviors.hpp"
#include "NPC/Behaviors/PlayAnimation.hpp"
#include "NPC/Behaviors/Selector.hpp"
#include "NPC/Behaviors/Sequence.hpp"
#include "NPC/Behaviors/Wait.hpp"
#include "NPC/Behaviors/WalkTo.hpp"
#include "NPC/Locators/CustomTargetLocator.hpp"
#include "NPC/Locators/DynamicTargetLocator.hpp"

namespace sl
{
namespace ai
{

static std::unique_ptr<BehaviorTreeNode> getInCarNode()
{
	auto root = unique<Selector>();
	root->addChild(unique<CustomCondition>([](BehaviorContext& context) {
		return context.getNPC().getCar() != nullptr;
	}));
	auto getInCar = unique<Sequence>();
	{
		getInCar->addChild(unique<WalkTo>(unique<DynamicTargetLocator>("patrol_car")));
		getInCar->addChild(unique<CustomCondition>([](BehaviorContext& context) {
			context.getNPC().enterCar(*static_cast<Car*>(context.getNPC().getSoul()->getKnownEntity("patrol_car")));
			return BehaviorTreeNode::Result::Success;
		}));
	}
	root->addChild(std::move(getInCar));
	return std::move(root);
}

static std::unique_ptr<BehaviorTreeNode> createCopBehaviorTree()
{
	auto root = unique<ForgetfulSelector>();

	auto chasingSameScene = unique<ForgetfulSelector>();
	{
		// Report player via radio
		chasingSameScene->addChild(unique<CustomCondition>([](BehaviorContext& context) {
			PoliceForce& police = context.getWorld()->getPolice();
			if (context.getNPC().canSee(context.getPlayer()->getPos()))
			{
				if (police.getLastKnownLocationAge() > 15 * 30)
				{
					context.getNPC().makeTextBubble("Suspect sighted!");
				}
				police.updateLastKnownLocation();
			}
			// always 'fail' so that the next nodes run too.
			return BehaviorTreeNode::Result::Failure;
		}));
		// shoot them if close
		auto shoot = unique<ai::Sequence>();
		{
			shoot->addChild(unique<CustomCondition>([](BehaviorContext& context) {
				NPC& npc = context.getNPC();
				Player& player = *context.getPlayer();
				const float dist = pointDistance(npc.getPos(), player.getPos());
				const bool canSee = npc.canSee(player.getPos());
				if (dist <= 169 && canSee)
				{
					Random& rng = Random::get();
					const int n = rng.randomInclusive(9, 13);
					for (int i = 0; i < n; ++i)
					{
						const Vector2f veloc = lengthdir(7.f + rng.randomFloat(7.f), pointDirection(npc.getPos(), player.getPos()))
						                     + lengthdir(4, Random::get().random(360));
						const float s = Random::get().randomFloat(1);
						npc.getScene()->addEntityToList(new Blood(player.getPos().x, player.getPos().y, veloc.x * s, veloc.y * s));
					}
					player.getStats().soul->getStatus().hp -= 5;
					return BehaviorTreeNode::Result::Success;
				}
				return BehaviorTreeNode::Result::Failure;
			}));
			shoot->addChild(unique<ai::PlayAnimation>([](BehaviorContext& context) {
				return context.getPlayer()->getPos().x < context.getNPC().getPos().x ? "shoot_west" : "shoot_east";
			}));
			shoot->addChild(unique<CustomCondition>([](BehaviorContext& context) {
				// make them no longer invisible...
				// TOOD: instead of this hack, make a proper animation system.
				context.getNPC().setDrawableSet(context.getPlayer()->getPos().x < context.getNPC().getPos().x ? "west" : "east");
				return BehaviorTreeNode::Result::Success;
			}));
			shoot->addChild(unique<Wait>(30*1));
		}
		chasingSameScene->addChild(std::move(shoot));
		// Run to player
		chasingSameScene->addChild(unique<WalkTo>(unique<CustomTargetLocator>([](BehaviorContext& context) {
			return context.getWorld()->getPolice().getLastKnownLocation();
		})));
	}

	auto patrolling = unique<Selector>();
	{
		auto patrolInCar = unique<Sequence>();
		{
			patrolInCar->addChild(goToTownNode());
			patrolInCar->addChild(getInCarNode());
			patrolInCar->addChild(unique<WalkTo>(unique<CustomTargetLocator>(
			[](BehaviorContext& context) {
				Town *town = static_cast<Town*>(context.getScene());
				const pf::VertexBasedPathManager& pathMan = town->getRoadPathManager();
				pf::Node node = pathMan.getNodeFromID(pathMan.getRandomID());
				return Target(node.getPos().x, node.getPos().y, town);
			})));
		}
		patrolling->addChild(std::move(patrolInCar));

		auto patrolOnFoot = unique<Sequence>();
		{
			patrolOnFoot->addChild(goToTownNode());
			patrolOnFoot->addChild(unique<WalkTo>(unique<CustomTargetLocator>(
			[](BehaviorContext& context) {
				Town *town = static_cast<Town*>(context.getScene());
				const Vector2f pos = town->getRandomSidewalkTile(Random::get());
				return Target(pos.x, pos.y, town);
			})));
		}
		patrolling->addChild(std::move(patrolOnFoot));
	}

	root->addChild(unique<IfElse>(
		unique<CustomCondition>([](BehaviorContext& context) {
			if (context.getPlayer() == nullptr)
			{
				return false;
			}
			static std::set<DNA::SkinColor> attackOnSight = {
				DNA::SkinColor::DarkMoreno2,
				DNA::SkinColor::Nigger,
				DNA::SkinColor::DarkMoreno,
				DNA::SkinColor::Tropical,
				DNA::SkinColor::ExtraNigger,
				DNA::SkinColor::GiggaNigga,
				DNA::SkinColor::Moreno,
				DNA::SkinColor::Moreno2
			};
			if (attackOnSight.count(context.getPlayer()->getStats().soul->getDNA().getSkinColor()))
			{
				if (context.rng().chance(1.f / 300.f))
				{
					context.getNPC().makeTextBubble("Shitskin sighted! All units attack!");
				}
				return true;
			}
			return context.getWorld()->getPolice().getSeverity() != PoliceForce::SeverityThresholds::None;
		}),
		unique<IfElse>(
			unique<CustomCondition>([](BehaviorContext& context) { return context.getScene() == context.getPlayer()->getScene(); }),
			std::move(chasingSameScene),
			unique<WalkTo>(unique<CustomTargetLocator>([](BehaviorContext& context) {
				// Run to player if we're not in the same scene as him.
				return context.getWorld()->getPolice().getLastKnownLocation();
			})),
			true
		),
		std::move(patrolling),
		true
	));

	return std::move(root);
}

} // ai

CopPatrolState::CopPatrolState(NPC& owner, Car *car)
	:NPCState(owner)
	,car(car)
	,police(nullptr)
	,state(State::Normal)
	,cooldown(-1)
	,behaviorTree(ai::createCopBehaviorTree())
{
	const std::string dir = owner.getSoul()->appearance.getResourcePath() + "/shoot/" + owner.getSoul()->appearance.getClothing(ClothingType::Outfit) + "/";
	const ColorSwapID swaps = owner.getSoul()->appearance.getBodySwapID() | owner.getSoul()->appearance.getHairSwapID();

	loadWalkAnims();

	owner.addDrawableToSet("shoot_east", LayeredSpriteLoader::create(LayeredSpriteLoader::Config(owner.getSoul()->appearance, "shoot", 32, Vector2i(12, 16)).animate(3, false), "side", false));
	owner.addDrawableToSet("shoot_west", LayeredSpriteLoader::create(LayeredSpriteLoader::Config(owner.getSoul()->appearance, "shoot", 32, Vector2i(12, 16)).animate(3, false), "side", true));

	owner.setDrawableSet("south");

	if (car)
	{
		owner.getSoul()->setKnownEntity("patrol_car", car);
	}
#ifdef SL_DEBUG
	owner.debugDrawBehaviorTree = true;
#endif // SL_DEBUG
}

void CopPatrolState::update()
{
	const Town& town = *static_cast<Town*>(owner.getScene());
	Player *player = town.getPlayer();
	/*const bool chaseOnFoot = (police->getSeverity() != PoliceForce::None
					        && pointDistance(owner.getPos(), player->getPos()) < 269
					        && (!player->car || !isRoad(town.getTileAt(player->getPos().x, player->getPos().y))));
	*/
	//switch (police->getSeverity())
	//{
	//case PoliceForce::None:
	//	owner.minimapIcon.setColor(sf::Color::White);
	//	break;
	//case PoliceForce::Lookout:
	//	owner.minimapIcon.setColor(sf::Color::Yellow);
	//	break;
	//case PoliceForce::Pursuing:
	//	owner.minimapIcon.setColor(sf::Color(255, 128, 0));
	//	break;
	//case PoliceForce::Manhunt:
	//	owner.minimapIcon.setColor(sf::Color::Red);
	//	break;
	//}

	std::list<CrimeEntity*> crimes;
	owner.getScene()->cull(crimes, rectAround(owner.getPos(), 256));
	for (CrimeEntity *crime : crimes)
	{
		if (owner.canSee(crime->getPos()) && reportCrime(crime->getCrimeId()))
		{
			owner.makeTextBubble("Code red!");
		}
	}

#ifdef SL_DEBUG
	behaviorTree->debugExecute(owner.getBehaviorContext());
#else
	behaviorTree->execute(owner.getBehaviorContext());
#endif
#ifdef SL_DEBUG
	if (owner.debugDrawBehaviorTree && owner.getScene()->getViewRect().intersects(owner.getMask().getBoundingBox()))
	{
		behaviorTree->drawAsTree(*owner.getScene(), pos, 400);//owner.getScene()->getMousePos(), 400);
	}
#endif // SL_DEBUG
}

// private
void CopPatrolState::onEnter()
{
	police = &owner.getScene()->getWorld()->getPolice();
	setPathManager("StaticGeometry");
	loadWalkAnims();
}

void CopPatrolState::onExit()
{
	police = nullptr;
}

void CopPatrolState::handleEvent(const Event& event)
{
	if (event.getCategory() == Event::Category::NPCInteraction)
	{
		static int timer = 0;
		if (event.getName() == "talk" && --timer <= 0)
		{
			timer = 90;
			if (owner.getWorld()->getPolice().getSeverity() != PoliceForce::SeverityThresholds::None)
			{
				owner.makeTextBubble("Die, pedo scum!");
			}
			else
			{
				owner.makeTextBubble(Random::get().choose({ "Good day, Sir", "Keep an eye out for kidnappers, citizen" }));
			}
		}
	}
}

} // sl
