#include "NPC/Stats/Associations.hpp"

#include <cmath>
#include <queue>
#include <set>

#include "Engine/Serialization/AutoDefineObjectSerialization.hpp"
#include "Engine/Serialization/Impl/SerializeMatrix.hpp"
#include "NPC/Stats/Concepts.hpp"
#include "Utility/Assert.hpp"
#include "Utility/Mapping.hpp"

namespace sl
{
// TODO: why is this here? it's always empty...
const std::vector<std::string> Associations::allConcepts;



Associations::Associations()
	:Associations(Concepts::get())
{
}

Associations::Associations(const Concepts& concepts)
	:associationList(concepts.getConceptCount())
	,totalEdgeWeights(0.f)
	,concepts(concepts)
{
}

DEF_SERIALIZE_OBJ_FIELDS(Associations::Arc, &Associations::Arc::neighbor, &Associations::Arc::weight)
DEF_SERIALIZE_OBJ_FIELDS(Associations::Vertex, &Associations::Vertex::knowledge, &Associations::Vertex::associations)

void Associations::deserialize(Deserialize& in)
{
	ASSERT(&concepts == &Concepts::get());
	AutoDeserialize::fields(in, associationList, associationMatrix, totalEdgeWeights);
}

void Associations::serialize(Serialize& out) const
{
	ASSERT(&concepts == &Concepts::get());
	AutoSerialize::fields(out, associationList, associationMatrix, totalEdgeWeights);
}


// Edge predicates
Associations::EdgePred Associations::noInterCategoryEdges() const
{
	ERROR("not implemented");
	return[concepts = &this->concepts](const std::string& a, const std::string& b) -> bool
	{
		return true;
	};
}

// Rest
void Associations::associate(const std::string& concept1, const std::string& concept2, float weight)
{
	ASSERT(concepts.isRegisteredConcept(concept1));
	ASSERT(concepts.isRegisteredConcept(concept2));
	ASSERT(weight > 0.f && weight <= 1.f);
	uncheckedAssociate(concept1, concept2, std::clamp(weight, 0.f, 1.f));
}

void Associations::setConceptKnowledge(const std::string& concept, float newValue)
{
	ASSERT(0.f <= newValue && newValue <= 1.f);
	ASSERT(concepts.isRegisteredConcept(concept));
	const int id = concepts.getConceptIndex(concept);
	resizeIfNeeded(id);
	ASSERT(id >= 0 && id < associationList.size());
	associationList[id].knowledge = std::clamp(newValue, 0.f, 1.f);
}

void Associations::shiftConceptKnowledge(const std::string& concept, float amount)
{
	ASSERT(concepts.isRegisteredConcept(concept));
	const int id = concepts.getConceptIndex(concept);
	ASSERT(id >= 0 && id < associationList.size());
	associationList[id].knowledge = std::clamp(associationList[id].knowledge + amount, 0.f, 1.f);
}

float Associations::getConceptKnowledge(const std::string& concept) const
{
	ASSERT(concepts.isRegisteredConcept(concept));
	const int id = concepts.getConceptIndex(concept);
	ASSERT(id >= 0 && id < associationList.size());
	return associationList[id].knowledge;
}

float Associations::computeAssociation(const std::string& concept1, const std::string& concept2, const EdgePred& pred) const
{
	if (concept1 == concept2)
	{
		return 1.f;
	}
	return computeAssociation({ {concept1, 1.f} }, concept2, pred);
}

float Associations::computeAssociation(const std::vector<std::pair<std::string, float>>& rootConcepts, const std::string& target, const EdgePred& pred) const
{
	std::map<std::string, float> dist = computeCloselyAssociated(rootConcepts, {target}, 0.f, pred);
	auto it = dist.find(target);
	return it != dist.end() ? it->second : 0.f;
}

std::map<std::string, float> Associations::computeCloselyAssociated(const std::vector<std::pair<std::string, float>>& rootConcepts, const std::vector<std::string>& targetConcepts, float threshold, const EdgePred& pred) const
{
	Matrix<float> startState(1, associationList.size());
	float totalRootSum = 0.f;
	for (const std::pair<std::string, float>& rootConcept : rootConcepts)
	{
		totalRootSum += rootConcept.second;
	}
	for (const std::pair<std::string, float>& rootConcept : rootConcepts)
	{
		startState(0, concepts.getConceptIndex(rootConcept.first)) = rootConcept.second / totalRootSum;
	}
	std::map<std::string, float> combinedResults = markovFlow(startState, threshold, pred);
	for (const std::pair<std::string, float>& rootConcept : rootConcepts)
	{
		std::map<std::string, float> shortestResults =
			shortestDistance(
				concepts.getConceptIndex(rootConcept.first),
				threshold,
				mapSet(targetConcepts, std::bind(&Concepts::getConceptIndex, concepts, std::placeholders::_1)),
				pred);
		for (auto& a : shortestResults)
		{
			a.second /= totalRootSum;
			auto it = combinedResults.find(a.first);
			if (it != combinedResults.end())
			{
				it->second = (it->second + a.second) / 2.f;
			}
			else
			{
				combinedResults.insert(a);
			}
		}
	}
	return combinedResults;
}

// private
std::map<std::string, float> Associations::shortestDistance(int concept, float threshold, const std::set<int>& targets, const EdgePred& pred) const
{
	// Since all weights are in (0, 1] then we can do a log-transform to get multiplicative-longest-path
	// out of additive-shortest-path! Since if we maximize log(1/a) + log(1/b) + ... log(1/c)
	// (which is fine for dikjstra since log(1/x) is always non-negative for x in (0, 1] )
	// then it is maximizing log(1(a*b...c)) and thus 1/(a*b...c) and thus maximizing a*b*...c !

	std::map<int, float> dist;
	//std::map<std::string, int> undirectedDist;
	// in case it's an isolated concept
	if (associationList[concept].associations.empty())
	{
		return std::map<std::string, float>();
	}
	// since it's a max-heap, we use > instead of <
	auto distComp = [&dist](int lhs, int rhs) -> bool
	{
		return dist[lhs] > dist[rhs];
	};
	std::priority_queue<int, std::vector<int>, decltype(distComp)> queue(distComp);
	std::vector<int> visited(associationList.size(), 0);
	queue.push(concept);
	dist.insert(std::make_pair(concept, std::log(1.f)));
	int targetsVisited = targets.count(concept) == 0 ? 0 : 1;
	while (!queue.empty())
	{
		const int u = queue.top();
		queue.pop();
		if (visited[u] == 1)
		{
			continue;
		}
		visited[u] = 1;
		if (targets.count(u))
		{
			if (++targetsVisited == targets.size())
			{
				break;
			}
		}
		const float uDist = dist.at(u);
		// can be assumed to not be an isolated concept as graph is undirected and the root is checked earlier
		for (const Arc& arc : associationList[u].associations)
		{
			if (visited[arc.neighbor] == 0)
			{
				const float vDist = uDist + std::log(1.f / arc.weight);
				if (1.f / std::exp(vDist) >= threshold)
				{
					auto it = dist.find(arc.neighbor);
					if (it != dist.end())
					{
						if (vDist < it->second)
						{
							it->second = vDist;
							queue.push(arc.neighbor);
						}
					}
					else
					{
						dist[arc.neighbor] = vDist;
						queue.push(arc.neighbor);
					}
				}
			}
		}
	}
	// inverse everything to make 1 = closest, 0 = furthest and to undo the log transformation
	std::map<std::string, float> ret;
	if (targets.empty())
	{
		for (const auto& d : dist)
		{
			ret[concepts.getConcept(d.first)] = 1.f / std::exp(d.second);
		}
	}
	else
	{
		for (const int concept : targets)
		{
			auto it = dist.find(concept);
			ASSERT(it != dist.end());
			ret[concepts.getConcept(concept)] = 1.f / std::expf(it->second);
		}
	}
	return std::move(ret);
}

std::map<std::string, float> Associations::markovFlow(const Matrix<float>& startState, float threshold, const EdgePred& pred) const
{
	ASSERT(isMatrixComputed());
	const int n = associationList.size();
	ASSERT(startState.getRows() == 1 && startState.getCols() == n);
	const float p = 0.1f;
	Matrix<float> decay(n, n);
	for (int i = 0; i < n; ++i)
	{
		decay(i, i) += (1.f - p);
		for (int j = 0; j < n; ++j)
		{
			decay(i, j) += p * startState(0, j);
		}
	}
	Matrix<float> m(associationMatrix);
	// block out edges by pred
	for (int r = 0; r < n; ++r)
	{
		for (int c = 0; c < n; ++c)
		{
			if (!pred(concepts.getConcept(r), concepts.getConcept(c)))
			{
				m(r, c) = 0.f;
			}
		}
	}
	m *= decay;
	// needs to be normalized for it to be a valid markov transition matrix
	for (int r = 0; r < n; ++r)
	{
		float sum = 0.f;
		for (int c = 0; c < n; ++c)
		{
			sum += m(r, c);
		}
		for (int c = 0; c < n; ++c)
		{
			m(r, c) /= sum;
		}
	}
	m.exponentiate(23);
	Matrix<float> result = startState * m;
	std::map<std::string, float> ret;
	for (int i = 0; i < n; ++i)
	{
		if (result(0, i) >= threshold)
		{
			ret.insert(std::make_pair(concepts.getConcept(i), result(0, i)));
		}
	}
	return std::move(ret);
}


void Associations::uncheckedAssociate(const std::string& concept1, const std::string& concept2, float weight)
{
	const int c1 = concepts.getConceptIndex(concept1);
	const int c2 = concepts.getConceptIndex(concept2);
	resizeIfNeeded(std::max(c1, c2));
	float newWeight;
	float oldWeight;
	ASSERT(0 <= c1 && c1 < associationList.size());
	ASSERT(0 <= c2 && c2 < associationList.size());
	auto it = std::find_if(associationList[c1].associations.begin(), associationList[c1].associations.end(), [c2](const Arc& arc) { return arc.neighbor == c2; });
	if (it != associationList[c1].associations.end())
	{
		oldWeight = it->weight;
		newWeight = std::clamp(it->weight + weight, 0.f, 1.f);
		it->weight = newWeight;
		it = std::find_if(associationList[c2].associations.begin(), associationList[c2].associations.end(), [c1](const Arc& arc) { return arc.neighbor == c1; });
		ASSERT(it != associationList[c2].associations.end());
		it->weight = newWeight;
	}
	else
	{
		oldWeight = 0.f;
		newWeight = weight;
		associationList[c1].associations.push_back(Arc{c2, weight});
		associationList[c2].associations.push_back(Arc{c1, weight});
	}
	if (isMatrixComputed())
	{
		associationMatrix(c1, c2) = newWeight;
		associationMatrix(c2, c1) = newWeight;
	}
	totalEdgeWeights += (weight - oldWeight);
}

void Associations::computeMatrix() const
{
	const int n = concepts.getConceptCount();
	associationMatrix = Matrix<float>(n, n, 0.f);
	for (int u = 0; u < n; ++u)
	{
		associationMatrix(u, u) = 1.f;
		for (int i = 0; i < associationList[u].associations.size(); ++i)
		{
			const Arc& arc = associationList[u].associations[i];
			associationMatrix(u, arc.neighbor) = arc.weight;
			// ensure undirected
			ASSERT(associationMatrix(arc.neighbor, u) == 0 || associationMatrix(arc.neighbor, u) == arc.weight);
		}
	}
}

bool Associations::isMatrixComputed() const
{
	return associationMatrix.getCols() == concepts.getConceptCount();
}

void Associations::resizeIfNeeded(int id)
{
	if (associationList.size() <= id)
	{
		associationList.resize(id + 1);
	}
}

} // sl
