/**
 * @section DESCRIPTION
 * An association-graph for what concepts an NPC associates with other concepts
 * Associations between concepts are treated as 1 if they're completely related
 * and 0 if they're completely unrelated
 */
#ifndef SL_ASSOCIATIONS_HPP
#define SL_ASSOCIATIONS_HPP

#include <map>
#include <set>
#include <string>
#include <utility>
#include <functional>
#include <vector>

#include "Utility/Matrix.hpp"

namespace sl
{

class Concepts;
class Deserialize;
class Serialize;

namespace test
{
class AssociationsTest;
}

class Associations
{
public:
	
	Associations();

	Associations(const Concepts& concepts);
	

	void deserialize(Deserialize& in);

	void serialize(Serialize& out) const;



	// Edge predicates
	typedef std::function<bool(const std::string&, const std::string&)> EdgePred;
	
	static bool allEdges(const std::string&, const std::string&) { return true; }
	
	EdgePred noInterCategoryEdges() const;

	// Rest
	void associate(const std::string& concept1, const std::string& concept2, float weight);

	void setConceptKnowledge(const std::string& concept, float newValue);
	
	void shiftConceptKnowledge(const std::string& concept, float amount);

	float getConceptKnowledge(const std::string& concept) const;

	/**
	 * Computes the association between two concepts
	 * @param concept1 The first concept
	 * @param concept2 The second concept
	 * @param pred An edge-predicate that is expected to return true if a concept-concept direction association is allowable - defaults to allowing everything
	 * @return A closeness value with 1 being completely associated and 0 being completely unassociated.
	 */
	float computeAssociation(const std::string& concept1, const std::string& concept2, const EdgePred& pred = allEdges) const;
	
	/**
	* Computes the association between two concepts
	* @param rootConcepts The concepts to relate {@param target} to
	* @param target The concept to relate to
	* @param pred An edge-predicate that is expected to return true if a concept-concept direction association is allowable - defaults to allowing everything
	* @return A closeness value with 1 being completely associated and 0 being completely unassociated.
	*/
	float computeAssociation(const std::vector<std::pair<std::string, float>>& rootConcepts, const std::string& target, const EdgePred& pred = allEdges) const;

	static const std::vector<std::string> allConcepts;

	/**
	 * Computes a list of concepts related to the input concept subject to a given threshold and edge-predicate
	 * @param rootConcepts The root concepts to search from, and their weight (this will be normalized!)
	 * @param threshold The threshold to stop searching after. Only include <Concept, Association> pairs if Association >= theshhold
	 * @param targetHint If non-empty, the concepts you are interested in as a return. If empty, it returns all.
	 * @param pred An edge-predicate that is expected to return true if a concept-concept direction association is allowable - defaults to allowing everything
	 * @return A <Concept, Association> map of all closely related concepts to the input concept
	 */
	std::map<std::string, float> computeCloselyAssociated(const std::vector<std::pair<std::string, float>>& rootconcepts, const std::vector<std::string>& targetConcepts = allConcepts, float threshold = 0.f, const EdgePred& pred = allEdges) const;


	// Do not refer to these, only public for serialization
	struct Arc
	{
		//! Neighbor arc (Concepts' id)
		int neighbor;
		//! [0, 1] weight for the arc
		float weight;
	};
	struct Vertex
	{
		//! [0, 1] depending on how well the NPC knows the concept. Should usually be close to 0 or close to 1.
		float knowledge = 0.f;
		std::vector<Arc> associations;
	};
private:

	std::map<std::string, float> shortestDistance(int concept, float threshold, const std::set<int>& targets, const EdgePred& pred) const;

	std::map<std::string, float> markovFlow(const Matrix<float>& startState, float threshold, const EdgePred& pred) const;

	void uncheckedAssociate(const std::string& concept1, const std::string& concept2, float weight);

	// since we don't know how big the matrix should be while constructing
	void computeMatrix() const;

	bool isMatrixComputed() const;

	void resizeIfNeeded(int id);


	std::vector<Vertex> associationList;
	mutable Matrix<float> associationMatrix;
	float totalEdgeWeights;
	const Concepts& concepts;

	friend class Concepts;
	friend class test::AssociationsTest;
};

} // sl

#endif // SL_ASSOCIATIONS_HPP