#include "NPC/Stats/Concepts.hpp"

#include "Home/DungeonObjectDatabase.hpp"
#include "NPC/Stats/ColorPref.hpp"
#include "NPC/Stats/Mood.hpp"
#include "Utility/Assert.hpp"
#include "Utility/JsonUtil.hpp"

namespace sl
{

Concepts::Concepts(bool loadFile)
	:defaultAssociations(*this)
{
	if (loadFile)
	{
		// --------------------------------------------------------------------
		// load list of known concepts and their associations based on what
		// categories they fall into
		std::string err;
		auto knownConcepts = readJsonFile("data/concepts.json", err);
		if (!err.empty())
		{
			ERROR(err);
		}
		std::function<void(const json11::Json&, const char*)> addCategories = [&](const json11::Json& json, const char *parent)
		{
			if (json.is_string())
			{
				registerConcept(json.string_value());
				if (parent)
				{
					defaultAssociations.uncheckedAssociate(parent, json.string_value(), 1);
				}
			}
			else
			{
				ASSERT(json.is_object());
				const std::string& category = json.object_items().at("category").string_value();
				registerConcept(category);
				for (const json11::Json& subcategory : json.object_items().at("members").array_items())
				{
					addCategories(subcategory, category.c_str());
				}
			}
		};
		for (const json11::Json& rootItem : knownConcepts.array_items())
		{
			addCategories(rootItem, nullptr);
		}

		// --------------------------------------------------------------------
		//                  add in default associations
		auto knownAssociations = readJsonFile("data/base_associations.json", err);
		if (!err.empty())
		{
			ERROR(err);
		}
		for (const json11::Json& relation : knownAssociations.array_items())
		{
			const json11::Json::array& edge = relation.array_items();
			ASSERT(edge.size() == 2 || (edge.size() == 3 && edge[2].is_number()));
			ASSERT(edge[0].is_string() && edge[1].is_string());
			const int weight = edge.size() == 2 ? 1 : edge[2].number_value();
			defaultAssociations.uncheckedAssociate(edge[0].string_value(), edge[1].string_value(), weight);
		}

		// --------------------------------------------------------------------
		//         add in associations based on all furniture attributes
		for (const auto& dungeonObject : DungeonObjectDatabase::get())
		{
			registerConcept(dungeonObject.first);
			defaultAssociations.uncheckedAssociate(dungeonObject.first, getInfo(dungeonObject.second.color).name, 1);
		}

		// now do a sanity check that everything in defaultAssociations is stored as known concepts
		//for (const std::pair<int, float>& u : defaultAssociations.associationList)
		//{
		//	// graph is undirected so no need to check both vertices of an edge
		//	ASSERT(isRegisteredConcept(getConcept(u.first)));
		//}
		defaultAssociations.computeMatrix();
	}
}

const Associations& Concepts::makeDefault() const
{
	return defaultAssociations;
}

bool Concepts::isRegisteredConcept(const std::string& concept) const
{
	return registeredConcepts.find(concept) != registeredConcepts.end();
}

int Concepts::getConceptIndex(const std::string& concept) const
{
	auto it = registeredConcepts.find(concept);
	if (it == registeredConcepts.end())
	{
		ERROR("concept '" + concept + "' not registered");
		return -1;
	}
	return it->second;
}

const std::string& Concepts::getConcept(int index) const
{
	ASSERT(index >= 0 && index < registeredConceptsFromIndex.size());
	return registeredConceptsFromIndex[index];
}

int Concepts::getConceptCount() const
{
	return registeredConcepts.size();
}

/*static*/const Concepts& Concepts::get()
{
	static Concepts concepts(true);
	return concepts;
}

/*static*/Concepts Concepts::makeEmpty()
{
	return Concepts(false);
}

// private
void Concepts::registerConcept(const std::string& concept)
{
	auto result = registeredConcepts.insert(std::make_pair(concept, registeredConceptsFromIndex.size()));
	if (!result.second)
	{
		ERROR("concept '" + concept + "' already registered");
	}
	else
	{
		registeredConceptsFromIndex.push_back(concept);
	}
}

} // sl
