#include "NPC/Stats/DNADatabase.hpp"

#include "Utility/Assert.hpp"
#include "Utility/Random.hpp"
#include <fstream>
#include <iostream>
#include "NPC/Stats/Soul.hpp"
#include "Utility/Memory.hpp"
namespace sl
{

std::unique_ptr<DNADatabase> DNADatabase::database;

DNADatabase::DNADatabase()
{
	load();
}

DNA DNADatabase::generateRandomDNA(Random& random, GenePool genePool, bool isMale) const
{
	DNA ret;
	for (int i = 0; i < DNA_SIZE; ++i)
	{
		const float dominantProbability = getDominantProbability(static_cast<DNA::Index>(i), genePool);
		//ASSERT(dominantProbability >= 0.f && dominantProbability <= 1.f);
		if (random.chance(dominantProbability))
		{
			ret.genes.set(i);
		}
	}
	if (isMale)
	{
		ret.genes.set(DNA::Gender);
	}
	return ret;
}

const DNADatabase::Gene& DNADatabase::getEntry(DNA::Index index) const
{
	ASSERT(index >= 0 && index < DNA_SIZE);
	return entries[index];
}

const DNADatabase::Allele& DNADatabase::getGene(const DNA& dna, DNA::Index index) const
{
	ASSERT(index >= 0 && index < DNA_SIZE);
	return dna.genes[index] ? entries[index].dominant : entries[index].recessive;
}

void DNADatabase::load()
{
	//----------------------------------------------------------------------------
	//                            Gene info
	entries[DNA::Gender].dominant.name = "male (Y chromosome)";
	entries[DNA::Gender].recessive.name = "female (X chromosome)";

	// HAIR GENES
	entries[DNA::Redhead].dominant.name = "melanocortin-1 receptor absent";
	entries[DNA::Redhead].dominant.description = "non-red hair";
	entries[DNA::Redhead].recessive.name = "melanocortin-1 receptor present";
	entries[DNA::Redhead].recessive.description = "red hair";
	entries[DNA::Redhead].recessive.effects.strength = 1;
	entries[DNA::Redhead].recessive.effects.intelligence = -1;
	entries[DNA::Redhead].recessive.effects.willpower = 1;


	entries[DNA::Hair1].dominant.name = "Dark pigmentation #1";

	entries[DNA::Hair1].recessive.name = "Light pigmentation #1";
	entries[DNA::Hair1].recessive.effects.personalityModifiers.openToCautious = 0.02f;
	entries[DNA::Hair1].recessive.effects.beauty = 2;
	entries[DNA::Hair1].recessive.effects.intelligence = -2;


	entries[DNA::Hair2].dominant.name = "Dark pigmentation #2";

	entries[DNA::Hair2].recessive.name = "Light pigmentation #2";
	entries[DNA::Hair2].recessive.effects.personalityModifiers.openToCautious = 0.04f;
	entries[DNA::Hair2].recessive.effects.beauty = -1;
	entries[DNA::Hair2].recessive.effects.intelligence = 4;


	entries[DNA::Hair3].dominant.name = "Dark pigmentation #3";

	entries[DNA::Hair3].recessive.name = "Light pigmentation #3";
	entries[DNA::Hair3].recessive.effects.personalityModifiers.openToCautious = 0.03f;
	entries[DNA::Hair3].recessive.effects.beauty = 1;


	entries[DNA::Hair4].dominant.name = "Dark pigmentation #4";

	entries[DNA::Hair4].recessive.name = "Light pigmentation #4";
	entries[DNA::Hair4].recessive.effects.personalityModifiers.openToCautious = 0.05f;
	entries[DNA::Hair4].recessive.effects.beauty = 2;
	entries[DNA::Hair4].recessive.effects.tallness = 1;




	// SKIN GENES
	entries[DNA::SkinDark1].dominant.name = "Dark skin pigmentation #1";
	entries[DNA::SkinDark1].dominant.effects.intelligence = -3;
	entries[DNA::SkinDark1].dominant.effects.personalityModifiers.organizedToCareless = 0.06f;
	entries[DNA::SkinDark1].dominant.effects.analTightness = 1;

	entries[DNA::SkinDark1].recessive.name = "Light skin pigmentation #1";
	entries[DNA::SkinDark1].recessive.effects.intelligence = 2;
	entries[DNA::SkinDark1].recessive.effects.personalityModifiers.organizedToCareless = -0.06f;
	entries[DNA::SkinDark1].recessive.effects.personalityModifiers.extrovertToIntrovert = 0.8f;


	entries[DNA::SkinDark2].dominant.name = "Dark skin pigmentation #2";

	entries[DNA::SkinDark2].recessive.name = "Light skin pigmentation #2";


	entries[DNA::SkinHue1].dominant.name = "Skin hue dominant";

	entries[DNA::SkinHue1].recessive.name = "Skin hue recessive";


	entries[DNA::SkinTint1].dominant.name = "Skin tint dominant";

	entries[DNA::SkinTint1].recessive.name = "Skin tint recessive";



	// PERSONALITY GENES
	entries[DNA::Whore].recessive.name = "Reduced libido";
	entries[DNA::Whore].recessive.effects.vaginalDepth = -1;
	entries[DNA::Whore].recessive.effects.vaginalTightness = -2;
	entries[DNA::Whore].recessive.effects.analTightness = -2;
	entries[DNA::Whore].recessive.effects.oppai = -2;
	entries[DNA::Whore].recessive.effects.libido = -3;

	entries[DNA::Whore].dominant.name = "Increased libido";
	entries[DNA::Whore].dominant.effects.vaginalDepth = 8;
	entries[DNA::Whore].dominant.effects.vaginalTightness = 9;
	entries[DNA::Whore].dominant.effects.analTightness = 11;
	entries[DNA::Whore].dominant.effects.oppai = 5;
	entries[DNA::Whore].dominant.effects.libido = 23;
	entries[DNA::Whore].dominant.effects.gullibility = 3;
	entries[DNA::Whore].dominant.effects.beauty = 1;
	entries[DNA::Whore].dominant.effects.personalityModifiers.nervousToConfident = 0.13f;
	entries[DNA::Whore].dominant.effects.pubertyStart = -100;
	entries[DNA::Whore].dominant.effects.personalityModifiers.friendlyToUnkind = -0.13f;

	entries[DNA::Bimbo].recessive.name = "Bimbo";
	entries[DNA::Bimbo].recessive.effects.beauty = 5;
	entries[DNA::Bimbo].recessive.effects.intelligence = -5;
	entries[DNA::Bimbo].recessive.effects.oppai = 25;
	entries[DNA::Bimbo].recessive.effects.personalityModifiers.nervousToConfident = 0.2f;
	entries[DNA::Bimbo].recessive.effects.personalityModifiers.friendlyToUnkind = -0.2f;

	entries[DNA::Bimbo].dominant.name = "Not Bimbo";

	//-------------------------------------------------------------------------
	//                        Gene distribution pools

	// default = caucasian gene pool (at least for now?)
	for (int i = 0; i < DNA_SIZE; ++i)
	{
		defaultDominantProbabilities[i] = -1;
	}

	// roughly 4-6% of whites are redhead, let's go with 5%.
	// except double this since only 50% will have light enough hair to have red hair expressed
	// which means that 1/(x*x) = 10%, so we get 1/sqrt(10) as our probability
	defaultDominantProbabilities[DNA::Redhead] = 1 - (1 / sqrt(10.f));
	// hair colour here works so if you have 0 or 1 out of the 4 dark hair pigment genes
	// expressed then you're blonde. roughly 50% of europeans are blonde so we have
	// 4C0 + 4C1 (5) ways to get blonde and 2^4 (16) ways in total so 31% of combinations
	// result in blonde so we must multiply every probability by 5/8 if we want to get roughly
	// 50% of them to be the 8 in 16
	const float hairPercent = /*0.5f*/0.3f * 5.f / 8.f;
	defaultDominantProbabilities[DNA::Hair1] = hairPercent;
	defaultDominantProbabilities[DNA::Hair2] = hairPercent;
	defaultDominantProbabilities[DNA::Hair3] = hairPercent;
	defaultDominantProbabilities[DNA::Hair4] = hairPercent;
	defaultDominantProbabilities[DNA::SkinDark1] = 0.f;
	defaultDominantProbabilities[DNA::SkinDark2] = 0.f;
	defaultDominantProbabilities[DNA::SkinHue1] = 0.f;
	defaultDominantProbabilities[DNA::SkinTint1] = 0.25f;
	defaultDominantProbabilities[DNA::SkinTint2] = 0.5f;
	defaultDominantProbabilities[DNA::EyeCollagen] = 0.5f;
	defaultDominantProbabilities[DNA::LightEyes] = 0.5f;
	// 3/5 = 1/x*x  -->  x = sqrt(5 / 3) ( to give 60% blue)
	defaultDominantProbabilities[DNA::Herc2] = 1.f - 1.f / sqrt(5.f / 3.f);// 60%
	// is ~ 60% is light, then we want 50/60 -> 1 / (6/5) to have no Gey
	// to get 50% overall blue which means 1/6 to have Gey, so 1/x*x = 1/6, x = sqrt(6)
	defaultDominantProbabilities[DNA::Gey] = 1.f / sqrt(6.f);

	defaultDominantProbabilities[DNA::Whore] = 0.25f;
	defaultDominantProbabilities[DNA::Bimbo] = 0.5f;
	// TODO: implement blood types in earnest
	defaultDominantProbabilities[DNA::BloodA] = 0.5f;
	defaultDominantProbabilities[DNA::BloodB] = 0.5f;

	defaultDominantProbabilities[DNA::Defect1] = 0.01f;
	defaultDominantProbabilities[DNA::Defect2] = 0.01f;
	defaultDominantProbabilities[DNA::Defect3] = 0.01f;
	defaultDominantProbabilities[DNA::Defect4] = 0.01f;
	defaultDominantProbabilities[DNA::Defect5] = 0.01f;
	defaultDominantProbabilities[DNA::Defect6] = 0.01f;
	defaultDominantProbabilities[DNA::Defect7] = 0.01f;
	defaultDominantProbabilities[DNA::Defect8] = 0.01f;
	defaultDominantProbabilities[DNA::Defect9] = 0.01f;


	DominantProbabilityMap& nordicGenes = dominantProbabilitiesByGenePool[GenePool::NordicCaucasian];
	// pulled out of ass, let's just assume there's more blondes...
	nordicGenes[DNA::Redhead] = 1.f - 0.15f;
	// using a similar argument to the nordic eye stuff below, let's just set these the same
	nordicGenes[DNA::Hair1] = 0.3f;//0.33;
	nordicGenes[DNA::Hair2] = 0.1f;//0.33;
	nordicGenes[DNA::Hair3] = 0.3f;//0.33;
	nordicGenes[DNA::Hair4] = 0.0f;
	nordicGenes[DNA::EyeCollagen] = 0.8f;
	nordicGenes[DNA::LightEyes] = 0.4f;
	// 75% light eyes, so rate of dominant is 25% in expressed, or 1 in 4
	// so 1/x*x = 3/4  -> x = sqrt(4/3)
	nordicGenes[DNA::Herc2] = 1.f - 1.f / sqrt(4.f / 3.f);
	nordicGenes[DNA::Gey] = 0.1f;
	nordicGenes[DNA::SkinDark1] = 0.f;
	nordicGenes[DNA::SkinDark2] = 0.f;
	nordicGenes[DNA::SkinTint1] = 0.f;
	nordicGenes[DNA::SkinTint2] = 0.3f;
	nordicGenes[DNA::Whore] = 0.2f;

	DominantProbabilityMap& mediteranianGenes = dominantProbabilitiesByGenePool[GenePool::MediterraneanCaucasian];
	mediteranianGenes[DNA::Redhead] = 1.f - 0.1f;
	// if 50% of combinations result in blonde and we want 25% to be blonde then let's make the ratiof dark:light
	// genes 2:1
	mediteranianGenes[DNA::Hair1] = 0.67f;
	mediteranianGenes[DNA::Hair2] = 0.67f;
	mediteranianGenes[DNA::Hair3] = 0.67f;
	mediteranianGenes[DNA::SkinDark1] = 0.f;
	mediteranianGenes[DNA::SkinDark2] = 0.4f;
	mediteranianGenes[DNA::SkinHue1] = 0.f;
	mediteranianGenes[DNA::SkinTint1] = 0.25f;
	mediteranianGenes[DNA::SkinTint2] = 0.85f;
	// we also want ~20-25% to have light eyes so 1/x*x = 1/4 -> x =sqrt(4)
	mediteranianGenes[DNA::Herc2] = 1.f - 1.f / sqrt(4.f);
	// 18% are blue, so (25-18)/25 need to have no Gey, which means that
	// 1/x*x = 1/(25/7) -> x = sqrt(25/7);
	mediteranianGenes[DNA::Gey] = 1.f - 1.f / sqrt(25.f / 7.f);

	DominantProbabilityMap& asianGenes = dominantProbabilitiesByGenePool[GenePool::Asian];
	// ALL ROOK SAME
	asianGenes[DNA::Redhead] = 1.f;
	asianGenes[DNA::Hair1] = 1.f;
	asianGenes[DNA::Hair2] = 1.f;
	asianGenes[DNA::Hair3] = 0.7f; // let's just give them a small chance of dark brown instead of black
	asianGenes[DNA::Hair4] = 1.f;
	asianGenes[DNA::EyeCollagen] = 0.8f;
	asianGenes[DNA::LightEyes] = 0.1f;
	asianGenes[DNA::Herc2] = 0.95f;
	asianGenes[DNA::Gey] = 0.99f;
	asianGenes[DNA::SkinDark1] = 0.f;
	asianGenes[DNA::SkinDark2] = 0.f;
	asianGenes[DNA::SkinHue1] = 1.f;
	asianGenes[DNA::SkinTint1] = 0.05f;
	asianGenes[DNA::SkinTint2] = 0.3f;
	asianGenes[DNA::Whore] = 0.15f;

	DominantProbabilityMap& hispanicGenes = dominantProbabilitiesByGenePool[GenePool::Hispanic];
	hispanicGenes[DNA::Redhead] = 1.f;
	hispanicGenes[DNA::Hair1] = 0.9f;
	hispanicGenes[DNA::Hair2] = 0.9f;
	hispanicGenes[DNA::Hair3] = 0.9f;
	hispanicGenes[DNA::Hair4] = 1.f;
	hispanicGenes[DNA::EyeCollagen] = 0.1f;
	hispanicGenes[DNA::LightEyes] = 0.4f;
	hispanicGenes[DNA::Herc2] = 0.8f;
	hispanicGenes[DNA::Gey] = 0.9f;
	hispanicGenes[DNA::SkinDark1] = 0.9f;
	hispanicGenes[DNA::SkinDark2] = 0.2f;
	hispanicGenes[DNA::SkinHue1] = 0.f;
	hispanicGenes[DNA::SkinTint1] = 0.4f;
	hispanicGenes[DNA::SkinTint2] = 0.5f;
	hispanicGenes[DNA::Whore] = 0.35f;

	DominantProbabilityMap& tropicalGenes = dominantProbabilitiesByGenePool[GenePool::Tropical];
	tropicalGenes[DNA::Redhead] = 1.f;
	tropicalGenes[DNA::Hair1] = 0.5f;
	tropicalGenes[DNA::Hair2] = 1.f;
	tropicalGenes[DNA::Hair3] = 1.f;
	tropicalGenes[DNA::Hair4] = 1.f;
	tropicalGenes[DNA::EyeCollagen] = 0.9f;
	tropicalGenes[DNA::LightEyes] = 0.9f;
	tropicalGenes[DNA::Herc2] = 0.7f; // 90% chance of brown. otherwise likely green/hazel.
	tropicalGenes[DNA::Gey] = 0.9f; // 1% chance of light (blue/grey/etc) eyes.
	tropicalGenes[DNA::SkinDark1] = 1.f;
	tropicalGenes[DNA::SkinDark2] = 0.f;
	tropicalGenes[DNA::SkinHue1] = 0.9f;
	tropicalGenes[DNA::SkinTint1] = 0.05f;
	tropicalGenes[DNA::SkinTint2] = 0.15f;
	tropicalGenes[DNA::Whore] = 0.4f;

	DominantProbabilityMap& niggerGenes = dominantProbabilitiesByGenePool[GenePool::Nigger];
	niggerGenes[DNA::Redhead] = 1.f;
	niggerGenes[DNA::Hair1] = 1.f;
	niggerGenes[DNA::Hair2] = 0.9f;
	niggerGenes[DNA::Hair3] = 1.f;
	niggerGenes[DNA::Hair4] = 1.f;
	niggerGenes[DNA::EyeCollagen] = 0.1f;
	niggerGenes[DNA::LightEyes] = 0.1f;
	niggerGenes[DNA::Herc2] = 0.9f;
	niggerGenes[DNA::Gey] = 1.f;
	niggerGenes[DNA::SkinDark1] = 1.f;
	niggerGenes[DNA::SkinDark2] = 1.f;
	niggerGenes[DNA::SkinHue1] = 1.f;
	niggerGenes[DNA::SkinTint1] = 0.3f;
	niggerGenes[DNA::SkinTint2] = 0.4f;
	niggerGenes[DNA::Whore] = 0.5f;

	DominantProbabilityMap& indianGenes = dominantProbabilitiesByGenePool[GenePool::Indian];
	indianGenes[DNA::Redhead] = 1.f;
	indianGenes[DNA::Hair1] = 1.f;
	indianGenes[DNA::Hair2] = 1.f;
	indianGenes[DNA::Hair3] = 0.8f;
	indianGenes[DNA::Hair4] = 1.f;
	indianGenes[DNA::EyeCollagen] = 0.f;
	indianGenes[DNA::LightEyes] = 0.3f;
	indianGenes[DNA::Herc2] = 0.85f;
	indianGenes[DNA::Gey] = 0.95f;
	indianGenes[DNA::SkinDark1] = 0.f;
	indianGenes[DNA::SkinDark2] = 1.f;
	indianGenes[DNA::SkinHue1] = 1.f;
	indianGenes[DNA::SkinTint1] = 0.5f;
	indianGenes[DNA::SkinTint2] = 0.6f;
	indianGenes[DNA::Whore] = 0.2f;

	DominantProbabilityMap& arabGenes = dominantProbabilitiesByGenePool[GenePool::Arabic];
	arabGenes[DNA::Redhead] = 1.f;
	arabGenes[DNA::Hair1] = 1.f;
	arabGenes[DNA::Hair2] = 1.f;
	arabGenes[DNA::Hair3] = 0.5f;
	arabGenes[DNA::Hair4] = 1.f;
	arabGenes[DNA::EyeCollagen] = 0.2f;
	arabGenes[DNA::LightEyes] = 0.2f;
	arabGenes[DNA::Herc2] = 0.35f;
	arabGenes[DNA::Gey] = 0.95f;
	arabGenes[DNA::SkinDark1] = 0.f;
	arabGenes[DNA::SkinDark2] = 1.f;
	arabGenes[DNA::SkinHue1] = 1.f;
	arabGenes[DNA::SkinTint1] = 0.f;
	arabGenes[DNA::SkinTint2] = 0.35f;
	arabGenes[DNA::Whore] = 0.1f;

	DominantProbabilityMap& seasianGenes = dominantProbabilitiesByGenePool[GenePool::SouthEastAsian];
	seasianGenes[DNA::Redhead] = 1.f;
	seasianGenes[DNA::Hair1] = 1.f;
	seasianGenes[DNA::Hair2] = 1.f;
	seasianGenes[DNA::Hair3] = 0.8f;
	seasianGenes[DNA::Hair4] = 1.f;
	seasianGenes[DNA::EyeCollagen] = 0.1f;
	seasianGenes[DNA::LightEyes] = 0.5f;
	seasianGenes[DNA::Herc2] = 0.8f;
	seasianGenes[DNA::Gey] = 1.f;
	seasianGenes[DNA::SkinDark1] = 0.f;
	seasianGenes[DNA::SkinDark2] = 0.f;
	seasianGenes[DNA::SkinHue1] = 1.f;
	seasianGenes[DNA::SkinTint1] = 1.f;
	seasianGenes[DNA::SkinTint2] = 0.5f;
	seasianGenes[DNA::Whore] = 0.6f;
}


void DNADatabase::test(Random& random, int generations, int populationSize, const char *nationality) const
{
	typedef std::vector<std::unique_ptr<Soul>> Generation;

	ASSERT(generations >= 1);
	ASSERT(populationSize >= 1);

	std::ofstream out("genetic test.txt");
	out << "[Generations: " << generations << " | Population Size:" << populationSize << "]\n\n";

	std::vector<Generation> people;

	auto generationPrinter = [&] (int generation)
	{
		std::map<DNA::Index, int> RRCount;
		std::map<DNA::Index, int> rRCount;
		std::map<DNA::Index, int> RrCount;
		std::map<DNA::Index, int> rrCount;
		std::map<DNA::HairColor, int> hairCount;
		std::map<DNA::SkinColor, int> skinCount;
		std::map<DNA::EyeColor, int> eyeCount;

		for (const std::unique_ptr<Soul>& soul : people.back())
		{
			for (int g = static_cast<int>(DNA::Redhead); g <= static_cast<int>(DNA::Gey); ++g)
			{
				const DNA::Index gene = static_cast<DNA::Index>(g);
				if (soul->fatherDNA.genes[g])
				{
					if (soul->motherDNA.genes[g])
					{
						++RRCount[gene];
					}
					else
					{
						++RrCount[gene];
					}
				}
				else if (soul->motherDNA.genes[g])
				{
					++rRCount[gene];
				}
				else
				{
					++rrCount[gene];
				}
			}
			++hairCount[soul->expressedDNA.getHairColor()];
			++eyeCount[soul->expressedDNA.getEyeColor()];
			++skinCount[soul->expressedDNA.getSkinColor()];
		}

		out << "\n\nGeneration " << generation << std::endl;
		out << "\t[Legend: RR | Rr/rR | rr]\n";
		auto printer = [&](DNA::Index gene, const char* name)
		{
			out << name << ": " << RRCount[gene] * 100 / populationSize << "% | "<< (RrCount[gene] + rRCount[gene]) * 100 / populationSize << "% | "<< rrCount[gene] * 100 / populationSize << "%\n";
		};
		printer(DNA::Redhead, "Redhead");
		printer(DNA::Hair1, "Hair1");
		printer(DNA::Hair2, "Hair2");
		printer(DNA::Hair3, "Hair3");
		printer(DNA::Hair4, "Hair4");
		printer(DNA::SkinDark1, "SkinDark1");
		printer(DNA::SkinDark2, "SkinDark2");
		printer(DNA::SkinHue1, "SkinHue1");
		printer(DNA::SkinTint1, "SkinTint1");
		printer(DNA::Herc2, "Herc2");
		printer(DNA::Gey, "Gey");
		out << std::endl;

		out << "Displayed eye colour:\n";
		out << "\tBlue:  " << eyeCount[DNA::EyeColor::Blue] * 100 / populationSize << "%\n";
		out << "\tBrown: " << eyeCount[DNA::EyeColor::Brown] * 100 / populationSize << "%\n";
		out << "\tGreen: " << eyeCount[DNA::EyeColor::Green] * 100 / populationSize << "%\n";
		out << "\tBlack:  " << eyeCount[DNA::EyeColor::Black] * 100 / populationSize << "%\n";
		out << "\tGrey:  " << eyeCount[DNA::EyeColor::Grey] * 100 / populationSize << "%\n";
		out << "\tHazel: " << eyeCount[DNA::EyeColor::Hazel] * 100 / populationSize << "%\n";
		out << "\tCyan:  " << eyeCount[DNA::EyeColor::Cyan] * 100 / populationSize << "%\n";
		out << std::endl;

		out << "Displayed hair colour:\n";
		out << "\tPlatinum Blonde:      " << hairCount[DNA::HairColor::PlatinumBlonde] * 100 / populationSize << "%\n";
		out << "\tBlonde:              " << hairCount[DNA::HairColor::Blonde] * 100 / populationSize << "%\n";
		out << "\tGolden Blonde:       " << hairCount[DNA::HairColor::GoldenBlonde] * 100 / populationSize << "%\n";
		out << "\tLight Brown Greyish: " << hairCount[DNA::HairColor::LightBrownGreyish] * 100 / populationSize << "%\n";
		out << "\tLight Brown:         " << hairCount[DNA::HairColor::LightBrown] * 100 / populationSize << "%\n";
		out << "\tDark Brown:          " << hairCount[DNA::HairColor::DarkBrown] * 100 / populationSize << "%\n";
		out << "\tBlack:               " << hairCount[DNA::HairColor::Black] * 100 / populationSize << "%\n";
		out << "\tRed:                 " << hairCount[DNA::HairColor::Red] * 100 / populationSize << "%\n";
		out << "\tLight Red:           " << hairCount[DNA::HairColor::LightRed] * 100 / populationSize << "%\n";
		out << std::endl;

		out << "Displayed skin colour:\n";
		out << "\tNordic:           " << skinCount[DNA::SkinColor::Nordic] * 100 / populationSize << "%\n";
		out << "\tWhite:            " << skinCount[DNA::SkinColor::Caucasian] * 100 / populationSize << "%\n";
		out << "\tHispanic:         " << skinCount[DNA::SkinColor::Hispanic] * 100 / populationSize << "%\n";
		out << "\tAsian:            " << skinCount[DNA::SkinColor::Asian] * 100 / populationSize << "%\n";
		out << "\tNigger:           " << skinCount[DNA::SkinColor::Nigger] * 100 / populationSize << "%\n";
		out << "\tTropical:         " << skinCount[DNA::SkinColor::Tropical] * 100 / populationSize << "%\n";
		out << "\tMoreno:           " << skinCount[DNA::SkinColor::Moreno] * 100 / populationSize << "%\n";
		out << "\tIndian:           " << skinCount[DNA::SkinColor::Indian] * 100 / populationSize << "%\n";
		out << "\tArabic:           " << skinCount[DNA::SkinColor::Arabic] * 100 / populationSize << "%\n";
		out << "\tMediterranean:     " << skinCount[DNA::SkinColor::Mediterranean] * 100 / populationSize << "%\n";
		out << "\tSouth East Asian: " << skinCount[DNA::SkinColor::SEAsian] * 100 / populationSize << "%\n";
		out << "\tSuper Nigger:     " << skinCount[DNA::SkinColor::ExtraNigger] * 100 / populationSize << "%\n";
		out << "\tMoreno 2:          " << skinCount[DNA::SkinColor::Moreno2] * 100 / populationSize << "%\n";
		out << "\tDark Moreno:      " << skinCount[DNA::SkinColor::DarkMoreno] * 100 / populationSize << "%\n";
		out << "\tDark Moreno 2:    " << skinCount[DNA::SkinColor::DarkMoreno2] * 100 / populationSize << "%\n";
		out << "\tDark Indian:      " << skinCount[DNA::SkinColor::DarkIndian] * 100 / populationSize << "%\n";
		out << "\tDark Arabic:      " << skinCount[DNA::SkinColor::DarkArabic] * 100 / populationSize << "%\n";
		out << "\tDark SE Asian:    " << skinCount[DNA::SkinColor::DSEA] * 100 / populationSize << "%\n";
		out << "\tGigga Nigga:      " << skinCount[DNA::SkinColor::GiggaNigga] * 100 / populationSize << "%\n";
		out << "\tDark Asian:       " << skinCount[DNA::SkinColor::DarkAsian] * 100 / populationSize << "%\n";
	};

	people.push_back(Generation());
	people.back().reserve(populationSize);
	std::cout << "Creating initial population: ";
	std::cout.flush();
	int creationProgress = 0;
	for (int i = 0; i < populationSize; ++i)
	{
		if (creationProgress < ((float)i/populationSize) * 50)
		{
			++creationProgress;
			std::cout << "|";
			std::cout.flush();
		}
		ERROR("fix this");
		//people.back().push_back(unique<Soul>(random, nationality));
	}
	std::cout << std::endl;

	generationPrinter(0);

	for (int i = 0; i < generations; ++i)
	{
		// don't bother shuffling yet... yay incest!
		if (generations > 0)
		{
			std::cout << "Starting generation " << (i + 1) << std::endl;
			std::cout << "Mating progress: ";
			std::cout.flush();
			int matingProgress = 0;
			people.push_back(Generation());
			Generation& current = people[people.size() - 2];
			Generation& next = people.back();
			next.reserve(populationSize);
			// mate arbitrarily with the person next to you and produce
			for (int j = 1; j < populationSize; j += 2)
			{
				next.push_back(unique<Soul>(*current[j], *current[j - 1], random));
				next.push_back(unique<Soul>(*current[j], *current[j - 1], random));
				if (matingProgress < ((float)j/populationSize) * 50)
				{
					++matingProgress;
					std::cout << "|";
					std::cout.flush();
				}
			}
			std::cout << std::endl;
		}

		generationPrinter(i + 1);
	}
	out.close();
}


float DNADatabase::getDominantProbability(DNA::Index index, GenePool genePool) const
{
	auto it = dominantProbabilitiesByGenePool[genePool].find(index);
	if (it != dominantProbabilitiesByGenePool[genePool].end())
	{
		return it->second;
	}
	return defaultDominantProbabilities[index];
}


/*		static		*/
DNADatabase& DNADatabase::get()
{
	if (!database)
	{
		// can't make_unique since private
		database.reset(new DNADatabase());
	}
	return *database;
}

} // sl
