/**
 * @section DESCRIPTION
 * A database containing information about all the genes, alleles, and the probabilities
 * of getting each version of alleles for every gene based on the input gene pool
 */
#ifndef SL_DNADATABASE_HPP
#define SL_DNADATABASE_HPP

#include <map>
#include <memory>

#include "NPC/Stats/DNA.hpp"
#include "NPC/Stats/GeneticStats.hpp"

namespace sl
{

class Random;

class DNADatabase
{
public:
	struct Allele
	{
		//! Name of the allele
		std::string name;
		//! Description of the allele
		std::string description;
		//! Genetic effects of the allele
		GeneticStats effects;
	};
	struct Gene
	{
		//! The recessive allele of the gene
		Allele recessive;
		//! The dominant allele of the gene
		Allele dominant;
	};
	enum GenePool
	{
		Caucasian = 0,
		NordicCaucasian,
		MediterraneanCaucasian,
		Asian,
		Hispanic,
		Tropical,
		Nigger,
		Indian,
		SouthEastAsian,
		Arabic
	};
	static const int GENE_POOL_SIZE = 10;

	/**
	 * @param random The random number generator to use
	 * @param genePool The gene pool to get the allele probabilities from
	 * @param isMale true if male, false if female
	 * @return A random chromosome from the specified pool
	 */
	DNA generateRandomDNA(Random& random, GenePool genePool, bool isMale = false) const;

	/**
	 * @param index The gene to get info about
	 * @return Info about the dominant and recessive alleles for the given gene
	 */
	const Gene& getEntry(DNA::Index index) const;

	/**
	 * @param dna The chromosome to check for
	 * @param index The gene to check for
	 * @return Info about the expressed allele for the gene index in the input dna
	 */
	const Allele& getGene(const DNA& dna, DNA::Index index) const;

	/**
	 * Load the database (invalid if this has not been done)
	 */
	void load();

	/**
	 * Run a test using the database. Each pair of parents has 2 kids always.
	 * @param random The RNG to use
	 * @param generations How many generations to breed for
	 * @param populationSize The number of people at each generation
	 * @param nationality The natioonality to generate out of
	 */
	void test(Random& random, int generations, int populationSize, const char *nationality) const;

	/**
	 * @return singleton instance of a database. Lazy-loading and auto-initiating
	 */
	static DNADatabase& get();

private:
	DNADatabase();

	/**
	 * @param index The gene to check
	 * @param genePool The pool to check
	 * @return The probability that a given gene from a given pool will be the dominant allele
	 */
	float getDominantProbability(DNA::Index index, GenePool genePool) const;


	//! Information about all the genes
	Gene entries[DNA_SIZE];

	typedef std::map<DNA::Index, float> DominantProbabilityMap;
	//! Probability map for each pool for a gene having the dominant allele
	DominantProbabilityMap dominantProbabilitiesByGenePool[GENE_POOL_SIZE];
	//! Default values for each gene's dominant allele
	float defaultDominantProbabilities[DNA_SIZE];


	//! Singleton instance of database
	static std::unique_ptr<DNADatabase> database;
};

}

#endif
