#include "NPC/Stats/LoliGen.hpp"

#include <algorithm>
#include <cmath>
#include <initializer_list>
#include <vector>
#include "Utility/Assert.hpp"
#include "Utility/LinearInterpolator.hpp"

namespace sl
{

const float daysInYear = 365.f;

float expectedWeight(int ageDays, Gender gender)
{
	// @todo find polynomial for male weight
	ASSERT(ageDays >= 0);
	float ageYears = ageDays / daysInYear;
	/*
		This is a cubic function that approximates the curve found on statistics found at
		http://www.halls.md/on/girls-weight-w.htm from ages 4-18~
	*/
	if (ageYears < 3.7f)
	{
		// ????
		//ERROR("not defined");
	}
	else if (ageYears >= 18.f)
	{
		ageYears = 18.f;
	}
	return -0.02163f * pow(ageYears, 3) + 0.65609f * pow(ageYears, 2) - 2.20954f * ageYears + 13.77941f;
}

float expectedHeight(int ageDays, Gender gender)
{
	// @todo find polynomial for male height
	ASSERT(ageDays >= 0);
	/*
		This is a linear function that approximates the statistics found at
		http://www.halls.md/chart/girls-height-w.htm up to age 14~
	*/
	if (ageDays >= 14 * daysInYear)
	{
		ageDays = 14 * daysInYear;
	}
	return 102 + 59.f * (((ageDays - 60) / daysInYear - 4) / 10.f);
}

float bladderCapacity(int ageDays)
{
	ASSERT(ageDays >= 0);
	/*
		Taken from: https://www.brennerchildrens.org/Pediatric-Urology/HPT/How-Much-Should-a-Bladder-Hold.htm
		Along with data for adult bladder size from various sources (showing 400-600ml, so let's go with 500ml
		which gives the age of ~14.9 as the time when an adult size matches the formula below
		Note: same for both genders!
	*/
	return (std::min(ageDays / daysInYear, 14.9f) + 2.f) * 30.f;
}

float urineFlowRate(int ageDays, Gender gender)
{
	// @todo find data for toddlers? (data only gives for 4+yos)
	ASSERT(ageDays >= 0);
	const float ageYears = ageDays / daysInYear;
	/*
	 * Flow rate from: http://www.nlm.nih.gov/medlineplus/ency/article/003325.htm
	 */
	static const std::initializer_list<std::pair<float, float>> maleData = {
		std::make_pair(5.5f, 10.f),
		std::make_pair(10.5f, 12.f),
		std::make_pair(18.f, 21.f), // 12 -> 21 -> 12 was too big of a jump for LERP, so I put in 2 data points
		std::make_pair(40.f, 21.f),
		std::make_pair(55.f, 12.f),
		std::make_pair(73.f, 9.f)
	};
	static const LinearInterpolator<float> maleFlowRate(maleData);
	static const std::initializer_list<std::pair<float, float>> femaleData = {
		std::make_pair(5.5f, 10.f),
		std::make_pair(10.5f, 15.f),
		std::make_pair(30.f, 18.f),
		std::make_pair(55.f, 18.f),
		std::make_pair(73.f, 18.f)
	};
	static const LinearInterpolator<float> femaleFlowRate(femaleData);
	return 3.f * (gender == Gender::Male ? maleFlowRate : femaleFlowRate).interpolate(ageYears);
}

class CalorieReqs
{
public:
	CalorieReqs()
		:sedentary(-1)
		,moderatelyActive(-1)
		,active(-1)
	{
	}

	CalorieReqs(int all)
		:sedentary(all)
		,moderatelyActive(all)
		,active(all)
	{
	}

	CalorieReqs(int sedentary, int moderatelyActive, int active)
		:sedentary(sedentary)
		,moderatelyActive(moderatelyActive)
		,active(active)
	{
	}

	int sedentary;
	int moderatelyActive;
	int active;
};

CalorieReqs operator*(float scalar, const CalorieReqs& reqs)
{
	return CalorieReqs(reqs.sedentary * scalar, reqs.moderatelyActive * scalar, reqs.active * scalar);
}

CalorieReqs operator*(const CalorieReqs& reqs, float scalar)
{
	return operator*(scalar, reqs);
}

CalorieReqs operator+(const CalorieReqs& lhs, const CalorieReqs& rhs)
{
	return CalorieReqs(lhs.sedentary + rhs.sedentary, lhs.moderatelyActive + rhs.moderatelyActive, lhs.active + rhs.active);
}

CalorieReqs operator-(const CalorieReqs& lhs, const CalorieReqs& rhs)
{
	return CalorieReqs(lhs.sedentary - rhs.sedentary, lhs.moderatelyActive - rhs.moderatelyActive, lhs.active - rhs.active);
}

float calorieRequirements(int ageDays, ActivityLevel activityLevel, Gender gender)
{
	/*
	 * Based on stats from: http://www.webmd.com/diet/estimated-calorie-requirement
	 * and ones for <2 are gotten from 100kcal / kg of bodyweight and this: http://cdn5.kellymom.com/images/growth/growthcharts.gif
	 */
	static const std::initializer_list<std::pair<float, CalorieReqs>> maleData = {
		std::make_pair(0.f, CalorieReqs(350)), // first 3 derived from kellymom data
		std::make_pair(0.5f, CalorieReqs(800)),
		std::make_pair(1.f, CalorieReqs(950)),
		std::make_pair(2.5f, CalorieReqs(1000, 1150, 1250)), // derived from 2-3 from webmd
		std::make_pair(6.f, CalorieReqs(1400, 1500, 1800)), // derived from 4-8 from webmd
		std::make_pair(11.f, CalorieReqs(1800, 2000, 2300)), // derived from 9-13 from webmd
		std::make_pair(16.f, CalorieReqs(2200, 2600, 3000)), // derived from 14-18 from webmd
		std::make_pair(24.5f, CalorieReqs(2400, 2700, 3000)), // derived from 19-30 from webmd
		std::make_pair(40.5f, CalorieReqs(2200, 2500, 2900)), // derived from 31-50 from webmd
		std::make_pair(60.f, CalorieReqs(2000, 2300, 2600)) // derived from 50+ from webmd
	};
	static const LinearInterpolator<CalorieReqs> maleRequirements(maleData);
	static const std::initializer_list<std::pair<float, CalorieReqs>> femaleData = {
		std::make_pair(0.f, CalorieReqs(340)), // first 3 derived from kellymom data
		std::make_pair(0.5f, CalorieReqs(750)),
		std::make_pair(1.f, CalorieReqs(900)),
		std::make_pair(2.5f, CalorieReqs(1000, 1200, 1200)), // derived from 2-3 from webmd
		std::make_pair(6.f, CalorieReqs(1200, 1500, 1600)), // derived from 4-8 from webmd
		std::make_pair(11.f, CalorieReqs(1600, 1800, 2000)), // derived from 9-13 from webmd
		std::make_pair(16.f, CalorieReqs(1800, 2000, 2400)), // derived from 14-18 from webmd
		std::make_pair(24.5f, CalorieReqs(2000, 2100, 2400)), // derived from 19-30 from webmd
		std::make_pair(40.5f, CalorieReqs(1800, 2000, 2200)), // derived from 31-50 from webmd
		std::make_pair(60.f, CalorieReqs(1600, 1800, 2100)) // derived from 50+ from webmd
	};
	static const LinearInterpolator<CalorieReqs> femaleRequirements(femaleData);
	const float ageYears = ageDays / daysInYear;
	const CalorieReqs calorieReqs = (gender == Gender::Male ? maleRequirements : femaleRequirements).interpolate(ageYears);
	switch (activityLevel)
	{
	case ActivityLevel::Sedentary:
		return calorieReqs.sedentary;
	case ActivityLevel::ModeratelyActive:
		return calorieReqs.moderatelyActive;
	case ActivityLevel::Active:
		return calorieReqs.active;
	}
	ERROR("should not be reached");
	return -1.f;
}

float idealBodyfat(int ageDays, Gender gender)
{
	/**
	 * Data taken from the upper-Healthy ranges from:
	 * http://www.tanita.com/data/Children's_Body_Fat_Chart_OTHERS_others_OTHERS__011b.jpg?rev=E5A0
	 * and https://en.wikipedia.org/wiki/Body_fat_percentage#Typical_body_fat_amounts for adult
	 */
	static const std::initializer_list<std::pair<float, float>> maleData = {
		std::make_pair(5.f, 15.f),
		std::make_pair(7.f, 17.f),
		std::make_pair(9.f, 18.f),
		std::make_pair(13.f, 18.f),
		std::make_pair(16.f, 17.f),
		std::make_pair(18.f, 17.f),
		std::make_pair(23.f, 18.f),
		std::make_pair(35.f, 20.f),
		std::make_pair(55, 24.f),
		std::make_pair(80, 24.f)
	};
	static const LinearInterpolator<float> maleBodyfat(maleData);
	static const std::initializer_list<std::pair<float, float>> femaleData = {
		std::make_pair(5.f, 17.f),
		std::make_pair(7.f, 20.f),
		std::make_pair(9.f, 22.f),
		std::make_pair(13.f, 23.f),
		std::make_pair(16.f, 24.f),
		std::make_pair(18.f, 25.f),
		std::make_pair(23.f, 27.f),
		std::make_pair(35.f, 29.f),
		std::make_pair(55, 32.f),
		std::make_pair(80, 29.f)
	};
	static const LinearInterpolator<float> femaleBodyfat(femaleData);
	const float ageYears = ageDays / daysInYear;
	return  (gender == Gender::Male ? maleBodyfat : femaleBodyfat).interpolate(ageYears);
}

float idealMuscleMass(int ageDays, Gender gender)
{
	return 1.f;
}

} // sl
