#include "NPC/Stats/NationalityDatabase.hpp"

#include <fstream>
#include <iostream>

#include "Utility/Assert.hpp"
#include "Utility/Random.hpp"

namespace sl
{

std::unique_ptr<NationalityDatabase> NationalityDatabase::database;

DNADatabase::GenePool NationalityDatabase::getPoolFromNationality(const Nationality& nationality) const
{
	ASSERT(info.count(nationality));
	return info.at(nationality).genePool;
}

const std::vector<NationalityDatabase::Nationality>& NationalityDatabase::getNationalities(DNADatabase::GenePool genePool) const
{
	ASSERT(nationalities.count(genePool));
	return nationalities.at(genePool);
}

const NationalityDatabase::Info& NationalityDatabase::getInfo(const Nationality& nationality) const
{
	ASSERT(info.count(nationality));
	return info.at(nationality);
}
bool NationalityDatabase::exists(const Nationality& nationality) const
{
	return info.count(nationality) > 0;
}

void NationalityDatabase::load()
{
	loadGenePool(DNADatabase::GenePool::NordicCaucasian, "nordic");
	loadGenePool(DNADatabase::GenePool::Caucasian, "caucasian");
	loadGenePool(DNADatabase::GenePool::MediterraneanCaucasian, "mediterranean");
	loadGenePool(DNADatabase::GenePool::Hispanic, "hispanic");
	loadGenePool(DNADatabase::GenePool::Asian, "asian");
	loadGenePool(DNADatabase::GenePool::Tropical, "pacific");
	loadGenePool(DNADatabase::GenePool::Nigger, "black");
	loadGenePool(DNADatabase::GenePool::Indian, "indian");
	loadGenePool(DNADatabase::GenePool::Arabic, "arabic");
	loadGenePool(DNADatabase::GenePool::SouthEastAsian, "seasian");
}

// private
void NationalityDatabase::loadGenePool(DNADatabase::GenePool genePool, const std::string& dir)
{
	const std::string filename = "data/nationalities/" + dir + "/index.txt";
	std::ifstream file(filename.c_str());
	if (file.is_open() && !file.eof())
	{
		std::string buffer;
		std::vector<Nationality>& nationalityList = nationalities[genePool];
		while (!file.eof())
		{
			std::getline(file, buffer);
			if (buffer.back() == '\r')
				buffer.pop_back();
			nationalityList.push_back(buffer);
			loadNationality(genePool, dir, buffer);
		}
	}
	else
	{
		std::cerr << "\n\n\n" << filename << " is either missing, corrupted or empty. Using default names instead.\n\n";
	}
	file.close();
}

void NationalityDatabase::loadNationality(DNADatabase::GenePool genePool, const std::string& dir, const Nationality& nationality)
{
	Info& nationalityInfo = info.insert(std::make_pair(nationality, Info(genePool))).first->second;

	ASSERT(nationalityInfo.firstNames.empty());
	ASSERT(nationalityInfo.lastNames.empty());

	const std::string nationalityDir = "data/nationalities/" + dir + "/" + nationality + "/";

	// first names
	{
		const std::string filename = nationalityDir + "names.txt";
		std::ifstream names(filename.c_str());

		std::cerr << "dir: '" << dir << "'" << std::endl;
		std::cerr << "nationality: '" << nationality << "'" << std::endl;

		if (names.is_open() && !names.eof())
		{
			std::string buffer;
			while (!names.eof())
			{
				std::getline(names, buffer);
				if (buffer.back() == '\r')
					buffer.pop_back();
				nationalityInfo.firstNames.push_back(buffer);
			}
		}
		else
		{
			std::cerr << "\n\n\n" << filename << " is either missing, corrupted or empty. Using default names instead.\n\n" << nationalityDir << "'\n\n\n and '" << filename << "'\n\n";
		}
		names.close();
	}

	// last names
	{
		const std::string filename = nationalityDir + "surnames.txt";
		std::ifstream names(filename.c_str());

		if (names.is_open() && !names.eof())
		{
			std::string buffer;
			while (!names.eof())
			{
				std::getline(names, buffer);
				if (buffer.back() == '\r')
					buffer.pop_back();
				nationalityInfo.lastNames.push_back(buffer);
			}
		}
		else
		{
			std::cerr << "\n\n\n" << filename << " is either missing, corrupted or empty. Using default names instead.\n\n";
		}
		names.close();
	}
}

// static
NationalityDatabase& NationalityDatabase::get()
{
	if (!database)
	{
		database = std::make_unique<NationalityDatabase>();
		database->load(); // todo move
	}
	return *database;
}


// helper functions
const std::string& pickRandomName(Random& random, const NationalityDatabase::Info& info)
{
	return random.choose(info.firstNames);
}

const std::string& pickRandomLastName(Random& random, const NationalityDatabase::Info& info)
{
	return random.choose(info.lastNames);
}

} // sl
