#include "NPC/Stats/Preferences.hpp"

#include <algorithm>
#include <numeric>

#include "Engine/Serialization/AutoDefineObjectSerialization.hpp"
#include "Engine/Serialization/Impl/SerializeVector.hpp"
#include "NPC/Stats/Soul.hpp"
#include "Utility/LinearInterpolator.hpp"
#include "Utility/Math.hpp"

namespace sl
{

Preferences::Preferences(Soul& soul)
	:flavors()
	,adventerous(0.3f)
	,soul(soul)
{
	std::fill(std::begin(colors), std::end(colors), 0.f);
}

//DEF_SERIALIZE_OBJ_FIELDS(FlavorProfile,
//	&FlavorProfile::sweet,
//	&FlavorProfile::sour,
//	&FlavorProfile::salty,
//	&FlavorProfile::bitter,
//	&FlavorProfile::umami,
//	&FlavorProfile::spicy)
DEF_SERIALIZE_OBJ_FIELDS(Preferences::FlavorResponse,
	&Preferences::FlavorResponse::flavor,
	&Preferences::FlavorResponse::response,
	&Preferences::FlavorResponse::datapoints)

void Preferences::deserialize(Deserialize& in)
{
	AutoDeserialize::fields(in, flavors, colors, colorExperiences, adventerous);
}

void Preferences::serialize(Serialize& out) const
{
	AutoSerialize::fields(out, flavors, colors, colorExperiences, adventerous);
}


float Preferences::calculateResponse(const FlavorProfile& flavor) const
{
	return (calculateResponseBySimilarityMethod(flavor) + calculateResponseByLERPing(flavor)) / 2.f;
}

void Preferences::setFoodPreference(const FlavorProfile& flavor, float response)
{
	FlavorResponse f = { flavor, response, 1 };
	flavors.push_back(f);
}

void Preferences::tryFood(const FlavorProfile& food)
{
	const float similarityThreshold = 0.2f;
	const float response = calculateResponse(food);
	bool foundSimilar = false;
	for (FlavorResponse& resp : flavors)
	{
		const float dist = flavorDistance(resp.flavor, food);
		if (dist <= similarityThreshold)
		{
			// combine new point with any existing points if they're similar enough
			// this might make other points combine at some point, so maybe
			// check for that later
			++resp.datapoints;
			const float oldPart = (resp.datapoints - 1.f) / resp.datapoints;
			const float newPart = 1.f / resp.datapoints;
			resp.flavor = resp.flavor * oldPart + food * newPart;
			resp.response = resp.response * oldPart + response * newPart;
			// if their tastes are still malleable, make them like it more.. up to a point
			const float familiarityThreshold = 0.4f;
			const float tasteIncrement = 0.05f;
			if (resp.response <= familiarityThreshold)
			{
				resp.response = std::min(familiarityThreshold, resp.response + tasteMalleability() * tasteIncrement);
			}
			foundSimilar = true;
		}
	}
	if (!foundSimilar)
	{
		setFoodPreference(food, response);
	}
}


ColorPref Preferences::getFavoriteColor() const
{
	return static_cast<ColorPref>(std::distance(std::begin(colors), std::max_element(std::begin(colors), std::end(colors))));
}

float Preferences::getColorPreference(ColorPref color) const
{
	return colors[static_cast<int>(color)];
}



// private
float Preferences::calculateResponseBySimilarityMethod(const FlavorProfile& flavor) const
{
	float meanLike = 0.f;
	float meanDislike = 0.f;
	int liked = 0;
	int disliked = 0;
	for (const FlavorResponse& f : flavors)
	{
		// for likes base the radius for comparing things at 
		const float maxDist = f.response >= 0.f ? foodAdventurous() : 0.3f;
		const float factor = std::max(0.f, maxDist - flavorDistance(f.flavor, flavor));
		if (factor > 0.01f)
		{
			if (f.response >= 0.f)
			{
				meanLike += factor * f.response;
				++liked;
			}
			else
			{
				meanDislike += factor * f.response;
				++disliked;
			}
		}
	}
	meanLike /= liked;
	meanDislike /= disliked;
	return liked + disliked;
}

float Preferences::calculateResponseByLERPing(const FlavorProfile& flavor) const
{
	auto lerpAxis = [&](const float& (FlavorProfile::*axis)() const) -> float
	{
		LinearInterpolator<float> axisPredictor;
		for (const FlavorResponse& f : flavors)
		{
			axisPredictor.addPoint((f.flavor.*axis)(), f.response);
		}
		return axisPredictor.interpolate((flavor.*axis)());
	};
	return (lerpAxis(&FlavorProfile::sweet)  + lerpAxis(&FlavorProfile::sour)  + lerpAxis(&FlavorProfile::salty) +
	        lerpAxis(&FlavorProfile::bitter) + lerpAxis(&FlavorProfile::umami) + lerpAxis(&FlavorProfile::spicy)) / 6.f;
}

float Preferences::flavorDistance(const FlavorProfile& a, const FlavorProfile& b) const
{
	float tmp;
#define SQRDIF(attr) (tmp = (a.attr() - b.attr()), tmp*tmp)
	return std::sqrt(SQRDIF(sweet) + SQRDIF(sour) + SQRDIF(salty) + SQRDIF(bitter) + SQRDIF(umami) + SQRDIF(spicy));
#undef SQRDIF
}

float Preferences::foodAdventurous() const
{
	return adventerous;
}

float Preferences::tasteMalleability() const
{
	static bool lerpInit = false;
	static LinearInterpolator<float> malleabilityLerp;
	if (!lerpInit)
	{
		malleabilityLerp.addPoint(1.f, 1.f);
		malleabilityLerp.addPoint(3.f, 0.9f);
		malleabilityLerp.addPoint(5.f, 0.3f);
		malleabilityLerp.addPoint(10.f, 0.01f);
		malleabilityLerp.addPoint(15.f, 0.1f);
		malleabilityLerp.addPoint(20.f, 0.3f);
		malleabilityLerp.addPoint(40.f, 0.2f);
		malleabilityLerp.addPoint(100.f, 0.0f);
		lerpInit = true;
	}
	return malleabilityLerp.interpolate(soul.getAgeYears());
}

void Preferences::computeColorPreferences()
{
	// for each thing they have a strong opinion of, associate it with a color
	const float triggerThreshold = 0.3f;

	// disallow associations that use a path in the association graph between two colours
	// as that will make most colours get associated with similar things
	auto noColToCol = [](const std::string& a, const std::string& b) -> bool
	{
		bool aIsCol = false;
		bool bIsCol = false;
		for (int colIndex = 0; colIndex < COLOR_PREF_COUNT; ++colIndex)
		{
			const std::string& colName = getInfo(static_cast<ColorPref>(colIndex)).name;
			if (colName == a)
			{
				aIsCol = true;
			}
			if (colName == b)
			{
				bIsCol = true;
			}
		}
		return !aIsCol || !bIsCol;
	};

	for (int colIndex = 0; colIndex < COLOR_PREF_COUNT; ++colIndex)
	{
		const ColorPref col = static_cast<ColorPref>(colIndex);
		const std::map<std::string, float> associated =
			soul.associations.computeCloselyAssociated(
				{ {getInfo(col).name, 1.f} },
				Associations::allConcepts,
				triggerThreshold,
				noColToCol);
		// take the average opinion of all things associated with the color
		colors[colIndex] = 0.f;
		int associatedCount = 0;
		for (const auto& association : associated)
		{
			const float opinion = soul.getOpinion(association.first);
			if (opinion != 0.f)
			{
				colors[colIndex] += association.second * opinion;
				++associatedCount;
			}
		}
		colors[colIndex] /= associatedCount;
	}
}

} // sl
