#include "NPC/Stats/Soul.hpp"

#include <algorithm>
#include <cmath>
#include <initializer_list>
#include <numeric>


#include "Engine/Serialization/AutoDeserialize.hpp"
#include "Engine/Serialization/AutoSerialize.hpp"
#include "Engine/Serialization/Serialization.hpp"
#include "NPC/Stats/DNADatabase.hpp"
#include "NPC/Stats/NationalityDatabase.hpp"
#include "NPC/Stats/SoulGenerationSettings.hpp"
#include "NPC/ClothesDatabase.hpp"
#include "NPC/Locators/FixedTargetLocator.hpp"
#include "NPC/Stats/LoliGen.hpp"
#include "Player/Player.hpp"
#include "Town/World.hpp"
#include "Utility/Assert.hpp"
#include "Utility/Logger.hpp"
#include "Utility/Memory.hpp"
#include "Utility/Random.hpp"
#include "Utility/LinearInterpolator.hpp"

namespace sl
{

uint64_t Soul::souls = 0;

const float daysInYear = 365.f;
const uint32_t pubertyLength = 5 * daysInYear;

Soul::Soul(Deserialize& in)
	:preferences(*this)
	,lifeEvents(*this)
{
	deserialize(in);
}

Soul::Soul(Random& random, const SoulGenerationSettings& settings)
	:preferences(*this)
	,lifeEvents(*this)
	,id(souls++)
	,npc(nullptr)
	,nationality(settings.getNationality(random))
	,motherDNA(DNADatabase::get().generateRandomDNA(random, NationalityDatabase::get().getPoolFromNationality(nationality), false))
	,fatherDNA(DNADatabase::get().generateRandomDNA(random, NationalityDatabase::get().getPoolFromNationality(nationality), true))
	,expressedDNA(motherDNA, fatherDNA)
	,motherID(id)
	,fatherID(id)
{
	setAgeInDays(settings.getAge(random));
	const bool isMale = settings.isMale(random);
	if (isMale != expressedDNA.isMale())
	{
		expressedDNA.genes[DNA::Gender] = isMale;
		fatherDNA.genes[DNA::Gender] = isMale;
		expressedDNA = DNA(motherDNA, fatherDNA);
		geneticStats = GeneticStats(expressedDNA);
	}
	
	// uncomment for totally random dna
	//for (int i = 0; i < DNA_SIZE; ++i)
	//{
	//	motherDNA.genes[i] = random.random(2);
	//	fatherDNA.genes[i] = random.random(2);
	//}
	//expressedDNA = DNA(motherDNA, fatherDNA);

	//if (nationality == "french")
	//	ageInDays = 365 * 7;

	randomlyGenerate(random);
	
	//if (nationality == "french")
	//{
	//	bodySwapID = toColorSwapID(DNA::SkinColor::Caucasian) | toColorSwapID(DNA::EyeColor::Green);
	//	cachedHairSwapID = toColorSwapID(DNA::HairColor::Red);
	//	isHairSwapIDValid = true;
	//}

	appearance.init(*this);
}

Soul::Soul(const Soul& mother, const Soul& father, Random& random, float maleChance)
	:preferences(*this)
	,lifeEvents(*this)
	,id(souls++)
	,npc(nullptr)
	,nationality(father.nationality)
	,motherDNA(random.chance(0.5f) ? mother.fatherDNA : mother.motherDNA)
	,fatherDNA(random.chance(maleChance) ? father.fatherDNA : father.motherDNA)
	,expressedDNA(motherDNA, fatherDNA)
	,geneticStats(expressedDNA)
	,motherID(mother.id)
	,fatherID(father.id)
{
	setAgeInDays(0);
	lastName = father.lastName;
	randomlyGenerate(random);

	appearance.init(*this);
}

namespace AutoDeserialize
{
void deserialize(Deserialize& in, Soul::SuspiciousActivityExplanation& sae)
{
	//LOADVAR(f, sae.suspiciousness);
	//LOADVAR(f, sae.explanationLevel);
}
} // AutoDeserialize

namespace AutoSerialize
{
void serialize(Serialize& out, const Soul::SuspiciousActivityExplanation& sae)
{
	//SAVEVAR(f, sae.suspiciousness);
	//SAVEVAR(f, sae.explanationLevel);
}
} // AutoDeserialize

void Soul::serialize(Serialize& out) const
{
	out.startObject("Soul");
	
	items.serialize(out);
	locationPathCache.serialize(out);
	associations.serialize(out);
	preferences.serialize(out);
	// TODO: lifeEvents
	SAVEVAR(u64, id);
	// npc not saved as it'll get set upon deserializtion of the NPC itself
	SAVEVAR(s, firstName);
	SAVEVAR(s, lastName);
	SAVEVAR(s, nationality);
	SAVEVAR(s, type);
	flags.serialize(out);
	AutoSerialize::serialize(out, playerKnowledge);
	out.s("motherDNA", motherDNA.genes.to_string());
	out.s("fatherDNA", fatherDNA.genes.to_string());
	SAVEVAR(u64, motherID);
	SAVEVAR(u64, fatherID);
	out.u8("status?", status ? 1 : 0);
	if (status)
	{
		out.startObject("status");
		SAVEVAR(f, status->nourishment);
		SAVEVAR(f, status->storedFat);
		SAVEVAR(f, status->muscle);
		SAVEVAR(f, status->hp);
		SAVEVAR(f, status->maxhp);
		SAVEVAR(f, status->hunger);
		SAVEVAR(f, status->thirst);
		SAVEVAR(f, status->energy);
		SAVEVAR(f, status->stamina);
		SAVEVAR(u16, status->bladderContent.urobilin);
		SAVEVAR(u16, status->bladderContent.fluid);
		SAVEVAR(f, status->urinaryUrge);
		SAVEVAR(u16, status->waterInDigestiveSystem);
		AutoSerialize::serialize(out, status->digestingFood);
		SAVEVAR(u8, status->digestionIndex);
		AutoSerialize::serialize(out, status->storedFood);

		SAVEVAR(b, status->isSleeping);
		SAVEVAR(i, status->chloroformTimer);
		
		SAVEVAR(f, status->hope);
		SAVEVAR(f, status->trust);
		SAVEVAR(f, status->fear);
		SAVEVAR(f, status->love);
		status->defaultMood.serialize(out);
		status->currentMood.serialize(out);
		AutoSerialize::serialize(out, status->moodCooldown);
		AutoSerialize::serialize(out, status->personalityModifiers);

		SAVEVAR(f, status->sexDesires);
		SAVEVAR(i, status->vagDamage);
		SAVEVAR(i, status->assDamage);
		SAVEVAR(f, status->vagStretching);
		SAVEVAR(f, status->assStretching);
		AutoSerialize::serialize(out, status->spermInWomb);
		SAVEVAR(u8, status->spermIndex);
		SAVEVAR(i, status->hoursToOvulation);
		SAVEVAR(i, status->hoursToMenstration);
		SAVEVAR(u8, status->hour);
		SAVEVAR(f, status->hornyModifier);
		SAVEVAR(f, status->hornyModifierChange);
		SAVEVAR(f, status->arousal);
		SAVEVAR(f, status->orgasm);
		out.endObject("status");
	}
	SAVEVAR(u32, ageInDays);
	SAVEVAR(f, height);
	SAVEVAR(f, weight);
	AutoSerialize::serialize(out, orgasmCount);
	AutoSerialize::serialize(out, sexCount);
	AutoSerialize::serialize(out, sexSkillType);
	AutoSerialize::serialize(out, sexCooldownByAct);
	AutoSerialize::serialize(out, sexCooldownByScenario);
	SAVEVAR(f, sexSkillDom);
	SAVEVAR(f, sexSkillSub);
	SAVEVAR(f, vagTightness);
	SAVEVAR(f, assTightness);
	SAVEVAR(u32, puberty);
	out.u8("sexuality", static_cast<uint8_t>(sexuality));
	SAVEVAR(u16, pregnant);
	if (baby)
	{
		// Baby has not been added to the world yet - so we serialize it directly here
		ASSERT(pregnant >= 0);
		baby->serialize(out);
	}

	AutoSerialize::serialize(out, locations);
	AutoSerialize::serialize(out, knownEntities);
	AutoSerialize::serialize(out, knownNPCs);
	AutoSerialize::serialize(out, knownCrimes);
	AutoSerialize::serialize(out, suspiciousActivities);
	AutoSerialize::serialize(out, relationships);
	AutoSerialize::serialize(out, opinions);
	// TOOD: serialize schedule / interpolatedSchedule ???
	appearance.serialize(out);
	out.endObject("Soul");
}

void Soul::deserialize(Deserialize& in)
{
	in.startObject("Soul");
	items.deserialize(in);
	locationPathCache.deserialize(in);
	associations.deserialize(in);
	preferences.deserialize(in);
	// TODO: lifeEvents
	LOADVAR(u64, id);
	// npc not saved as it'll get set upon deserializtion of the NPC itself
	LOADVAR(s, firstName);
	LOADVAR(s, lastName);
	LOADVAR(s, nationality);
	LOADVAR(s, type);
	flags.deserialize(in);
	AutoDeserialize::deserialize(in, playerKnowledge);
	motherDNA.genes = decltype(motherDNA.genes)(in.s("motherDNA"));
	fatherDNA.genes = decltype(fatherDNA.genes)(in.s("fatherDNA"));
	expressedDNA = DNA(motherDNA, fatherDNA);
	geneticStats = GeneticStats(expressedDNA);
	LOADVAR(u64, motherID);
	LOADVAR(u64, fatherID);
	if (in.u8("status?"))
	{
		if (!status)
		{
			status = unique<Status>();
		}
		in.startObject("status");
		LOADVAR(f, status->nourishment);
		LOADVAR(f, status->storedFat);
		LOADVAR(f, status->muscle);
		LOADVAR(f, status->hp);
		LOADVAR(f, status->maxhp);
		LOADVAR(f, status->hunger);
		LOADVAR(f, status->thirst);
		LOADVAR(f, status->energy);
		LOADVAR(f, status->stamina);
		LOADVAR(u16, status->bladderContent.urobilin);
		LOADVAR(u16, status->bladderContent.fluid);
		LOADVAR(f, status->urinaryUrge);
		LOADVAR(u16, status->waterInDigestiveSystem);
		AutoDeserialize::deserialize(in, status->digestingFood);
		LOADVAR(u8, status->digestionIndex);
		AutoDeserialize::deserialize(in, status->storedFood);

		LOADVAR(b, status->isSleeping);
		LOADVAR(i, status->chloroformTimer);

		LOADVAR(f, status->hope);
		LOADVAR(f, status->trust);
		LOADVAR(f, status->fear);
		LOADVAR(f, status->love);
		status->defaultMood.deserialize(in);
		status->currentMood.deserialize(in);
		AutoDeserialize::deserialize(in, status->moodCooldown);
		AutoDeserialize::deserialize(in, status->personalityModifiers);

		LOADVAR(f, status->sexDesires);
		LOADVAR(i, status->vagDamage);
		LOADVAR(i, status->assDamage);
		LOADVAR(f, status->vagStretching);
		LOADVAR(f, status->assStretching);
		AutoDeserialize::deserialize(in, status->spermInWomb);
		LOADVAR(u8, status->spermIndex);
		LOADVAR(i, status->hoursToOvulation);
		LOADVAR(i, status->hoursToMenstration);
		LOADVAR(u8, status->hour);
		LOADVAR(f, status->hornyModifier);
		LOADVAR(f, status->hornyModifierChange);
		LOADVAR(f, status->arousal);
		LOADVAR(f, status->orgasm);
		in.endObject("status");
	}
	LOADVAR(u32, ageInDays);
	LOADVAR(f, height);
	LOADVAR(f, weight);
	AutoDeserialize::deserialize(in, orgasmCount);
	AutoDeserialize::deserialize(in, sexCount);
	AutoDeserialize::deserialize(in, sexSkillType);
	AutoDeserialize::deserialize(in, sexCooldownByAct);
	AutoDeserialize::deserialize(in, sexCooldownByScenario);
	LOADVAR(f, sexSkillDom);
	LOADVAR(f, sexSkillSub);
	LOADVAR(f, vagTightness);
	LOADVAR(f, assTightness);
	LOADVAR(u32, puberty);
	sexuality = static_cast<Sexuality>(in.u8("sexuality"));
	LOADVAR(u16, pregnant);
	if (pregnant != -1)
	{
		baby = unique<Soul>(in);
	}
	else
	{
		baby.reset();
	}

	AutoDeserialize::deserialize(in, locations);
	AutoDeserialize::deserialize(in, knownEntities);
	AutoDeserialize::deserialize(in, knownNPCs);
	AutoDeserialize::deserialize(in, knownCrimes);
	AutoDeserialize::deserialize(in, suspiciousActivities);
	AutoDeserialize::deserialize(in, relationships);
	AutoDeserialize::deserialize(in, opinions);
	// TOOD: serialize schedule / interpolatedSchedule ???
	appearance.deserialize(in);

	in.endObject("Soul");
}

Soul::ID Soul::getID() const
{
	return id;
}

void Soul::setNPC(NPC *npc)
{
	ASSERT((this->npc == nullptr) != (npc == nullptr));
	this->npc = npc;
}

NPC* Soul::getNPC()
{
	return npc;
}

const NPC* Soul::getNPC() const
{
	return npc;
}

const ls::Flags& Soul::getFlags() const
{
	return flags;
}

ls::Flags& Soul::getFlags()
{
	return flags;
}

void Soul::mate(World& world, Soul& partner, Random& random)
{
	const std::string fullName = getFirstName() + " " + getLastName();
	Logger::log(Logger::NPC, Logger::Debug) << fullName << "'s vagina was filled with semen" << std::endl;
	const float dayProgression = world.getGameTime().time.getHour() / 24.f;
	// 300 million sperm @TODO male sperm count
	// 65% is the sperm retention
	constexpr float spermCount = 0.3f * 0.65f;
	status->spermInWomb[partner.id][status->spermIndex] += spermCount * (1.f - dayProgression);
	status->spermInWomb[partner.id][(status->spermIndex + 1) % Status::spermLifespan] += spermCount * dayProgression;

	if (puberty >= MENARCHE && !baby)
	{
		pregnant = 9 * 30 * 24; // @todo add variance to pregnancy time
		baby = std::make_unique<Soul>(*this, partner, random);
	}
}

std::shared_ptr<Soul> Soul::giveBirth(World& world)
{
	std::shared_ptr<Soul> ret;

	if (pregnant == 0)
	{
		ASSERT(baby);
		ret = std::move(baby);
		pregnant = -1;
		ASSERT(!baby);
		world.recordBirth(ret);
	}

	return std::move(ret);
}

void Soul::ageOneDay()
{
	const int growthPeriod = 30;
	const std::string oldResourcePath = appearance.getResourcePath();
	setAgeInDays(getAgeDays() + 1);
	if (oldResourcePath != appearance.getResourcePath())
	{
		// need to reset animations!
		if (npc)
		{
			npc->handleEvent(Event(Event::Category::NonBroadcasted, "update_drawables"));
		}
	}
	if (getAgeDays() % growthPeriod == 0)
	{
		if (getAgeDays() < 14.5 * 365)
		{
			const Gender gender = expressedDNA.isMale() ? Gender::Male : Gender::Female;
			height += (1.f + std::min(getNourishment(), 0.f)) * (expectedHeight(ageInDays, gender) - expectedHeight(ageInDays - growthPeriod, gender));
			//weight += (1.f + std::min(getNourishment(), 0.f)) * (expectedWeight(ageInDays, gender) - expectedWeight(ageInDays - growthPeriod, gender));
		}
		const float yearsF = getAgeYears();
		const float progressionThroughPuberty = (float)puberty / pubertyLength;
		if (puberty < pubertyLength) // if not finished puberty
		{
			if (getAgeDays() > pubertyAge())
			{
				puberty = std::min(puberty + growthPeriod, pubertyLength);
			}
		}
		currentOppai = geneticStats.oppai * (progressionThroughPuberty / 100.f) * ((status->nourishment + 100) / 100.f);

		// Orifice healing/stretching
		const float vagCircGrowth = computeVagCircumference() - vagTightness;
		status->vagStretching -= 0.8f * vagCircGrowth;
		if (status->vagStretching < 0.f)
		{
			status->vagStretching = 0.f;
		}
		vagTightness += vagCircGrowth;
		//Logger::log(Logger::NPC, Logger::Debug) << getFirstName() << "(" << 0.1f *(int)(ageInDays / 36.5) << ") now has a " << vagTightness << "cm vag (+" << vagCircGrowth << ")\n";

		const float assCircGrowth = computeAssCircumference() - assTightness;
		status->assStretching -= 0.7f * assCircGrowth;
		if (status->assStretching < 0.f)
		{
			status->assStretching = 0.f;
		}
		assTightness += assCircGrowth;
	}
	if (ageInDays % 365 == 0)
	{
		//Logger::log(Logger::NPC, Logger::Debug) << getFullName() << "(" << 0.1f *(int)(ageInDays / 36.5) << ") now has a " << vagTightness << "cm vag and " << "\n";
	}
}

void Soul::hourlyUpdate(World& world, Random& rng, bool timeSkip)
{
	// DEMO HACK
	// sometimes NPCs get 'stuck' so let's just unstick them if they're off-screen
	if (npc && npc->getScene() && npc->getWorld() && npc->getWorld()->getPlayer())
	{
		if (npc->getScene() != npc->getWorld()->getPlayer()->getScene() || !npc->getScene()->getViewRect().intersects(npc->getMask().getBoundingBox()))
		{
			npc->initDefaultState();
		}
	}

	status->hour = (status->hour + 1) % 24;

	//--------------------- Urinary System -----------------
	// DEMO HACK too much urine all the time
	if (false)//!timeSkip)
	{
		const uint16_t maxUrineProduction = 833;
		const uint16_t minUrineProduction = 33;
		// the minimum amount of urine produced here gets done even without much fluid
		// intake, as it is necessary for other bodily functions
		// this has a negative affect on hydration obviously
		// 100ug / 100ml is the normal amount of urobilin, and assuming 2400ml of water per day that's
		// 100ml / hour, so we 100 ug / hour should be the amount produced
		status->bladderContent.urobilin += 100;
		const uint16_t urineProduced = std::clamp(status->waterInDigestiveSystem, minUrineProduction, maxUrineProduction);
		const uint16_t extraUrinaryRequirements = minUrineProduction - status->waterInDigestiveSystem;
		status->bladderContent.fluid += urineProduced;
		status->waterInDigestiveSystem -= std::min(status->waterInDigestiveSystem, urineProduced);
		const float bladderRatio = status->bladderContent.fluid / bladderCapacity(ageInDays);
		const float urgeDiff = 0.3f - bladderRatio;
		if (urgeDiff >= 0.f)
		{
			status->urinaryUrge = std::fminf(1.f, status->urinaryUrge + urgeDiff);
		}
	}


	//------------------ hunger and digestion -------------
	if (!timeSkip)
	{
		// takes 6-8 (so let's say 7) hours for food to digest
		status->storedFood += status->digestingFood[status->digestionIndex];
		status->digestingFood[status->digestionIndex] = Nutrition();
		status->digestionIndex = (status->digestionIndex + 1) % Status::digestionTime;

		//ASSERT(status->storedFood.getTotalCalories() >= 0.f);
		const float hourlyBMR = calorieRequirements(ageInDays, ActivityLevel::Sedentary, expressedDNA.isMale() ? Gender::Male : Gender::Female) / 24.f;
		const Nutrition expendedFood(hourlyBMR, status->storedFood);
		status->storedFood -= expendedFood;



		// @todo take into account the ratios for macronutrients for fat gain/loss and muscle gain/loss and stuff

		// ~500g of glucogen stores in muscles
		const float maxStoredCalories = 500 * 4.f;
		const float desiredMuscleMass = 5.f;
		if (status->storedFood.getTotalCalories() > maxStoredCalories)
		{
			// store extra in fat
			const Nutrition extraFood(status->storedFood.getTotalCalories() - maxStoredCalories, status->storedFood);
			status->storedFood -= extraFood;
			status->storedFat += extraFood.getTotalCalories() / 9.f; // 9kcal per gram of fat

		}
		else if (status->storedFood.getTotalCalories() < 0.f)
		{
			// make the loli feel drained and feel like shit for starving
			const float deficitRatio = 1.f + (status->storedFood.getTotalCalories() / hourlyBMR);
			ASSERT(deficitRatio >= 0.f && deficitRatio <= 1.f);
			//@todo adjust -3*Happiness and Relaxation, 1*Seriousness, 2*Anger previously here
			// takes extra energy to break down your body into food
			status->energy -= deficitRatio * (4.f / 100.f);

			// break down some fat/muscle and feed ourself with it
			//auto sigmoid = [](float x) -> float
			//{
			//	//1 / (1 + e^ -x)
			//	return 1.f / std::expf();
			//};
			const float fatBurnRatioAtEssential = 0.3f;
			const float fatRatioRatioAtIdeal = 0.7f;
			const float essentialFat = 0.1f;
			const float idealFat = idealBodyfat(ageInDays, Gender::Female);
			const float slope = (fatRatioRatioAtIdeal - fatBurnRatioAtEssential) / (idealFat - essentialFat);
			const float currentBodyfat = (status->storedFat * 1000.f) / getWeight();
			const float linearRatio = fatBurnRatioAtEssential + std::max(0.f, currentBodyfat - essentialFat) * slope;
			const float minFatBurnRatio = 0.05f;
			const float maxFatBurnRatio = 0.95f;
			const float muscleFatBreakdownRatio = std::clamp(linearRatio, minFatBurnRatio, maxFatBurnRatio);
			// 4kcal per gram of protein/carb, but is this a good use? there is water too in muscle...
			// however the glycogen/etc could be factored into the rest of the weight, not the muscle mass
			status->muscle -= muscleFatBreakdownRatio * status->storedFood.getTotalCalories() / 4.f;
			status->storedFat -= (1.f - muscleFatBreakdownRatio) * status->storedFood.getTotalCalories() / 9.f; // 9kcal per gram of fat
			status->storedFood = Nutrition(); // reset food
		}

		const float baseDrainRate = 0.03f;
		status->energy += -baseDrainRate;
	}

	//-------------------- orifice healing ----------------
	// TODO different rates for both
	const int vagHealRate = 1.01f - status->vagDamage;
	status->vagDamage -= std::min(status->vagDamage, vagHealRate);
	const int assHealRate = 1.01f - status->assDamage;
	status->assDamage -= std::min(status->assDamage, assHealRate);



	//------------------------- mood ----------------------
	adjustMoodTowardsDefault(status->isSleeping ? 2.f : 0.5f);
	if (status->hornyModifier < 0.2f)
	{
		status->hornyModifierChange += 0.05f;
	}
	else if (status->hornyModifier > 0.8f)
	{
		status->hornyModifierChange -= 0.05f;
	}
	else
	{
		status->hornyModifierChange += rng.randomFloat(-0.05f, 0.05f);
	}
	status->hornyModifier = std::clamp(status->hornyModifier + status->hornyModifierChange, 0.f, 1.f);

	//--------------------- misc sexual -------------------
	// @TODO different rates for diff things?
	for (int i = 0; i < static_cast<int>(SexType::COUNT); ++i)
	{
		sexCooldownByAct[i] = std::fmaxf(0.f, sexCooldownByAct[i] - 0.1f);
	}
	// this should be okay to remove in-place, as we know that all other iterators are kept valid
	// and also even though there might be rebalancing, iteration is sorted for std::map
	// so this shouldn't change the order of decrementing.
	const float cooldown = 0.1f;
	for (auto it = sexCooldownByScenario.begin(); it != sexCooldownByScenario.end(); /*incremented in body*/)
	{
		if (it->second > cooldown)
		{
			it->second -= cooldown;
			++it;
		}
		else
		{
			it = sexCooldownByScenario.erase(it);
		}
	}

	//---------------------- fertility --------------------
	if (pregnant > 0)
	{
		ASSERT(baby);
		--pregnant;
	}
	///*
	//* Fertility chance from https://www.cyclebeads.com/blog/694/fertility-myth_-i-can-get-pregnant-any-day-of-my-cycle
	//*/
	//static const std::initializer_list<std::pair<float, float>> fertilityRelativeToOvulationData = {
	//	std::make_pair(-7.f, 0.f),
	//	std::make_pair(-6.f, 0.f),
	//	std::make_pair(-5.f, 0.05f),
	//	std::make_pair(-4.f, 0.12f),
	//	std::make_pair(-3.f, 0.09f),
	//	std::make_pair(-2.f, 0.30f),
	//	std::make_pair(-1.f, 0.27f),
	//	std::make_pair(0.f, 0.08f),
	//	std::make_pair(1.f, 0.f),
	//	std::make_pair(2.f, 0.f),
	//};
	//static const LinearInterpolator<float> fertilityRelativeToOvulation(fertilityRelativeToOvulationData);
	if (status->hour % 24 == 0)
	{
		status->spermIndex = (status->spermIndex + 1) % Status::spermLifespan;
		std::vector<decltype(Status::spermInWomb)::iterator> toRemove;
		for (auto it = status->spermInWomb.begin(), end = status->spermInWomb.end(); it != end; ++it)
		{
			bool remove = true;
			// remove old sperm
			it->second[status->spermIndex] = 0.f;
			for (float sperm : it->second)
			{
				if (sperm != 0.f)
				{
					remove = false;
					break;
				}
			}
			if (remove)
			{
				toRemove.push_back(it);
			}
		}
		for (auto it : toRemove)
		{
			status->spermInWomb.erase(it);
		}
	}

	if (puberty >= MENARCHE && !baby)
	{
		if (status->hoursToOvulation > 0)
		{
			--status->hoursToOvulation;
			//Logger::log(Logger::NPC, Logger::Debug) << fullName << " was not ovulating so the sperm went to waste" << std::endl;
		}
		else if (status->hoursToOvulation == 0)
		{
			for (const auto& spermInWomb : status->spermInWomb)
			{
				// on average out of a usual ejaculation (300m), only 20 make it to the egg
				constexpr double chanceOfReachingEgg = 20.f / 300'000'000;
				// sperm from day + lifepsan - offset days ago
#define SPERM(offset) spermInWomb.second[(status->spermIndex + Status::spermLifespan - offset) % Status::spermLifespan]
				const double spermReachedEgg = chanceOfReachingEgg * (SPERM(0) * 0.5f + SPERM(1) + SPERM(2) * 0.5f + SPERM(3) * 0.25f + SPERM(4) * 0.125f);
#undef SPERM
				const Soul& partner = *world.getSoul(spermInWomb.first);
				const std::string fullName = firstName + " " + lastName;
				const std::string& partnerName = partner.getFirstName() + " " + partner.getLastName();
				Logger::log(Logger::NPC, Logger::Debug) << spermReachedEgg << " of " << partnerName << "'s sperm cells have reached " << fullName << "'s egg" << std::endl;
				// out of the 20 that rech there, should only be a 25% chance of getting pregnant
				// and default fertility is 50% (thus 25% when multiplied), so divide by 20
				const double chance = spermReachedEgg * partner.getFertility() * getFertility() * (1.f / 20.f);
				if (rng.chance((float) chance))
				{
					Logger::log(Logger::NPC, Logger::Debug) << fullName << " has become pregnant with " << partnerName << "'s baby" << std::endl;
					pregnant = 9 * 30 * 24; // @todo add variance to pregnancy time
					baby = std::make_unique<Soul>(*this, partner, rng);
				}
			}

			status->hoursToOvulation = -1;
		}

		if (status->hoursToMenstration > 0)
		{
			--status->hoursToMenstration;
		}
		else
		{
			status->hoursToOvulation = rng.randomInclusive(12 * 24, 16 * 24);
			status->hoursToMenstration = rng.randomInclusive(21 * 24, 35 * 24);
		}
	}
}

// ---------------------------------------------------------------------------
// ------------------------------- appearance --------------------------------
const std::string& Soul::getType() const
{
	ASSERT(type == appearance.getType());
	return type;
}

void Soul::setType(const std::string& newType)
{
	type = newType;
	appearance.setType(newType);
}

// ---------------------------------------------------------------------------
// --------------------------- appearance / main -----------------------------
const std::string& Soul::getFirstName() const
{
	return firstName;
}

const std::string& Soul::getLastName() const
{
	return lastName;
}

float Soul::getAgeYears() const
{
	return ageInDays / daysInYear;
}

int Soul::getAgeDays() const
{
	return ageInDays;
}

int Soul::getHeight() const
{
	return height;
}

int Soul::getWeight() const
{
	return weight;
}

const DNA& Soul::getDNA() const
{
	return expressedDNA;
}

const std::string& Soul::getNationality() const
{
	return nationality;
}

// ---------------------------------------------------------------------------
// -------------------------------- stats ------------------------------------
int Soul::strength() const
{
	return geneticStats.strength;
}

int Soul::luck() const
{
	return 0;
}

int Soul::intelligence() const
{
	return geneticStats.intelligence;
}

int Soul::willpower() const
{
	return geneticStats.willpower;
}

int Soul::charisma() const
{
	return geneticStats.charisma;
}

int Soul::dexterity() const
{
	return geneticStats.dexterity;
}

int Soul::beauty() const
{
	return geneticStats.beauty;
}

// ---------------------------------------------------------------------------
// ----------------------------------- personality ---------------------------
float Soul::gullibility() const
{
	return geneticStats.gullibility / 10.f;
}

float Soul::dominance() const
{
	// @TODO: should this be derived from other stats?
	return status->personalityModifiers.dominance;
}

float Soul::openToCautious() const
{
	// https://en.wikipedia.org/wiki/Big_Five_personality_traits#Heritability
	// As seen here, the big five have a genetic component specified there.
	constexpr float geneticComponent = 0.57f;
	return status->personalityModifiers.openToCautious * (1.f - geneticComponent)
	     + geneticStats.personalityModifiers.openToCautious * geneticComponent;
}

float Soul::organizedToCareless() const
{
	// https://en.wikipedia.org/wiki/Big_Five_personality_traits#Heritability
	// As seen here, the big five have a genetic component specified there.
	constexpr float geneticComponent = 0.49f;
	return status->personalityModifiers.organizedToCareless * (1.f - geneticComponent)
		+ geneticStats.personalityModifiers.organizedToCareless * geneticComponent;
}

float Soul::extrovertToIntrovert() const
{
	// https://en.wikipedia.org/wiki/Big_Five_personality_traits#Heritability
	// As seen here, the big five have a genetic component specified there.
	constexpr float geneticComponent = 0.54f;
	return status->personalityModifiers.extrovertToIntrovert * (1.f - geneticComponent)
		+ geneticStats.personalityModifiers.extrovertToIntrovert * geneticComponent;
}

float Soul::friendlyToUnkind() const
{
	// https://en.wikipedia.org/wiki/Big_Five_personality_traits#Heritability
	// As seen here, the big five have a genetic component specified there.
	constexpr float geneticComponent = 0.42f;
	return status->personalityModifiers.friendlyToUnkind * (1.f - geneticComponent)
		+ geneticStats.personalityModifiers.friendlyToUnkind * geneticComponent;
}

float Soul::nervousToConfident() const
{
	// https://en.wikipedia.org/wiki/Big_Five_personality_traits#Heritability
	// As seen here, the big five have a genetic component specified there.
	constexpr float geneticComponent = 0.48f;
	return status->personalityModifiers.nervousToConfident * (1.f - geneticComponent)
		+ geneticStats.personalityModifiers.nervousToConfident * geneticComponent;
}

// ---------------------------------------------------------------------------
// ----------------------------------- mood ----------------------------------
void Soul::affectMood(Mood::Axis moodAxis, float value)
{
	constexpr float permanenceScalar = 1.f / 100.f;

	float& cooldown = status->moodCooldown[moodAxis * 2 + (value >= 0.f ? 0 : 1)];

	const float cooldownScalar = (Mood::max - Mood::rawToResult(cooldown)) / Mood::max;

	ASSERT(cooldownScalar >= 0.f && cooldownScalar <= 1.f);

	status->currentMood.adjustRawValue(moodAxis, value * cooldownScalar);
	status->defaultMood.adjustRawValue(moodAxis, value * cooldownScalar * permanenceScalar);

	cooldown = std::min(cooldown + value, Mood::resultToRaw(Mood::max));
}

float Soul::getHope() const
{
	return status->hope;
}

float Soul::getWillingness() const
{
	const float fearContr = 0.4f;
	const float trustContr = 0.5f;
	return 0.2f * getLove()
		+ 0.8f * ((getFear() * fearContr + (1.f + getTrust()) * 0.5f * trustContr) - 0.5f) / (0.5f * trustContr + fearContr);
}

float Soul::getTrust() const
{
	return status->trust;
}

float Soul::getFear() const
{
	return status->fear;
}

float Soul::getLove() const
{
	return status->love;
}

void Soul::setFear(float fear)
{
	status->fear = std::clamp(status->fear + fear, 0.f, 1.f);
}


// ---------------------------------------------------------------------------
// ---------------------------- physical status ------------------------------
const Soul::Status& Soul::getStatus() const
{
	ASSERT(status);
	return *status;
}

Soul::Status& Soul::getStatus()
{
	ASSERT(status);
	return *status;
}

bool Soul::isAlive() const
{
	return status != nullptr;
}

void Soul::kill()
{
	status.reset();
}

float Soul::getNourishment() const
{
	return status->nourishment;
}

float Soul::getHydration() const
{
	ERROR("unimplemented");
	return -1.f;
}

Urine Soul::urinate()
{
	const uint16_t flowRate = urineFlowRate(ageInDays, expressedDNA.isMale() ? Gender::Male : Gender::Female); // in ml/s
	const float saltRatio = (float) status->bladderContent.urobilin / status->bladderContent.fluid;
	const int urineVolume = std::min(status->bladderContent.fluid, flowRate);
	Urine ret;
	ret.fluid = urineVolume;
	ret.urobilin = saltRatio * urineVolume;
	status->bladderContent -= ret;
	const float bladderRatio = status->bladderContent.fluid / bladderCapacity(ageInDays);
	status->urinaryUrge = bladderRatio;
	if (ret.fluid != 0)
	{
		std::cout << "just released " << ret.fluid << "ml  of urine - " << status->bladderContent.fluid << "ml left\n";
	}
	return ret;
}

float Soul::getUrinaryUrges() const
{
	return status->urinaryUrge;
}

void Soul::sleep()
{
	ASSERT(!status->isSleeping);
	status->isSleeping = true;
}

void Soul::wakeup()
{
	ASSERT(status->isSleeping);
	status->isSleeping = false;
}


//----------misc shit idk where to put
void Soul::giftItem(Item item)
{
	// improve their mood maybe, but this depends on the context?
	// maybe we could have some kind of base amount here, then for context
	// sensitive things code the change in the dialogue via affectMood()



	items.addItem(std::move(item));
}

void Soul::consume(Item item)
{
	if (item.hasComponent(Item::Edible))
	{
		//@todo adjust 2*response to happiness and relaxation previously here
		const float response = status->hunger * preferences.calculateResponse(item.getFlavor());
		preferences.tryFood(item.getFlavor());
		status->digestingFood[(status->digestionIndex + status->digestionTime - 1) % status->digestionTime] += item.getNutrition();
	}
	else
	{
		// ???
		ERROR("should this happen?");
	}
}

Schedule::InterpolatedSchedule Soul::getInterpolatedSchedule(const GameTime& time)
{
	auto interpolation = interpolatedSchedule.interpolate(time);
	// out of date (TODO: do this for everyone asynch ahead of time?)
	if (interpolation.progress == 1.f || interpolation.progress == -1.f)
	{
		// to ensure we don't miss an entire event (ie sleep) and end up querying in-between.
		GameTime yesterday(time);
		yesterday.date.advance(-1);
		GameTime nextWeek(time);
		nextWeek.date.advance(7);
		interpolatedSchedule = schedule.computeSchedule(*this, yesterday, nextWeek);
	}
	else
	{
		ASSERT(interpolation.progress >= 0 && interpolation.progress < 1.f);
	}
	return interpolatedSchedule;
}

const Schedule& Soul::getSchedule() const
{
	return schedule;
}

// ---------------------------------------------------------------------------
// -------------------------------- relations --------------------------------
bool Soul::playerKnows(const std::string& fact) const
{
	return playerKnowledge.count(fact) != 0;
}

void Soul::setPlayerKnowledge(const std::string& fact, bool value)
{
	if (value)
	{
		playerKnowledge.insert(fact);
	}
	else
	{
		playerKnowledge.erase(fact);
	}
}

const TargetLocator* Soul::getLocation(const std::string& locationName) const
{
	auto it = locations.find(locationName);
	if (it != locations.end())
	{
		return it->second.get();
	}
	return nullptr;
}

void Soul::setLocation(const std::string& locationName, std::unique_ptr<TargetLocator> location)
{
	ASSERT(location->shouldSerialize());
	ASSERT(knownEntities.count(locationName) == 0);
	locations[locationName] = std::move(location);
}

// TODO: getKnownEntities()? what if there's multiple...
Entity* Soul::getKnownEntity(const std::string& entityName)
{
	auto it = knownEntities.find(entityName);
	if (it != knownEntities.end())
	{
		return it->second;
	}
	return nullptr;
}

const Entity* Soul::getKnownEntity(const std::string& entityName) const
{
	auto it = knownEntities.find(entityName);
	if (it != knownEntities.end())
	{
		return it->second;
	}
	return nullptr;
}

void Soul::setKnownEntity(const std::string& entityName, Entity *entity)
{
	ASSERT(locations.count(entityName) == 0);
	//ASSERT(knownNPCs.count(entityName) == 0);
	locations[entityName] = unique<FixedTargetLocator>(Target(entity));
	knownEntities[entityName] = entity;
}

bool Soul::isCrimeReported(int crimeId) const
{
	auto it = knownCrimes.find(crimeId);
	return it != knownCrimes.end() && it->second;
}

bool Soul::isCrimeKnown(int crimeId) const
{
	return knownCrimes.count(crimeId) != 0;
}

void Soul::setCrimeReported(int crimeId, bool reported)
{
	knownCrimes[crimeId] = reported;
}

std::vector<int> Soul::getUnreportedCrimes() const
{
	std::vector<int> unreportedCrimes;
	for (const std::pair<int, bool>& crime : knownCrimes)
	{
		if (!crime.second)
		{
			unreportedCrimes.push_back(crime.first);
		}
	}
	return unreportedCrimes;
}

const std::map<int, bool>& Soul::getKnownCrimes() const
{
	return knownCrimes;
}

int Soul::hasNPC(const std::string& npcName) const
{
	auto it = knownNPCs.find(npcName);
	return it == knownNPCs.end() ? 0 : it->second.size();
}

Soul::ID Soul::getNPC(const std::string& npcName) const
{
	auto it = knownNPCs.find(npcName);
	ASSERT(it != knownNPCs.end());
	ASSERT(it->second.size() == 1);
	return it->second.front();
}

void Soul::setNPC(const std::string& npcName, ID id)
{
	knownNPCs[npcName] = std::vector<ID>(1, id);
}

const std::vector<Soul::ID>& Soul::getNPCs(const std::string& npcCategory) const
{
	auto it = knownNPCs.find(npcCategory);
	ASSERT(it != knownNPCs.end());
	return it->second;
}

void Soul::addNPC(const std::string& npcCategory, Soul::ID id)
{
	knownNPCs[npcCategory].push_back(id);
}

float Soul::computeSuspicionOfPlayer() const
{
	float sum = 0.f;
	for (const auto& activity : suspiciousActivities)
	{
		sum += activity.second.suspiciousness;
	}
	return sum * ((1.f / suspiciousActivities.size()) + 0.3f);
}

void Soul::suspectPlayer(const std::string& id, float level)
{
	auto insert = suspiciousActivities.insert(std::make_pair(id, SuspiciousActivityExplanation{ false, level }));
	if (!insert.second)
	{
		if (level > insert.first->second.suspiciousness)
		{
			insert.first->second.suspiciousness = level;
		}
	}
}

void Soul::shiftNPCRelationship(ID id, float amount)
{
	relationships[id] += amount;
}

float Soul::getNPCRelationship(ID id) const
{
	// TODO: sigmoid function this instead of linear?
	auto it = relationships.find(id);
	return it != relationships.end() ? std::clamp(it->second, 0.f, 1.f) : 0.f;
}

void Soul::setOpinion(const std::string& thing, float value)
{
	opinions[thing] = value;
}

void Soul::shiftOpinion(const std::string& thing, float value)
{
	opinions[thing] += value;
}

float Soul::getOpinion(const std::string& thing) const
{
	// TODO: sigmoid function this instead of linear?
	auto it = opinions.find(thing);
	return it != opinions.end() ? std::clamp(it->second, 0.f, 1.f) : 0.f;
}

// ---------------------------------------------------------------------------
// ---------------------------------- sexual ---------------------------------
int Soul::oppai() const
{
	return currentOppai;
}

int Soul::getHoursUntilPregnancy() const
{
	return pregnant;
}

int Soul::getPubertyDays() const
{
	return puberty;
}

float Soul::getPuberty() const
{
	return ((float)puberty) / pubertyLength;
}

int Soul::getVagTightness() const
{
	return vagTightness + status->vagStretching;
}

int Soul::getAssTightness() const
{
	return assTightness + status->assStretching;
}

float Soul::getFertility() const
{
	// @TODO vary this
	return 0.5f;
}

float Soul::getArousal() const
{
	// we already affected the rate at which arousal changes by puberty, so let's not be too harsh on the max
	const float pubertalContrib = computePubertalContributionToArousal();
	const float maxArousalByPuberty = 0.5f + 0.5f * pubertalContrib;
	// to simulate a minimal amount of random horniness
	const float raw = std::fminf(status->arousal * 0.9f + status->hornyModifier * pubertalContrib * 0.1f, maxArousalByPuberty);
	if (raw < 0.f)
	{
		// avoid NaN from pow
		return 0.05f;
	}
	const float exponent = 0.7f;
	const float mult = 19.f;
	static const float div = std::pow(mult, exponent);
	return std::pow(raw * mult, exponent) / div;
}

void Soul::arouse(float diff)
{
	// make it so that when you're in a horny mood, it not only increases faster,
	// but will decrease slower and vice versa
	const float modifier = computePubertalContributionToArousal()
		* (diff >= 0.f ? status->hornyModifier : 1.f - status->hornyModifier);
	status->arousal = std::clamp(status->arousal + diff * (0.25f + modifier), 0.f, 1.f);
}

void Soul::stimulate(float amount)
{
	status->orgasm += amount * getArousal();
	status->orgasm = std::clamp(status->orgasm, 0.f, 1.f);
}

float Soul::getOrgasm() const
{
	return status->orgasm;
}

void Soul::haveOrgasm(SexType cause)
{
	ASSERT(getOrgasm() >= 0.97f);
	status->orgasm = 0.f;
	arouse(-0.5f);
	++orgasmCount[static_cast<int>(cause)];
}

int Soul::getOrgasmCount(SexType cause) const
{
	return orgasmCount[static_cast<int>(cause)];
}

int Soul::getTotalOrgasmCount() const
{
	return std::accumulate(std::begin(orgasmCount), std::end(orgasmCount), 0);
}

void Soul::addSexExp(SexType type, const std::string *scenario, float xp, float dominance)
{
	if (xp <= 0.f)
	{
		ERROR("invalid sexExp amount: " + std::to_string(xp));
		return;
	}
	ASSERT(dominance >= -1.f && dominance <= 1.f);
	const int typeIndex = static_cast<int>(type);
	float cooldown = sexCooldownByAct[typeIndex];
	if (scenario)
	{
		auto it = sexCooldownByScenario.find(*scenario);
		if (it != sexCooldownByScenario.end())
		{
			cooldown += it->second;
		}
	}
	const float scaledXp = xp / (1.f + (std::pow(cooldown, 1.2f) / xp));
	sexSkillType[typeIndex] += scaledXp;
	if (dominance > 0.f)
	{
		sexSkillDom += scaledXp * dominance;
	}
	else
	{
		sexSkillSub -= scaledXp * dominance;
	}
}

float Soul::getSexExp(SexType type) const
{
	return std::fmaxf(sexSkillType[static_cast<int>(type)], getMaxSexExp());
}

void Soul::incrementSex(SexType type)
{
	++sexCount[static_cast<int>(type)];
}

int Soul::getSexCount(SexType type) const
{
	return sexCount[static_cast<int>(type)];
}

int Soul::getTotalSexCount() const
{
	return std::accumulate(std::begin(sexCount), std::end(sexCount), 0);
}

float Soul::getSexAbility(SexType type, float dominance) const
{
	ASSERT(dominance >= -1.f && dominance <= 1.f);
	float categoryBonus = 0.f;
	auto ability = [&](SexType t) -> float {
		const float linear = fmaxf(sqrtf(sexSkillType[static_cast<int>(t)]) / getMaxSexExp(), 1.f);
		static const float maxVal = std::log(18.f);
		return std::log(17.f * linear + 1) / maxVal;
	};
	switch (type)
	{
	case SexType::Vaginal:
	case SexType::Anal:
	case SexType::Thighsex:
		categoryBonus = (ability(SexType::Vaginal) + ability(SexType::Anal) + ability(SexType::Thighsex)) / 5.f;
		break;
	case SexType::Oral:
		categoryBonus = ability(SexType::Handjob) * 0.5;
		break;
	case SexType::Handjob:
		categoryBonus = ability(SexType::Footjob) * 0.15 + ability(SexType::Oral) * 0.4;
		break;
	case SexType::Footjob:
		categoryBonus = ability(SexType::Handjob) * 0.3;
		break;
	case SexType::Masturbation:
		categoryBonus = ability(SexType::Handjob) * 0.1
		              + ability(SexType::Molestation) * 0.5f; // she knows from receiving
		break;
	case SexType::Molestation:
		categoryBonus = ability(SexType::Handjob) * 0.1
		              + ability(SexType::Masturbation) * 0.5f; // she knows from giving
		break;
	case SexType::OralReceive:
		categoryBonus = ability(SexType::Oral) * 0.3
		              + ability(SexType::Masturbation) * 0.3;
		break;
	default:
		ERROR("unhandled type");
	}
	const float base = (categoryBonus + ability(type)) / getMaxSexExp();
	const float dominanceBoost = dominance < 0.f
	                           ? ((1.f + dominance) - (dominance * sexSkillSub))
	                           : ((1.f - dominance) + (dominance * sexSkillDom));
	return std::clamp(base * dominanceBoost, 0.f, 1.f);
}

float Soul::checkVaginalFit(float girth) const
{
	return girth / getVagTightness();
}

float Soul::performVaginalInsertion(float girth)
{
	const float fit = checkVaginalFit(girth);
	if (fit > 2.51f)
	{
		return 0.f;
	}
	if (fit > 1.f)
	{
		// @TODO: figure out what kind of relation/numbers make sense here
		status->vagDamage += powf(fit, 1.5f);
		status->vagStretching += 0.2f * powf(fit, 1.5f);
	}
	return fit;
}

float Soul::checkAnalFit(float girth) const
{
	return girth / getAssTightness();
}

float Soul::performAnalInsertion(float girth)
{
	const float fit = checkAnalFit(girth);
	if (fit > 2.51f)
	{
		return 0.f;
	}
	if (fit > 1.f)
	{
		// @TODO: figure out what kind of relation/numbers make sense here
		status->assDamage += fit * fit;
		status->assStretching += 0.4f * fit * fit;
	}
	return fit;
}

/*		private			*/
void Soul::randomlyGenerate(Random& random)
{
	const Gender gender = expressedDNA.isMale() ? Gender::Male : Gender::Female;

	setType("loli");


	const NationalityDatabase& nationDB = NationalityDatabase::get();

	const NationalityDatabase::Info& nationalityInfo = nationDB.getInfo(nationality);
	firstName = pickRandomName(random, nationalityInfo);
	if (lastName.empty())
	{
		lastName = pickRandomLastName(random, nationalityInfo);
	}
	//Logger::log(Logger::Area::NPC, Logger::Priority::Debug) << "loli names '" << getName() << "' generated\n";


	//-------------------------------------------------------------------------
	//                                Physical stuff

	/*
		Standard deviations for height and weight are derived from the stats mentioned in
		expectedWeight() and expectedHeight()
	*/
	//tallness = 100 + random.randomFromNormalDistribution(0, 2.75);
	//tallness += geneticStats.tallness;
	//heaviness = 100 + random.randomFromNormalDistribution(0, 7.14);
	//heaviness += geneticStats.heaviness;

	//testosteroneNormal = random.randomFloat(-1, 1);
	//estrogenNormal = random.randomFloat(-1, 1);
	height = expectedHeight(ageInDays, gender) * (100 + geneticStats.tallness) / 100.f;
	weight = expectedWeight(ageInDays, gender) * (100 + geneticStats.heaviness) / 100.f;//* ((100 - nourishment) / 100.f);
	//oppai = random.randomFloat(100) * heaviness;


	ASSERT(!status);
	status = unique<Status>();
	//-------------------------------------------------------------------------
	//                              physical status
	status->nourishment = 0;
	status->storedFat = idealBodyfat(ageInDays, gender) * weight;
	status->muscle = idealMuscleMass(ageInDays, gender) * weight;
	status->hp = 100.f;
	status->maxhp = status->hp;
	status->hunger = 0.f;
	status->thirst = 0.f;
	status->energy = 1.f;
	status->bladderContent.fluid = 0;
	status->bladderContent.urobilin = 0;
	status->waterInDigestiveSystem = 0;
	std::fill(std::begin(status->digestingFood), std::end(status->digestingFood), Nutrition());
	status->digestionIndex = 0;

	status->isSleeping = false;

	status->chloroformTimer = 0;

	//-------------------------------------------------------------------------
	//                              mental status
	status->hope = 0.7f + random.randomFloat(0.3f);
	status->trust = random.randomFloat(0.15f);
	status->fear = random.randomFloat(0.1f);
	status->love = random.randomFloat(0.05f);
	std::fill(std::begin(status->moodCooldown), std::end(status->moodCooldown), 0.f);
	// TODO: achieve this through life events instead (or both?)
	status->personalityModifiers.dominance = random.randomFloat(-0.5f, 0.5f);
	status->personalityModifiers.extrovertToIntrovert = random.randomFloat(-0.5f, 0.5f);
	status->personalityModifiers.friendlyToUnkind = random.randomFloat(-0.5f, 0.5f);
	status->personalityModifiers.nervousToConfident = random.randomFloat(-0.5f, 0.5f);
	status->personalityModifiers.openToCautious = random.randomFloat(-0.5f, 0.5f);
	status->personalityModifiers.organizedToCareless = random.randomFloat(-0.5f, 0.5f);

	//-------------------------------------------------------------------------
	//                                preferences
	// @TODO add in more foods, or maybe have them be slowly introduced during a girl's life
	preferences.setFoodPreference(Item(Item::strToType("candy")).getFlavor(), random.randomFloat(0.3f, 0.6f));
	preferences.setFoodPreference(Item(Item::strToType("chocolatebar")).getFlavor(), random.randomFloat(0.5f, 0.8f));

	//-------------------------------------------------------------------------
	//                              reproductive/sexual
	status->sexDesires = 0;
	status->vagDamage = 0;
	status->assDamage = 0;
	status->vagStretching = 0;
	status->assStretching = 0;
	status->spermIndex = 0;
	status->hoursToOvulation = -1;
	status->hoursToMenstration = 0;
	status->hour = 0;
	status->hornyModifier = 0.5f;
	status->hornyModifierChange = 0.f;
	status->arousal = 0.f;
	
	for (int i = 0; i < static_cast<int>(SexType::COUNT); ++i)
	{
		orgasmCount[i] = 0;
		sexCount[i] = 0;
		sexSkillType[i] = 0.f;
		sexCooldownByAct[i] = 0.f;
	}
	sexSkillDom = 0.f;
	sexSkillSub = 0.f;
	pregnant = -1;
	puberty = 0;
	if (getAgeDays() > pubertyAge())
	{
		puberty = std::min(ageInDays - pubertyAge(), pubertyLength);
	}
	vagTightness = computeVagCircumference();
	assTightness = computeAssCircumference();
	// According to many studies, roughly 1% of the population is asexual
	sexuality = Sexuality::Asexual;
	if (random.chance(0.99f))
	{
		/*
			And now we give them a sexuality. based on page 39/43, Figure 7 of:
			http://psyc158-sp09.pbworks.com/f/ARTICLE%25C2%25A0-%25C2%25A0Sexual%25C2%25A0Orientation%25C2%25A0Lies%25C2%25A0Smoothly%25C2%25A0on%25C2%25A0a%25C2%25A0Continuum%25C2%25A0~%25C2%25A0Epstein%25C2%25A02007.pdf
		*/
		struct SexualityPerRespondents
		{
			Sexuality sexuality;
			int respondants;
		};
		// these numbers only apply for females
		static const SexualityPerRespondents respondantNumbers[7] = {
			{Sexuality::Straight,                50},
			{Sexuality::AlmostEntirelyStraight, 300},
			{Sexuality::MostlyStraight,         425},
			{Sexuality::Bisexual,               275},
			{Sexuality::MostlyGay,              150},
			{Sexuality::AlmostEntirelyGay,      100},
			{Sexuality::Gay,                     25}
		};
		static const int total = std::accumulate(std::begin(respondantNumbers), std::end(respondantNumbers), 0,
			[](int count, const SexualityPerRespondents& resp) -> int
			{
				return count + resp.respondants;
			}
		);
		int sexualityIndex = random.random(total);
		for (const SexualityPerRespondents& resp : respondantNumbers)
		{
			sexualityIndex -= resp.respondants;
			if (sexualityIndex <= 0)
			{
				sexuality = resp.sexuality;
				break;
			}
		}
		ASSERT(sexuality != Sexuality::Asexual);
		// and then figure out if they're a complete nympho slut loli or not...
	}

	for (int i = 0; i < 69; ++i) // @todo remove and init vag/ass/etc here
	{
		ageOneDay();
	}
}

int Soul::pubertyAge() const
{
	const int daysInYear = 365;
	if (status->nourishment > 0)//fatty mc fat fat
	{
		return (10 - status->nourishment / 50.f) * daysInYear + geneticStats.pubertyStart; // (8-10 + deviation)
	}
	else
	{
		return (10 - status->nourishment / 25.f) * daysInYear + geneticStats.pubertyStart; // (10-14 + deviation)
	}
}

float Soul::computePubertalContributionToArousal() const
{
	const float base = 0.4f;
	static const float div = base + powf(3.f, 0.3f);
	return base + powf(3.f * getPuberty(), 0.3f) / div;
}

float Soul::computeArousalContributionToVaginalSize() const
{
	const float base = 0.2f;
	static const float div = base + powf(3.f, 0.3f);
	return base + powf(3.f * getPuberty(), 0.3f) / div;
}

float Soul::computeVagCircumference() const
{
	// I have no idea what most of these numbers are anymore since I wrote them at the start of the project and never commented them...
	// but I do remember that they didn't really represent anything in particular and were just things I fiddled
	// with to make it seem to give reasonable sizes at various ages
	const float yearsF = getAgeYears();
	return ((height + (100.f + geneticStats.heaviness) * 1.2 + std::min(yearsF, 18.f) * 4.f + std::min(yearsF, 10.f) * 12.f) / 30.f)
	     * (0.8 + 1.f * getPuberty()) / 2.f
	     * ((100.f + geneticStats.vaginalTightness) / 100.f);
}

float Soul::computeAssCircumference() const
{
	// see comment for computeVagCircumference()
	const float yearsF = getAgeYears();
	return ((height * 2.f + ((100.f + geneticStats.heaviness) / 100.f) * 69.f + 32 + std::min(yearsF, 18.f) * 6) / 30.f)
	     * (1.f + (getPuberty() / 3.f)) / 2.f
	     * ((100.f + geneticStats.analTightness) / 100.f);
}

void Soul::adjustMoodTowardsDefault(float rate)
{
	const float cooldownRate = 1.f;
	status->currentMood.gravitateTowards(status->defaultMood, rate);
	for (float& cooldown : status->moodCooldown)
	{
		cooldown = std::max(0.f, cooldown - cooldownRate);
	}
}

void Soul::setAgeInDays(int newAge)
{
	this->ageInDays = newAge;
	appearance.setAge(newAge);
}

}// End of sl namespace
