#include "NPC/ThrownItem.hpp"

#include "Engine/Graphics/Sprite.hpp"
#include "Engine/Scene.hpp"
#include "Engine/Serialization/AutoDeserialize.hpp"
#include "Engine/Serialization/AutoSerialize.hpp"
#include "Engine/Serialization/Impl/SerializeVector.hpp"
#include "Items/ItemPickup.hpp"
#include "Items/ItemToDrawables.hpp"
#include "NPC/Stealth/Noise.hpp"
#include "Utility/Random.hpp"
#include "Utility/Memory.hpp"
#include "Utility/Trig.hpp"

namespace sl
{
const float g = -0.25f;
const float throwSpeed = 3.f;

ThrownItem::ThrownItem(const Vector2f& start, const Vector2f& target, Item item)
	:Entity(start.x, start.y, 6, 6, "ThrownItem")
	,veloc(lengthdir(throwSpeed, pointDirection(start, target)))
	,currentHeight(0)
	,item(std::move(item))
{
	// @HACK: this is here to let us call getCurrentDrawable() when we rotate
	addDrawableToSet("base", itemToDrawable(item.getType(), true));
	setDrawableSet("base");

	// we start with y''(t) = g, for some gravity constant
	// which gives    y'(t) = gt + c_1, for some constant c_1
	// thus we have    y(t) = (g/2)t^2 + c_1t + c_2 for some constant c_2
	// now known conditions for w = amount of time it takes to reach target's x/y
	// y(0) = y(w) = 0 since item is on ground and also
	// y'(w/2) = 0 since in the middle of the arc the projectile isn't falling
	// so y'(w/2) = g(w/2) + c_1 = 0, thus c_1 = -g(w/2)
	// now to find c_2 we use y(0) = (g/2)0^2 + c_10 + c_2 = c_2 = 0
	// so y(t) = (g/2)t^2 - g(w/2)t and y'(t) = gt - g(w/2)
	// and we want y'(0) the initial velocity which is -g(w/2)
	const float w = (start - target).length() / throwSpeed;
	heightVeloc = -g * w / 2;
}

ThrownItem::ThrownItem(DeserializeConstructor)
	:Entity(0, 0, 6, 6, "ThrownItem")
	,item(Item(0))
{
}

void ThrownItem::deserialize(Deserialize& in)
{
	AutoDeserialize::fields(in, veloc, heightVeloc, currentHeight, item);
}

void ThrownItem::serialize(Serialize& out) const
{
	AutoSerialize::fields(out, veloc, heightVeloc, currentHeight, item);
}

const ThrownItem::Type ThrownItem::getType() const
{
	return Entity::makeType("ThrownItem");
}


void ThrownItem::onUpdate()
{
	if (veloc.x || veloc.y)
	{
		pos += veloc;
		currentHeight += heightVeloc;
		heightVeloc += g;
		if (currentHeight < 0.f)
		{
			// and release a sound
			scene->addEntityToList(new Noise(pos, Noise::Level::concerning, 32.f - 8.f * heightVeloc));
			// hit ground, stop moving if not bouncing much
			if (heightVeloc <= -1.5f)
			{
				currentHeight = 0.f;
				veloc = lengthdir(veloc.length() * 0.3f, direction(veloc) + 15.f - Random::get().randomFloat(30.f));
				heightVeloc *= -0.5f;
			}
			else
			{
				//currentHeight = 0.f;
				//veloc.x = 0.f;
				//veloc.y = 0.f;
				//heightVeloc = 0.f;
				scene->addEntityToList(new ItemPickup(pos, std::move(item)));
				destroy();
			}
			
		}
		setHeight(currentHeight);
		// @HACK ew
		Sprite *sprite = (Sprite*)getCurrentDrawable();
		sprite->rotate(sprite->getRotation() + 13.f);
	}
}

} // sl
