#ifndef SL_LS_AUTOMATICFUNCTIONWRAPPING_HPP
#define SL_LS_AUTOMATICFUNCTIONWRAPPING_HPP

#include <functional>
#include <tuple>
#include <type_traits>
#include <utility>

#include "Parsing/Binding/Arguments.hpp"
#include "Parsing/Binding/OptionalParam.hpp"
#include "Parsing/Binding/Value.hpp"
#include "Parsing/Binding/ValueConversion.hpp"

namespace sl
{
namespace ls
{
namespace afw
{

///
template <typename F> struct FunctorTraits;

template <typename R, typename ...Args>
struct FunctorTraits<R(Args...)>
{
	static constexpr size_t arity = sizeof...(Args);
	typedef R Return;
	typedef std::tuple<Args...> ArgsTuple;
};

template <typename F>
struct FunctorTraits : public FunctorTraits<decltype(&F::operator())> {};
template <typename R, typename ...Args>
struct FunctorTraits<R(&)(Args...)> : public FunctorTraits<R(Args...)> {};
template <typename Class, typename R, typename ...Args>
struct FunctorTraits<R(Class::*)(Args...)> : public FunctorTraits<R(Args...)> {};
template <typename Class, typename R, typename ...Args>
struct FunctorTraits<R(Class::*)(Args...) const> : public FunctorTraits<R(Args...)> {};
/*template <typename R, typename ...Args>
struct FunctorTraits<std::function<R(Args...)>> : public FunctorTraits<R(Args...)> {};*/

///
template <typename Return>
struct CallToValue
{
	template <typename F, typename Args>
	Value operator()(const F& f, const Args& args) {
		return Value(std::apply(f, args));
	}
};
template <>
struct CallToValue<void> {
	template <typename F, typename Args>
	Value operator()(const F& f, const Args& args) {
        std::apply(f, args);
        return Value();
    }
};

///Recursive tuple setup (does not work in MSVC++)
//template <int Index, typename... Args>
//inline std::tuple<Args...>& convertArgs(const Arguments &args, std::tuple<Args...> &tuple)
//{
//	if constexpr(sizeof...(Args))
//	{
//		std::get<Index>(tuple) = convert<typename std::tuple_element<Index,std::tuple<Args...>>::type>(args[Index]);
//		if constexpr(Index+1 < sizeof...(Args))
//		{
//			return convertArgs<Index+1>(args, tuple);
//		}
//	}
//	return tuple;
//}
///Classic SFINAE
template <int Index, typename std::enable_if< Index == 0, int>::type=0, typename... Args>
inline std::tuple<Args...>& convertArgs(const Arguments &args, std::tuple<Args...> &tuple) {
	return tuple;
}

template <int Index, typename std::enable_if< 0 < Index, int>::type=0, typename... Args>
inline std::tuple<Args...>& convertArgs(const Arguments &args, std::tuple<Args...> &tuple) {
	std::get<Index-1>(tuple) =
		convert<typename std::tuple_element<Index-1,std::tuple<Args...>>::type>(
			Index-1 < args.count() ? args[Index-1] : Value());
	return convertArgs<Index-1>(args, tuple);
}

template <typename... Args>
inline std::tuple<Args...> convertArgs(const Arguments& args)
{
	// TODO: uncomment + make it work with OptionalParam (it was commented out before this though)
	//VERIFY_ARG_COUNT(args, sizeof...(Args));
	std::tuple<Args...> tuple;
	return convertArgs<sizeof...(Args)>(args, tuple);
}

template <typename Tuple>
inline Tuple convertFunctorArgs(const Arguments& args)
{
	Tuple tuple;
	return convertArgs<std::tuple_size<Tuple>::value>(args, tuple);
}

// It already had a loliscript parameter signature - do nothing in this case.
template <>
inline std::tuple<const Arguments&> convertFunctorArgs<std::tuple<const Arguments&>>(const Arguments& args)
{
	return args;
}

} // afw

/**
 * Converts a C++-function to one that automatically grabs the correct ParameterList parameters, verifies them, etc that an be bound to loliscript
 * @param function The C++ function to bind
 * @return A loliscript-callable version of it
 */
template <typename Return, typename... Args>
std::function<Value(const Arguments&)> sepplesToLoliscript(Return(*function)(Args...))
{
	return [function](const Arguments& args) -> Value
	{
		return afw::CallToValue<Return>()(function, afw::convertArgs<Args...>(args));
	};
}

/**
* Converts a C++ method and class instance to one that automatically grabs the correct ParameterList parameters, verifies them, etc that an be bound to loliscript
* @param obj The instance to bind it to
* @param function The C++ function to bind
* @return A loliscript-callable version of it
*/
template <typename Class, typename Return, typename... Args>
std::function<Value(const Arguments&)> sepplesToLoliscript(Class *instance, Return(Class::*function)(Args...))
{
	return [instance, function](const Arguments& args) -> Value
	{
		return afw::CallToValue<Return>()(
			[instance, function](Args... args) { return ((*instance).*function)(args...); },
			afw::convertArgs<Args...>(args));
	};
}
// const version
template <typename Class, typename Return, typename... Args>
std::function<Value(const Arguments&)> sepplesToLoliscript(const Class *instance, Return(Class::*function)(Args...) const)
{
	return [instance, function](const Arguments& args) -> Value
	{
		return afw::CallToValue<Return>()(
			[instance, function](Args... args) { return ((*instance).*function)(args...); },
			afw::convertArgs<Args...>(args));
	};
}

/**
 * Converts a C++ functor (includes lambdas, std::function. etc) to one that automatically grabs the correct ParameterList parameters,
 * verifies them (length for now), etc that an be bound to loliscript
 * @param functor The functor to be bound.
 * @return A loliscript-callable version of it
 */
template <typename F>
std::function<Value(const Arguments&)> sepplesToLoliscript(F functor)
{
	return [f(std::move(functor))](const Arguments& args) -> Value
	{
		return afw::CallToValue<typename afw::FunctorTraits<F>::Return>()(f, afw::convertFunctorArgs<typename afw::FunctorTraits<F>::ArgsTuple>(args));
	};
}

} // ls
} // sl

#endif // SL_LS_AUTOMATICFUNCTIONWRAPPING_HPP