/**
 * @section DESCRIPTION
 * All registered functions and variables
 */

#ifndef SL_LS_BINDINGS_HPP
#define SL_LS_BINDINGS_HPP

#include <iostream>
#include <functional>
#include <map>
#include <sstream>
#include <stack>
#include <string>
#include <unordered_map>
#include <vector>

#include "Parsing/Binding/Arguments.hpp"
#include "Parsing/Binding/AutomaticFunctionWrapping.hpp"
#include "Parsing/Binding/BoundVar.hpp"
#include "Parsing/Binding/Parameters.hpp"
#include "Parsing/Binding/Value.hpp"
#include "Parsing/Errors/RuntimeError.hpp"

namespace sl
{

extern void VERIFY_ARG_COUNT(const ls::Arguments& args, int desiredCount);
extern void VERIFY_ARG_COUNT(const ls::Arguments& args, int minParamCount, int maxParamCount);

namespace ls
{

class SyntaxTreeNode;

class Bindings
{
public:
	/**
	 * Constructs a Bindings with nothing bound to it yet.
	 */
	Bindings();
	/**
	 * Sets a variable's value.
	 * If the variable doesn't exist, an error is generated unless it's in the
	 * scope 'local' in which case the variable is created and set as well.
	 * @param scope The scope that the varible resides in
	 * @param name The name of the variable
	 * @param value The value to set the variable as
	 */
	void setVar(const std::string& scope, const std::string& name, const Value& value);
	/**
	 * Gets the value of a variable stored in this context.
	 * An error is generated if the variable doesn't exist.
	 * @param scope The scope that the variable resides in
	 * @param name The name of the variable
	 * @return The value of the variable
	 */
	Value getVar(const std::string& scope, const std::string& name) const;
	/**
	 * Checks to see if there is a variable registered to the input scope/name
	 * @param scope The scope to check in
	 * @param name The name to check for
	 * @return True if the variable exists, false otherwise.
	 */
	bool variableExists(const std::string& scope, const std::string& name) const;

	/**
	 * Binds a C++ variable/something that acts like one to this Bindings's variable lists
	 * so that it can be accessed from within the scripting language. If a variable is
	 * already bound to the name/scope it is deleted and replaced.
	 * @param scope The scope to beind it to
	 * @param name The name to bind it to
	 * @param var The variable to bind
	 */
	void registerVar(const std::string& scope, const std::string& name, std::unique_ptr<BoundVar> var);

	bool functionExists(const std::string& scope, const std::string& name) const;

	/**
	 * Convenience converter of arbitrary function(-pointer) to loliscript function, then registers it.
	 * @param scope The scope to register the function to
	 * @param name The name to expose to loliscript
	 * @param function The function to convert then bind
	 */
	template <typename Return, typename... Args>
	void registerFunction(const std::string& scope, const std::string& name, Return(*function)(Args...))
	{
		registerFunction(scope, name, sepplesToLoliscript(function));
	}

	/**
	* Convenience converter of arbitrary member function(-pointer) to loliscript function, then registers it.
	* @param scope The scope to register the function to
	* @param name The name to expose to loliscript
	* @param instance The instance to call the method on. This must outlive the use of the function, as a pointer is stored to it.
	* @param function The method pointer to convert.
	*/
	template <typename Class, typename Return, typename... Args>
	void registerFunction(const std::string& scope, const std::string& name, Class& instance, Return(Class::*function)(Args...))
	{
		registerFunction(scope, name, sepplesToLoliscript(&instance, function));
	}

	/**
	 * Const version of the above.
	 */
	template <typename Class, typename Return, typename... Args>
	void registerFunction(const std::string& scope, const std::string& name, const Class& instance, Return(Class::*function)(Args...) const)
	{
		registerFunction(scope, name, sepplesToLoliscript(&instance, function));
	}

	/**
	 * Convenience converter of arbitrary functor (std::function, lambda, etc) to loliscript function, then registers it.
	 * @param scope The scope to register the function to
	 * @param name The name to expose to loliscript
	 * @param function The functor to convert then bind
	 */
	template <typename F>
	void registerFunction(const std::string& scope, const std::string& name, F function)
	{
		registerFunction(scope, name, sepplesToLoliscript(std::move(function)));
	}

	/**
	 * Registers a function to be accessible from loliscript.
	 * This is the actual implementation of registerFunction -- all other template fuckery ones simply convert to stuff to pass into here.
	 * @param scope The scope to register the function to
	 * @param name The name to expose to loliscript
	 * @param function The function to bind
	 */
	void registerFunction(const std::string& scope, const std::string& name, std::function<Value(const Arguments&)> function)
	{
		functions[scope][name] = std::move(function);
	}

	/**
	 * Evaluates a function with the given parameters, then return the result.
	 */
	Value evaluateFunction(const std::string& scope, const std::string& name, const Parameters& parameters) const;

	/**
	 * Unregisters all varibles / flags / functions in a given scope.
	 * @param scope The scope to unregister, if it existed
	 */
	void unregisterScope(const std::string& scope);
	/**
	 * Unregisters a certain variable in a given scope
	 * @param scope The scope the variable is in
	 * @param name The name of the variable
	 */
	void unregisterVar(const std::string& scope, const std::string& name);

#ifdef SL_DEBUG
	/**
	 * Debug print method that prints out all variables/flags/functions bound/stored in this context.
	 */
	void print();
#endif

	const SyntaxTreeNode* getCurrentNode() const;

	class Mark
	{
	public:
		Mark(const Bindings& context, const SyntaxTreeNode *cur)
			:prev(context.getCurrentNode())
			,context(context)
		{
			const_cast<Bindings&>(context).currentNode = cur;
		}

		~Mark()
		{
			const_cast<Bindings&>(context).currentNode = prev;
		}

	private:
		Mark(const Mark&) = delete;
		Mark& operator=(const Mark&) = delete;

		const SyntaxTreeNode *prev;
		const Bindings& context;
	};

private:
	//!	A lookup table of all variables bound to this Bindings
	std::unordered_map<std::string, std::unordered_map<std::string, std::unique_ptr<BoundVar>>> variables;
	//!	A map of all functions bound to this Bindings
	std::unordered_map<std::string, std::map<std::string, std::function<Value(const Arguments&)>>> functions;
	//!	The local variables currently on the stack. Only the ones on the top are accessable.
	std::stack<std::map<std::string, Value>> localVars;
	//! Named constants.
	std::map<std::string, Value> constants;
	//!
	const SyntaxTreeNode *currentNode;

	friend class Mark;
};

} // ls
} // sl

#endif // SL_LS_BINDINGS_HPP
