#include "Parsing/Binding/Flags.hpp"

#include "Engine/Serialization/AutoDeserialize.hpp"
#include "Engine/Serialization/AutoSerialize.hpp"
#include "Engine/Serialization/Serialization.hpp"
#include "Parsing/Binding/Bindings.hpp"
#include "Utility/Assert.hpp"

namespace sl
{

// TODO: move?
void serialize(Serialize& out, const ls::Value& val)
{
	out.u8("Type", static_cast<uint8_t>(val.getType()));
	switch (val.getType())
	{
	case ls::Value::Type::Uninitialized:
		break;
	case ls::Value::Type::Bool:
		out.u8("Value", val.asBool() ? 1 : 0);
		break;
	case ls::Value::Type::Int:
		out.i("Value", val.asInt());
		break;
	case ls::Value::Type::Float:
		out.f("Value", val.asFloat());
		break;
	case ls::Value::Type::String:
		out.s("Value", val.asString());
		break;
	default:
		ERROR("could not serialize ls::Value of type " + val.typeToStr(val.getType()));
	}
}

void deserialize(Deserialize& in, ls::Value& val)
{
	const ls::Value::Type type = static_cast<ls::Value::Type>(in.u8("Type"));
	switch (type)
	{
	case ls::Value::Type::Uninitialized:
		val = ls::Value();
		break;
	case ls::Value::Type::Bool:
		val = ls::Value(in.u8("Value") == 0 ? false : true);
		break;
	case ls::Value::Type::Int:
		val = ls::Value(in.i("Value"));
		break;
	case ls::Value::Type::Float:
		val = ls::Value(in.f("Value"));
		break;
	case ls::Value::Type::String:
		val = ls::Value(in.s("Value"));
		break;
	default:
		ERROR("Unsupported type: " + ls::Value::typeToStr(type));
	}
}

namespace ls
{

void Flags::set(const std::string& name, Value value)
{
	flags[name] = std::move(value);
}

const Value& Flags::get(const std::string& name) const
{
	// default initialize things to 0 to allow scripts to make their own flags
	static const Value notFound(0);
	auto it = flags.find(name);
	return it != flags.end() ? it->second : notFound;
}

bool Flags::exists(const std::string& name) const
{
	return flags.find(name) != flags.cend();
}

void Flags::serialize(Serialize& out) const
{
	AutoSerialize::serialize(out, flags);
}

void Flags::deserialize(Deserialize& in)
{
	AutoDeserialize::deserialize(in, flags);
}

// non-class functions
void registerFlags(Bindings& context, Flags& flags, const std::string& scope)
{
	context.registerFunction(scope, "flag", [&flags](std::string name) -> Value {
		return flags.get(name);
	});
	context.registerFunction(scope, "setFlag", [&flags](const Arguments& args) {
		VERIFY_ARG_COUNT(args, 1, 2);
		const ls::Value val = args.count() >= 2
		                    ? args[1]
		                    : ls::Value(true);
		flags.set(args[0].asString(), std::move(val));
	});
	context.registerFunction(scope, "incrFlag", [&flags](const Arguments& args) {
		VERIFY_ARG_COUNT(args, 1, 2);
		const Value incrAmount = args.count() == 1 ? Value(1) : args[1];
		const std::string key = args[0].asString();
		flags.set(key, flags.get(key) + incrAmount);
	});
	context.registerFunction(scope, "flagExists", [&flags](std::string name) -> bool {
		return flags.exists(name);
	});
}

} // ls
} // sl
