#include "Parsing/Binding/Value.hpp"

#include "Parsing/Errors/RuntimeError.hpp"
#include "Utility/Assert.hpp"
#include "Utility/DynamicStore.hpp"
#include "Utility/ToString.hpp"

#include <algorithm>
#include <memory>

namespace sl
{
namespace ls
{

/*static*/ const std::string& Value::typeToStr(Type type)
{
	static const std::string names[] = {
		"Uninitialized",
		"Bool",
		"Int",
		"Float",
		"String",
		"Ptr",
		"Data"
	};
	return names[static_cast<int>(type)];
}

Value::Value()
	:type(Type::Uninitialized)
{
}

Value::Value(int val)
	:type(Type::Int)
	,iVal(val)
{
}

Value::Value(float val)
	:type(Type::Float)
	,fVal(val)
{
}

Value::Value(double val)
	:Value(static_cast<float>(val))
{
}

Value::Value(bool val)
	:type(Type::Bool)
	,bVal(val)
{
}

Value::Value(std::string val)
	:type(Type::String)
{
	sVal = new std::string(std::move(val));
}

Value::Value(std::unique_ptr<DynamicStore> data)
	:type(Type::Data)
	,dVal(data.release())
{
}

Value::Value(const Value& copy)
	:type(copy.type)
{
	switch (type)
	{
	case Type::Uninitialized:
		break;
	case Type::Float:
		fVal = copy.fVal;
		break;
	case Type::Int:
		iVal = copy.iVal;
		break;
	case Type::Bool:
		bVal = copy.bVal;
		break;
	case Type::String:
		sVal = new std::string(*copy.sVal);
		break;
	case Type::Ptr:
		pVal = copy.pVal;
		break;
	case Type::Data:
		dVal = copy.dVal->clone().release();
		break;
	default:
		ERROR("unimplemented");
	}
}

Value::Value(Value&& move)
	:type(move.type)
{
	switch (type)
	{
	case Type::Uninitialized:
		break;
	case Type::Float:
		fVal = move.fVal;
		break;
	case Type::Int:
		iVal = move.iVal;
		break;
	case Type::Bool:
		bVal = move.bVal;
		break;
	case Type::String:
		sVal = move.sVal;
		break;
	case Type::Ptr:
		pVal = move.pVal;
		break;
	case Type::Data:
		dVal = move.dVal;
		break;
	default:
		ERROR("unimplemented");
	}
	move.type = Type::Uninitialized;
}

Value::~Value()
{
	cleanup();
}

Value& Value::operator=(const Value& rhs)
{
	cleanup();
	type = rhs.type;
	switch (type)
	{
	case Type::Uninitialized:
		break;
	case Type::Float:
		fVal = rhs.fVal;
		break;
	case Type::Int:
		iVal = rhs.iVal;
		break;
	case Type::Bool:
		bVal = rhs.bVal;
		break;
	case Type::String:
		sVal = new std::string(*rhs.sVal);
		break;
	case Type::Ptr:
		pVal = rhs.pVal;
		break;
	case Type::Data:
		dVal = rhs.dVal->clone().release();
		break;
	default:
		ERROR("unimplemented");
	}
	return *this;
}

Value& Value::operator=(Value&& rhs)
{
	if (&rhs != this)
	{
		cleanup();
		type = rhs.type;
		switch (type)
		{
		case Type::Uninitialized:
			break;
		case Type::Float:
			fVal = rhs.fVal;
			break;
		case Type::Int:
			iVal = rhs.iVal;
			break;
		case Type::Bool:
			bVal = rhs.bVal;
			break;
		case Type::String:
			sVal = rhs.sVal;
			break;
		case Type::Ptr:
			pVal = rhs.pVal;
			break;
		case Type::Data:
			dVal = rhs.dVal;
			break;
		default:
			ERROR("unimplemented");
		}
		rhs.type = Type::Uninitialized;
	}
	return *this;
}

float Value::asFloat() const
{
	switch (type)
	{
	case Type::Uninitialized:
		break;
	case Type::Float:
		return fVal;
		break;
	case Type::Int:
		return iVal;
		break;
	case Type::Bool:
		return bVal ? 1.f : 0.f;
		break;
	case Type::String:
	case Type::Ptr:
	case Type::Data:
		throw RuntimeError("Invalid conversion: " + typeToStr(type) + " -> Float");
	}
	ASSERT("control flow should not reach here");
	return -1.f;
}

int Value::asInt() const
{
	switch (type)
	{
	case Type::Uninitialized:
		break;
	case Type::Float:
		return fVal;
		break;
	case Type::Int:
		return iVal;
	case Type::Bool:
		return bVal ? 1 : 0;
	case Type::String:
	case Type::Ptr:
	case Type::Data:
		throw RuntimeError("Invalid conversion: " + typeToStr(type) + " -> Int");
	}
	ASSERT("control flow should not reach here");
	return -1;
}

bool Value::asBool() const
{
	switch (type)
	{
	case Type::Uninitialized:
		break;
	case Type::Float:
		// @TODO some epsilon check?
		return fVal != 0.f;
	case Type::Int:
		return iVal != 0;
	case Type::Bool:
		return bVal;
	case Type::String:
		return !sVal->empty();//throw RuntimeError("Invalid conversion: String -> Float");
	case Type::Ptr:
		return pVal != nullptr;
	case Type::Data:
		throw RuntimeError("Invalid conversion: Data -> Bool");
	}
	ASSERT("control flow should not reach here");
	return false;
}

std::string Value::asString() const
{
	switch (type)
	{
	case Type::Bool:
		return bVal ? "true" : "false";
	case Type::Int:
		return std::to_string(iVal);
	case Type::Float:
		return std::to_string(fVal);
	case Type::String:
		return *sVal;
	case Type::Ptr:
		return "ptr:" + std::to_string((uintptr_t)pVal);
	case Type::Data:
		throw RuntimeError("Invalid conversion: Data -> String");
	}
	throw RuntimeError("Cannot convert uninitialized to String");
}

DynamicStore& Value::asData() const
{
	if (type != Type::Data)
	{
		throw RuntimeError("Invalid conversion: " + typeToStr(type) + " -> Data");
	}
	return *dVal;
}

Value::Type Value::getType() const
{
	return type;
}

Value Value::asType(Type convertTo) const
{
	if (convertTo == type)
	{
		return *this;
	}
	switch (convertTo)
	{
	case Type::Bool:
		return Value(asBool());
	case Type::Int:
		return Value(asInt());
	case Type::Float:
		return Value(asFloat());
	case Type::String:
		return Value(asString());
	case Type::Ptr:
	case Type::Data:
		throw RuntimeError("Cannot convert to " + typeToStr(convertTo));
	}
	throw RuntimeError("Trying to convert a value to uninitialized");
}

static inline void throwCantCompare()
{
	throw RuntimeError("Cannot compare String to non-String, Ptr to non-Ptr, or cannot compare data/uninitialized values");
}

bool Value::operator==(const Value& rhs) const
{
	if (type == Type::String && rhs.type == Type::String)
	{
		return asString() == rhs.asString();
	}
	const Type commonType = convertToType(type, rhs.type);
	switch (commonType)
	{
	case Type::Float:
		return asFloat() == rhs.asFloat();
	case Type::Int:
		return asInt() == rhs.asInt();
	case Type::Bool:
		return asBool() == rhs.asBool();
	case Type::Ptr:
		return pVal == rhs.pVal;
	default:
		throwCantCompare();
	}
}

bool Value::operator!=(const Value& rhs) const
{
	return !(*this == rhs);
}

bool Value::operator<(const Value& rhs) const
{
	if (type == Type::String && rhs.type == Type::String)
	{
		return asString() < rhs.asString();
	}
	const Type commonType = convertToType(type, rhs.type);
	switch (commonType)
	{
	case Type::Float:
		return asFloat() < rhs.asFloat();
	case Type::Int:
		return asInt() < rhs.asInt();
	case Type::Bool:
		return asBool() < rhs.asBool();
	case Type::Ptr:
		return pVal < rhs.pVal;
	default:
		throwCantCompare();
	}
}

Value& Value::operator+=(const Value& rhs)
{
	if (type == Type::String)
	{
		*sVal += rhs.asString();
	}
	else if (rhs.type == Type::String)
	{
		sVal = new std::string(asString() + rhs.asString());
		type = Type::String;
	}
	else
	{
		switch (convertToType(type, rhs.type))
		{
		case Type::Int:
			iVal = asInt() + rhs.asInt();
			type = Type::Int;
			break;
		case Type::Float:
			fVal = asFloat() + rhs.asFloat();
			type = Type::Float;
			break;
		default:
			throw RuntimeError("Cannot perform: " + typeToStr(type) + " + " + typeToStr(rhs.type));
		}
	}
	return *this;
}

Value& Value::operator-=(const Value& rhs)
{
	switch (convertToType(type, rhs.type))
	{
	case Type::Int:
		iVal = asInt() - rhs.asInt();
		type = Type::Int;
		break;
	case Type::Float:
		fVal = asFloat() - rhs.asFloat();
		type = Type::Float;
		break;
	default:
		throw RuntimeError("Cannot perform: " + typeToStr(type) + " - " + typeToStr(rhs.type));
	}
	return *this;
}

Value& Value::operator*=(const Value& rhs)
{
	switch (convertToType(type, rhs.type))
	{
	case Type::Int:
		iVal = asInt() * rhs.asInt();
		type = Type::Int;
		break;
	case Type::Float:
		fVal = asFloat() * rhs.asFloat();
		type = Type::Float;
		break;
	default:
		throw RuntimeError("Cannot perform: " + typeToStr(type) + " * " + typeToStr(rhs.type));
	}
	return *this;
}

Value& Value::operator/=(const Value& rhs)
{
	switch (convertToType(type, rhs.type))
	{
	case Type::Int:
		// if you want int, call floor()
	case Type::Float:
		if (rhs.asFloat() == 0.f)
		{
			throw RuntimeError("Division by zero");
		}
		fVal = asFloat() / rhs.asFloat();
		type = Type::Float;
		break;
	default:
		throw RuntimeError("Cannot perform: " + typeToStr(type) + " / " + typeToStr(rhs.type));
	}
	return *this;
}

// private
/*static*/Value::Type Value::convertToType(Type lhs, Type rhs)
{
	if (lhs >= Type::Bool && lhs <= Type::Float &&
	    rhs >= Type::Bool && rhs <= Type::Float)
	{
		return std::max(lhs, rhs);
	}
	return Type::Uninitialized;
}

void Value::cleanup()
{
	if (type == Type::String)
	{
		delete sVal;
	}
	else if (type == Type::Data)
	{
		delete dVal;
	}
	// pointers are non-owned
	type = Type::Uninitialized;
}

} // sl
} // sl
