#ifndef SL_LS_VALUE_HPP
#define SL_LS_VALUE_HPP

#include <string>
#include <type_traits>

#include "Parsing/Errors/RuntimeError.hpp"
#include "Utility/Assert.hpp"

namespace sl
{

class DynamicStore;

namespace ls
{

class Value
{
public:
	enum class Type
	{
		Uninitialized = 0,
		Bool,
		Int,
		Float,
		String,
		Ptr,
		Data
	};
	static const std::string& typeToStr(Type type);

	template <typename T>
	static Value make(const T& arg)
	{
		return Value(arg);
	}

	//template <>
	//static Value make<void>(void)
	//{
	//	return Value();
	//}

	Value();

	explicit Value(int val);

	explicit Value(float val);

	explicit Value(double val);

	explicit Value(bool val);

	explicit Value(std::string val);

	template <typename T>
	explicit Value(T *ptr)
		:type(Type::Ptr)
		,pVal(ptr)
	{
	}

	// otherwise nullptr matches the C-String constructor...
	explicit Value(nullptr_t)
		:type(Type::Ptr)
		,pVal(nullptr)
	{
	}

	explicit Value(const char *str)
		:Value(std::string(str))
	{
	}

	explicit Value(std::unique_ptr<DynamicStore> data);

	Value(const Value& copy);

	Value(Value&& move);

	~Value();

	Value& operator=(const Value& rhs);

	Value& operator=(Value&& rhs);

	float asFloat() const;

	int asInt() const;

	bool asBool() const;

	std::string asString() const;
	
	template <typename T>
	T* asPtr() const
	{
		ASSERT(type == Type::Ptr);
		return reinterpret_cast<T*>(pVal);
	}

	DynamicStore& asData() const;

	Type getType() const;

	Value asType(Type convertTo) const;

	bool operator==(const Value& rhs) const;

	bool operator!=(const Value& rhs) const;

	bool operator<(const Value& rhs) const;

	Value& operator+=(const Value& rhs);

	Value& operator-=(const Value& rhs);

	Value& operator*=(const Value& rhs);

	Value& operator/=(const Value& rhs);

private:
	static Type convertToType(Type lhs, Type rhs);

	void cleanup();

	Type type;
	union
	{
		float fVal;
		int iVal;
		bool bVal;
		std::string *sVal;
		//std::map<Value> *mVal;
		void *pVal;
		DynamicStore *dVal;
	};
};

inline Value operator+(const Value& lhs, const Value& rhs)
{
	Value ret(lhs);
	ret += rhs;
	return std::move(ret);
}

inline Value operator-(const Value& lhs, const Value& rhs)
{
	Value ret(lhs);
	ret -= rhs;
	return std::move(ret);
}

inline Value operator*(const Value& lhs, const Value& rhs)
{
	Value ret(lhs);
	ret *= rhs;
	return std::move(ret);
}

inline Value operator/(const Value& lhs, const Value& rhs)
{
	Value ret(lhs);
	ret /= rhs;
	return std::move(ret);
}

} // ls
} // sl

#endif // SL_LS_VALUE_HPP