#include "Parsing/Parsers/DialogueParser.hpp"

#include <cstdlib>
#include <cstring>
#include <limits>
#include <sstream>

#include "Parsing/SyntaxTree/Expressions/Literal.hpp"
#include "Parsing/SyntaxTree/Statements/Proc.hpp"
#include "Parsing/SyntaxTree/Statements/SetVariable.hpp"
#include "Parsing/SyntaxTree/Statements/Statement.hpp"
#include "Parsing/Errors/SyntaxError.hpp"
#include "Parsing/Parsers/ExpressionParser.hpp"
#include "Parsing/Parsers/StatementParser.hpp"
#include "Parsing/Parsers/ParserUtil.hpp"

#include "NPC/Dialogue/Outcome.hpp"
#include "NPC/Dialogue/AttributeRange.hpp"
#include "Utility/FileUtil.hpp"
#include "Utility/Logger.hpp"
#include "Utility/Memory.hpp"

namespace sl
{
namespace ls
{

static std::unique_ptr<const Outcome> parseOutcome(ExpressionParser& expressionParser, const std::string& scenarioId, const std::vector<std::string>& strings, std::size_t first, std::size_t last);

/*	  extern functions		*/

std::unique_ptr<const Scenario> parseDialogueFile(const std::string& filename)
{
	const std::string fullFilename = "resources/dlg/" + filename + ".dlg";
	const std::vector<std::string> lines = fileToLines(fullFilename.c_str());
	if (lines.empty())
	{
		throw SyntaxError("Dialogue file (" + fullFilename + ") does not exist or is empty");
	}
	ExpressionParser expressionParser(fullFilename);
	return parseDialogueStrings(expressionParser, filename, lines);
}

std::unique_ptr<const Scenario> parseDialogueStrings(ExpressionParser& expressionParser, const std::string& identifier, const std::vector<std::string>& strings)
{
	enum DType
	{
		Void,
		Title,
		Requirements,
		OutcomeBlock,
		Links
	};
	DType lastType = Void;
	int lineOfLast = 0;
	const char *str_title       = "title",
	           *str_requirement = "requirements",
	           *str_outcome     = "outcome",
	           *str_extern      = "extern",
	           *str_links       = "links";

	std::unique_ptr<const Statement> title;
	std::unique_ptr<const Statement> requirements;
	std::unique_ptr<const Statement> links; // links common to all outcomes
	std::vector<std::unique_ptr<const Outcome>> outcomes;
	std::string outcomeId;
	const char *c;
	int depth = 0;
	for (std::size_t i = 0; i < strings.size(); ++i)
	{
		c = strings[i].c_str();
		eatWhitespace(c);
		Logger::log(Logger::Parsing) << "parse[" << (i>8?"":" ") << (i+1) << "|" << depth << "]: " << c << "\n";
		if (depth == 0)
		{
			if (strncmp(c, str_title, strlen(str_title)) == 0)
			{
				Logger::log(Logger::Parsing) << "<title>\n";
				lastType = Title;
				c += strlen(str_title);
				eatWhitespace(c);
				// not end of line - thus inline message
				if (*c != '\0')
				{
					if (*c == '{')
					{
						throw SyntaxError("CURLY BRACES GO ON THE NEXT LINE FAGGOT");
					}
					title = unique<SetVariable>(identifier, i, "scenario", "title", expressionParser.parseAsConcatenatedStrings(c));
					continue;
				}
			}
			else if (strncmp(c, str_requirement, strlen(str_requirement)) == 0)
			{
				Logger::log(Logger::Parsing) << "<requirements>\n";
				lastType = Requirements;
				c += strlen(str_requirement);
				eatWhitespace(c);
				// not end of line - thus inline message
				if (*c != '\0')
				{
					if (*c == '{')
					{
						throw SyntaxError("CURLY BRACES GO ON THE NEXT LINE FAGGOT");
					}
					requirements = unique<SetVariable>(identifier, i, "scenario", "allow", expressionParser.parseEntireExpression(c));
					continue;
				}
			}
			else if (strncmp(c, str_outcome, strlen(str_outcome)) == 0)
			{
				Logger::log(Logger::Parsing) << "<outcome>\n";
				lastType = OutcomeBlock;
				c += strlen(str_outcome);
				eatWhitespace(c);
				outcomeId = c;
			}
			else if (strncmp(c, str_links, strlen(str_links)) == 0)
			{
				Logger::log(Logger::Parsing) << "<links>\n";
				lastType = Links;
			}
			else if (strncmp(c, str_extern, strlen(str_extern)) == 0)
			{
				c += strlen(str_extern);
				eatWhitespace(c);
				if (strncmp(c, str_outcome, strlen(str_outcome)) == 0)
				{
					c += strlen(str_outcome);
					eatWhitespace(c);
					std::string filename("resources/dlg/");
					filename += c;
					filename += ".lso";
					auto lines = fileToLines(filename.c_str());
					outcomes.push_back(parseOutcome(expressionParser, std::string(c), lines, 0, lines.size()));
				}
				else if (strncmp(c, str_links, strlen(str_links)) == 0)
				{
					c += strlen(str_links);
					eatWhitespace(c);
					std::string filename(c);
					auto lines = fileToLines(filename.c_str());
					links = parseFile(filename.c_str());
				}
				else
				{
					std::stringstream ss;
					ss << "Unknown statement: \""
					   << c
					   << "\" after extern keyword on line "
					   << (i + 1);
					throw SyntaxError(ss.str());
				}
			}
			else if (strncmp(c, "//", 2) == 0 || *c == '\0' || *c == '\n' || *c == '\r')
			{
				Logger::log(Logger::Parsing) << "<comment>\n";
				//  comments and empty lines are okay
			}
			else if (*c == '{')
			{
				// curly braces are okay, caught later
			}
			else
			{
				std::stringstream ss;
				ss << "Unknown statement :\""
				   << c
				   << "\" on line "
				   << (i + 1);
				throw SyntaxError(ss.str());
			}
			lineOfLast = i + 1;
		}
		if (*c == '{')
		{
			if (++depth == 1)
			{
				lineOfLast = i + 1;
			}
		}
		else if (*c == '}')
		{
			if (--depth == 0)
			{
				switch (lastType)
				{
					case Void:
						//  Why would you even bother to make an empty block?
						break;
					case Title:
						if (requirements)
						{
							throw SyntaxError("Additional title block encountered on line " + std::to_string(i + 1));
						}
						title = parseStrings(expressionParser, strings, lineOfLast, i, identifier);
						break;
					case Requirements:
						if (requirements)
						{
							throw SyntaxError("Additional requirements block encountered on line " + std::to_string(i + 1));
						}
						requirements = parseStrings(expressionParser, strings, lineOfLast, i, identifier);
						break;
					case OutcomeBlock:
						outcomes.push_back(parseOutcome(expressionParser, identifier + ":" + outcomeId, strings, lineOfLast, i));
						break;
				}
				lastType = Void;
			}
			if (depth < 0)
			{
				std::stringstream ss;
				ss << "Extra } encountered: \""
				   << strings[i]
				   << "\" on line"
				   << (i + 1);
				throw SyntaxError(ss.str());
			}
		}
	}
	if (depth != 0)
	{
		//  uh oh, probably complain or something here
		throw SyntaxError("Unmatched curly brace at end of file");
	}
	if (outcomes.empty())
	{
		throw SyntaxError("There aren't any fucking outcomes you dumbass");
	}
	// Should we allow this?
	//if (!title)
	//{
	//	throw SyntaxError("Missing title block in scenario");
	//}
	return std::make_unique<Scenario>(identifier, std::move(title), std::move(requirements), std::move(links), outcomes);
}

std::unique_ptr<const Outcome> parseOutcome(ExpressionParser& expressionParser, const std::string& identifier, const std::vector<std::string>& strings, std::size_t first, std::size_t last)
{
	const std::string fullFilename = fullPathFromDialogueIdentifier(identifier);
	for (int i = first; i < last; ++i)
		Logger::log(Logger::Parsing) << "po: " << strings[i] << std::endl;
	enum DType
	{
		Void,
		Result,
		Links,
		Message,
		Attributes,
		Requirements,
		NestedOutcome
	};
	DType lastType = Void;
	std::size_t lineOfLast = first;
	const char *str_result      = "result",
	           *str_links       = "links",
	           *str_message     = "message",
	           *str_attributes  = "attributes",
	           *str_requirement = "requirements",
	           *str_outcome     = "outcome",
	           *str_priority    = "priority";

	std::vector<std::string> optionFilenames;
	std::unique_ptr<const Statement> message;
	std::unique_ptr<const Statement> result;
	std::unique_ptr<const Statement> links;
	std::unique_ptr<const Statement> requirements;
	std::vector<std::unique_ptr<const Outcome>> outcomes;
	std::vector<std::pair<std::string, AttributeRange>> attributes;
	std::unique_ptr<const Expression> priority;
	std::string outcomeId;
	const char *c;
	int depth = 0;
	for (std::size_t i = first; i < last; ++i)
	{
		c = strings[i].c_str();
		eatWhitespace(c);
		Logger::log(Logger::Parsing) << "Oparse[" << (i>8?"":" ") << (i+1) << "|" << depth << "]: " << c << "\n";
		if (depth == 0)
		{
			if (strncmp(c, str_result, strlen(str_result)) == 0)
			{
				lastType = Result;
				c += strlen(str_result);
				eatWhitespace(c);
				// not end of line - thus inline message
				if (*c != '\0')
				{
					if (*c == '{')
					{
						throw SyntaxError("CURLY BRACES GO ON THE NEXT LINE FAGGOT");
					}
					result = parseSingleStatement(c);
					continue;
				}
			}
			else if (strncmp(c, str_links, strlen(str_links)) == 0)
			{
				lastType = Links;
			}
			else if (strncmp(c, str_message, strlen(str_message)) == 0)
			{
				lastType = Message;
				c += strlen(str_message);
				eatWhitespace(c);
				// not end of line - thus inline message
				if (*c != '\0')
				{
					if (*c == '{')
					{
						throw SyntaxError("CURLY BRACES GO ON THE NEXT LINE FAGGOT");
					}
					auto params = unique<Parameters>();
					params->addParameter(expressionParser.parseAsConcatenatedStrings(c));
					message = unique<Proc>(fullFilename, i, "local", "msg", std::move(params));
					continue;
				}
			}
			else if (strncmp(c, str_attributes, strlen(str_attributes)) == 0)
			{
				lastType = Attributes;
				c += strlen(str_attributes);
				eatWhitespace(c);
				if (strncmp(c, str_priority, strlen(str_priority)) == 0)
				{
					c += strlen(str_priority);
					eatWhitespace(c);
					priority = expressionParser.parseEntireExpression(c);
				}
			}
			else if (strncmp(c, str_requirement, strlen(str_requirement)) == 0)
			{
				lastType = Requirements;
				c += strlen(str_requirement);
				eatWhitespace(c);
				// not end of line - thus inline message
				if (*c != '\0')
				{
					if (*c == '{')
					{
						throw SyntaxError("CURLY BRACES GO ON THE NEXT LINE FAGGOT");
					}
					requirements = unique<SetVariable>(fullFilename, i, "outcome", "allow", expressionParser.parseEntireExpression(c));
					continue;
				}
			}
			else if (strncmp(c, str_outcome, strlen(str_outcome)) == 0)
			{
				lastType = NestedOutcome;
				c += strlen(str_outcome);
				eatWhitespace(c);
				outcomeId = c;
			}
			else if (strncmp(c, "//", 2) == 0 || *c == '\0' || *c == '\n' || *c == '\r' || *c == '{' || *c == '}')
			{
				//  comments and empty lines are okay
			}
			else
			{
				std::stringstream ss;
				ss << "Unknown statement: \""
				   << strings[i]
				   << "\" on line "
				   << (i + 1);
				throw SyntaxError(ss.str());
			}
			lineOfLast = i + 1;
		}
		if (*c == '{')
		{
			if (++depth == 1)
			{
				lineOfLast = i + 1;
			}
		}
		else if (*c == '}')
		{
			if (--depth == 0)
			{
				switch (lastType)
				{
					case Void:
						//  Why would you even bother to make an empty block?
						break;
					case Result:
						if (result)
						{
							throw SyntaxError("Additional result block encountered on line " + std::to_string(i + 1));
						}
						result = parseStrings(expressionParser, strings, lineOfLast, i, fullFilename);
						break;
					case Links:
						if (links)
						{
							throw SyntaxError("Additional links block encountered on line " + std::to_string(i + 1));
						}
						links = parseStrings(expressionParser, strings, lineOfLast, i, fullFilename);
						break;
					case Message:
						if (message)
						{
							throw SyntaxError("Additional message block encountered on line " + std::to_string(i + 1));
						}
						Logger::log(Logger::Parsing) << "Passing the following to statement parser:\n";
						for (int m = lineOfLast; m < i; ++m)
							Logger::log(Logger::Parsing) << "\t| " << strings[m] << std::endl;
						message = parseStrings(expressionParser, strings, lineOfLast, i, fullFilename);
						break;
					case Attributes:
					{
						if (!attributes.empty())
						{
							throw SyntaxError("Additional attributes block encountered on line " + std::to_string(i + 1));
						}
						const char *startOfc;
						float scalar = 1.f;
						Logger::log(Logger::Parsing) << "attributes:\n";
						for (int j = lineOfLast; j < i; ++j)
							Logger::log(Logger::Parsing) << "\t" << strings[j] << std::endl;
						Logger::log(Logger::Parsing) << "--end of attributes--\n";
						for (int j = lineOfLast; j < i; ++j)
						{
							Logger::log(Logger::Parsing) << "prs attribute: " << strings[j] << std::endl;
							c = strings[j].c_str();
							eatWhitespace(c);
							if (*c == '*')
							{
								++c;
								startOfc = c;
								eatWhitespace(c);
								for (; *c != '\n' && *c != '\0'; ++c)
								{
									if (isspace(*c))
									{
										scalar = atof(std::string(startOfc, c - startOfc).c_str());
										break;
									}
									else if (!isdigit(*c) && *c != '.')
									{
										std::stringstream ss;
										ss << "Expected scalar number after * in attribrute list, but found \"" << strings[i] << "\" on line" << (i + 1);
										throw SyntaxError(ss.str());
									}
								}
							}
							else if (strcmp(c, "//") == 0)
							{
								continue;
							}
							eatWhitespace(c);
							std::string attributeName;
							for (startOfc = c; *c != '\0'; ++c)
							{
								if (iswspace(*c))
								{
									attributeName.append(startOfc, c);
									eatWhitespace(c);
									break;
								}
							}
							if (attributeName.empty())
							{
								throw SyntaxError("Empty attribute value");
							}
							if (*c == '>')
							{
								++c;
								attributes.push_back(
									std::make_pair(
										std::move(attributeName),
										AttributeRange(
											expressionParser.parseEntireExpression(c),
											std::make_unique<Literal>(
												identifier,
												expressionParser.nextId(),
												std::numeric_limits<float>::max()),
											scalar)));
							}
							else if (*c == '<')
							{
								++c;
								attributes.push_back(
									std::make_pair(
										std::move(attributeName),
										AttributeRange(
											std::make_unique<Literal>(
												identifier,
												expressionParser.nextId(),
												std::numeric_limits<float>::min()),
											expressionParser.parseEntireExpression(c),
											scalar)));
							}
							else
							{
								auto attributeValue = expressionParser.parseExpression(c);
								c = attributeValue.second;

								// single
								if (c == nullptr)
								{
									attributes.push_back(std::make_pair(std::move(attributeName), AttributeRange(std::move(attributeValue.first), scalar)));
								}
								else // range
								{

									eatWhitespace(c);
									if (strncmp(c, "to", 2) == 0 && iswspace(*(c + 2)))
									{
										c += 2;
										eatWhitespace(c);
										std::unique_ptr<Expression> maxValue = expressionParser.parseEntireExpression(c);
										attributes.push_back(std::make_pair(std::move(attributeName), AttributeRange(std::move(attributeValue.first), std::move(maxValue), scalar)));
									}
									else
									{
										std::stringstream ss;
										ss << "Expected \'to\' keyword after non-terminal attribute expression, but found \"" << c << "\" on line" << (i + 1);
										throw SyntaxError(ss.str());
									}
								}
							}
						}
						break;
					}
					case Requirements:
					{
						if (requirements)
						{
							throw SyntaxError("Additional requirements block encountered on line " + std::to_string(i + 1));
						}
						requirements = parseStrings(expressionParser, strings, lineOfLast, i, fullFilename);
						break;
					}
					case NestedOutcome:
					{
						outcomes.push_back(parseOutcome(expressionParser, identifier + ":" + outcomeId, strings, lineOfLast, i));
						break;
					}
				}
				lastType = Void;
			}
			if (depth < 0)
			{
				std::stringstream ss;
				ss << "Extra } encountered: \""
				   << strings[i]
				   << "\" on line"
				   << (i + 1);
				throw SyntaxError(ss.str());
			}
		}
	}
	if (depth != 0)
	{
		//  uh oh, probbly complain or something here
		throw SyntaxError("Unmatched curly brace at end of file");
	}
	if (outcomes.empty())
	{
		if (!message)
		{
			throw SyntaxError("message and requirements block mandatory in leaf outcome blocks");
		}
	}
	else
	{
		if (result || message)
		{
			// why? just define an order...
			throw SyntaxError("message and result blocks are not allowed in non-leaf outcome blocks");
		}
	}
	return std::make_unique<const Outcome>(
		identifier,
		std::move(message),
		std::move(result),
		std::move(links),
		std::move(attributes),
		std::move(requirements),
		std::move(outcomes),
		std::move(priority));
}

std::string fullPathFromDialogueIdentifier(const std::string& dialogueName)
{
	return "resources/dlg/" + dialogueName + ".dlg";
}

} // ls
} // sl
