#include "ExpressionParser.hpp"

#include <algorithm>
#include <cstdlib>
#include <initializer_list>
#include <memory>
#include <numeric>
#include <vector>

#include "Parsing/SyntaxTree/Expressions/Variable.hpp"
#include "Parsing/SyntaxTree/Expressions/Literal.hpp"
#include "Parsing/SyntaxTree/Expressions/Arithmetic.hpp"
#include "Parsing/SyntaxTree/Expressions/Logical.hpp"
#include "Parsing/SyntaxTree/Expressions/Relational.hpp"
#include "Parsing/Errors/SyntaxError.hpp"
#include "Parsing/SyntaxTree/Expressions/FuncCall.hpp"
#include "Parsing/Parsers/ParamParser.hpp"
#include "Parsing/Parsers/ParserUtil.hpp"
#include "Parsing/Binding/Arguments.hpp"

#include "Utility/Assert.hpp"
#include "Utility/Logger.hpp"
#include "Utility/Mapping.hpp"

namespace sl
{
namespace ls
{

ExpressionParser::EToken::EToken(TokenType type, const char *startOfExpr)
	:type(type)
	,startOfExpr(startOfExpr)
	,unary(nullptr)
{
}

ExpressionParser::ENode::ENode(EToken type, ENode*& head)
	:type(type.type)
	,startOfExpr(type.startOfExpr)
	,exp(std::move(type.unary))
	,prev(nullptr)
	,next(nullptr)
	,head(head)
{
}

std::unique_ptr<Expression> ExpressionParser::ENode::createExpression(ExpressionParser& parser)
{
	if (!exp)
	{
		switch (type)
		{
			case Add:
				exp = std::make_unique<Addition>(parser.getFile(), parser.nextId(), prev->createExpression(parser), next->createExpression(parser));
				break;
			case Sub:
				exp = std::make_unique<Subtraction>(parser.getFile(), parser.nextId(), prev->createExpression(parser), next->createExpression(parser));
				break;
			case Mul:
				exp = std::make_unique<Multiplication>(parser.getFile(), parser.nextId(), prev->createExpression(parser), next->createExpression(parser));
				break;
			case Div:
				exp = std::make_unique<Division>(parser.getFile(), parser.nextId(), prev->createExpression(parser), next->createExpression(parser));
				break;
			case Gre:
				exp = std::make_unique<GreaterThan>(parser.getFile(), parser.nextId(), prev->createExpression(parser), next->createExpression(parser));
				break;
			case Les:
				exp = std::make_unique<LessThan>(parser.getFile(), parser.nextId(), prev->createExpression(parser), next->createExpression(parser));
				break;
			case GrE:
				exp = std::make_unique<GreaterThanOrEqual>(parser.getFile(), parser.nextId(), prev->createExpression(parser), next->createExpression(parser));
				break;
			case LeE:
				exp = std::make_unique<LessThanOrEqual>(parser.getFile(), parser.nextId(), prev->createExpression(parser), next->createExpression(parser));
				break;
			case Eq:
				exp = std::make_unique<Equals>(parser.getFile(), parser.nextId(), prev->createExpression(parser), next->createExpression(parser));
				break;
			case NEq:
				exp = std::make_unique<Inverse>(parser.getFile(), parser.nextId(), std::make_unique<Equals>(parser.getFile(), parser.nextId(), prev->createExpression(parser), next->createExpression(parser)));
				break;
			case And:
				exp = std::make_unique<LogicalAnd>(parser.getFile(), parser.nextId(), prev->createExpression(parser), next->createExpression(parser));
				break;
			case Or:
				exp = std::make_unique<LogicalOr>(parser.getFile(), parser.nextId(), prev->createExpression(parser), next->createExpression(parser));
				break;
			case Inv:
				exp = std::make_unique<Inverse>(parser.getFile(), parser.nextId(), next->createExpression(parser));
				break;
			case UMi:
				exp = std::make_unique<UnaryMinus>(parser.getFile(), parser.nextId(), next->createExpression(parser));
				break;
			default:
				break;
		}
		if (isBinaryOperator(type))
		{
			consomePrev();
			consumeNext();
		}
		else if (isUnaryOperator(type))
		{
			consumeNext();
		}
	}
	return std::move(exp);
}

void ExpressionParser::ENode::consomePrev()
{
	if (!prev)
	{
		return;
	}
	startOfExpr = std::min(startOfExpr, prev->leftMostStartOfExpr());
	if (prev == head)
	{
		head = this;
	}
	ENode *temp = prev;
	if (prev->prev)
	{
		prev->prev->next = this;
	}
	prev = prev->prev;

	delete temp;
}

void ExpressionParser::ENode::consumeNext()
{
	if (!next)
	{
		return;
	}
	startOfExpr = std::min(startOfExpr, next->leftMostStartOfExpr());
	ENode *temp = next;
	if (next->next)
	{
		next->next->prev = this;
	}
	next = next->next;

	delete temp;
}

const char* ExpressionParser::ENode::leftMostStartOfExpr() const
{
	// TODO: why does this work? because root nodes always consume the prev?
	const char *leftStart = startOfExpr;//prev ? prev->leftMostStartOfExpr() : startOfExpr;
	const char *rightStart = next ? next->leftMostStartOfExpr() : startOfExpr;
	return std::min(std::min(leftStart, rightStart), startOfExpr);
}

void ExpressionParser::lex(std::vector<EToken>& tokens, const char *exp)
{
	tokens.clear();
	bool singleChar; //  whether or not a single-char token was parsed on a given iteration
	for (const char *c = exp; *c != '\0';)
	{
		const char *startOfC = c;
		singleChar = true;
		if (*c == '"')
		{
			singleChar = true;
			++c;
			std::string buffer;
			while (*c != '\0' && (*c != '"' || *(c - 1) == '\\'))
			{
				if (*c != '\\')
				{
					buffer += *c;
				}
				++c;
			}
			if (*c == '\0')
			{
				throw SyntaxError("Unterminated string literal");
			}
			else
			{
				Logger::log(Logger::Parsing) << "string literal found: [" << buffer << "]" << std::endl;
				tokens.push_back(EToken{ Str, startOfC });
				tokens.back().unary = std::make_unique<Literal>(filename, nextId(), Value(std::move(buffer)));
			}
		}
		else if (*c == '+')							  //  Arithmetic operators
		{
			tokens.push_back(EToken{ Add, startOfC});
		}
		else if (*c == '-')
		{
			if (tokens.empty() || !isOperand(tokens.back().type))
			{
				tokens.push_back(EToken{ UMi, startOfC });
			}
			else
			{
				tokens.push_back(EToken{ Sub, startOfC });
			}
		}
		else if (*c == '*')
		{
			tokens.push_back(EToken{ Mul, startOfC });
		}
		else if (*c == '/')
		{
			tokens.push_back(EToken{ Div, startOfC });
		}
		else if (*c == '=')					 //  Realational operators
		{
			while (*(c + 1) == '=')
			{
				++c;
			}
			tokens.push_back(EToken{ Eq, startOfC });
		}
		else if (*c == '>')
		{
			if (*(c + 1) == '=')
			{
				tokens.push_back(EToken{ GrE, startOfC });
				++c;
			}
			else
			{
				tokens.push_back(EToken{ Gre, startOfC });
			}
		}
		else if (*c == '<')
		{
			if (*(c + 1) == '=')
			{
				tokens.push_back(EToken{ LeE, startOfC });
				++c;
			}
			else
			{
				tokens.push_back(EToken{ Les, startOfC });
			}
		}
		else if (*c == '!')
		{
			if (*(c + 1) == '=')
			{
				while (*(c + 1) == '=')
				{
					++c;
				}
				tokens.push_back(EToken{ NEq, startOfC });
			}
			else								//  Logical operators
			{
				tokens.push_back(EToken{ Inv, startOfC });
			}
		}
		else if (isKeyword(c, "or", 2))
		{
			tokens.push_back(EToken{ Or, startOfC });
			++c;
		}
		else if (*c == '|')
		{
			if (*(c + 1) == '|')
			{
				++c;
			}
			tokens.push_back(EToken{ Or, startOfC });
		}
		else if (isKeyword(c, "and", 3))
		{
			tokens.push_back(EToken{ And, startOfC });
			c += 2;
		}
		else if (*c == '&' && *(c + 1) == '&')
		{
			++c;
			tokens.push_back(EToken{ And, startOfC });
		}
		else if (*c == '(')
		{
			tokens.push_back(EToken{ LBr, startOfC });
		}
		else if (*c == ')')
		{
			tokens.push_back(EToken{ RBr, startOfC });
		}
		else if (isValidIdentifierStart(*c))  //  variables / functions
		{
			singleChar = false;
			bool onScope = true;
			std::string scope, name;
			while (isValidIdentifierBody(*c) || *c == ':')
			{
				if (*c == ':')
				{
					onScope = false;
				}
				else if (onScope)
				{
					scope += *c;
				}
				else
				{
					name += *c;
				}
				++c;
			}
			if (onScope)	//  Must be a local var then
			{
				name = scope;
				scope = "local";
			}

			if (*c == '(')	// might be a function
			{
				tokens.push_back(EToken{ Fun, startOfC });
				int depthLevel = 0;
				const char *startOfParam = c + 1;
				bool stringLitInParams = false;
				while (*c != '\0')
				{
					if (stringLitInParams)
					{
						if (*c == '"' && *(c - 1) != '\\') // it's okay to access c-1 since this can't be 1st char
						{
							stringLitInParams = false;
						}
					}
					else
					{
						if (*c == '"')
						{
							stringLitInParams = true;
						}
						else if (*c == '(')
						{
							++depthLevel;
						}
						else if (*c == ')')
						{
							if (--depthLevel <= 0)
							{
								std::string buffer(startOfParam, c - startOfParam);
								Logger::log(Logger::Parsing) << "arglist of " << scope << "." << name << "() is [" << buffer << "]" << std::endl;
								tokens.back().unary = std::make_unique<FuncCall>(filename, nextId(), scope, name, parseFunctionParameters(*this, buffer.c_str()));
								break;
							}
						}
					}
					++c;
				}
			}
			else
			{
				tokens.push_back(EToken{ Var, startOfC });
				tokens.back().unary = std::make_unique<Variable>(filename, nextId(), scope, name);
			}
		}
		else if (*c >= '0' && *c <= '9')		//  Numerical constants
		{
			singleChar = false;
			tokens.push_back(EToken{ Num, startOfC });
			std::string number;
			bool decimal = false;
			while ((*c >= '0' && *c <= '9') || *c == '.')
			{
				if (*c == '.')
				{
					decimal = true;
				}
				number += *c;
				++c;
			}
			if (decimal)
			{
				tokens.back().unary = std::make_unique<Literal>(filename, nextId(), Value((float)atof(number.c_str())));
			}
			else
			{
				tokens.back().unary = std::make_unique<Literal>(filename, nextId(), Value((int)atof(number.c_str())));
			}
		}
		if (singleChar)
		{
			++c;
		}
	}
}

std::vector<std::pair<std::unique_ptr<Expression>, const char*>> ExpressionParser::parse(std::vector<EToken>& tokens, std::size_t start, std::size_t end)
{
	ENode *head = nullptr, *cur = nullptr, *temp, *tail = nullptr;
	try
	{
		//handle parenthesis fuckery
		for (std::size_t i = start; i < end; ++i)
		{
			if (tokens[i].type == RBr)
			{
				continue;
			}
			temp = new ENode(std::move(tokens[i]), head);
			if (!head)
			{
				head = temp;
			}
			if (cur)
			{
				cur->next = temp;
				temp->prev = cur;
			}
			cur = tail = temp;

			if (tokens[i].type == LBr)
			{
				std::size_t newStart = i + 1;
				int depth = 0;
				while (i < end)
				{
					if (tokens[i].type == LBr)
					{
						++depth;
					}
					else if (tokens[i].type == RBr)
					{
						--depth;
						if (depth == 0)
						{
							break;
						}

					}
					++i;
				}
				//skip ahead until Rbr then call this on that
				auto curExp = parse(tokens, newStart, i);
				if (curExp.size() != 1)
				{
					throw SyntaxError("Multiple expressions found within an expression");
				}
				cur->exp = std::move(curExp.front().first);
			}
		}

		if (!head)
		{
			// failure?
			ERROR("is this a failure?");
			//return std::vector<std::unique_ptr<Expression>>({ std::make_unique<Literal>(filename, nextId(), Value()) });
			throw SyntaxError("no expression found");
		}

		//  Iterate next-to-prev to parse unary ops
		for (ENode *it = tail; it != nullptr; it = it->prev)
		{
			if (isUnaryOperator(it->type))
			{
				it->exp = it->createExpression(*this);
			}
		}


		parseLevel({Mul, Div}, head);
		parseLevel({Add, Sub}, head);
		parseLevel({Gre, Les, GrE, LeE}, head);
		parseLevel({Eq, NEq}, head);
		parseLevel({And}, head);
		parseLevel({Or}, head);
	}
	catch (const SyntaxError& see)
	{
		//  Free all the dynamically allocated memory in the node-chain before throwing
		for (ENode *it = head, *temp = nullptr; it != nullptr; )
		{
			/*if (it->exp)
			{
				delete it->exp;
			}*/
			temp = it;
			it = it->next;
			delete temp;
		}
		throw see;
	}

	std::vector<std::pair<std::unique_ptr<Expression>, const char*>> expressions;
	do
	{
		expressions.push_back(std::make_pair(head->createExpression(*this), head->leftMostStartOfExpr()));
		head = head->next;
		if (head)
		{
			head->prev = nullptr; // prevent it from using the just-parsed stuff.
		}
	}
	while (head);
	return expressions;
}

void ExpressionParser::parseLevel(std::initializer_list<TokenType> il, ENode*& head)
{
	ENode *it = head;

	while (it != nullptr)
	{
		for (auto tokType = il.begin(); tokType != il.end(); ++tokType)
		{
			if (it->type == *tokType)
			{
				it->exp = it->createExpression(*this);
				break;  //  Already matched one type, no need to check rest
			}
		}
		it = it->next;
	}
}

/*static*/bool ExpressionParser::isBinaryOperator(TokenType type)
{
	return (!isUnaryOperator(type) && !isOperand(type));
}
/*static*/bool ExpressionParser::isUnaryOperator(TokenType type)
{
	return (type == Inv || type == UMi);
}
/*static*/bool ExpressionParser::isOperand(TokenType type)
{
	return (type == Var || type == Num || type == Str || type == RBr || type == LBr || type == Fun);
}

/*static*/char ExpressionParser::tokenToChar(TokenType type)
{
	switch (type)
	{
		case Add:
			return '+';
		case Sub:
			return '-';
		case Mul:
			return '*';
		case Div:
			return '/';
		case Var:
			return 'V';
		case Num:
			return 'C';
		case Str:
			return 'S';
		case RBr:
			return ')';
		case LBr:
			return '(';
		case Inv:
			return '!';
		case UMi:
			return 'm';
		case Gre:
			return '>';
		case Les:
			return '<';
		case GrE:
			return 'g';
		case LeE:
			return 'l';
		case Eq:
			return '=';
		case NEq:
			return 'n';
		case And:
			return '&';
		case Or:
			return '|';
		default:
			return '?';
	}
}

ExpressionParser::ExpressionParser(const std::string& filename)
	:filename(filename)
	,uniqueId(0)
{
}

std::unique_ptr<Expression> ExpressionParser::parseEntireExpression(const char *exp)
{
	std::vector<std::unique_ptr<Expression>> expressions = parseExpressions(exp);
	if (expressions.empty())
	{
		//  This shouldn't happen. I don't think it can, actually.
		ASSERT(":(");
		return std::make_unique<Literal>(getFile(), nextId(), Value(0));
	}
	if (expressions.size() > 1)
	{
		throw SyntaxError("Multiple expressions found");
	}
	return std::move(expressions.front());
}

std::pair<std::unique_ptr<Expression>, const char*> ExpressionParser::parseExpression(const char *exp)
{
	std::vector<EToken> tokens;

	lex(tokens, exp);
	std::vector<std::pair<std::unique_ptr<Expression>, const char*>> expressions = parse(tokens, 0, tokens.size());
	if (expressions.empty())
	{
		//  This shouldn't happen. I don't think it can, actually.
		ASSERT(":(");
		return std::make_pair(std::make_unique<Literal>(getFile(), nextId(), Value(0)), nullptr);
	}
	return std::make_pair(
		std::move(expressions.front().first),
		expressions.size() > 1 ? expressions[1].second : nullptr);
}

std::vector<std::unique_ptr<Expression>> ExpressionParser::parseExpressions(const char *exp)
{
	std::vector<EToken> tokens;

	lex(tokens, exp);

	return mapVecMove(parse(tokens, 0, tokens.size()), [](auto&& p) { return std::move(p.first); });
}

std::unique_ptr<Expression> ExpressionParser::parseAsConcatenatedStrings(const char *exp)
{
	auto strings = splitIntoConcatenatedStrings(exp);
	return concatenatedStrings(strings.cbegin(), strings.cend());
}

std::unique_ptr<Expression> ExpressionParser::concatenatedStrings(std::vector<std::string>::const_iterator begin, std::vector<std::string>::const_iterator end)
{
	std::vector<std::unique_ptr<Expression>> expressions = mapVec(begin, end, [this](const std::string& string) { return parseEntireExpression(string.c_str()); });
	std::unique_ptr<Expression> ret;
	if (!expressions.empty())
	{
		ret = std::move(expressions.front());
		for (int i = 1; i < expressions.size(); ++i)
		{
			ret = std::make_unique<Addition>(filename, nextId(), std::move(ret), std::move(expressions[i]));
		}
	}
	return ret;
}

std::vector<std::string> ExpressionParser::splitIntoConcatenatedStrings(const char *exp) const
{
	bool inStringLiteral = false;
	const char *start = exp;
	const char *c;
	std::vector<std::string> strings;
	std::string buff;
	auto append = [&] {
		bool hasStuff = false;
		for (char ch : buff)
		{
			if (ch != ' ' && ch != '\t')
			{
				hasStuff = true;
				break;
			}
		}
		if (hasStuff)
		{
			// trim trailing whitespace
			buff.erase(buff.find_last_not_of(" \t") + 1);
			//std::cout << "string concat[" << buff << "]\n";
			strings.push_back(std::move(buff));
			buff.clear();
			start = c;
		}
	};
	int depth = 0;
	for (c = exp; *c; ++c)
	{
		if (!inStringLiteral && depth == 0 && *c == '+')
		{
			append();
			continue;
		}
		// trim leading whitespace
		if (!isWhitespace(*c) || !buff.empty())
		{
			buff += *c;
		}
		if (*c == '\"' && depth == 0)
		{
			if (c == exp || c[-1] != '\\')
			{
				if (inStringLiteral)
				{
					append();
				}
				inStringLiteral = !inStringLiteral;
				continue;
			}
		}
		if (!inStringLiteral)
		{
			if (*c == '(')
			{
				++depth;
			}
			else if (*c == ')')
			{
				--depth;
			}
			if (depth == 0 && *c != ':' && (!isValidIdentifierBody(*c) || (*c != '(' && c[1] == '\"')))
			{
				append();
			}
		}
	}
	append();
	return strings;
}

const std::string& ExpressionParser::getFile() const
{
	return filename;
}

int ExpressionParser::nextId()
{
	return uniqueId++;
}

} // ls
} // sl
