#ifndef SL_LS_EXPRESSIONPARSER_HPP
#define SL_LS_EXPRESSIONPARSER_HPP

#include "Parsing/SyntaxTree/Expressions/Expression.hpp"
#include "Parsing/Binding/Bindings.hpp"

#include <memory>

namespace sl
{
namespace ls
{

class ExpressionParser
{
public:
	ExpressionParser(const std::string& filename);
	
	/**
	 * Parses just a single expression - throws SyntaxError if extra successive expressions exist.
	 * @return expression parsed from {\param exp}
	 */
	std::unique_ptr<Expression> parseEntireExpression(const char *exp);

	/**
	 * Parses the first expression and ignores successive ones.
	 * @return <e, c> where e is the FIRST expression parsed from {\param exp} and c is the first character AFTER the string used to parse e if additional expressions exist, else it is nullptr
	 */
	std::pair<std::unique_ptr<Expression>, const char*> parseExpression(const char *exp);

	/**
	 * @return All expressions <e, c> parsed from {\param exp} where c points to the FIRST char of the expression e was parsed from.
	 */
	std::vector<std::unique_ptr<Expression>> parseExpressions(const char *exp);

	std::unique_ptr<Expression> parseAsConcatenatedStrings(const char *exp);

	std::unique_ptr<Expression> concatenatedStrings(std::vector<std::string>::const_iterator begin, std::vector<std::string>::const_iterator end);

	std::vector<std::string> splitIntoConcatenatedStrings(const char *exp) const;

	const std::string& getFile() const;

	int nextId();

private:
	enum TokenType
	{
		Add,
		Sub,
		Mul,
		Div,
		Var,
		Num,
		Str,
		RBr, 
		LBr,
		Inv,
		UMi,
		Gre,
		Les,
		GrE,
		LeE,
		Eq,
		NEq,
		And,
		Or,
		Fun
	};

	struct EToken
	{
		EToken(const TokenType type, const char *startOfExpr);


		const TokenType type;
		const char *startOfExpr;
		std::unique_ptr<Expression> unary; // constants or vars
	};


	struct ENode
	{
		ENode(const EToken type, ENode*& head);

		std::unique_ptr<Expression> createExpression(ExpressionParser& parser);

		void consomePrev();
		
		void consumeNext();

		const char* leftMostStartOfExpr() const;

		const TokenType type;
		
		std::unique_ptr<Expression> exp;
		ENode *prev, *next;
		ENode*& head;

	private:

		const char *startOfExpr;
	};

	void lex(std::vector<EToken>& tokens, const char *exp);

	std::vector<std::pair<std::unique_ptr<Expression>, const char*>> parse(std::vector<EToken>& tokens, std::size_t start, std::size_t end);

	void parseLevel(std::initializer_list<TokenType> il, ENode*& head);

	static bool ExpressionParser::isBinaryOperator(TokenType type);

	static bool ExpressionParser::isUnaryOperator(TokenType type);

	static bool ExpressionParser::isOperand(TokenType type);

	static char ExpressionParser::tokenToChar(TokenType type);


	std::string filename;
	int uniqueId;
};


} // ls
} // sl

#endif // SL_LS_EXPRESSIONPARSER_HPP
