#include "Parsing/Parsers/StatementParser.hpp"

#include <initializer_list>
#include <sstream>
#include <fstream>
#include <cstdlib>
#include <cstring>
#include <set>

#include "Parsing/Parsers/ExpressionParser.hpp"
#include "Parsing/Errors/SyntaxError.hpp"
#include "Parsing/Parsers/ParamParser.hpp"
#include "Parsing/Parsers/ParserUtil.hpp"

// for the Fork statement
#include "Parsing/SyntaxTree/Expressions/FuncCall.hpp"
#include "Parsing/SyntaxTree/Expressions/Literal.hpp"

#include "Parsing/SyntaxTree/Statements/If.hpp"
#include "Parsing/SyntaxTree/Statements/ChanceProbability.hpp"
#include "Parsing/SyntaxTree/Statements/Block.hpp"
#include "Parsing/SyntaxTree/Statements/Proc.hpp"
#include "Parsing/SyntaxTree/Statements/SetVariable.hpp"
#include "Parsing/SyntaxTree/Statements/SwitchCase.hpp"
#include "Parsing/SyntaxTree/Statements/WhileLoop.hpp"

//  Because of +=, -=, *=, and /=
#include "Parsing/SyntaxTree/Expressions/Arithmetic.hpp"

#include "Parsing/SyntaxTree/Expressions/Variable.hpp"

#include "Utility/FileUtil.hpp"
#include "Utility/Logger.hpp"

namespace sl
{
namespace ls
{

enum class TokenType
{
	If,
	Else,
	While,
	Switch,
	Case,
	Default,
	Choice,
	Option,
	Chance,
	Probability,
	SetVar,
	ProcedureCall,
	OpenBrace,
	CloseBrace,
	Message
};

struct SToken
{
	SToken(TokenType type, std::size_t lineNumber, const std::string& line);

	TokenType type;
	std::unique_ptr<Statement> statement;//in the case of constants, vars, and flags
	std::size_t lineNumber;
	std::string line;
	std::unique_ptr<Expression> expression, extraExpression;
};


struct SNode
{
	SNode(SToken type, SNode*& head);

	std::unique_ptr<Statement> createStatement(const std::string& filename);
	void consomePrev();
	void consumeNext();

	TokenType type;
	std::unique_ptr<Statement> statement;
	SNode *prev, *next;
	SNode*& head;
	const std::size_t lineNumber;
	const std::string line;
	std::unique_ptr<const Expression> expression, extraExpression;

private:
	void creteStatementForOption(const std::string& filename, std::initializer_list<TokenType> toEat)
	{
		std::vector<std::unique_ptr<Statement>> caseStatements;
		while (next && std::find(toEat.begin(), toEat.end(), next->type) != toEat.end())
		{
			caseStatements.push_back(next->createStatement(filename));
			consumeNext();
		}
		statement = std::make_unique<Block>(filename, lineNumber, std::move(caseStatements));
	}
};


static void lex(ExpressionParser& expressionParser, std::vector<SToken>& tokens, const std::vector<std::string>& strings, std::size_t start, std::size_t end, const std::string& filename);

static std::unique_ptr<Statement> parse(ExpressionParser& expressionParser, std::vector<SToken>& tokens, std::size_t start, std::size_t end, const std::string& filename);

static char tokenToChar(TokenType type);



std::unique_ptr<Statement> parseSingleStatement(const char *str)
{
	std::vector<std::string> lines = {
		std::string(str)
	};
	return parseStrings(ExpressionParser("n/a"), lines);
}

std::unique_ptr<Statement> parseFile(const char *filename)
{
	eatWhitespace(filename);
	std::unique_ptr<Statement> logic;
	try
	{
		logic = parseStrings(ExpressionParser(filename), fileToLines(filename), 0, 0, filename);
	}
	catch (const SyntaxError& see)
	{
		throw SyntaxError(see, " in " + std::string(filename));
	}
	return std::move(logic);
}

std::unique_ptr<Statement> parseStrings(ExpressionParser& expressionParser, const std::vector<std::string>& strings, std::size_t first, std::size_t last, const std::string& filename)
{
	if (last == 0)
	{
		last = strings.size();
	}

	std::vector<SToken> tokens;

	std::unique_ptr<Statement> retVal;

	for (int i = first; i < last; ++i)
		Logger::log(Logger::Parsing) << "ps:" << strings[i] << "\n";

	lex(expressionParser, tokens, strings, first, last, filename);

	retVal = parse(expressionParser, tokens, 0, tokens.size(), filename);

	if (!retVal)
	{
		//  I hope this never happens
		retVal = std::make_unique<Block>(filename, 0);  //  empty block
		ERROR("empty result for parseStrings()");
	}

	return std::move(retVal);
}

void executeFile(const char *filename, Bindings& context)
{
	parseFile(filename)->execute(context);
	Logger::log(Logger::Parsing) << "doneparsing" << filename << "\n";
}

void executeStrings(const std::vector<std::string>& strings, Bindings& context)
{
	parseStrings(ExpressionParser("n/a"), strings)->execute(context);
}

SToken::SToken(TokenType type, std::size_t lineNumber, const std::string& line)
	:type(type)
	,statement(nullptr)
	,lineNumber(lineNumber)
	,line(line)
	,expression(nullptr)
{
}



SNode::SNode(SToken type, SNode*& head)
	:type(type.type)
	,statement(std::move(type.statement))
	,prev(nullptr)
	,next(nullptr)
	,head(head)
	,lineNumber(type.lineNumber)
	,line(type.line)
	,expression(std::move(type.expression))
	,extraExpression(std::move(type.extraExpression))
{
}

std::unique_ptr<Statement> SNode::createStatement(const std::string& filename)
{
	Logger::log(Logger::Parsing) << "   createStatement()[" << tokenToChar(type) << "]:   ";
	if (!statement)
	{
		Logger::log(Logger::Parsing) << "\tcreating statement on heap\n";
		if (type == TokenType::If)
		{
			//  I don't think it's physically possible to cram more 'next's into the next couple lines...
			if (next && next->next && next->next->next && next->next->type == TokenType::Else)
			{
				statement = std::make_unique<If>(filename, lineNumber, std::move(expression), next->createStatement(filename), next->next->next->createStatement(filename));
				next->next->consumeNext();
				next->consumeNext();
				consumeNext();
			}
			else
			{
				if (!next)
				{
					std::stringstream ss;
					ss << "If statement followed by no other statements: \""
					   << line
					   << "\" on line #"
					   << lineNumber;

					throw SyntaxError(ss.str());
				}
				statement = std::make_unique<If>(filename, lineNumber, std::move(expression), next->createStatement(filename));
				consumeNext();
			}
		}
		else if (type == TokenType::While)
		{
			if (next)
			{
				statement = std::make_unique<WhileLoop>(filename, lineNumber, std::move(expression), next->createStatement(filename));
				consumeNext();
			}
			else
			{
				std::stringstream ss;
				ss << "While statement followed by no other statements: \""
					<< line
					<< "\" on line #"
					<< lineNumber;

				throw SyntaxError(ss.str());
			}
		}
		else if (type == TokenType::Case)
		{
			creteStatementForOption(filename, { TokenType::CloseBrace, TokenType::Case, TokenType::Default });
		}
		else if (type == TokenType::Default)
		{
			creteStatementForOption(filename, { TokenType::CloseBrace, TokenType::Case, TokenType::Default });
		}
		else if (type == TokenType::Option)
		{
			creteStatementForOption(filename, { TokenType::CloseBrace, TokenType::Option });
		}
		else if (type == TokenType::Probability)
		{
			creteStatementForOption(filename, { TokenType::CloseBrace, TokenType::Probability });
		}
		else if (type == TokenType::Message)
		{
			statement = std::make_unique<Proc>(filename, lineNumber, "local", "msg", std::make_unique<Parameters>(std::move(expression)));
		}
	}
	return std::move(statement);
}

void SNode::consomePrev()
{
	if (!prev)
	{
		return;
	}
	if (prev == head)
	{
		head = this;
	}
	SNode *temp = prev;
	if (prev->prev)
	{
		prev->prev->next = this;
	}
	prev = prev->prev;
	Logger::log(Logger::Parsing) << "killed prev\n";
	delete temp;
}

void SNode::consumeNext()
{
	if (!next)
	{
		Logger::log(Logger::Parsing) << "couldn't kill next (\n";
		return;
	}
	SNode *temp = next;
	if (next->next)
	{
		next->next->prev = this;
	}
	next = next->next;
	Logger::log(Logger::Parsing) << "killed next\n";
	delete temp;
}


/*		  END OF TOKEN METHODS			*/


////////////////////////////////////////////////////
//		  STATIC FUNCTION DEFINITIONS		   //
////////////////////////////////////////////////////

void lex(ExpressionParser& expressionParser, std::vector<SToken>& tokens, const std::vector<std::string>& strings, std::size_t first, std::size_t last, const std::string& filename)
{
	tokens.clear(); //  memory leaks lol
	bool parsedElse;
	Logger::log(Logger::Parsing) << "Parsing statements: ";
	for (std::size_t line = first; line < last; ++line)
	{
		const char *c = strings[line].c_str();
		parsedElse = false;

		Logger::log(Logger::Parsing) << "[" << strings[line] << "]";

		eatWhitespace(c);

		//	ignore blank lines or comments
		if (isEnd(*c) || *c == '@' || strncmp(c, "//", 2) == 0)
		{
			continue;
		}

		if (*c == '{')
		{
			tokens.push_back(SToken(TokenType::OpenBrace, line + 1, strings[line]));
			continue;
		}
		else if (*c == '}')
		{
			tokens.push_back(SToken(TokenType::CloseBrace, line + 1, strings[line]));
			continue;
		}
		if (isKeyword(c, "else", 4))
		{
			tokens.push_back(SToken(TokenType::Else, line + 1, strings[line]));
			c += 4;
			parsedElse = true;
		}
		eatWhitespace(c);
		if (isKeyword(c, "if", 2))
		{
			tokens.push_back(SToken(TokenType::If, line + 1, strings[line]));
			c += 2;
			tokens.back().expression = expressionParser.parseEntireExpression(c);
		}
		else if (isKeyword(c, "while", 5))
		{
			tokens.push_back(SToken(TokenType::While, line + 1, strings[line]));
			c += 5;
			tokens.back().expression = expressionParser.parseEntireExpression(c);
		}
		else if (isKeyword(c, "switch", 6))
		{
			tokens.push_back(SToken(TokenType::Switch, line + 1, strings[line]));
			c += 6;
			tokens.back().expression = expressionParser.parseEntireExpression(c);
		}
		else if (isKeyword(c, "case", 4))
		{
			tokens.push_back(SToken(TokenType::Case, line + 1, strings[line]));
			c += 4;
			tokens.back().expression = expressionParser.parseEntireExpression(c);
		}
		else if (isKeyword(c, "default", 7))
		{
			tokens.push_back(SToken(TokenType::Default, line + 1, strings[line]));
			c += 7;
		}
		else if (isKeyword(c, "choice", 6))
		{
			tokens.push_back(SToken(TokenType::Choice, line + 1, strings[line]));
			c += 6;
			eatWhitespace(c);
			if (*c != '\n' && *c != '\r' && *c != '\0')
			{
				tokens.back().expression = expressionParser.parseAsConcatenatedStrings(c);
			}
		}
		else if (isKeyword(c, "option", 6))
		{
			tokens.push_back(SToken(TokenType::Option, line + 1, strings[line]));
			c += 6;
			auto expressionStrings = expressionParser.splitIntoConcatenatedStrings(c);
			auto ifString = std::find(expressionStrings.cbegin(), expressionStrings.cend(), "if");
			if (ifString == expressionStrings.cend())
			{
				tokens.back().expression = expressionParser.concatenatedStrings(expressionStrings.cbegin(), expressionStrings.cend());
				tokens.back().extraExpression = std::make_unique<Literal>(expressionParser.getFile(), expressionParser.nextId(), ls::Value(true));
			}
			else
			{
				std::string boolPartString;
				for (auto it = std::next(ifString); it != expressionStrings.cend(); ++it)
				{
					boolPartString += *it + " ";
				}
				tokens.back().expression = expressionParser.concatenatedStrings(expressionStrings.cbegin(), ifString);
				tokens.back().extraExpression = expressionParser.parseEntireExpression(boolPartString.c_str());
			}
		}
		else if (isKeyword(c, "chance", 6))
		{
			tokens.push_back(SToken(TokenType::Chance, line + 1, strings[line]));
			c += 6;
			eatWhitespace(c);
			if (*c != '\n' && *c != '\r' && *c != '\0')
			{
				// TODO: different types
			}
		}
		else if (isKeyword(c, "probability", 11))
		{
			tokens.push_back(SToken(TokenType::Probability, line + 1, strings[line]));
			c += 11;
			tokens.back().expression = expressionParser.parseEntireExpression(c);
		}
		else if (isKeyword(c, "m", 1))
		{
			tokens.push_back(SToken(TokenType::Message, line + 1, strings[line]));
			c += 1;
			tokens.back().expression = expressionParser.parseAsConcatenatedStrings(c);
			ASSERT(tokens.back().expression);
		}
		else if (isValidIdentifierStart(*c))  //  variables / flags
		{
			bool onScope = true;
			std::string scope, name;
			while (isValidIdentifierBody(*c) || *c == '.' || *c == ':')
			{
				if (*c == ':')
				{
					onScope = false;
				}
				else if (onScope)
				{
					scope += *c;
				}
				else
				{
					name += *c;
				}
				++c;
			}
			if (onScope)	//  local var
			{
				name = scope;
				scope = "local";
			}
			eatWhitespace(c);
			if (*c == '(')	// might be a function
			{
				tokens.push_back(SToken(TokenType::ProcedureCall, line + 1, strings[line]));
				int depthLevel = 0;
				const char *startOfParam = c + 1;
				bool isInStringLiteral = false;
				while (*c != '\0')
				{
					if (isInStringLiteral)
					{
						if (*c == '"' && *(c - 1) != '\\') // safe since never in string literal in first char
						{
							isInStringLiteral = false;
						}
					}
					else
					{
						if (*c == '"')
						{
							isInStringLiteral = true;
						}
						else if (*c == '(')
						{
							++depthLevel;
						}
						else if (*c == ')')
						{
							if (--depthLevel <= 0)
							{
								std::string buffer(startOfParam, c - startOfParam);
								//std::cout << "arglist of " << scope << "." << name << "() is [" << buffer << "]" << std::endl;
								tokens.back().statement = std::make_unique<Proc>(filename, line, scope, name, parseFunctionParameters(expressionParser, buffer.c_str()));
								break;
							}
						}
					}
					++c;
				}
			}
			else
			{
				tokens.push_back(SToken(TokenType::SetVar, line + 1, strings[line]));
				while (*c != '=' && *c != '\0')
				{
					++c;
				}
				if (*c == '\0')
				{
					std::stringstream ss;
					ss << "Missing expression in \""
					   << strings[line]
					   << "\" on line #"
					   << (line + 1);
					throw SyntaxError(ss.str());
				}
				++c;
				std::unique_ptr<Expression> exp = expressionParser.parseEntireExpression(c);
				std::unique_ptr<Expression> self = std::make_unique<Variable>(expressionParser.getFile(), expressionParser.nextId(), scope, name);
				if (*(c - 2) == '+')
				{
					exp = std::make_unique<Addition>(expressionParser.getFile(), expressionParser.nextId(), std::move(self), std::move(exp));
				}
				else if (*(c - 2) == '-')
				{
					exp = std::make_unique<Subtraction>(expressionParser.getFile(), expressionParser.nextId(), std::move(self), std::move(exp));
				}
				else if (*(c - 2) == '*')
				{
					exp = std::make_unique<Multiplication>(expressionParser.getFile(), expressionParser.nextId(), std::move(self), std::move(exp));
				}
				else if (*(c - 2) == '/')
				{
					exp = std::make_unique<Division>(expressionParser.getFile(), expressionParser.nextId(), std::move(self), std::move(exp));
				}
				eatWhitespace(c);
				tokens.back().statement = std::make_unique<SetVariable>(filename, line, scope, name, std::move(exp));
			}
		}
		else if (isEnd(*c) || !parsedElse || strncmp(c, "//", 2) == 0)
		{
			// empty line, comment, etc
		}
		else
		{
			std::stringstream ss;
			ss << "Encountered unknown statement \""
			   << strings[line]
			   << "\" on line #"
			   << (line + 1);
			throw SyntaxError(ss.str());
		}
	}
	Logger::log(Logger::Parsing) << "\ndone creating tokens" << std::endl;
}

std::unique_ptr<Statement> parse(ExpressionParser& expressionParser, std::vector<SToken>& tokens, std::size_t start, std::size_t end, const std::string& filename)
{
	Logger::log(Logger::Parsing) << "parse() on:";
	for (std::size_t i = start; i < end; ++i)
	{
		Logger::log(Logger::Parsing) << tokenToChar(tokens[i].type) << " ";
	}
	Logger::log(Logger::Parsing) << "\n";
	if (tokens.empty())
	{
		Logger::log(Logger::Parsing) << "empty tokens?\n";
		return std::make_unique<Block>(filename, start);
	}
	SNode *head = nullptr, *cur = nullptr, *temp;
	std::vector<std::unique_ptr<Statement>> statements;
	try
	{
		for (std::size_t i = start; i < end; ++i)
		{
			Logger::log(Logger::Parsing) << tokenToChar(tokens[i].type) << ",";
			if (tokens[i].type == TokenType::CloseBrace)
			{
				continue;
			}
			temp = new SNode(std::move(tokens[i]), head);
			if (!head)
			{
				head = temp;
			}
			if (cur)
			{
				cur->next = temp;
				temp->prev = cur;
			}
			cur = temp;
			if (tokens[i].type == TokenType::OpenBrace)
			{
				std::size_t newStart = i + 1;
				int depth = 0;
				while (i < end)
				{
					if (tokens[i].type == TokenType::OpenBrace)
					{
						++depth;
					}
					else if (tokens[i].type == TokenType::CloseBrace)
					{
						--depth;
						if (depth == 0)
						{
							break;
						}
					}
					++i;
				}
				//skip ahead until CloseBrace then call this on that
				cur->statement = parse(expressionParser, tokens, newStart, i, filename);
			}
			else if (tokens[i].type == TokenType::Switch)
			{
				std::size_t newStart = i + 2;
				int depth = 0;
				std::set<int> validCases; // since we want to ignore case statements that are nested!
				while (i < end)
				{
					if (depth == 1 && (tokens[i].type == TokenType::Case || tokens[i].type == TokenType::Default))
					{
						validCases.insert(i);
					}
					if (tokens[i].type == TokenType::OpenBrace)
					{
						++depth;
					}
					else if (tokens[i].type == TokenType::CloseBrace)
					{
						--depth;
						if (depth == 0)
						{
							break;
						}
					}
					++i;
				}
				//skip ahead until CloseBrace
				auto switchCase = std::make_unique<SwitchCase>(filename, tokens[i].lineNumber, std::move(cur->expression));
				size_t caseStart = 0, caseEnd;
				auto handleEnd = [&] {
					if (caseStart != 0)
					{
						switch (tokens[caseStart].type)
						{
						case TokenType::Case:
							switchCase->addCase(std::move(tokens[caseStart].expression), parse(expressionParser, tokens, caseStart + 1, caseEnd, filename));
							break;
						case TokenType::Default:
							ASSERT(tokens[caseStart].expression == nullptr);
							if (!switchCase->setDefault(parse(expressionParser, tokens, caseStart + 1, caseEnd, filename)))
							{
								std::stringstream ss;
								ss << "Multiple default case statements in a switch on line " << tokens[caseStart].lineNumber;
								throw SyntaxError(ss.str());
							}
							break;
						default:
							ERROR(std::string("unknown token type handled in a switch: ") + tokenToChar(tokens[caseStart].type));
						}
					}
				};
				for (caseEnd = newStart; caseEnd < i; ++caseEnd)
				{
					if ((tokens[caseEnd].type == TokenType::Case || tokens[caseEnd].type == TokenType::Default) && validCases.count(caseEnd))
					{
						handleEnd();
						caseStart = caseEnd;
					}
				}
				handleEnd();
				cur->statement = std::move(switchCase);
			}
			else if (tokens[i].type == TokenType::Choice)
			{
				std::size_t newStart = i + 2;
				int depth = 0;
				std::set<int> validOptions; // since we want to ignore case statements that are nested!
				while (i < end)
				{
					if (depth == 1 && tokens[i].type == TokenType::Option)
					{
						validOptions.insert(i);
					}
					if (tokens[i].type == TokenType::OpenBrace)
					{
						++depth;
					}
					else if (tokens[i].type == TokenType::CloseBrace)
					{
						--depth;
						if (depth == 0)
						{
							break;
						}
					}
					++i;
				}
				//skip ahead until CloseBrace
				std::vector<std::unique_ptr<Expression>> optionExpressions;
				std::vector<std::unique_ptr<Expression>> optionConditions;
				std::vector<std::unique_ptr<const Statement>> optionBodies;
				size_t caseStart = 0, caseEnd;
				auto handleEnd = [&] {
					if (caseStart != 0)
					{
						optionExpressions.push_back(std::move(tokens[caseStart].expression));
						optionConditions.push_back(std::move(tokens[caseStart].extraExpression));
						optionBodies.push_back(parse(expressionParser, tokens, caseStart + 1, caseEnd, filename));
					}
				};
				for (caseEnd = newStart; caseEnd < i; ++caseEnd)
				{
					if (tokens[caseEnd].type == TokenType::Option && validOptions.count(caseEnd))
					{
						handleEnd();
						caseStart = caseEnd;
					}
				}
				handleEnd();
				ASSERT(optionExpressions.size() == optionBodies.size());
				ASSERT(optionConditions.size() == optionBodies.size());
				auto forkParams = std::make_unique<Parameters>();
				for (int paramIndex = 0; paramIndex < optionBodies.size(); ++paramIndex)
				{
					forkParams->addParameter(std::move(optionConditions[paramIndex]));
					forkParams->addParameter(std::move(optionExpressions[paramIndex]));
				}
				auto forkCall = std::make_unique<FuncCall>(expressionParser.getFile(), expressionParser.nextId(), "local", "forkWithConds", std::move(forkParams));
				auto switchCase = std::make_unique<SwitchCase>(filename, tokens[i].lineNumber, std::move(forkCall));
				for (int optionIndex = 0; optionIndex < optionBodies.size(); ++optionIndex)
				{
					switchCase->addCase(std::make_unique<Literal>(expressionParser.getFile(), expressionParser.nextId(), Value(optionIndex)), std::move(optionBodies[optionIndex]));
				}
				if (cur->expression) // choice had a message
				{
					auto msgCall = std::make_unique<Proc>(filename, tokens[i].lineNumber, "local", "msg", std::make_unique<Parameters>(std::move(cur->expression)));
					std::vector<std::unique_ptr<Statement>> block;
					block.push_back(std::move(msgCall));
					block.push_back(std::move(switchCase));
					cur->statement = std::make_unique<Block>(filename, tokens[i].lineNumber, std::move(block));
				}
				else
				{
					cur->statement = std::move(switchCase);
				}
			}
			else if (tokens[i].type == TokenType::Chance)
			{
				std::size_t newStart = i + 2;
				int depth = 0;
				std::set<int> validCases; // since we want to ignore case statements that are nested!
				while (i < end)
				{
					if (depth == 1 && tokens[i].type == TokenType::Probability)
					{
						validCases.insert(i);
					}
					if (tokens[i].type == TokenType::OpenBrace)
					{
						++depth;
					}
					else if (tokens[i].type == TokenType::CloseBrace)
					{
						--depth;
						if (depth == 0)
						{
							break;
						}
					}
					++i;
				}
				//skip ahead until CloseBrace
				auto switchCase = std::make_unique<ChanceProbability>(filename, tokens[i].lineNumber);
				size_t caseStart = 0, caseEnd;
				auto handleEnd = [&] {
					if (caseStart != 0)
					{
						switchCase->addCase(std::move(tokens[caseStart].expression), parse(expressionParser, tokens, caseStart + 1, caseEnd, filename));
					}
				};
				for (caseEnd = newStart; caseEnd < i; ++caseEnd)
				{
					if (tokens[caseEnd].type == TokenType::Probability && validCases.count(caseEnd))
					{
						handleEnd();
						caseStart = caseEnd;
					}
				}
				handleEnd();
				cur->statement = std::move(switchCase);
			}
		}
		Logger::log(Logger::Parsing) << "{";
		for (SNode *it = head; it != nullptr; it = it->next)
		{
			Logger::log(Logger::Parsing) << "eval" << tokenToChar(it->type) << std::endl;
			if (it->type == TokenType::Else)
			{
				//  All Else tokens should've been consumed by If
				std::stringstream ss;
				ss << "Unexpected else statement in \""
				   << it->line
				   << "\" on line #"
				   << it->lineNumber;
				throw SyntaxError(ss.str());
			}
			else if (it->type == TokenType::Case)
			{
				//  All Case tokens should've been consumed by Switch
				std::stringstream ss;
				ss << "Unexpected case statement in \""
					<< it->line
					<< "\" on line #"
					<< it->lineNumber;
				throw SyntaxError(ss.str());
			}
			else if (it->type == TokenType::Default)
			{
				//  All Default tokens should've been consumed by Switch
				std::stringstream ss;
				ss << "Unexpected default statement in \""
					<< it->line
					<< "\" on line #"
					<< it->lineNumber;
				throw SyntaxError(ss.str());
			}
			else if (it->type == TokenType::Option)
			{
				//  All Case tokens should've been consumed by Choice
				std::stringstream ss;
				ss << "Unexpected option statement in \""
					<< it->line
					<< "\" on line #"
					<< it->lineNumber;
				throw SyntaxError(ss.str());
			}
			//it->statement = it->createStatement();
			statements.push_back(it->createStatement(filename));
		}
		Logger::log(Logger::Parsing) << "}";
	}
	catch (const SyntaxError& see)
	{
		//  Free all the dynamically allocated memory in the node-chain before throwing
		for (SNode *it = head, *temp = nullptr; it != nullptr; )
		{
			temp = it;
			it = it->next;
			delete temp;
		}
		throw see;
	}
	if (statements.size() == 1)
	{
		return std::move(statements.front());
	}
	return std::make_unique<Block>(filename, start, std::move(statements));
}

char tokenToChar(TokenType type)
{
	switch (type)
	{
		case TokenType::If:
			return 'i';
		case TokenType::Else:
			return 'e';
		case TokenType::While:
			return 'w';
		case TokenType::Switch:
			return 's';
		case TokenType::Case:
			return 'c';
		case TokenType::Default:
			return 'd';
		case TokenType::Choice:
			return 'f';
		case TokenType::Option:
			return 'o';
		case TokenType::Chance:
			return 'C';
		case TokenType::Probability:
			return 'P';
		case TokenType::SetVar:
			return 'v';
		case TokenType::ProcedureCall:
			return 'p';
		case TokenType::OpenBrace:
			return '{';
		case TokenType::CloseBrace:
			return '}';
		case TokenType::Message:
			return 'm';
		default:
			return '?';
	}
}

} // ls
} // sl
