#include "Parsing/SyntaxTree/Expressions/Relational.hpp"

namespace sl
{
namespace ls
{

/*			Equal to				*/
Equals::Equals(const std::string& filename, int uniqueId, std::unique_ptr<const Expression> lhs, std::unique_ptr<const Expression> rhs)
	:Expression(filename, uniqueId)
	,lhs(std::move(lhs))
	,rhs(std::move(rhs))
{
}

Value Equals::onEvaluate(const Bindings& context) const
{
	return Value(lhs->evaluate(context) == rhs->evaluate(context));
}

void Equals::lint(LintBindings& context) const
{
	lhs->lint(context);
	rhs->lint(context);
}




/*			Less than				*/
LessThan::LessThan(const std::string& filename, int uniqueId, std::unique_ptr<const Expression> lhs, std::unique_ptr<const Expression> rhs)
	:Expression(filename, uniqueId)
	,lhs(std::move(lhs))
	,rhs(std::move(rhs))
{
}

Value LessThan::onEvaluate(const Bindings& context) const
{
	return Value(lhs->evaluate(context) < rhs->evaluate(context));
}

void LessThan::lint(LintBindings& context) const
{
	lhs->lint(context);
	rhs->lint(context);
}




/*			Less than or equal		*/
LessThanOrEqual::LessThanOrEqual(const std::string& filename, int uniqueId, std::unique_ptr<const Expression> lhs, std::unique_ptr<const Expression> rhs)
	:Expression(filename, uniqueId)
	,lhs(std::move(lhs))
	,rhs(std::move(rhs))
{
}

Value LessThanOrEqual::onEvaluate(const Bindings& context) const
{
	const Value left = lhs->evaluate(context);
	const Value right = rhs->evaluate(context);
	return Value(left < right || left == right);
}

void LessThanOrEqual::lint(LintBindings& context) const
{
	lhs->lint(context);
	rhs->lint(context);
}




/*			Greater than			*/
GreaterThan::GreaterThan(const std::string& filename, int uniqueId, std::unique_ptr<const Expression> lhs, std::unique_ptr<const Expression> rhs)
	:Expression(filename, uniqueId)
	,lhs(std::move(lhs))
	,rhs(std::move(rhs))
{
}

Value GreaterThan::onEvaluate(const Bindings& context) const
{
	return Value(rhs->evaluate(context) < lhs->evaluate(context));
}

void GreaterThan::lint(LintBindings& context) const
{
	lhs->lint(context);
	rhs->lint(context);
}




/*			Greater than or equal	*/
GreaterThanOrEqual::GreaterThanOrEqual(const std::string& filename, int uniqueId, std::unique_ptr<const Expression> lhs, std::unique_ptr<const Expression> rhs)
	:Expression(filename, uniqueId)
	,lhs(std::move(lhs))
	,rhs(std::move(rhs))
{
}

Value GreaterThanOrEqual::onEvaluate(const Bindings& context) const
{
	const Value left = lhs->evaluate(context);
	const Value right = rhs->evaluate(context);
	return Value(right < left || left == right);
}

void GreaterThanOrEqual::lint(LintBindings& context) const
{
	lhs->lint(context);
	rhs->lint(context);
}

} // ls
} // sl
