/**
 * @section DESCRIPTION
 * A configuration for a driving car. Used for pathfinding.
 * It represents the possible (discretized, minimalized) states a car can be in.
 * This includes position, angle, and speed.
 * NOTE: The coordinates here are HALF-tiles in the Town space, rather than pixels or whole-tiles.
 */
#ifndef SL_PF_CARCONFIG_HPP
#define SL_PF_CARCONFIG_HPP

#include "Pathfinding/NodeID.hpp"
#include "Utility/Vector.hpp"

namespace sl
{

class Car;
class Scene;

namespace pf
{

class CarConfig
{
public:
	enum class Speed
	{
		REVERSE = 0,
		SLOW,
		MEDIUM,
		FAST
	};
	static const float speedToRealSpeed[4];
	static const float angleToRealAngle[16];

	enum class Angle
	{
		East = 0,
		North1East2,
		NorthEast,
		North2East1,
		North,
		North2West1,
		NorthWest,
		North1West2,
		West,
		South1West2,
		SouthWest,
		South2West1,
		South,
		South2East1,
		SouthEast,
		South1East2,
	};
	CarConfig(Vector2i pos, Speed speed, Angle angle);

	NodeID toNodeID() const;

	float distance(const CarConfig& rhs) const;

	float distance(const NodeID& rhs) const;

	bool operator==(const CarConfig& rhs) const {
		return pos == rhs.pos && speed == rhs.speed && angle == rhs.angle;
	}

	bool operator!=(const CarConfig& rhs) const {
		return !(*this == rhs);
	}

	bool operator<(const CarConfig& rhs) const {
		if (pos < rhs.pos)
		{
			return true;
		}
		else if (rhs.pos < pos)
		{
			return false;
		}
		if (speed < rhs.speed)
		{
			return true;
		}
		else if (rhs.speed < speed)
		{
			return false;
		}
		return angle < rhs.angle;
	}

	const Vector2i& getPos() const { return pos; }

	Speed getSpeed() const { return speed; }

	Angle getAngle() const { return angle; }

	/**
	 * @return The coordinates of this configuration in AbstractScene-space
	 */
	Vector2f getTownPos() const { return Vector2f(pos.x, pos.y) * halfTownTileSize; }

	


	static CarConfig carToConfig(const Car& car);
	
	// Only call this from pathfind() I guess?
	static Scene* getTown() { return town; }

	static void setTown(Scene *newTown);

private:
	// coordinate in HALF node/tile coords
	Vector2i pos;
	Speed speed;
	Angle angle;


	static Scene *town;
	static float halfTownTileSize;
};

} // pf
} // sl

#endif // SL_PF_CARCONFIG_HPP