#include "Pathfinding/GridPathManager.hpp"

#include <unordered_map>

#include "Engine/Door.hpp"
#include "NPC/Locators/Target.hpp"
#include "Engine/Scene.hpp"
#include "Pathfinding/Pathfind.hpp"
#include "Utility/Assert.hpp"
#include "Utility/Random.hpp"
#include "Utility/Logger.hpp"
#include "Utility/Memory.hpp"
#include <iostream>// @todo pls remove
namespace sl
{
namespace pf
{

GridPathManager::GridPathManager()
#ifndef SL_NOMULTITHREADED
	:calculationThreads()
#endif // SL_NOMULTITHREADED
{
#ifndef SL_NOMULTITHREADED
	// no copy ctor defined, so can't use that
	for (int i = 0; i < 3; ++i)
	{
		calculationThreads.push_back(PathThread(*this));
	}
#endif // SL_NOMULTITHREADED
}

std::shared_ptr<Path> GridPathManager::getPath(const PathSpec& spec, Priority priority)
{
	ASSERT(spec.heuristic);
	std::shared_ptr<Path> ret = std::make_shared<Path>(*this);
#ifndef SL_NOMULTITHREADED
	if (priority != Priority::Realtime)
	{
		// TODO: differentiate between non-real-time priorities.
		//       possible to simply turn the calculationQueue into a priority queue, and additional to always
		//       launch a new thread when high priority requests come in, and to not launch calculations on lower priorities
		//       while there are >= normal paththread count high-priority threads, to ensure the high-priority ones always
		//       have maximum CPU usage... or even just PQs might be fine.
		//std::lock_guard<std::mutex> guard(calQueueLock);
		calQueueLock.lock();
		calculationQueue.push_back(Args(ret, spec));
		Logger::log(Logger::Pathfinding, Logger::Debug) << "<GridPathManager>  -  Adding calcuation, now " << calculationQueue.size() << " calculations [" << std::this_thread::get_id() << "]\n";
		calQueueLock.unlock();

		// if there is a free thread, restart it
		for (PathThread& pathThread : calculationThreads)
		{
			if (pathThread.refresh())
			{
				break;
			}
		}
	}
	else
#endif // SL_NOMULTITHREADED
	// Real-time process
	{
		const bool success = findPathImpl(spec, *ret);
	}
	return ret;
}

bool GridPathManager::cancelCalculation(std::shared_ptr<Path>& path)
{
#ifndef SL_NOMULTITHREADED
	try
	{
		std::lock_guard<std::mutex> guard(calQueueLock);
		bool isFirst = true;
		for (auto it = calculationQueue.begin(); it != calculationQueue.end(); )
		{
			if (isFirst)
			{
				isFirst = false;
				//	maybe do something here later
				if (it->path == path)
				{
					// it's being calculated already so just let it finish
					return true;
				}
			}
			else
			{
				if (it->path == path)
				{
					it = calculationQueue.erase(it);
					return true;//break;//continue;
				}
			}
			++it;
		}
	}
	catch (std::exception& e)
	{
		Logger::log(Logger::Pathfinding, Logger::Serious) << "exception in path threaded loop:\n" << e.what();
	}
	catch (...)
	{
		Logger::log(Logger::Pathfinding, Logger::Serious) << "unknown exception in path threaded loop\n";
	}
	//calculationThread.join(); // why the fuck was this join call here?!
#endif // SL_NOMULTITHREADED
	return false;
}

void GridPathManager::cancelAllCalculations()
{
#ifndef SL_NOMULTITHREADED
	calQueueLock.lock();
	calculationQueue.clear();
	calQueueLock.unlock();
	// wait for the calculation queue to join so it isn't still processing its current item when we leave since
	// cancellAllCalculations() is likely being called on shut-down so we don't want pathing to be
	// collision-detecting against freed scenes or other bad things like that
	for (PathThread& pathThread : calculationThreads)
	{
		pathThread.cancel();
	}
#endif // SL_NOMULTITHREADED
}

NodeID GridPathManager::getRandomID() const
{
	Random& rng = Random::get();
	auto it = pathGrids.begin();
	std::advance(it, rng.random(pathGrids.size()));
	Scene* scene = it->first;
	const int ts = scene->getTileSize();
	return NodeID(rng.random(scene->getWidth() / ts), rng.random(scene->getHeight() / ts), scene);
}

NodeID GridPathManager::getClosestNodeID(const Target& target) const
{
	Scene *scene = target.getScene();
	const int tileSize = scene->getTileSize();
	const int width = scene->getWidth() / tileSize;
	const int height = scene->getHeight() / tileSize;
	int x = target.getX() / tileSize;
	if (x < 0)
	{
		x = 0;
	}
	else if (x >= width)
	{
		x = width - 1;
	}
	int y = target.getY() / tileSize;
	if (y < 0)
	{
		y = 0;
	}
	else if (y >= height)
	{
		y = height - 1;
	}
	return NodeID(x, y, scene);
}

Node GridPathManager::getNodeFromID(NodeID id) const
{
	return Node(id.getX(), id.getY(), id.getScene());
}

void GridPathManager::update(int64_t timestamp)
{
	for (auto& sceneGrids : pathGrids)
	{
		for (auto& pathGrid : sceneGrids.second)
		{
			pathGrid->update(timestamp);
		}
	}
}


void GridPathManager::registerGrid(Scene *scene, std::shared_ptr<PathGrid> pathGrid)
{
	auto& vec = pathGrids[scene];
#ifdef SL_DEBUG
	if (!vec.empty())
	{
		ASSERT(vec.front()->getWidth() == pathGrid->getWidth());
		ASSERT(vec.front()->getHeight() == pathGrid->getHeight());
		ASSERT(vec.front()->getTileSize() == pathGrid->getTileSize());
	}
#endif // SL_DEBUG
	vec.push_back(pathGrid);
}

#ifdef SL_DEBUG
static std::map<Vector2i, std::set<Vector2i>> extractGraphAsAdjList(const PathGrid& grid)
{
	std::map<Vector2i, std::set<Vector2i>> adjlist;
	for (int y = 0; y < grid.getHeight(); ++y)
	{
		for (int x = 0; x < grid.getWidth(); ++x)
		{
			//if (!grid.isSolid(x, y))
			{
				const Vector2i pos(x, y);
				if (y < grid.getHeight() - 1 && grid.canGoDown(x, y))
				{
					adjlist[pos].insert(Vector2i(x, y + 1));
				}
				if (x > 0 && grid.canGoLeft(x, y))
				{
					adjlist[pos].insert(Vector2i(x - 1, y));
				}
				if (x < grid.getWidth() - 1 && grid.canGoRight(x, y))
				{
					adjlist[pos].insert(Vector2i(x + 1, y));
				}
				if (y > 0 && grid.canGoUp(x, y))
				{
					adjlist[pos].insert(Vector2i(x, y - 1));
				}
			}
		}
	}
	return adjlist;
}

std::vector<std::map<Vector2i, std::set<Vector2i>>> GridPathManager::exctractGraphsAsAdjlists(Scene *scene) const
{
	std::vector<std::map<Vector2i, std::set<Vector2i>>> output;
	auto it = pathGrids.find(scene);
	if (it != pathGrids.end())
	{
		for (const auto& pathGrid : it->second)
		{
			auto adj = extractGraphAsAdjList(*pathGrid);
			if (!adj.empty())
			{
				output.push_back(std::move(adj));
			}
		}
	}
	return output;
}
#endif // SD_DEBUG

// private
#ifndef SL_NOMULTITHREADED
GridPathManager::PathThread::PathThread(GridPathManager& manager)
	:man(&manager)
	,lock(unique<std::mutex>())
{
}

GridPathManager::PathThread::PathThread(PathThread&& other)
	:man(other.man)
	,thread(std::move(other.thread))
	,lock(std::move(other.lock))
{
}

GridPathManager::PathThread& GridPathManager::PathThread::operator=(PathThread&& rhs)
{
	man = rhs.man;
	thread = std::move(rhs.thread);
	lock = std::move(rhs.lock);
	return *this;
}

GridPathManager::PathThread::~PathThread()
{
	cancel();
}

bool GridPathManager::PathThread::refresh()
{
	if (lock->try_lock())
	{
		//std::cout <<  "creating new path thread[" << std::this_thread::get_id() << "]\n";
		lock->unlock();
		if (thread.joinable())
		{
			thread.join();
		}
		thread = std::thread(&GridPathManager::PathThread::calculatePathLoop, this, 0);
		return true;
	}
	return false;
}

void GridPathManager::PathThread::cancel()
{
	if (thread.joinable())
	{
		thread.join();
	}
}

void GridPathManager::PathThread::calculatePathLoop(int id)
{
	Logger::log(Logger::Pathfinding, Logger::Debug) << "<GridPathManager>  -  Starting: calculatePathLoop()[" << std::this_thread::get_id() << "]\n";
	//try
	{
		//std::cout << "Starting: calculatePathLoop()[" << std::this_thread::get_id() << "]\n";
		std::lock_guard<std::mutex> guard(*lock);
		for ( ; ; )
		{
			if (sf::Keyboard::isKeyPressed(sf::Keyboard::Key::P))
				std::cout << "pathQueue.size() = " << man->calculationQueue.size() << std::endl;
			//std::lock_guard<std::mutex> qlock(calQueueLock);
			man->calQueueLock.lock();
			if (man->calculationQueue.empty())
			{
				man->calQueueLock.unlock();
				break;
			}
			const Args top = std::move(man->calculationQueue.front());
			man->calculationQueue.pop_front();
			man->calQueueLock.unlock();

			const bool success = man->findPathImpl(top.spec, *top.path);
			//Logger::log(Logger::Pathfinding, Logger::Debug) << "<GridPathManager>  - Path from ("
			//	<< top.from.getX() << ", " << top.from.getY() << ") to (" << top.path->lastNode().getX() << ", " << top.path->lastNode().getY() << ") computed by [" << std::this_thread::get_id() << "] now " << calculationQueue.size() << " calculations" << std::endl;
		}

		Logger::log(Logger::Pathfinding, Logger::Debug) << "<GridPathManager>  -  Finishing: calculatePathLoop()[" << std::this_thread::get_id() << "]\n";
	}
	//catch (std::exception& e)
	//{
	//	Logger::log(Logger::Pathfinding, Logger::Serious) << "FUCKING THREAD CRASHED MOTHERFUCKER: " << e.what() << std::endl;
	//	Logger::log(Logger::Pathfinding, Logger::Serious) << "STOP MOTHERFUCKER\n";
	//	man->calQueueLock.unlock();
	//}
}
#endif // SL_NOMULTITHREADED

bool GridPathManager::findPathImpl(const PathSpec& spec, Path& output) const
{
	output.phase = Path::Phase::Calculating;
	std::map<const NodeID, const NodeID> portals;
	for (auto& pgrid : pathGrids)
	{
		std::list<Door*> doors;
		pgrid.first->cull(doors, Rect<int>(pgrid.first->getWidth(), pgrid.first->getHeight()));
		for (const Door *door : doors)
		{
			auto ret = portals.insert(std::make_pair(door->getEntranceNodeID(), door->getExitNodeID()));
			ASSERT(ret.second);
		}
	}
	std::deque<NodeID> nodes;
	const bool success = findShortestPathOnGrid(nodes, spec.from, spec.to, pathGrids, *spec.weights, portals, spec.heuristic, spec.context);
	if (success)
	{
		for (const NodeID& node : nodes)
		{
			output.path.push_back(Node(node.getX(), node.getY(), node.getScene()));
		}
		if (spec.postProcess)
		{
			spec.postProcess(output.path);
		}
	}
	output.phase = nodes.empty() ? Path::Phase::DNE : Path::Phase::Calculated;
	return success;
}

} // pf
} // sl
