#include "Pathfinding/Path.hpp"

#include "Pathfinding/PathManager.hpp"
#include "Utility/Assert.hpp"
#include "Utility/FlatMap.hpp"
#include "Utility/Trig.hpp"

namespace sl
{
namespace pf
{

static constexpr int TOTAL_DISTANCE_NOT_CALCULATED = -1.f;

Path::Path(PathManager& manager)
	:path()
	,phase(Phase::Queued)
	,manager(&manager)
	,totalDistCache(TOTAL_DISTANCE_NOT_CALCULATED)
	,lookup()
{
}

Path::Path(std::deque<Node>&& path)
	:path(path)
	,phase(Phase::Calculated)
	,manager(nullptr)
	,totalDistCache(TOTAL_DISTANCE_NOT_CALCULATED)
	,lookup()
{
}

Path::Path(Path&& other)
	:Path(std::move(other.path))
{
	ASSERT(other.isCalculated());
	other.phase = Phase::DNE;
}

Path::~Path()
{
	// TODO: Why is this commented out?
	//manager.cancelCalculation(this);
}

bool Path::isCalculated() const
{
	return phase == Phase::Calculated;
}

bool Path::isFinished() const
{
	return path.empty();
}

int Path::size() const
{
	return path.size();
}

float Path::totalSceneDistance() const
{
	if (totalDistCache == TOTAL_DISTANCE_NOT_CALCULATED && isCalculated())
	{
		totalDistCache = 0.f;
		if (!path.empty())
		{
			auto prev = path.begin();
			for (auto cur = std::next(prev), end = path.end(); cur != end; ++prev, ++cur)
			{
				if (cur->getScene() == prev->getScene())
				{
					totalDistCache += pointDistance(prev->getPos(), cur->getPos());
				}
			}
		}
	}
	return totalDistCache;
}

ScenePos Path::PathNodeInterpolation::position() const
{
	if (start.getScene() != end.getScene())
	{
		return progress >= 0.5f ? end.getScenePos() : start.getScenePos();
	}
	return ScenePos(
		*start.getScene(),
		start.getPos() * (1.f - progress) + end.getPos() * progress);
}

Path::PathNodeInterpolation Path::interpolate(float sceneDistance) const
{
	float distSoFar = 0.f;
	auto prev = path.begin();
	for (auto cur = std::next(prev), end = path.end(); cur != end; ++prev, ++cur)
	{
		if (cur->getScene() == prev->getScene())
		{
			const float prevToCur = pointDistance(prev->getPos(), cur->getPos());
			distSoFar += prevToCur;
			if (distSoFar >= sceneDistance)
			{
				return PathNodeInterpolation{*prev, *cur, 1.f - (distSoFar - sceneDistance) / prevToCur};
			}
		}
	}
	return PathNodeInterpolation{path.back(), path.back(), 1.f};
}

const Node& Path::nextNode() const
{
	ASSERT(!path.empty());
	ASSERT(phase == Phase::Calculated);

	return path.front();
}

const Node& Path::lastNode() const
{
	ASSERT(!path.empty());
	ASSERT(phase == Phase::Calculated);

	return path.back();
}

const Path::LookUp& Path::getLookup() const
{

	if (!lookup)
	{
		ASSERT(!path.empty());
		ASSERT(phase == Phase::Calculated);

		LookUp::Data data;
		data.reserve(path.size());
		for (auto it = path.cbegin(); it != path.cend(); ++it)
		{
			data.push_back(std::make_pair(it->getID(), it));
		}
		lookup = std::make_unique<LookUp>(std::move(data));
	}
	return *lookup;
}

void Path::pop()
{
	ASSERT(!path.empty());
	ASSERT(phase == Phase::Calculated);

	path.pop_front();
}

const std::deque<Node>& Path::getNodes() const
{
	return path;
}

bool Path::didPathingFail() const
{
	return phase == Phase::DNE;
}

/*static*/std::shared_ptr<Path> Path::merge(std::initializer_list<std::pair<std::deque<Node>::const_iterator, std::deque<Node>::const_iterator>> paths)
{
	std::deque<Node> ret;
	for (const auto& path : paths)
	{
		ret.insert(ret.end(), path.first, path.second);
	}
	return std::make_shared<Path>(std::move(ret));
}


} // pf
} // sl