#ifndef SL_PF_PATH_HPP
#define SL_PF_PATH_HPP

#include <deque>

#include "Engine/ScenePos.hpp"
#include "Pathfinding/Node.hpp"

namespace sl
{

template <typename K, typename V>
class FlatMap;

namespace pf
{

class PathManager;
class VertexBasedPathManager;

class Path
{
public:
	Path(PathManager& manager);

	/**
	 * Explicitly construct a path with a pre-computed path and no manager involved.
	 */
	Path(std::deque<Node>&& path);
	/*
	 * Only works on already-calculated paths
	 */
	Path(Path&& other);

	~Path();

	bool isCalculated() const;

	bool isFinished() const;
	
	int size() const;

	/**
	 * @return Total distance in scene-space only (not worldspace!). does not adjust to the Scene's scale, nor does it count any distance for inter-scene movement
	 */
	float totalSceneDistance() const;

	struct PathNodeInterpolation
	{
		ScenePos position() const;

		Node start;
		Node end;
		float progress;
	};
	PathNodeInterpolation interpolate(float sceneDistance) const;

	const Node& nextNode() const;

	const Node& lastNode() const;

	using LookUp = FlatMap<NodeID, std::deque<Node>::const_iterator>;
	const LookUp& getLookup() const;

	void pop();

	const std::deque<Node>& getNodes() const;

	bool didPathingFail() const;

	static std::shared_ptr<Path> merge(std::initializer_list<std::pair<std::deque<Node>::const_iterator, std::deque<Node>::const_iterator>> paths);

private:
	/*
	 * Deleted due to no valid behavour if we are currently calculating, since the PathManager would be referencing this object
	 */
	Path(const Path& other) = delete;
	/*
	 * Deleted due to no valid behavour if we are currently calculating, since the PathManager would be referencing this object
	 */
	Path& operator=(Path&& rhs) = delete;
	/*
	 * Deleted due to no valid behavour if we are currently calculating, since the PathManager would be referencing this object
	 */
	Path& operator=(const Path& rhs) = delete;

	enum class Phase
	{
		Queued,
		Calculating,
		Calculated,
		DNE
	};

	std::deque<Node> path;
	Phase phase;
	PathManager *manager; // TODO: remove?
	mutable float totalDistCache;
	mutable std::unique_ptr<LookUp> lookup;

	friend class VertexBasedPathManager;
	friend class GridPathManager;
};

} // pf
} // sl

#endif // SL_PF_PATH_HPP