#include "Pathfinding/Pathfind.hpp"

#pragma optimize( "gty", on )

#include <chrono>
#include <climits>
#include <queue>
#include <set>

#include "Engine/Scene.hpp"

#include "Pathfinding/PathGrid.hpp"
#include "Utility/Assert.hpp"
#include "Utility/Logger.hpp"
#include "Utility/Trig.hpp"

#include <SFML/Graphics.hpp>

namespace sl
{
namespace pf
{

void decreaseKey(Vertex **array, unsigned long int *distances, std::size_t index)
{
	Vertex **parent;
	while (index > 0)
	{
		parent = array + (index - 1) / 2;
		if (distances[array[index]->distIndex] >= distances[(*parent)->distIndex])
			break;
		Vertex *vertexSwap = array[index];
		array[index] = *parent;
		*parent = vertexSwap;

		array[index]->heapIndex = index;
		(*parent)->heapIndex = parent - array;

		index = (index - 1) / 2;
	}
}

void increaseKey(Vertex **array, unsigned long int *distances, std::size_t index, std::size_t size)
{
	Vertex **smallest;
	while (index * 2 + 1 < size)
	{
		smallest = array + (index * 2) + 1;   //  &array[2 * index + 1]
		if (index * 2 + 2 < size && distances[(*(smallest + 1))->distIndex] < distances[(*smallest)->distIndex])
			++smallest;
		if (distances[array[index]->distIndex] <= distances[(*smallest)->distIndex])
			break;
		Vertex *vertexSwap = array[index];
		array[index] = *smallest;
		*smallest = vertexSwap;

		array[index]->heapIndex = index;
		(*smallest)->heapIndex = smallest - array;

		index = smallest - array;
	}
}

void heapify(Vertex **array, unsigned long int *distances, std::size_t size)
{
	for (std::size_t i = 0; i < size; ++i)
	{
		decreaseKey(array, distances, i);
		array[i]->heapIndex = i;
	}
}

void pop(Vertex **array, unsigned long int *distances, std::size_t size)
{
	array[size - 1]->heapIndex = 0;
	array[0] = array[size - 1];
	increaseKey(array, distances, 0, size - 1);
}

bool pathExists(bool *visited, const Vertex *current, const Vertex *end)
{
	visited[current->distIndex] = true;
	for (const std::pair<const Vertex*, int>& vertex : current->edges)
	{
		if (vertex.first == end)
		{
			return true;
		}
		if (!visited[vertex.first->distIndex])
		{
			if (pathExists(visited, vertex.first, end))
			{
				return true;
			}
		}
	}
	return false;
}

bool testPath(const std::vector<std::unique_ptr<Vertex>>& source, const Vertex *start, const Vertex *end)
{
	bool *visited = new bool[source.size()];

	for (int i = 0; i < source.size(); ++i)
	{
		source[i]->distIndex = i;
		visited[i] = false;
	}

	bool found = pathExists(visited, start, end);

	delete visited;

	return found;
}

bool findShortestPath(std::deque<const Vertex*>& destination, const std::vector<std::unique_ptr<Vertex>>& source, const Vertex *start, const Vertex *end)
{
	std::unique_ptr<unsigned long int[]> distance = std::make_unique<unsigned long int[]>(source.size());
	std::unique_ptr<Vertex*[]> previous = std::make_unique<Vertex*[]>(source.size());
	std::uninitialized_fill_n(previous.get(), source.size(), nullptr);

	bool startExists = false;
	bool endExists = false;

	for (std::size_t i = 0; i < source.size(); ++i)
	{
		if (source[i].get() == start)
		{
			startExists = true;
			distance[i] = 0;
		}
		else
		{
			distance[i] = ULONG_MAX;
		}
		if (source[i].get() == end)
		{
			endExists = true;
		}
	}

	if (!startExists || !endExists)
	{
		if (!startExists)
			Logger::log(Logger::Pathfinding, Logger::Serious) << "start isn't in the graph" << std::endl;
		if (!endExists)
			Logger::log(Logger::Pathfinding, Logger::Serious) << "end isn't in the graph" << std::endl;
		return false;
	}

	std::unique_ptr<Vertex*[]> vertices = std::make_unique<Vertex*[]>(source.size());

	for (std::size_t i = 0; i < source.size(); ++i)
	{
		source[i]->distIndex = i;
		vertices[i] = source[i].get();
	}

	std::size_t heapSize = source.size();

	heapify(vertices.get(), distance.get(), heapSize);

	bool found = true;

	while (heapSize > 0)
	{
		Vertex *v = vertices[0];

		if (v == end)
		{
			break;
		}

		if (distance[v->distIndex] == ULONG_MAX)
		{
			//  ruh roh
			found = false;
			//Logger::log(Logger::Pathfinding, Logger::Serious) << "path not found in pathfind" << std::endl;
			break;
		}

		for (auto it = v->edges.begin(); it != v->edges.end(); ++it)
		{
			const unsigned long int alt = distance[v->distIndex] + it->second;
			if (alt < distance[it->first->distIndex])
			{
				distance[it->first->distIndex] = alt;
				previous[it->first->distIndex] = v;
				decreaseKey(vertices.get(), distance.get(), it->first->heapIndex);
			}
		}
		pop(vertices.get(), distance.get(), heapSize--);
	}

	destination.clear();

	const Vertex *crawl = end;
	while (previous[crawl->distIndex] != nullptr)
	{
		destination.push_front(crawl);
		crawl = previous[crawl->distIndex];
	}
	if (previous[end->distIndex])
	{
		destination.push_front(start);
	}

	return found;
}

class NodeHash
{
public:
	inline uint32_t operator()(const NodeID& node) const
	{
		//return node.hash();
		// moved into NodeID itself to make == and < faster too.
		return (node.getX() << 9) | node.getY() | (0xFFFFC000 & (node.getScene()->getSceneID() << 18));
	}
};

// since looking it up dynamically will cause a) some lag and b) an invalid heap for the
// priority queue implementation, we have to cache the distance for each node id
class NodeIDWithDistCached
{
public:
	NodeIDWithDistCached(const NodeID& id, float dist)
		:id(id)
		,dist(dist)
	{
	}
	bool operator<(const NodeIDWithDistCached& rhs) const
	{
		return dist > rhs.dist;
	}

	NodeID id;
	float dist;
};

static WeightFunc defaultWeightFunc{uniformWeight};


bool findShortestPathOnGrid(std::deque<NodeID>& output, const NodeID& start, const NodePredicate& isEnd, const std::map<Scene*, std::vector<std::shared_ptr<PathGrid>>>& grids, const WeightMap& weights, const Portals& portals, const WeightFunc& heuristic, ai::BehaviorContext *context)
{
	//for (const std::shared_ptr<const PathGrid>& grid : grids.at(end.getScene()))
	//{
	//	if (grid->isSolid(end.getX(), end.getY()))
	//	{
	//		Logger::log(Logger::Pathfinding, Logger::Serious) << "endpoint is solid!" << std::endl;
	//		return false;
	//	}
	//}
	const std::chrono::high_resolution_clock::time_point startTime = std::chrono::high_resolution_clock::now();

	// @todo get a decent hash function for nodes
	std::set<NodeID> visited;
	std::map<NodeID, NodeID> prev;
	std::unordered_map<NodeID, float, NodeHash> distances(1 << 16);
	std::priority_queue<NodeIDWithDistCached> queue;

	for (const auto& gridList : grids)
	{
		const int widthMinusOne = (gridList.second.front())->getWidth() - 1;
		const int heightMinusOne = (gridList.second.front())->getHeight() - 1;
		const int tileSize = (gridList.second.front())->getTileSize();
#ifdef SL_DEBUG
		for (const std::shared_ptr<const PathGrid>& grid : gridList.second)
		{
			ASSERT(widthMinusOne == grid->getWidth() - 1);
			ASSERT(heightMinusOne == grid->getHeight() - 1);
			ASSERT(tileSize == grid->getTileSize());
		}
#endif // SL_DEBUG
	}

	const float startWeight = heuristic(start, context);
	distances.insert(std::make_pair(start, startWeight));
	queue.push(NodeIDWithDistCached(start, startWeight));
	
	Scene *scene = nullptr;
	const std::vector<std::shared_ptr<PathGrid>> *currentSceneGrids = nullptr;
	const WeightFunc *weight = nullptr;
	int widthMinusOne;
	int heightMinusOne;
	int tileSize;

	auto debugOutput = [&](const char *str) {
		std::map<const Scene*, sf::Image> images;
		std::set<const Scene*> endScenes;
		for (const NodeID& v : visited)
		{
			if (images.count(v.getScene()) == 0)
			{
				const int ts = v.getScene()->getTileSize();
				images[v.getScene()].create(v.getScene()->getWidth() / ts, v.getScene()->getHeight() / ts, sf::Color::Black);
			}
			sf::Color col = sf::Color::Yellow;
			if (isEnd(v, context))
			{
				endScenes.insert(v.getScene());
				col = sf::Color::Red;
			}
			else if (v == start)
			{
				col = sf::Color::Green;
			}
			images[v.getScene()].setPixel(v.getX(), v.getY(), col);
		}
		for (const NodeID& v : output)
		{
			sf::Image& image = images.at(v.getScene());
			const int vx = v.getX();
			const int vy = v.getY();
			if (image.getPixel(vx, vy) == sf::Color::Yellow)
			{
				image.setPixel(v.getX(), v.getY(), sf::Color::Blue);
			}
		}
		for (const auto& portal : portals)
		{
			auto it = images.find(portal.first.getScene());
			if (it == images.end())
			{
				continue;
			}
			const int px = portal.first.getX();
			const int py = portal.first.getY();
			it->second.setPixel(px, py, it->second.getPixel(px, py) == sf::Color::Black ? sf::Color::Magenta : sf::Color::Cyan);
		}
		for (auto& image : images)
		{
			// TEMPORARY HERE REMOVE LATER
			const int ts = image.first->getTileSize();
			for (int y = 0; y < (int)image.second.getSize().y; ++y)
			{
				for (int x = 0; x < (int)image.second.getSize().x; ++x)
				{
					if (image.second.getPixel(x, y) == sf::Color::Black && ((Scene*)image.first)->checkCollision(Rect<int>((x+0.25f)*ts, (y+0.25f)*ts, ts/2, ts/2), "StaticGeometry") == nullptr)
					{
						image.second.setPixel(x, y, sf::Color(64, 64, 64));
					}
				}
			}
			// END TEMPORARY
			std::stringstream ss;
			ss << "pathfinddebug/" << str << "/";// << time(0) << "/";
			if (image.first == start.getScene())
			{
				ss << "[start]";
			}
			if (endScenes.count(image.first))
			{
				ss << "[end]";
			}
			ss << image.first << ".png";
			image.second.saveToFile(ss.str());
		}
	};

	int count1 = 0, count2 = 0;
	while (!queue.empty())
	{
		auto duration = std::chrono::duration_cast<std::chrono::duration<double>>(std::chrono::high_resolution_clock::now() - startTime);
		if (duration.count() >= 60.0)
		{
			Logger::log(Logger::Pathfinding, Logger::Debug) << "Calculation timed out after exploring " << distances.size() << " vertices." << std::endl;
#ifdef SL_DEBUG
			debugOutput("timeout");
#endif // SL_DEBUG
			return false;
		}
		NodeID v = queue.top().id;
		const float dist = queue.top().dist;
		queue.pop();
		++count1;
		// this is because I cbf to implement proper dijkstra with a decrease_key for path updates which would require lots of stupid shit
		// and make me construct additional things when that would actually be less efficient
		// so let's just keep track of how many redundant pops we do if we're in debug mode instead
		auto visit = visited.insert(v);
		if (!visit.second)
		{
			continue;
		}
		++count2;
		const int x = v.getX();
		const int y = v.getY();
	
		const float realDist = distances.at(v);
		//ASSERT(realDist == dist);
		
		if (isEnd(v, context))
		{
			while (v != start)
			{
				output.push_front(v);
				v = prev.at(v);
			}
			output.push_front(start);

			Logger::log(Logger::Pathfinding, Logger::Debug) << "Success after " << duration.count() << "s. pathlen = " << output.size() << " explored: " << distances.size() << "pq: " << count1 << " / " << count2 << " = " << ((float) count1 / count2) << std::endl;
#ifdef SL_DEBUG
			//debugOutput("success");
#endif // SL_DEBUG
			return true;
		}

		if (v.getScene() != scene)
		{
			scene = v.getScene();
			currentSceneGrids = &grids.at(scene);
			widthMinusOne = currentSceneGrids->front()->getWidth() - 1;
			heightMinusOne = currentSceneGrids->front()->getHeight() - 1;
			tileSize = currentSceneGrids->front()->getTileSize();
			auto it = weights.find(scene->getType());
			if (it == weights.end())
			{
				weight = &defaultWeightFunc;
				ERROR("please define a weight for " + scene->getType());
			}
			else
			{
				weight = &it->second;
			}
		}

		// check to see if we're at a portal vertex - in which case link up to the node in the other scene
		auto it = portals.find(v);
		if (it != portals.end())
		{
			const NodeID& newNode = it->second;
			const float newWeight = dist/* + (*weight)(newNode, context) + heuristic(newNode, context)*/;
			auto it = distances.insert(std::make_pair(newNode, newWeight));
			if (newWeight < it.first->second)
			{
				it.first->second = newWeight;
				prev.at(newNode) = v;
			}
			if (it.second)
			{
				prev.insert(std::make_pair(newNode, v));
				queue.push(NodeIDWithDistCached(newNode, newWeight));
			}
		}

		// spread out in 4 directions
		if (x > 0)
		{
			const NodeID left(x - 1, y, scene);
			bool canGoLeft = true;
			for (const std::shared_ptr<const PathGrid>& grid : *currentSceneGrids)
			{
				if (!grid->canGoLeft(x, y))
				{
					canGoLeft = false;
					break;
				}
			}
			if (canGoLeft)
			{
				const float leftWeight = dist + (*weight)(left, context) + heuristic(left, context);
				auto it = distances.insert(std::make_pair(left, leftWeight));
				// check if element was NOT just inserted (since a < a is always false) AND we found a shorter path
				// (map::insert doesn't update the value if it's already there and returns an iterator to it,
				// so this is the cheapest way I can think of to allow only 1 map lookup
				if (leftWeight < it.first->second)
				{
					it.first->second = leftWeight;
					prev.at(left) = v;
				}
				if (it.second)
				{
					prev.insert(std::make_pair(left, v));
					queue.push(NodeIDWithDistCached(left, leftWeight));
				}
			}
		}

		if (x < widthMinusOne)
		{
			const NodeID right(x + 1, y, scene);
			bool canGoRight = true;
			for (const std::shared_ptr<const PathGrid>& grid : *currentSceneGrids)
			{
				if (!grid->canGoRight(x, y))
				{
					canGoRight = false;
					break;
				}
			}
			if (canGoRight)
			{
				const float rightWeight = dist + (*weight)(right, context) + heuristic(right, context);
				auto it = distances.insert(std::make_pair(right, rightWeight));
				if (rightWeight < it.first->second)
				{
					it.first->second = rightWeight;
					prev.at(right) = v;
				}
				if (it.second)
				{
					prev.insert(std::make_pair(right, v));
					queue.push(NodeIDWithDistCached(right, rightWeight));
				}
			}
		}

		if (y > 0)
		{
			const NodeID up(x, y - 1, scene);
			bool canGoUp = true;
			for (const std::shared_ptr<const PathGrid>& grid : *currentSceneGrids)
			{
				if (!grid->canGoUp(x, y))
				{
					canGoUp = false;
					break;
				}
			}
			if (canGoUp)
			{
				const float upWeight = dist + (*weight)(up, context) + heuristic(up, context);
				auto it = distances.insert(std::make_pair(up, upWeight));
				if (upWeight < it.first->second)
				{
					it.first->second = upWeight;
					prev.at(up) = v;
				}
				if (it.second)
				{
					prev.insert(std::make_pair(up, v));
					queue.push(NodeIDWithDistCached(up, upWeight));
				}
			}
		}

		if (y < heightMinusOne)
		{
			const NodeID down(x, y + 1, scene);
			bool canGoDown = true;
			for (const std::shared_ptr<const PathGrid>& grid : *currentSceneGrids)
			{
				if (!grid->canGoDown(x, y))
				{
					canGoDown = false;
					break;
				}
			}
			if (canGoDown)
			{
				const float downWeight = dist + (*weight)(down, context) + heuristic(down, context);
				auto it = distances.insert(std::make_pair(down, downWeight));
				if (downWeight < it.first->second)
				{
					it.first->second = downWeight;
					prev.at(down) = v;
				}
				if (it.second)
				{
					prev.insert(std::make_pair(down, v));
					queue.push(NodeIDWithDistCached(down, downWeight));
				}
			}
		}
	}
	Logger::log(Logger::Pathfinding, Logger::Debug) << "Pathspace exhausted! explored: " << distances.size() << std::endl;
#ifdef SL_DEBUG
	debugOutput("fail");
#endif // SL_DEBUG
	return false;
}



class CarConfigHash
{
public:
	uint32_t operator()(const CarConfig& node) const
	{
		// first 22 bits are coordinates, then 4 bits of angle, and 2 bits of speed
		// last 6 bits are unused
		// (11 bits -> coordinates of 0-1023, thus 0-511 full-tiles, the town is 320 wide so this should be fine)
		return (node.getPos().x << 17)
		     | (node.getPos().y << 6)
		     | (static_cast<int>(node.getAngle()) << 2)
		     | (static_cast<int>(node.getSpeed()));
	}
};

// since looking it up dynamically will cause a) some lag and b) an invalid heap for the
// priority queue implementation, we have to cache the distance for each node id
class CarConfigWithDistCached
{
public:
	CarConfigWithDistCached(const CarConfig& id, float dist)
		:id(id)
		,dist(dist)
	{
	}
	bool operator<(const CarConfigWithDistCached& rhs) const
	{
		return dist > rhs.dist;
	}

	CarConfig id;
	float dist;
};


bool findShortestCarPath(std::deque<CarConfig>& output, const CarConfig& start, const NodePredicate& isEnd,
                         std::function<float(const CarConfig&, ai::BehaviorContext *context)> heuristic, ai::BehaviorContext *context)
{
	// @todo get a decent hash function for nodes
	std::set<CarConfig> visited;
	std::map<CarConfig, CarConfig> prev;
	std::unordered_map<CarConfig, float, CarConfigHash> distances;
	std::priority_queue<CarConfigWithDistCached> queue;

	const float startWeight = heuristic(start, context);
	distances.insert(std::make_pair(start, startWeight));
	queue.push(CarConfigWithDistCached(start, startWeight));

	Scene& town = *CarConfig::getTown();
	ASSERT((2 * town.getTileSize()) / 2 == town.getTileSize()); // must be >= 2 and power of two.
	// remember, CarConfig space is at double the resolution of the tile sizes
	const int widthMinusOne = 2 * town.getWidth() / town.getTileSize();
	const int heightMinusOne = 2 * town.getHeight() / town.getTileSize();

	int count1 = 0, count2 = 0;
	while (!queue.empty())
	{
		CarConfig v = queue.top().id;
		const float dist = queue.top().dist;
		queue.pop();
		++count1;
		// this is because I cbf to implement proper dijkstra with a decrease_key for path updates which would require lots of stupid shit
		// and make me construct additional things when that would actually be less efficient
		// so let's just keep track of how many redundant pops we do if we're in debug mode instead
		auto visit = visited.insert(v);
		if (!visit.second)
		{
			continue;
		}
		++count2;

		const float realDist = distances.at(v);
		//ASSERT(realDist == dist);

		if (isEnd(v.toNodeID(), context))
		{
			while (v != start)
			{
				output.push_front(v);
				v = prev.at(v);
			}
			output.push_front(start);
			return true;
		}

		static const Vector2i offsetsFrom[16] = {
			Vector2i(-2, 0),
			Vector2i(-2, -1),
			Vector2i(-1, -1),
			Vector2i(-1, -2),
			Vector2i(0, -2),
			Vector2i(1, -2),
			Vector2i(1, -1),
			Vector2i(2, -1),
			Vector2i(2, 0),
			Vector2i(2, 1),
			Vector2i(1, 1),
			Vector2i(1, 2),
			Vector2i(0, 2),
			Vector2i(-1, 2),
			Vector2i(-1, 1),
			Vector2i(-2, 1),
		};

		// all possible turns (TODO: 5 turns instead of 3?)
		for (int turn = -1; turn <= 1; ++turn)
		{
			// all possible speed changes
			for (float speed = -1; speed <= 2; ++speed)
			{
				const int newTurn = (static_cast<int>(v.getAngle()) + turn) % 16;
				const int newSpeed = speed + (v.getSpeed() == CarConfig::Speed::REVERSE
					? -1 : static_cast<int>(v.getSpeed()));
				const Vector2i newPos = v.getPos() + (offsetsFrom[newTurn] * speed);
				if ((turn == 0 || newSpeed <= 1) && // disable turning at high speeds
					(newSpeed >= -1 && newSpeed <= 3) && // legal driving speed
					(newPos.x > 0 && newPos.x < widthMinusOne && newPos.y > 0 && newPos.y < heightMinusOne)) // legal position
				{
					const CarConfig::Speed speedConfig = newSpeed == -1
						? CarConfig::Speed::REVERSE : static_cast<CarConfig::Speed>(newSpeed);
					CarConfig newConfig(newPos, speedConfig, static_cast<CarConfig::Angle>(newTurn));
					// check if you will hit something here. @TODO: maybe don't use AABB checks, and use collisionLine() or something else instead once that works?
					const int halfTileSize = town.getTileSize() / 2;
					const int minX = std::min(newPos.x, v.getPos().x) * halfTileSize;
					const int maxX = std::max(newPos.x, v.getPos().x) * halfTileSize;
					const int minY = std::min(newPos.y, v.getPos().y) * halfTileSize;
					const int maxY = std::max(newPos.y, v.getPos().y) * halfTileSize;
					const Rect<int> newConfigQuery(minX, minY, maxX - minX, maxY - minY);
					if (town.checkCollision(newConfigQuery, "StaticGeometry") == nullptr)
					{
						const float newWeight = dist + heuristic(newConfig, context);
						auto it = distances.insert(std::make_pair(newConfig, newWeight));
						// check if element was NOT just insert (since a < a is always false) AND we found a shorter path
						// (map::insert doesn't update the value if it's already there and returns an iterator to it,
						// so this is the cheapest way I can think of to allow only 1 map lookup
						if (newWeight < it.first->second)
						{
							it.first->second = newWeight;
							prev.at(newConfig) = v;
						}
						if (it.second)
						{
							prev.insert(std::make_pair(newConfig, v));
							// @TODO: shouldn't this get pushed in the update case too?
							queue.push(CarConfigWithDistCached(newConfig, newWeight));
						}
					}
				}
			}
		}
	}
	Logger::log(Logger::Pathfinding, Logger::Debug) << "[CarConfig] Pathspace exhausted!" << std::endl;
#ifdef SL_DEBUG
	//	debugOutput("fail");
#endif // SL_DEBUG
	return false;
}

} // pf
} // sl

#pragma optimize( "", off )  