#include "Pathfinding/PrecomputedPathGrid.hpp"

#include "Engine/Game.hpp"
#include "Engine/Scene.hpp"
#include "Utility/Assert.hpp"

namespace sl
{
namespace pf
{

static constexpr float gap = 0.4f;

PrecomputedPathGrid::PrecomputedPathGrid(Scene& scene, std::string type)
	:tileSize(scene.getTileSize())
	,grid(scene.getWidth() / tileSize, scene.getHeight() / tileSize, 0)
	,scene(scene)
	,type(std::move(type))
{
	// TODO: figure out which of these makes more sense to do. Then again this isn't really important since it only happens upon creation.
	//std::list<Entity*> entities;
	//scene.cull(entities, Rect<int>(scene.getWidth(), scene.getHeight()), type);
	//for (Entity *entity : entities)
	//{
	//	recompute(entity->getMask().getBoundingBox());
	//}
	recompute(Rect<int>(scene.getWidth(), scene.getHeight()));
}

void PrecomputedPathGrid::recompute(const Rect<int>& area)
{
	const int left   = std::max(0, static_cast<int>(std::floor((float) area.left / tileSize)) - 1);
	const int right  = std::min(getWidth() - 1, static_cast<int>(std::ceil((float) area.right() / tileSize)) + 1);
	const int top    = std::max(0, static_cast<int>(std::floor((float) area.top / tileSize)) - 1);
	const int bottom = std::min(getHeight() - 1, static_cast<int>(std::ceil((float) area.bottom() / tileSize)) + 1);
	for (int y = top; y < bottom; ++y)
	{
		for (int x = left; x < right; ++x)
		{
			if (isSolidImpl(x, y))
			{
				grid(x, y) = SOLID; // can't go left/right/etc either
			}
			else
			{
				if (canGoRightImpl(x, y))
				{
					grid(x, y) |= RIGHT;
				}
				else
				{
					grid(x, y) &= ~RIGHT; // turn off right bit
				}
				if (canGoRightImpl(x - 1, y))
				{
					grid(x, y) |= LEFT;
				}
				else
				{
					grid(x, y) &= ~LEFT; // turn off left bit
				}
				if (canGoDownImpl(x, y))
				{
					grid(x, y) |= DOWN;
				}
				else
				{
					grid(x, y) &= ~DOWN; // turn off down bit
				}
				if (canGoDownImpl(x, y - 1))
				{
					grid(x, y) |= UP;
				}
				else
				{
					grid(x, y) &= ~UP; // turn off up bit
				}
			}
		}
	}
}

bool PrecomputedPathGrid::isSolid(int x, int y) const
{
	return grid(x, y) & SOLID;
}

bool PrecomputedPathGrid::canGoRight(int x, int y) const
{
	return grid(x, y) & RIGHT;
}

bool PrecomputedPathGrid::canGoLeft(int x, int y) const
{
	return grid(x, y) & LEFT;
}

bool PrecomputedPathGrid::canGoUp(int x, int y) const
{
	return grid(x, y) & UP;
}

bool PrecomputedPathGrid::canGoDown(int x, int y) const
{
	return grid(x, y) & DOWN;
}

int PrecomputedPathGrid::getWidth() const
{
	return grid.getWidth();
}

int PrecomputedPathGrid::getHeight() const
{
	return grid.getHeight();
}

int PrecomputedPathGrid::getTileSize() const
{
	return tileSize;
}

#ifdef SL_DEBUG
void PrecomputedPathGrid::debugSaveImage(const std::string& sceneType) const
{
	std::stringstream ss;
	ss << "pathfinddebug/precompute/" << sceneType << "+" << (int)this << "+" << type << ".png";
	sf::Image img;
	img.create(getWidth() * 3, getHeight() * 3, sf::Color::Black);
	for (int y = 0; y < getHeight(); ++y)
	{
		for (int x = 0; x < getWidth(); ++x)
		{
			if (isSolid(x, y))
			{
				img.setPixel(3 * x + 1, 3 * y + 1, sf::Color(128, 128, 128));
			}
			if (!canGoDown(x, y))
			{
				for (int i = 0; i < 3; ++i)
				{
					img.setPixel(3 * x + i, 3 * y + 2, sf::Color::Red);
				}
			}
			if (!canGoUp(x, y))
			{
				for (int i = 0; i < 3; ++i)
				{
					img.setPixel(3 * x + i, 3 * y, sf::Color::Green);
				}
			}
			if (!canGoRight(x, y))
			{
				for (int i = 0; i < 3; ++i)
				{
					img.setPixel(3 * x + 2, 3 * y + i, sf::Color::Blue);
				}
			}
			if (!canGoLeft(x, y))
			{
				for (int i = 0; i < 3; ++i)
				{
					img.setPixel(3 * x, 3 * y + i, sf::Color::Yellow);
				}
			}
		}
	}
	img.saveToFile(ss.str());
}
#endif // SL_DEBUG

// private
bool PrecomputedPathGrid::isSolidImpl(int x, int y) const
{
	return scene.checkCollision(Rect<int>((x + gap) * tileSize, (y + gap) * tileSize, (1 - 2 * gap) * tileSize, (1 - 2 * gap) * tileSize), type) != nullptr;
}

bool PrecomputedPathGrid::canGoRightImpl(int x, int y) const
{
	return scene.checkCollision(Rect<int>((x + gap) * tileSize, (y + gap) * tileSize, (2 - 2 * gap) * tileSize, (1 - 2 * gap) * tileSize), type) == nullptr;
}

bool PrecomputedPathGrid::canGoDownImpl(int x, int y) const
{
	return scene.checkCollision(Rect<int>((x + gap) * tileSize, (y + gap) * tileSize, (1 - 2 * gap) * tileSize, (2 - 2 * gap) * tileSize), type) == nullptr;
}

} // pf
} // sl