#include "Pathfinding/RealTimeCachedPathGrid.hpp"

#include "Engine/Game.hpp"
#include "Engine/Scene.hpp"
#include "Utility/Assert.hpp"

namespace sl
{
namespace pf
{

const float gap = 0.3f;

RealTimeCachedPathGrid::RealTimeCachedPathGrid(Scene& scene, const std::string& type)
	:scene(scene)
	,type(type)
	,tileSize(scene.getTileSize())
	,halfTileSize(tileSize / 2)
	,quarterTileSize(halfTileSize / 2)
	,updateTime(256)
	,timestamp(scene.getGame()->getTimestamp())
	,queryRect(halfTileSize, halfTileSize)
	,cacheStatus(scene.getWidth() / tileSize, scene.getHeight() / tileSize, CacheStatus(false, timestamp - updateTime))
{
	// for quarter-tiles to work
	ASSERT(tileSize >= 4);
}

bool RealTimeCachedPathGrid::isSolid(int x, int y) const
{
	return check(2 * x, 2 * y);
}

bool RealTimeCachedPathGrid::canGoRight(int x, int y) const
{
	x *= 2;
	y *= 2;
	for (int i = 0; i < 3; ++i)
	{
		if (!check(x + i, y))
		{
			return false;
		}
	}
	return true;
}

bool RealTimeCachedPathGrid::canGoLeft(int x, int y) const
{
	x *= 2;
	y *= 2;
	for (int i = 0; i < 3; ++i)
	{
		if (!check(x - i, y))
		{
			return false;
		}
	}
	return true;
}

bool RealTimeCachedPathGrid::canGoUp(int x, int y) const
{
	x *= 2;
	y *= 2;
	for (int i = 0; i < 3; ++i)
	{
		if (!check(x, y - i))
		{
			return false;
		}
	}
	return true;
}

bool RealTimeCachedPathGrid::canGoDown(int x, int y) const
{
	x *= 2;
	y *= 2;
	for (int i = 0; i < 3; ++i)
	{
		if (!check(x, y + i))
		{
			return false;
		}
	}
	return true;
}

int RealTimeCachedPathGrid::getWidth() const
{
	return scene.getWidth() / tileSize;
}

int RealTimeCachedPathGrid::getHeight() const
{
	return scene.getHeight() / tileSize;
}

int RealTimeCachedPathGrid::getTileSize() const
{
	return tileSize;
}

void RealTimeCachedPathGrid::update(int64_t newTimestamp)
{
	timestamp = newTimestamp;
}

// private
bool RealTimeCachedPathGrid::check(int x, int y) const
{
	CacheStatus& status = cacheStatus.at(x, y);
	if (timestamp - status.timestamp > updateTime)
	{
		queryRect.left = x * halfTileSize + quarterTileSize;
		queryRect.top = y * halfTileSize + quarterTileSize;
		status.timestamp = timestamp;
		status.hit = scene.checkCollision(queryRect, type) != nullptr;
	}
	return status.hit;
}

} // pf
} // sl