#include "Pathfinding/VertexBasedPathManager.hpp"

#include "Engine/Serialization/Serialization.hpp"
#include "NPC/Locators/Target.hpp"
#include "Pathfinding/Pathfind.hpp"
#include "Utility/Assert.hpp"
#include "Utility/Logger.hpp"
#include "Utility/Memory.hpp"
#include "Utility/Random.hpp"
#include "Utility/Trig.hpp"

#include <limits>

namespace sl
{
namespace pf
{

VertexBasedPathManager::VertexBasedPathManager(int tileSize)
	:tileSize(tileSize)
#ifndef SL_NOMULTITHREADED
	,calculatingPaths(false)
#endif
{
}

std::shared_ptr<Path> VertexBasedPathManager::getPath(NodeID from, NodeID to)
{
	std::shared_ptr<Path> ret = std::make_shared<Path>(*this);

//#ifndef SL_NOMULTITHREADED
//	//std::lock_guard<std::mutex> guard(calQueueLock);
//	calQueueLock.lock();
//	calculationQueue.push_back(Args(ret, from, to));
//	calQueueLock.unlock();
//
//	if (lock.try_lock())
//	{
//		//std::cout <<  "creating new path thread[" << std::this_thread::get_id() << "]\n";
//		lock.unlock();
//		if (calculationThread.joinable())
//		{
//			calculationThread.join();
//		}
//		calculationThread = std::thread(&VertexBasedPathManager::calculatePathLoop, this, 0);
//	}
//#else
	this->getPath(ret->path, from, to);
	ret->phase = ret->path.empty() ? Path::Phase::DNE : Path::Phase::Calculated;
//#endif
	return ret;
}

bool VertexBasedPathManager::cancelCalculation(std::shared_ptr<Path>& path)
{
#ifndef SL_NOMULTITHREADED
	bool found = false;
	try
	{
		std::lock_guard<std::mutex> guard(calQueueLock);
		bool isFirst = true;
		for (auto it = calculationQueue.begin(); it != calculationQueue.end(); )
		{
			if (isFirst)
			{
				isFirst = false;
				//	maybe do something here later
			}
			else
			{
				if (it->path == path)
				{
					it = calculationQueue.erase(it);
					found = true;
					break;
				}
			}
			++it;
		}
	}
	catch (std::exception& e)
	{
		Logger::log(Logger::Pathfinding, Logger::Serious) << "exception in path threaded loop:\n" << e.what();
		calculationThread.join();
	}
	catch (...)
	{
		Logger::log(Logger::Pathfinding, Logger::Serious) << "unknown exception in path threaded loop\n";
		calculationThread.join();
	}
	//calculationThread.join();
	return found;
#endif
	return false;
}

void VertexBasedPathManager::cancelAllCalculations()
{
#ifndef SL_NO_MULTITHREADED
	calQueueLock.lock();
	(void)calculationQueue.empty();
	calQueueLock.unlock();
	if (calculationThread.joinable())
	{
		calculationThread.join();
	}
#endif // SL_NOMULTITHREADED
}

#ifndef SL_NOMULTITHREADED
void VertexBasedPathManager::calculatePathLoop(int id)
{
	try
	{
	//std::cout << "Starting: calculatePathLoop()[" << std::this_thread::get_id() << "]\n";
	std::lock_guard<std::mutex> guard(lock);
	for ( ; ; )
	{
		calQueueLock.lock();
		if (calculationQueue.empty())
		{
			calQueueLock.unlock();
			break;
		}
		const Args top = calculationQueue.front();
		calculationQueue.pop_front();
		calQueueLock.unlock();

		top.path->phase = Path::Phase::Calculating;
		getPath(top.path->path, top.to, top.from);
		top.path->phase = top.path->path.empty() ? Path::Phase::DNE : Path::Phase::Calculated;
		//std::cout << "-[" << std::this_thread::get_id() << "] ";
	}

	//std::cout << "Finishing: calculatePathLoop()[" << std::this_thread::get_id() << "]\n";
	}
	catch (std::exception& e)
	{
		Logger::log(Logger::Pathfinding, Logger::Serious) << "FUCKING THREAD CRASHED MOTHERFUCKER: " << e.what() << std::endl;
		Logger::log(Logger::Pathfinding, Logger::Serious) << "STOP MOTHERFUCKER\n";
		calQueueLock.unlock();
	}
}
#endif // SL_NOMULTITHREADED

bool VertexBasedPathManager::getPath(std::deque<Node>& results, NodeID from, NodeID to)
{
	ASSERT(from.getScene() == to.getScene());
	results.clear();
	if (vertices.count(from) == 0 || vertices.count(to) == 0)
	{
		return false;
	}
	bool pathFound = false;
	std::pair<NodeID, NodeID> key = std::make_pair(from, to);
	auto it = computedPaths.find(key);
	if (it == computedPaths.end())
	{
		std::deque<const Vertex*> path;
		pathFound = findShortestPath(path, vertexList, vertices.find(from)->second, vertices.find(to)->second);
		if (path.empty())
		{
			//	I guess with directed edges you can have a weakly connected graph that isn't strongly connected
			//ASSERT(!pathExists(from, to));
			return false;
		}
		const Vertex *prev = path.front();
		for (const Vertex *vertex : path)
		{
			if (vertex != prev)	//not the first one
			{
				auto it = conditionalEdges.find(prev);
				if (it != conditionalEdges.end())	//	from vertex found
				{
					auto checker = it->second.find(vertex);
					if (checker != it->second.end())	//	to found (ie edge found)
					{
						results.push_back(Node(vertex->x, vertex->y, from.getScene(), checker->second));
						prev = vertex;
						continue;
					}
				}
			}
			results.push_back(Node(vertex->x, vertex->y, from.getScene()));
			prev = vertex;
		}
		computedPaths.insert(std::make_pair(key, results));
	}
	else
	{
		results = it->second;
		pathFound = true;
	}
	return pathFound;
}

NodeID VertexBasedPathManager::getClosestNodeID(const Target& target) const
{
	ASSERT(!vertices.empty());
	Vertex* closest = nullptr;
	NodeID ret(-1, -1, target.getScene());
	float closestDist = std::numeric_limits<float>::max(); // @todo max dist
	for (const auto& it : vertices)
	{
		const float dist = pointDistance(it.second->x * tileSize, it.second->y * tileSize, target.getX(), target.getY());
		if (closestDist > dist)
		{
			closest = it.second;
			closestDist = dist;
			ret = it.first;
		}
	}
	return ret;
}

/*bool PathManager::pathExists(NodeID from, NodeID to) const
{
	return testPath(vertexList, vertices.find(from)->second, vertices.find(to)->second);
}*/

void VertexBasedPathManager::addDirectedEdge(NodeID from, NodeID to, float weight)
{
	if (vertices.count(from) == 0)
	{
		auto vertex = unique<Vertex>(from.getX(), from.getY());
		vertices.insert(std::make_pair(from, vertex.get()));
		vertexList.push_back(std::move(vertex));
	}
	if (vertices.count(to) == 0)
	{
		auto vertex = unique<Vertex>(to.getX(), to.getY());
		vertices.insert(std::make_pair(to, vertex.get()));
		vertexList.push_back(std::move(vertex));
	}
	vertices[from]->addVertex(vertices[to]);
}

void VertexBasedPathManager::addDirectedConditionalEdge(NodeID from, NodeID to, const NodeAvailabilityChecker *checker, float weight)
{
	ASSERT(checker != nullptr);
	addDirectedEdge(from, to, weight);
	conditionalEdges[vertices[from]][vertices[to]] = checker;
	condedges.push_back(std::make_pair(sf::VertexArray(sf::Lines, 2), checker));
	condedges.back().first[0].color = sf::Color::Magenta;
	condedges.back().first[0].color = sf::Color::Cyan;
	condedges.back().first[0].position = sf::Vector2f(from.getX() * tileSize, from.getY() * tileSize);
	condedges.back().first[1].position = sf::Vector2f(to.getX() * tileSize, to.getY() * tileSize);
}

NodeID VertexBasedPathManager::getRandomID() const
{
	ASSERT(vertices.size() != 0);
	int i = 0;
	int n = Random::get().random(vertices.size());
	for (std::map<NodeID, Vertex*>::const_iterator it = vertices.begin(); it != vertices.end(); ++it)
	{
		if (i++ == n)
		{
			return it->first;
		}
	}
	ERROR("dude wat");
	return NodeID(-1, -1, nullptr);//to stop warnings
}

Node VertexBasedPathManager::getNodeFromID(NodeID id) const
{
	return Node(id.getX(), id.getY(), id.getScene());
}

void VertexBasedPathManager::draw(RenderTarget& target) const
{
	for (const std::unique_ptr<Vertex>& vertex : vertexList)
	{
		vertex->draw(target);
	}
	for (auto& pair : condedges)
	{
		const bool isAvailable = pair.second && pair.second->getAvailability();
		pair.first[0].color = isAvailable ? sf::Color::Magenta : sf::Color::Black;
		pair.first[1].color = isAvailable ? sf::Color::Cyan    : sf::Color::Black;
		//if (pair.second->getAvailability())
		{
			target.draw(pair.first);
		}
	}
}

void VertexBasedPathManager::serialize(Serialize& out) const
{
	out.startObject("VertexBasedPathManager");
	std::map<const Vertex*, int> indices;
	for (int i = 0; i < vertexList.size(); ++i)
	{
		indices[vertexList[i].get()] = i;
	}
	out.i("n", vertexList.size());
	for (const std::unique_ptr<Vertex>& vertex : vertexList)
	{
		out.i("x", vertex->x);
		out.i("y", vertex->y);
		out.i("deg", vertex->edges.size());
		for (const auto& edge : vertex->edges)
		{
			out.i("neighbor", indices.at(edge.first));
			out.i("weight", edge.second);
		}
	}
	// TODO serialize conditional edges
	out.endObject("VertexBasedPathManager");
}

void VertexBasedPathManager::deserialize(Deserialize& in)
{
	in.startObject("VertexBasedPathManager");
	const int n = in.i("n");
	for (int i = 0; i < n; ++i)
	{
		vertexList.push_back(unique<Vertex>(-1, -1));
	}
	for (int i = 0; i < n; ++i)
	{
		const int x = in.i("x");
		const int y = in.i("y");
		vertexList[i]->x = x;
		vertexList[i]->y = y;
		vertices.insert(std::make_pair(NodeID(x, y, nullptr), vertexList[i].get()));
		const int deg = in.i("deg");
		for (int i = 0; i < deg; ++i)
		{
			const int neighbor = in.i("neighbor");
			const int weight = in.i("weight");
			vertexList[i]->addVertex(vertexList[neighbor].get(), weight);
		}
	}
	in.endObject("VertexBasedPathManager");
}

} // pf
} // sl
