#include "Player/Actions/ActionChooser.hpp"

#include "Engine/GUI/DrawableContainer.hpp"
#include "Engine/GUI/NormalBackground.hpp"
#include "Engine/GUI/WidgetList.hpp"
#include "Engine/AbstractScene.hpp"
#include "Engine/Game.hpp"
#include "Items/ItemToDrawables.hpp"
#include "Player/Actions/PlayerActionFactory.hpp"
#include "Player/Player.hpp"
#include "Player/PlayerStats.hpp"
#include "Utility/Ranges.hpp"

namespace sl
{
namespace gui
{

enum class TopLevel
{
	Base = 0,
	Items
};

static const NormalBackground normalBG(sf::Color::Black, sf::Color::White);

ActionChooser::ActionChooser(Player& player, CustomAction& actionToPerform)
	:GUIWidget(Rect<int>(512, 384))
	,root(*this, nullptr)
	,actionToBind(NO_CUSTOM_ACTION)
	,actionToPerform(actionToPerform)
	,player(player)
{
	actionToPerform = NO_CUSTOM_ACTION;
	auto makeTerminal = [this](int depth) {
		return [this,depth](CustomAction id) {
			this->actionToPerform = id;
			close();
		};
	};

	auto actionName = [](CustomAction id) {
		return playerActionFromId(id)->getShortName();
	};
	// TODO: (maybe in ScrollArea???/WidgetList?) make it so it will make room for the scrollbar if it's there.
	auto topLevelList = unique<WidgetList<TopLevel>>(Rect<int>(96-8, 384),
	[](const TopLevel& level) {
		return level == TopLevel::Base ? "Abilities" : "Items";
	}, [this,actionName, makeTerminal](const TopLevel& topLevel) {
		actionToBind = NO_CUSTOM_ACTION; // TODO: centralize?
		switch (topLevel)
		{
		case TopLevel::Base:
		{
			auto abilitiesList = unique<WidgetList<CustomAction>>(Rect<int>(96-16, 0, 160, 384), actionName, makeTerminal(2));
			abilitiesList->setOnHover([this](CustomAction actionId) {
				actionToBind = actionId;
			});
			abilitiesList->setPreview([](CustomAction id) {
				auto container = unique<gui::DrawableContainer>(Rect<int>(20, 32), playerActionFromId(id)->getIcon());
				container->applyBackground(gui::NormalBackground(sf::Color::Black, sf::Color::White));
				return container;
			});
			abilitiesList->registerAcceptButton(Key::GUIRight);
			abilitiesList->updateItems(getUniqueActions()).trim();
			layer(1) = unique<Layer>(*this, &root, std::move(abilitiesList));
		}
			break;
		case TopLevel::Items:
		{
			auto itemList = unique<WidgetList<Item::Type>>(Rect<int>(96-16, 0, 160, 384),
			[](Item::Type typeId) {
				return Item(typeId).getName();
			}, [this,actionName,makeTerminal](Item::Type typeId) {
				actionToBind = NO_CUSTOM_ACTION; // TODO: centralize?
				auto itemActionsList = unique<WidgetList<CustomAction>>(Rect<int>(96-16 + 160 -16, 0, 160, 384), actionName, makeTerminal(3));
				itemActionsList->setOnHover([this](CustomAction actionId) {
					actionToBind = actionId;
				});
				itemActionsList->registerAcceptButton(Key::GUIRight);
				itemActionsList->updateItems(getItemActions(typeId)).trim();
				layer(2) = unique<Layer>(*this, layer(1).get(), std::move(itemActionsList));
			});
			itemList->setPreview(ItemToBoxedGUI(Rect<int>(20, 32)));
			itemList->registerAcceptButton(Key::GUIRight);
			itemList->updateItems(this->player.getStats().soul->items.getTypesContained()).trim();
			layer(1) = unique<Layer>(*this, &root, std::move(itemList));
		}
			break;
		}
	});
	topLevelList->registerAcceptButton(Key::GUIRight);
	topLevelList->updateItems(enumRangeInclusive(TopLevel::Base, TopLevel::Items)).trim();
	root.setList(std::move(topLevelList));

	registerInputEvent(Key::GUIBack, InputEventType::Released, [this] {
		close();
	});
	registerInputEvent(Key::ActionChooser, InputEventType::Released, [this] {
		close();
	});
	//setFocus(true);
}



// private
void ActionChooser::onUpdate(int x, int y)
{
	if (actionToBind != NO_CUSTOM_ACTION)
	{
		const std::unique_ptr<KeyBind> justPressed = scene->getPlayerController().getJustPressedUnboundKey();
		if (justPressed)
		{
			scene->getGame()->getMutablePlayerControls().registerCustomAction(actionToBind, *justPressed);
			close();
		}
	}
}

void ActionChooser::close()
{
	this->getScene()->destroy();
}


std::unique_ptr<ActionChooser::Layer>& ActionChooser::layer(int depth)
{
	ASSERT(depth > 0);
	std::unique_ptr<Layer> *cur = &root.child;
	for (int i = 1; i < depth; ++i)
	{
		cur = &(*cur)->child;
		ASSERT(&*cur);
	}
	return *cur;
}



// Layer
ActionChooser::Layer::Layer(GUIWidget& base, Layer *parent)
	:parent(parent)
	,child(nullptr)
	,base(base)
	,list(nullptr)
{
}

ActionChooser::Layer::Layer(GUIWidget& base, Layer *parent, std::unique_ptr<GUIWidget> w)
	:ActionChooser::Layer(base, parent)
{
	setList(std::move(w));
}

ActionChooser::Layer::~Layer()
{
	ASSERT(list);
	if (parent)
	{
		parent->list->removeWidget(*list);
	}
}


void ActionChooser::Layer::setList(std::unique_ptr<GUIWidget> w)
{
	list = w.get();
	if (parent)
	{
		parent->list->addWidget(std::move(w));
		list->registerInputEvent(Key::GUIBack, InputEventType::Released, [this] {
			parent->child.reset();
		});
		list->registerInputEvent(Key::GUILeft, InputEventType::Released, [this] {
			parent->child.reset();
		});
	}
	else
	{
		base.addWidget(std::move(w));
	}
	//list->setFocus(true);
}

} // gui
} // sl