#include "Player/Actions/PlayerActionFactory.hpp"

#include "Player/Actions/PlayerConsumeItemAction.hpp"
#include "Player/Actions/PlayerDropItemAction.hpp"
#include "Player/Actions/PlayerStalkItemAction.hpp"
#include "Player/Actions/PlayerSubdueItemAction.hpp"
#include "Player/Actions/PlayerThrowItemAction.hpp"
#include "Player/Actions/PlayerGrabAction.hpp"
#include "Player/Actions/PlayerPunchAction.hpp"
#include "Player/Actions/PlayerStripAction.hpp"
#include "Items/Item.hpp"
#include "Utility/Memory.hpp"
#include "Utility/Ranges.hpp"

namespace sl
{

enum class UniqueAction : CustomAction
{
	Grab,
	Punch,
	Strip
};
static constexpr CustomAction UNIQUE_ACTIONCOUNT = 3;
enum class ItemUseType : CustomAction
{
	Consume = 0,
	Stalk,
	Subdue,
	Throw,
	Drop
};
static constexpr int itemUseTypes = 5;

std::vector<CustomAction> getUniqueActions()
{
	return rangeInclusive<CustomAction>(1, UNIQUE_ACTIONCOUNT);
}

std::vector<CustomAction> getItemActions(int itemId)
{
	Item item(itemId);
	std::vector<CustomAction> actionIds;
	const CustomAction itemBaseId = 1 + UNIQUE_ACTIONCOUNT + itemUseTypes * item.getType();
	if (item.hasComponent(Item::Category::Consumable))
	{
		actionIds.push_back(itemBaseId + static_cast<CustomAction>(ItemUseType::Consume));
	}
	if (item.getStalkStats())
	{
		actionIds.push_back(itemBaseId + static_cast<CustomAction>(ItemUseType::Stalk));
	}
	if (item.getSubdueStats())
	{
		actionIds.push_back(itemBaseId + static_cast<CustomAction>(ItemUseType::Subdue));
	}
	actionIds.push_back(itemBaseId + static_cast<CustomAction>(ItemUseType::Throw));
	actionIds.push_back(itemBaseId + static_cast<CustomAction>(ItemUseType::Drop));
	return actionIds;
}

std::unique_ptr<PlayerAction> playerActionFromId(CustomAction id)
{
	std::unique_ptr<PlayerAction> action;
	if (id != NO_CUSTOM_ACTION)
	{
		--id;
		if (id < UNIQUE_ACTIONCOUNT)
		{
			switch (static_cast<UniqueAction>(id))
			{
			case UniqueAction::Grab:
				action = unique<PlayerGrabAction>();
				break;
			case UniqueAction::Punch:
				action = unique<PlayerPunchAction>();
				break;
			case UniqueAction::Strip:
				action = unique<PlayerStripAction>();
				break;
			}
		}
		else
		{
			// assume it's an item
			id -= UNIQUE_ACTIONCOUNT;
			const Item::Type itemId = id / itemUseTypes;
			const Item item(itemId);
			const ItemUseType useType = static_cast<ItemUseType>(id % itemUseTypes);
			switch (useType)
			{
			case ItemUseType::Consume:
				if (item.hasComponent(Item::Category::Consumable))
				{
					action = unique<PlayerConsumeItemAction>(itemId);
				}
				break;
			case ItemUseType::Stalk:
				if (item.getStalkStats())
				{
					action = unique<PlayerStalkItemAction>(itemId);
				}
				break;
			case ItemUseType::Subdue:
				if (item.getSubdueStats())
				{
					action = unique<PlayerSubdueItemAction>(itemId);
				}
				break;
			case ItemUseType::Throw:
				action = unique<PlayerThrowItemAction>(itemId);
				break;
			case ItemUseType::Drop:
				action = unique<PlayerDropItemAction>(itemId);
				break;
			}
		}
	}
	return action;
}

} // sl