#include "Player/Player.hpp"

#include "Engine/Door.hpp"
#include "Engine/Game.hpp"
#include "Engine/Input/Input.hpp"
#include "Town/World.hpp"
#include "Engine/Graphics/SFDrawable.hpp"
#include "Engine/Graphics/SpriteAnimation.hpp"
#include "Engine/Graphics/Sprite.hpp"
#include "Engine/GUI/NormalBackground.hpp"
#include "Menus/InGameMenuController.hpp"
#include "Menus/GameOver.hpp"
#include "NPC/NPC.hpp"
#include "NPC/Blood.hpp"
#include "NPC/States/IdleState.hpp"
#include "NPC/LayeredSpriteLoader.hpp"
#include "NPC/Stats/NationalityDatabase.hpp"
#include "NPC/Stats/SoulGenerationSettings.hpp"
#include "Engine/Scene.hpp"
#include "Player/States/PlayerWalkState.hpp"
#include "Utility/Random.hpp"
#include "Utility/TypeToString.hpp"
#include "Utility/Memory.hpp"

#include <initializer_list>

#include <SFML/Graphics/CircleShape.hpp>

namespace sl
{

DECLARE_TYPE_TO_STRING(Player)

Player::Player(int x, int y, std::shared_ptr<PlayerStats> stats)
	:Entity(x, y, 8, 8, "Player")
	,stats(stats)
	,state()
	,minimapIcon(pos, sf::Color(0, 255, 64))
	,menuController(nullptr)
	,gameOverTimer(-1)
	,currentlySelectedCallout(callouts.end())
	,currentlySelectedCalloutIndex(0)
	,staminaCooldown(0)
	,calloutTargetCircle(nullptr)
{
	auto circle = std::make_unique<SFDrawable<sf::CircleShape>>();
	circle->getData().setFillColor(sf::Color(255, 192, 64, 48));
	circle->getData().setOutlineColor(sf::Color(255, 164, 32, 164));
	circle->getData().setOutlineThickness(1.f);
	calloutTargetCircle = &circle->getData();
	addDrawable(std::move(circle));

	mask.setOrigin(4, 4);

	minimapIcon.setDetailedIcon(map::LOD::Map, std::make_unique<Sprite>("map/player_icon.png", 8, 8));
}





const Types::Type Player::getType() const
{
	return Entity::makeType("Player");
}

void Player::handleEvent(const Event& event)
{
	switch (event.getCategory())
	{
	case Event::Category::Debug:
		{
			if (event.getName() == "spawn_loli")
			{
				const std::string nationality = event.getFlag("nationality").asString();
				Random& rng = Random::get();
				if (NationalityDatabase::get().exists(nationality) || nationality == "random")
				{
					SoulGenerationSettings genSettings;
					if (nationality != "random")
					{
						genSettings.addNationality(nationality);
					}
					auto spawnedLoliSoul = scene->getWorld()->createSoul(rng, genSettings);
					NPC *obj = new NPC(pos.x, pos.y, spawnedLoliSoul);
					obj->initStates(unique<IdleState>(*obj));
					scene->addEntityToQuad(obj);
				}
			}
			else if (event.getName() == "player's toggle debug btree")
			{
				const Vector2f mPos(scene->getMousePos().x, scene->getMousePos().y);
				scene->registerLocalEvent(unique<Event>(Event::Category::Debug, "toggle debug btree", mPos));
			}
		}
		break;
	default:
		break;
	}

	state.handleEvent(event);
}

PlayerStats& Player::getStats()
{
	ASSERT(stats);
	return *stats;
}

void Player::addActionCallout(CalloutType type, const std::string& text, std::function<void()> result, Entity *target)
{
	callouts[std::make_pair(type, text)].push_back({ std::move(result), target });
}

void Player::serialize(Serialize& out) const
{
	serializeEntity(out);
	stats->serialize(out);
}

void Player::deserialize(Deserialize& in)
{
	deserializeEntity(in);
	stats = unique<PlayerStats>(nullptr);
	stats->deserialize(in);
	state.initStates(unique<PlayerWalkState>(*this));
}

std::string Player::serializationType() const
{
	return "Player";
}

bool Player::shouldSerialize() const
{
	return true;
}


/*		private			*/
void Player::onUpdate()
{
	//if (controller.pressed("test_randomcrime"))
	//{
	//	Random& rng = Random::get();
	//	const unsigned long int victim = rng.choose(NPC::registry).first;
	//	const int crimes = rng.choose({1, 1, 1, 2, 2, 3});
	//	for (int j = 0; j < crimes; ++j)
	//	{
	//		const int times = rng.random(4);
	//		const Crime::Type crimeType = rng.choose({
	//			Crime::Type::SexualAssault,
	//			Crime::Type::Murder,
	//			Crime::Type::Kidnapping,
	//			Crime::Type::Rape,
	//			Crime::Type::StatRape,
	//			Crime::Type::AttemptedRape,
	//			Crime::Type::AttemptedKidnapping
	//		});
	//		for (int i = 0; i < times; ++i)
	//		{
	//			stats->recordCrime(victim, crimeType);
	//		}
	//	}
	//}

	calloutGUI.disable();

	state.update();

	// to preserve currentlySelectedCallout iterator validity, don't clear entire map.
	if (currentlySelectedCallout == callouts.end())
	{
		callouts.clear();
	}
	else
	{
		for (auto& callout : callouts)
		{
			callout.second.clear();
		}
	}

	if (gameOverTimer > 0)
	{
		if (--gameOverTimer == 0)
		{
			Game *game = scene->getGame();

			World *world = scene->getWorld();
			game->addScene(unique<GameOver>(game, world, scene->getPlayer()->getStats()));
			world->stop();
		}
	}
	else if (stats->soul->getStatus().hp <= 0)
	{
		int n = Random::get().randomInclusive(24, 48);
		for (int i = 0; i < n; ++i)
		{
			Vector2<float> veloc(lengthdir(Random::get().randomInclusive(3, 12), Random::get().random(360)));
			getScene()->addEntityToList(new Blood(pos.x, pos.y, veloc.x, veloc.y));
		}
		gameOverTimer = 128;
		state.initStates(nullptr);
		clearDrawablesInSet("bleedout");
		addDrawableToSet("bleedout", unique<SpriteAnimation>("guy/bleedout.png", 24, 23, false, ColorSwapID(), 12, 21));
		setDrawableSet("bleedout");
	}

	

	//Door *door = (Door*) scene->checkCollision(*this, "Door");
	//if (door)
	//{
	//	door->transfer(this);
	//	door->enter();
	//}
	calculateHeight();

#ifdef SL_DEBUG
	static constexpr int hm_dist = 64;
	static const Vector2i offsets[4] = {
		Vector2i(0, 1),
		Vector2i(1, 0),
		Vector2i(1, 1),
		Vector2i(1, -1)
	};
	std::vector<Vector2f> points;
	points.reserve(128);
	for (int i = 0; i < 4; ++i)
	{
		points.clear();
		//Rect<int> bounds(getMask().getBoundingBox());
		//bounds.left -= 64;
		for (int j = -hm_dist; j < hm_dist; ++j)
		{
			const Vector2i p = pos + offsets[i] * j;
			points.push_back(Vector2f(p.x, p.y - scene->getHeightAt(p)));
			//++bounds.left;
		}
		scene->debugDrawLines("draw_heightmaps", points, sf::Color::Green);
	}
#endif
}

void Player::onEnter()
{
	scene->setPlayer(this);
	scene->getWorld()->setPlayer(this);

	minimapIcon.setScene(scene);

	state.onSceneEnter();

	ASSERT(menuController == nullptr);
	menuController = new InGameMenuController(*stats);
	scene->addEntityToList(menuController);

	calloutGUI.setScene(scene);
	calloutGUI.setHandle(unique<gui::TextLabel>(Rect<int>(512 - 64, 24), "UNINITIALIZED_STRING", gui::TextLabel::Config().centered()));
	calloutGUI->applyBackground(gui::NormalBackground(sf::Color::Black, sf::Color::White));
}

void Player::onExit()
{
	calloutGUI.disable();
	calloutGUI.setScene(nullptr);

	ASSERT(menuController != nullptr);
	menuController->destroy();
	menuController = nullptr;

	state.onSceneExit();

	scene->getWorld()->setPlayer(nullptr);
	scene->setPlayer(nullptr);
}

void Player::loadWalkAnims()
{
	if (!scene || scene->getGame()->getSettings().get("loli_mode") == 0)
	{
		loadPlayerWalkAnims();
	}
	else
	{
		loadLoliWalkAnims();
	}
}

void Player::loadPlayerWalkAnims()
{
	LayeredSpriteLoader::load(*this, LayeredSpriteLoader::Config(getStats().soul->appearance, "walk", 32, Vector2i(18, 28)).animate(5));
	setDrawableSet("south");
}

void Player::loadLoliWalkAnims()
{
	Soul& soul = *getStats().soul;
	soul.setType("loli");
	LayeredSpriteLoader::load(*this, LayeredSpriteLoader::Config(getStats().soul->appearance, "walk", 32, Vector2i(16, 28)).animate(5));
	soul.setType("player");
}

void Player::handleButtonCallouts()
{
	// hacky way to just make it not display
	calloutTargetCircle->setOrigin(-1337666, -1337666);
	if (callouts.empty())
	{
		calloutGUI.disable();
	}
	else
	{
		calloutGUI.enable();
		ASSERT(currentlySelectedCalloutIndex >= 0);

		auto callout = callouts.begin();

		if (currentlySelectedCallout != callouts.end())
		{
			if (currentlySelectedCallout->second.empty())
			{
				currentlySelectedCallout = callouts.end();
			}
			else
			{
				callout = currentlySelectedCallout;
			}
		}

		if (currentlySelectedCalloutIndex >= callout->second.size())
		{
			currentlySelectedCalloutIndex = 0;
		}

		if (callout->second.empty())
		{
			calloutGUI.disable();
		}
		else
		{
			std::stringstream ss;
			ss << "Press " << scene->getGame()->getPlayerControls().getBindsString(Key::Interact) << " to " << callout->first.second;
			if (callouts.size() > 1 || callouts.begin()->second.size() > 1)
			{
				if (scene->getPlayerController().pressed(Key::CycleCallout))
				{
					// if we hadn't tabbed to anything before, just start at the highest priority action (start)
					if (++currentlySelectedCalloutIndex >= callout->second.size())
					{
						currentlySelectedCalloutIndex = 0;
						if (++callout == callouts.end())
						{
							callout = callouts.begin();
						}
					}
					currentlySelectedCallout = callout;
				}

				if (callout->second.size() > 1)
				{
					ss << " (" << (currentlySelectedCalloutIndex + 1) << "/" << callout->second.size() << ")";
				}
				ss << " - Press " << scene->getGame()->getPlayerControls().getBindsString(Key::CycleCallout) << " to toggle to next action";
			}
			calloutGUI->setString(ss.str());

			if (scene->getPlayerController().pressed(Key::Interact))
			{
				callout->second[currentlySelectedCalloutIndex].callback();
			}

			const Entity *target = callout->second[currentlySelectedCalloutIndex].target;
			if (target)
			{
				const auto& maskOrigin = target->getMask().getOrigin();
				const auto& targetRect = target->getMask().getBoundingBox();
				const float targetRad = std::max(targetRect.width, targetRect.height) / 2.f;
				calloutTargetCircle->setRadius(targetRad);
				calloutTargetCircle->setOrigin(
					(getPos().x - target->getPos().x) + (maskOrigin.x),
					(getPos().y - target->getPos().y) + (maskOrigin.y));
			}
		}
	}
	calloutGUI->update((scene->getViewWidth() - calloutGUI->getBox().width) / 2, scene->getViewHeight() - calloutGUI->getBox().height - 48);
}

}// End of sl namespace
