/**
 * @section DESCRIPTION
 * This class represents only the physical representation of the player within a Scene. See PlayerStats for the player's actual data.
 */
#ifndef SL_PLAYER_HPP
#define SL_PLAYER_HPP

#include <memory>

#include "Engine/Input/Controller.hpp"
#include "Engine/GUI/TextLabel.hpp"
#include "Engine/GUI/WidgetHandle.hpp"
#include "Engine/Entity.hpp"
#include "Engine/Scene.hpp"
#include "Engine/StateManager.hpp"
#include "NPC/FacingDirection.hpp"
#include "Player/PlayerStats.hpp"
#include "Player/PlayerState.hpp"
#include "Town/Map/Icon.hpp"

namespace sf
{
class CircleShape;
} // sf

namespace sl
{

class Car;
class InGameMenuController;

class Player : public Entity
{
public:
	/**
	 * Constructs a Player into a level
	 * @param x The x position of the player
	 * @param y The y position of the player
	 * @param stats Player data
	 */
	Player(int x, int y, std::shared_ptr<PlayerStats> stats);


	/**
	 * Returns the Entity Type
	 * @return Returns the Player Entity Type
	 */
	const Types::Type getType() const override;

	void handleEvent(const Event& event) override;

	StateManager<PlayerState>& getState() { return state; }

	PlayerStats& getStats();

	//@todo refactor
	Car *car;

	// This is here so we can add in a prioritization of callouts
	enum class CalloutType
	{
		PickUp,
		Open,
		Fill,
		Enter,
		Use,
		Drop,
		Talk,
	};
	void addActionCallout(CalloutType type, const std::string& text, std::function<void()> result, Entity *target = nullptr);

	void serialize(Serialize& out) const override;

	void deserialize(Deserialize& in) override;

	std::string serializationType() const override;

	bool shouldSerialize() const override;

	FacingDirection facing() const { return facingDir; }



private:
	/**
	 * Updates the Player. Gets called automatically by Entity::update()
	 */
	void onUpdate() override;

	void onEnter() override;

	void onExit() override;

	void loadWalkAnims();

	void loadPlayerWalkAnims();

	void loadLoliWalkAnims();

	void handleButtonCallouts();



	//!The Player's data
	std::shared_ptr<PlayerStats> stats;

	StateManager<PlayerState> state;

	map::Icon minimapIcon;

	InGameMenuController *menuController;

	// @todo refactor somewhere else?
	int gameOverTimer;

	typedef std::pair<CalloutType, std::string> CalloutCategory;
	struct CalloutAction
	{
		std::function<void()> callback;
		Entity *target;
	};
	//! This automatically prioritizes based on CalloutType (and then in lex order)! :D
	std::map<CalloutCategory, std::vector<CalloutAction>> callouts;
	//! Do not remove keys from callouts as long as this iterator is non-end
	decltype(callouts)::iterator currentlySelectedCallout;
	int currentlySelectedCalloutIndex;
	gui::WidgetHandle<gui::TextLabel> calloutGUI;
	sf::CircleShape *calloutTargetCircle;

	int staminaCooldown;

	FacingDirection facingDir;


	friend class PlayerState;
};

}// End of sl namespace

#endif
