#include "Player/PlayerState.hpp"

#include "Engine/Door.hpp"
#include "Engine/Game.hpp"
#include "NPC/Stealth/Noise.hpp"
#include "Player/Player.hpp"
#include "Town/Container.hpp"

namespace sl
{

PlayerState::PlayerState(Player& player)
	:player(player)
{
}

Vector2<float>& PlayerState::pos()
{
	return player.pos;
}

PlayerStats& PlayerState::stats()
{
	return *player.stats;
}

const Controller& PlayerState::controls() const
{
	return player.getScene()->getPlayerController();
}

void PlayerState::changeState(std::unique_ptr<PlayerState> newState)
{
	player.state.changeState(std::move(newState));
	player.calloutGUI.disable();
}

PlayerState::AttemptWalkResult PlayerState::attemptWalk(float scalar)
{
	const int boundary = 32;

	Scene& scene = *player.getScene();

	const float walkSpeed = scene.getGame()->getSettings().get("player_walk_speed") / 10.f;
	float speed = walkSpeed;
	float animSpeed = 0.75f;
	float& stamina = player.getStats().soul->getStatus().stamina;
	if (controls().held(Key::Sprint))
	{
		if (stamina > 0.f || scene.getGame()->getSettings().get("unlimited_sprint"))
		{
			speed = scene.getGame()->getSettings().get("player_sprint_speed") / 10.f;
			animSpeed = 0.5f * speed / walkSpeed;
		}
	}
	else if (controls().held(Key::Sneak))
	{
		speed = scene.getGame()->getSettings().get("player_sneak_speed") / 10.f;
		animSpeed = 0.15f * speed / walkSpeed;
	}

	const float startSpeed = speed;
	bool moved = false;
	bool attempted = false;
	const bool noclip = scene.getInputHandle().isKeyHeld(Input::KeyboardKey::N);
	Door *door = nullptr;
	while (!moved && speed > 0.f)
	{
		Rect<int> boundsCheck = player.getMask().getBoundingBox();
		if (controls().held(Key::MoveUp))
		{
			// add an extra one otherwise when speed != floor(speed) the player ends up 1 pixel into walls sometimes
			boundsCheck.top -= speed + 1;
			if (pos().y > boundary && ((door = static_cast<Door*>(player.getScene()->checkCollision(boundsCheck, "Door"))) || noclip || scene.checkCollision(boundsCheck, "StaticGeometry") == nullptr))
			{
				player.pos.y -= speed;
				moved = true;
			}
			face(FacingDirection::North);
			attempted = true;
		}
		if (!moved && controls().held(Key::MoveDown))
		{
			boundsCheck.top += speed + 1;
			if (pos().y < scene.getHeight() - boundary && ((door = static_cast<Door*>(player.getScene()->checkCollision(boundsCheck, "Door"))) || noclip || scene.checkCollision(boundsCheck, "StaticGeometry") == nullptr))
			{
				player.pos.y += speed;
				moved = true;
			}
			face(FacingDirection::South);
			attempted = true;
		}
		if (!moved && controls().held(Key::MoveRight))
		{
			boundsCheck.left += speed + 1;
			if (pos().x < scene.getWidth() - boundary && ((door = static_cast<Door*>(player.getScene()->checkCollision(boundsCheck, "Door"))) || noclip || scene.checkCollision(boundsCheck, "StaticGeometry") == nullptr))
			{
				player.pos.x += speed;
				moved = true;
			}
			face(FacingDirection::East);
			attempted = true;
		}
		if (!moved && controls().held(Key::MoveLeft))
		{
			boundsCheck.left -= speed + 1;
			if (pos().x > boundary && ((door = static_cast<Door*>(player.getScene()->checkCollision(boundsCheck, "Door"))) || noclip || scene.checkCollision(boundsCheck, "StaticGeometry") == nullptr))
			{
				player.pos.x -= speed;
				moved = true;
			}
			face(FacingDirection::West);
			attempted = true;
		}
		speed -= 1.f;
	}

	if (door)
	{
		door->transfer(&player);
		door->enter();
	}

	player.setAnimationSpeed(moved ? 0.25f + 0.5f * startSpeed / walkSpeed : 0.f);

	if (moved)
	{
		scene.addEntityToList(new Noise(pos(), Noise::Level::concerning, startSpeed * 10));
		if (controls().held(Key::Sprint))
		{
			if (stamina > 0.f || scene.getGame()->getSettings().get("unlimited_sprint"))
			{
				stamina = std::max(0.f, stamina - 0.015f);
			}
			player.staminaCooldown = 45;
		}
	}

	const int staminaCooldownDrop = moved ? 1 : 3;
	if (player.staminaCooldown >= staminaCooldownDrop)
	{
		player.staminaCooldown -= staminaCooldownDrop;
	}
	else
	{
		stamina = std::min(1.f, stamina + (moved ? 0.01f : 0.05f));
	}

	return moved
		? AttemptWalkResult::Walked
		: (attempted ? AttemptWalkResult::Attempted : AttemptWalkResult::NotAttempted);
}

void PlayerState::loadWalkAnims()
{
	player.loadWalkAnims();
}

void PlayerState::face(FacingDirection dir)
{
	player.facingDir = dir;
	player.setDrawableSet(facingDirToStr(dir));
}

void PlayerState::handleButtonCallouts()
{
	player.handleButtonCallouts();
}

} // sl
