/**
 * @section DESCRIPTION
 * The abstract base class for Player states.
 */
#ifndef SL_PLAYERSTATE_HPP
#define SL_PLAYERSTATE_HPP

#include "NPC/FacingDirection.hpp"
#include "Player/PlayerStats.hpp"
#include "Utility/Vector.hpp"

namespace sl
{

class Controller;
class Event;
class Player;

class PlayerState
{
public:
	/**
	 * Destructor for PlayerState
	 */
	virtual ~PlayerState() {}
	/**
	 * Code to be executed when the STATE is entered.
	 */
	virtual void onEnter() {/*std::cout << "Enter\n";*/ }
	/**
	 * Code to be executed when the STATE exits
	 */
	virtual void onExit() {/*std::cout << "Exit\n";*/ }
	/**
	 * To be executed when the PLAYER'S SCENE is entered
	 */
	virtual void onSceneEnter() {}
	/**
	 * To be executed when the PLAYER'S SCENE exits
	 */
	virtual void onSceneExit() {}
	/**
	 * Handle an event
	 * @param event The event to handle
	 */
	virtual void handleEvent(const Event& event) {}

	/**
	 * Updates the state.
	 */
	virtual void update() {}
protected:
	/**
	 * Initializes the state
	 * @param owner The NPC this state belongs to
	 */
	PlayerState(Player& player);
	/**
	 * Access the owner's position in subclasses.
	 * @return Returns a reference to the owner's position vector.
	 */
	Vector2<float>& pos();

	PlayerStats& stats();

	const Controller& controls() const;

	/**
	 * Changes state
	 * @param newState The new state
	 */
	void changeState(std::unique_ptr<PlayerState> newState);

	enum class AttemptWalkResult
	{
		NotAttempted,
		Attempted,
		Walked
	};
	/**
	 * Attempts to walk
	 * @param scalar How fast relative to normal walking speed to walk, ie 0.5f = 50% speed
	 * @return NotAttempted if no input received, Attempted if input but couldn't physically move, Walked if moved.
	 */
	AttemptWalkResult attemptWalk(float scalar = 1.f);

	void loadWalkAnims();

	void face(FacingDirection dir);

	void handleButtonCallouts();



	Player& player;
};

}

#endif // SL_PLAYERSTATE_HPP
