#include "Player/StalkingData.hpp"

#include "Engine/Serialization/AutoDefineObjectSerialization.hpp"
#include "Engine/Serialization/Serialization.hpp"
#include "Utility/Assert.hpp"
#include "Utility/Functions/ConvergentSeries.hpp"
#include "Utility/Mapping.hpp"

namespace sl
{

float StalkingData::getFamiliarity(Soul::ID id, const GameDate& today) const
{
	auto it = data.find(id);
	if (it != data.end())
	{
		return it->second.computeFamiliarity(today);
	}
	return 0.f;
}

std::vector<Soul::ID> StalkingData::topStalked(int n) const
{
	const GameDate mostRecentDate = std::max_element(data.begin(), data.end(), [](const auto& lhs, const auto& rhs) {
		return lhs.second.lastStalked() < rhs.second.lastStalked();
	})->second.lastStalked();
	// construct as map to be able to pull in order by familiarity
	const std::map<float, Soul::ID> byFamiliarity = mapMap(data, [&](const auto& d) {
		return std::make_pair(d.second.computeFamiliarity(mostRecentDate), d.first);
	});
	const int actualN = std::min<int>(n, data.size());
	std::vector<Soul::ID> topN;
	topN.reserve(actualN);
	auto rit = byFamiliarity.rbegin();
	for (int i = 0; i < actualN; ++i)
	{
		topN.push_back(rit->second);
		++rit;
	}
	return topN;
}

void StalkingData::stalk(Soul::ID id, const GameDate& date)
{
	float& stalked = data[id].stalkedByDay[date];
	stalked = std::clamp(stalked + 0.25f, 0.f, 1.f);
}

void StalkingData::stalk(Soul::ID id, const GameTime& date, NPCPhoto photo)
{
	stalk(id, date.date);
	photos.push_back(PhotoInfo{ date, std::move(photo), id });
}

DEF_SERIALIZE_OBJ_CTOR(TimeOfDay, &TimeOfDay::getTotalSeconds)
DEF_SERIALIZE_OBJ_CTOR(GameDate, &GameDate::year, &GameDate::day)
DEF_SERIALIZE_OBJ_CTOR(GameTime, &GameTime::date, &GameTime::time)
DEF_SERIALIZE_OBJ_CTOR(sf::Color, &sf::Color::toInteger)
DEF_SERIALIZE_OBJ_MANUAL(NPCAppearance)
DEF_SERIALIZE_OBJ_FIELDS(NPCPhoto,
	&NPCPhoto::background,
	&NPCPhoto::data,
	&NPCPhoto::pose,
	&NPCPhoto::action,
	&NPCPhoto::face)
DEF_SERIALIZE_OBJ_CTOR(StalkingData::PhotoInfo,
	&StalkingData::PhotoInfo::time,
	&StalkingData::PhotoInfo::photo,
	&StalkingData::PhotoInfo::id)
DEF_SERIALIZE_OBJ_FIELDS(StalkingData::Data, &StalkingData::Data::stalkedByDay)

void StalkingData::serialize(Serialize& out) const
{
	AutoSerialize::fields(out, data, photos);
}

void StalkingData::deserialize(Deserialize& in)
{
	AutoDeserialize::fields(in, data, photos);
}

// Data
float StalkingData::Data::computeFamiliarity(const GameDate& today) const
{
	float familiarity = 0.f;
	for (const std::pair<GameDate, float>& byDay : stalkedByDay)
	{
		familiarity += byDay.second * geometricSeries(1 + today.u64() - byDay.first.u64());
	}
	return familiarity;
}

const GameDate& StalkingData::Data::lastStalked() const
{
	// should never happen, since stalking is only way to insert
	ASSERT(!stalkedByDay.empty());
	// since we store the dates in a map, rbegin() gets us the one with the most-recent date.
	return stalkedByDay.rbegin()->first;
}

} // sl