#include "Player/States/PlayerCarryingState.hpp"

#include "Engine/Graphics/Sprite.hpp"
#include "Engine/Graphics/SpriteAnimation.hpp"
#include "NPC/Stats/Soul.hpp"
#include "NPC/NPC.hpp"
#include "NPC/States/LoliRoutineState.hpp"
#include "NPC/LayeredSpriteLoader.hpp"
#include "Player/Player.hpp"
#include "Player/States/PlayerWalkState.hpp"
#include "Town/Cars/Car.hpp"
#include "Town/Container.hpp"

namespace sl
{

PlayerCarryingState::PlayerCarryingState(Player& player, NPC& kidnapee)
	:PlayerState(player)
	,kidnapee(kidnapee)
	,kidnappingCrime(nullptr)
{
	if (player.getScene())
	{
		onSceneEnter();
	}
	LayeredSpriteLoader::load(player, LayeredSpriteLoader::Config(player.getStats().soul->appearance, "dragging", 32, Vector2i(18, 28)).animate(8));
}

void PlayerCarryingState::update()
{
	attemptWalk(0.6f);

	if (kidnapee.getSoul()->getStatus().chloroformTimer <= 0 && Random::get().chance(1.f / 12.f))
	{
		const Vector2f impulse = lengthdir(4.f, Random::get().randomFloat(360.f));
		if (!player.getScene()->checkCollision(player.getMask().getBoundingBox() + impulse.to<int>(), "StaticGeometry"))
		{
			player.setPos(player.getPos() + impulse);
		}
	}

	kidnapee.update();

	std::list<Entity*> carsNear;
	player.getScene()->cull(carsNear, player, Entity::makeType("Car"));
	for (Entity *e : carsNear)
	{
		Car *car = static_cast<Car*>(e);
		if (car != nullptr)
		{
			if (car->isPlayerOwned() && car->isParked())
			{
				if (car->isTrunkFull())
				{
					player.addActionCallout(Player::CalloutType::Fill, "Can't place in car (Trunk is full)", std::function<void()>(), car);//[] {/* do nothing */});
				}
				else
				{
					player.addActionCallout(Player::CalloutType::Fill, "Dump body in trunk", [&,car]
					{
						kidnapee.getSoul()->setNPC(nullptr);
						car->storePersonInTrunk(kidnapee.getSoul());
						kidnapee.destroy();
						changeState(unique<PlayerWalkState>(player));
					}, car);
				}
			}
		}
	}

	std::list<Container*> containersNear;
	player.getScene()->cull(containersNear, player);
	for (Container *container : containersNear)
	{
		// make sure we prioritize it this way to make interacting less shitty for dumping bodies
		if (container->isClosed())
		{
			player.addActionCallout(Player::CalloutType::Open, "Open", [container]
			{
				container->setClosed(false);
			}, container);
		}
		else
		{
			player.addActionCallout(Player::CalloutType::Fill, "Dump body in container", [&, container]
			{
				kidnapee.getSoul()->setNPC(nullptr);
				container->addNPC(kidnapee.getSoul());
				kidnapee.destroy();
				changeState(unique<PlayerWalkState>(player));
			}, container);
			player.addActionCallout(Player::CalloutType::Use, "Close", [container]
			{
				container->setClosed(true);
			}, container);
		}
	}

	if (player.getScene()->checkCollision(Rect<int>(player.getScene()->getWidth(), player.getScene()->getHeight()), "Dungeon"))
	{
		player.addActionCallout(Player::CalloutType::Drop, "Put her in the dungeon", [&]
		{
			auto& flags = kidnapee.getSoul()->getFlags();
			if (!flags.get("captive").asBool())
			{
				flags.set("captive", ls::Value(true));
				flags.set("captive_where_am_i", ls::Value(1.f));
				flags.set("captive_what_you_do", ls::Value(1.f));
			}
			kidnapee.initDefaultState();
			changeState(unique<PlayerWalkState>(player));
		});
	}
	else
	{
		player.addActionCallout(Player::CalloutType::Drop, "Put her down", [&]
		{
			kidnapee.initDefaultState();
			changeState(unique<PlayerWalkState>(player));
		});
	}

	handleButtonCallouts();

	player.getScene()->setView(pos().x, pos().y);
}

void PlayerCarryingState::onSceneEnter()
{
	player.getScene()->transferEntity(&kidnapee);
}

void PlayerCarryingState::onSceneExit()
{
}

void PlayerCarryingState::onExit()
{
	// TODO: I think tihs was just here for when it gave the player a sideways loli sprite???
	//player.clearNonSetDrawables();
}

} // namespace sl
