#include "Player/States/PlayerDrivingState.hpp"

#include "Engine/Game.hpp"
#include "Player/Player.hpp"
#include "Player/States/PlayerWalkState.hpp"
#include "Town/Cars/Car.hpp"
#include "Town/Cars/CarControls.hpp"

//remove:
#include "Pathfinding/Pathfind.hpp"
#include "Town/RoadSegment.hpp"

namespace sl
{

PlayerDrivingState::PlayerDrivingState(Player& player, Car& car, CarControls& carControls)
	:PlayerState(player)
	,carControls(carControls)
	,car(car)
{
	if (player.getScene())
	{
		onSceneEnter();
	}
	player.car = &car;
}

void PlayerDrivingState::update()
{
	carControls.reset();

	if (controls().pressed(Key::ExitCar))
	{
		car.removeDriver();
		if (player.getScene()->checkCollision(player, "StaticGeometry") == nullptr)
		{
			changeState(unique<PlayerWalkState>(player));
			player.car = nullptr;
		}
		else
		{
			car.setDriver(&player, &player.getState());
		}
	}
	else
	{
		if (controls().pressed(Key::Cruise))
		{
			carControls.toggleCruiseSpeed();
		}

		if (controls().held(Key::SteerRight))
		{
			carControls.turnRight();
		}
		if (controls().held(Key::SteerLeft))
		{
			carControls.turnLeft();
		}
		if (controls().held(Key::Accelerate))
		{
			carControls.accelerate();
		}
		if (controls().held(Key::Reverse))
		{
			carControls.reverse();
		}
		if (controls().held(Key::Brake))
		{
			carControls.brake();
		}
		
		car.driftTowardsCenter(0.003f * (2.f + car.getSpeed()), 1.f);

		// remove
		if (controls().pressed(Key::Autopilot))
		{
			auto isEnd = [&](const pf::NodeID& node, ai::BehaviorContext*) {
				return pointDistance(car.getPos(), Vector2f(node.getX() * 16.f, node.getY() * 16.f)) > 640.f;
			};
			auto heuristic = [&](const pf::CarConfig& config, ai::BehaviorContext*) {
				const RoadSegment *road = static_cast<const RoadSegment*>(car.getScene()->checkCollision(Rect<int>(config.getTownPos().x - 12.f, config.getTownPos().y - 12.f, 24.f, 24.f), "RoadSegment"));
				if (!road)
				{
					return 20000.f;
				}
				pf::CarConfig::Angle roadAngle;
				switch (road->getOrientation())
				{
				case RoadSegment::Orientation::North:
					roadAngle = pf::CarConfig::Angle::North;
					break;
				case RoadSegment::Orientation::East:
					roadAngle = pf::CarConfig::Angle::East;
					break;
				case RoadSegment::Orientation::West:
					roadAngle = pf::CarConfig::Angle::West;
					break;
				case RoadSegment::Orientation::South:
					roadAngle = pf::CarConfig::Angle::South;
					break;
				}
				pf::CarConfig idealRoadConfig(config.getPos(), config.getSpeed(), roadAngle);
				return idealRoadConfig.distance(config) * 40.f;;
			};
			configPath.clear();
			if (!pf::findShortestCarPath(configPath, pf::CarConfig::carToConfig(car), isEnd, heuristic, nullptr))
			{
				configPath.clear();
				Logger::log(Logger::Area::Pathfinding, Logger::Debug) << "autopilot failed :(" << std::endl;
			}
			else
			{
				Logger::log(Logger::Area::Pathfinding, Logger::Debug) << "autopilot generated a path of " << configPath.size() << " nodes :)" << std::endl;
			}
		}
		// TODO: remove.
		if (!configPath.empty())
		{
			for (int i = 1; i < configPath.size(); ++i)
			{
				car.getScene()->debugDrawLine("CarConfig", configPath[i - 1].getTownPos(), configPath[i].getTownPos(), sf::Color::Red);
			}
		}

		car.viewFollowCar();
	}
}

void PlayerDrivingState::onSceneEnter()
{
}

void PlayerDrivingState::onSceneExit()
{
}

} // namespace sl
