#include "Player/States/PlayerShoppingState.hpp"

#include "Engine/Game.hpp"
#include "Engine/Graphics/Sprite.hpp"
#include "Engine/GUI/NormalBackground.hpp"
#include "Engine/GUI/WidgetList.hpp"
#include "NPC/NPC.hpp"
#include "Player/Player.hpp"
#include "Player/PlayerStats.hpp"
#include "Player/States/PlayerWalkState.hpp"
#include "Utility/Assert.hpp"
#include "Utility/Memory.hpp"
#include "Utility/ToString.hpp"

namespace sl
{

PlayerShoppingState::PlayerShoppingState(Player& player, NPC& npc)
	:PlayerState(player)
	,npc(npc)
	,selectedItem()
{
	loadWalkAnims();

	player.setDrawableSet("north");

	const int xOffset = 64;
	const int yOffset = 48;
	const int itemListWidth = 160;
	const int headerHeight = 32;
			
	gui::NormalBackground blackWhiteBG(sf::Color::Black, sf::Color::White);

	auto menu = unique<gui::GUIWidget>(Rect<int>(xOffset, yOffset, player.getScene()->getViewWidth() - 2 * xOffset, player.getScene()->getViewHeight() - 2 * yOffset));
	menu->applyBackground(gui::NormalBackground(sf::Color(14, 14, 14), sf::Color::White));


	// @todo refactor this + Inventory item info box into separate class
	// --------- item info box ---------
	const int itemBoxHGap = 24;
	const int itemBoxVGap = 24;
	const int itemGap = 24;
	const int itemBoxX = itemListWidth + xOffset + itemBoxHGap;
	const int itemBoxWidth = menu->getBox().width - (itemBoxX - xOffset) - itemBoxHGap;
	const int itemBoxY = yOffset + headerHeight;
	//const int itemBoxHeight = menu->getBox().height - (itemBoxY - yOffset) - itemBoxVGap;
	int ySoFar = itemBoxY;
	const int nameBoxHeight = 32;
	gui::TextLabel *nameBoxPtr = new gui::TextLabel(Rect<int>(itemBoxX, ySoFar, itemBoxWidth, nameBoxHeight),
		gui::TextLabel::Config().valign(gui::TextLabel::VerticalPolicy::Middle).pad(32, 0));
	nameBoxPtr->applyBackground(blackWhiteBG);
	menu->addWidget(std::unique_ptr<gui::TextLabel>(nameBoxPtr));
	ySoFar += nameBoxHeight + itemGap;

	const int previewBoxHeight = 96;
	auto previewBox = unique<gui::GUIWidget>(Rect<int>(itemBoxX, ySoFar, itemBoxWidth, previewBoxHeight));
	previewBox->applyBackground(blackWhiteBG);
	menu->addWidget(std::move(previewBox));
	ySoFar += previewBoxHeight + itemGap;

	const int descBoxHeight = 160;
	gui::TextLabel *descBoxPtr = new gui::TextLabel(Rect<int>(itemBoxX, ySoFar, itemBoxWidth, descBoxHeight),
		gui::TextLabel::Config().pad(16, 16));
	descBoxPtr->applyBackground(blackWhiteBG);
	menu->addWidget(std::unique_ptr<gui::TextLabel>(descBoxPtr));
	ySoFar += nameBoxHeight + itemGap;



	// --------- actions ----------
	
	const int actionButtonGap = 8;
	const int actionButtonHeight = 32;
	const int actionButtonWidth = itemListWidth - 2 * actionButtonGap;
	const int actions = 3;
	auto actionListHeightAt = [&](int index)
	{
		return (index + 1) * actionButtonGap + index * actionButtonHeight;
	};
	const int actionListHeight = actionListHeightAt(actions); 
	const int actionListY = menu->getBox().top + menu->getBox().height - actionListHeight;
	auto actionList = unique<gui::GUIWidget>(Rect<int>(menu->getBox().left, actionListY, itemListWidth, actionListHeight));
	

	int actionIndex = 0;
	auto createActionButton = [&](const std::string& text, std::function<void()> callback) -> gui::TextLabel*
	{
		const Rect<int> box(menu->getBox().left + actionButtonGap, actionListY + actionListHeightAt(actionIndex++), actionButtonWidth, actionButtonHeight);
		auto button = gui::Button::createGenericButton(box, text, callback);
		gui::TextLabel *ret = &button->getText();
		actionList->addWidget(std::move(button));
		return ret;
	};
	buyButton = createActionButton("Buy", [&]
	{
		ASSERT(selectedItem);
		if(player.getStats().money >= (*selectedItem)->getValue())
		{
			npc.getSoul()->items.transferItem(*selectedItem, player.getStats().soul->items);
			player.getStats().money -= (*selectedItem)->getValue();
			// @TODO check how many they have left, and only unselect if there's 0 now
			// OR give an option to buy multiple
			selectedItem.reset();
			updateItemList();
		}
	});
	createActionButton("???", []
	{

	});
	createActionButton("Exit", [&]
	{
		changeState(unique<PlayerWalkState>(player));
	});
	actionList->applyBackground(blackWhiteBG);
	menu->addWidget(std::move(actionList));

	// ---------   item list   ---------
	const int itemListY = menu->getBox().top;
	const Rect<int> itemListBox(menu->getBox().left, itemListY, itemListWidth, menu->getBox().height - actionListHeight);
	itemList = new gui::WidgetList<ItemListContainer::value_type>(
		itemListBox,
		[](const std::pair<Inventory::ItemRef, int>& item) -> std::string
		{
			return item.first->getName();
		},
			[&, nameBoxPtr, descBoxPtr](const std::pair<Inventory::ItemRef, int>& item)
		{
			nameBoxPtr->setString(item.first->getName());
			descBoxPtr->setString(item.first->getDesc());
			// @todo update preview later
			selectedItem = unique<Inventory::ItemRef>(item.first);
			const int cost = item.first->getValue();
			if (player.getStats().money >= cost)
			{
				buyButton->setString("Buy ($" + std::to_string(cost) + ")");
			}
			else
			{
				buyButton->setString("Insufficient funds ($" + std::to_string(cost) + ")");
			}
		}
	);
	std::unique_ptr<gui::WidgetList<ItemListContainer::value_type>> itemListWidget(itemList);
	itemList->applyBackground(blackWhiteBG);

	updateItemList();

	menu->addWidget(std::move(itemListWidget));

	widget.setHandle(std::move(menu));
	AbstractScene *scene = player.getScene();
	widget.setScene(scene);
	widget.enable();
				
	const int horizBorder = (player.getScene()->getViewWidth() - widget->getBox().width) / 2;
	const int vertBorder = (player.getScene()->getViewHeight() - widget->getBox().height) / 2;
				
	widget->move(horizBorder, vertBorder);
	widget->update();
}

void PlayerShoppingState::update()
{
	widget->update();
}

void PlayerShoppingState::onSceneEnter()
{
	ERROR("Why is a shopping player moving between scenes?");
}

void PlayerShoppingState::onSceneExit()
{
	ERROR("Why is a shopping player moving between scenes?");
}

void PlayerShoppingState::handleEvent(const Event& event)
{
}

// private
void PlayerShoppingState::updateItemList()
{
	std::vector<Inventory::ItemRef> items;
	npc.getSoul()->items.queryItems(items);
	ItemListContainer itemsByType;
	for (const Inventory::ItemRef& item : items)
	{
		++itemsByType[item];
	}
	itemList->updateItems(itemsByType);
}

} // namespace sl
