#include "Player/States/PlayerTalkState.hpp"

#include "Engine/Game.hpp"
#include "Engine/Graphics/Sprite.hpp"
#include "Engine/GUI/NormalBackground.hpp"
#include "NPC/NPC.hpp"
#include "Player/Player.hpp"
#include "Player/States/PlayerCarryingState.hpp"
#include "Player/States/PlayerShoppingState.hpp"
#include "Player/States/PlayerWalkState.hpp"
#include "Utility/Assert.hpp"
#include "Utility/Sequencer/SequencedFunction.hpp"

namespace sl
{

const int dialogueHBorder = 6;
const int dialogueVBorder = 4;
const int dialogueHeight = 216;

PlayerTalkState::PlayerTalkState(Player& player, NPC& npc, const std::string& startingSituation)
	:PlayerState(player)
	,npc(npc)
	,currentSequence(nullptr)
{
	loadWalkAnims();

	ls::Bindings& context = player.getScene()->getGame()->getContext();
	AbstractScene *scene = player.getScene();
	dialogue.setScene(scene);
	dialogue.setHandle(std::make_unique<DialogueGUI>(scene->getGame()->getContext(), Rect<int>(scene->getViewWidth() - 2 * dialogueHBorder, dialogueHeight), npc, player.getStats()));
	dialogue->applyBackground(gui::NormalBackground(sf::Color::Black, sf::Color::White));
	dialogue.enable();
	dialogue->setScenario(startingSituation);
	context.registerFunction("local", "changeState", std::function<ls::Value(const ls::Arguments&)>(
		[&](const ls::Arguments& args) -> ls::Value
		{
			Event event(Event::Category::NPCInteraction, "changeState_" + args[0].asString());
			dialogue->runAfterMessageDone([ev(std::move(event)),&npc,&player] {
				npc.handleEvent(ev);
				player.handleEvent(ev);
				//dialogue.disable();
			});
			return ls::Value();
		}
	));
	context.registerFunction("local", "action", std::function<ls::Value(const ls::Arguments&)>(
		[&](const ls::Arguments& args) -> ls::Value
	{
		Event event(Event::Category::NPCInteraction, "dialogue_subtree_" + args[0].asString());
		event.setFlag("paramCount", args.count() - 1);
		for (int i = 1; i < args.count(); ++i)
		{
			event.setFlag(std::to_string(i - 1), args[i]);
		}
		dialogue->runAfterMessageDone([ev(std::move(event)), &npc, &player] {
			npc.handleEvent(ev);
			player.getState().changeState(unique<PlayerWalkState>(player));
			//dialogue.disable();
		});
		return ls::Value();
	}));

	player.setAnimationSpeed(0.f);
}

void PlayerTalkState::update()
{
	if (currentSequence)
	{
		if (currentSequence->tick())
		{
			currentSequence = nullptr;
		}
	}

	dialogue->update(dialogueHBorder, player.getScene()->getViewHeight() - dialogue->getBox().height - dialogueVBorder);

	// offset view a bit so you're not covered by the dialogue gui
	player.getScene()->setView(pos().x, pos().y + 72);
}

void PlayerTalkState::onSceneEnter()
{
	ERROR("Why is a talking player moving between scenes?");
}

void PlayerTalkState::onSceneExit()
{
	ERROR("Why is a talking player moving between scenes?");
}

void PlayerTalkState::handleEvent(const Event& event)
{
	if (event.getCategory() == Event::Category::NPCInteraction)
	{
		const std::string& eventName = event.getName();

		if (eventName == "changeState_kidnapped")
		{
			changeState(unique<PlayerCarryingState>(player, npc));
		}
		else if (eventName == "changeState_deadloli")
		{
			changeState(unique<PlayerWalkState>(player));
		}
		else if (eventName == "changeState_idle")
		{
			changeState(unique<PlayerWalkState>(player));
		}
		else if (eventName == "changeState_shopping")
		{
			changeState(unique<PlayerShoppingState>(player, npc));
		}
	}
}

} // namespace sl
