#include "Player/States/PlayerWalkState.hpp"

#include "Engine/Game.hpp"
#include "Engine/GUI/NormalBackground.hpp"
#include "Items/ItemPickup.hpp"
#include "Menus/PauseScene.hpp"
#include "Menus/PhotoViewer.hpp"
#include "NPC/States/DraggedState.hpp"
#include "NPC/NPC.hpp"
#include "Player/Player.hpp"
#include "Player/Actions/ActionChooser.hpp"
#include "Player/Actions/PlayerActionFactory.hpp"
#include "Player/States/PlayerCarryingState.hpp"
#include "Player/States/PlayerDrivingState.hpp"
#include "Player/States/PlayerPerformingActionState.hpp"
#include "Player/States/PlayerTalkState.hpp"
#include "Town/Cars/Car.hpp"
#include "Town/Container.hpp"
#include "Town/World.hpp"
#include "Utility/Trig.hpp"


// toddler testing @REMOVE
#include "NPC/Stats/SoulGenerationSettings.hpp"
#include "NPC/States/ToddlerState.hpp"
#include "NPC/States/IdleState.hpp"
#include "NPC/ClothesDatabase.hpp"

namespace sl
{

PlayerWalkState::PlayerWalkState(Player& player)
	:PlayerWalkState(player, playerActionFromId(player.getStats().currentActionId))
{
}

PlayerWalkState::PlayerWalkState(Player& player, std::unique_ptr<PlayerAction> action)
	:PlayerState(player)
	,actionChooserResult(NO_CUSTOM_ACTION)
	,currentAction(std::move(action))
{
	loadWalkAnims();

	if (player.getScene())
	{
		onSceneEnter();
	}
}

void PlayerWalkState::update()
{
	NPC *near = player.getScene()->checkCollision<NPC>(rectAround(player.getPos(), 96));
	if (near)
	{
		//near->getSoul()->getStatus().urinaryUrge = std::fminf(1.f, near->getSoul()->getStatus().urinaryUrge + 0.005f);
		//near->getSoul()->getStatus().bladderContent.fluid = 400;
		//near->getSoul()->getStatus().bladderContent.fluid = 600;
		//std::cout << "wetting urges now: " << near->getSoul()->getStatus().urinaryUrge << std::endl;
	}

	std::list<ItemPickup*> itemsNear;
	player.getScene()->cull(itemsNear, player);
	for (ItemPickup *item : itemsNear)
	{
		player.addActionCallout(Player::CalloutType::PickUp, "Pick up " + item->getItem().getName(), [item, this]
		{
			player.getStats().soul->items.addItem(item->getItem());
			item->destroy();
		}, item);
	}

	std::list<Car*> carsNear;
	player.getScene()->cull(carsNear, player);
	for (Car *car : carsNear)
	{
		if (car->isPlayerOwned() && car->isParked())
		{
			if (!car->isTrunkEmpty())
			{
				player.addActionCallout(Player::CalloutType::PickUp, "Remove NPCs from trunk", [car, this]
				{
					NPC *npc = new NPC(player.getPos().x, player.getPos().y, car->takePersonFromTrunk());
					npc->changeState(unique<DraggedState>(*npc, player));
					player.getScene()->addImmediately(npc, StorageType::InQuad);
					changeState(unique<PlayerCarryingState>(player, *npc));
				}, car);
			}
			player.addActionCallout(Player::CalloutType::Enter, "Get in", [car, this]
			{
				CarControls *carControls;
				if (car->setDriver(&player, &player.getState(), carControls))
				{
					changeState(unique<PlayerDrivingState>(player, *car, *carControls));
				}
			}, car);
		}
	}

	std::list<Container*> containersNear;
	player.getScene()->cull(containersNear, player);
	for (Container *container : containersNear)
	{
		// make sure we prioritize it this way to make interacting less shitty for dumping bodies
		if (container->isClosed())
		{
			player.addActionCallout(Player::CalloutType::Open, "Open", [container]
			{
				container->setClosed(false);
			}, container);
		}
		else
		{
			if (!container->isClosed() && container->hasNPCs())
			{
				player.addActionCallout(Player::CalloutType::PickUp, "Remove NPC from dumpster", [container, this]
				{
					NPC *npc = new NPC(player.getPos().x, player.getPos().y, container->takeNPC());
					npc->changeState(unique<DraggedState>(*npc, player));
					player.getScene()->addImmediately(npc, StorageType::InQuad);
					changeState(unique<PlayerCarryingState>(player, *npc));
				}, container);
			}
			player.addActionCallout(Player::CalloutType::Use, "Close", [container]
			{
				container->setClosed(true);
			}, container);
		}
	}

	std::list<NPC*> npcsNear;
	player.getScene()->cull(npcsNear, player);
	for (NPC *npc : npcsNear)
	{
		const char *calloutMessage = npc->getDialogueStartOverride()
			? "Respond"
			: "Talk";
		player.addActionCallout(Player::CalloutType::Talk, calloutMessage, [&, npc]
		{
			npc->handleEvent(Event(Event::Category::NPCInteraction, "talk",
				[&, npc](const std::string& startNode)
			{
				//changeState(unique<PlayerTalkState>(player, *(NPC*)(npc), startNode));
			}
			));
		}, npc);
	}

	// DEMO HACK?
	if (controls().released(Key::GUIBack))
	{
		switchCurrentAction(NO_CUSTOM_ACTION);
	}

	CustomAction actionToPerform = controls().customActionPressed();
	bool perform = false;
	if (actionChooserResult != NO_CUSTOM_ACTION)
	{
		actionToPerform = actionChooserResult;
		actionChooserResult = NO_CUSTOM_ACTION;
		perform = true;
	}
	if (controls().released(Key::ActionChooser))
	{
		player.getScene()->getGame()->addScene(unique<PauseScene>(player.getWorld(), unique<gui::ActionChooser>(player, actionChooserResult)));
		return;
	}
	attemptWalk();

	// user-invoked action perform (via double-tap or menu)
	if (actionToPerform != NO_CUSTOM_ACTION)
	{
		if (Settings::fuckYouSingletonHack()->get("bind_first_use"))
		{
			perform = true;
		}
		if (actionToPerform != player.getStats().currentActionId)
		{
			switchCurrentAction(actionToPerform);
		}
		else
		{
			perform = true;
		}
		if (perform)
		{
			std::vector<NPC*> validTargets;
			const PlayerAction::Usability usability = getActionTargets(validTargets);
			if (usability == PlayerAction::Usability::Useable)
			{
				ASSERT(!validTargets.empty());
				performAction(validTargets.front());
			}
			else
			{
				// TODO: display some kind of thing saying why it couldn't be used
			}
		}
		
	}
	// no action invoked - but let's make some pop-ups anyway if there are valid targets
	else if (currentAction)
	{
		std::vector<NPC*> validTargets;
		const PlayerAction::Usability usability = getActionTargets(validTargets);
		if (usability == PlayerAction::Usability::Useable)
		{
			if (validTargets.size() > 1 || (validTargets.size() == 1 && validTargets.front() != nullptr))
			{
				for (NPC *target : validTargets)
				{
					player.addActionCallout(Player::CalloutType::Use, currentAction->getName(), std::bind(&PlayerWalkState::performAction, this, target), target);
				}
			}
		}
	}

	handleButtonCallouts();

	player.getScene()->setView(pos().x, pos().y);

	// @TODO: remove this once 100% convinced this is correct line-checking
	//const Vector2f a = player.getPos(), b = Vector2f(player.getScene()->getMousePos().x, player.getScene()->getMousePos().y);
	//std::list<Entity*> results;
	//player.getScene()->cullFiniteLine(results, a, b, "Car");
	//player.getScene()->debugDrawLine("draw_collision_masks", a, b,
	//	results.empty() ? sf::Color::Green : sf::Color::Red);
	Vector2f cast;
	auto result = player.getScene()->raycast(player.getPos(), { "StaticGeometry", "Car", }, Vector2f(player.getScene()->getMousePos().x - player.getPos().x, player.getScene()->getMousePos().y - player.getPos().y), cast);
	player.getScene()->debugDrawRect("draw_collision_masks", Rect<int>(cast.x - 1, cast.y - 1, 2, 2), result == Scene::RaycastResult::MovedFully ? sf::Color::Green : sf::Color::Red);
	player.getScene()->debugDrawLine("draw_collision_masks", player.getPos(), cast, result == Scene::RaycastResult::MovedFully ? sf::Color::Green : sf::Color::Red);

	//if (player.getScene()->getInputHandle().isKeyPressed(Input::KeyboardKey::T))
	//{
	//	// @REMOVE
	//	NPC *toddler = new NPC(player.getPos().x, player.getPos().y + 9, player.getScene()->getWorld()->createSoul(Random::get(),
	//		SoulGenerationSettings()
	//		.setAgeRange(4 * 365, 16 * 365)
	//		//.setAgeRange(1.5*365, 2.5*365)
	//			/*.addGenePool(DNADatabase::GenePool::NordicCaucasian)
	//			.addGenePool(DNADatabase::GenePool::Caucasian)
	//			.addGenePool(DNADatabase::GenePool::Arabic, 0.3)
	//			.addGenePool(DNADatabase::GenePool::Tropical)
	//			.addGenePool(DNADatabase::GenePool::Indian, 0.5)
	//			.addGenePool(DNADatabase::GenePool::Hispanic, 0.5)*/
	//		.addGenePool(DNADatabase::GenePool::Tropical, 2)
	//		.addGenePool(DNADatabase::GenePool::Nigger)
	//	));
	//	ClothesDatabase::get().dressLoliAppropriately(Random::get(), toddler->getSoul()->appearance);
	//	//toddler->initStates(unique<ToddlerState>(*toddler));
	//	toddler->initStates(unique<IdleState>(*toddler));
	//	
	//	player.getScene()->addEntityToQuad(toddler);
	//}

	// TODO: move?
	if (player.getScene()->getInputHandle().isKeyPressed(Input::KeyboardKey::P))
	{
		player.getScene()->getGame()->addScene(unique<PauseScene>(player.getWorld(), unique<gui::PhotoViewer>(player)));
		return;
	}
}

void PlayerWalkState::switchCurrentAction(CustomAction id)
{
	player.getStats().currentActionId = id;
	currentAction = playerActionFromId(id);
}

void PlayerWalkState::performAction(NPC *target)
{
	changeState(unique<PlayerPerformingActionState>(player, target, std::move(currentAction)));
}

PlayerAction::Usability PlayerWalkState::getActionTargets(std::vector<NPC*>& valid) const
{
	std::list<NPC*> actionTargetableNpcsNear;
	player.getScene()->cull(actionTargetableNpcsNear, rectAround(player.getPos(), currentAction->usageRadius()) + facingDirToVector(player.facing()).to<int>() * 20);
	// (try and) perform the action on the closest NPC, then if all else fails, try the no-target one.
	std::vector<NPC*> npcsVec(actionTargetableNpcsNear.begin(), actionTargetableNpcsNear.end());
	std::sort(npcsVec.begin(), npcsVec.end(), [&](const NPC *a, const NPC *b) -> bool {
		return pointDistance(player.getPos(), a->getPos()) < pointDistance(player.getPos(), b->getPos());
	});
	npcsVec.push_back(nullptr); // also the non-targeted one
	PlayerAction::Usability failReason = PlayerAction::Usability::Useable;
	for (NPC *npc : npcsVec)
	{
		const PlayerAction::Usability canUse = currentAction->canUse(player, npc);
		if (canUse == PlayerAction::Usability::Useable)
		{
			valid.push_back(npc);
		}
		else if (valid.empty() && failReason == PlayerAction::Usability::Useable)
		{
			failReason = canUse;
		}
	}
	return failReason;
}

} // namespace sl
