#include "Police/PoliceForce.hpp"

#include <map>

#include "Town/World.hpp"
#include "NPC/Stats/NationalityDatabase.hpp"
#include "NPC/States/Police/CopPatrolState.hpp"
#include "NPC/Stats/SoulGenerationSettings.hpp"
#include "NPC/NPC.hpp"
#include "Player/Fetishes.hpp"
#include "Player/Player.hpp"
#include "Police/CrimeDatabase.hpp"
#include "Town/Cars/Car.hpp"
#include "Town/Town.hpp"
#include "Utility/Assert.hpp"
#include "Utility/Random.hpp"

namespace sl
{

PoliceForce::PoliceForce()
	:severity(0)
	,lastKnownLocation()
	,town(nullptr)
	,updatesSincePlayerSeen(-1)
{
}

void PoliceForce::setTown(Town& town)
{
	this->town = &town;
}

void PoliceForce::update()
{
	std::map<PoliceForce::SeverityThresholds, int> severityCopMap;
	severityCopMap[PoliceForce::SeverityThresholds::None] = 3;
	severityCopMap[PoliceForce::SeverityThresholds::Lookout] = 9;
	severityCopMap[PoliceForce::SeverityThresholds::Pursuing] = 23;
	severityCopMap[PoliceForce::SeverityThresholds::Manhunt] = 42;
	int requiredCops = (Fetishes::get().isEnabled("ntr") ? 100 : severityCopMap[getSeverity()]) - cops.size();
	if (requiredCops > 0)
	{
		// add enough cops
		while (requiredCops > 0)
		{
			--requiredCops;
			const auto& randomNode = town->getRoadPathManager().getRandomID();
			const int x = randomNode.getX() * town->getTileSize();
			const int y = randomNode.getY() * town->getTileSize();
			Car *car = nullptr;
			// TODO: reenable cars when you get their AI to do path or whatever (you'll get an assert)
			//if (Random::get().chance(0.5f))
			//{
			//	car = new Car(x, y, CarSpecs::CopCar, town, false);
			//	town->addImmediately(car, StorageType::InQuad);
			//}
			SoulGenerationSettings genSettings;
			genSettings.setMaleChance(1.f);
			genSettings.addGenePool(DNADatabase::GenePool::Caucasian);
			NPC *npc = new NPC(x, y, town->getWorld()->createSoul(Random::get(), genSettings));
			npc->getSoul()->setType("cop");
			npc->getSoul()->appearance.setClothing(ClothingType::Outfit, "uniform");
			npc->getSoul()->appearance.setClothing(ClothingType::Accessory, "cophat");
			npc->getSoul()->appearance.setHair("short");
			npc->initStates(unique<CopPatrolState>(*npc, car));
			town->addImmediately(npc, StorageType::InQuad);
			cops.insert(npc->getMultiLink());
		}
	}
	else
	{
		// remove cop if he's offscreen
		for (Entity& cop : cops)
		{
			
			if (requiredCops >= 0)
			{
				break;
			}
			else if (!cop.getMask().getBoundingBox().intersects(town->getViewRect()))
			{
				++requiredCops;
				auto& copNPC = static_cast<NPC&>(cop);
				if (copNPC.getCar())
				{
					copNPC.getCar()->destroy();
				}
				copNPC.destroy();
			}
		}
	}
}


void PoliceForce::updateLastKnownLocation()
{
	Player *player = town->getWorld()->getPlayer();
	lastKnownLocation = Target(player->getPos(), player->getScene());
	updatesSincePlayerSeen = 0;
	//for (Entity& cops : cops)
	{

	}
}

const Target& PoliceForce::getLastKnownLocation() const
{
	return lastKnownLocation;
}

int PoliceForce::getLastKnownLocationAge() const
{
	return updatesSincePlayerSeen;
}

PoliceForce::SeverityThresholds PoliceForce::getSeverity() const
{
	const SeverityThresholds severityThresholds[3] = {Lookout, Pursuing, Manhunt};
#ifdef SL_DEBUG
	for (int i = 1; i < sizeof(severityThresholds) / sizeof(*severityThresholds); ++i)
	{
		ASSERT(severityThresholds[i - 1] < severityThresholds[i]);
	}
#endif
	SeverityThresholds severityLevel = None;
	for (SeverityThresholds severityThreshold : severityThresholds)
	{
		if (severity > severityThreshold)
		{
			severityLevel = severityThreshold;
		}
	}
	return severityLevel;
}

const Crime* PoliceForce::reportCrime(int crimeID)
{
	const Crime *crime = records.reportCrime(crimeID);
	if (crime != nullptr)
	{
		// @todo change to use a specific severity per crime and maybe do a cooldown instead of just using the months punishment
		severity += CrimeDatabase::get().getEntry(crime->type).sentenceLengthInMonths;
	}
	return crime;
}

void PoliceForce::recordCrime(const Crime& crime)
{
	records.recordCrime(crime);
}

const CriminalRecords& PoliceForce::getCriminalRecords() const
{
	return records;
}

void PoliceForce::serialize(Serialize& out) const
{
	SAVEVAR(i, severity);
	// @TODO last known location not serialized
	SAVEPTR(town);
	// @TODO cop list not serialized
	records.serialize(out);
	SAVEVAR(i, updatesSincePlayerSeen);
}

void PoliceForce::deserialize(Deserialize& in)
{
	LOADVAR(i, severity);
	LOADPTR(town);
	records.deserialize(in);
	LOADVAR(i, updatesSincePlayerSeen);
}

} // sl
