#ifndef SL_TEST_DIALOGUELINTTEST_HPP
#define SL_TEST_DIALOGUELINTTEST_HPP

#include <algorithm>
#include <iostream>
#include <filesystem>
#include <functional>
#include <map>
#include <set>
#include <string>
#include <stack>
#include <vector>

#include "NPC/Dialogue/DialogueBindings.hpp"
// TODO: try and refactor dependency out of here.
#include "NPC/Dialogue/DialogueGUI.hpp"
#include "NPC/Dialogue/ItemQuery.hpp"
#include "NPC/Dialogue/Popup.hpp"
#include "NPC/Stats/Soul.hpp"
#include "NPC/Stats/SoulGenerationSettings.hpp"
#include "NPC/NPC.hpp"
#include "Parsing/Binding/Bindings.hpp"
#include "Parsing/Binding/LintBindings.hpp"
#include "Parsing/Errors/SyntaxError.hpp"
#include "Parsing/Parsers/FileParser.hpp"
#include "Player/PlayerStats.hpp"
#include "Tests/Test.hpp"

namespace sl
{
namespace test
{

class DialogueLintTest : public Test
{
public:
	bool run(std::ostream& err) override
	{
		static const char *eqLine = "===============================================================================\n";
		// This is a two-phase test.
		std::vector<std::string> prevs;
		std::vector<std::string> backs;
		bool success = true;
		ls::LintBindings lintContext;
		try
		{
			// Phase 1: check every single file for syntax errors:
			// if an exception was thrown, then something went wrong in the construction
			// ie invalid syntax in a file.
			// results are cached by FileParser anyway so in Phase 2 we don't triple-parse every file
			namespace fs = std::filesystem;
			const std::string dlgDir = "resources\\dlg\\";
			std::set<std::string> checkedDlgFiles;
			// item query stuff does not impact which nodes get visited
			// just other side effects, so it's okay to ignore this.
			ItemQuery ignoredQuery;
			auto ignoredSoul = std::make_shared<Soul>(Random::get(), SoulGenerationSettings());
			ignoredQuery.registerInventory("player", &ignoredSoul->items, ignoredSoul.get());
			ignoredQuery.registerInventory("loli", &ignoredSoul->items, ignoredSoul.get());
			ls::Bindings context;
			// and these should only be called (hopefully) from non-EXECUTED (only parsed) code
			// so that we can just use the existence of functions for linting without executing them
			// If in the future someone puts these into links() forms we might need to ensure valid data...
			Popup popupIgnored{Vector2i(), NPCAppearance()};
			popupIgnored.registerFunctions(context, "loli");
			NPC npcIgnored{-1, -1, ignoredSoul};
			PlayerStats playerStats(ignoredSoul);
			DialogueGUI gui(context, Rect<int>(), npcIgnored, playerStats);
			lintContext.setBindings(context);
			lintContext.addFunction("local", "action");
			lintContext.addFunction("local", "changeState");
			// TODO: remove these 2. only necessary since UNUSED files (fight/punch) use them.
			lintContext.addVariable("player", "hp");
			lintContext.addVariable("loli", "hp");
			for (const fs::directory_entry& file : fs::recursive_directory_iterator(dlgDir))
			{
				const std::string filePath = file.path().string(); // to maintain iterator integrity
				const std::string ext = file.path().extension().string();
				const std::string filePathInDlg(
					std::next(filePath.begin(), dlgDir.size()),
					std::prev(filePath.end(), ext.size()));
				std::cout << "Checking '" << filePathInDlg << "' for syntax errors." << std::endl;
				lintContext.markFile(filePathInDlg);
				try
				{
					if (ext == ".dlg")
					{
						// to check for syntax errors
						const Scenario& scenario = ls::FileParser::get().readDialogueFile(filePathInDlg);
						// then runtime errors
						scenario.lint(lintContext);
						checkedDlgFiles.insert(canonifyDlgFilename(filePathInDlg));
					}
					else if (ext == ".ls")
					{
						// to check for syntax errors
						const ls::Statement& raw = ls::FileParser::get().readRawLoliscriptFile(filePathInDlg);
						// then runtime errors
						raw.lint(lintContext);
					}
					else if (ext == ".lso")
					{
						// TODO: lint these?
					}
				}
				catch (const ls::SyntaxError& e)
				{
					err << eqLine;
					err << e.what() << std::endl;
					err << "Skipping scenario [ " << filePathInDlg << " ] due to error parsing it." << std::endl;
					err << eqLine;
					success = false;
					continue;
				}
			}
			if (!success)
			{
				err << "Skipping structural check due to syntax errors in Phase 1 (syntax checking)" << std::endl;
			}
			else
			{
				// Phase 2: structural checks of graph
				// construct the graph via DFS-ing the dialogue files.
				std::set<std::string> visitedFiles = {
					"begin",
					"captivity/begin",
					"toddler/begin",
					"chloro_begin",
					"sleeping_begin"
				};
				std::stack<std::string> queue;

				for (const std::string& root : visitedFiles)
				{
					queue.push(root);
				}
				while (!queue.empty())
				{
					std::string cur = queue.top();
					queue.pop();
					//std::cout << "  --  visiting file " << cur << " --" << std::endl;
					auto& neighbors = arcs[cur];
					const Scenario *scenario;

					scenario = &ls::FileParser::get().readDialogueFile(cur);
					ASSERT(checkedDlgFiles.erase(canonifyDlgFilename(cur)));

					for (auto& outcome : scenario->outcomes)
					{
						std::vector<std::pair<std::string, std::string>> links;
						try
						{
							scenario->getLinks(*outcome, context, ignoredQuery, links);
						}
						catch (const std::exception& e)
						{
							// includes RuntimeError / SyntaxError
							err << eqLine;
							err << e.what() << std::endl;
							err << "Skipping scenario [ " << cur << " ] due to runtime error in links block." << std::endl;
							err << eqLine;
							success = false;
							continue;
						}
						for (const auto& link : links)
						{
							const std::string& file = link.first;
							if (file == "?PREV?")
							{
								prevs.push_back(cur);
							}
							else if (file == "?BACK?")
							{
								backs.push_back(cur);
							}
							else
							{
								// more context if you check if it doesn't exist here instead of waiting to pop it.
								if (std::filesystem::exists("resources/dlg/" + file + ".dlg"))
								{
									if (visitedFiles.insert(file).second)
									{
										queue.push(file);
									}
									neighbors.insert(file);
								}
								else
								{
									err << eqLine;
									err << "Could not find file " << file << ", linked to from " << cur << std::endl;
									err << eqLine;
									success = false;
								}
							}
						}
					}
				}

				if (!checkedDlgFiles.empty())
				{
					err << eqLine;
					// not an error, but worthwhile to report unreachable files.
					for (const std::string& unusedFile : checkedDlgFiles)
					{
						err << "Warning: Unused .dlg file: " << unusedFile << std::endl;
					}
					err << eqLine;
				}
			}
		}
		catch (const std::exception& e)
		{
			err << eqLine;
			// includes RuntimeError / SyntaxError
			err << e.what();
			err << eqLine;
			return false;
		}
		catch (...)
		{
			err << "unknown object thrown" << std::endl;
			return false;
		}
		for (const std::string& back : backs)
		{
			// not efficient but whatever.
			// if it takes too long we can optimize it later by building a reverse-graph as we go
			for (const auto& u : arcs)
			{
				if (u.second.count(back) != 0)
				{
					arcs[back].insert(u.first);
				}
			}
		}
		for (const std::string& prev : prevs)
		{
			for (const auto& u : arcs)
			{
				if (u.second.count(prev) != 0)
				{
					arcs[prev].insert(u.second.begin(), u.second.end());
				}
			}
		}

		// immediate failure upon any exception in construction is necessary, since there's no point
		// running the other tests on a malformed graph.
		// however, everything beyond this point isn't insta-fail.
		if (!checkPathToExits(err))
		{
			success = false;
		}
		// Report runtime errors found
		if (!lintContext.getErrors().empty())
		{
			err << "\n\n\n" << eqLine << "Runtime Errors fond:" << std::endl;
			for (const auto& error : lintContext.getErrors())
			{
				error.print(err);
				err << std::endl;
			}
			err << eqLine << std::endl;
			success = false;
		}
		// more tests here later
		return success;
	}
private:
	bool checkPathToExits(std::ostream& err) const
	{
		std::set<std::string> visited;
		const auto reverse = reverseArcs(arcs);
		std::function<void(const std::string&)> dfs = [&](const std::string& u) {
			visited.insert(u);
			for (const auto& v : reverse.at(u))
			{
				if (visited.count(v) == 0)
				{
					dfs(v);
				}
			}
		};
		// @TODO: automatically determine if this an exit based on the precense of an exit call
		const std::string exits[] = {
			"exit",
			"exit_direct",
			//"chloro_pickup",
			//"force/fight_punch",
			//"force/fight_choke",
			//"force/fight_kick",
			"playdoctor/follow_1",
			"playdoctor/reverse/follow_1"
		};
		for (const std::string& exit : exits)
		{
			dfs(exit);
		}
		for (const auto& u : arcs)
		{
			if (visited.count(u.first) == 0)
			{
				err << u.first << " does not contain any possible path to an exit-node" << std::endl;
				return false;
			}
		}
		return true;
	}

	std::map<std::string, std::set<std::string>> reverseArcs(const std::map<std::string, std::set<std::string>>& input) const
	{
		std::map<std::string, std::set<std::string>> output;
		for (const auto& u : input)
		{
			output[u.first]; // insert default if doesn't exist.
			for (const auto& v : u.second)
			{
				output[v].insert(u.first);
			}
		}
		return std::move(output);
	}

	static std::string canonifyDlgFilename(std::string str)
	{
		std::replace(str.begin(), str.end(), '\\', '/');
		return str;
	}

	std::map<std::string, std::set<std::string>> arcs;
};

} // test
} // sl

#endif // SL_TEST_DIALOGUELINTTEST_HPP