#include "Town/Cars/Car.hpp"

#include <algorithm>
#include <cmath>

#include "Engine/Input/Input.hpp"
#include "Engine/Graphics/Sprite.hpp"
#include "Engine/GUI/NormalBackground.hpp"
#include "Engine/Serialization/Impl/SerializeSoul.hpp"
#include "Engine/Serialization/AutoDeserialize.hpp"
#include "Engine/Serialization/AutoSerialize.hpp"
#include "NPC/Blood.hpp"
#include "NPC/NPC.hpp"
#include "Engine/Physics/PhysicsComponent.hpp"
#include "Player/Player.hpp"
#include "Police/CrimeUtil.hpp"
#include "Town/RoadSegment.hpp"
#include "Town/Cars/TireSkid.hpp"
#include "Utility/Logger.hpp"
#include "Utility/Random.hpp"
#include "Utility/Trig.hpp"
#include "Utility/TypeToString.hpp"
#include "Utility/Vector.hpp"

namespace sl
{

DECLARE_TYPE_TO_STRING(Car)

Car::Car(int x, int y, bool playerOwned)
	:Entity(x, y, 32, 32, "Car")
	,specs(nullptr)
	,speed(0)
	,carAngle(0)
	,slipping(0.f)
	,isSliding(false)
	,lastSkid(nullptr)
	,physics(this->getPhysics())
	,prevPos(pos)
	,parked(true)
	,playerOwned(playerOwned)
	,bloodbath(0)
	,controls()
	,driver(nullptr)
	,driverState(nullptr)
	,passengers()
	,trunk()
	,minimapIcon(pos, playerOwned ? sf::Color(0, 255, 255) : sf::Color::Transparent, nullptr)
	,model(model)
{
	mask.setOrigin(16, 16);
}

Car::Car(DeserializeConstructor)
	:Car(0, 0, true)
{
}

Car::Car(int x, int y, CarSpecs::Model model, Town *town, bool playerOwned)
	:Car(x, y, playerOwned)
{
	this->town = town;
	this->model = model;

	paintjob = Random::get().random(7);
	if (model == CarSpecs::Van)
	{
		paintjob = Random::get().choose({1, 6, 6, 6, 6, 6});
	}
	else if (model == CarSpecs::LuxerySedan)
	{
		paintjob = 1;
	}
	else if (model == CarSpecs::CopCar)
	{
		paintjob = -1; // no paintjob
	}
	//else if (model == CarSpecs::Supercar)
	//{
	//	paintjob = 5;
	//}

	init();
}

const Types::Type Car::getType() const
{
	return Entity::makeType("Car");
}

void Car::setDriver(Entity *driver, GenericStateManager *driverState)
{
	ASSERT(driver);
	ASSERT(driverState);

	this->driver = driver;
	this->driverState = driverState;
	town->deactivateEntity(driver);
	minimapIcon.hide();

	ASSERT(parked);
	parked = false;
}

bool Car::setDriver(Entity *driver, GenericStateManager *driverState, CarControls *&carControls)
{
	if (!this->driver)
	{
		setDriver(driver, driverState);
		carControls = &controls;
		return true;
	}

	return false;
}

void Car::removeDriver()
{
	ASSERT(!parked);
	parked = true;
	town->activateEntity(driver);
	driver->setPos(pos + Vector2f(24.f, 24.f) + lengthdir(32.f, carAngle + 90.f));
	driver = nullptr;
	driverState = nullptr;

	minimapIcon.show();
}

std::shared_ptr<Soul> Car::takePersonFromTrunk()
{
	std::shared_ptr<Soul> ret;
	if (!trunk.empty())
	{
		ret = trunk.back();
		trunk.pop_back();
	}
	return ret;
}

bool Car::storePersonInTrunk(std::shared_ptr<Soul> soul)
{
	if (!isTrunkFull())
	{
		trunk.push_back(soul);
		return true;
	}
	return false;
}

bool Car::isTrunkFull() const
{
	ASSERT(trunk.size() <= specs->trunkCapacity);
	return trunk.size() >= specs->trunkCapacity;
}

bool Car::isTrunkEmpty() const
{
	ASSERT(trunk.size() >= 0);
	return trunk.size() <= 0;
}

bool Car::isPlayerOwned() const
{
	return playerOwned;
}

bool Car::isParked() const
{
	return parked;
}

float Car::getAngle() const
{
	return carAngle;
}

float Car::getSpeed() const
{
	return speed;
}

void Car::stopForReal()
{
	pos = prevPos;
}

void Car::driftTowardsCenter(float factor, float dotFactor)
{
	// drift towards middle of the lane
	std::list<RoadSegment*> segments;
	scene->cull(segments, getMask().getBoundingBox());
	if (!segments.empty())
	{
		ASSERT(carAngle >= 0 && carAngle < 360);
		ASSERT(segments.size() == 2);
		const RoadSegment::Orientation o1 = segments.front()->getOrientation();
		const RoadSegment::Orientation o2 = (*std::next(segments.begin()))->getOrientation();
		if (o1 == RoadSegment::Orientation::North || o1 == RoadSegment::Orientation::South)
		{
			
			ASSERT(o1 != o2 && (o2 == RoadSegment::Orientation::North || o2 == RoadSegment::Orientation::South));
			const RoadSegment& segment = (carAngle >= 180 && carAngle < 360) == (o1 == RoadSegment::Orientation::South)
				? *segments.front()
				: **std::next(segments.begin());
			const float dot = std::max(0.f, physics.getSpeedVector().normalized().dot(segment.getDirection()));
			pos.x += (segment.getMagnet() - pos.x) * factor * (1.f - dotFactor + dot * dotFactor);
		}
		else
		{
			ASSERT(o1 != o2 && (o2 == RoadSegment::Orientation::West || o2 == RoadSegment::Orientation::East));
			const RoadSegment& segment = (carAngle < 90 || carAngle >= 270) == (o1 == RoadSegment::Orientation::East)
				? *segments.front()
				: **std::next(segments.begin());
			const float dot = std::max(0.f, physics.getSpeedVector().normalized().dot(segment.getDirection()));
			pos.y += (segment.getMagnet() - pos.y) * factor * (1.f - dotFactor + dot * dotFactor);
		}
	}
}

std::string Car::serializationType() const
{
	return "Car";
}

bool Car::shouldSerialize() const
{
	return isPlayerOwned();
}

void Car::serialize(Serialize& out) const
{
	serializeEntity(out);
	SAVEPTR(town);
	// ignore all physics-y stuff
	// ignore view /etc
	ASSERT(parked);
	ASSERT(playerOwned);
	// ignore bloodbath
	// ignore controls
	ASSERT(driver == nullptr);
	ASSERT(driverState == nullptr);
	ASSERT(passengers.empty()); // is this correct? probably not TODO: figure out how to handle this, ie figure out if these entities are still in a scene or not
	AutoSerialize::serialize(out, trunk);
	// ignore icon
	out.u32("model", static_cast<uint32_t>(model));
	SAVEVAR(i, paintjob);
}

void Car::deserialize(Deserialize& in)
{
	deserializeEntity(in);
	LOADPTR(town);
	AutoDeserialize::deserialize(in, trunk);
	model = static_cast<CarSpecs::Model>(in.u32("model"));
	LOADVAR(i, paintjob);

	init();
}


/*		private methods			*/
void Car::onEnter()
{
}

void Car::onExit()
{
}

void Car::handleEvent(const Event& event)
{
	if (driverState)
	{
		driverState->handleEvent(event);
	}
}

void Car::onUpdate()
{
	if (!parked)
	{
		controls.reset();
		driver->setPos(pos);
		driverState->update();

		// @todo I think this is causing an issue - make it so for now only the player can kill people
		std::list<NPC*> deathrow;
		town->cull<NPC, Entity&>(deathrow, *this);
		for (NPC *temp : deathrow)
		{
			if (physics.getSpeed() > 4 && isPlayerOwned()) //  kill
			{
				const int n = Random::get().randomInclusive(24, 32);
				for (int i = 0; i < n; ++i)
				{
					Vector2<float> veloc = Vector2<float>(physics.getSpeedVector().x * 2, physics.getSpeedVector().y * 2);
					veloc += lengthdir(4, Random::get().random(360));
					const float s = Random::get().randomFloat(1);
					town->addEntityToList(new Blood(temp->getPos().x, temp->getPos().y, veloc.x * s, veloc.y * s));
					if(bloodbath < 50)
					{
						++bloodbath;
					}
				}
				commitCrime(*temp, Crime::Type::Murder);
				temp->destroy();
			}
			else	//  knock back
			{
				const float angle = direction(physics.getSpeedVector());
				const Vector2f carToNPC = temp->getPos() - pos;
				const Vector2f directionNormal = lengthdir(1.f, angle + 90.f);
				const bool bumpLeft = carToNPC.dot(directionNormal) >= 0.f;
				const Vector2f bump = lengthdir(physics.getSpeed() * 0.5f, bumpLeft ? angle + 90.f : angle - 90.f);
				temp->setPos(temp->getPos() + physics.getSpeedVector() + bump);
			}
		}


	}
	if (parked)
	{
		stopForReal();
	}
	else
	{
		this->handlePhysics();
	}

	for (int i = 39; i > 0; i -= 1)
	{
		prevangle[i] = prevangle[i - 1];
	}
	prevangle[0] = carAngle;

	if (playerOwned && !parked)
	{
		Logger::log(Logger::Town) << "carAngle = " << carAngle << std::endl;
		Logger::log(Logger::Town) << "speed = " << speed << std::endl;
		Logger::log(Logger::Town) << "speed() = " << physics.getSpeed() << std::endl;
		Logger::log(Logger::Town) << "xspeed = " << physics.getSpeedVector().x << std::endl;
		Logger::log(Logger::Town) << "yspeed = " << physics.getSpeedVector().y << std::endl;
	}

	//Entity *temp = town->checkCollision(*this, Entity::makeType("NPC"));
	//std::cout << "temp = " << temp << std::endl;

	//if (temp)

	prevPos = pos;
}

/*		Car private methods			*/
void Car::handlePhysics()
{
	// TURNING
	const float turnThreshold = 0.1f;
	const float turnDirection = controls.getWheelRotation();
	const bool isTurning = (turnDirection >= turnThreshold) || (turnDirection <= -turnThreshold);
	isSliding = false;
	if (fabsf(physics.getSpeed()) > 0.25 && isTurning)
	{
		const float actualTurnSpeed = specs->turnSpeed * turnDirection;
		carAngle += std::clamp((physics.getSpeed() + 1) * actualTurnSpeed, -specs->turnMax, specs->turnMax);
		if (fabsf((physics.getSpeed() + 1) * actualTurnSpeed) > specs->turnMax * specs->grip / 10 + 5)
		{
			slipping += (physics.getSpeed() + 1) * fabsf(actualTurnSpeed) / 40 * (2 - specs->grip);
			isSliding = true;
		}
	}
	if (bloodbath > 0 && physics.getSpeed() > 0.9)
	{
		--bloodbath;
	}
	if (slipping > 10)
	{
		slipping = 10;
	}
	if (slipping > 3.f || bloodbath > 0)
	{
		TireSkid *s = new TireSkid(pos.x, pos.y, bloodbath > 0);
		scene->addImmediately(s, StorageType::InList);
		if (lastSkid)
		{
			s->join(lastSkid);
		}
		lastSkid = s;
	}
	else
	{
		lastSkid = nullptr;
	}

	if (controls.isBraking() && speed > 3)
	{
		slipping += 0.35f;
	}
	else if (!isSliding)
	{
		if (slipping > 1)
		{
			slipping -= 1;
		}
		else
		{
			slipping = 0;
		}
	}

	if (physics.getSpeed() > 0)
	{
		physics.setSpeed((speed*specs->grip)+(1-specs->grip)/100);

		const int motionImpulseIndices[] = {1, 2, 3, 5, 7, 9, 11, 14, 17};
		for (int offset : motionImpulseIndices)
		{
			physics.addMotion(prevangle[static_cast<int>(slipping + 1)], (speed*((1-specs->grip)/10))+(1-specs->grip)/100);
		}
	}
	else
	{
		physics.setSpeed(speed);
	}

	//carAngle = (pointDirection(Vector2f(0, 0), physics.getSpeedVector()) + carAngle) / 2.f;

	while (carAngle >= 360)
	{
		carAngle -= 360;
	}
	while (carAngle < 0)
	{
		carAngle += 360;
	}

	const float accelthreshold = 0.1f;
	const float pedalPos = controls.getPedalPosition();
	const bool isAccelerating = pedalPos >= accelthreshold;
	if (isAccelerating)
	{
		speed += specs->acceleration * pedalPos * 0.7;
	}
	if (controls.isReverse())
	{
		speed -= specs->accelerationReverse;
	}

	if (controls.isBraking())
	{
		if (speed > specs->brakes)
		{
			speed -= specs->brakes;
		}
		else if (speed < -specs->brakes)
		{
			speed += specs->brakes;
		}
		else
		{
			speed = 0;
		}
	}

	const float speedLimit = 5.f;
	if (speed > specs->topSpeed * 0.7)
	{
		speed = specs->topSpeed * 0.7;
	}
	if (controls.isSpeedLimited() && speed > speedLimit)
	{
		const float decrease = 0.5f;
		if (speed - decrease > speedLimit)
		{
			speed -= decrease;
		}
		else
		{
			speed = speedLimit;
		}
	}
	if (speed < specs->topSpeedReverse)
	{
		speed = specs->topSpeedReverse;
	}

	if (!isAccelerating && !controls.isReverse())
	{
		speed = speed * 0.98;
		if (speed < 3 && speed > -3)
		{
			if (speed > 0)
			{
				speed -= 0.25;
			}
			else
			{
				speed += 0.25;
			}
		}
		if (speed < 0.5 && speed > -0.5)
		{
			speed = 0;
		}
	}

	const int snap = 16;

	if (!isTurning)
	{
		if (carAngle >= 360 - snap || carAngle<= snap)
		{
			carAngle = 0;
		}
		if (carAngle >= 90 - snap && carAngle <= 90 + snap)
		{
			carAngle = 90;
		}
		if (carAngle >= 180 - snap && carAngle <= 180 + snap)
		{
			carAngle = 180;
		}
		if (carAngle >= 270 - snap && carAngle <= 270 + snap)
		{
			carAngle = 270;
		}
	}
	/*
	if (carAngle >= 45 && carAngle < 135)
	{
		pos.y -= speed / 5;
	}
	else if (carAngle >= 135 && carAngle < 225)
	{
		pos.x -= speed / 5;
	}
	else if (carAngle >= 225 && carAngle < 315)
	{
		pos.y += speed / 5;
	}
	else
	{
		pos.x += speed / 5;
	}
	*/
	if (carAngle >= 22.5 && carAngle < 67.5)
	{
		setDrawableSet("northeast");
	}
	else if (carAngle >= 67.5 && carAngle < 112.5)
	{
		setDrawableSet("north");
	}
	else if (carAngle >= 112.5 && carAngle < 157.5)
	{
		setDrawableSet("northwest");
	}
	else if (carAngle >= 157.5 && carAngle < 202.5)
	{
		setDrawableSet("west");
	}
	else if (carAngle >= 202.5 && carAngle < 247.5)
	{
		setDrawableSet("southwest");
	}
	else if (carAngle >= 247.5 && carAngle < 292.5)
	{
		setDrawableSet("south");
	}
	else if (carAngle >= 292.5 && carAngle < 337.5)
	{
		setDrawableSet("southeast");
	}
	else
	{
		setDrawableSet("east");
	}
}

void Car::viewFollowCar()
{
	const float targetDist = 32.f + (specs->topSpeed + 40.f) / (4.f * specs->topSpeed);
	const float to16 = std::min(physics.getSpeed() / 16.f, 1.f);
	const float toMax = physics.getSpeed() / specs->topSpeed;
	const float viewPush = 0.8f * to16 + 0.2f * toMax;
	const Vector2f dirVector = lengthdir(viewPush * 9.f,
	                                     direction(physics.getSpeedVector()) + 45.f * controls.getWheelRotation());
	const int viewOffsetTargetX = dirVector.x * targetDist;
	const int viewOffsetTargetY = dirVector.y * 0.75f * targetDist;
	viewOffsets[viewOffsetIndex] = Vector2f(viewOffsetTargetX, viewOffsetTargetY);
	if (++viewOffsetIndex >= VIEW_OFFSET_COUNT)
	{
		viewOffsetIndex = 0;
	}

	Vector2f viewOffset(0.f, 0.f);
	float total = 0.f;
	for (int i = 0; i < VIEW_OFFSET_COUNT; ++i)
	{
		const Vector2f& offset = viewOffsets[(viewOffsetIndex + i) % VIEW_OFFSET_COUNT];
		const float factor = powf(i + 1.f, 1.3f);
		viewOffset += offset * factor;
		total += factor;
	}
	viewOffset *= (1.f / total);

	town->setView(pos.x + viewOffset.x, pos.y + viewOffset.y);
}

void Car::init()
{
	prevPos = pos;
	minimapIcon.setScene(town);
	if (playerOwned)
	{
		minimapIcon.setDetailedIcon(map::LOD::Minimap, std::make_unique<Sprite>("map/player_car_icon_mm.png", 5, 5));
		minimapIcon.setDetailedIcon(map::LOD::Map, std::make_unique<Sprite>("map/player_car_icon.png", 8, 8));
	}

	ColorSwapID swap;
	sf::Color light(240, 240, 240);
	sf::Color dark(212, 212, 212);
	sf::Color outline(72, 72, 72);
	if (paintjob == 0)
	{
		swap = ColorSwapID({
			std::make_pair(light, sf::Color(0, 0, 205)),
			std::make_pair(dark, sf::Color(0, 0, 148)),
			std::make_pair(outline, sf::Color(0, 0, 64))
		});
	}
	else if (paintjob == 1)
	{
		swap = ColorSwapID({
			std::make_pair(light, sf::Color(32, 32, 32)),
			std::make_pair(dark, sf::Color(16, 16, 16)),
			std::make_pair(outline, sf::Color(0, 0, 0))
		});
	}
	else if (paintjob == 2)
	{
		swap = ColorSwapID({
			std::make_pair(light, sf::Color(0, 128, 0)),
			std::make_pair(dark, sf::Color(0, 64, 0)),
			std::make_pair(outline, sf::Color(0, 48, 0))
		});
	}
	else if (paintjob == 3)
	{
		swap = ColorSwapID({
			std::make_pair(light, sf::Color(240, 240, 0)),
			std::make_pair(dark, sf::Color(164, 164, 0)),
			std::make_pair(outline, sf::Color(72, 72, 0))
		});
	}
	else if (paintjob == 4)
	{
		swap = ColorSwapID({
			std::make_pair(light, sf::Color(192, 192, 211)),
			std::make_pair(dark, sf::Color(128, 128, 140)),
			std::make_pair(outline, sf::Color(64, 64, 96))
		});
	}
	else if (paintjob == 5)
	{
		swap = ColorSwapID({
			std::make_pair(light, sf::Color(164, 0, 0)),
			std::make_pair(dark, sf::Color(96, 0, 0)),
			std::make_pair(outline, sf::Color(64, 0, 0))
		});
	}
	int xoffset, yoffset;
	xoffset = yoffset = 48 / 2;
	if (model == CarSpecs::Schoolbus)
	{
		xoffset = yoffset = 64 / 2;
	}

	specs = &CarSpecs::getSpecs(model);

	// for serialization, where it would have created these twice. TODO: only do it once? serialization ctor?
	clearDrawablesInSet("north");
	clearDrawablesInSet("south");
	clearDrawablesInSet("west");
	clearDrawablesInSet("east");
	clearDrawablesInSet("northwest");
	clearDrawablesInSet("northeast");
	clearDrawablesInSet("southwest");
	clearDrawablesInSet("southeast");
	addDrawableToSet("north", std::make_unique<Sprite>(("cars/" + specs->filename + "/" + specs->filename + "_north.png").c_str(), swap, xoffset, yoffset));
	addDrawableToSet("south", std::make_unique<Sprite>(("cars/" + specs->filename + "/" + specs->filename + "_south.png").c_str(), swap, xoffset, yoffset));
	addDrawableToSet("west", std::make_unique<Sprite>(("cars/" + specs->filename + "/" + specs->filename + "_west.png").c_str(), swap, xoffset, yoffset));
	addDrawableToSet("east", std::make_unique<Sprite>(("cars/" + specs->filename + "/" + specs->filename + "_east.png").c_str(), swap, xoffset, yoffset));
	addDrawableToSet("northwest", std::make_unique<Sprite>(("cars/" + specs->filename + "/" + specs->filename + "_northwest.png").c_str(), swap, xoffset, yoffset));
	addDrawableToSet("northeast", std::make_unique<Sprite>(("cars/" + specs->filename + "/" + specs->filename + "_northeast.png").c_str(), swap, xoffset, yoffset));
	addDrawableToSet("southwest", std::make_unique<Sprite>(("cars/" + specs->filename + "/" + specs->filename + "_southwest.png").c_str(), swap, xoffset, yoffset));
	addDrawableToSet("southeast", std::make_unique<Sprite>(("cars/" + specs->filename + "/" + specs->filename + "_southeast.png").c_str(), swap, xoffset, yoffset));
	setDrawableSet("east");


	std::fill(std::begin(prevangle), std::end(prevangle), carAngle);

	this->enablePhysics();

	physics.setOnVelocChange([&](const Vector2f& newVeloc, PhysicsComponent::BounceType bounceType) {
		if (bounceType == PhysicsComponent::BounceType::Back)
		{
			carAngle = pointDirection(newVeloc, Vector2f(0.f, 0.f));
			speed = -newVeloc.length();
			for (int i = 0; i < 19; ++i)
			{
				prevangle[i] = carAngle;
			}
		}
		else
		{
			carAngle = pointDirection(Vector2f(0.f, 0.f), newVeloc);
			speed = newVeloc.length();
			for (int i = 0; i < 19; ++i)
			{
				prevangle[i] = carAngle;
			}
		}
	});
}

} // sl
