#ifndef SL_CAR_HPP
#define SL_CAR_HPP

#include "Engine/Input/Controller.hpp"
#include "Engine/StateManager.hpp"
#include "Engine/Entity.hpp"
#include "Engine/Serialization/DeserializeConstructor.hpp"
#include "NPC/Stats/Soul.hpp"
#include "Town/Cars/CarControls.hpp"
#include "Town/Cars/CarSpecs.hpp"
#include "Town/Town.hpp"
#include "Town/Map/Icon.hpp"

namespace sl
{

class TireSkid;

class Car : public Entity
{
public:
	Car(DeserializeConstructor);
	/**
	 * @param x The x position of the car
	 * @param y The y position of the car
	 * @param model The model of the car (determines sprite/stats)
	 * @param town The Town this car exists in
	 * @param playerOwned Whether the player owns this car (true) or NPC owned (false)
	 */
	Car(int x, int y, CarSpecs::Model model, Town *town, bool playerOwned);

	/**
	 * See: Entity::getType()
	 * @return "Car"
	 */
	const Types::Type getType() const;

	void setDriver(Entity *driver, GenericStateManager *driverState);

	bool setDriver(Entity *driver, GenericStateManager *driverState, CarControls *&carControls);

	void removeDriver();

	/**
	 * Takes a stored person from the trunk (removes from Car!)
	 * @return The Soul of the person stored in the trunk (nullptr if empty)
	 */
	std::shared_ptr<Soul> takePersonFromTrunk();

	/**
	 * Attempts to store a person in the trunk
	 * @param soul The person's Soul to store
	 * @return True if the person was stored, False if trunk was full
	 */
	bool storePersonInTrunk(std::shared_ptr<Soul> soul);

	/**
	 * @return True if trunk is full, Flase if trunk can store more NPCs
	 */
	bool isTrunkFull() const;

	/**
	 * @return True if trunk is empty, False if trunk has NPCs in it
	 */
	bool isTrunkEmpty() const;

	/**
	 * @return True if it's the player's car, False if an NPC
	 */
	bool isPlayerOwned() const;

	/**
	 * @return True if the car is parked, False if someone is driving it
	 */
	bool isParked() const;

	float getAngle() const;

	float getSpeed() const;

	/**
	 * Updates the Town's view to follow this car
	 */
	void viewFollowCar();

	void stopForReal();

	void driftTowardsCenter(float factor, float dotFactor);

	// Serializable
	std::string serializationType() const override;

	bool shouldSerialize() const override;

	void serialize(Serialize& out) const override;

	void deserialize(Deserialize& in) override;


private:
	// for delegation only
	Car(int x, int y, bool playerOwned);

	Car(const Car&) = delete;
	Car& operator=(const Car&) = delete;

	void onEnter() override;

	void onExit() override;

	void handleEvent(const Event& event) override;
	/**
	 * See: Entity::onUpdate()
	 */
	void onUpdate() override;

	/**
	 * Handles all the car driving physics after input has been calculated
	 */
	void handlePhysics();

	/**
	 * Creates the graphical / stats / etc components
	 */
	void init();



	//!The stats (accel,handling,topspeed, etc) of the car
	const CarSpecs *specs;
	//!The town this car exists in
	Town *town;

	/*   physics stuff   */
	//!The values of car_angle from the previous 40 frames
	float prevangle[40];
	//!The speed the car is actually moving
	float speed;
	//! Direction the car is actually facing
	float carAngle;
	//! The amount the cars tires are slipping
	float slipping;
	//! Whether the slipping has hit a certain threshold and the car starts to slip
	bool isSliding;
	//! The current tire marks effect when you're skidding - nullptr if not skidding
	TireSkid *lastSkid;
	//! Physics component to the car. responsible for bouncing and stuff
	PhysicsComponent& physics;

	/*  view stuff  */
	static const int VIEW_OFFSET_COUNT = 45;
	int viewOffsetIndex = 0;
	Vector2f viewOffsets[VIEW_OFFSET_COUNT];
		
	Vector2<float> prevPos;
	

	/* other stuff */
	bool parked;
	bool playerOwned;
	int bloodbath;

	/*  controls	*/
	CarControls controls;

	/*  driver/passanger/cargo  */
	Entity *driver;
	GenericStateManager *driverState;
	std::vector<Entity*> passengers;
	std::vector<std::shared_ptr<Soul>> trunk;

	map::Icon minimapIcon;
	/* kept for serialization only */
	CarSpecs::Model model;
	int paintjob;
};

}

#endif
