#include "Town/Generation/BlockGenerator.hpp"

#include <algorithm>

#include "Utility/ProportionList.hpp"

namespace sl
{
namespace towngen
{

BlockGenerator::BlockGenerator(BlockOfBuildings::FenceType fenceType, SoulGenerationSettings npcSettings)
	:buildingGenerator()
	,fenceType(fenceType)
	,npcSettings(std::move(npcSettings))
{
}


void BlockGenerator::addFiller(std::unique_ptr<YardBlueprint> filler, FillerPlacement fillerPlacement)
{
	fillers.push_back(std::make_pair(fillerPlacement, std::move(filler)));
}

BuildingGenerator& BlockGenerator::buildings()
{
	return buildingGenerator;
}

void BlockGenerator::generate(Random& rng, const Rect<int>& zoneArea, std::vector<BlockOfBuildings>& buildings, std::vector<std::tuple<int, int, int>>& roadSplitLocations, RequiredCountList *requiredBuildings) const
{
	std::vector<BlockOfBuildings> buildingsTemp;
	std::vector<std::tuple<int, int, int>> roadSplitsTemp;

	const int roadHeight = 4 + 5; // 2 * sidewalkThickness + roadWidth
	int heightUsed = 0;
	int lastHeight = 0;
	const int heightBudget = zoneArea.height;

	ASSERT(buildingGenerator.getMinHeight() <= zoneArea.height);

	constexpr int FENCE_HEIGHT = 1;
	while (heightUsed + buildingGenerator.getMinHeight() + FENCE_HEIGHT + (buildingsTemp.empty() ? 0 : roadHeight) <= heightBudget)
	{
		if (!buildingsTemp.empty())
		{
			roadSplitsTemp.push_back(std::make_tuple(zoneArea.left, zoneArea.top + heightUsed, zoneArea.width));
			heightUsed += roadHeight;
		}
		//blocks.push_back(Block(zones.at(x, y), Rect<int>(zoneArea.left, zoneArea.top, zoneArea.width, zoneArea.height)));
		buildingsTemp.emplace_back(BlockOfBuildings(zoneArea.left, zoneArea.top + heightUsed, fenceType, npcSettings));
		lastHeight = buildingGenerator.getRandomBuildings(rng, buildingsTemp.back(), zoneArea.width, heightBudget - heightUsed, requiredBuildings);
		heightUsed += lastHeight;
	}
	////  Perhaps we overshot...
	//if (heightUsed > heightBudget)
	//{
	//	//  undo adding of block (and remove required buildings count)
	//	buildingsTemp.back().clear(requiredBuildings);
	//	buildingsTemp.pop_back();
	//	heightUsed -= lastHeight;

	//	//  But maybe there's enough space to the smallest kind!
	//	const int roadHeightUsed = (buildingsTemp.empty() ? 0 : roadHeight);
	//	if (heightUsed + buildingGenerator.getMinHeight() + roadHeightUsed <= heightBudget)
	//	{
	//		buildingsTemp.push_back(BlockOfBuildings(zoneArea.left, zoneArea.top + heightUsed, fenceType, npcSettings));
	//		lastHeight = buildingGenerator.getRandomBuildings(rng, buildingsTemp.back(), zoneArea.width, zoneArea.height - heightUsed, requiredBuildings);
	//		heightUsed += lastHeight + roadHeightUsed;
	//	}
	//	else
	//	{
	//		//  we got rid of the last block of buildings, so get rid of the split too
	//		if (!roadSplitsTemp.empty())
	//		{
	//			roadSplitsTemp.pop_back();
	//			heightUsed -= lastHeight + roadHeight;
	//		}
	//	}
	//}
	ASSERT(!buildingsTemp.empty());
	std::map<int, int> extraBlockBudget;
	//  choose random blocks in the zone to give the leftover height to
	const int extraHeight = heightBudget - heightUsed;
	for (int i = 0, chosenBlock; i < extraHeight; ++i)
	{
		chosenBlock = rng.random(buildingsTemp.size());
		++extraBlockBudget[chosenBlock];
	}
	for (const auto& extra : extraBlockBudget)
	{
		int extraBudget = extra.second;
		const int chosenBlock = extra.first;
		// fill it with the biggest possible filler
		ProportionList<std::pair<FillerPlacement, YardBlueprint*>> possibleFillers;
		for (const auto& filler : fillers)
		{
			if (filler.second->getWidth() <= zoneArea.width && filler.second->getHeight() <= extraBudget)
			{
				const int weight = pow(filler.second->getHeight(), 1.5); // to make a much higher weighting for the larger ones
				possibleFillers.insert(std::make_pair(filler.first, filler.second.get()), weight);
			}
		}
		auto moveBlock = [&](int blockIndex, int yMove)
		{
			std::get<1>(roadSplitsTemp[blockIndex]) += yMove;
			buildingsTemp[blockIndex + 1].move(0, yMove);
		};
		if (!possibleFillers.empty())
		{
			const auto& chosenFiller = possibleFillers.get(rng);
			YardBlueprint filler(*chosenFiller.second);
			FillerPlacement fillerPlacement = chosenFiller.first;
			const int fillerHeight = std::min(filler.getMaxHeight(), extraBudget);
			filler.adjustYardSize(zoneArea.width - filler.getWidth(), fillerHeight - filler.getHeight());
			buildingsTemp[chosenBlock].setFiller(filler, fillerPlacement);
			extraBudget -= filler.getHeight();
			// move the current row down to accomodate the filler
			if (fillerPlacement == FillerPlacement::Above)
			{
				//moveBlock(0, fillerHeight);
			}
		}

		//  extend the yard with any leftovers
		if (extraBudget > 0)
		{
			buildingsTemp[chosenBlock].extendVertically(rng, extraBudget);
		}
		//  make sure to adjust the positions of all roads and buildings below that one
		for (int i = chosenBlock, size = roadSplitsTemp.size(); i < size; ++i)
		{
			// extra.second to include both the leftovers AND the fillerheight
			moveBlock(i, extra.second);
		}
	}
	for (std::tuple<int, int, int>& split : roadSplitsTemp)
	{
		roadSplitLocations.push_back(split);
	}
	for (BlockOfBuildings& block : buildingsTemp)
	{
		buildings.push_back(std::move(block));
	}
}


} // towngen
} // sl
