#include "Town/Generation/BlockOfBuildings.hpp"

#include <algorithm>

#include "Town/Town.hpp"
#include "Utility/Assert.hpp"
#include "Utility/Random.hpp"

namespace sl
{
namespace towngen
{

BlockOfBuildings::BlockOfBuildings(int x, int y, FenceType fenceType, const SoulGenerationSettings& npcSettings)
	:x(x)
	,y(y)
	,width(FENCE_WIDTH)
	,height(FENCE_HEIGHT)
	,fillerPlacement(FillerPlacement::Above)
	,fenceType(fenceType)
	,npcSettings(npcSettings)
{
}

BlockOfBuildings::BlockOfBuildings(BlockOfBuildings&& other)
	:x(other.x)
	,y(other.y)
	,width(other.width)
	,height(other.height)
	,blueprints(std::move(other.blueprints))
	,filler(std::move(other.filler))
	,fillerPlacement(other.fillerPlacement)
	,fenceType(other.fenceType)
	,npcSettings(other.npcSettings)
{
}
BlockOfBuildings::BlockOfBuildings(BlockOfBuildings& other)
	:x(other.x)
	,y(other.y)
	,width(other.width)
	,height(other.height)
	,blueprints(std::move(other.blueprints))
	,filler(std::move(other.filler))
	,fillerPlacement(other.fillerPlacement)
	,fenceType(other.fenceType)
	,npcSettings(other.npcSettings)
{
}

BlockOfBuildings& BlockOfBuildings::operator=(BlockOfBuildings&& rhs)
{
	x = rhs.x;
	y = rhs.y;
	width = rhs.width;
	height = rhs.height;
	blueprints = std::move(rhs.blueprints);
	filler = std::move(rhs.filler);
	fillerPlacement = rhs.fillerPlacement;
	fenceType = rhs.fenceType;
	return *this;
}


void BlockOfBuildings::generate(Random& rng, Town& town, Grid<TileType>& tiles, int tileSize, map::MapStaticEntityData& minimapStaticEntityData, bool& isPlayerGenerated)
{
	if (blueprints.empty())
	{
		return;
	}
	const int fillerOffset = filler && fillerPlacement == FillerPlacement::Below ? getHeight() : 0;
	const int buildingsOffset = FENCE_HEIGHT + (filler && fillerPlacement == FillerPlacement::Above ? filler->getHeight() : 0);
	const int yardsHeight = std::max_element(blueprints.cbegin(), blueprints.cend(), [](const YardBlueprint& a, const YardBlueprint& b) {
		return a.getHeight() < b.getHeight();
	})->getHeight();
	std::vector<StaticEntityPrototype> entities;
	for (int i = y + height - yardsHeight - FENCE_HEIGHT; i < y + height - FENCE_HEIGHT; ++i)
	{
		entities.push_back(StaticEntityPrototype(x, i + buildingsOffset, vertFenceType()));
	}
	int xSoFar = x + FENCE_WIDTH;
	// do fences at first so they're there when we generate the yards (to avoid trees/etc running into them)
	for (int bi = 0; bi < blueprints.size(); ++bi)
	{
		const int fenceHeight = FENCE_HEIGHT + ((bi < blueprints.size() - 1)
			? std::max(blueprints[bi].getHeight(), blueprints[bi + 1].getHeight())
			: blueprints[bi].getHeight());
		for (int i = xSoFar - FENCE_WIDTH; i < xSoFar + blueprints[bi].getWidth(); ++i)
		{
			entities.push_back(StaticEntityPrototype(i + FENCE_WIDTH, y + buildingsOffset - FENCE_HEIGHT, horizFenceType()));
			//entities.push_back(StaticEntityPrototype(i + 1, y + height, StaticEntityPrototype::FenceH));
		}
		for (int i = y + height - yardsHeight - FENCE_HEIGHT; i < y + height - FENCE_HEIGHT; ++i)
		{
			entities.push_back(StaticEntityPrototype(xSoFar + blueprints[bi].getWidth(), i + buildingsOffset, vertFenceType()));
		}

		xSoFar += blueprints[bi].getWidth();
		xSoFar += FENCE_WIDTH;
	}
	xSoFar = x + FENCE_WIDTH;
	for (int bi = 0; bi < blueprints.size(); ++bi)
	{
		ASSERT(height >= blueprints[bi].getHeight() + FENCE_HEIGHT);
		//	place buildings
		const int heightDiff = 0;//height - FENCE_HEIGHT - blueprints[bi].getHeight();
		blueprints[bi].generate(Vector2i(xSoFar, y + buildingsOffset + heightDiff) * tileSize, rng, town, tiles, tileSize, minimapStaticEntityData, isPlayerGenerated, npcSettings);

		xSoFar += blueprints[bi].getWidth();
		xSoFar += FENCE_WIDTH;
	}

	if (filler)
	{
		filler->generate(Vector2i(x, y + fillerOffset) * tileSize, rng, town, tiles, tileSize, minimapStaticEntityData, isPlayerGenerated, npcSettings);
	}

	for (const StaticEntityPrototype& entity : entities)
	{
		entity.generate(rng, town, tileSize, minimapStaticEntityData);
	}
}

int BlockOfBuildings::getX() const
{
	return x;
}

int BlockOfBuildings::getY() const
{
	return y;
}

int BlockOfBuildings::buildings() const
{
	return blueprints.size();
}

int BlockOfBuildings::getWidth() const
{
	return width;
}

int BlockOfBuildings::getHeight() const
{
	return height;
}

void BlockOfBuildings::add(const YardBlueprint& blueprint)
{
	width += blueprint.getWidth() + FENCE_WIDTH;
	if (blueprint.getHeight() + FENCE_HEIGHT > height)
	{
		height = blueprint.getHeight() + FENCE_HEIGHT;
	}
	blueprints.push_back(blueprint);
}

void BlockOfBuildings::setFiller(const YardBlueprint& blueprint, FillerPlacement fillerPlacement)
{
	//if (filler)
	//{
	//	height -= filler->getHeight();
	//}
	//height += blueprint.getHeight();
	filler = unique<YardBlueprint>(blueprint);
	this->fillerPlacement = fillerPlacement;
}

void BlockOfBuildings::remove(RequiredCountList *required)
{
	ASSERT(!blueprints.empty());
	width -= blueprints.back().getWidth() + FENCE_WIDTH;
	if (required)
	{
		(*required)[blueprints.back().getName()].created--;
	}
	blueprints.pop_back();
}

void BlockOfBuildings::clear(RequiredCountList *required)
{
	if (required)
	{
		for (const auto& blueprint : blueprints)
		{
			(*required)[blueprint.getName()].created--;
		}
	}
	blueprints.clear();
	width = FENCE_WIDTH;
	height = FENCE_HEIGHT;
}

const YardBlueprint& BlockOfBuildings::last() const
{
	return blueprints.back();
}

void BlockOfBuildings::extendVertically(Random& rng, int height)
{
	ASSERT(height >= 0);
	// because some yards will have a max height limit, we need to check what the maximum we added to every block is
	// since if they all have a height limit we might add less than we tried, even to the one that was extended the most
	int maxHeightAdded = 0;
	for (YardBlueprint& blueprint : blueprints)
	{
		const int heightToAdd = this->height + height - blueprint.getHeight() - FENCE_HEIGHT;
		ASSERT(heightToAdd > 0);
		if (heightToAdd > 0)
		{
			//const int topHeight = rng.random(heightToAdd);
			//blueprint.extend(0, 0, topHeight, heightToAdd - topHeight);
			maxHeightAdded = std::max(maxHeightAdded, blueprint.adjustYardSize(0, heightToAdd));
		}
	}
	this->height += height;
}

void BlockOfBuildings::extendHorizontally(Random& rng, int width)
{
	ASSERT(width >= 0);
	this->width += width;
	for (int i = 0; i < width; ++i)
	{
		YardBlueprint& yard = blueprints[rng.random(blueprints.size())];
		yard.adjustYardSize(1, 0);
		//if (rng.chance(0.5))
		//{
		//	yard.extend(1, 0, 0, 0);
		//}
		//else
		//{
		//	yard.extend(0, 1, 0, 0);
		//}
	}
#ifdef ASSERT
	int cumulWidth = FENCE_WIDTH;
	for (const YardBlueprint& yard : blueprints)
	{
		cumulWidth += FENCE_WIDTH + yard.getWidth();
	}
	ASSERT(cumulWidth == this->width);
#endif
}

void BlockOfBuildings::move(int x, int y)
{
	this->x += x;
	this->y += y;
}

//-------------------------  BlockOfBuildings::Yard  --------------------------
//BlockOfBuildings::Yard::Yard(YardBlueprint& blueprint)
//	:blueprint(&blueprint)
//	,x(0)
//	,y(0)
//	,width(blueprint.getWidth())
//	,height(blueprint.getHeight())
//{
//}
//
//
//void BlockOfBuildings::Yard::generate(const Vector2i& pos, Random& rng, Town& town)
//{
//	blueprint->generate(Vector2i(pos.x + x * 32, pos.y + y * 32), rng, town);
//}
//
//void BlockOfBuildings::Yard::extend(int left, int right, int top, int bottom)
//{
//	ASSERT(left >= 0 && right >= 0 && top >= 0 && bottom >= 0);
//	if (left)
//	{
//		x += left;
//		width += left;
//	}
//	if (right)
//	{
//		width += right;
//	}
//	if (top)
//	{
//		y += top;
//		height += top;
//	}
//	if (bottom)
//	{
//		height += bottom;
//	}
//}
//
//int BlockOfBuildings::Yard::getWidth() const
//{
//	return width;
//}
//
//int BlockOfBuildings::Yard::getHeight() const
//{
//	return height;
//}



// private
StaticEntityPrototype::Type BlockOfBuildings::horizFenceType() const
{
	switch (fenceType)
	{
	case FenceType::Wooden:
		return StaticEntityPrototype::FenceH;
	case FenceType::Chainlink:
		return StaticEntityPrototype::ChainLinkFenceH;
	}
	return StaticEntityPrototype::Type::Tree; // lol
}

StaticEntityPrototype::Type BlockOfBuildings::vertFenceType() const
{
	switch (fenceType)
	{
	case FenceType::Wooden:
		return StaticEntityPrototype::FenceV;
	case FenceType::Chainlink:
		return StaticEntityPrototype::ChainLinkFenceV;
	}
	return StaticEntityPrototype::Type::Tree; // lol
}


} // towngen
} // sl
