#include "Town/Generation/BuildingBlueprint.hpp"

#include <algorithm>
#include <numeric>
#include <set>

#include "Engine/Door.hpp"
#include "Town/World.hpp"
#include "Engine/Graphics/Sprite.hpp"
#include "Home/BuildingLevel.hpp"
#include "NPC/Stats/SoulGenerationSettings.hpp"
#include "Player/Player.hpp"
#include "Player/PlayerGen.hpp"
#include "Player/States/PlayerWalkState.hpp"
#include "Town/Generation/Interiors/HomeConfig.hpp"
#include "Town/LocationMarker.hpp"
#include "Town/Cars/Car.hpp"
#include "Town/StaticGeometry.hpp"
#include "Town/Town.hpp"
#include "Town/Map/LOD.hpp"
#include "Town/Map/MapStaticEntityData.hpp"
#include "Utility/Increment.hpp"
#include "Utility/Random.hpp"
#include "Utility/Memory.hpp"

namespace sl
{
namespace towngen
{

BuildingBlueprint::BuildingBlueprint(const std::string& filename, const Vector2i& imageOrigin, const Vector2i& size, InsideType insideType, std::vector<ColorSwapID>&& swaps)
	:insideType(insideType)
	,filename(filename)
	,swapIDs(swaps)
	,imageOrigin(imageOrigin)
{
	solidGeo.push_back(Rect<int>(imageOrigin.x, imageOrigin.y, size.x * 16, size.y * 16));
	std::fill(std::begin(maxProtrusion), std::end(maxProtrusion), 0);
}

//BuildingBlueprint::BuildingBlueprint(BuildingBlueprint&& other)
//	:insideType(other.insideType)
//	,filename(std::move(other.filename))
//	,swapIDs(std::move(other.swapIDs))
//	,imageOrigin(other.imageOrigin)
//	,solidGeo(std::move(other.solidGeo))
//{
//	for (int i = 0; i < 4; ++i)
//	{
//		attachments[i] = std::move(other.attachments[i]);
//		maxProtrusion[i] = other.maxProtrusion[i];
//	}
//}
//
//BuildingBlueprint& BuildingBlueprint::operator=(BuildingBlueprint&& rhs)
//{
//	insideType = rhs.insideType;
//	filename = std::move(rhs.filename);
//	swapIDs = std::move(rhs.swapIDs);
//	imageOrigin = rhs.imageOrigin;
//	solidGeo = std::move(rhs.solidGeo);
//	for (int i = 0; i < 4; ++i)
//	{
//		attachments[i] = std::move(rhs.attachments[i]);
//		maxProtrusion[i] = rhs.maxProtrusion[i];
//	}
//	return *this;
//}


void BuildingBlueprint::setDetailed(std::initializer_list<Rect<int>> regions)
{
	solidGeo.clear();
	for (const Rect<int>& region : regions)
	{
		solidGeo.push_back(region);
	}
}

void BuildingBlueprint::addAttachment(const AttachmentConfig& config, int sides)
{
	auto addAttachmentImpl = [&](AttachedSide side)
	{
		const int sideIndex = sideToIndex(side);
		const bool isSides = (side == Left || side == Right);
		const int newProtrusion = isSides ? config.attachment->getWidth() : config.attachment->getHeight();
		maxProtrusion[sideIndex] = std::max(maxProtrusion[sideIndex], newProtrusion);
		attachments[sideIndex].push_back(config);
		ASSERT(computeMaxAttachmentSize(side) <= (isSides ? getHeight() : getWidth()));
	};
	for (int i = 0; i < 4; ++i)
	{
		const AttachedSide side = indexToSide(i);
		if (side & sides)
		{
			addAttachmentImpl(side);
		}
	}
}

void BuildingBlueprint::addPathSpawn(const Vector2i& pathStart)
{
	pathSpawns.push_back(pathStart);
}

const std::vector<Vector2i>& BuildingBlueprint::getPathSpawns() const
{
	return pathSpawns;
}

ConditionalPrototypeLimit* BuildingBlueprint::createNewLimit(int count)
{
	limits.push_front(ConditionalPrototypeLimit(count));
	return &limits.front();
}



void BuildingBlueprint::generate(const Vector2i& pos, Random& rng, Town& town, map::MapStaticEntityData& minimapStaticEntityData, bool& isPlayerGenerated, const SoulGenerationSettings& npcSettings, const ColorSwapID *forceSwap) const
{
	const int ts = town.getTileSize();
	const Vector2i& housePos = pos + Vector2i(ts * maxProtrusion[sideToIndex(Left)], ts * maxProtrusion[sideToIndex(Top)]);

	// ------------------------------------------------------------------------
	//                   create the exterior of the building

	// make sure to reset all generation-limits in case this is not the first time this blueprint is being used
	for (ConditionalPrototypeLimit& limit : limits)
	{
		limit.reset();
	}

	std::unique_ptr<Sprite> houseSprite;
	if (swapIDs.empty() && !forceSwap)
	{
		houseSprite = unique<Sprite>(filename.c_str(), imageOrigin.x, imageOrigin.y);
	}
	else
	{
		if (!forceSwap)
		{
			forceSwap = &swapIDs[rng.random(swapIDs.size())];
		}
		houseSprite = unique<Sprite>(filename.c_str(), *forceSwap, imageOrigin.x, imageOrigin.y);
	}
	for (const Rect<int>& region : solidGeo)
	{
		if (houseSprite)
		{
			houseSprite->setBottomPointOffset(region.bottom() - houseSprite->getCurrentBounds().height);
		}
		Entity *house = new StaticGeometry(housePos.x + region.left - imageOrigin.x, housePos.y + region.top - imageOrigin.y, region.width, region.height, std::move(houseSprite));
		town.addImmediately(house, StorageType::InGrid);
		minimapStaticEntityData.set(*house, map::MapStaticEntityData::StaticEntity::Building);
	}

	const int tx = housePos.x / ts;
	const int ty = housePos.y / ts;



	// ------------------------------------------------------------------------
	//              generate the interiors of the buildings
	if (insideType == InsideType::HouseBasic && !isPlayerGenerated)
	{
		isPlayerGenerated = true;

		// @TODO add player's dungeon back in here
		Player *player = new Player(housePos.x + 64, housePos.y + 128, createPlayer(*town.getWorld(), rng));
		player->getState().initStates(unique<PlayerWalkState>(*player));
		town.addEntityToList(player);

		// give him other shit to drive too
		town.addEntityToQuad(new Car(housePos.x - 64, housePos.y + 196, CarSpecs::Model::SportsCar, &town, true));
		//town.addEntityToQuad(new Car(housePos.x - 160, housePos.y + 196, CarSpecs::Model::Schoolbus, &town, true));
		town.addEntityToQuad(new Car(housePos.x - 125, housePos.y + 196, CarSpecs::Model::Van, &town, true));
		
		// and a sign to help him find his way home
		town.addEntityToGrid(new StaticGeometry(housePos.x + 8, housePos.y + 48, 0, 0, unique<Sprite>("homesign.png")));

		LocationMarker *homeMarker = new LocationMarker(housePos.x, housePos.y, "Player's House", "home");
		town.addEntityToList(homeMarker);

		HomeConfig config(Rect<int>(tx, ty, HomeConfig::AUTOMATIC, HomeConfig::AUTOMATIC), npcSettings);
		config.addAboveLevel(HomeConfig::LevelConfig("player/ground.json", Vector2i(1, 1)));
		config.addAboveLevel(HomeConfig::LevelConfig("player/upper.json", Vector2i(1, 1)));
		config.addBelowLevel(HomeConfig::LevelConfig("player/basement.json", Vector2i(1, 1)));
		config.generate(town, rng);
	}
	else
	{
		switch (insideType)
		{
		case InsideType::HouseBasic:
			{
				HomeConfig config(Rect<int>(tx, ty, HomeConfig::AUTOMATIC, HomeConfig::AUTOMATIC), npcSettings);
				config.addAboveLevel(HomeConfig::LevelConfig("basic_3x2/above1_a.json", Vector2i(1, 1)));
				config.addAboveLevel(HomeConfig::LevelConfig("basic_3x2/above2_a.json", Vector2i(1, 1)));
				config.generate(town, rng);
			}
			break;
		case InsideType::HouseBiggerBasic:
			{
				HomeConfig config(Rect<int>(tx, ty, HomeConfig::AUTOMATIC, HomeConfig::AUTOMATIC), npcSettings);
				config.addAboveLevel(HomeConfig::LevelConfig("basic_4x3/above1_a.json", Vector2i(1, 1)));
				config.addAboveLevel(HomeConfig::LevelConfig("basic_4x3/above2_a.json", Vector2i(1, 1)));
				config.generate(town, rng);
			}
			break;
		
		case InsideType::HouseMiddleClassBasic:
			{
				HomeConfig config(Rect<int>(tx, ty, HomeConfig::AUTOMATIC, HomeConfig::AUTOMATIC), npcSettings);
				config.addAboveLevel(HomeConfig::LevelConfig("basic_6x4/above1_a.json", Vector2i(2, 1)));
				config.addAboveLevel(HomeConfig::LevelConfig("basic_6x4/above2_terminal_a.json", Vector2i(2, 1)));
				config.generate(town, rng);
			}
			break;
		case InsideType::HouseMiddleClassBasicTall:
			{
				HomeConfig config(Rect<int>(tx, ty, HomeConfig::AUTOMATIC, HomeConfig::AUTOMATIC), npcSettings);
				config.addAboveLevel(HomeConfig::LevelConfig("basic_6x4/above1_a.json", Vector2i(2, 1)));
				config.addAboveLevel(HomeConfig::LevelConfig("basic_6x4/above2_a.json", Vector2i(2, 1)));
				config.addAboveLevel(HomeConfig::LevelConfig("basic_6x4/above3_a.json", Vector2i(2, 1)));
				config.generate(town, rng);
			}
			break;
		case InsideType::School:
			{
				HomeConfig config(Rect<int>(tx, ty, HomeConfig::AUTOMATIC, HomeConfig::AUTOMATIC), npcSettings);
				config.addAboveLevel(HomeConfig::LevelConfig("school/ground.json", Vector2i(1, 1)));
				config.addAboveLevel(HomeConfig::LevelConfig("school/upper.json", Vector2i(1, 1)));
				config.generate(town, rng);
				const auto& entrancePos = housePos;
				LocationMarker *schoolMarker = new LocationMarker(entrancePos.x + 164, entrancePos.y + 148, "SchoolMarker", "school");
				schoolMarker->editIcon().setLabel("school");
				town.addEntityToQuad(schoolMarker);
			}
			break;
		case InsideType::SchoolSmall:
		{
			HomeConfig config(Rect<int>(tx, ty, HomeConfig::AUTOMATIC, HomeConfig::AUTOMATIC), npcSettings);
			config.addAboveLevel(HomeConfig::LevelConfig("school_small/ground.json", Vector2i(1, 1)));
			config.addAboveLevel(HomeConfig::LevelConfig("school_small/upper.json", Vector2i(1, 1)));
			config.generate(town, rng);
			const auto& entrancePos = housePos;
			LocationMarker *schoolMarker = new LocationMarker(entrancePos.x + 164, entrancePos.y + 148, "SchoolMarker", "school");
			schoolMarker->editIcon().setLabel("school");
			town.addEntityToQuad(schoolMarker);
		}
		break;
		case InsideType::CandyShop:
			{
				//groundFloor->loadCandyStore(rng);

				const auto& entrancePos = housePos;
				LocationMarker *schoolMarker = new LocationMarker(entrancePos.x, entrancePos.y + 64, "ShopMarker", "shop");
				schoolMarker->editIcon().setLabel("candy shop");
				town.addEntityToQuad(schoolMarker);
			}
			break;
		}
	}

	// ------------------------------------------------------------------------
	//             add on attachments to the sides if any exist
	for (int i = 0; i < 4; ++i)
	{
		const AttachedSide side = indexToSide(i);
		if (attachments[i].empty())
		{
			continue;
		}
		const bool isSide = (side == Left || side == Right);
		int budget = (isSide ? getHeight() : getWidth());

		// figure out which attachments will be used (by probability / limits)
		std::vector<int> chosenAttachments;
		// @todo stick some <Side, Index> pairs into a vector and permute them to get a more side-balanced random allowing with regard to limits
		for (int j = 0; j < attachments[i].size(); ++j)
		{
			const AttachmentConfig& config = attachments[i][j];
			if (rng.chance(config.probability) && (!config.limit || config.limit->use()))
			{
				chosenAttachments.push_back(j);
				if (isSide)
				{
					budget -= config.attachment->getHeight();
				}
				else
				{
					budget -= config.attachment->getHeight();
				}
			}
		}

		if (chosenAttachments.empty())
		{
			continue;
		}

		const int minSpacing = std::min(8.f, 0.25f * budget / (chosenAttachments.size() + 1.f));
		budget -= minSpacing * (chosenAttachments.size() + 1);

		// randomly distribute the free space between attachments
		std::vector<int> spacing(chosenAttachments.size() + 1, minSpacing);
		for (int j = 0; j < budget; ++j)
		{
			++spacing[rng.random(spacing.size())];
		}

		// randomly order the buildings
		std::vector<int> attachOrder(chosenAttachments.size());
		std::generate(attachOrder.begin(), attachOrder.end(), Increment<int>());
		rng.shuffle(attachOrder.begin(), attachOrder.end());

		int offsetSoFar = spacing[0];
		for (int j = 0; j < chosenAttachments.size(); ++j)
		{
			// j -> attach order (to permute the order via the random shuffled permutation
			//   -> chosenAttachments (so only chosen ones are hit)
			const BuildingAttachment& building = *attachments[i][chosenAttachments[attachOrder[j]]].attachment;
			Vector2i offset;
			switch (side)
			{
			case Left:
				offset.x = -building.getWidth();
				offset.y = offsetSoFar;
				break;
			case Right:
				offset.x = getWidth();
				offset.y = offsetSoFar;
				break;
			case Top:
				offset.x = offsetSoFar;
				offset.y = -building.getHeight();
				break;
			case Bottom:
				offset.x = offsetSoFar;
				offset.y = getHeight();
				break;
			}
			
			building.generate(housePos + offset * ts, rng, town, minimapStaticEntityData, isPlayerGenerated, npcSettings, forceSwap);

			offsetSoFar += spacing[j + 1];
		}
	}
}

int BuildingBlueprint::getWidth() const
{
	int width = 0;
	if (!solidGeo.empty())
	{
		int left = solidGeo[0].left;
		int right = solidGeo[0].left + solidGeo[0].width;
		for (int i = 1; i < solidGeo.size(); ++i)
		{
			if (solidGeo[i].left < left)
			{
				left = solidGeo[i].left;
			}
			if (solidGeo[i].left + solidGeo[i].width > right)
			{
				right = solidGeo[i].left + solidGeo[i].width;
			}
		}
		width = (right - left) / 16; // @HACK
	}
	return width + maxProtrusion[sideToIndex(Left)] + maxProtrusion[sideToIndex(Right)];
}

int BuildingBlueprint::getHeight() const
{
	int height = 0;
	if (!solidGeo.empty())
	{
		int top = solidGeo[0].top;
		int bottom = solidGeo[0].top + solidGeo[0].height;
		for (int i = 1; i < solidGeo.size(); ++i)
		{
			if (solidGeo[i].top < top)
			{
				top = solidGeo[i].top;
			}
			if (solidGeo[i].top + solidGeo[i].height > bottom)
			{
				bottom = solidGeo[i].top + solidGeo[i].height;
			}
		}
		height = (bottom - top) / 16; // @HACK
	}
	return height + maxProtrusion[sideToIndex(Top)] + maxProtrusion[sideToIndex(Bottom)];
}



/*                                 static                                    */
static void fillRegularPaintColors(std::vector<ColorSwapID>& swaps)
{
	const sf::Color normal(209, 208, 203);
	const sf::Color dark(185, 184, 176);
	const sf::Color outline(73, 72, 65);

	const ColorSwapID noSwap;
	const ColorSwapID yellowSwap = ColorSwapID({
		std::make_pair(normal, sf::Color(204, 201, 140)),
		std::make_pair(dark, sf::Color(165, 162, 88)),
		std::make_pair(outline, sf::Color(99, 97, 52))
	});
	const ColorSwapID greenSwap = ColorSwapID({
		std::make_pair(normal, sf::Color(148, 163, 146)),
		std::make_pair(dark, sf::Color(123, 141, 120)),
		std::make_pair(outline, sf::Color(48, 72, 54))
	});
	const ColorSwapID blueSwap = ColorSwapID({
		std::make_pair(normal, sf::Color(180, 202, 214)),
		std::make_pair(dark, sf::Color(121, 170, 196)),
		std::make_pair(outline, sf::Color(49, 75, 102))
	});

	swaps.push_back(noSwap);
	swaps.push_back(noSwap);
	swaps.push_back(noSwap);
	swaps.push_back(noSwap);
	swaps.push_back(yellowSwap);
	swaps.push_back(yellowSwap);
	swaps.push_back(yellowSwap);
	swaps.push_back(greenSwap);
	swaps.push_back(greenSwap);
	swaps.push_back(blueSwap);
}

BuildingBlueprint BuildingBlueprint::create(OutsideType type)
{
	switch (type)
	{
		case OutsideType::HouseBasic:
		{
			std::vector<ColorSwapID> swaps;
			fillRegularPaintColors(swaps);
			BuildingBlueprint building("buildings/3by2housebasic.png", Vector2i(0, 32), Vector2i(4, 3), InsideType::HouseBasic, std::move(swaps));
			building.addPathSpawn(Vector2i(2, 2));
			return building;
		}
		case OutsideType::HouseBiggerBasic:
		{
			std::vector<ColorSwapID> swaps;
			fillRegularPaintColors(swaps);
			BuildingBlueprint building("buildings/4by3housebasic.png", Vector2i(0, 32), Vector2i(5, 3), InsideType::HouseBiggerBasic, std::move(swaps));
			building.addPathSpawn(Vector2i(3, 2));
			return building;
		}
		case OutsideType::HouseMiddleClass:
		{
			std::vector<ColorSwapID> swaps;
			fillRegularPaintColors(swaps);
			BuildingBlueprint building("buildings/6by4housebasic.png", Vector2i(0, 32), Vector2i(7, 4), InsideType::HouseMiddleClassBasic, std::move(swaps));
			building.addPathSpawn(Vector2i(3, 3));
			return building;
		}
		case OutsideType::HouseMiddleClassTall:
		{
			std::vector<ColorSwapID> swaps;
			fillRegularPaintColors(swaps);
			BuildingBlueprint building("buildings/6by4housebasic_tall.png", Vector2i(0, 48), Vector2i(7, 4), InsideType::HouseMiddleClassBasicTall, std::move(swaps));
			building.addPathSpawn(Vector2i(3, 3));
			return building;
		}
		case OutsideType::School:
		{
			BuildingBlueprint building("buildings/school.png", Vector2i(9, 70), Vector2i(477, 104) / 16, InsideType::School);
			building.setDetailed({
				Rect<int>(8, 74, 478, 108), // main
				Rect<int>(8, 182, 146, 114), // left wing
				Rect<int>(340, 182, 146, 114), // right wing
				Rect<int>(218, 182, 59, 11) // entranceway
			});
			return building;
		}
		case OutsideType::SchoolSmall:
		{
			BuildingBlueprint building("buildings/school_small.png", Vector2i(0, 68), Vector2i(24, 10), InsideType::SchoolSmall);
			building.setDetailed({
				Rect<int>(0, 68, 384, 112), // main
				Rect<int>(240, 182, 144, 40) // right wing
			});
			return building;
		}
		case OutsideType::AbandonedHouse:
		{
			BuildingBlueprint building("buildings/abandonedhouse.png", Vector2i(0, 32), Vector2i(4, 3), InsideType::None);
			return building;
		}
		case OutsideType::SmallGarage:
		{
			BuildingBlueprint building("buildings/garage_small.png", Vector2i(4, 128 - 44), Vector2i(56, 44) / 16, InsideType::None);
			return building;
		}
		case OutsideType::OldApartment:
		{
			//--------------------  Paintjobs  --------------------
			const sf::Color light(148, 120, 114);
			const sf::Color normal(117, 92, 89);
			const sf::Color dark(91, 74, 68);
			const sf::Color outline(56, 44, 44);

			const ColorSwapID noSwap;
			const ColorSwapID yellowSwap = ColorSwapID({
				std::make_pair(light, sf::Color(148, 146, 114)),
				std::make_pair(normal, sf::Color(116, 114, 90)),
				std::make_pair(dark, sf::Color(90, 91, 68)),
				std::make_pair(outline, sf::Color(56, 54, 44))
			});

			std::vector<ColorSwapID> swaps;
			swaps.push_back(noSwap);
			swaps.push_back(noSwap);
			swaps.push_back(noSwap);
			swaps.push_back(yellowSwap);
			swaps.push_back(yellowSwap);

			//--------------------  Building  --------------------
			BuildingBlueprint building("buildings/oldapartment.png", Vector2i(3, 210), Vector2i(187, 109) / 16, InsideType::None, std::move(swaps));

			return building;
		}
		case OutsideType::CandyShop:
		{
			BuildingBlueprint building("buildings/candyshop.png", Vector2i(2, 128 - 45), Vector2i(126, 45) / 16, InsideType::CandyShop);
			return building;
		}
		default:
			ERROR("Implement this you fuck");
	}
	ERROR("Ya dun fucked up");
	return BuildingBlueprint("This is to get the compiler to stop bitching even though control flow can never reach here", Vector2i(-1, -1), Vector2i(-1, -1));
}

// private
int BuildingBlueprint::computeMaxAttachmentSize(AttachedSide side) const
{
	const int sideIndex = sideToIndex(side);
	if (side == Left || side == Right)
	{
		return std::accumulate(attachments[sideIndex].begin(), attachments[sideIndex].end(), 0, attachmentHeightAccumulator);
	}
	else
	{
		ASSERT(side == Top || side == Bottom);
		return std::accumulate(attachments[sideIndex].begin(), attachments[sideIndex].end(), 0, attachmentWidthAccumulator);
	}
}

/*static*/ int BuildingBlueprint::attachmentHeightAccumulator(int x, const AttachmentConfig& a)
{
	return x + a.attachment->getHeight();
}

/*static*/ int BuildingBlueprint::attachmentWidthAccumulator(int x, const AttachmentConfig& a)
{
	return x + a.attachment->getWidth();
}

/*static*/ int BuildingBlueprint::sideToIndex(AttachedSide side)
{
	switch (side)
	{
	case Left:
		return 0;
	case Right:
		return 1;
	case Top:
		return 2;
	case Bottom:
		return 3;
	}
	ERROR("invalid side");
	return -1;
}

/*static*/ BuildingBlueprint::AttachedSide BuildingBlueprint::indexToSide(int index)
{
	switch (index)
	{
	case 0:
		return Left;
	case 1:
		return Right;
	case 2:
		return Top;
	case 3:
		return Bottom;
	}
	ERROR("invalid index");
	return Left;
}

} // towngen
} // sl
