#ifndef SL_TOWNGEN_BUILDING_BLUEPRINT_HPP
#define SL_TOWNGEN_BUILDING_BLUEPRINT_HPP

#include <initializer_list>
#include <list>
#include <memory>
#include <string>
#include <vector>

#include "Engine/Graphics/ColorSwapID.hpp"
#include "Town/Generation/BuildingAttachment.hpp"
#include "Town/Generation/ConditionalPrototypeLimit.hpp"
#include "Utility/Rect.hpp"
#include "Utility/Vector.hpp"

namespace sl
{

namespace map
{
class MapStaticEntityData;
}

class Random;
class Town;

namespace towngen
{

class BuildingBlueprint : public BuildingAttachment
{
public:
	enum class InsideType
	{
		None,
		HouseBasic,
		HouseBiggerBasic,
		HouseMiddleClassBasic,
		HouseMiddleClassBasicTall,
		School,
		SchoolSmall,
		CandyShop
	};
	/**
	 * @param filename Filename of the image for the building
	 * @param imageOrigin Offset in PIXELS within the image for the (0, 0) (top left) of the building's physical portion
	 * @param size The size IN TILES of the physical part of the building
	 * @param insideType What kind of interior to generate for the building
	 * @param swaps Which palette swaps to apply to the building. If nullptr, it will default to some single choice. You'd mostly only override this for attachments that must share the same colour as the building they're attached to
	 */
	BuildingBlueprint(const std::string& filename, const Vector2i& imageOrigin, const Vector2i& size, InsideType insideType = InsideType::None, std::vector<ColorSwapID>&& swaps = std::vector<ColorSwapID>());

	BuildingBlueprint(BuildingBlueprint&& other) = default;

	BuildingBlueprint& operator=(BuildingBlueprint&& rhs) = default;

	/**
	 * Gets rid of the exisitng geometry and create based on a list of geometry.
	 * Note: for convenience, this is in terms of position in the IMAGE, not relative to the actual house.
	 * @params regions The list of the rectangular solid regions
	 */
	void setDetailed(std::initializer_list<Rect<int>> regions);

	enum AttachedSide
	{
		Top = 1,
		Bottom = 2,
		Left = 4,
		Right = 8
	};
	struct AttachmentConfig
	{
		AttachmentConfig(std::shared_ptr<BuildingAttachment> attachment, float probability = 1.f, ConditionalPrototypeLimit *limit = nullptr)
			:attachment(std::move(attachment))
			,probability(probability)
			,limit(limit)
		{
		}

		std::shared_ptr<BuildingAttachment> attachment;
		float probability;
		ConditionalPrototypeLimit *limit;
	};
	/**
	 * Attachs an attachment to the building on one of the sides listed
	 * @param attachment Configuration to attach (attachment/probability/limit/etc)
	 * @param side A mask of AttachedSides - will place one copy in each side listed
	 */
	void addAttachment(const AttachmentConfig& attachment, int side);

	/**
	 * @param pathStart The coordinates in tiles relative to the building to place a path gen spawn at
	 */
	void addPathSpawn(const Vector2i& pathStart);

	/**
	 * @return The coordinates in tiles relative to the building to start path generation from
	 */
	const std::vector<Vector2i>& getPathSpawns() const;

	ConditionalPrototypeLimit* createNewLimit(int count);

	void generate(const Vector2i& pos, Random& rng, Town& town, map::MapStaticEntityData& minimapStaticEntityData, bool& isPlayerGenerated, const SoulGenerationSettings& npcSettings, const ColorSwapID *forceSwap = 0) const override;

	/**
	 * @return Width IN TILES of the bounding box that bounds the building and all its attachments
	 */
	int getWidth() const override;

	/**
	* @return Height IN TILES of the bounding box that bounds the building and all its attachments
	*/
	int getHeight() const override;



	enum class OutsideType
	{
		// suburban
		HouseBasic,
		HouseBiggerBasic,
		HouseMiddleClass,
		HouseMiddleClassTall,
		HouseRich,
		School,
		SchoolSmall,
		AbandonedHouse,

		// attachments
		SmallGarage,

		// urban
		OldApartment,
		LoliCo,
		HomeGrown,
		CandyShop
	};
	static BuildingBlueprint create(OutsideType type);

private:

	int computeMaxAttachmentSize(AttachedSide side) const;

	static int attachmentHeightAccumulator(int x, const AttachmentConfig& a);

	static int attachmentWidthAccumulator(int x, const AttachmentConfig& a);

	static int sideToIndex(AttachedSide side);

	static AttachedSide indexToSide(int index);

	BuildingBlueprint(const BuildingBlueprint&);
	BuildingBlueprint& operator=(const BuildingBlueprint&);



	InsideType insideType;
	std::string filename;
	std::vector<ColorSwapID> swapIDs;
	Vector2i imageOrigin;
	std::vector<Rect<int>> solidGeo;
	std::vector<AttachmentConfig> attachments[4];
	int maxProtrusion[4];
	mutable std::list<ConditionalPrototypeLimit> limits;
	std::vector<Vector2i> pathSpawns;
};

} // towngen
} // sl

#endif // SL_TOWNGEN_BUILDING_BLUEPRINT_HPP
