/**
 * @section DESCRIPTION
 * A generic content generator for town generation. It occupies a region and generates things within that region.
 */
#ifndef SL_TOWNGEN_CONTENTGENERATOR_HPP
#define SL_TOWNGEN_CONTENTGENERATOR_HPP

#include "Town/TownTiles.hpp"
#include "Utility/Grid.hpp"
#include "Utility/Rect.hpp"
#include "Utility/Vector.hpp"

namespace sl
{


class Random;
class Town;

namespace map
{
class MapStaticEntityData;
} // map

namespace towngen
{

class ContentGenerator
{
public:
	virtual ~ContentGenerator() {}

	/**
	 * Generates some kind of content at the specified position
	 * @param pos The position in tiles to generate the content at (top left)
	 * @param rng The random number generator to use in generation if necessary
	 * @param town The town to create objects in
	 * @param tiles The tiles of the town in case your generator places tiles
	 * @param tileSize The tile size in pixels of each tile
	 * @param minimapStaticEntityData The minimap to update if any non-tile static modifications need to be done to reflect the changes
	 */
	virtual void generate(const Vector2i& pos, Random& rng, Town& town, Grid<TileType>& tiles, int tileSize, map::MapStaticEntityData& minimapStaticEntityData) const;

	/**
	 * Generates some kind of content in the specified area. Default behaviour is to place the content in the position function 
	 * @param area The region in tiles in the town to generate in
	 * @param rng The random number generator to use in generation if necessary
	 * @param town The town to create objects in
	 * @param tiles The tiles of the town in case your generator places tiles
	 * @param tileSize The tile size in pixels of each tile
	 * @param minimapStaticEntityData The minimap to update if any non-tile static modifications need to be done to reflect the changes
	 */
	virtual void generate(const Rect<int>& region, Random& rng, Town& town, Grid<TileType>& tiles, int tileSize, map::MapStaticEntityData& minimapStaticEntityData) const;


	/**
	 * @return minimum width in tiles of the region this will generate (VARIABLE_WIDTH by default)
	 */
	virtual int getWidth() const;

	/**
	 * @return minimum height in tiles of the region this will generate (VARIABLE_HEIGHT by default)
	 */
	virtual int getHeight() const;

	bool isFixedSize() const;

	static const int VARIABLE_WIDTH = -1;
	static const int VARIABLE_HEIGHT = -1;
};

} // towngen
} // sl

#endif // SL_TOWNGEN_CONTENTGENERATOR_HPP