#include "Town/Generation/Interiors/HomeConfig.hpp"

#include <initializer_list>
#include <fstream>
#include <map>

#include "Engine/Door.hpp"
#include "Engine/Game.hpp"
#include "Town/World.hpp"
#include "Engine/Graphics/Sprite.hpp"
#include "Home/BuildingLevel.hpp"
#include "Home/Furniture.hpp"
#include "NPC/Stealth/SuspiciousLocations.hpp"
#include "NPC/FamilyGen.hpp"
#include "NPC/Region.hpp"
#include "Town/StaticGeometry.hpp"
#include "Town/Town.hpp"
#include "Utility/Memory.hpp"
#include "Utility/JsonUtil.hpp"

namespace sl
{
namespace towngen
{

const int HomeConfig::AUTOMATIC = -1;
// HomeConfig::LevelConfig
HomeConfig::LevelConfig::LevelConfig(std::string name, const Vector2i& tileOffset)
	:name(std::move(name))
	,tileOffset(tileOffset)
{
}

bool HomeConfig::LevelConfig::operator<(const LevelConfig& rhs) const
{
	if (name == rhs.name)
	{
		if (tileOffset.x == rhs.tileOffset.x)
		{
			return tileOffset.y < rhs.tileOffset.y;
		}
		else
		{
			return tileOffset.x < rhs.tileOffset.x;
		}
	}
	else
	{
		return name < rhs.name;
	}
}



// HomeConfig
HomeConfig::HomeConfig(const Rect<int>& worldBounds, const SoulGenerationSettings& familySettings)
	:worldBounds(worldBounds)
	,familySettings(familySettings)
{
}


void HomeConfig::addAboveLevel(LevelConfig&& config)
{
	above.push_back(config);
}

void HomeConfig::addBelowLevel(LevelConfig&& config)
{
	below.push_back(config);
}

void HomeConfig::generate(Town& town, Random& rng) const
{
	// @todo move this cache somewhere else?
	static std::map<LevelConfig, json11::Json> cachedJson;

	FamilyGen family(familySettings, *town.getWorld(), rng);

	std::vector<std::map<LevelConfig, json11::Json>::const_iterator> json;
	auto readJson = [&](const std::vector<LevelConfig>& configs)
	{
		for (const LevelConfig& config : configs)
		{
			std::map<LevelConfig, json11::Json>::const_iterator it = cachedJson.find(config);
			if (it == cachedJson.cend())
			{
				std::string err;
				json11::Json data = readJsonFile("data/gendata/homes/" + config.name, err);
				if (err.empty())
				{
					it = cachedJson.insert(std::make_pair(config, std::move(data))).first;
				}
				else
				{
					ERROR(err);
				}
			}
			json.push_back(it);
		}
	};
	readJson(below);
	readJson(above);

	Rect<int> worldBounds(this->worldBounds);
	if (worldBounds.width == AUTOMATIC || worldBounds.height == AUTOMATIC)
	{
		const int UNKNOWN = -2;
		int biggestKnownWidth = UNKNOWN;
		int biggestKnownHeight = UNKNOWN;
		for (const auto it : json)
		{
			const LevelConfig& config = it->first;
			const json11::Json& levelJson = it->second;
			const int tw = levelJson.object_items().at("walkableWidth").int_value();
			const int th = levelJson.object_items().at("walkableHeight").int_value();
			// assume that there's an equal amount of spacing on left as right and same for top/bottom
			const int width = config.tileOffset.x * 2 + tw;
			const int height = config.tileOffset.y * 2 + th;
			if (biggestKnownWidth == UNKNOWN || biggestKnownWidth < width)
			{
				biggestKnownWidth = width;
			}
			if (biggestKnownHeight == UNKNOWN || biggestKnownHeight < height)
			{
				biggestKnownHeight = height;
			}
		}
		if (worldBounds.width == AUTOMATIC)
		{
			worldBounds.width = biggestKnownWidth;
		}
		if (worldBounds.height == AUTOMATIC)
		{
			worldBounds.height = biggestKnownHeight;
		}
	}

	std::vector<std::unique_ptr<BuildingLevel>> levels;
	for (const auto it : json)
	{
		const float buildingScale = 0.5f;
		const LevelConfig& config = it->first;
		const json11::Json& levelJson = it->second;
		const int width = levelJson.object_items().at("width").int_value();
		const int height = levelJson.object_items().at("height").int_value();
		const int xOff = levelJson.object_items().at("xoff").int_value();
		const int yOff = levelJson.object_items().at("yoff").int_value();
		const int worldXoff = worldBounds.left + (config.tileOffset.x - xOff) * buildingScale;
		const int worldYoff = worldBounds.top + (config.tileOffset.y - yOff) * buildingScale;
		
		const int ts = town.getTileSize();
		const Rect<int> sceneBounds(xOff - config.tileOffset.x, yOff - config.tileOffset.y, worldBounds.width, worldBounds.height);
		levels.push_back(unique<BuildingLevel>(town.getGame(), width * ts, height * ts, worldXoff, worldYoff, sceneBounds, buildingScale, false));
		// HAAAAAAAACK! but I can't think of a good way to create the dungeon before any other things are spawned/processed
		// so the dungeonedit object in the editor is down the drain...
		if (it->first.name == "player/basement.json")
		{
			levels.back()->createDungeon(26, 22);
		}
	}
	auto level = levels.begin();
	BuildingLevel *lastLevel = nullptr;
	for (const auto it : json)
	{
		BuildingLevel *nextLevel = (level + 1) != levels.end() ? (level + 1)->get() : nullptr;
		(*level)->setConnections(lastLevel, nextLevel, &town);
		// this NEEDs to be called before addObject, or world will be null for the scene!
		BuildingLevel *curLevel = level->get();
		town.getWorld()->addScene(std::move(*level), true);
		const json11::Json& levelJson = it->second;
		for (const auto& rootdata : levelJson.object_items())
		{
			if (rootdata.first == "objects")
			{
				for (const json11::Json& object : rootdata.second.array_items())
				{
					const json11::Json::object& vars = object.object_items();
					const int x = vars.at("x").int_value();
					const int y = vars.at("y").int_value();
					const std::string& type = vars.at("type").string_value();
					curLevel->addObject(Vector2i(x, y), type, vars.count("attributes") ? &vars.at("attributes").object_items() : nullptr, &family);
				}
			}
			else if (rootdata.first == "tiles")
			{
				for (const json11::Json& tile : rootdata.second.array_items())
				{
					const json11::Json::object& vars = tile.object_items();
					const int x = vars.at("x").int_value();
					const int y = vars.at("y").int_value();
					const std::string& id = vars.at("id").string_value();
					curLevel->setTile(Vector2i(x, y), id);
				}
			}
			else if (rootdata.first == "walls")
			{
				for (const json11::Json& wall : rootdata.second.array_items())
				{
					const json11::Json::object& vars = wall.object_items();
					//@todo handle id
					curLevel->addWall(Vector2i(vars.at("x").int_value(), vars.at("y").int_value()), vars.at("id").string_value());
				}
			}
			else if (rootdata.first == "regions")
			{
				for (const json11::Json& region : rootdata.second.array_items())
				{
					const json11::Json::object& vars = region.object_items();
					const std::string& type = vars.at("type").string_value();
					Region *regionEntity = new Region(parseRect(vars), type);
					regionEntity->setSuspiciousScenario(SuspiciousLocations::byKey(type));
					curLevel->addEntityToQuad(regionEntity);
				}
			}
			else if (rootdata.first == "width" || rootdata.first == "height" || rootdata.first == "xoff" || rootdata.first == "yoff" || rootdata.first == "walkableWidth" || rootdata.first == "walkableHeight")
			{
				// these are checked earlier
			}
			else
			{
				ERROR("undefined data");
			}
		}
		curLevel->createPrecomputedPathgrid();
		lastLevel = curLevel;
		++level;
	}
}


// private
void loadJson()
{
}

} // towngen
} // sl
