#include "Town/Generation/Interiors/HomeObjectDatabase.hpp"

#include <fstream>

#include "Utility/Assert.hpp"
#include "Utility/JsonUtil.hpp"

namespace sl
{
namespace towngen
{

void HomeObjectDatabase::reload()
{
	const std::string objectFile("data/gendata/homeobjects.json");
	std::string err;
	json11::Json data = readJsonFile(objectFile, err);
	if (err.empty())
	{
		for (const json11::Json& object : data.array_items())
		{
			const json11::Json::object& vars = object.object_items();
			const std::string& type = vars.at("type").string_value();
			HomeObjectPrototype::CustomCode customCode; // @todo add in custom code for things like computers, etc that can be interacted with
			std::map<std::string, std::string> attributes;
			if (vars.count("attributes") != 0)
			{
				for (const auto& attribute : vars.at("attributes").object_items())
				{
					attributes.insert(std::make_pair(attribute.first, attribute.second.string_value()));
				}
			}
			std::vector<HeightRect> heightRects;
			auto it = vars.find("heightRects");
			if (it != vars.end())
			{
				for (const json11::Json& rect : it->second.array_items())
				{
					const json11::Json::object& rectVars = rect.object_items();
					heightRects.push_back(HeightRect(parseRect(rectVars), rectVars.at("topleft").int_value(), rectVars.at("topright").int_value(), rectVars.at("botleft").int_value(), rectVars.at("botright").int_value()));
				}
			}
			std::vector<Rect<int>> geometry;
			it = vars.find("geometry");
			if (it != vars.end())
			{
				for (const json11::Json& rect : it->second.array_items())
				{
					geometry.push_back(parseRect(rect.object_items()));
				}
			}
			std::vector<HomeObjectPrototype::ImageInfo> images;
			it = vars.find("images");
			if (it != vars.end())
			{
				for (const json11::Json& image : it->second.array_items())
				{
					const json11::Json::object& imageVars = image.object_items();
					auto bottom = imageVars.find("bottomOffset");
					const int bottomOffset = bottom == imageVars.end() ? 0 : bottom->second.int_value();
					images.push_back(HomeObjectPrototype::ImageInfo(imageVars.at("file").string_value(), Vector2i(imageVars.at("iofx").int_value(), imageVars.at("iofy").int_value()), bottomOffset));
				}
			}
			Rect<int> tileArea = parseRect(vars.at("tileArea").object_items());
			if (!prototypes.insert(std::make_pair(type, HomeObjectPrototype(std::move(type), std::move(images), std::move(geometry), customCode, std::move(attributes), std::move(heightRects), tileArea))).second)
			{
				ERROR("duplicate definition for " + type);
			}
		}
	}
	else
	{
		ERROR("error opening or parsing " + objectFile + ":  " + err);
	}
}

const HomeObjectPrototype* HomeObjectDatabase::getPrototype(const std::string& id) const
{
	auto it = prototypes.find(id);
	if (it != prototypes.end())
	{
		return &it->second;
	}
	return nullptr;
}

HomeObjectDatabase::Data::const_iterator HomeObjectDatabase::begin() const
{
	return prototypes.cbegin();
}

HomeObjectDatabase::Data::const_iterator HomeObjectDatabase::end() const
{
	return prototypes.cend();
}

int HomeObjectDatabase::size() const
{
	return prototypes.size();
}



/*static*/ HomeObjectDatabase& HomeObjectDatabase::get()
{
	static HomeObjectDatabase database;
	return database;
}

HomeObjectDatabase::HomeObjectDatabase()
{
	reload();
}

} // towngen
} // sl
