#include "Town/Generation/Interiors/HomeObjectPrototype.hpp"

#include <algorithm>

#include "Engine/Door.hpp"
#include "Engine/Game.hpp"
#include "Town/World.hpp"
#include "Engine/Graphics/DrawableCollection.hpp"
#include "Engine/Graphics/Sprite.hpp"
#include "Home/BuildingLevel.hpp"
#include "Home/Internet/Computer.hpp"
#include "Home/Dungeon.hpp"
#include "Home/Furniture.hpp"
#include "Home/OpenableDoor.hpp"
#include "Home/PlayerBed.hpp"
#include "NPC/Bed.hpp"
#include "NPC/FamilyGen.hpp"
#include "NPC/Locators/FixedTargetLocator.hpp"
#include "NPC/NamedEntity.hpp"
#include "NPC/NPC.hpp"
#include "NPC/States/LoliRoutineState.hpp"
#include "NPC/States/AdultWanderState.hpp"
#include "Engine/Scene.hpp"
#include "Town/Town.hpp"
#include "Town/StaticGeometry.hpp"
#include "Town/Generation/Interiors/EditingUndo.hpp"
#include "Town/Generation/Interiors/HomeObjectPrototypeInstance.hpp"
#include "Utility/Memory.hpp"

// remove
#include "Engine/Graphics/SFDrawable.hpp"
#include <SFML/Graphics/RectangleShape.hpp>
namespace sl
{
namespace towngen
{

HomeObjectPrototype::HomeObjectPrototype(std::string type, std::vector<ImageInfo> images, std::vector<Rect<int>> geometry, const CustomCode& customCode, std::map<std::string, std::string> attributes, std::vector<HeightRect> heightRects, const Rect<int>& tileArea)
	:type(std::move(type))
	,images(std::move(images))
	,geometry(std::move(geometry))
	,customCode(customCode)
	,attributes(std::move(attributes))
	,heightRects(std::move(heightRects))
	,tileArea(tileArea)
{
}

std::unique_ptr<EditingUndo> HomeObjectPrototype::generate(Random& rng, Scene& scene, const Vector2i& pos, Town *town, BuildingLevel *below, BuildingLevel *above, const json11::Json::object *instanceAttributes, FamilyGen *family) const
{
	const bool isInEditor = (town == nullptr);
	std::unique_ptr<EditingUndo> undo = unique<EditingUndo>(HomeObjectPrototypeInstance(pos.x, pos.y, type), undoRect(pos));
	StaticGeometry *geo = nullptr;
	for (const Rect<int>& collisionRect : geometry)
	{
		geo = new StaticGeometry(pos.x + collisionRect.left, pos.y + collisionRect.top, collisionRect.width, collisionRect.height, nullptr);
		undo->addEntityToUndo(geo);
		scene.addImmediately(geo, StorageType::InList);
		if (customCode)
		{
			customCode(rng, geo, *undo);
		}
	}
	// if we placed geo already, add all sprites to last one, else create one just for drawing sprites
	if (geo)
	{
		// since geo's (x, y) might not have been (0, 0) we need to manually offset it
		// geo->getPos() - pos = geometry's position
		geo->addDrawable(createPreview(1.f, geo->getPos().to<int>() - pos));
	}
	else
	{
		geo = new StaticGeometry(pos.x, pos.y, 0, 0, createPreview());
		undo->addEntityToUndo(geo);
		scene.addImmediately(geo, StorageType::InList);
		if (customCode)
		{
			customCode(rng, geo, *undo);
		}
	}
	for (const HeightRect& heightRect : heightRects)
	{
		// stored in a QuadTree for auto memory managment
		HeightRect *rect = new HeightRect(heightRect);
		rect->setPos(pos + rect->getBounds().topLeft());
		scene.addHeightMap(rect);
		undo->addHeightMapToUndo(rect);
	}

	// for non-generic behaviour for interactable things that need custom code
	if (!attributes.empty())
	{
		const std::string& specialType = attributes.at("type");
		if (specialType == "door")
		{
			auto scenePosInWorld = [](const Scene& s, const Vector2f& p) -> Vector2f
			{
				return Vector2f(s.getWorldCoords().x, s.getWorldCoords().y) * static_cast<float>(s.getTileSize()) + (p * s.getWorldScale());
			};
			const Vector2f posf(pos.x, pos.y);
			const Vector2f posInWorld = scenePosInWorld(scene, posf);
			const std::string& dir = attributes.at("dir");
			const std::string& doorType = attributes.at("doortype");
			if (doorType == "out")
			{
				if (!isInEditor)
				{
					const Vector2f targetInWorld = scenePosInWorld(*town, Vector2f(0.f, 0.f));
					const Vector2f targetPos = (posInWorld - targetInWorld) * town->getWorldScale();
					const int xOverride = instanceAttributes && instanceAttributes->count("x_override") ? instanceAttributes->at("x_override").int_value() : 0;
					const int yOverride = instanceAttributes && instanceAttributes->count("y_override") ? instanceAttributes->at("y_override").int_value() : 0;
					//@todo handle other door orientations here if ever implemented
					const int ts = scene.getTileSize();
					auto toGrid = [ts](Vector2i p) -> Vector2f
					{
						return (ts * (p / ts)).to<float>();
					};
					auto toGridCentered = [ts](Vector2i p) -> Vector2f
					{
						return (ts * (p / ts) + Vector2i(ts / 2, ts / 2)).to<float>();
					};
					const Vector2f interiorPos = toGrid(pos);
					const Vector2f interiorTarget = toGridCentered(targetPos.to<int>()) + Vector2f(xOverride, yOverride);
					scene.addEntityToList(
						new Door(
							interiorPos,
							ts,
							ts / 4,
							interiorTarget,
							town,
							pf::NodeID(interiorPos.to<int>() / scene.getTileSize() + Vector2i(0, -1), &scene),
							pf::NodeID(interiorTarget.to<int>() / town->getTileSize(), town)));
					const Vector2f exteriorPos = toGrid(targetPos.to<int>()) + Vector2f(xOverride, yOverride - ts / 4);
					const Vector2f exteriorTarget = toGridCentered(Vector2i(pos.x, pos.y - ts / 2));
					town->addEntityToList(
						new Door(
							exteriorPos,
							ts,
							ts / 4,
							exteriorTarget,
							&scene,
							pf::NodeID(exteriorPos.to<int>() / town->getTileSize() + Vector2i(0, 1), town),
							pf::NodeID(exteriorTarget.to<int>() / scene.getTileSize(), &scene)));
					//town->addEntityToQuad(new StaticGeometry(toGrid(targetPos.x), toGrid(targetPos.y), 0, 0, unique<Sprite>("door.png", 8, 32)));
					
					// DEBUG STUFF
					//const BuildingLevel& building = static_cast<const BuildingLevel&>(scene);
					//const int whyx = building.getWorldspaceRect().left * town->getTileSize();
					//const int whyy = building.getWorldspaceRect().top * town->getTileSize();
					//const int whyw = building.getWorldspaceRect().width * town->getTileSize();
					//const int whyh = building.getWorldspaceRect().height * town->getTileSize();
					//std::unique_ptr<SFDrawable<sf::RectangleShape>> houserect(new SFDrawable<sf::RectangleShape>(sf::RectangleShape(sf::Vector2f(whyw, whyh))));
					//houserect->getData().setFillColor(sf::Color(255, 0, 0, 64));
					//houserect->getData().setOutlineColor(sf::Color(255, 0, 0));
					//houserect->getData().setOutlineThickness(1.f);
					//town->addEntityToList(new StaticGeometry(whyx, whyy, 0, 0, std::move(houserect)));

					//std::unique_ptr<SFDrawable<sf::RectangleShape>> scenerec(new SFDrawable<sf::RectangleShape>(sf::RectangleShape(sf::Vector2f(scene.getWidth(), scene.getHeight()))));
					//scenerec->getData().setFillColor(sf::Color(0, 255, 0, 64));
					//scenerec->getData().setOutlineColor(sf::Color(0, 128, 0));
					//scenerec->getData().setOutlineThickness(1.f);
					//const int whyx2 = scene.getWorldCoords().x * scene.getTileSize();
					//const int whyy2 = scene.getWorldCoords().y * scene.getTileSize();
					//town->addEntityToList(new StaticGeometry(whyx2, whyy2, 0, 0, std::move(scenerec)));
					// END DEBUG STUFF
				}
			}
			else
			{
				const bool up = doorType == "up";
				Scene *target = up ? above : below;
				ASSERT(up || doorType == "down");
				if (target)
				{
					//@todo handle other door orientations
					const Vector2f targetInWorld = scenePosInWorld(*target, Vector2f(0.f, 0.f));
					const Vector2f targetPos = (posInWorld - targetInWorld) / target->getWorldScale();
					if (up)
					{
						createStairsUpLeft(posf + Vector2f(16.f, 0.f), scene, targetPos + Vector2f(8, 10), *target);
					}
					else
					{
						createStairsDownLeft(posf, scene, targetPos + Vector2f(6, 10), *target);
					}
				}
			}
		}
		else if (specialType == "bed"  && family)
		{
			// allow overriding via instanceAttribrutes
			const std::string& owner = (instanceAttributes && instanceAttributes->count("spawn")) ?
				instanceAttributes->at("spawn").string_value() :
				attributes.at("spawn");
			Bed *bed = new Bed(Rect<int>(pos.x, pos.y, 16, 16)); // TODO: make it read based on the actual size
			scene.addEntityToList(bed);
			auto createNPC = [&](std::shared_ptr<Soul> soul)
			{
				NPC *npc = new NPC(bed->getPos().x + 12, bed->getPos().y + 12, soul);
				if (soul->getType() == "loli")
				{
					npc->initStates(unique<LoliRoutineState>(*npc));

					scene.getWorld()->getSchoolRegistry().registerNPC(*npc);
				}
				else
				{
					npc->initStates(unique<AdultWanderState>(*npc));
				}
				soul->setLocation("bedroom", unique<FixedTargetLocator>(Target(bed->getPos().x, bed->getPos().y, &scene)));
				soul->setKnownEntity("bed", bed);
				scene.addImmediately(npc, StorageType::InQuad);
			};
			if (owner == "daughter")
			{
				static int debugLoliLimit = 0;
				if (debugLoliLimit++ < INT_MAX)
				{
					createNPC(family->createChild(4*365, 14*365, 0.f));
				}
			}
			else if (owner == "parents")
			{
				createNPC(family->getMother());
				createNPC(family->getFather());
			}
			else if (owner == "parent")
			{
				createNPC(family->getParent());
			}
			else if (owner == "none")
			{
				// do nothing - bed has no owner
			}
			else if (owner == "player")
			{
				PlayerBed *playerBed = new PlayerBed(pos.x, pos.y);
				undo->addEntityToUndo(playerBed);
				scene.addEntityToList(playerBed);
			}
			else
			{
				ERROR("Unknown bed owner type: " + owner);
			}
		}
		else if (specialType == "celldoor")
		{
			Dungeon *dungeon = static_cast<BuildingLevel&>(scene).getDungeon();
			ASSERT(dungeon != nullptr);
			if (dungeon)
			{
				dungeon->placeDoor(pos.x, pos.y, RoomDoor::Orientation::South);
				undo->addUndoFunc([pos, dungeon](){
					dungeon->removeDoorAt(pos.x, pos.y);
				});
			}
		}
		else if (specialType == "sliding_bookcase")
		{
			if (town)
			{
				geo->destroy();
				OpenableDoor *slidingBookcase = new OpenableDoor(pos.x, pos.y, OpenableDoor::DoorType::SlidingBookcase);
				undo->addEntityToUndo(slidingBookcase);
				scene.addEntityToList(slidingBookcase);
			}
		}
		else if (specialType == "computer")
		{
			Computer *computer = new Computer(pos.x, pos.y);
			undo->addEntityToUndo(computer);
			scene.addEntityToList(computer);
		}
		else if (specialType == "toilet")
		{
			NamedEntity::tag("Toilet", geo, Vector2f(8.f, 24.f));
		}
		else if (specialType == "school_desk")
		{
			if (!isInEditor)
			{
				scene.getWorld()->getSchoolRegistry().registerDesk(geo);
			}
		}
	}
	return std::move(undo);
}

std::unique_ptr<Drawable> HomeObjectPrototype::createPreview(float alpha, const Vector2i& extraOffset) const
{
	auto drawable = unique<DrawableCollection>();
	for (const ImageInfo& info : images)
	{
		auto sprite = unique<Sprite>(info.imageName.c_str(), info.imageOffset.x + extraOffset.x, info.imageOffset.y + extraOffset.y);
		sprite->setBottomPointOffset(info.bottomOffset);
		// need to update it once since otherwise the sprite is never updated, since StaticGeometry is never updated!
		//sprite->update(0, 0);//info.imageOffset.x, info.imageOffset.y);
		ASSERT(alpha >= 0.f && alpha <= 1.f);
		if (alpha < 1.f)
		{
			sprite->setAlpha(alpha);
		}
		drawable->addDrawable(std::move(sprite));
	}
	return std::move(drawable);
}

const std::string& HomeObjectPrototype::getAttribute(const std::string& name) const
{
	return attributes.at(name);
}

Rect<int> HomeObjectPrototype::undoRect(const Vector2i& pos) const
{
	// use the bounding box of all geometry for the undo area (or maybe we should use the bounding box of the graphics?)
	int left = 0;
	int right = 0;
	int top = 0;
	int bottom = 0;
	for (const Rect<int>& collisionRect : geometry)
	{
		top = std::min(top, collisionRect.top);
		bottom = std::max(bottom, collisionRect.top + collisionRect.height);
		left = std::min(left, collisionRect.left);
		right = std::max(right, collisionRect.left + collisionRect.width);
	}
	constexpr int minUndoSize = 4;
	if (right - left < minUndoSize)
	{
		right = left + minUndoSize;
	}
	if (bottom - top < minUndoSize)
	{
		bottom = top + minUndoSize;
	}
	return Rect<int>(pos.x + left, pos.y + top, right - left, bottom - top);
}

const Rect<int>& HomeObjectPrototype::getTileArea() const
{
	return tileArea;
}

} // towngen
} // sl
