#ifndef SL_TOWNGEN_HOMEOBJECTPROTOTYPE_HPP
#define SL_TOWNGEN_HOMEOBJECTPROTOTYPE_HPP

#include <json11/json11.hpp>

#include <functional>
#include <map>
#include <memory>
#include <string>
#include <vector>

#include "Engine/Physics/HeightRect.hpp"
#include "Utility/Rect.hpp"
#include "Utility/Vector.hpp"

namespace sl
{

class BuildingLevel;
class Drawable;
class FamilyGen;
class Scene;
class Random;
class Sprite;
class StaticGeometry;
class Town;

namespace towngen
{

class EditingUndo;

class HomeObjectPrototype
{
public:
	typedef std::function<void(Random&, StaticGeometry*, EditingUndo&)> CustomCode;
	
	class ImageInfo
	{
	public:
		ImageInfo(std::string imageName, Vector2i imageOffset, int bottomOffset)
			:imageName(std::move(imageName))
			,imageOffset(imageOffset)
			,bottomOffset(bottomOffset)
		{
		}

		std::string imageName;
		Vector2i imageOffset;
		int bottomOffset;
	};

	HomeObjectPrototype(std::string type, std::vector<ImageInfo> images, std::vector<Rect<int>> geometry, const CustomCode& customCode, std::map<std::string, std::string> attributes, std::vector<HeightRect> heightRects, const Rect<int>& tileArea);

	/**
	 * Instanties the object into the world
	 * @param rng Random number generator
	 * @param scene The scene to instantiate the object into
	 * @param pos The position in scene to place it at
	 * @param town The town the scene is connected to
	 * @param below The floor below us. Use nullptr if this is the bottom level
	 * @param above The floor above us. Use nullptr if this is the top level
	 * @param instanceAttributes An optional map of attributers specific to the instance. An example overriding town position for exterior doors
	 * @param family Info about the family if placing an object (ie bed) should lead to family creation. Use nullptr if you don't want NPCs created (ie HomeEditor)
	 * @return An Entity allowing you to easily undo all effects of this function. Must be manually added to the world
	 */
	std::unique_ptr<EditingUndo> generate(Random& rng, Scene& scene, const Vector2i& pos, Town *town, BuildingLevel *below, BuildingLevel *above, const json11::Json::object *instanceAttributes, FamilyGen *family) const;

	std::unique_ptr<Drawable> createPreview(float alpha = 1.f, const Vector2i& extraOffset = Vector2i(0, 0)) const;

	const std::string& getAttribute(const std::string& name) const;

	/**
	 * @param pos The position in the scene it would be placed at
	 * @return The EditingUndo area that would be taken up by this object if it were spawned
	 */
	Rect<int> undoRect(const Vector2i& pos = Vector2i(0, 0)) const;

	const Rect<int>& getTileArea() const;


private:
	
	std::string type;
	std::vector<ImageInfo> images;
	std::vector<Rect<int>> geometry;	
	CustomCode customCode;
	std::map<std::string, std::string> attributes;
	std::vector<HeightRect> heightRects;
	Rect<int> tileArea;
};

} // towngen
} // sl

#endif // SL_TOWNGEN_HOMEOBJECTPROTOTYPE_HPP
