#include "Town/Generation/ParkPathGen.hpp"

#include "Town/Generation/PathGenUtil.hpp"
#include "Utility/Random.hpp"

namespace sl
{
namespace towngen
{

ParkPathGen::ParkPathGen(Rect<int> bounds, Random& rng)
	:bounds(bounds)
	,pathGrid(bounds.width + 2, bounds.height + 2, false)
	,rng(rng)
{
	for (int x = 1; x < bounds.width; ++x)
	{
		pathGrid.at(x, 2) = true;
		pathGrid.at(x, bounds.height - 1) = true;
	}
	for (int y = 2; y < bounds.height - 1; ++y)
	{
		pathGrid.at(2, y) = true;
		pathGrid.at(bounds.width - 1, y) = true;
	}
}


void ParkPathGen::walk(DirectionalTendency dirTend, Direction dir, int x, int y)
{
	walkers.push_back(Walker(pathGrid, rng, dirTend, dir, x, y));
}

void ParkPathGen::sprawlFromFountain(int x, int y)
{

}

void ParkPathGen::generate()
{
	while (!walkers.empty())
	{
		for (int i = 0; i < walkers.size(); ++i)
		{
			// kill all dead walkers and remove from queue
			while (i < walkers.size() && !walkers[i].isAlive())
			{
				walkers[i] = walkers.back();
				walkers.pop_back();
			}

			if (i < walkers.size())
			{
				walkers[i].walk();
			}
		}
	}
	cleanUpClumps(pathGrid);
}

Vector2i ParkPathGen::addEntrance()
{
	const int entranceX = rng.randomInclusive(2, bounds.width - 1);
	for (int y = pathGrid.getHeight() - 1; y >= 0; --y)
	{
		if (pathGrid.at(entranceX, y))
		{
			break;
		}
		pathGrid.at(entranceX, y) = true;
	}
	return Vector2i(bounds.left + entranceX, bounds.top + pathGrid.getHeight() - 2);
}

Vector2i ParkPathGen::addTopEntrance()
{
	const int entranceX = rng.randomInclusive(2, bounds.width - 1);
	for (int y = 1; y < pathGrid.getHeight(); ++y)
	{
		if (pathGrid.at(entranceX, y))
		{
			break;
		}
		pathGrid.at(entranceX, y) = true;
	}
	return Vector2i(bounds.left + entranceX, bounds.top + 2);
}

void ParkPathGen::extractTiles(Grid<TileType>& grid, Town& town) const
{
	extractParkPathTiles(grid, Vector2i(bounds.left, bounds.top), pathGrid, Rect<int>(1, 1, bounds.width, bounds.height));
}


//-----------------------------------------------------------------------------
//--------------------------------  Walker  -----------------------------------


ParkPathGen::Walker::Walker(Grid<bool>& pathGrid, Random& rng, DirectionalTendency dirTend, Direction dir, int x, int y)
	:pathGrid(&pathGrid)
	,rng(&rng)
	,tendency(dirTend)
	,direction(dir)
	,x(x)
	,y(y)
	,alive(true)
{
}


void ParkPathGen::Walker::walk()
{
	switch (direction)
	{
	case Direction::North:
		--y;
		break;
	case Direction::East:
		++x;
		break;
	case Direction::West:
		--x;
		break;
	case Direction::South:
		++y;
		break;
	}

	if (x < 0 || y < 0 || x >= pathGrid->getWidth() || y >= pathGrid->getHeight())
	{
		alive = false;
	}
	else if (pathGrid->at(x, y))
	{
		alive = false;
	}
	else
	{
		pathGrid->at(x, y) = true;

		switch (tendency)
		{
		case DirectionalTendency::DrunkenWalk:
			if (rng->chance(0.1f))
			{
				switch (direction)
				{
				case Direction::North:
					direction = rng->choose({Direction::West, Direction::East});
					break;
				case Direction::East:
					direction = rng->choose({Direction::North, Direction::South});
					break;
				case Direction::West:
					direction = rng->choose({Direction::North, Direction::South});
					break;
				case Direction::South:
					direction = rng->choose({Direction::West, Direction::East});
					break;
				}
			}
			break;
		}
	}
}

bool ParkPathGen::Walker::isAlive() const
{
	return alive;
}


} // towngen
} // sl
