#include "Town/Generation/ParkingLotGenerator.hpp"

#include <cmath>

#include "Town/Town.hpp"
#include "Town/Map/MapStaticEntityData.hpp"

namespace sl
{
namespace towngen
{

ParkingLotGenerator::ParkingLotGenerator(int layers, bool isAbove)
	:layers(layers)
	,isAbove(isAbove)
{
}

void ParkingLotGenerator::generate(const Rect<int>& area, Random& rng, Town& town, Grid<TileType>& tiles, int tileSize, map::MapStaticEntityData& minimapStaticEntityData) const
{
	ASSERT(area.height == getHeight());
	bool isTopSection = true;
	int topOffset = isAbove ? 0 : 1;
	for (int layer = 0; layer < layers; ++layer, isTopSection = !isTopSection)
	{
		// place parking lots
		const int y = area.top + getHeight(layer) - (isTopSection ? 0 : 1);
		for (int x = -1; x <= area.width; ++x)
		{
				
			const TileType lotPaint = x % 2 == 0 ? // figure out whether we need left or right paint
				(
					x == area.width ? // don't place left if we're the last stall and there would be no matching right paint
					TileType::RoadRest :
					TileType::ParkingLotL
				) :
				TileType::ParkingLotR;
			if (isTopSection)
			{
				if (!isAbove)
				{
					tiles.at(area.left + x, y) = TileType::ParkingLotM;
				}
				if (!isAbove && layer == layers - 1) // last layer on bottom
				{
					tiles.at(area.left + x, y + topOffset) = TileType::RoadRest;
					tiles.at(area.left + x, y + topOffset + 1) = TileType::RoadRest;
					tiles.at(area.left + x, y + topOffset + 2) = lotPaint;
					tiles.at(area.left + x, y + topOffset + 3) = lotPaint;
				}
				else
				{
					tiles.at(area.left + x, y + topOffset) = lotPaint;
					tiles.at(area.left + x, y + topOffset + 1) = lotPaint;
					tiles.at(area.left + x, y + topOffset + 2) = TileType::RoadRest;
					tiles.at(area.left + x, y + topOffset + 3) = TileType::RoadRest;
				}
			}
			else
			{
				if (isAbove && layer != layers - 1) // don't add it for the last one, since that gets added later
				{
					tiles.at(area.left + x, y - 1) = TileType::ParkingLotM;
				}
				tiles.at(area.left + x, y + topOffset) = lotPaint;
				tiles.at(area.left + x, y + topOffset + 1) = lotPaint;
			}
		}
		// add entrance/exit on each side
		if (isTopSection)
		{
			const int yOffset = (!isAbove && layer == layers - 1) ? // move entrance up 2 tiles on last layer on bottom so that we can reverse the order for that
				y + topOffset - 2:
				y + topOffset;
			// left exit
			tiles.at(area.left - 2, yOffset + 1) = TileType::SidewalkSW;
			tiles.at(area.left - 2, yOffset + 2) = TileType::RoadRest;
			tiles.at(area.left - 2, yOffset + 3) = TileType::RoadRest;
			tiles.at(area.left - 2, yOffset + 4) = TileType::SidewalkNW;
			// right exit
			tiles.at(area.right() + 1, yOffset + 1) = TileType::SidewalkSE;
			tiles.at(area.right() + 1, yOffset + 2) = TileType::RoadRest;
			tiles.at(area.right() + 1, yOffset + 3) = TileType::RoadRest;
			tiles.at(area.right() + 1, yOffset + 4) = TileType::SidewalkNE;
		}
	}
	// add in sidewalk between parking lot and building
	if (isAbove)
	{
		for (int x = 0; x < area.width; ++x)
		{
			tiles.at(area.left + x, area.top + getHeight() - 1) = TileType::SidewalkN;
		}
	}
}

int ParkingLotGenerator::getHeight() const
{
	return getHeight(layers);
}

// static
int ParkingLotGenerator::getHeight(int layerCount)
{
	return layerCount * 2                                     // parking spot contribution
	     + 2 * static_cast<int>(std::ceil(layerCount / 2.f))  // road between spots
	     + static_cast<int>(std::ceil((layerCount - 2)/ 2.f)) // dividor between layers
	     + 1;                                                 // top
}

} // towngen
} // sl
