#include "Town/Generation/RoadVertex.hpp"

#include "Pathfinding/NodeID.hpp"
#include "Pathfinding/PathManager.hpp"
#include "Town/TrafficLight.hpp"
#include "Town/FourWayTrafficController.hpp"
#include "Town/ThreeWayTrafficController.hpp"
#include "Town/LoadBasedFourWayTrafficController.hpp"
#include "Town/RoadSegment.hpp"
#include "Town/Road.hpp"
#include "Town/Town.hpp"

namespace sl
{
namespace towngen
{

static Road addRoadsLater("add in roads later", Road::Orientation::Horizontal, 0, 0, 0);

void RoadVertex::buildPathing(Town& town) const
{
	//         |  nl   nr  |
	//_________| = = = = = |________
	//         = nwo   neo =
	//     wu  =   nw ne   = eu
	//     wd  =   sw se   = ed
	//_________= swo   seo =________
	//         | = = = = = |
	//         |  sl  sr   |
	const int tileSize = town.getTileSize();
	pf::VertexBasedPathManager& pathManager = town.getRoadPathManager();
	int directions = 0;
	if (west)
	{
		++directions;
	}
	if (east)
	{
		++directions;
	}
	if (north)
	{
		++directions;
	}
	if (south)
	{
		++directions;
	}
	const pf::NodeAvailabilityChecker *northStraightChecker = nullptr,
	                                  *eastStraightChecker  = nullptr,
	                                  *westStraightChecker  = nullptr,
	                                  *southStraightChecker = nullptr,
	                                  *northLeftChecker     = nullptr,
	                                  *eastLeftChecker      = nullptr,
	                                  *westLeftChecker      = nullptr,
	                                  *southLeftChecker     = nullptr;
	if (directions > 2)	//	3-way or 4-way, so we need traffic lights
	{
		TrafficLight *westLight  = nullptr,
		             *eastLight  = nullptr,
		             *northLight = nullptr,
		             *southLight = nullptr;
		if (west)
		{
			westLight = new TrafficLight((x - leftRoadSize - 1) * tileSize, (y - aboveRoadSize - 1) * tileSize, TrafficLight::West, south);
			westStraightChecker = &westLight->getStraightChecker();
			westLeftChecker = &westLight->getLeftChecker();
			town.addEntityToGrid(westLight);
		}
		if (east)
		{
			eastLight = new TrafficLight((x + rightRoadSize + 1) * tileSize, (y + belowRoadSize + 1) * tileSize, TrafficLight::East, north);
			eastStraightChecker = &eastLight->getStraightChecker();
			eastLeftChecker = &eastLight->getLeftChecker();
			town.addEntityToGrid(eastLight);
		}
		if (north)
		{
			northLight = new TrafficLight((x + rightRoadSize + 1) * tileSize, (y - aboveRoadSize - 1) * tileSize, TrafficLight::North, west);
			northStraightChecker = &northLight->getStraightChecker();
			northLeftChecker = &northLight->getLeftChecker();
			town.addEntityToGrid(northLight);

		}
		if (south)
		{
			southLight = new TrafficLight((x - leftRoadSize - 1) * tileSize, (y + belowRoadSize + 1) * tileSize, TrafficLight::South, east);
			southStraightChecker = &southLight->getStraightChecker();
			southLeftChecker = &southLight->getLeftChecker();
			town.addEntityToGrid(southLight);
		}

		if (directions == 4)
		{
			town.addEntityToList(new FourWayTrafficController(*northLight, *eastLight, *westLight, *southLight));
			//town.addEntityToList(new LoadBasedFourWayTrafficController(northLight, eastLight, westLight, southLight));
		}
		else //	three way
		{
			if (!west)
			{
				town.addEntityToList(new ThreeWayTrafficController(*eastLight, *southLight, *northLight));
			}
			else if (!east)
			{
				town.addEntityToList(new ThreeWayTrafficController(*westLight, *northLight, *southLight));
			}
			else if (!north)
			{
				town.addEntityToList(new ThreeWayTrafficController(*southLight, *westLight, *eastLight));
			}
			else if (!south)
			{
				town.addEntityToList(new ThreeWayTrafficController(*northLight, *eastLight, *westLight));
			}
		}
	}

	if (west)
	{
		if (south)	//	right turn
		{
			pathManager.addDirectedEdge(wd(&town), swo(&town));
			pathManager.addDirectedEdge(swo(&town), sl(&town));
		}
		if (east)	//	forward
		{
			if (directions > 2)	//	hook the edge up to east traffic light
			{
				pathManager.addDirectedConditionalEdge(wd(&town), ed(&town), eastStraightChecker);
			}
			else
			{
				pathManager.addDirectedEdge(wd(&town), ed(&town));
			}
		}
		if (north)	//	left turn
		{
			if (directions > 2)	//	hook the edge up to north traffic light
			{
				pathManager.addDirectedConditionalEdge(wd(&town), nw(&town), westLeftChecker);
			}
			else
			{
				pathManager.addDirectedEdge(wd(&town), nw(&town));
			}
			pathManager.addDirectedEdge(nw(&town), nr(&town));
		}
		pathManager.addDirectedEdge(wu(&town), west->eu(&town));
		const int ts = town.getTileSize();
		town.addEntityToQuad(new RoadSegment(Rect<int>((west->x + rightRoadSize + 3)*ts, (y - aboveRoadSize - 1)*ts, (x - west->x - roadWidth - 6)*ts, (roadWidth + 2)*ts), wu(&town).getY()*ts, RoadSegment::Orientation::West, addRoadsLater));
	}
	if (east)
	{
		if (north)	//	right turn
		{
			pathManager.addDirectedEdge(eu(&town), neo(&town));
			pathManager.addDirectedEdge(neo(&town), nr(&town));
		}
		if (west)	//	forward
		{
			if (directions > 2)	//	hook the edge up to west traffic light
			{
				pathManager.addDirectedConditionalEdge(eu(&town), wu(&town), westStraightChecker);
			}
			else
			{
				pathManager.addDirectedEdge(eu(&town), wu(&town));
			}
		}
		if (south)	//	left turn
		{
			if (directions > 2)	//	hook the edge up to south traffic light
			{
				pathManager.addDirectedConditionalEdge(eu(&town), se(&town), eastLeftChecker);
			}
			else
			{
				pathManager.addDirectedEdge(eu(&town), se(&town));
			}
			pathManager.addDirectedEdge(se(&town), sl(&town));
		}
		pathManager.addDirectedEdge(ed(&town), east->wd(&town));
		const int ts = town.getTileSize();
		town.addEntityToQuad(new RoadSegment(Rect<int>((x + rightRoadSize + 3)*ts, (y - aboveRoadSize - 1)*ts, (east->x - x - roadWidth - 6)*ts, (roadWidth + 2)*ts), ed(&town).getY()*ts, RoadSegment::Orientation::East, addRoadsLater));
	}
	if (north)
	{
		if (west)	//	right turn
		{
			pathManager.addDirectedEdge(nl(&town), nwo(&town));
			pathManager.addDirectedEdge(nwo(&town), wu(&town));
		}
		if (south)	//	forward
		{
			if (directions > 2)	//	hook the edge up to south traffic light
			{
				pathManager.addDirectedConditionalEdge(nl(&town), sl(&town), southStraightChecker);
			}
			else
			{
				pathManager.addDirectedEdge(nl(&town), sl(&town));
			}
		}
		if (east)	//	left turn
		{
			if (directions > 2)	//	hook the edge up to east traffic light
			{
				pathManager.addDirectedConditionalEdge(nl(&town), ne(&town), northLeftChecker);
			}
			else
			{
				pathManager.addDirectedEdge(nl(&town), ne(&town));
			}
			pathManager.addDirectedEdge(ne(&town), ed(&town));
		}
		pathManager.addDirectedEdge(nr(&town), north->sr(&town));
		const int ts = town.getTileSize();
		town.addEntityToQuad(new RoadSegment(Rect<int>((x - leftRoadSize - 1)*ts, (north->y + belowRoadSize + 3)*ts, (roadWidth + 2)*ts, (y - north->y - roadWidth - 6)*ts), nr(&town).getX()*ts, RoadSegment::Orientation::North, addRoadsLater));
	}
	if (south)
	{
		if (east)	//	right turn
		{
			pathManager.addDirectedEdge(sr(&town), seo(&town));
			pathManager.addDirectedEdge(seo(&town), ed(&town));
		}
		if (north)	//	forward
		{
			if (directions > 2)	//	hook the edge up to north traffic light
			{
				pathManager.addDirectedConditionalEdge(sr(&town), nr(&town), northStraightChecker);
			}
			else
			{
				pathManager.addDirectedEdge(sr(&town), nr(&town));
			}
		}
		if (west)	//	left turn
		{
			if (directions > 2)	//	hook the edge up to west traffic light
			{
				pathManager.addDirectedConditionalEdge(sr(&town), sw(&town), southLeftChecker);
			}
			else
			{
				pathManager.addDirectedEdge(sr(&town), sw(&town));
			}
			pathManager.addDirectedEdge(sw(&town), wu(&town));
		}
		pathManager.addDirectedEdge(sl(&town), south->nl(&town));
		const int ts = town.getTileSize();
		town.addEntityToQuad(new RoadSegment(Rect<int>((x - leftRoadSize - 1)*ts, (y + belowRoadSize + 3)*ts, (roadWidth + 2)*ts, (south->y - y - roadWidth - 6)*ts), sl(&town).getX()*ts, RoadSegment::Orientation::South, addRoadsLater));
	}
}

// private
pf::NodeID RoadVertex::sl(Scene *town) const
{
	return pf::NodeID(x - leftRoadSize / 2, y + belowRoadSize + 5, town);
}
pf::NodeID RoadVertex::sr(Scene *town) const
{
	return pf::NodeID(x + 1 + rightRoadSize / 2, y + belowRoadSize + 5, town);
}
pf::NodeID RoadVertex::nl(Scene *town) const
{
	return pf::NodeID(x - leftRoadSize / 2, y - aboveRoadSize - 5, town);
}
pf::NodeID RoadVertex::nr(Scene *town) const
{
	return pf::NodeID(x + 1 + rightRoadSize / 2, y - aboveRoadSize - 5, town);
}
pf::NodeID RoadVertex::eu(Scene *town) const
{
	return pf::NodeID(x + rightRoadSize + 5, y - aboveRoadSize / 2, town);
}
pf::NodeID RoadVertex::ed(Scene *town) const
{
	return pf::NodeID(x + rightRoadSize + 5, y + 1 + belowRoadSize / 2, town);
}
pf::NodeID RoadVertex::wu(Scene *town) const
{
	return pf::NodeID(x - leftRoadSize - 5, y - aboveRoadSize / 2, town);
}
pf::NodeID RoadVertex::wd(Scene *town) const
{
	return pf::NodeID(x - leftRoadSize - 5, y + 1 + belowRoadSize / 2, town);
}
pf::NodeID RoadVertex::nw(Scene *town) const
{
	return pf::NodeID(x - leftRoadSize / 2, y - aboveRoadSize / 2, town);
}
pf::NodeID RoadVertex::ne(Scene *town) const
{
	return pf::NodeID(x + 1 + rightRoadSize / 2, y - aboveRoadSize / 2, town);
}
pf::NodeID RoadVertex::sw(Scene *town) const
{
	return pf::NodeID(x - leftRoadSize / 2, y + 1 + belowRoadSize / 2, town);
}
pf::NodeID RoadVertex::se(Scene *town) const
{
	return pf::NodeID(x + 1 + rightRoadSize / 2, y + 1 + belowRoadSize / 2, town);
}
pf::NodeID RoadVertex::nwo(Scene *town) const
{
	return pf::NodeID(x - leftRoadSize / 2 - 2, y - aboveRoadSize / 2 - 2, town);
}
pf::NodeID RoadVertex::neo(Scene *town) const
{
	return pf::NodeID(x + 3 + rightRoadSize / 2, y - aboveRoadSize / 2 - 2, town);
}
pf::NodeID RoadVertex::swo(Scene *town) const
{
	return pf::NodeID(x - leftRoadSize / 2 - 2, y + 3 + belowRoadSize / 2, town);
}
pf::NodeID RoadVertex::seo(Scene *town) const
{
	return pf::NodeID(x + 3 + rightRoadSize / 2, y + 3 + belowRoadSize / 2, town);
}

} // towngen
} // sl
