#include "Town/Generation/TownGenMinimap.hpp"

#include "Town/Map/MapData.hpp"
#include "Town/Map/MapStaticEntityData.hpp"

namespace sl
{
namespace towngen
{

static void addMapData(sf::Uint8 *mapTexData, const Grid<TileType>& mapData);

static void addStaticEntities(sf::Uint8 *mapTexData, const int tilesWide, const int tilesHigh, const std::vector<StaticEntityPrototype>& staticEntities);

static void addStaticEntities(sf::Uint8 *mapTexData, const map::MapStaticEntityData& minimapStaticEntityData);

sf::Image createMapTexture(const map::MapData& mapData)
{
	//  Pixel data for minimap
	const int tilesWide = mapData.getWidth();
	const int tilesHigh = mapData.getHeight();
	std::unique_ptr<sf::Uint8[]> mapTexData = std::make_unique<sf::Uint8[]>(4 * tilesWide * tilesHigh);
	
	if (mapData.getTiles())
	{
		addMapData(mapTexData.get(), *mapData.getTiles());
	}
	if (mapData.getPrototypes())
	{
		//addStaticEntities(mapTexData, tilesWide, tilesHigh, *mapData.getPrototypes());
	}
	if (mapData.getEntities())
	{
		addStaticEntities(mapTexData.get(), *mapData.getEntities());
	}

	sf::Image img;
	img.create(tilesWide, tilesHigh, mapTexData.get());

	return img;
}

sf::Image createMapTexture(const Grid<TileType>& mapData)
{
	//  Pixel data for minimap
	const int tilesWide = mapData.getWidth();
	const int tilesHigh = mapData.getHeight();
	std::unique_ptr<sf::Uint8[]> mapTexData = std::make_unique<sf::Uint8[]>(4 * tilesWide * tilesHigh);
	
	addMapData(mapTexData.get(), mapData);

	sf::Image img;
	img.create(tilesWide, tilesHigh, mapTexData.get());

	return img;
}

sf::Image createMapTexture(const Grid<TileType>& mapData, const std::vector<StaticEntityPrototype>& staticEntities)
{
	//  Pixel data for minimap
	const int tilesWide = mapData.getWidth();
	const int tilesHigh = mapData.getHeight();
	std::unique_ptr<sf::Uint8[]> mapTexData = std::make_unique<sf::Uint8[]>(4 * tilesWide * tilesHigh);
	
	addMapData(mapTexData.get(), mapData);
	addStaticEntities(mapTexData.get(), tilesWide, tilesHigh, staticEntities);

	sf::Image img;
	img.create(tilesWide, tilesHigh, mapTexData.get());

	return img;
}


//-----------------------------  static functions  ----------------------------

void addMapData(sf::Uint8 *mapTexData, const Grid<TileType>& mapData)
{
	const int tilesWide = mapData.getWidth();
	const int tilesHigh = mapData.getHeight();
	std::size_t offset;
	TileType t;
	for (int x = 0; x < tilesWide; ++x)
	{
		for (int y = 0; y < tilesHigh; ++y)
		{
			offset = 4 * x + (y * tilesWide * 4);
			t = mapData.at(x, y);
			if (t == Grass)
			{
				mapTexData[offset] = 82;
				mapTexData[offset + 1] = 135;
				mapTexData[offset + 2] = 58;
			}
			else if (isSidewalk(t))
			{
				mapTexData[offset] = 192;
				mapTexData[offset + 1] = 192;
				mapTexData[offset + 2] = 192;
			}
			else if (t == Pavement)
			{
				mapTexData[offset] = 164;
				mapTexData[offset + 1] = 164;
				mapTexData[offset + 2] = 164;
			}
			else if (t == Dirt)
			{
				mapTexData[offset] = 103;
				mapTexData[offset + 1] = 77;
				mapTexData[offset + 2] = 56;
			}
			else if (t == Intersection)
			{
				mapTexData[offset] = 64;
				mapTexData[offset + 1] = 64;
				mapTexData[offset + 2] = 64;
			}
			else if (t == CrosswalkH || t == CrosswalkV)
			{
				mapTexData[offset] = 255;
				mapTexData[offset + 1] = 255;
				mapTexData[offset + 2] = 255;
			}
			else if (isRoad(t) || isParkingLot(t))
			{
				mapTexData[offset] = 64;
				mapTexData[offset + 1] = 64;
				mapTexData[offset + 2] = 64;
			}
			else if (isParkPath(t))
			{
				mapTexData[offset] = 106;
				mapTexData[offset + 1] = 85;
				mapTexData[offset + 2] = 53;
			}
			else if (t == HACK)
			{
				mapTexData[offset] = 35;
				mapTexData[offset + 1] = 80;
				mapTexData[offset + 2] = 5;
			}
			else
			{
				//ERROR("unhandled tile");
			}
			mapTexData[offset + 3] = 255;
		}
	}
}

void addStaticEntities(sf::Uint8 *mapTexData, const int tilesWide, const int tilesHigh, const std::vector<StaticEntityPrototype>& staticEntities)
{
	std::size_t offset;
	for (const StaticEntityPrototype& sep : staticEntities)
	{
		offset = 4 * sep.getX() + (sep.getY() * tilesWide * 4);
		switch (sep.getType())
		{
			case StaticEntityPrototype::Tree:
			{
				mapTexData[offset] = 35;
				mapTexData[offset + 1] = 80;
				mapTexData[offset + 2] = 5;
				break;
			}
			case StaticEntityPrototype::FenceH:
			{
				mapTexData[offset] = 164;
				mapTexData[offset + 1] = 132;
				mapTexData[offset + 2] = 0;
				break;
			}
			case StaticEntityPrototype::FenceV:
			{
				mapTexData[offset] = 148;
				mapTexData[offset + 1] = 116;
				mapTexData[offset + 2] = 0;
				break;
			}
		}
		mapTexData[offset + 3] = 255;
	}
}

void addStaticEntities(sf::Uint8 *mapTexData, const map::MapStaticEntityData& minimapStaticEntityData)
{
	std::size_t offset;
	const int tilesWide = minimapStaticEntityData.getWidth();
	const int tilesHigh = minimapStaticEntityData.getHeight();
	for (int x = 0; x < tilesWide; ++x)
	{
		for (int y = 0; y < tilesHigh; ++y)
		{
			if (minimapStaticEntityData.get(x, y) != map::MapStaticEntityData::StaticEntity::None)
			{
				offset = 4 * x + (y * tilesWide * 4);
				switch (minimapStaticEntityData.get(x, y))
				{
				case map::MapStaticEntityData::StaticEntity::Tree:
					mapTexData[offset] = 35;
					mapTexData[offset + 1] = 80;
					mapTexData[offset + 2] = 5;
					break;
				case map::MapStaticEntityData::StaticEntity::Fence:
					mapTexData[offset] = 164;
					mapTexData[offset + 1] = 132;
					mapTexData[offset + 2] = 0;
					break;
				case map::MapStaticEntityData::StaticEntity::Building:
					mapTexData[offset] = 222;
					mapTexData[offset + 1] = 222;
					mapTexData[offset + 2] = 222;
					break;
				}
				mapTexData[offset + 3] = 255;
			}
		}
	}
}

} // towngen
} // sl
