#include "Town/Generation/TownGenerator.hpp"

#include <algorithm>
#include <tuple>

#include "Town/World.hpp"
#include "Pathfinding/PathManager.hpp"
#include "Town/Town.hpp"
#include "Town/Generation/BlockGeneratorCreation.hpp"
#include "Town/Generation/BlockOfBuildings.hpp"
#include "Town/Generation/BuildingBlueprint.hpp"
#include "Town/Generation/ConditionalPrototypeLimit.hpp"
#include "Town/Generation/GenerationTimer.hpp"
#include "Town/Generation/TownGenCommon.hpp"
#include "Town/Generation/TownGenMinimap.hpp"
#include "Town/Generation/YardBlueprint.hpp"
#include "Town/Generation/YardObjectPrototype.hpp"
#include "Utility/Assert.hpp"
#include "Utility/Logger.hpp"
#include "Utility/Memory.hpp"
#include "Utility/ToString.hpp"

namespace sl
{
namespace towngen
{

// definitions for TownGenCommon.hpp
const int roadWidth = 5;
const int leftRoadSize = std::floor(roadWidth / 2.f);
const int rightRoadSize = std::ceil(roadWidth / 2.f);
const int aboveRoadSize = leftRoadSize;
const int belowRoadSize = rightRoadSize;
const int sidewalkThickness = 2; // must be at least 2

RoadVertex TownGenerator::roadEdge(-1, -1);

TownGenerator::TownGenerator(int zonesWide, int zonesHigh, int zoneWidth, int zoneHeight, int tileSize, Random& random)
	:zonesWide(zonesWide)
	,zonesHigh(zonesHigh)
	,zoneWidth(zoneWidth)
	,zoneHeight(zoneHeight)
	,tilesWide(zonesWide * zoneWidth)
	,tilesHigh(zonesHigh * zoneHeight)
	,tileSize(tileSize)
	,zoneWidths(zonesWide)
	,zoneHeights(zonesHigh)
	,zoneStartX(zonesWide)
	,zoneStartY(zonesHigh)
	,random(random)
	,mapData(tilesWide, tilesHigh)
	,zones(zonesWide, zonesHigh, BlockType::Forest)
	,valid(false)
	,blockGenerators()
	,minimapStaticEntityData(tilesWide, tilesHigh, tileSize)
	,minimapData(tilesWide, tilesHigh, &mapData, &staticEntities, &minimapStaticEntityData)
	,roadVertexGrid(tilesWide, tilesHigh, nullptr) 
{
	makeBlockGenerators(blockGenerators, requiredBuildings, tileSize);
}

void TownGenerator::generate(const GenerationListener&  listener)
{
	valid = false;
	staticEntities.clear();
	intersections.clear();
	sidewalksNW.clear();
	sidewalksNE.clear();
	sidewalksSW.clear();
	sidewalksSE.clear();
	blocks.clear();
	for (auto& createdRequired : requiredBuildings)
	{
		createdRequired.second.created = 0;
	}

	//  Okay, now let's randomize the distances between zones a bit!
	const int minZoneWidth  = zoneWidth * 0.8f,  maxZoneWidth  = zoneWidth * 1.2f,
	          minZoneHeight = zoneHeight * 0.8f, maxZoneHeight = zoneHeight * 1.2f;
	int looped = 0;
	bool boundsViolated;
	const int maxAttempts = 255;

	GenerationTimer generationTimer;

	listener(0, "Plotting Zones", minimapData);

	//  Calculate widths
	do
	{
		//  Just in fucking case, so this thing won't be O(infinity) :3
		if (looped > maxAttempts)
		{
			//  If you somehow did maxAttempts attempts here, then you better enjoy your
			//  perfectly symmetrical zone layout!
			for (int i = 0; i < zonesWide; ++i)
			{
				zoneWidths[i] = zoneWidth;
			}
			break;
		}

		//  Reset everything to 0
		for (int i = 0; i < zonesWide; ++i)
		{
			zoneWidths[i] = 0;
		}

		//  Distribute the tiles randomly
		for (int i = 0; i < tilesWide; ++i)
		{
			++zoneWidths[random.random(zonesWide)];
		}
		//  Check to make sure the bounds haven't been violated
		boundsViolated = false;
		for (int i = 0; i < zonesWide; ++i)
		{
			if (zoneWidths[i] < minZoneWidth || zoneWidths[i] > maxZoneWidth)
			{
				boundsViolated = true;
				break;  //  No need to check the rest
			}
		}

		++looped;
	}
	while (boundsViolated);

	//  Calculate heights
	do
	{
		//  Just in fucking case, so this thing won't be O(infinity) :3
		if (looped > maxAttempts)
		{
			//  If you somehow did maxAttemps attempts here, then you better enjoy your
			//  perfectly symmetrical zone layout!
			for (int i = 0; i < zonesHigh; ++i)
			{
				zoneHeights[i] = zoneHeight;
			}
			break;
		}

		//  Reset everything to 0
		for (int i = 0; i < zonesHigh; ++i)
		{
			zoneHeights[i] = 0;
		}

		//  Distribute the tiles randomly
		for (int i = 0; i < tilesHigh; ++i)
		{
			++zoneHeights[random.random(zonesHigh)];
		}

		//  Check to make sure the bounds haven't been violated
		boundsViolated = false;
		for (int i = 0; i < zonesHigh; ++i)
		{
			if (zoneHeights[i] < minZoneHeight || zoneHeights[i] > maxZoneHeight)
			{
				boundsViolated = true;
				break;  //  No need to check the rest
			}
		}

		++looped;
	}
	while (boundsViolated);

	//  Calculate zone positions
	zoneStartX[0] = 0;
	for (int i = 1; i < zonesWide; ++i)
	{
		zoneStartX[i] = zoneStartX[i - 1] + zoneWidths[i - 1];
	}

	zoneStartY[0] = 0;
	for (int i = 1; i < zonesHigh; ++i)
	{
		zoneStartY[i] = zoneStartY[i - 1] + zoneHeights[i - 1];
	}

	//  Generate the blocks
	int n, x, y;
	zones.resize(zonesWide, zonesHigh, BlockType::Forest);

	//  Sprawl out with tendrils from the center of the city
	int totalZones = 0;
	const int repCount = 6;
	for (int reps = 0; reps < repCount; ++reps)
	{
		//  This ensures that there are two kinds of tendrils: one that prefers horizontal by 300%
		//  and one that prefers vertical
		for (int focusY = 0; focusY < 2; ++focusY)
		{
			//  These next two loops are to get the 8 direction:
			//  North, South, West, East, NorthWest, NorthEast, SouthWest, Southeast
			for (int xdir = -1; xdir < 2; ++xdir)
			{
				for (int ydir = -1; ydir < 2; ++ydir)
				{
					const int sprawlLength = 10;
					const int sprawlGirth = 4;
					n = random.randomInclusive(std::min(zonesWide, zonesHigh) / sprawlLength, std::min(zonesWide, zonesHigh) / sprawlGirth) + std::min(zonesWide, zonesHigh) / repCount;
					x = zonesWide / 2;
					y = zonesHigh / 2;
					for (int steps = 0; steps < n; ++steps)
					{
						//  If prefer y && 1/3 chance || prefer x and no 1/3 chance
						//  The chance being true means you deviate from your preference
						if (random.chance(1 / 3.f) == (focusY != 0))
						{
							x += xdir;
						}
						else
						{
							y += ydir;
						}
						if (x > 0 && x < zonesWide - 1 && y > 0 && y < zonesHigh - 1)
						{
							if (zones.at(x, y) != BlockType::Residential)
							{
								++totalZones;
								zones.at(x, y) = BlockType::Residential;
							}
						}
					}
				}
			}
		}
	}
	generationTimer("Plotting Zones");

	listener(1, "Designating Zones", minimapData);

	//  Fill in the center
	zones.at(zonesWide / 2, zonesHigh / 2) = BlockType::Residential;
	//  Now fill blocks in as commercial until the minimum commerical block quota is met
	int filled = 0;
	const int quota = std::max(1, totalZones / 5);
	do
	{
		//  This ensures that there are two kinds of tendrils: one that prefers horizontal by 300%
		//  and one that prefers vertical
		for (int focusY = 0; focusY < 2; ++focusY)
		{
			//  These next two loops are to get the 8 direction:
			//  North, South, West, East, NorthWest, NorthEast, SouthWest, Southeast
			for (int xdir = -1; xdir < 2; ++xdir)
			{
				for (int ydir = -1; ydir < 2; ++ydir)
				{
					x = random.randomInclusive(zonesWide / 4, zonesWide - zonesWide / 4);
					y = random.randomInclusive(zonesHigh / 4, zonesHigh - zonesHigh / 4);
					for (int steps = 0; steps < 2; ++steps)
					{
						//  If prefer y && 1/3 chance || prefer x and no 1/3 chance
						//  The chance being true means you deviate from your preference
						if (random.chance(1 / 3.f) == (focusY != 0))
						{
							x += xdir;
						}
						else
						{
							y += ydir;
						}
						if (x >= 0 && x < zonesWide && y >= 0 && y < zonesHigh)
						{
							//  Check to see if there's a city block here
							if (zones.at(x, y) == BlockType::Residential && filled < quota + 1)
							{
								//  Turn it into commercial
								zones.at(x, y) = BlockType::Commercial;
								++filled;
							}
						}
					}
				}
			}
		}
	}
	while (filled < quota);


	auto zonesAround = [](const Grid<BlockType>::Iterator& zone, BlockType type) {
		int count = 0;
		for (int ydiff = -1; ydiff <= 1; ++ydiff)
		{
			for (int xdiff = -1; xdiff <= 1; ++xdiff)
			{
				if ((xdiff != 0 || ydiff != 0) && zone.offset(xdiff, ydiff) != zone.getGrid().end() && *zone.offset(xdiff, ydiff) == type)
				{
					++count;
				}
			}
		}
		return count;
	};

	// make some adjustments (insert school zones + random forests)
	std::vector<Grid<BlockType>::Iterator> possibleSchoolZones;
	for (auto zone = zones.begin(); zone != zones.end(); ++zone)
	{
		if (random.chance(0.2f))
		{
			*zone = BlockType::Forest;
		}
		else
		{
			if (*zone == BlockType::Residential || *zone == BlockType::PoorResidential || *zone == BlockType::Commercial)
			{
				possibleSchoolZones.push_back(zone);
			}
			else if (zone.getX() > 0 && zone.getX() < zones.getWidth() - 1 && zone.getY() > 0 && zone.getY() < zones.getHeight() - 1)
			{
				const int forests = zonesAround(zone, BlockType::Forest);
				if (forests < 8)
				{
					if (random.chance(forests / 40.f))
					{
						*zone = BlockType::Residential; // TODO: make a more rural one?
						possibleSchoolZones.push_back(zone);
					}
				}
			}
		}
	}



	auto zonePriority = [this](const Grid<BlockType>::Iterator& zone) -> float {
		float p = 0.f;
		for (int xoff = -2; xoff <= 2; ++xoff)
		{
			for (int yoff = -2; yoff <= 2; ++yoff)
			{
				if (zone.offset(xoff, yoff) != zone.getGrid().end())
				{
					const float distFactor = 3.f / sqrtf(xoff*xoff + yoff*yoff);
					switch (*zone.offset(xoff, yoff))
					{
					case BlockType::Residential:
						p += distFactor * 1.f;
						break;
					case BlockType::PoorResidential:
						p += distFactor * 0.5f;
						break;
					case BlockType::Forest:
						break;
					case BlockType::Commercial:
						p += distFactor * 0.25f;
						break;
					case BlockType::School:
						p -= distFactor * 10.f;
					}
				}
			}
		}
		const int xLack = 20 - this->zoneWidths[zone.getX()];
		if (xLack > 0)
		{
			p -= xLack * 100.f;
		}
		else
		{
			p -= xLack;
		}
		const int yLack = 20 - this->zoneHeights[zone.getY()];
		if (yLack > 0)
		{
			p -= yLack * 100.f;
		}
		else
		{
			p -= yLack;
		}
		return p;
	};
	
	const int requiredSchoolZones = 2;
	for (int i = 0; i < requiredSchoolZones; ++i)
	{
		auto zone = std::max_element(possibleSchoolZones.begin(), possibleSchoolZones.end(),
			[&zonePriority](const Grid<BlockType>::Iterator& lhs, const Grid<BlockType>::Iterator& rhs) {
				return zonePriority(lhs) < zonePriority(rhs);
			}
		);
		if (zone != possibleSchoolZones.end())
		{
			**zone = BlockType::School;
			*zone = possibleSchoolZones.back();
			possibleSchoolZones.pop_back();
		}
	}

	auto centralPark = std::min_element(possibleSchoolZones.begin(), possibleSchoolZones.end(),
		[&zonePriority](const Grid<BlockType>::Iterator& lhs, const Grid<BlockType>::Iterator& rhs) {
		return zonePriority(lhs) < zonePriority(rhs);
	});
	ASSERT(centralPark != possibleSchoolZones.end());
	**centralPark = BlockType::Park;


	//  Do the default fill for the zone type
	for (int x = 0; x < zonesWide; ++x)
	{
		for (int y = 0; y < zonesHigh; ++y)
		{
			switch (zones.at(x, y))
			{
			case BlockType::Forest:
				fillBlock(x, y, Dirt);
				break;
			case BlockType::Residential:
				fillBlock(x, y, Grass);
				// @todo move this somewhere else or do something else for income variance?
				if (random.chance(0.35f))
				{
					zones.at(x, y) = BlockType::PoorResidential;
				}
				break;
			case BlockType::Commercial:
				fillBlock(x, y, Pavement);
				break;
			case BlockType::School:
				fillBlock(x, y, Grass);
				break;
			case BlockType::Park:
				fillBlock(x, y, Grass);
				break;
			}
		}
	}

	for (int x = 0; x < zonesWide; ++x)
	{
		for (int y = 0; y < zonesHigh; ++y)
		{
			fillRoad(x, y);
		}
	}

	//  And create some blocks based on the block layout
	//std::vector<std::tuple<int, int, std::vector<YardBlueprint> > > buildings;
	//std::vector<BlockOfBuildings> buildings;

	generationTimer("Designating Zones");

	listener(2, "Splitting Roads", minimapData);
	//  std::tuple<x, y, width>
	std::vector<std::tuple<int, int, int>> roadSplitLocations;
	//fs << "{" << zonesWide << "," << zonesHigh << "}\n";
	std::vector<std::pair<int, int>> zoneCoordinates;
	for (int x = 0; x < zonesWide; ++x)
	{
		for (int y = 0; y < zonesHigh; ++y)
		{
			zoneCoordinates.push_back(std::make_pair(x, y));
		}
	}
	// shuffle zone coordinates so that buildings get constructed in a random order to avoid having all required buildings in top left of town
	random.shuffle(zoneCoordinates.begin(), zoneCoordinates.end());
	for (const std::pair<int, int>& zoneCoordinate : zoneCoordinates)
	{
		const int x = zoneCoordinate.first;
		const int y = zoneCoordinate.second;
		//fs << "[" << x << "," << y << "]\n";
		//  if for some reason we don't find it...
		auto it = blockGenerators.find(zones.at(x, y));
		if (it != blockGenerators.end())
		{
			const Rect<int> zoneArea(
				zoneStartX[x] + leftRoadSize + sidewalkThickness,
				zoneStartY[y] + aboveRoadSize + sidewalkThickness,
				zoneWidths[x] - roadWidth - 4,
				zoneHeights[y] - roadWidth - 4);
			it->second.generate(random, zoneArea, buildings, roadSplitLocations, &requiredBuildings);
		}
	}

	for (const auto& required : requiredBuildings)
	{
		Logger::log(Logger::TownGen, Logger::Debug) << "created " << required.second.created << " / " << required.second.required << " of " << required.first << std::endl;
	}

	random.shuffle(roadSplitLocations.begin(), roadSplitLocations.end());

	//  Now we need to determine if roads can join properly without having ugly intersections
	//        ____|   |                    ____|   |
	//  Bad:          |________    Good:           |    _______
	//        ____                         ____    |   | This is now a dead end
	//            |    ________                |   |   |_______
	//            |   |                        |   |
	//  if they can't then we must make that end of the split a dead end.
	//  If a double-dead-end is made then we must try to add in a vertical split to connect the road
	//  like this:
	//      ___| |    __________   | |____        ___| |   ___________   | |____
	//      ___  |   |__________|  |  ____        ___  |  |____   ____|  |  ____
	//  Bad:   | |_________________| |        Good:  | |_______| |_______| |
	//         |  _________________  |               |  _________________  |

	bool leftIsFucked, rightIsFucked;
	int tx, ty, twidth; //  stuff from tuples
	for (const auto& splitLocation : roadSplitLocations)
	{
		std::tie(tx, ty, twidth) = splitLocation;
		int leftStart = tx - sidewalkThickness + 1;
		int rightStart = tx + twidth + sidewalkThickness - 2;
		leftIsFucked = false;
		rightIsFucked = false;
		//	Figure out if the left/right ends are bad (first case)
		for (int i = 0; i < 4 + roadWidth; ++i)
		{
			if (isRoad(mapData.at(leftStart - 6 - roadWidth, ty + i)))
			{
				leftIsFucked = true;
			}
			if (isRoad(mapData.at(rightStart + 6 + roadWidth, ty + i)))
			{
				rightIsFucked = true;
			}
		}

		if (leftIsFucked)
		{
			leftStart += 6;
		}
		if (rightIsFucked)
		{
			rightStart -= 6;
		}

		// 12 is just some arbitrary minimum amount I'd ever like the distance to be...
		ASSERT(rightStart - leftStart > 12);

		for (int x = leftStart; x <= rightStart; ++x)
		{
			placeSouthSidewalk(x, ty + sidewalkThickness - 1);
			for (int i = 0; i < roadWidth; ++i)
			{
				addRoad(x, ty + sidewalkThickness + i, RoadRest);
			}
			addRoad(x, ty + aboveRoadSize + sidewalkThickness, RoadH);
			placeNorthSidewalk(x, ty + sidewalkThickness + roadWidth);
		}

		if (leftIsFucked)
		{
			
			for (int i = 1; i <= roadWidth + 2 * sidewalkThickness; ++i)
			{
				placeEastSidewalk(leftStart, ty - 1 + i);
			}
			mapData.at(leftStart, ty + sidewalkThickness - 1) = SidewalkNWI;
			mapData.at(leftStart, ty + sidewalkThickness + roadWidth) = SidewalkSWI;
			addIntersection(leftStart + rightRoadSize, ty + sidewalkThickness + aboveRoadSize);
		}
		else
		{
			addSidewalkCorner(leftStart, ty - 2 + sidewalkThickness, SidewalkSW);
			addSidewalkCorner(leftStart, ty + 1 + sidewalkThickness + roadWidth, SidewalkNW);
			addIntersection(leftStart - 1 - rightRoadSize, ty + sidewalkThickness + aboveRoadSize);
		}
		if (rightIsFucked)
		{
			
			for (int i = 1; i <= roadWidth + 2 * sidewalkThickness; ++i)
			{
				placeWestSidewalk(rightStart, ty - 1 + i);
			}
			mapData.at(rightStart, ty + sidewalkThickness - 1) = SidewalkNEI;
			mapData.at(rightStart, ty + sidewalkThickness + roadWidth) = SidewalkSEI;
			addIntersection(rightStart - rightRoadSize, ty + sidewalkThickness + aboveRoadSize);
		}
		else
		{
			addSidewalkCorner(rightStart, ty - 2 + sidewalkThickness, SidewalkSE);
			addSidewalkCorner(rightStart, ty + 1 + sidewalkThickness + roadWidth, SidewalkNE);
			addIntersection(rightStart + leftRoadSize + sidewalkThickness, ty + sidewalkThickness + aboveRoadSize);
		}

	}
	generationTimer("Splitting Roads");

	placeCrosswalks();

	placeIntersections();

	for (int x = 0; x < zonesWide; ++x)
	{
		for (int y = 0; y < zonesHigh; ++y)
		{
			generateBlock(x, y, zones.at(x, y));
		}
	}

	valid = true;
}

void TownGenerator::exportMap(World& world, const GenerationListener& listener)
{
	ASSERT(isValid());
	auto townOwned = unique<Town>(world.getGame(), tilesWide * tileSize, tilesHigh * tileSize, 512, 384);
	Town *town = townOwned.get();
	//town->setTiles(mapData); // @TODO remove this (it's needed because minimap drawable won't be alive until it's called)
	world.addScene(std::move(townOwned), true);
	world.getPolice().setTown(*town);

	bool isPlayerGenerated = false;

	GenerationTimer generationTimer;

	for (int i = 0; i < staticEntities.size(); ++i)
	{
		if (i % ((int) staticEntities.size() / 10) == 0)
		{
			std::ostringstream ss;
			ss << "Placing Static Entities [" << (100 * i / staticEntities.size()) << "%]";
			listener(10 + 10 * (float)i / staticEntities.size(), ss.str().c_str(), minimapData);
		}
		staticEntities[i].generate(random, *town, tileSize, minimapStaticEntityData);
	}
	generationTimer("Placing Static Entities");


	// shuffle buildings for two reasons:
	// 1) since we choose the 1st generated residential basic house as the player's house, this leads to a random position for it
	// 2) this makes the map generation look more random in the preview window instead of being from top left to bottom rights
	random.shuffle(buildings.begin(), buildings.end());
	const int buildingsPerUpdate = buildings.size() < 10 ? 1 : ((int)buildings.size() / 10);
	for (int i = 0; i < buildings.size(); ++i)
	{
		if (i % buildingsPerUpdate == 0)
		{
			std::ostringstream ss;
			ss << "Generating Buildings [" << (100 * i / buildings.size()) << "%]";
			listener(20 + 55 * (float)i / buildings.size(), ss.str().c_str(), minimapData);
		}
		buildings[i].generate(random, *town, mapData, tileSize, minimapStaticEntityData, isPlayerGenerated);
	}
	generationTimer("Generating Buildings");


	// reset tiles for any added during yard generation
	town->setTiles(mapData);


	listener(75, "Calculating Car Pathing", minimapData);
	calculateCarPathing(*town);
	generationTimer("Calculating Car Pathing");

	
	town->makeMap(minimapData);
	//town->spawnAdults(64);


	world.finalizeGeneration(minimapData, listener, 80, 100);
	generationTimer("Calculating NPC schedules / pathing");
	

	listener(100, "Done!", minimapData);
}

const Grid<TileType>& TownGenerator::extractTiles() const
{
	return mapData;
}

bool TownGenerator::isValid() const
{
	return valid;
}

const map::MapData& TownGenerator::getMinimapData() const
{
	return minimapData;
}


/*		  private methods			 */

void TownGenerator::generateBlock(int x, int y, BlockType type)
{
	//return;
	switch (type)
	{
		case BlockType::Residential:
		{
			break;
		}
		case BlockType::Commercial:
		{
			break;
		}
		case BlockType::Forest:
		{
			const float borderForests = getNeighborBlocks(x, y, BlockType::Forest, 0.5);
			const int xMax = zoneStartX[x] + zoneWidths[x], yMax = zoneStartY[y] + zoneHeights[y];
			if (borderForests >= 0.99)
			{
				for (int i = zoneStartX[x]; i < xMax; ++i)
				{
					for (int j = zoneStartY[y]; j < yMax; ++j)
					{
						mapData.at(i, j) = HACK;
					}
				}
			}
			else
			{
				return;
				for (int i = zoneStartX[x]; i < xMax; ++i)
				{
					for (int j = zoneStartY[y]; j < yMax; ++j)
					{
						//if (mapData.at(i, j) == Dirt && random.chance(borderForests))
						//{
						//	staticEntities.push_back(StaticEntityPrototype(i, j, StaticEntityPrototype::Tree));
						//}
						if (mapData.at(i, j) == Dirt)
						{
							auto weight = [&](int zx, int zy) -> float
							{
								if (zx <= 0 || zx >= zonesWide || zy <= 0 || zy >= zonesHigh)
								{
									return 1.f;
								}
								return zones.at(zx, zy) == BlockType::Forest ? 1.f : 0.1f;
							};
							const float xRatio = (i - zoneStartX[x]) / (float)zoneWidths[x];
							const float yRatio = (j - zoneStartY[y]) / (float)zoneHeights[y];
							ASSERT(xRatio >= 0.f && xRatio <= 1.f);
							ASSERT(yRatio >= 0.f && yRatio <= 1.f);
							const float probability = 0.5f * (weight(x - 1, y) * (1.f - xRatio) + weight(x + 1, y) * xRatio)
							                        + 0.5f * (weight(x, y - 1) * (1.f - yRatio) + weight(x, y + 1) * yRatio);
							if (random.chance(probability))
							{
								staticEntities.push_back(StaticEntityPrototype(i, j, StaticEntityPrototype::Tree));
							}
						}
					}
				}
			}
			break;
		}
	}
}

void TownGenerator::fillBlock(int x, int y, TileType fillType)
{
	const int xMax = zoneStartX[x] + zoneWidths[x], yMax = zoneStartY[y] + zoneHeights[y];

	for (int i = zoneStartX[x]; i < xMax; ++i)
	{
		for (int j = zoneStartY[y]; j < yMax; ++j)
		{
			mapData.at(i, j) = fillType;
		}
	}
}

void TownGenerator::fillRoad(int x, int y)
{
	// Top left corner of block
	const int sx = zoneStartX[x];
	const int sy = zoneStartY[y];
	// Bottom right corner of block
	const int ex = sx + zoneWidths[x] - 1;
	const int ey = sy + zoneHeights[y] - 1;
	//  Top
	const bool top = (zones.at(x, y) != BlockType::Forest || (y > 1 && zones.at(x, y - 1) != BlockType::Forest));
	if (top)
	{
		for (int i = aboveRoadSize + 1; i < zoneWidths[x] - belowRoadSize - 1; ++i)
		{
			for (int j = 0; j < aboveRoadSize; ++j)
			{
				addRoad(sx + i, sy + j, RoadRest);
			}
			placeNorthSidewalk(sx + i, sy + aboveRoadSize);
		}
	}
	const bool bottom = (zones.at(x, y) != BlockType::Forest || (y < zonesHigh - 1 && zones.at(x, y + 1) != BlockType::Forest));
	if (bottom)
	{
		for (int i = aboveRoadSize + 1; i < zoneWidths[x] - belowRoadSize - 1; ++i)
		{
			for (int j = 0; j < belowRoadSize; ++j)
			{
				addRoad(sx + i, ey - j - 1, RoadRest);
			}
			placeSouthSidewalk(sx + i, ey - belowRoadSize);
			addRoad(sx + i, ey, RoadH);
		}
	}
	const bool left = (zones.at(x, y) != BlockType::Forest || (x > 1 && zones.at(x - 1, y) != BlockType::Forest));
	if (left)
	{
		for (int i = leftRoadSize + 1; i < zoneHeights[y] - rightRoadSize - 1; ++i)
		{
			for (int j = 0; j < leftRoadSize; ++j)
			{
				addRoad(sx + j, sy + i, RoadRest);
			}
			placeWestSidewalk(sx + leftRoadSize, sy + i);
		}
	}
	const bool right = (zones.at(x, y) != BlockType::Forest || (x < zonesWide - 1 && zones.at(x + 1, y) != BlockType::Forest));
	if (right)
	{
		for (int i = leftRoadSize + 1; i < zoneHeights[y] - rightRoadSize - 1; ++i)
		{
			for (int j = 0; j < rightRoadSize; ++j)
			{
				addRoad(ex - j, sy + i + 1, RoadRest);
			}
			placeEastSidewalk(ex - rightRoadSize, sy + i);
			addRoad(ex, sy + i, RoadV);
		}
	}
	if (zones.at(x, y) != BlockType::Forest)
	{
		addIntersection(ex, ey);
		addSidewalkCorner(sx + leftRoadSize + 1, sy + aboveRoadSize + 1, SidewalkNW);
		addSidewalkCorner(ex - rightRoadSize - 1, sy + aboveRoadSize + 1, SidewalkNE);
		addSidewalkCorner(sx + leftRoadSize + 1, ey - belowRoadSize - 1, SidewalkSW);
		addSidewalkCorner(ex - rightRoadSize - 1, ey - belowRoadSize - 1, SidewalkSE);
	}
	else	//  We're a forest, so TIME TO TAKE CARE OF LOTS AND LOTS OF FUCKING EDGE CASES! YAY!
	{
		if (left && top)
		{
			addSidewalkCorner(sx + leftRoadSize + 1, sy + aboveRoadSize + 1, SidewalkNW);
		}
		if (right && top)
		{
			addSidewalkCorner(ex - rightRoadSize - 1, sy + aboveRoadSize + 1, SidewalkNE);
		}
		if (left && bottom)
		{
			addSidewalkCorner(sx + leftRoadSize + 1, ey - belowRoadSize - 1, SidewalkSW);
		}
		if (right && bottom)
		{
			addSidewalkCorner(ex - rightRoadSize - 1, ey - belowRoadSize - 1, SidewalkSE);
		}
		if (left)
		{
			if (!top)
			{
				for (int i = -1; i < aboveRoadSize + 1; ++i)
				{
					placeWestSidewalk(sx + leftRoadSize, sy + i);
				}
			}
			if (!bottom)
			{
				for (int i = -1; i < belowRoadSize + 1; ++i)
				{
					placeWestSidewalk(sx + leftRoadSize, ey - i);
				}
			}
		}
		if (right)
		{
			if (!top)
			{
				for (int i = -1; i < aboveRoadSize + 1; ++i)
				{
					placeEastSidewalk(ex - rightRoadSize, sy + i);
				}
			}
			if (!bottom)
			{
				for (int i = -1; i < belowRoadSize + 1; ++i)
				{
					placeEastSidewalk(ex - rightRoadSize, ey - i);
				}
			}
		}
		if (top)
		{
			if (!left)
			{
				for (int i = -1; i < leftRoadSize + 1; ++i)
				{
					placeNorthSidewalk(sx + i, sy + aboveRoadSize);
				}
			}
			if (!right)
			{
				for (int i = -1; i < rightRoadSize + 1; ++i)
				{
					placeNorthSidewalk(ex - i, sy + aboveRoadSize);
				}
			}
		}
		if (bottom)
		{
			if (!left)
			{
				for (int i = -1; i < leftRoadSize + 1; ++i)
				{
					placeSouthSidewalk(sx + i, ey - belowRoadSize);
				}
			}
			if (!right)
			{
				for (int i = -1; i < rightRoadSize + 1; ++i)
				{
					placeSouthSidewalk(ex - i, ey - belowRoadSize);
				}
			}
		}
		const bool topleft = (x > 1 && y > 1 && zones.at(x - 1, y - 1) != BlockType::Forest);
		const bool topright = (x < zonesWide - 1 && y > 1 && zones.at(x + 1, y - 1) != BlockType::Forest);
		const bool bottomleft = (x > 1 && y < zonesHigh - 1 && zones.at(x - 1, y + 1) != BlockType::Forest);
		const bool bottomright = (x < zonesWide - 1 && y < zonesHigh - 1 && zones.at(x + 1, y + 1) != BlockType::Forest);
		// the following inner loops are starting at 1 because that area is already covered in the above section of code
		if (topleft && !left && !top)
		{
			for (int i = 1; i <= aboveRoadSize; ++i)
			{
				placeWestSidewalk(sx + leftRoadSize, sy + i);
			}
			for (int i = 1; i <= leftRoadSize; ++i)
			{
				placeNorthSidewalk(sx + i, sy + aboveRoadSize);
			}
			mapData.at(sx + leftRoadSize, sy + aboveRoadSize) = SidewalkSEI;
			mapData.at(sx + leftRoadSize + 1, sy + aboveRoadSize + 1) = SidewalkInner;
		}
		if (topright && !right && !top)
		{
			for (int i = 1; i <= aboveRoadSize; ++i)
			{
				placeEastSidewalk(ex - rightRoadSize, sy + i);
			}
			for (int i = 1; i <= rightRoadSize; ++i)
			{
				placeNorthSidewalk(ex - i, sy + aboveRoadSize);
			}
			mapData.at(ex - rightRoadSize, sy + aboveRoadSize) = SidewalkSWI;
			mapData.at(ex - rightRoadSize - 1, sy + aboveRoadSize + 1) = SidewalkInner;
		}
		if (bottomleft && !left && !bottom)
		{
			for (int i = 1; i <= belowRoadSize; ++i)
			{
				placeWestSidewalk(sx + leftRoadSize, ey - i);
			}
			for (int i = 1; i <= leftRoadSize; ++i)
			{
				placeSouthSidewalk(sx + i, ey - belowRoadSize);
			}
			mapData.at(sx + leftRoadSize, ey - belowRoadSize) = SidewalkNEI;
			mapData.at(sx + leftRoadSize + 1, ey - belowRoadSize - 1) = SidewalkInner;
		}
		if (bottomright && !right && !bottom)
		{
			for (int i = 1; i <= belowRoadSize; ++i)
			{
				placeEastSidewalk(ex - rightRoadSize, ey - i);
			}
			for (int i = 1; i <= rightRoadSize; ++i)
			{
				placeSouthSidewalk(ex - i, ey - belowRoadSize);
			}
			mapData.at(ex - rightRoadSize, ey - belowRoadSize) = SidewalkNWI;
			mapData.at(ex - rightRoadSize - 1, ey - belowRoadSize - 1) = SidewalkInner;
		}
		if (bottom || right || bottomright)
		{
			addIntersection(ex, ey);
		}
	}
}

float TownGenerator::getNeighborBlocks(int x, int y, BlockType type, float diagWeight) const
{
	float neighborsMatched = 0, neighborsTotal = 0;
	if (x > 1)
	{
		neighborsMatched += (zones.at(x - 1, y) == type);
		++neighborsTotal;
		if (y > 1)
		{
			neighborsMatched += (zones.at(x - 1, y - 1) == type) * diagWeight;
			neighborsTotal += diagWeight;
		}
		if (y < zonesHigh - 1)
		{
			neighborsMatched += (zones.at(x - 1, y + 1) == type) * diagWeight;
			neighborsTotal += diagWeight;
		}
	}
	if (x < zonesWide - 1)
	{
		neighborsMatched += (zones.at(x + 1, y) == type);
		++neighborsTotal;
		if (y > 1)
		{
			neighborsMatched += (zones.at(x + 1, y - 1) == type) * diagWeight;
			neighborsTotal += diagWeight;
		}
		if (y < zonesHigh - 1)
		{
			neighborsMatched += (zones.at(x + 1, y + 1) == type) * diagWeight;
			neighborsTotal += diagWeight;
		}
	}
	if (y > 1)
	{
		neighborsMatched += (zones.at(x, y - 1) == type);
		++neighborsTotal;
	}
	if (y < zonesHigh - 1)
	{
		neighborsMatched += (zones.at(x, y + 1) == type);
		++neighborsTotal;
	}
	return neighborsMatched / neighborsTotal;
}

/* block stuff */
void TownGenerator::resolveIntersections()
{
	for (std::vector<Block>::iterator it = blocks.begin(); it != blocks.end(); ++it)
	{

	}
}

void TownGenerator::splitSubroads()
{
	for (std::vector<Block>::iterator it = blocks.begin(); it != blocks.end(); ++it)
	{
		if (random.chance(9))
		{
			//  Try three times to create a split
			for (int i = 0; i < 3; ++i)
			{
				//int xoffset = 4;
			}
		}
	}
}

/*  Intersection stuff */
void TownGenerator::addIntersection(int x, int y)
{
	intersections.push_back(std::make_pair(x, y));
}

void TownGenerator::placeIntersections()
{
	int x, y;
	for (std::vector<std::pair<int, int>>::iterator it = intersections.begin(); it != intersections.end(); ++it)
	{
		x = it->first;
		y = it->second;
		roadVertices.push_back(unique<RoadVertex>(x, y));
		for (int i = -leftRoadSize; i < rightRoadSize; ++i)
		{
			for (int j = -aboveRoadSize; j < belowRoadSize; ++j)
			{
				mapData.at(x + i, y + j) = Intersection;
				roadVertexGrid.at(x + i, y + j) = roadVertices.back().get();
			}
		}
	}
	// join up intersection vertices
	for (std::vector<std::pair<int, int>>::iterator it = intersections.begin(); it != intersections.end(); ++it)
	{
		x = it->first;
		y = it->second;
		RoadVertex *start = roadVertexGrid.at(x, y);
		while (roadVertexGrid.at(x, y) == start || roadVertexGrid.at(x, y) == &roadEdge)
		{
			++x;
		}
		if (roadVertexGrid.at(x, y))
		{
			roadVertexGrid.at(x, y)->west = start;
			start->east = roadVertexGrid.at(x, y);
		}
		x = it->first;
		y = it->second;
		while (roadVertexGrid.at(x, y) == start || roadVertexGrid.at(x, y) == &roadEdge)
		{
			--x;
		}
		if (roadVertexGrid.at(x, y))
		{
			roadVertexGrid.at(x, y)->east = start;
			start->west = roadVertexGrid.at(x, y);
		}
		x = it->first;
		y = it->second;
		while (roadVertexGrid.at(x, y) == start || roadVertexGrid.at(x, y) == &roadEdge)
		{
			++y;
		}
		if (roadVertexGrid.at(x, y))
		{
			roadVertexGrid.at(x, y)->north = start;
			start->south = roadVertexGrid.at(x, y);
		}
		x = it->first;
		y = it->second;
		while (roadVertexGrid.at(x, y) == start || roadVertexGrid.at(x, y) == &roadEdge)
		{
			--y;
		}
		if (roadVertexGrid.at(x, y))
		{
			roadVertexGrid.at(x, y)->south = start;
			start->north = roadVertexGrid.at(x, y);
		}
	}
}

void TownGenerator::createRoadSystem(Town& town) const
{
	for (const std::pair<int, int>& intersection : intersections)
	{
		const int x = intersection.first;
		const int y = intersection.second;
		ERROR("pls implement?");
		const bool north = mapData.at(x, y - 2) == CrosswalkH;
		const bool south = mapData.at(x, y + 3) == CrosswalkH;
		const bool west = mapData.at(x - 2, y) == CrosswalkV;
		const bool east = mapData.at(x + 3, y) == CrosswalkV;
	}
}

void TownGenerator::calculateCarPathing(Town& town) const
{
	ASSERT(valid);
	for (const std::unique_ptr<RoadVertex>& roadVertex : roadVertices)
	{
		roadVertex->buildPathing(town);
	}
}

/*  Sidewalk stuff  */
void TownGenerator::addSidewalkCorner(int x, int y, TileType sidewalkType)
{
	ASSERT(sidewalkType == SidewalkNW || sidewalkType == SidewalkNE ||
	       sidewalkType == SidewalkSW || sidewalkType == SidewalkSE);
	//  Place corner
	mapData.at(x, y) = SidewalkInner;
	//  Build corner vectors so that they're ready for placeCrosswalks()
	if (sidewalkType == SidewalkNW)
	{
		mapData.at(x - 1, y - 1) = sidewalkType;
		sidewalksNW.push_back(std::make_pair(x, y));
	}
	else if (sidewalkType == SidewalkNE)
	{
		mapData.at(x + 1, y - 1) = sidewalkType;
		sidewalksNE.push_back(std::make_pair(x, y));
	}
	else if (sidewalkType == SidewalkSW)
	{
		mapData.at(x - 1, y + 1) = sidewalkType;
		sidewalksSW.push_back(std::make_pair(x, y));
	}
	else if (sidewalkType == SidewalkSE)
	{
		mapData.at(x + 1, y + 1) = sidewalkType;
		sidewalksSE.push_back(std::make_pair(x, y));
	}
}

//  Must call addSidewalkCorner() to build up the vectors used here
void TownGenerator::placeCrosswalks()
{
	std::vector<std::pair<int, int> >::iterator it;
	int x, y;   //  position of sidewalk corner
	int i;      //  used as an iterator for placing crosswalks
	for (it = sidewalksNW.begin(); it != sidewalksNW.end(); ++it)
	{
		x = it->first;
		y = it->second;
		i = x - 1;
		//  Keep moving left and painting crosswalks as long as there's road
		while (--i >= 0 && !isSidewalk(mapData.at(i, y)))// == RoadV || mapData.at(i, y) == RoadRest))
		{
			addRoad(i, y,  CrosswalkH);
			addRoad(i, y - 1, RoadRest);
		}
		i = y - 1;
		//  Keep moving up and painting crosswalks as long as there's road
		while (--i >= 0 && !isSidewalk(mapData.at(x, i)))// == RoadH || mapData.at(x, i) == RoadRest))
		{
			addRoad(x, i, CrosswalkV);
			addRoad(x - 1, i, RoadRest);
		}
	}
	for (it = sidewalksNE.begin(); it != sidewalksNE.end(); ++it)
	{
		x = it->first;
		y = it->second;
		i = x + 1;
		//  Keep moving right and painting crosswalks as long as there's road
		while (++i < tilesWide && !isSidewalk(mapData.at(i, y)))//(mapData.at(i, y) == RoadV || mapData.at(i, y) == RoadRest))
		{
			addRoad(i, y, CrosswalkH);
			addRoad(i, y - 1,  RoadRest);
		}
		i = y - 1;
		//  Keep moving up and painting crosswalks as long as there's road
		while (--i > 0 && !isSidewalk(mapData.at(x, i)))//(mapData.at(x, i) == RoadH || mapData.at(x, i) == RoadRest))
		{
			addRoad(x, i, CrosswalkV);
			addRoad(x + 1, i, RoadRest);
		}
	}
	for (it = sidewalksSW.begin(); it != sidewalksSW.end(); ++it)
	{
		x = it->first;
		y = it->second;
		i = x - 1;
		//  Keep moving left and painting crosswalks as long as there's road
		while (--i >= 0 && !isSidewalk(mapData.at(i, y)))//&& (mapData.at(i, y) == RoadV || mapData.at(i, y) == RoadRest))
		{
			addRoad(i, y, CrosswalkH);
			addRoad(i, y + 1, RoadRest);
		}
		i = y + 1;
		//  Keep moving down and painting crosswalks as long as there's road
		while (++i < tilesHigh && !isSidewalk(mapData.at(x, i)))//(mapData.at(x, i) == RoadH || mapData.at(x, i) == RoadRest))
		{
			addRoad(x, i, CrosswalkV);
			addRoad(x - 1, i, RoadRest);
		}
	}
	for (it = sidewalksSE.begin(); it != sidewalksSE.end(); ++it)
	{
		x = it->first;
		y = it->second;
		i = x + 1;
		//  Keep moving right and painting crosswalks as long as there's road
		while (++i < tilesWide && !isSidewalk(mapData.at(i, y)))//(mapData.at(i, y) == RoadV || mapData.at(i, y) == RoadRest))
		{
			addRoad(i, y, CrosswalkH);
			addRoad(i, y + 1, RoadRest);
		}
		i = y + 1;
		//  Keep moving down and painting crosswalks as long as there's road
		while (++i < tilesHigh && !isSidewalk(mapData.at(x, i)))//(mapData.at(x, i) == RoadH || mapData.at(x, i) == RoadRest))
		{
			addRoad(x, i, CrosswalkV);
			addRoad(x + 1, i, RoadRest);
		}
	}
}

void TownGenerator::calculateSidewalkPathing()
{
	ASSERT(valid);
}

void TownGenerator::placeNorthSidewalk(int x, int y)
{
	if (x >= 0 && x < mapData.getWidth())
	{
		mapData.at(x, y) = SidewalkN;
		for (int j = 1; j < sidewalkThickness; ++j)
		{
			mapData.at(x, y + j) = SidewalkInner;
		}
	}
}

void TownGenerator::placeEastSidewalk(int x, int y)
{
	if (y >= 0 && y < mapData.getHeight())
	{
		mapData.at(x, y) = SidewalkE;
		for (int j = 1; j < sidewalkThickness; ++j)
		{
			mapData.at(x - j, y) = SidewalkInner;
		}
	}
}

void TownGenerator::placeWestSidewalk(int x, int y)
{
	if (y >= 0 && y < mapData.getHeight())
	{
		mapData.at(x, y) = SidewalkW;
		for (int j = 1; j < sidewalkThickness; ++j)
		{
			mapData.at(x + j, y) = SidewalkInner;
		}
	}
}

void TownGenerator::placeSouthSidewalk(int x, int y)
{
	if (x >= 0 && x < mapData.getWidth())
	{
		mapData.at(x, y) = SidewalkS;
		for (int j = 1; j < sidewalkThickness; ++j)
		{
			mapData.at(x, y - j) = SidewalkInner;
		}
	}
}

void TownGenerator::addRoad(int x, int y, TileType type)
{
	ASSERT(isRoad(type) || type == CrosswalkH || type == CrosswalkV);
	mapData.at(x, y) = type;
	roadVertexGrid.at(x, y) = &roadEdge;
}

}//towngen namespace

}//sl namespace
