/**
 * @section DESCRIPTION
 * This classes generates towns. Who would've thought?
 */
#ifndef SL_TOWNGEN_TOWNGENERATOR_HPP
#define SL_TOWNGEN_TOWNGENERATOR_HPP

#include <SFML/Graphics.hpp>

#include <functional>
#include <map>
#include <memory>
#include <utility>
#include <vector>

#include "Town/Generation/BlockGenerator.hpp"
#include "Town/Generation/GenerationListener.hpp"
#include "Town/Generation/RequiredBuildings.hpp"
#include "Town/Generation/RoadVertex.hpp"
#include "Town/TownTiles.hpp"
#include "Town/Generation/StaticEntityPrototype.hpp"
#include "Town/Map/MapData.hpp"
#include "Town/Map/MapStaticEntityData.hpp"
#include "Utility/Grid.hpp"
#include "Utility/Random.hpp"
#include "Utility/Rect.hpp"

namespace sl
{

class Town;
class World;

namespace towngen
{

class RoadVertex;

class TownGenerator
{
public:
	/**
	 * Constructs a TownGenerator
	 * @param zonesWide The number of zones (big blocks, excludes subroads) wide city is
	 * @param zonesHigh The number of zones (big blocks, eclludes subroads) high city is
	 * @param zoneWidth How many tiles each zone is wide
	 * @param zoneHeight How many tiles each zone is high
	 * @param random The random number generator to use for generation
	 */
	TownGenerator(int zonesWide, int zonesHigh, int zoneWidth, int zoneHeight, int tileSize, Random& random = Random::get());


	/**
	 * Executes all the code to entirely generate a town
	 */
	void generate(const GenerationListener& listener);
	/**
	 * Creates a town out of the generated town
	 * @param town The town to export to
	 */
	void exportMap(World& world, const GenerationListener& listener);

	const Grid<TileType>& extractTiles() const;
	/**
	 * Checks if the generator has a valid Town generated
	 * @return true if the town can be exported, false if it can't
	 */
	bool isValid() const;

	const map::MapData& getMinimapData() const;

private:
	struct Block
	{
		Block(BlockType type, const Rect<int>& bounds)
			:type(type)
			,bounds(bounds)
		{
		}

		BlockType type;
		Rect<int> bounds;
	};
	/*  Block stuff */
	void resolveIntersections();
	void splitSubroads();
	

	/*  Intersection stuff  */
	/**
	 * Adds an intersection to the list of intersections for later use
	 */
	void addIntersection(int x, int y);
	/**
	 * Goes through the list of intersections and places the proper tiles down.
	 */
	void placeIntersections();
	/**
	 * Creates a road network from interections and adds them to the town
	 */
	void createRoadSystem(Town& town) const;
	/**
	 * Calculates pathing for cars.
	 */
	void calculateCarPathing(Town& town) const;

	/*  Sidewalk stuff  */
	/**
	 * Adds a sidewalk corner tile and places the corner in a list
	 * for later use (ie crosswalk creation, pathing)
	 * @param x The x position of the corner in tiles
	 * @param y The y position of the corner in tiles
	 * @param sidewalkType The type of corner
	 */
	void addSidewalkCorner(int x, int y, TileType sidewalkType);
	/**
	 * Goes through the list of corners from addSidewalkCorner and creates
	 * crosswalk tiles for them all.
	 */
	void placeCrosswalks();
	/**
	 * Calculates NPC pathing.
	 */
	void calculateSidewalkPathing();


	/**
	 * I dunno lol
	 */
	void generateBlock(int x, int y, BlockType type);
	/**
	 * Convenience function to fill an entire zone with one tile type
	 * @param x The x position of the block in zones
	 * @param y The y position of the block in zones
	 * @param fillType The type of tile to use for fill
	 */
	void fillBlock(int x, int y, TileType fillType);
	/**
	 * Generates a road around the block.
	 * @param x The x position of the block in zones
	 * @param y The y position of the block in zones
	 */
	void fillRoad(int x, int y);
	/**
	 * Lists the proportion of neighboring blocks with the given type
	 * @param x The x position of the block in zones
	 * @param y The y position of the block in zones
	 * @param type The type of block to give the proportion of
	 * @param diagWeight How to weight the diagonal neighbors (1 = same as other neighbors)
	 * @return The proportion of neighors of the given type compared to total neighbors, answer is [0,1]
	 */
	float getNeighborBlocks(int x, int y, BlockType type, float diagWeight = 0) const;

	void placeNorthSidewalk(int x, int y);

	void placeEastSidewalk(int x, int y);

	void placeWestSidewalk(int x, int y);

	void placeSouthSidewalk(int x, int y);

	void addRoad(int x, int y, TileType type);



	//! The texture to use for generating minimap previews
	sf::Texture mapTex;
	//! How many zones wide the town is
	const int zonesWide;
	//! How many zones high the town is
	const int zonesHigh;
	//! How many tiles wide each zone is on average
	const int zoneWidth;
	//! How many tiles high each zone is on average
	const int zoneHeight;
	//! How many tiles wide the entire town is
	const int tilesWide;
	//! How many tiles high the entire town is
	const int tilesHigh;
	//! How many pixels big a tile is
	const int tileSize;
	//! The actual widths of each zone
	std::vector<int> zoneWidths;
	//! The actual heights of each zone
	std::vector<int> zoneHeights;
	//! The sum of all zone widths before the current one
	std::vector<int> zoneStartX;
	//! The sum of all zone heights before the current one
	std::vector<int> zoneStartY;
	//! The random number generator to use
	Random& random;
	//! A 2D array of all tiles in the town
	Grid<TileType> mapData;
	//! A 2D array of the types for each zone
	Grid<BlockType> zones;
	//! A list of all static entities in the town (fences, trees)
	std::vector<StaticEntityPrototype> staticEntities;
	//! Whether or not a valid town is currently generated
	bool valid;
	//! A list of all blocks in the town
	std::vector<Block> blocks;
	//! A lookup table of all databases to get BuildingBuildprints from
	std::map<BlockType, BlockGenerator> blockGenerators;
	//! Databases for filler buildings
	//! A count of which buildings are required
	RequiredCountList requiredBuildings;
	//! Used for car+NPC pathing as well as traffic lights and stuff
	std::vector<std::pair<int, int>> intersections;
	//! Used for placing crosswalks
	std::vector<std::pair<int, int>> sidewalksNW;
	//! Used for placing crosswalks
	std::vector<std::pair<int, int>> sidewalksNE;
	//! Used for placing crosswalks
	std::vector<std::pair<int, int>> sidewalksSW;
	//! Used for placing crosswalks
	std::vector<std::pair<int, int>> sidewalksSE;
	//! Record of what kind of static entities have been placed already for minimap gen
	map::MapStaticEntityData minimapStaticEntityData;
	//! Info used to create minimaps (contains mapData/staticEntities/etc)
	map::MapData minimapData;

	std::vector<BlockOfBuildings> buildings;

	//! Which road-intersection-vertex each tile belongs to (used to construct road pathing and road systems)
	Grid<RoadVertex*> roadVertexGrid;
	std::vector<std::unique_ptr<RoadVertex>> roadVertices;

	static RoadVertex roadEdge;
};

}//towngen namespace
}//sl namespace

#endif // SL_TOWNGEN_TOWNGENERATOR_HPP
