#include "Town/Generation/YardBlueprint.hpp"

#include "Engine/Graphics/Sprite.hpp"
#include "Town/Generation/ParkPathGen.hpp"
#include "Town/Generation/PathGenUtil.hpp"
#include "Town/StaticGeometry.hpp"
#include "Town/Map/MapStaticEntityData.hpp"
#include "Town/Town.hpp"
#include "Town/LocationMarker.hpp"
#include "Utility/Random.hpp"

#include <queue>

namespace sl
{

namespace towngen
{

YardBlueprint::YardBlueprint(const std::string& name, int width, int height, FillType fillType, int maxHeight)
	:masterData(new SharedData(name, fillType))
	,data(masterData)
	,width(width)
	,height(height)
	,extraYardWidth(0)
	,extraYardHeight(0)
	//,yardWidthLimit(maxWidth)
	,yardHeightLimit(maxHeight)
{
	//ASSERT(getWidth() <= yardWidthLimit);
	ASSERT(getHeight() <= yardHeightLimit);
}

YardBlueprint::YardBlueprint(const YardBlueprint& other)
	:masterData(nullptr)
	,data(other.data)
	,width(other.width)
	,height(other.height)
	,extraYardWidth(other.extraYardWidth)
	,extraYardHeight(other.extraYardHeight)
	//,yardWidthLimit(other.yardWidthLimit)
	,yardHeightLimit(other.yardHeightLimit)
{
}



int YardBlueprint::adjustYardSize(int widthToAdd, int heightToAdd)
{
	extraYardWidth += widthToAdd;
	ASSERT(extraYardWidth >= 0);
	if (extraYardWidth < 0)
	{
		extraYardWidth = 0;
	}
	const int oldExtraHeight = extraYardHeight;
	extraYardHeight += heightToAdd;
	//ASSERT(getHeight() <= yardHeightLimit);
	if (getHeight() > yardHeightLimit)
	{
		extraYardHeight = yardHeightLimit - height;
	}
	ASSERT(extraYardHeight >= 0);
	if (extraYardHeight < 0)
	{
		extraYardHeight = 0;
	}
	return extraYardHeight - oldExtraHeight;
}

void YardBlueprint::addObject(StaticEntityPrototype::Type type, int min, int max)
{
	ASSERT(masterData);
	masterData->entities[type] = std::make_pair(min, max);
}

void YardBlueprint::addContentGenerator(const Vector2i& offset, std::unique_ptr<ContentGenerator>&& generator, int maxWidth, int maxHeight)
{
	ASSERT(masterData);
	// check that we don't place it outside the bounds of the yard
	ASSERT(generator->getWidth() == ContentGenerator::VARIABLE_WIDTH || offset.x + generator->getWidth() <= getWidth());
	ASSERT(generator->getHeight() == ContentGenerator::VARIABLE_HEIGHT || offset.y + generator->getHeight() <= getHeight());
	masterData->contentGenerators.push_back(SharedData::ContentArea(std::move(generator), offset, maxWidth, maxHeight));
}

void YardBlueprint::addBuilding(BuildingBlueprint&& building)
{
	ASSERT(masterData);
	masterData->buildings.push_back(std::move(building));
}

int YardBlueprint::getWidth() const
{
	return width + extraYardWidth;
}

int YardBlueprint::getHeight() const
{
	return height + extraYardHeight;
}

int YardBlueprint::getMaxHeight() const
{
	return yardHeightLimit;
}

const std::string& YardBlueprint::getName() const
{
	return data->name;
}

void YardBlueprint::generate(const Vector2i& pos, Random& rng, Town& town, Grid<TileType>& tiles, int tileSize, map::MapStaticEntityData& minimapStaticEntityData, bool& isPlayerGenerated, const SoulGenerationSettings& npcSettings) const
{
	const int yardWidth = getWidth();
	const int yardHeight = getHeight();
	
	const int px = pos.x / tileSize;
	const int py = pos.y / tileSize;

	std::vector<Vector2i> pathSpawns;
	for (const BuildingBlueprint& building : data->buildings)
	{
		const Vector2i move(rng.random(yardWidth - building.getWidth()), rng.random(yardHeight - building.getHeight()));
		building.generate(pos + move * tileSize, rng, town, minimapStaticEntityData, isPlayerGenerated, npcSettings);
		for (const Vector2i& pathRelativeToBuilding : building.getPathSpawns())
		{
			pathSpawns.push_back(move + pathRelativeToBuilding);
		}
	}

	for (const SharedData::ContentArea& content : data->contentGenerators)
	{
		const ContentGenerator& generator = *content.content;
		if (generator.isFixedSize())
		{
			generator.generate(pos / tileSize + content.offset, rng, town, tiles, tileSize, minimapStaticEntityData);
		}
		else
		{
			Rect<int> contentArea(px + content.offset.x, py + content.offset.y, generator.getWidth(), generator.getHeight());
			if (contentArea.width == ContentGenerator::VARIABLE_WIDTH)
			{
				contentArea.width = getWidth() - content.offset.x;
				if (content.maxWidth != ContentGenerator::VARIABLE_WIDTH && contentArea.width > content.maxWidth)
				{
					contentArea.width = content.maxWidth;
				}
			}
			if (contentArea.height == ContentGenerator::VARIABLE_HEIGHT)
			{
				contentArea.height = getHeight() - content.offset.y;
				if (content.maxHeight != ContentGenerator::VARIABLE_HEIGHT && contentArea.height > content.maxHeight)
				{
					contentArea.height = content.maxHeight;
				}
			}
			generator.generate(contentArea, rng, town, tiles, tileSize, minimapStaticEntityData);
		}
	}

	for (const auto& entity : data->entities)
	{
		const int n = rng.randomInclusive(entity.second.first, entity.second.second);
		placeEntities(entity.first, TileType::Pavement, pos, rng, town, tiles, tileSize, minimapStaticEntityData, n);
	}

	switch (data->fillType)
	{
		case FillType::None:
			break;
		case FillType::Park:
		case FillType::ParkAbove:
		{
			ParkPathGen paths(Rect<int>(px, py, yardWidth, yardHeight), rng);
			const int pathCount = yardWidth * yardHeight / tileSize / 3;
			for (int i = 0; i < pathCount; ++i)
			{
				paths.walk(ParkPathGen::DirectionalTendency::DrunkenWalk,
					rng.choose({
						ParkPathGen::Direction::North,
						ParkPathGen::Direction::East,
						ParkPathGen::Direction::West,
						ParkPathGen::Direction::South
					}),
					5 + rng.random(yardWidth - 8),
					5 + rng.random(yardHeight - 8));
			}
			paths.generate();
			const Vector2i entrancePos = data->fillType == FillType::Park ? paths.addEntrance() : paths.addTopEntrance();;
			town.addImmediately(new LocationMarker(entrancePos.x * tileSize, (entrancePos.y - 2) * tileSize, "ParkMarker"), StorageType::InList);
			paths.extractTiles(tiles, town);

			placeSomeTrees(pos, rng, town, tiles, tileSize, minimapStaticEntityData, 1 / 12.f);
			
			// Place some benches
			for (int bx = px; bx < px + yardWidth; ++bx)
			{
				for (int by = py; by < py + yardHeight - 1; ++by)
				{
					if (!isParkPath(tiles.at(bx, by)) && isParkPath(tiles.at(bx, by + 1)))
					{
						Rect<int> bBox(bx * tileSize, by * tileSize, 16, 16);
						if (!town.checkCollision(bBox, "StaticGeometry") && rng.chance(0.1f))
						{
							town.addEntityToGrid(new StaticGeometry(bBox.left, bBox.top, bBox.width, bBox.height, unique<Sprite>("bench.png", 2, 0)));
						}
					}
				}
			}


			
			// Try 100 times to place a fountain somewhere
			for (int attempts = 0; attempts < 100; ++attempts)
			{
				const int fx = rng.random(yardWidth - 1) + px;
				const int fy = rng.random(yardHeight - 1) + py;
				//minimapStaticEntityData.set(fx, fy, map::MapStaticEntityData::StaticEntity::Building);
				Rect<int> bBox(fx * tileSize, fy * tileSize, 64, 64);
				if (tiles.at(fx, fy) == Grass && tiles.at(fx + 1, fy) == Grass && tiles.at(fx, fy + 1) == Grass && tiles.at(fx + 1, fy + 1) == Grass &&
				    !town.checkCollision(bBox, "StaticGeometry"))
				{
					town.addEntityToGrid(new StaticGeometry(bBox.left, bBox.top, bBox.width, bBox.height, unique<Sprite>("fountain.png")));
					break;
				}
			}

		}
			break;
		case FillType::Forest:
			placeSomeTrees(pos, rng, town, tiles, tileSize, minimapStaticEntityData, 1 / 4.f);
			break;
		case FillType::LightTrees:
			placeSomeTrees(pos, rng, town, tiles, tileSize, minimapStaticEntityData, 1 / 18.f);
			break;
	}
	const Vector2i posTiles(px, py);
	// all coordinates are relative to the yard until the final conversion step at the end of the method
	Grid<bool> isPath(getWidth(), getHeight(), false);
	// BFS to nearest path OR to the sidewalk (bottom only)
	auto computePath = [&](const Vector2i& start) {
		std::queue<Vector2i> queue;
		queue.push(start);
		// doubles as visited set
		std::map<Vector2i, Vector2i> prev = { std::make_pair(start, start) };
		while (!queue.empty())
		{
			const Vector2i u = queue.front();
			queue.pop();

			static const Vector2i offsets[4] = {
				Vector2i(0, 1),
				Vector2i(1, 0),
				Vector2i(0, -1),
				Vector2i(-1, 0)
			};
			for (const Vector2i& offset : offsets)
			{
				const Vector2i v = u + offset;
				const bool vInBounds = v.x >= 0 && v.x < getWidth() && v.y >= 0 && v.y < getHeight();

				if (isSidewalk(tiles[posTiles + v]) || (vInBounds && isPath[v]))
				{
					for (Vector2i cur = u; prev.at(cur) != cur; cur = prev.at(cur))
					{
						ASSERT(!isPath[cur]);
						isPath[cur] = true;
					}
					return;
				}

				const Rect<int> bBox((px + v.x) * tileSize + 1, (py + v.y) * tileSize + 2, tileSize - 2, tileSize - 3);
				if (vInBounds && !town.checkCollision(bBox, "StaticGeometry"))
				{
					if (prev.insert(std::make_pair(v, u)).second)
					{
						queue.push(v);
					}
				}
			}
		}
		isPath[start] = true;
	};
	for (const Vector2i& spawn : pathSpawns)
	{
		computePath(spawn);
	}
	extractFixedPathTiles(tiles, Walkway, posTiles, isPath, Rect<int>(getWidth(), getHeight()));
}

/*				private					*/

void YardBlueprint::placeSomeTrees(const Vector2i& pos, Random& rng, Town& town, Grid<TileType>& tiles, int tileSize, map::MapStaticEntityData& minimapStaticEntityData, float treeDensity) const
{
	const int treeCount = getWidth() * getHeight() * treeDensity;
	placeEntities(StaticEntityPrototype::Tree, Grass, pos, rng, town, tiles, tileSize, minimapStaticEntityData, treeCount);
}

void YardBlueprint::placeEntities(StaticEntityPrototype::Type type, TileType tileOn, const Vector2i& pos, Random& rng, Town& town, Grid<TileType>& tiles, int tileSize, map::MapStaticEntityData& minimapStaticEntityData, int amount) const
{
	const int yardWidth = getWidth();
	const int yardHeight = getHeight();

	const int ew = StaticEntityPrototype::size(type).x;
	const int eh = StaticEntityPrototype::size(type).y;

	const int px = pos.x / tileSize;
	const int py = pos.y / tileSize;

	for (int i = 0; i < amount; ++i)
	{
		for (int attempts = 0; attempts < 1000; ++attempts)
		{
			const int tx = rng.random(yardWidth + 1 - ew) + px;
			const int ty = rng.random(yardHeight + 1 - eh) + py;

			const Rect<int> bBox((tx - 1) * tileSize, (ty - 1) * tileSize, tileSize * (2 + ew), tileSize * (2 + eh));
			if (tiles.at(tx, ty) == tileOn && !town.checkCollision(bBox, "StaticGeometry"))
			{
				StaticEntityPrototype(tx, ty, type).generate(rng, town, tileSize, minimapStaticEntityData);
				break;
			}
		}
	}
}

YardBlueprint::SharedData::~SharedData()
{
	//ERROR("wtf");
}

}//towngen namespace
}//sl namespace
