#ifndef SL_TOWNGEN_BUILDINGBLUEPRINT_HPP
#define SL_TOWNGEN_BUILDINGBLUEPRINT_HPP

#include <limits>
#include <utility>
#include <map>
#include <vector>

#include "Town/Generation/ContentGenerator.hpp"
#include "Town/Generation/BuildingBlueprint.hpp"
#include "Town/Generation/StaticEntityPrototype.hpp"
#include "Town/TownTiles.hpp"
#include "Utility/Grid.hpp"
#include "Utility/Vector.hpp"

namespace sl
{

class Random;
class SoulGenerationSettings;
class Town;

namespace map
{
class MapStaticEntityData;
} // map

namespace towngen
{

class YardBlueprint
{
public:
	enum class FillType
	{
		None,
		Park,
		ParkAbove,// @todo remove this and figure out a better way to handle park exits
		Forest,
		LightTrees
	};
	YardBlueprint(const std::string& name, int width, int height, FillType fillType = FillType::None, int maxHeight = std::numeric_limits<int>::max());
	
	YardBlueprint(const YardBlueprint& other);


	/**
	 * Adjusts the yard size by the given width/height
	 * @param widthToAdd The amount to extend the yard horizontally (in tiles)
	 * @param heightToAdd The amount to extend the yard vertically (in tiles)
	 * @return The amount successfully extended vertically (since only height has a max height for yards)
	 */
	int adjustYardSize(int widthToAdd, int heightToAdd);

	void addObject(StaticEntityPrototype::Type type, int min, int max);

	void addContentGenerator(const Vector2i& offset, std::unique_ptr<ContentGenerator>&& generator, int maxWidth = ContentGenerator::VARIABLE_WIDTH, int maxHeight = ContentGenerator::VARIABLE_HEIGHT);

	void addBuilding(BuildingBlueprint&& building);


	int getWidth() const;
	int getHeight() const;
	int getMaxHeight() const;

	const std::string& getName() const;

	void generate(const Vector2i& pos, Random& rng, Town& town, Grid<TileType>& tiles, int tileSize, map::MapStaticEntityData& minimapStaticEntityData, bool& isPlayerGenerated, const SoulGenerationSettings& npcSettings) const;

private:
	class SharedData
	{
	public:
		class ContentArea
		{
		public:
			ContentArea(std::unique_ptr<ContentGenerator>&& content, const Vector2i& offset = Vector2i(0, 0), int maxWidth = ContentGenerator::VARIABLE_WIDTH, int maxHeight = ContentGenerator::VARIABLE_HEIGHT)
				:offset(offset)
				,maxWidth(maxWidth)
				,maxHeight(maxHeight)
				,content(std::move(content))
			{
			}
			
			ContentArea(ContentArea&& other)
				:offset(other.offset)
				,maxWidth(other.maxWidth)
				,maxHeight(other.maxHeight)
				,content(std::move(other.content))
			{
			}

			ContentArea& operator=(ContentArea&& rhs)
			{
				offset = rhs.offset;
				maxWidth = rhs.maxWidth;
				maxHeight = rhs.maxHeight;
				content = std::move(rhs.content);
				return *this;
			}


			Vector2i offset;
			int maxWidth;
			int maxHeight;
			std::unique_ptr<ContentGenerator> content;
		private:
			ContentArea(const ContentArea&);
			ContentArea& operator=(const ContentArea&);
		};

		SharedData(const std::string& name, FillType fillType)
			:fillType(fillType)
			,name(name)
		{
		}

		~SharedData();

		std::vector<BuildingBlueprint> buildings;
		//! Static entities to place randomly in a level <Type, <Min, Max>>
		std::map<StaticEntityPrototype::Type, std::pair<int, int>> entities;
		std::vector<ContentArea> contentGenerators;
		FillType fillType;
		std::string name;
	};

	void placeSomeTrees(const Vector2i& pos, Random& rng, Town& town, Grid<TileType>& tiles, int tileSize, map::MapStaticEntityData& minimapStaticEntityData, float treeDensity) const;

	void placeEntities(StaticEntityPrototype::Type type, TileType tileOn, const Vector2i& pos, Random& rng, Town& town, Grid<TileType>& tiles, int tileSize, map::MapStaticEntityData& minimapStaticEntityData, int amount) const;



	SharedData * masterData;
	std::shared_ptr<const SharedData> data;


	/* stuff */
	int width;
	int height;
	
	int extraYardWidth;
	int extraYardHeight;

	//int yardWidthLimit;
	int yardHeightLimit;
};

}//towngen namespace

}//sl namespace

#endif // SL_TOWNGEN_BUILDINGBLUEPRINT_HPP
