#include "Town/StaticGeometry.hpp"

#include "Engine/Serialization/AutoDeserialize.hpp"
#include "Engine/Serialization/AutoSerialize.hpp"
#include "Engine/Scene.hpp"

namespace sl
{

static const char *behindKey = "player_behind";
static const char *notBehindKey = "player_not_behind";

StaticGeometry::StaticGeometry(int x, int y, int w, int h, std::unique_ptr<Drawable> drawable)
	:Entity(x, y, w, h, "StaticGeometry")
	,policy(PlayerBehindPolicy::DoNothing)
	,behindCheckBox()
	,alpha(1.f)
	,serializeThis(true)
{
	if (drawable)
	{
		addDrawable(std::move(drawable));
	}
}

StaticGeometry::StaticGeometry(int x, int y, int w, int h, const Rect<int>& behindCheckBox, std::initializer_list<Drawable*> drawables, float transparencyWhenBehind)
	:Entity(x, y, w, h, "StaticGeometry")
	,policy(PlayerBehindPolicy::BecomeTransparent)
	,behindCheckBox(behindCheckBox)
	,alpha(transparencyWhenBehind)
	,serializeThis(true)
{
	for (Drawable* drawable : drawables)
	{
		addDrawable(std::unique_ptr<Drawable>(drawable));
	}
}

StaticGeometry::StaticGeometry(int x, int y, int w, int h, const Rect<int>& behindCheckBox, std::initializer_list<Drawable*> drawables, std::initializer_list<Drawable*> drawablesWhenBehind)
	:Entity(x, y, w, h, "StaticGeometry")
	,policy(PlayerBehindPolicy::AlternativeDrawables)
	,behindCheckBox(behindCheckBox)
	,alpha(1.f)
	,serializeThis(true)
{
	for (Drawable* drawable : drawables)
	{
		this->addDrawableToSet(notBehindKey, std::unique_ptr<Drawable>(drawable));
	}
	for (Drawable* drawable : drawablesWhenBehind)
	{
		this->addDrawableToSet(behindKey, std::unique_ptr<Drawable>(drawable));
	}
	this->setDrawableSet(notBehindKey);
}

const Types::Type StaticGeometry::getType() const
{
	return Entity::makeType("StaticGeometry");
}

bool StaticGeometry::isStatic() const
{
	return true;
}

void StaticGeometry::doNotSerialize()
{
	serializeThis = false;
}

void StaticGeometry::serialize(Serialize& out) const
{
	serializeEntity(out);
	serializeDrawables(out);
	AutoSerialize::serialize(out, mask.getBoundingBox());
}

void StaticGeometry::deserialize(Deserialize& in)
{
	deserializeEntity(in);
	deserializeDrawables(in);
	Rect<int> bounds;
	AutoDeserialize::deserialize(in, bounds);
	mask.set(bounds.left, bounds.top, bounds.width, bounds.height);
}

std::string StaticGeometry::serializationType() const
{
	return "StaticGeometry";
}

bool StaticGeometry::shouldSerialize() const
{
	return serializeThis;
}

void StaticGeometry::onUpdate()
{
	if (policy != PlayerBehindPolicy::DoNothing)
	{
		bool playerIsBehind = scene->checkCollision(behindCheckBox, "Player") != nullptr;
		if (policy == PlayerBehindPolicy::BecomeTransparent)
		{
		}
		else if (policy == PlayerBehindPolicy::AlternativeDrawables)
		{
			if (playerIsBehind)
			{
				this->setDrawableSet(behindKey);
			}
			else
			{
				this->setDrawableSet(notBehindKey);
			}
		}
	}
}

} // sl