#include "Town/Town.hpp"

#include <cmath>

#include "Engine/Game.hpp"
#include "Engine/Serialization/TileSerialization.hpp"
#include "Engine/Input/Input.hpp"
#include "Town/World.hpp"
#include "Engine/Graphics/Sprite.hpp"
#include "Engine/Graphics/TextureManager.hpp"
#include "NPC/NPC.hpp"
#include "NPC/States/IdleState.hpp"
#include "NPC/States/LoliRoutineState.hpp"
#include "NPC/States/KidnappedState.hpp"
#include "NPC/States/Police/CopPatrolState.hpp"
#include "NPC/States/AdultWanderState.hpp"
#include "NPC/Stats/NationalityDatabase.hpp"
#include "NPC/Stats/SoulGenerationSettings.hpp"
#include "NPC/ClothesDatabase.hpp"
#include "Pathfinding/CarConfig.hpp"
#include "Player/Player.hpp"
#include "Town/Cars/Car.hpp"
#include "Town/Generation/StaticEntityPrototype.hpp"
#include "Utility/Memory.hpp"
#include "Utility/Random.hpp"

#include <SFML/Graphics/RenderTexture.hpp>

namespace sl
{

// @TODO move?
void serialize(Serialize& out, const TileType& val)
{
	out.u8("", static_cast<uint8_t>(val));
}

void deserialize(Deserialize& in, TileType& val)
{
	val = static_cast<TileType>(in.u8(""));
}

Town::Town(Game *game, unsigned int width, unsigned int height, unsigned int viewWidth, unsigned int viewHeight)
	:Scene(game, width, height, viewWidth, viewHeight, "Town")
	,tiles()
	,tileTexture(TextureManager::getTexture("tiles/outside_tiles.png"))
	,tileGrid(0, 0, width / getTileSize(), height / getTileSize(), getTileSize(), getTileSize())
	,logger(getGame()->getLogger())
	,roadPathManager(getTileSize())
	,weatherMask(sf::Vector2f(viewWidth, viewHeight))
{
	setTileTexture(Dirt, { Vector2i(1, 3), Vector2i(1, 4) });
	setTileTexture(Grass, /*{ Vector2i(0, 6), Vector2i(1, 6) });*/{ Vector2i(0, 3), Vector2i(0, 4), Vector2i(0, 5) });
	setTileTexture(Pavement, Vector2i(1, 1));
	setTileTexture(RoadH, Vector2i(1, 0));
	setTileTexture(RoadV, Vector2i(0, 1));
	setTileTexture(RoadRest, Vector2i(1, 1));
	setTileTexture(SidewalkN, Vector2i(4, 2));
	setTileTexture(SidewalkS, Vector2i(4, 0));
	setTileTexture(SidewalkW, Vector2i(5, 1));
	setTileTexture(SidewalkE, Vector2i(3, 1));
	setTileTexture(SidewalkNW, Vector2i(6, 0));
	setTileTexture(SidewalkNE, Vector2i(8, 0));
	setTileTexture(SidewalkSW, Vector2i(6, 2));
	setTileTexture(SidewalkSE, Vector2i(8, 2));
	setTileTexture(SidewalkNWI, Vector2i(3, 0));
	setTileTexture(SidewalkNEI, Vector2i(5, 0));
	setTileTexture(SidewalkSWI, Vector2i(3, 2));
	setTileTexture(SidewalkSEI, Vector2i(5, 2));
	setTileTexture(SidewalkInner, Vector2i(7, 1));
	setTileTexture(Intersection, Vector2i(1, 1));
	setTileTexture(CrosswalkH, Vector2i(10, 2));
	setTileTexture(CrosswalkV, Vector2i(9, 2));
	setTileTexture(DrivewayC, Vector2i(1, 1));
	setTileTexture(DrivewayL, Vector2i(0, 9)); // fix
	setTileTexture(DrivewayR, Vector2i(0, 9)); // fix
	setTileTexture(DrivewayJoinL, Vector2i(0, 9)); // fix
	setTileTexture(DrivewayJoinR, Vector2i(0, 9)); // fix
	setTileTexture(Walkway, Vector2i(11, 2));
	setTileTexture(ParkPathN, Vector2i(7, 5));
	setTileTexture(ParkPathE, Vector2i(6, 4));
	setTileTexture(ParkPathW, Vector2i(8, 4));
	setTileTexture(ParkPathS, Vector2i(7, 3));
	setTileTexture(ParkPathNS, Vector2i(3, 4));
	setTileTexture(ParkPathEW, Vector2i(4, 3));
	setTileTexture(ParkPathNE, Vector2i(3, 5));
	setTileTexture(ParkPathNW, Vector2i(5, 5));
	setTileTexture(ParkPathES, Vector2i(3, 3));
	setTileTexture(ParkPathWS, Vector2i(5, 3));
	setTileTexture(ParkPathNEW, Vector2i(10, 3));
	setTileTexture(ParkPathNWS, Vector2i(9, 4));
	setTileTexture(ParkPathNES, Vector2i(11, 4));
	setTileTexture(ParkPathEWS, Vector2i(10, 5));
	setTileTexture(ParkPathNEWS, Vector2i(7, 4));
	setTileTexture(ParkingLotL, Vector2i(11, 0));
	setTileTexture(ParkingLotR, Vector2i(12, 0));
	setTileTexture(ParkingLotM, Vector2i(11, 1));
	setTileTexture(HACK, Vector2i(0, 9)); // fix

	//registerActivity(unique<PoliceRadio>());
	
	// @todo make a NPC non-cop driver state AI
	//for (int i = 0; i < 30; ++i)
	//{
	//	int x, y;
	//	TileType t;
	//	for (int j = 0; j < 100000; ++j)
	//	{
	//		x = Random::get().random(width);
	//		y = Random::get().random(height);
	//		t = tiles.at(x / 32, y / 32);
	//		if (t == RoadN || t == RoadS || t == RoadW || t == RoadE || t == RoadRest || t == Intersection || t == CrosswalkH || t == CrosswalkV)
	//		{
	//			 
	//			if (!this->checkCollision(Rect<int>(x, y, 32, 32), Entity::makeType("Car")))
	//			{
	//				const CarSpecs::Model model = Random::get().choose({
	//					CarSpecs::Van,         CarSpecs::Van,         CarSpecs::Van,
	//					CarSpecs::SportsCar,   CarSpecs::SportsCar,   CarSpecs::SportsCar,   CarSpecs::SportsCar,
	//					CarSpecs::LuxerySedan, CarSpecs::LuxerySedan, CarSpecs::LuxerySedan, CarSpecs::LuxerySedan,
	//					CarSpecs::Schoolbus
	//				});
	//				Car *car = new Car(x, y, model, this, false);
	//				addImmediately(car, StorageType::InQuad);
	//				const NationalityDatabase::Nationality& nationality = Random::get().choose(NationalityDatabase::get().getNationalities(DNADatabase::GenePool::Caucasian));
	//				NPC *npc = new NPC(x, y, shared<Soul>(Random::get(), nationality));
	//				npc->getSoul()->setType("cop");
	//				npc->getSoul()->setClothing("uniform");
	//				npc->initStates(new CopPatrolState(*npc, *car));
	//				addImmediately(npc, StorageType::InQuad);
	//				break;
	//			}
	//		}
	//	}
	//}
}

Town::~Town()
{
	pf::CarConfig::setTown(nullptr);
}

void Town::onUpdate(int64_t timestamp)
{
	updateEntities();

	if (input.isKeyHeld(Input::KeyboardKey::L))
	{
		if (input.isMousePressed(Input::MouseButton::Left))
		{
			if (input.isKeyHeld(Input::KeyboardKey::LShift) || input.isKeyHeld(Input::KeyboardKey::RShift))
			{
				static const char *poolNames[DNADatabase::GENE_POOL_SIZE] = {"white", "nordic", "med", "asian", "hisp", "trop", "nig", "ind", "sea", "ar"};
				static int poolIndex = 0;
				const DNADatabase::GenePool pool = static_cast<DNADatabase::GenePool>(poolIndex);
				Logger::log(Logger::Area::Town, Logger::Priority::Debug) << "spawning lolis from pool" << poolNames[poolIndex] << std::endl;
				poolIndex = (poolIndex + 1) % DNADatabase::GENE_POOL_SIZE;
				for (int i = 0; i < 8; ++i)
				{
					const int rad = 128;
					const int x = this->getMousePos().x - rad / 2 + Random::get().random(rad), y = this->getMousePos().y - rad / 2 + Random::get().random(rad);
					SoulGenerationSettings genSettings;
					genSettings.addGenePool(pool);
					NPC *obj = new NPC(x, y, getWorld()->createSoul(Random::get(), genSettings));
					ClothesDatabase::get().dressLoliAppropriately(Random::get(), obj->getSoul()->appearance);
					obj->initStates(unique<IdleState>(*obj));
					addEntityToQuad(obj);
				}
			}
			else
			{
				const int x = this->getMousePos().x, y = this->getMousePos().y;
				logger.log(Logger::Town) << "creating Loli at (" << x << ", " << y << ")" << std::endl;
				NPC *obj = new NPC(x, y, getWorld()->createSoul(Random::get(), SoulGenerationSettings().setAgeRange(4,12)));
				/*if (Random::get().chance(0.3f))
				{
					obj->getSoul()->setClothing(ClothingType::Accessory, "ghost");
				}
				else
				{
					obj->getSoul()->setClothing(ClothingType::Accessory, "witch_hat");
					obj->getSoul()->setClothing(ClothingType::Outfit, "witch_robe");
					obj->getSoul()->setClothingColour(ClothingType::Accessory, obj->getSoul()->getClothingColour(ClothingType::Outfit));
				}*/
				obj->initStates(unique<IdleState>(*obj));
				addEntityToQuad(obj);
			}
		}
	}

	updateLighting();
}

void Town::onDraw(RenderTarget& target)
{
	target.draw(tileGrid);

	if (getGame()->getSettings().get("draw_car_path_graph"))
	{
		roadPathManager.draw(target);
	}

	// @todo @hack fix culling properly instead of hacking the cull rect bigger for demo day (180 pixels is the height of the appartments which would get cut off)
	const Rect<int> cullRect(viewPos.x - viewWidth / 2, viewPos.y - viewHeight / 2 - 64, viewWidth, viewHeight + 180 + 64);
	drawEntities(target, cullRect, getGame()->getRawInput().mousePosF() + viewPos - sf::Vector2f(viewWidth / 2, viewHeight / 2));

	target.draw(weatherMask, sf::RenderStates(sf::BlendAlpha));
}


void Town::setView(int x, int y)
{
	Scene::setView(x, y);
	const int halfViewWidth = viewWidth / 2;
	const int halfViewHeight = viewHeight / 2;
	tileGrid.setVisibleArea(Rect<int>(viewPos.x - halfViewWidth, viewPos.y - halfViewHeight, viewWidth, viewHeight));
}

pf::VertexBasedPathManager& Town::getRoadPathManager()
{
	return roadPathManager;
}

const pf::VertexBasedPathManager& Town::getRoadPathManager() const
{
	return roadPathManager;
}

void Town::setTiles(const Grid<TileType>& tiles)
{
	static sf::RenderTexture debugTiles;
	const int tw = tileTexture.get().getSize().x;
	const int th = tileTexture.get().getSize().y;
	debugTiles.create(tw, th);
	debugTiles.draw(sf::Sprite(tileTexture.get()));
	sf::Vertex line[2];
	line[0].color = line[1].color = sf::Color::Cyan;
	for (int i = 0; i < tw; i += 16)
	{
		line[0].position.x = line[1].position.x = i;
		line[0].position.y = 0;
		line[1].position.y = th - 1;
		debugTiles.draw(line, 2, sf::PrimitiveType::Lines);
	}
	for (int i = 0; i < th; i += 16)
	{
		line[0].position.y = line[1].position.y = i;
		line[0].position.x = 0;
		line[1].position.x = tw - 1;
		debugTiles.draw(line, 2, sf::PrimitiveType::Lines);
	}
	debugTiles.display();
	if (&this->tiles != &tiles)
	{
		// this could not be hit if we're deserializing, but we still want to run the rest
		this->tiles = tiles;
	}
	for (auto it = tiles.begin(), end = tiles.end(); it != end; ++it)
	{
		const auto tex = tileRects.find(*it);
		if (tex != tileRects.end())
		{
			const sf::IntRect& rect = Random::get().choose(tex->second);
			
			//tileGrid.setTexture(it.getX(), it.getY(), debugTiles.getTexture(), rect);
			tileGrid.setTexture(it.getX(), it.getY(), tileTexture.get(), rect);
		}
		else
		{
			//ERROR("unknown tile texture");
		}
	}
}

TileType Town::getTileAt(int x, int y) const
{
	return tiles.at(x / getTileSize(), y / getTileSize());
}

TileType Town::getTileAtInTileCoords(int tx, int ty) const
{
	return tiles.at(tx, ty);
}

void Town::spawnAdults(int adultCount)
{
	ASSERT(getWorld() != nullptr);
	for (int i = 0; i < adultCount; ++i)
	{
		int x, y;
		TileType t;
		for (int attempts = 0; attempts < 10000; ++attempts)
		{	
			x = Random::get().random(width);
			y = Random::get().random(height);
			t = tiles.at(x / getTileSize(), y / getTileSize());
			if ((isSidewalk(t) || isRoad(t)) && !checkCollision(Rect<int>(x - 64, y - 64, 128, 128), "StaticGeometry"))
			{
				NPC *npc = new NPC(x, y, getWorld()->createSoul(Random::get(), SoulGenerationSettings()));
				npc->initStates(unique<AdultWanderState>(*npc));
				addImmediately(npc, StorageType::InQuad);
				break;
			}
		}
	}
}

Vector2f Town::getRandomSidewalkTile(Random& rng) const
{
	// TODO: optimize?
	int x, y;
	TileType t;
	for (int attempts = 0; attempts < 100000; ++attempts)
	{
		x = Random::get().random(width);
		y = Random::get().random(height);
		t = tiles.at(x / getTileSize(), y / getTileSize());
		if (isSidewalk(t))
		{
			return Vector2f(x, y);
		}
	}
	ERROR("well that was unlucky :(");
	return Vector2f(0.f, 0.f);
}

void Town::serialize(Serialize& out) const
{
	out.startObject("Town");
	this->serializeScene(out);
	serializeTiles(out, tiles);
	roadPathManager.serialize(out);
	// rest are ignored...
	out.endObject("Town");
}

void Town::deserialize(Deserialize& in)
{
	in.startObject("Town");
	this->deserializeScene(in);
	deserializeTiles(in, tiles);
	tileGrid.set(0, 0, width / getTileSize(), height / getTileSize(), getTileSize(), getTileSize());
	setTiles(tiles);
	roadPathManager.deserialize(in);
	in.endObject("Town");

	init();
}

std::string Town::serializationType() const
{
	return "Town";
}

bool Town::shouldSerialize() const
{
	return true;
}



/*		Protected methods			*/
void Town::onDrawGUI(RenderTarget& target)
{
}


/*	  	Private methods				 */

void Town::handleEvent(const Event& event)
{
	switch (event.getCategory())
	{
	case Event::Category::Debug:
		if (event.getName() == "spawn_car")
		{
			Player *player = getPlayer();
			if (player)
			{
				Car *car = new Car(player->getPos().x, player->getPos().y, static_cast<CarSpecs::Model>(event.getFlag("model").asInt()), this, true);
				addEntityToQuad(car);
			}
		}
		break;
	default:
		break;
	}
}

void Town::onWorldAdd()
{
	init();
}

void Town::updateLighting()
{
	weatherMask.setPosition(getView().x - viewWidth / 2, getView().y - viewHeight / 2);
	int maskAlpha;
	const float maxValue = 0.85f;
	const float minValue = 0.f;
	const int hour = getWorld()->getGameTime().time.getHour();
	const int minute = getWorld()->getGameTime().time.getMinutes();
	const int sunriseStart = 6;
	const int sunriseEnd = 8;
	const int sunsetStart = 18;
	const int sunsetEnd = 20;
	if (hour >= sunriseStart && hour < sunriseEnd)
	{
		// progression of sunrise - gives value between 0.f and 1.f
		const float theta = pi * ((hour - sunriseStart) * 60 + minute) / (60.f * (sunriseEnd - sunriseStart));
		const float multiplier = (cos(theta) + 1.f) / 2.f;
		maskAlpha = 255 * multiplier * ((maxValue - minValue) + minValue);
	}
	else if (hour >= sunsetStart && hour < sunsetEnd)
	{
		// progression of sunset - gives value between 0.f and 1.f
		const float theta = pi * ((hour - sunsetStart) * 60 + minute) / (60.f * (sunsetEnd - sunsetStart));
		const float multiplier = (-1.f * cos(theta) + 1.f) / 2.f;
		maskAlpha = 255 * multiplier * ((maxValue - minValue) + minValue);
	}
	else if (hour >= sunriseEnd && hour < sunsetStart)
	{
		maskAlpha = 255 * minValue;
	}
	else
	{
		maskAlpha = 255 * maxValue;
	}
	weatherMask.setFillColor(sf::Color(16, 20, 32, maskAlpha));
}

void Town::setTileTexture(TileType type, const Vector2i& offsetInTiles)
{
	const int ts = getTileSize();
	tileRects[type].push_back(sf::IntRect(offsetInTiles.x * ts, offsetInTiles.y * ts, ts, ts));
}

void Town::setTileTexture(TileType type, std::initializer_list<Vector2i> offsetsInTiles)
{
	for (const Vector2i& offset : offsetsInTiles)
	{
		setTileTexture(type, offset);
	}
}

void Town::init()
{
	// make sure the car configurations know we're in this town. there should only be one town.
	pf::CarConfig::setTown(this);

	//createAndRegisterRealTimePathGrid({"StaticGeometry"}, "StaticGeometry");
	createAndRegisterPrecomputedPathGrid("StaticGeometry", "StaticGeometry");
}

}// End of sl namespace
