/**
 * @section DESCRIPTION
 * This is the class to represent the Scene that the player will kidnap lolis in.
 */
#ifndef SL_TOWN_HPP
#define SL_TOWN_HPP

#include <SFML/Graphics.hpp>

#include <initializer_list>
#include <map>

#include "Engine/Graphics/TileGrid.hpp"
#include "Engine/Graphics/Texture.hpp"
#include "Engine/GUI/DrawableContainer.hpp"
#include "Engine/GUI/WidgetHandle.hpp"
#include "Engine/Scene.hpp"
#include "Pathfinding/VertexBasedPathManager.hpp"
#include "Town/TownTiles.hpp"
#include "Utility/Grid.hpp"

namespace sl
{

class Logger;
//class World;
class Game;

namespace map
{
class MapData;
class Map;
class Icon;
} // map

class Town : public Scene
{
public:
	/**
	 * Constructs a Town
	 * @param game The Game this level exists in
	 * @param width How wide the Town will be
	 * @param height How high the Town will be
	 */
	Town(Game *game, unsigned int width, unsigned int height, unsigned int viewWidth, unsigned int viewHeight);
	/**
	 * Destructs the Town
	 */
	~Town();

	/* AbstractScene interface */
	/**
	 * Draws an overlay over the Town.
	 * @param target The RenderTarget to draw to
	 */
	void onDrawGUI(RenderTarget& target) override;
	/**
	 * Changes where the camera looks in the Town.
	 * @param x The x position
	 * @param y The y position
	 */
	void setView(int x, int y) override;


	pf::VertexBasedPathManager& getRoadPathManager();

	const pf::VertexBasedPathManager& getRoadPathManager() const;

	void setTiles(const Grid<TileType>& tiles);

	/**
	 * @param x The position (in worldspace, not tile!) to query
	 * @param y The position (in worldspace)
	 * @return The Tile type at that position
	 */
	TileType getTileAt(int x, int y) const;

	TileType getTileAtInTileCoords(int tx, int ty) const;

	/**
	 * Spawns some adults to patrol the town (DO NOT CALL BE FORE THE TOWN IS FULLY GENERATED!)
	 */
	void spawnAdults(int adultCount);

	Vector2f getRandomSidewalkTile(Random& rng) const;




	void serialize(Serialize& out) const override;

	void deserialize(Deserialize& in) override;

	std::string serializationType() const override;

	bool shouldSerialize() const override;

private:
	/**
	 * Updates the Town. Called automatically by AbstractScene::update()
	 * @param timestamp The current time (in seconds since game start)
	 */
	void onUpdate(int64_t timestamp) override;
	/**
	 * Draws the Town
	 * @param target The RenderTarget to draw to
	 */
	void onDraw(RenderTarget& target) override;

	void handleEvent(const Event& event) override;

	void onWorldAdd() override;

	void updateLighting();

	void setTileTexture(TileType type, const Vector2i& offsetInTiles);

	void setTileTexture(TileType type, std::initializer_list<Vector2i> offsetsInTiles);

	/**
	 * Stuff to call on either onWorldAdd() or after deserialization
	 */
	void init();



	Grid<TileType> tiles;
	std::map<TileType, std::vector<sf::IntRect>> tileRects;
	Texture tileTexture;
	TileGrid tileGrid;
	const Logger& logger;
	pf::VertexBasedPathManager roadPathManager;
	sf::RectangleShape weatherMask;
};

}// End of sl namespace

#endif
