#include "Town/TrafficLight.hpp"

#include "Engine/Graphics/Animation.hpp"
#include "Engine/Graphics/Sprite.hpp"
#include "Engine/Scene.hpp"
#include "Utility/Assert.hpp"

namespace sl
{

TrafficLight::TrafficLight(int x, int y, Direction direction, bool canGoLeft)
	:Entity(x, y, 4, 13, "TrafficLight")
	,straightAvailabilityChecker(false)
	,leftAvailabilityChecker(false)
	,direction(direction)
	,canGoLeft(canGoLeft)
	,cachedLeftLoad(-1)
	,cachedStraightLoad(-1)
	,cachedRightLoad(-1)
	,polePos(x, y)
{
	switch (direction)
	{
	case North:
		this->setPos(pos.x - 34, pos.y - 25);
		break;
	case South:
		this->setPos(pos.x + 32, pos.y - 24);
		break;
	case West:
		this->setPos(pos.x, pos.y + 10);
		break;
	case East:
		this->setPos(pos.x - 2, pos.y - 46);
		break;
	}
}


const Types::Type TrafficLight::getType() const
{
	return makeType("TrafficLight");
}

bool TrafficLight::isStatic() const
{
	return true;
}

const pf::NodeAvailabilityChecker& TrafficLight::getStraightChecker() const
{
	return straightAvailabilityChecker;
}

const pf::NodeAvailabilityChecker& TrafficLight::getLeftChecker() const
{
	return leftAvailabilityChecker;
}

void TrafficLight::setColour(Colour colour)
{
	switch (colour)
	{
		case Green:
			this->setDrawableSet("green");
			straightAvailabilityChecker.setAvailability(true);
			break;
		case Yellow:
			this->setDrawableSet("yellow");
			straightAvailabilityChecker.setAvailability(false);
			break;
		case Red:
			this->setDrawableSet("red");
			straightAvailabilityChecker.setAvailability(false);
			break;
	}
}

void TrafficLight::setLeftColour(Colour colour)
{
	//ASSERT(canGoLeft);
	switch (colour)
	{
		case Green:
			this->setDrawableSet("green");
			leftAvailabilityChecker.setAvailability(true);
			break;
		case Yellow:
			this->setDrawableSet("yellow");
			leftAvailabilityChecker.setAvailability(false);
			break;
		case Red:
			this->setDrawableSet("red");
			leftAvailabilityChecker.setAvailability(false);
			break;
	}
}

int TrafficLight::getLeftLoad() const
{
	if (cachedLeftLoad == -1)
	{
		computeLoads();
	}
	return cachedLeftLoad;
}

int TrafficLight::getStraightLoad() const
{
	if (cachedStraightLoad == -1)
	{
		computeLoads();
	}
	return cachedStraightLoad;
}

int TrafficLight::getRightLoad() const
{
	if (cachedRightLoad == -1)
	{
		computeLoads();
	}
	return cachedRightLoad;
}


/*				private					*/
void TrafficLight::onUpdate()
{
	// flush cached loads
	cachedLeftLoad = -1;
	cachedStraightLoad = -1;
	cachedRightLoad = -1;
}

void TrafficLight::onEnter()
{
	createDrawables();
	scene->addEntityToGridNoUpdate(new Pole(polePos.x, polePos.y, direction));
}

void TrafficLight::onExit()
{
	//	maybe remove that pole? but who would move traffic lights between worlds...?
}

void TrafficLight::computeLoads() const
{
	// TODO: write this
	cachedLeftLoad = 1;
	cachedStraightLoad = 2;
	cachedRightLoad = 1;
}

void TrafficLight::createDrawables()
{
	std::string fname;
	switch (direction)
	{
	case North:
		fname = "north";
		break;
	case South:
		fname = "south";
		break;
	case West:
		fname = "west";
		break;
	case East:
		fname = "east";
		break;
	}
	this->addDrawableToSet("green", std::make_unique<Sprite>(("trafficlight_" + fname + "_green.png").c_str()));
	this->addDrawableToSet("yellow", std::make_unique<Sprite>(("trafficlight_" + fname + "_yellow.png").c_str()));
	this->addDrawableToSet("red", std::make_unique<Sprite>(("trafficlight_" + fname + "_red.png").c_str()));
	this->setDrawableSet("green");
}

void TrafficLight::deserialize(Deserialize& in)
{
	deserializeEntity(in);
	direction = static_cast<Direction>(in.u8("dir"));
	LOADVAR(b, canGoLeft);
	LOADVAR(i, polePos.x);
	LOADVAR(i, polePos.y);
	// drawables + Pole should get created via onEnter()
}

void TrafficLight::serialize(Serialize& out) const
{
	serializeEntity(out);
	out.u8("dir", static_cast<uint8_t>(direction));
	SAVEVAR(b, canGoLeft);
	SAVEVAR(i, polePos.x);
	SAVEVAR(i, polePos.y);
}

/*		Pole		*/
TrafficLight::Pole::Pole(int x, int y, Direction direction)
	:Entity(x, y, 2, 2, "TrafficLightPole")
{
	int xoffset = 0;
	int yoffset = 0;
	std::string fname;
	switch (direction)
	{
		case North:
			fname = "north";
			xoffset = 34;
			yoffset = 35;
			break;
		case South:
			fname = "south";
			xoffset = 28;
			yoffset = 27;
			break;
		case West:
			fname = "west";
			xoffset = 34;
			yoffset = 28;
			break;
		case East:
			fname = "east";
			xoffset = 28;
			yoffset = 66;
			break;
	}
	this->addDrawable(std::make_unique<Sprite>(("trafficlight_pole_" + fname + ".png").c_str(), xoffset, yoffset));
}

const Types::Type TrafficLight::Pole::getType() const
{
	return Entity::makeType("TrafficLightPole");
}

bool TrafficLight::Pole::isStatic() const
{
	return true;
}


void TrafficLight::Pole::onUpdate()
{
}


} 