#include "Town/World.hpp"

#include "Engine/Game.hpp"
#include "Engine/Scene.hpp"
#include "Engine/Serialization/AutoDeserialize.hpp"
#include "Engine/Serialization/AutoSerialize.hpp"
#include "Engine/Serialization/Serialization.hpp"
#include "NPC/ScheduleGeneration.hpp"
#include "NPC/Stats/Soul.hpp"
#include "Pathfinding/GridPathManager.hpp"
#include "Utility/Assert.hpp"
#include "Utility/IteratorUtil.hpp"
#include "Utility/Memory.hpp"
#include "Utility/Random.hpp"

#ifndef SL_NOMULTITHREADED
	#include <atomic>
	#include <thread>
#endif // SL_NOMULTITHREADED

namespace sl
{

World::World(Game *game, int viewWidth, int viewHeight, const GameTime& time)
	:AbstractScene(game, viewWidth, viewHeight)
	,scenes()
	,currentScene(nullptr)
	,time(time)
	,secondsPerUpdate(4.f) // 4 seconds per update, so 2 minutes per second and 2 hours per minute
	,timeResidue(0.f)
	,player(nullptr)
{
}

World::~World()
{
	// cancel pathmanagers first since they could be accessing things from the world's scenes
	// for collision detection and stuff and can be run on separate threads
	for (auto& manager : pathManagers)
	{
		manager.second->cancelAllCalculations();
	}
}

void World::addScene(std::unique_ptr<AbstractScene> scene, bool alwaysUpdate)
{
	ASSERT(scene.get() != this);
	ASSERT(scene != nullptr && scene->world == nullptr);
	scene->world = this;
	if (scenes.empty())
	{
		ASSERT(currentScene == nullptr);
		currentScene = scene.get();
	}
	scene->onWorldAdd();
	ASSERT(std::find_if(scenes.begin(), scenes.end(), [s = scene.get()](const auto& p) { return p.first.get() == s; }) == scenes.end());
	scenes.push_back(std::make_pair(std::move(scene), alwaysUpdate));
}

void World::setCurrentScene(AbstractScene *scene)
{
	ASSERT(scene != nullptr);
	for (auto& s : scenes)
	{
		if (scene == s.first.get())
		{
			currentScene = scene;
			return;
		}
	}
	ERROR("Cannot set current scene because the AbstractScene is not in this World.");
}

const sf::Vector2f World::getView() const
{
	if (currentScene)
	{
		return currentScene->getView();
	}
	else
	{
		return sf::Vector2f(0, 0);
	}
}


void World::setView(int x, int y)
{
	if (currentScene)
	{
		currentScene->setView(x, y);
	}
}

const GameTime& World::getGameTime() const
{
	return time;
}

//void World::setTimeRate(float secondsPerUpdate)
//{
//	this->secondsPerUpdate = secondsPerUpdate;
//}

int World::getTimeRate() const
{
	return secondsPerUpdate;
}

PoliceForce& World::getPolice()
{
	return police;
}

SchoolRegistry& World::getSchoolRegistry()
{
	return schoolRegistry;
}

std::shared_ptr<Soul> World::getSoul(uint64_t id)
{
	const auto it = registry.find(id);
	if (it != registry.end())
	{
		return it->second;
	}
	return nullptr;
}

std::shared_ptr<Soul> World::createSoul(Random& rng, const SoulGenerationSettings& settings)
{
	auto soul = shared<Soul>(rng, settings);
	recordBirth(soul);
	return soul;
}

void World::recordBirth(std::shared_ptr<Soul> baby)
{
	ASSERT(registry.count(baby->getID()) == 0);
	registry.insert(std::make_pair(baby->getID(), baby));
}

pf::PathManager* World::getPathManager(const std::string& name) const
{
	auto it = pathManagers.find(name);
	if (it != pathManagers.end())
	{
		return it->second.get();
	}
	ERROR("Could not find path manager'" + name + "' :(");
	return nullptr;
}

void World::registerPathManager(const std::string& name, std::shared_ptr<pf::PathManager> pathManager)
{
	pathManagers[name] = pathManager;
}

void World::addGridToPathManager(const std::string& name, Scene *scene, std::shared_ptr<pf::PathGrid> pathGrid)
{
	auto& manager = pathManagers[name];
	if (!manager)
	{
		manager = shared<pf::GridPathManager>();
	}
	auto gridManager = dynamic_cast<pf::GridPathManager*>(manager.get());
	ASSERT(gridManager);
	if (gridManager)
	{
		gridManager->registerGrid(scene, pathGrid);
	}
}

void World::advanceDate(int daysPassed, bool timeSkip)
{
	time.date.advance(daysPassed);
	for (int i = 0; i < daysPassed; ++i)
	{
		for (auto& npc : registry)
		{
			// TODO: something better than this? might be weird for stuff like food/etc if it's all done at once
			for (int h = 0; h < 24; ++h)
			{
				npc.second->hourlyUpdate(*this, Random::get(), timeSkip);
			}
			npc.second->ageOneDay();
		}
	}
}

void World::serialize(Serialize& out) const
{
	ASSERT(registry.size() < 0x7FFFFFFF);
	out.startArray("registry", registry.size());
	for (const auto& soul : registry)
	{
		soul.second->serialize(out);
	}
	out.endArray("registry");
	int sceneCount = 0;
	for (const std::pair<std::unique_ptr<AbstractScene>, bool>& scene : scenes)
	{
		if (scene.first->shouldSerialize())
		{
			++sceneCount;
		}
	}
	out.startArray("scenes", sceneCount);
	for (const std::pair<std::unique_ptr<AbstractScene>, bool>& scene : scenes)
	{
		if (scene.first->shouldSerialize())
		{
			out.b("active", scene.second);
			out.saveSerializable("scene", scene.first.get());
		}
	}
	out.endArray("scenes");
	out.ptr("currentScene", currentScene);
	out.raw("time", &time, sizeof(GameTime));
	SAVEVAR(f, secondsPerUpdate);
	// don't bother saving time residue
	police.serialize(out);
	schoolRegistry.serialize(out);
	out.u64("nextSoulID", Soul::souls);
	// path managers???
}

void World::deserialize(Deserialize& in)
{
	in.setWorld(this);
	const int registryLen = in.startArray("registry");
	registry.clear();
	for (int i = 0; i < registryLen; ++i)
	{
		auto soul = shared<Soul>(in);
		registry.insert(std::make_pair(soul->getID(), std::move(soul)));
	}
	in.endArray("registry");
	const int sceneCount = in.startArray("scenes");
	for (int i = 0; i < sceneCount; ++i)
	{
		const bool active = in.b("active");
		auto scene = loadSerializable<AbstractScene>(in, "scene");
		// don't call addScene() - we don't want onWorldAdd(), etc invoked - just insert it directly.
		scenes.push_back(std::make_pair(std::move(scene), active));
	}
	in.endArray("scenes");
	currentScene = static_cast<AbstractScene*>(in.ptr("currentScene"));
	in.raw("time", &time, sizeof(GameTime));
	LOADVAR(f, secondsPerUpdate);
	// time residue not saved
	police.deserialize(in);
	schoolRegistry.deserialize(in);
	Soul::souls = in.u64("nextSoulID");
	// path managers???
}

std::string World::serializationType() const
{
	return "World";
}

bool World::shouldSerialize() const
{
	return true;
}

Player* World::getPlayer()
{
	return player;
}

void World::setPlayer(Player *player)
{
	ASSERT((this->player == nullptr) != (player == nullptr));
	this->player = player;
}

void World::getScenesWithin(std::vector<Scene*>& output, const Rect<int>& worldspaceRegion, bool activeOnly)
{
	for (const std::pair<std::unique_ptr<AbstractScene>, bool>& scene : scenes)
	{
		if (scene.second || !activeOnly)
		{
			Scene *s = dynamic_cast<Scene*>(scene.first.get());
			if (s && worldspaceRegion.intersects(worldRegion(*s)))
			{
				output.push_back(s);
			}
		}
	}
}

void World::finalizeGeneration(const map::MapData& mapData, const towngen::GenerationListener& listener, int startPercent, int endPercent)
{
	ASSERT(startPercent < endPercent);
	const int percents = endPercent - startPercent;
	const int npcsPerUpdate = registry.size() < percents ? 1 : ((int)registry.size() / percents);
#ifdef SL_NOMULTITHREADED
	int done = 0;
#else
	std::atomic<int> done;
	std::mutex updateListenerMutex;
#endif
	
	auto finalizeNPC = [&](Soul& npc, int i) {
		// @HACK to stop the OpenGL warnings/errors in the console.
		// This is because you shouldn't be updating OpenGL stuff from another thread.
		// We will need to fix this by indirectly relaying this info back to the main thread somehow
		// and doing the actual drawable updates there.
		//if (++done >= npcsPerUpdate)
		//{
		//	std::lock_guard<std::mutex> lock(updateListenerMutex);
		//	done = 0;
		//	std::stringstream ss;
		//	ss << "Calculating NPC schedules / pathing [" << (100 * i / registry.size()) << "%]";
		//	listener(80 + percents * (float)i / registry.size(), ss.str().c_str(), mapData);
		//}
		if (npc.isAlive())
		{
			registerLocations(npc, *this);

			//// TODO: schedules for others
			// leave commented out? for explanation see NPC::onUpdate() @HACK
			/*if (npc.getType() == "loli")
			{
				npc.schedule = makeSchoolSchedule(*npc.getNPC(), *this);
			}*/
		}
	};
#ifdef SL_NOMULTITHREADED
	int i = 0;
	for (auto& npc : registry)
	{
		finalizeNPC(*npc.second, i);
		++i;
	}
#else
	const int threadCount = std::thread::hardware_concurrency();
	std::vector<std::thread> threads;
	for (int offset = 0; offset < threadCount; ++offset)
	{
		threads.push_back(std::thread([&, offset, this]() {
			int i = offset;
			for (auto it = iter::tryNext(registry.begin(), registry.end(), offset);
			     it != registry.end();
			     iter::tryAdvance(it, registry.end(), threadCount))
			{
				finalizeNPC(*it->second, i);
				i += threadCount;
			}
		}));
	}
	for (std::thread& thread : threads)
	{
		thread.join();
	}
#endif // SL_NOMULTITHREADED
}

/*				private					*/
void World::onUpdate(Timestamp timestamp)
{
	auto& input = getGame()->getRawInput();
	if ((input.isKeyHeld(Input::KeyboardKey::LShift) || input.isKeyHeld(Input::KeyboardKey::RShift)) && (input.isKeyHeld(Input::KeyboardKey::LControl) || input.isKeyHeld(Input::KeyboardKey::RControl)))
	{
		if (input.isKeyPressed(Input::KeyboardKey::H))
		{
			timeResidue += 3600.f;
		}
		if (input.isKeyPressed(Input::KeyboardKey::Y))
		{
			int total;
			float avg;
#define PRINTSTAT(stat) do { std::cout << #stat << ":"; \
	avg = 0.f; \
	total = 0; \
	for (const auto& npc : registry) { \
		if (npc.second->getType() == "loli") { \
			++total; \
			avg += npc.second->stat(); \
			std::cout << " " << npc.second->stat(); \
		} \
	} \
	std::cout << "\navg of " << (avg / total) << "\n\n" <<  std::endl; \
	} while (false)

			PRINTSTAT(gullibility);
			PRINTSTAT(getWillingness);
			PRINTSTAT(getHope);
			PRINTSTAT(getLove);
			PRINTSTAT(dominance);
			PRINTSTAT(openToCautious);
			PRINTSTAT(organizedToCareless);
			PRINTSTAT(extrovertToIntrovert);
			PRINTSTAT(friendlyToUnkind);
			PRINTSTAT(nervousToConfident);
		}
	}

	timeResidue += secondsPerUpdate;
	static float hourresidue = 0;
	// @hack for demo
	hourresidue += secondsPerUpdate / 3600.f;
	while (hourresidue >= 1.f)
	{
		hourresidue -= 1.f;
		for (auto& npc : registry)
		{
			if (npc.second->isAlive())
			{
				// @TODO pass specific Random instance?
				npc.second->hourlyUpdate(*this, Random::get(), false);
			}
		}
	}
	const int secondsToAdvance = static_cast<int>(timeResidue);
	if (secondsToAdvance)
	{
		timeResidue -= secondsToAdvance;
		
		const int daysPassed = time.time.advance(secondsToAdvance);
		if (daysPassed)
		{
			advanceDate(daysPassed, false);
		}
	}

	for (auto& pm : pathManagers)
	{
		pm.second->update(timestamp);
	}

	for (std::pair<std::unique_ptr<AbstractScene>, bool>& scene : scenes)
	{
		if (scene.second)
		{
			scene.first->update(timestamp);
		}
	}
	police.update();
}


void World::onDraw(RenderTarget& target)
{
	if (currentScene)
	{
		currentScene->draw(target);
	}
}

void World::onDrawGUI(RenderTarget& target)
{
	if (currentScene)
	{
		currentScene->drawGUI(target);
	}
}

void World::onHandleEvents(const EventQueue& events)
{
	for (std::pair<std::unique_ptr<AbstractScene>, bool>& scene : scenes)
	{
		if (scene.second)
		{
			scene.first->handleEvents(events);
		}
	}
}

void World::handleEvent(const Event& event)
{
	for (std::pair<std::unique_ptr<AbstractScene>, bool>& scene : scenes)
	{
		if (scene.second)
		{
			scene.first->handleEvent(event);
		}
	}
}

} // sl
