/**
 * @section DESCRIPTION
 * This represents a collection of Scenes of which you can choose
 * whether all are updated or only the current once is updated. Only
 * the current one is rendered however.
 */
#ifndef SL_WORLD_HPP
#define SL_WORLD_HPP


#include <cstdint>
#include <utility>
#include <vector>

#include "Engine/AbstractScene.hpp"
#include "NPC/SchoolRegistry.hpp"
#include "Pathfinding/PathGrid.hpp"
#include "Pathfinding/PathManager.hpp"
#include "Police/PoliceForce.hpp"
#include "Town/Generation/GenerationListener.hpp"
#include "Utility/Time/GameTime.hpp"

/**
 * A note about WORLDSPACE:
 * World space is essentially a standardized-tilespace.
 * If the scene has worldSacale == 1 then this is essentially tile-space.
 */

namespace sl
{

class Player;
class Random;
class Scene;
class Soul;
class SoulGenerationSettings;
class Deserialize;
class Serialize;

class World : public AbstractScene
{
public:
	World(Game *game, int viewWidth, int viewHeight, const GameTime& time = GameTime(0, 0, TimeOfDay(0)));

	~World();

	/**
	 * Note: thie assumes ownership over scene
	 * Adds a AbstractScene into this World.
	 * @param scene The AbstractScene to add
	 * @param alwaysUpdate Whether or not this AbstractScene should be updated even if it's not the current one
	 */
	void addScene(std::unique_ptr<AbstractScene> scene, bool alwaysUpdate);
	/**
	 * Sets the current AbstractScene. This is the only AbstractScene that is rendered
	 * Additionally, scenes not set to alwaysUpdate will only update
	 * when they are the current AbstractScene.
	 * @param scene The AbstractScene to set to current (error if not in this World)
	 */
	void setCurrentScene(AbstractScene *scene);

	const sf::Vector2f getView() const override;

	void setView(int x, int y) override;

	const GameTime& getGameTime() const;

	//void setTimeRate(float secondsPerUpdate);

	int getTimeRate() const;

	PoliceForce& getPolice();

	SchoolRegistry& getSchoolRegistry();

	std::shared_ptr<Soul> getSoul(uint64_t id);

	std::shared_ptr<Soul> createSoul(Random& rng, const SoulGenerationSettings& settings);

	void recordBirth(std::shared_ptr<Soul> baby);

	/**
	 * Gets a PathManager for use based on its name
	 * @param name The name it was registered with
	 * @return The PathManager -- will be null if none was registered
	 */
	pf::PathManager* getPathManager(const std::string& name) const;

	/**
	 *
	 */
	void registerPathManager(const std::string& name, std::shared_ptr<pf::PathManager> pathManager);

	void addGridToPathManager(const std::string& name, Scene *scene, std::shared_ptr<pf::PathGrid> pathGrid);

	void advanceDate(int daysPassed, bool timeSkip);

	void serialize(Serialize& out) const override;

	void deserialize(Deserialize& in) override;

	std::string serializationType() const override;

	bool shouldSerialize() const override;

	Player* getPlayer();

	/**
	 * Sets the Player. Should only be called from Player::onEnter()/onExit()
	 */
	void setPlayer(Player *player);

	void getScenesWithin(std::vector<Scene*>& output, const Rect<int>& worldspaceRegion, bool activeOnly = true);

	void finalizeGeneration(const map::MapData& mapData, const towngen::GenerationListener& listener, int startPercent, int endPercent);

private:

	void onUpdate(Timestamp timestamp) override;

	void onDraw(RenderTarget& target) override;

	void onDrawGUI(RenderTarget& target) override;

	void onHandleEvents(const EventQueue& events) override;

	void handleEvent(const Event& event) override;


	std::vector<std::pair<std::unique_ptr<AbstractScene>, bool>> scenes;
	AbstractScene *currentScene;
	GameTime time;
	float secondsPerUpdate;
	float timeResidue;
	PoliceForce police;
	SchoolRegistry schoolRegistry;
	std::map<uint64_t, std::shared_ptr<Soul>> registry;
	std::map<std::string, std::shared_ptr<pf::PathManager>> pathManagers;
	Player *player;
};

} // sl

#endif // SL_WORLD