#ifndef SL_DYNAMICSTORE_HPP
#define SL_DYNAMICSTORE_HPP

#include <memory>
#include <utility> // for std::forward

namespace sl
{

class DynamicStore
{
public:
	virtual ~DynamicStore() = default;

	template <typename T>
	static std::unique_ptr<DynamicStore> make();

	template <typename T, typename... Args>
	static std::unique_ptr<DynamicStore> make(Args... args);

	template <typename T>
	T* as();

	template <typename T>
	const T* as() const;

	virtual std::unique_ptr<DynamicStore> clone() const = 0;

protected:
	DynamicStore() = default;

private:
	DynamicStore(const DynamicStore&) = delete;
	DynamicStore& operator=(const DynamicStore&) = delete;
};

namespace impl
{
template <typename T>
class DynamicStoreData : public DynamicStore
{
public:
	DynamicStoreData()
		:data()
	{
	}

	template <typename... Args>
	DynamicStoreData(Args... args)
		:data(std::forward<Args>(args)...)
	{
	}

	std::unique_ptr<DynamicStore> clone() const override
	{
		return std::make_unique<DynamicStoreData<T>>(data);
	}

	T data;
};
} // impl

template <typename T>
/*static*/ std::unique_ptr<DynamicStore> DynamicStore::make()
{
	return std::make_unique<impl::DynamicStoreData<T>>();
}

template <typename T, typename... Args>
/*static*/ std::unique_ptr<DynamicStore> DynamicStore::make(Args... args)
{
	return std::make_unique<impl::DynamicStoreData<T>>(std::forward<Args>(args)...);
}

template <typename T>
T* DynamicStore::as()
{
	auto self = dynamic_cast<impl::DynamicStoreData<T>*>(this);
	return self ? &self->data : nullptr;
}

template <typename T>
const T* DynamicStore::as() const
{
	auto self = dynamic_cast<const impl::DynamicStoreData<T>*>(this);
	return self ? self->data : self;
}

} // sl

#endif // SL_DYNAMICSTORE_HPP