#ifndef SL_FLATMAP_HPP
#define SL_FLATMAP_HPP

#include <algorithm>
#include <string>
#include <vector>

#include "Utility/Assert.hpp"

namespace sl
{

template <typename K, typename V>
class FlatMap
{
public:
	using Data = std::vector<std::pair<K, V>>;
	using Iter = typename Data::iterator;
	using ConstIter = typename Data::const_iterator;

	FlatMap();

	template <typename T>
	FlatMap(const T& begin, const T& end);

	FlatMap(Data&& data);

	template <typename T>
	void assign(const T& begin, const T& end);

	void assign(Data&& data);

	Iter begin() { return data.begin(); }

	Iter end() { return data.end(); }

	ConstIter begin() const { return data.begin(); }

	ConstIter end() const { return data.end(); }

	int size() const { return data.size(); }

	void clear() { data.clear(); }

	bool empty() const { return data.empty(); }

	/**
	 * @param key The key to look for
	 * @return value associated with key, if it exists. Throws otherwise.
	 */
	V& operator[](const K& key);

	/**
	* @param key The key to look for
	* @return value associated with key, if it exists. Throws otherwise.
	*/
	const V& operator[](const K& key) const;


	Iter find(const K& key);
	
	ConstIter find(const K& key) const;

private:
	void sort();

	static bool compare(const std::pair<K, V>& lhs, const K& rhs);

	Data data;
};

template <typename K, typename V>
FlatMap<K, V>::FlatMap()
	:data()
{
}

template <typename K, typename V>
template <typename T>
FlatMap<K, V>::FlatMap(const T& begin, const T& end)
	:data()
{
	assign(begin, end);
}

template <typename K, typename V>
FlatMap<K, V>::FlatMap(Data&& data)
	:data(data)
{
	sort();
}

template <typename K, typename V>
template <typename T>
void FlatMap<K, V>::assign(const T& begin, const T& end)
{
	data.clear();
	data.insert_range(begin, end);
	sort();
}

template <typename K, typename V>
void FlatMap<K, V>::assign(Data&& data)
{
	this->data = data;
	sort();
}

template <typename T>
static inline void throwInvalidKey(const T&)
{
	throw std::out_of_range("key not found in FlatMap");
}

template<>
static inline void throwInvalidKey(const std::string& key)
{
	throw std::out_of_range("key '" + key + "' not found in FlatMap");
}

template <typename K, typename V>
V& FlatMap<K, V>::operator[](const K& key)
{
	auto it = find(key);
	if (it != end())
	{
		return it->second;
	}
	throwInvalidKey(key);
}

template <typename K, typename V>
const V& FlatMap<K, V>::operator[](const K& key) const
{
	auto it = find(key);
	if (it != end())
	{
		return it->second;
	}
	throwInvalidKey(key);
}

template <typename K, typename V>
typename FlatMap<K, V>::Iter FlatMap<K, V>::find(const K& key)
{
	auto it = std::lower_bound(data.begin(), data.end(), key, compare);
	if (it != end() && it->first == key)
	{
		return it;
	}
	auto it2 = std::find_if(data.begin(), data.end(), [&](const std::pair<K, V>& rhs) {
		return key == rhs.first;
	});
	ASSERT(it2 == end());
	return end();
}

template <typename K, typename V>
typename FlatMap<K, V>::ConstIter FlatMap<K, V>::find(const K& key) const
{
	auto it = std::lower_bound(data.begin(), data.end(), key, compare);
	if (it != end() && it->first == key)
	{
		return it;
	}
	auto it2 = std::find_if(data.begin(), data.end(), [&](const std::pair<K, V>& rhs) {
		return key == rhs.first;
	});
	ASSERT(it2 == end());
	return end();
}



// private
template <typename K, typename V>
/*static*/ bool FlatMap<K, V>::compare(const std::pair<K, V>& lhs, const K& rhs)
{
	return lhs.first < rhs;
}

template <typename K, typename V>
void FlatMap<K, V>::sort()
{
	std::sort(data.begin(), data.end(), [](const auto& lhs, const auto& rhs) {
		return lhs.first < rhs.first;
	});
	auto it = std::unique(data.begin(), data.end(), [](const auto& lhs, const auto& rhs) {
		return lhs.first == rhs.first;
	});
	ASSERT(it == end());
}

} // ls

#endif // SL_FLATMAP_HPP