#include "Utility/Functions/ConvergentSeries.hpp"

#include "Utility/Assert.hpp"
#include "Utility/Trig.hpp"

#include <limits>

namespace sl
{

float geometricSeries(int n)
{
	ASSERT(n >= 1);
	return 1.f / std::powf(2, n);
}


float baselSeries(int n)
{
	ASSERT(n >= 1);
	const float nf = static_cast<float>(n);
	// This series converges to pi^2 / 6, so we must normalize all terms
	static constexpr float normalizationScalar = 6.f / (pi*pi);
	return normalizationScalar / (nf*nf);
}

FiniteSeries::FiniteSeries(std::function<float(int)> f)
	:FiniteSeries(f, computeNonZeroTermsIndex(f))
{
}

FiniteSeries::FiniteSeries(std::function<float(int)> f, int nonZeroTerms)
	:f(std::move(f))
	,nonZeroTerms(nonZeroTerms)
	,normalizationScalar(computeNormalizationScalar())
{
}

float FiniteSeries::operator()(int n) const
{
	return n > nonZeroTerms ? 0.f : (f(n) * normalizationScalar);
}

float FiniteSeries::computeNormalizationScalar() const
{
	float sum = 0.f;
	for (int n = 1; n <= nonZeroTerms; ++n)
	{
		sum += f(n);
	}
	return 1.f / sum;
}

/*static*/int FiniteSeries::computeNonZeroTermsIndex(const std::function<float(int)>& f)
{
	int n = 1;
	while (f(n + 1) > std::numeric_limits<float>::epsilon())
	{
		++n;
	}
	return n;
}

} // sl
