#include "Utility/GraphClustering.hpp"

#include <algorithm>
#include <cmath>
#include <map>
#include <queue>
#include <set>

#include "Utility/Increment.hpp"
#include "Utility/JoinableSet.hpp"

namespace sl
{

static std::vector<std::pair<float, std::vector<int>>> connectedComponentsFromThreshold(const Matrix<float>& adjacencyMatrix, const std::vector<int>& vertices, float threshold)
{
	std::vector<std::pair<float, std::vector<int>>> components;
	std::queue<int> queue; // by index into vertices
	std::vector<int> visited(vertices.size(), 0); // indexed by index into vertices
	for (int i = 0; i < vertices.size(); ++i)
	{
		if (visited[i] == 0)
		{
			components.push_back(std::make_pair(std::numeric_limits<float>::max(), std::vector<int>()));
			visited[i] = 1;
			queue.push(i);
			while (!queue.empty())
			{
				const int u = vertices[queue.front()];
				queue.pop();
				components.back().second.push_back(u);
				for (int j = 0; j < vertices.size(); ++j)
				{
					const int v = vertices[j];
					if (u != v && adjacencyMatrix(u, v) >= threshold && visited[j] == 0)
					{
						components.back().first = std::min(components.back().first, adjacencyMatrix(u, v));
						visited[j] = 1;
						queue.push(j);
					}
				}
			}
		}
	}
	return components;
}

static void splitOversizeComponents(std::vector<std::vector<int>>& output, const std::vector<int>& vertices, const Matrix<float>& adjacencyMatrix, int maxGroupSize, float joinThreshold)
{
	auto components = connectedComponentsFromThreshold(adjacencyMatrix, vertices, joinThreshold);
	for (auto& component : components)
	{
		if (component.second.size() > maxGroupSize)
		{
			splitOversizeComponents(output, component.second, adjacencyMatrix, maxGroupSize, std::nextafterf(component.first, std::numeric_limits<float>::max()));
		}
		else
		{
			output.push_back(std::move(component.second));
		}
	}
}

std::vector<std::vector<int>> weightedDirectedClusters(const Matrix<float>& adjacencyMatrix, int maxGroupSize, float joinThreshold)
{
	ASSERT(adjacencyMatrix.getRows() == adjacencyMatrix.getCols());
	const int n = adjacencyMatrix.getRows();
	std::vector<int> allVertices(n);
	std::generate(allVertices.begin(), allVertices.end(), Increment<int>());
	std::vector<std::vector<int>> firstPhaseComponents;
	// group into components based on the threshold, then start removing the lightest weight edge in each oversize component until it splits into ones small enough
	splitOversizeComponents(firstPhaseComponents, allVertices, adjacencyMatrix, maxGroupSize, joinThreshold);
	return firstPhaseComponents;
}

std::vector<std::vector<int>> mergeLeftoversGreedily(const Matrix<float>& adjacencyMatrix, int maxGroupSize, float mergeThreshold, const std::vector<std::vector<int>>& firstPhaseComponents)
{
	ASSERT(adjacencyMatrix.getRows() == adjacencyMatrix.getCols());
	const int n = adjacencyMatrix.getRows();
	JoinableSet mergeSets(n);
	// copy computed sets -> JoinableSet
	for (const std::vector<int>& component : firstPhaseComponents)
	{
		for (int i = 1; i < component.size(); ++i)
		{
			mergeSets.join(component[0], component[i]);
		}
	}
	// merge the smaller ones subject to the original threshold
	for (const std::vector<int>& component : firstPhaseComponents)
	{
		if (component.size() < maxGroupSize)
		{
			for (int u : component)
			{
				bool componentMerged = false;
				for (int v = 0; v < n; ++v)
				{
					if (adjacencyMatrix(u, v) >= mergeThreshold)
					{
						const int uComp = mergeSets.getComponent(u);
						const int vComp = mergeSets.getComponent(v);
						if (mergeSets.componentSize(uComp) + mergeSets.componentSize(vComp) <= maxGroupSize)
						{
							mergeSets.join(uComp, vComp);
							componentMerged = true;
							break;
						}
					}
				}
				if (componentMerged)
				{
					continue;
				}
			}
		}
	}
	std::vector<std::vector<int>> output;
	// and JoinableSet -> output
	std::map<int, std::set<int>> componentToVertices;
	for (int u = 0; u < n; ++u)
	{
		componentToVertices[mergeSets.getComponent(u)].insert(u);
	}
	for (const auto& component : componentToVertices)
	{
		output.push_back(std::vector<int>());
		for (int u : component.second)
		{
			output.back().push_back(u);
		}
	}
	return output;
}

} // sl