#ifndef SL_GRID_HPP
#define SL_GRID_HPP

#include "Utility/Assert.hpp"
#include "Utility/Vector.hpp"

namespace sl
{

template <typename T>
class Grid
{
public:
	class Iterator;
	class ConstIterator
	{
	public:
		ConstIterator(const Grid<T>& owner, int x = 0, int y = 0);

		bool operator==(const Iterator& rhs) const;

		bool operator!=(const Iterator& rhs) const;

		bool operator==(const ConstIterator& rhs) const;

		bool operator!=(const ConstIterator& rhs) const;

		const T& operator*() const;

		const T* operator->() const;

		ConstIterator& operator++();

		ConstIterator operator++(int);

		ConstIterator& operator--();

		ConstIterator operator--(int);

		int getX() const;

		int getY() const;

		ConstIterator offset(int xoff, int yoff) const;

		const Grid<T>& getGrid() const;

	private:
		const Grid<T> *owner;
		int x, y;
	};
	class Iterator
	{
	public:
		Iterator(Grid<T>& owner, int x = 0, int y = 0);

		bool operator==(const Iterator& rhs) const;

		bool operator!=(const Iterator& rhs) const;

		bool operator==(const ConstIterator& rhs) const;

		bool operator!=(const ConstIterator& rhs) const;

		T& operator*() const;

		T* operator->() const;

		Iterator& operator++();

		Iterator operator++(int);

		Iterator& operator--();

		Iterator operator--(int);

		int getX() const;

		int getY() const;

		Iterator offset(int xoff, int yoff) const;

		Grid<T>& getGrid();

		const Grid<T>& getGrid() const;

		operator typename Grid<T>::ConstIterator() const
		{
			return Grid<T>::ConstIterator(&owner, x, y);
		}

	private:
		Grid<T> *owner;
		int x, y;
	};

	Grid();

	Grid(int width, int height);

	Grid(int width, int height, const T& value);

	Grid(const Grid<T>& other);

	Grid(Grid<T>&& move);

	~Grid();

	Grid<T>& operator=(const Grid<T>& rhs);

	Grid<T>& operator=(Grid<T>&& rhs);


	Iterator begin();
	Iterator end();
	ConstIterator begin() const;
	ConstIterator end() const;
	ConstIterator cbegin() const;
	ConstIterator cend() const;

	void resize(int width, int height, const T& defVal = T());
	void clear();

	T& at(int x, int y);
	const T& at(int x, int y) const;

	T& operator()(int x, int y)
	{
		return at(x, y);
	}
	const T& operator()(int x, int y) const
	{
		return at(x, y);
	}
	T& at(const Vector2i& pos)
	{
		return at(pos.x, pos.y);
	}
	const T& at(const Vector2i& pos) const
	{
		return at(pos.x, pos.y);
	}
	T& operator()(const Vector2i& pos)
	{
		return at(pos.x, pos.y);
	}
	const T& operator()(const Vector2i& pos) const
	{
		return at(pos.x, pos.y);
	}
	T& operator[](const Vector2i& pos)
	{
		return at(pos.x, pos.y);
	}
	const T& operator[](const Vector2i& pos) const
	{
		return at(pos.x, pos.y);
	}

	int getWidth() const;
	int getHeight() const;

	bool isValid(int x, int y) const;
	bool isValid(const Vector2i& pos) const
	{
		return isValid(pos.x, pos.y);
	}

private:
	T *data;
	int width;
	int height;
};

template <typename T>
Grid<T>::Grid()
	:data(nullptr)
	,width(0)
	,height(0)
{
}

template <typename T>
Grid<T>::Grid(int width, int height)
	:data(new T[width * height])
	,width(width)
	,height(height)
{
}

template <typename T>
Grid<T>::Grid(int width, int height, const T& value)
	:data((T*) operator new[](width * height * sizeof(T)))
	,width(width)
	,height(height)
{
	T *iter = data;
	for (int i = 0, size = width * height; i < size; ++i)
	{
		new (iter++) T(value);
	}
}

template <typename T>
Grid<T>::Grid(const Grid<T>& other)
	:data(new T[other.width * other.height])
	,width(other.width)
	,height(other.height)
{
	int size = width * height;
	for (int i = 0; i < size; ++i)
	{
		data[i] = other.data[i];
	}
}

template <typename T>
Grid<T>::Grid(Grid<T>&& move)
	:data(move.data)
	,width(move.width)
	,height(move.height)
{
	move.width = 0;
	move.height = 0;
	move.data = nullptr;
}

template <typename T>
Grid<T>::~Grid()
{
	this->clear();
}

template <typename T>
Grid<T>& Grid<T>::operator=(const Grid<T>& rhs)
{
	if (rhs.width * rhs.height != width * height)
	{
		if (data)
		{
			delete[] data;
		}
		data = new T[rhs.width * rhs.height];
	}
	width = rhs.width;
	height = rhs.height;
	int size = width * height;
	for (int i = 0; i < size; ++i)
	{
		data[i] = rhs.data[i];
	}
	return *this;
}

template <typename T>
Grid<T>& Grid<T>::operator=(Grid<T>&& rhs)
{
	clear();
	width = rhs.width;
	height = rhs.height;
	data = rhs.data;
	rhs.width = 0;
	rhs.height = 0;
	rhs.data = nullptr;
	return *this;
}

template <typename T>
typename Grid<T>::Iterator Grid<T>::begin()
{
	return Iterator(*this, 0, 0);
}

template <typename T>
typename Grid<T>::Iterator Grid<T>::end()
{
	return Iterator(*this, 0, height);
}

template <typename T>
typename Grid<T>::ConstIterator Grid<T>::begin() const
{
	return ConstIterator(*this, 0, 0);
}

template <typename T>
typename Grid<T>::ConstIterator Grid<T>::end() const
{
	return ConstIterator(*this, 0, height);
}

template <typename T>
typename Grid<T>::ConstIterator Grid<T>::cbegin() const
{
	return ConstIterator(*this, 0, 0);
}

template <typename T>
typename Grid<T>::ConstIterator Grid<T>::cend() const
{
	return ConstIterator(*this, 0, height);
}

template <typename T>
void Grid<T>::resize(int width, int height, const T& defVal)
{
	ASSERT(width > 0 && height > 0);
	T* oldData = data;
	data = new T[width * height];
	for (int y = 0; y < height; ++y)
	{
		for (int x = 0; x < width; ++ x)
		{
			// copy over all old elements then pad grid with defVal
			if (y < this->height && x < this->width)
			{
				data[x + y * width] = oldData[x + y * this->width];
			}
			else
			{
				data[x + y * width] = defVal;
			}
		}
	}
	delete[] oldData;
	this->width = width;
	this->height = height;
}

template <typename T>
void Grid<T>::clear()
{
	if (data)
	{
		delete[] data;
	}
	data = nullptr;
	width = 0;
	height = 0;
}

template <typename T>
T& Grid<T>::at(int x, int y)
{
	ASSERT(isValid(x, y));
	return data[x + width * y];
}

template <typename T>
const T& Grid<T>::at(int x, int y) const
{
	ASSERT(isValid(x, y));
	return data[x + width * y];
}

template <typename T>
int Grid<T>::getWidth() const
{
	return width;
}

template <typename T>
int Grid<T>::getHeight() const
{
	return height;
}

template <typename T>
bool Grid<T>::isValid(int x, int y) const
{
	return x >= 0 && y >= 0 && x < width && y < height;
}



// -------------------------------- Iterator ----------------------------------
template <typename T>
Grid<T>::Iterator::Iterator(Grid<T>& owner, int x, int y)
	:owner(&owner)
	,x(x)
	,y(y)
{
}

template <typename T>
bool Grid<T>::Iterator::operator==(const Iterator& rhs) const
{
	return x == rhs.x && y == rhs.y && owner == rhs.owner;
}

template <typename T>
bool Grid<T>::Iterator::operator!=(const Iterator& rhs) const
{
	return x != rhs.x || y != rhs.y || owner != rhs.owner;
}

template <typename T>
bool Grid<T>::Iterator::operator==(const ConstIterator& rhs) const
{
	return x == rhs.getX() && y == rhs.getY() && owner == &rhs.getGrid();
}

template <typename T>
bool Grid<T>::Iterator::operator!=(const ConstIterator& rhs) const
{
	return x != rhs.getX() || y != rhs.getY() || owner != &rhs.getGrid();
}

template <typename T>
T& Grid<T>::Iterator::operator*() const
{
	ASSERT(owner->isValid(x, y));
	return owner->at(x, y);
}

template <typename T>
T* Grid<T>::Iterator::operator->() const
{
	ASSERT(owner->isValid(x, y));
	return &owner->at(x, y);
}

template <typename T>
typename Grid<T>::Iterator& Grid<T>::Iterator::operator++()
{
	++x;
	while (x >= owner->width)
	{
		x -= owner->width;
		++y;
	}
	return *this;
}

template <typename T>
typename Grid<T>::Iterator Grid<T>::Iterator::operator++(int)
{
	Iterator ret(*this);
	++x;
	while (x >= owner->width)
	{
		x -= owner->width;
		++y;
	}
	return ret;
}

template <typename T>
typename Grid<T>::Iterator& Grid<T>::Iterator::operator--()
{
	--x;
	while (x < 0)
	{
		x += owner->width;
		--y;
	}
	return *this;
}

template <typename T>
typename Grid<T>::Iterator Grid<T>::Iterator::operator--(int)
{
	Iterator ret(*this);
	--x;
	while (x < 0)
	{
		x += owner->width;
		--y;
	}
	return ret;
}

template <typename T>
int Grid<T>::Iterator::getX() const
{
	return x;
}

template <typename T>
int Grid<T>::Iterator::getY() const
{
	return y;
}

template <typename T>
typename Grid<T>::Iterator Grid<T>::Iterator::offset(int xoff, int yoff) const
{
	const int nx = x + xoff;
	const int ny = y + yoff;
	if (owner->isValid(nx, ny))
	{
		return Iterator(*owner, nx, ny);
	}
	return owner->end();
}


template <typename T>
Grid<T>& Grid<T>::Iterator::getGrid()
{
	return *owner;
}

template <typename T>
const Grid<T>& Grid<T>::Iterator::getGrid() const
{
	return *owner;
}

// ----------------------------- ConstIterator --------------------------------
template <typename T>
Grid<T>::ConstIterator::ConstIterator(const Grid<T>& owner, int x, int y)
	:owner(&owner)
	,x(x)
	,y(y)
{
}

template <typename T>
bool Grid<T>::ConstIterator::operator==(const Iterator& rhs) const
{
	return x == rhs.getX() && y == rhs.getY() && owner == &rhs.getGrid();
}

template <typename T>
bool Grid<T>::ConstIterator::operator!=(const Iterator& rhs) const
{
	return x != rhs.getX() || y != rhs.getY() || owner != &rhs.getGrid();
}

template <typename T>
bool Grid<T>::ConstIterator::operator==(const ConstIterator& rhs) const
{
	return x == rhs.x && y == rhs.y && owner == rhs.owner;
}

template <typename T>
bool Grid<T>::ConstIterator::operator!=(const ConstIterator& rhs) const
{
	return x != rhs.x || y != rhs.y || owner != rhs.owner;
}

template <typename T>
const T& Grid<T>::ConstIterator::operator*() const
{
	ASSERT(owner->isValid(x, y));
	return owner->at(x, y);
}

template <typename T>
const T* Grid<T>::ConstIterator::operator->() const
{
	ASSERT(owner->isValid(x, y));
	return &owner->at(x, y);
}

template <typename T>
typename Grid<T>::ConstIterator& Grid<T>::ConstIterator::operator++()
{
	++x;
	while (x >= owner->width)
	{
		x -= owner->width;
		++y;
	}
	return *this;
}

template <typename T>
typename Grid<T>::ConstIterator Grid<T>::ConstIterator::operator++(int)
{
	ConstIterator ret(*this);
	++x;
	while (x >= owner->width)
	{
		x -= owner->width;
		++y;
	}
	return ret;
}

template <typename T>
typename Grid<T>::ConstIterator& Grid<T>::ConstIterator::operator--()
{
	--x;
	while (x < 0)
	{
		x += owner->width;
		--y;
	}
	return *this;
}

template <typename T>
typename Grid<T>::ConstIterator Grid<T>::ConstIterator::operator--(int)
{
	ConstIterator ret(*this);
	--x;
	while (x < 0)
	{
		x += owner->width;
		--y;
	}
	return ret;
}

template <typename T>
int Grid<T>::ConstIterator::getX() const
{
	return x;
}

template <typename T>
int Grid<T>::ConstIterator::getY() const
{
	return y;
}

template <typename T>
typename Grid<T>::ConstIterator Grid<T>::ConstIterator::offset(int xoff, int yoff) const
{
	const int nx = x + xoff;
	const int ny = y + yoff;
	if (owner->isValid(nx, ny))
	{
		return ConstIterator(*owner, nx, ny);
	}
	return owner->cend();
}

template <typename T>
const Grid<T>& Grid<T>::ConstIterator::getGrid() const
{
	return *owner;
}

} // sl

#endif // SL_GRID_HPP
