#ifndef SL_INTRUSIVELIST_HPP
#define SL_INTRUSIVELIST_HPP

#include "Engine/Serialization/Serialization.hpp"
#include "Utility/Assert.hpp"

#ifdef SL_VALIDATE_INTRUSIVE_LIST
	#include <iostream>
	#include <vector>
#endif // SL_DEBUG

namespace sl
{

enum class IntrusiveListOwnershipType : bool
{
	Owned = 0, // automatically deletes contents upon going out of scope
	Unowned             // forbids deleteAll()
};

template <typename T, IntrusiveListOwnershipType O>
class IntrusiveList;

template <typename T>
class IntrusiveListNode
{
public:
	IntrusiveListNode(T& data);

	IntrusiveListNode(IntrusiveListNode<T>&& move) = delete;

	~IntrusiveListNode();

	IntrusiveListNode<T>& operator=(IntrusiveListNode<T>&& rhs) = delete;

//	void swap(sl::IntrusiveListNode<T>& b)
//	{
//#ifdef SL_DEBUG
//		owner->validate();
//		b.owner->validate();
//#endif // SL_DEBUG
//		if (prev)
//		{
//			ASSERT(next);
//			prev->next = &b;
//			next->prev = &b;
//		}
//		if (b.prev)
//		{
//			ASSERT(b.next);
//			b.prev->next = this;
//			b.next->prev = this;
//		}
//		std::swap(prev, b.prev);
//		std::swap(next, b.next);
//		std::swap(data, b.data);
//#ifdef SL_DEBUG
//		std::swap(owner, b.owner);
//#endif // SL_DEBUG
//#ifdef SL_DEBUG
//		owner->validate();
//		b.owner->validate();
//#endif // SL_DEBUG
//	}


	void serialize(Serialize& out) const
	{
		out.ptr("data", data);
		if (prev->data == nullptr)
		{
			// skip head nodes (null data) since we'll join them into them
			// when we insert them into the list during deserialization
			// and until then they're floating cycles
			ASSERT(prev->prev->data != nullptr);
			out.ptr("prev", prev->prev->data);
		}
		else
		{
			out.ptr("prev", prev->data);
		}
		if (next->data)
		{
			out.ptr("next", next->data);
		}
	}

	void deserialize(Deserialize& in)
	{
		data = static_cast<T*>(in.ptr("data"));
		prev = &static_cast<T*>(in.ptr("prev"))->getLink();
		next = &static_cast<T*>(in.ptr("next"))->getLink();
	}

private:
	IntrusiveListNode();
	IntrusiveListNode(const IntrusiveListNode<T>&) = delete;
	IntrusiveListNode<T>& operator=(const IntrusiveListNode<T>&) = delete;


	void unlink();


	T *data;
	IntrusiveListNode<T> *prev, *next;
#ifdef SL_VALIDATE_INTRUSIVE_LIST
	// arbitrarily picked one of them...
	IntrusiveList<T, IntrusiveListOwnershipType::Owned> *owner;
#endif // SL_VALIDATE_INTRUSIVE_LIST

	// sometimes I hate C++...
	friend class IntrusiveList<T, IntrusiveListOwnershipType::Owned>;
	friend class IntrusiveList<T, IntrusiveListOwnershipType::Unowned>;
};

template <typename T, IntrusiveListOwnershipType O = IntrusiveListOwnershipType::Unowned>
class IntrusiveList
{
public:
	typedef T value_type;
	class Iterator
	{
	public:
		Iterator(IntrusiveListNode<T> *node);

		bool operator==(const Iterator& rhs) const;

		bool operator!=(const Iterator& rhs) const;

		T& operator*() const;

		T* operator->() const;

		Iterator& operator++();

		Iterator operator++(int);

		Iterator& operator--();

		Iterator operator--(int);

		/**
		 * Deletes the data and returns an iterator to the node previous to this one.
		 */
		Iterator deleteData() const;

	private:
		IntrusiveListNode<T> *node;
	};
	class ConstIterator
	{
	public:
		ConstIterator(const IntrusiveListNode<T> *node);

		bool operator==(const ConstIterator& rhs) const;

		bool operator!=(const ConstIterator& rhs) const;

		const T& operator*() const;

		const T* operator->() const;

		ConstIterator& operator++();

		ConstIterator operator++(int);

		ConstIterator& operator--();

		ConstIterator operator--(int);

	private:
		const IntrusiveListNode<T> *node;
	};

	IntrusiveList();

	~IntrusiveList();

	IntrusiveList(IntrusiveList<T, O>&& other);

	IntrusiveList<T, O>& operator=(IntrusiveList<T, O>&& rhs);

	Iterator begin();
	Iterator end();
	ConstIterator begin() const;
	ConstIterator end() const;
	ConstIterator cbegin() const;
	ConstIterator cend() const;

	void clear();

	inline void deleteAll();

	bool empty() const;

	void splice(IntrusiveList<T, O>& donor);

	void insert(IntrusiveListNode<T>& node);

	//template <typename U>
	//void insert(U& object);

	int size() const;

private:
	//	disallow copy ctor
	IntrusiveList(const IntrusiveList<T, O>& copy) = delete;
	//	disallow copy operator
	IntrusiveList<T, O>& operator=(const IntrusiveList<T, O>& copy) = delete;

#if SL_VALIDATE_INTRUSIVE_LIST
public:
	void validate() const;
#endif // SL_VALIDATE_INTRUSIVE_LIST




	IntrusiveListNode<T> head;

	friend class IntrusiveListNode<T>;
};

//} // sl
//
//
//namespace std
//{
//
//template <typename T>
//void swap(sl::IntrusiveListNode<T>& a, sl::IntrusiveListNode<T>& b)
//{
//	a.swap(b);
//}
//
//} // std
//
//namespace sl
//{

/*
	----------------------------------
			LIST IMPLEMENTATION
	----------------------------------
*/

template <typename T, IntrusiveListOwnershipType O>
IntrusiveList<T, O>::IntrusiveList()
	:head()
{
#ifdef SL_VALIDATE_INTRUSIVE_LIST
	head.owner = reinterpret_cast<IntrusiveList<T, IntrusiveListOwnershipType::Owned>*>(this);
#endif // SL_VALIDATE_INTRUSIVE_LIST
}


template <typename T, IntrusiveListOwnershipType O>
IntrusiveList<T, O>::~IntrusiveList()
{
	if (O == IntrusiveListOwnershipType::Owned)
	{
		this->deleteAll();
	}
	else
	{
		this->clear();
	}
}

template <typename T, IntrusiveListOwnershipType O>
IntrusiveList<T, O>::IntrusiveList(IntrusiveList<T, O>&& other)
	:head()
{
#ifdef SL_VALIDATE_INTRUSIVE_LIST
	head.owner = reinterpret_cast<IntrusiveList<T, IntrusiveListOwnershipType::Owned>*>(this);
	validate();
	other.validate();
#endif // SL_VALIDATE_INTRUSIVE_LIST
	if (!other.empty())
	{
		head.next = other.head.next;
		head.prev = other.head.prev;
		head.next->prev = &head;
		head.prev->next = &head;
		other.head.next = &other.head;
		other.head.prev = &other.head;
	}
#if SL_VALIDATE_INTRUSIVE_LIST
	for (IntrusiveListNode<T> *it = head.next; it != &head; it = it->next)
	{
		it->owner = head.owner;
	}
	validate();
	other.validate();
#endif // SL_VALIDATE_INTRUSIVE_LIST
}

template <typename T, IntrusiveListOwnershipType O>
IntrusiveList<T, O>& IntrusiveList<T, O>::operator=(IntrusiveList<T, O>&& rhs)
{
#if SL_VALIDATE_INTRUSIVE_LIST
	validate();
	rhs.validate();
#endif // SL_VALIDATE_INTRUSIVE_LIST
	if (O == IntrusiveListOwnershipType::Owned)
	{
		this->deleteAll();
	}
	else
	{
		this->clear();
	}
	if (!rhs.empty())
	{
		head.next = rhs.head.next;
		head.prev = rhs.head.prev;
		head.next->prev = &head;
		head.prev->next = &head;
		rhs.head.next = &rhs.head;
		rhs.head.prev = &rhs.head;
	}
#if SL_VALIDATE_INTRUSIVE_LIST
	for (IntrusiveListNode<T> *it = head.next; it != &head; it = it->next)
	{
		it->owner = head.owner;
	}
	validate();
	rhs.validate();
#endif // SL_VALIDATE_INTRUSIVE_LIST
	return *this;
}

template <typename T, IntrusiveListOwnershipType O>
typename IntrusiveList<T, O>::Iterator IntrusiveList<T, O>::begin()
{
	return Iterator(head.next);
}

template <typename T, IntrusiveListOwnershipType O>
typename IntrusiveList<T, O>::Iterator IntrusiveList<T, O>::end()
{
	return Iterator(&head);
}

template <typename T, IntrusiveListOwnershipType O>
typename IntrusiveList<T, O>::ConstIterator IntrusiveList<T, O>::begin() const
{
	return ConstIterator(head.next);
}

template <typename T, IntrusiveListOwnershipType O>
typename IntrusiveList<T, O>::ConstIterator IntrusiveList<T, O>::end() const
{
	return ConstIterator(&head);
}

template <typename T, IntrusiveListOwnershipType O>
typename IntrusiveList<T, O>::ConstIterator IntrusiveList<T, O>::cbegin() const
{
	return ConstIterator(head.next);
}

template <typename T, IntrusiveListOwnershipType O>
typename IntrusiveList<T, O>::ConstIterator IntrusiveList<T, O>::cend() const
{
	return ConstIterator(&head);
}

template <typename T, IntrusiveListOwnershipType O>
void IntrusiveList<T, O>::clear()
{
#if SL_VALIDATE_INTRUSIVE_LIST
	validate();
#endif // SL_VALIDATE_INTRUSIVE_LIST
	IntrusiveListNode<T> *temp = nullptr, *it = head.next;
	head.prev = &head;
	head.next = &head;
	while (it != &head)
	{
#ifdef SL_VALIDATE_INTRUSIVE_LIST
		ASSERT(reinterpret_cast<decltype(it->owner)>(this) == it->owner);
#endif // SL_VALIDATE_INTRUSIVE_LIST
		temp = it->next;
		it->unlink();
		it = temp;
	}
#if SL_VALIDATE_INTRUSIVE_LIST
	validate();
#endif // SL_VALIDATE_INTRUSIVE_LIST
}

template <typename T, IntrusiveListOwnershipType O>
void IntrusiveList<T, O>::deleteAll()
{
#if SL_VALIDATE_INTRUSIVE_LIST
	validate();
#endif // SL_VALIDATE_INTRUSIVE_LIST
	ASSERT(O == IntrusiveListOwnershipType::Owned);
	//static_assert(O == IntrusiveListOwnershipType::Owned,
	//              "Can only delete from lists with ownership.");

	IntrusiveListNode<T> *temp = nullptr, *it = head.next;
	head.prev = &head;
	head.next = &head;
	while (it != &head)
	{
#ifdef SL_VALIDATE_INTRUSIVE_LIST
		ASSERT(reinterpret_cast<decltype(it->owner)>(this) == it->owner);
		it->owner = nullptr;
#endif // SL_VALIDATE_INTRUSIVE_LIST
		temp = it->next;
		it->prev = nullptr;
		it->next = nullptr;
		delete it->data;
#if SL_VALIDATE_INTRUSIVE_LIST
		validate();
		temp->owner->validate();
#endif // SL_VALIDATE_INTRUSIVE_LIST
		it = temp;
	}
#if SL_VALIDATE_INTRUSIVE_LIST
	validate();
#endif // SL_VALIDATE_INTRUSIVE_LIST
}

template <typename T, IntrusiveListOwnershipType O>
bool IntrusiveList<T, O>::empty() const
{
	return &head == head.next;
}

template <typename T, IntrusiveListOwnershipType O>
void IntrusiveList<T, O>::splice(IntrusiveList<T, O>& donor)
{
#if SL_VALIDATE_INTRUSIVE_LIST
	validate();
	donor.validate();
#endif // SL_VALIDATE_INTRUSIVE_LIST
	if (!donor.empty())
	{
#ifdef SL_VALIDATE_INTRUSIVE_LIST
		for (IntrusiveListNode<T> *it = donor.head.next; it != &donor.head; it = it->next)
		{
			ASSERT(it->owner == reinterpret_cast<decltype(it->owner)>(&donor));
			it->owner = this;
		}
#endif // SL_VALIDATE_INTRUSIVE_LIST
		donor.head.next->prev = &this->head;
		donor.head.prev->next = this->head.next;
		this->head.next->prev = donor.head.prev;
		this->head.next = donor.head.next;

		donor.head.prev = donor.head.next = &donor.head;
	}
#if SL_VALIDATE_INTRUSIVE_LIST
	validate();
	donor.validate();
	ASSERT(donor.empty());
#endif // SL_VALIDATE_INTRUSIVE_LIST
}

template <typename T, IntrusiveListOwnershipType O>
void IntrusiveList<T, O>::insert(IntrusiveListNode<T>& node)
{
#if SL_VALIDATE_INTRUSIVE_LIST
	validate();
#endif // SL_VALIDATE_INTRUSIVE_LIST
	ASSERT(node.data);//I really hope someone isn't inserting list heads...
	node.unlink();
	node.next = this->head.next;
	node.prev = &this->head;
	this->head.next->prev = &node;
	this->head.next = &node;
#ifdef SL_VALIDATE_INTRUSIVE_LIST
	node.owner = reinterpret_cast<IntrusiveList<T, IntrusiveListOwnershipType::Owned>*>(this);
	validate();
#endif // SL_VALIDATE_INTRUSIVE_LIST
}

/*template <typename T, typename U>
void IntrusiveList<T, O>::insert(U& object)
{
	this->insert(object.getLink());
}*/

template <typename T, IntrusiveListOwnershipType O>
int IntrusiveList<T, O>::size() const
{
	int elems = 0;
	for (IntrusiveListNode<T> *it = head.next; it != &head; it = it->next)
	{
		++elems;
	}
	return elems;
}

// private
#if SL_VALIDATE_INTRUSIVE_LIST
template <typename T, IntrusiveListOwnershipType O>
void IntrusiveList<T, O>::validate() const
{
	std::vector<IntrusiveListNode<T>*> forward, back;
	ASSERT(this == reinterpret_cast<decltype(this)>(head.owner));
	ASSERT(head.data == nullptr);
	for (IntrusiveListNode<T> *it = head.next; it != &head; it = it->next)
	{
		ASSERT(this == reinterpret_cast<decltype(this)>(it->owner));
		ASSERT(it->data != nullptr);
		forward.push_back(it);
	}
	for (IntrusiveListNode<T> *it = head.prev; it != &head; it = it->prev)
	{
		back.push_back(it);
	}
	if (forward.size() != back.size())
	{
		std::cout << "Uh oh, forward / back size misatch!" << std::endl;
		std::cout << "Forward[" << forward.size() << "] = {";
		for (IntrusiveListNode<T> *e : forward)
		{
			std::cout << (int)e << ", ";
		}
		std::cout << std::endl;
		std::cout << "Back[" << back.size() << "] = {";
		for (IntrusiveListNode<T> *e : back)
		{
			std::cout << (int)e << ", ";
		}
		std::cout << std::endl;
	}
	ASSERT(forward.size() == back.size());
	for (int i = 0; i < forward.size(); ++i)
	{
		ASSERT(forward[i] == back[back.size() - i - 1]);
	}
}
#endif // SL_VALIDATE_INTRUSIVE_LIST

/*
	----------------------------------
			NODE IMPLEMENTATION
	----------------------------------
*/

template <typename T>
IntrusiveListNode<T>::IntrusiveListNode(T& data)
	:data(&data)
	,prev(nullptr)
	,next(nullptr)
#ifdef SL_VALIDATE_INTRUSIVE_LIST
	,owner(nullptr)
#endif // SL_VALIDATE_INTRUSIVE_LIST
{
}

template <typename T>
IntrusiveListNode<T>::IntrusiveListNode()
	:data(nullptr)
	,prev(this)
	,next(this)
#ifdef SL_VALIDATE_INTRUSIVE_LIST
	,owner(nullptr)
#endif // SL_VALIDATE_INTRUSIVE_LIST
{
}

//template <typename T>
//IntrusiveListNode<T>::IntrusiveListNode(IntrusiveListNode<T>&& move)
//	:IntrusiveListNode()
//{
//	std::swap(*this, move);
//}

template <typename T>
IntrusiveListNode<T>::~IntrusiveListNode()
{
	if (data) // regular node
	{
		unlink();
	}
	else // head of a list
	{
		ASSERT(prev == this && next == this);
	}
}

//template <typename T>
//IntrusiveListNode<T>& IntrusiveListNode<T>::operator=(IntrusiveListNode<T>&& rhs)
//{
//#if SL_DEBUG
//	owner->validate();
//#endif // SL_DEBUG
//	std::swap(*this, rhs);
//#if SL_DEBUG
//	owner->validate();
//#endif // SL_DEBUG
//	return *this;
//}

template <typename T>
void IntrusiveListNode<T>::unlink()
{
	if (prev == nullptr || next == nullptr)
	{
		ASSERT(prev == nullptr && next == nullptr);
	}
	else
	{
		ASSERT((prev != this) == (next != this));
		if (prev != this)
		{
			prev->next = next;
			next->prev = prev;
			prev = nullptr;
			next = nullptr;
#ifdef SL_VALIDATE_INTRUSIVE_LIST
			owner = nullptr;
#endif // SL_VALIDATE_INTRUSIVE_LIST SL_DEBUG
		}
	}
#ifdef SL_VALIDATE_INTRUSIVE_LIST
	ASSERT(owner == nullptr);
#endif // SL_VALIDATE_INTRUSIVE_LIST
}




/*
	--------------------------------
		ITERATOR IMPLEMENTATION
	--------------------------------
*/
template <typename T, IntrusiveListOwnershipType O>
IntrusiveList<T, O>::Iterator::Iterator(IntrusiveListNode<T> *node)
	:node(node)
{
}

template <typename T, IntrusiveListOwnershipType O>
bool IntrusiveList<T, O>::Iterator::operator==(const Iterator& rhs) const
{
	return this->node == rhs.node;
}

template <typename T, IntrusiveListOwnershipType O>
bool IntrusiveList<T, O>::Iterator::operator!=(const Iterator& rhs) const
{
	return this->node != rhs.node;
}

template <typename T, IntrusiveListOwnershipType O>
T& IntrusiveList<T, O>::Iterator::operator*() const
{
	ASSERT(node != nullptr);
	ASSERT(node->data != nullptr);//can't dereference past-end iterators
	return *(node->data);
}

template <typename T, IntrusiveListOwnershipType O>
T* IntrusiveList<T, O>::Iterator::operator->() const
{
	ASSERT(node != nullptr);
	ASSERT(node->data != nullptr);//can't dereference past-end iterators
	return node->data;
}

template <typename T, IntrusiveListOwnershipType O>
typename IntrusiveList<T, O>::Iterator& IntrusiveList<T, O>::Iterator::operator++()
{
	ASSERT(node != nullptr);
	node = node->next;
	return *this;
}

template <typename T, IntrusiveListOwnershipType O>
typename IntrusiveList<T, O>::Iterator IntrusiveList<T, O>::Iterator::operator++(int)
{
	ASSERT(node != nullptr);
	node = node->next;
	return IntrusiveList<T, O>::Iterator(node->prev);
}

template <typename T, IntrusiveListOwnershipType O>
typename IntrusiveList<T, O>::Iterator& IntrusiveList<T, O>::Iterator::operator--()
{
	ASSERT(node != nullptr);
	node = node->prev;
	return *this;
}

template <typename T, IntrusiveListOwnershipType O>
typename IntrusiveList<T, O>::Iterator IntrusiveList<T, O>::Iterator::operator--(int)
{
	ASSERT(node != nullptr);
	node = node->prev;
	return IntrusiveList<T, O>::Iterator(node->next);
}

template <typename T, IntrusiveListOwnershipType O>
typename IntrusiveList<T, O>::Iterator IntrusiveList<T, O>::Iterator::deleteData() const
{
	ASSERT(node != nullptr);
	IntrusiveListNode<T> *prev = node->prev;
	delete &(node->data);
	return IntrusiveList<T, O>::Iterator(prev);
}
/*			const iterator			*/
template <typename T, IntrusiveListOwnershipType O>
IntrusiveList<T, O>::ConstIterator::ConstIterator(const IntrusiveListNode<T> *node)
	:node(node)
{
}

template <typename T, IntrusiveListOwnershipType O>
bool IntrusiveList<T, O>::ConstIterator::operator==(const ConstIterator& rhs) const
{
	return this->node == rhs.node;
}

template <typename T, IntrusiveListOwnershipType O>
bool IntrusiveList<T, O>::ConstIterator::operator!=(const ConstIterator& rhs) const
{
	return this->node != rhs.node;
}

template <typename T, IntrusiveListOwnershipType O>
const T& IntrusiveList<T, O>::ConstIterator::operator*() const
{
	ASSERT(node != nullptr);
	ASSERT(node->data != nullptr);//can't dereference past-end iterators
	return *(node->data);
}

template <typename T, IntrusiveListOwnershipType O>
const T* IntrusiveList<T, O>::ConstIterator::operator->() const
{
	ASSERT(node != nullptr);
	ASSERT(node->data != nullptr);//can't dereference past-end iterators
	return node->data;
}

template <typename T, IntrusiveListOwnershipType O>
typename IntrusiveList<T, O>::ConstIterator& IntrusiveList<T, O>::ConstIterator::operator++()
{
	ASSERT(node != nullptr);
	node = node->next;
	return *this;
}

template <typename T, IntrusiveListOwnershipType O>
typename IntrusiveList<T, O>::ConstIterator IntrusiveList<T, O>::ConstIterator::operator++(int)
{
	ASSERT(node != nullptr);
	node = node->next;
	return IntrusiveList<T, O>::ConstIterator(node->prev);
}

template <typename T, IntrusiveListOwnershipType O>
typename IntrusiveList<T, O>::ConstIterator& IntrusiveList<T, O>::ConstIterator::operator--()
{
	ASSERT(node != nullptr);
	node = node->prev;
	return *this;
}

template <typename T, IntrusiveListOwnershipType O>
typename IntrusiveList<T, O>::ConstIterator IntrusiveList<T, O>::ConstIterator::operator--(int)
{
	ASSERT(node != nullptr);
	node = node->prev;
	return IntrusiveList<T, O>::ConstIterator(node->next);
}

} // sl


#pragma optimize( "", on )

#endif // SL_INTRUSIVELIST_HPP
