/**
 * @secton DESCRIPTION All IntrusiveMultiLists are assuemd to be non-owning. If this is needed we could specify it like in IntrusiveList
 */
#ifndef SL_INTRUSIVEMULTILIST_HPP
#define SL_INTRUSIVEMULTILIST_HPP

#include <map>

#include "Utility/Assert.hpp"

namespace sl
{

template <typename T>
class IntrusiveMultiList;

template <typename T>
class IntrusiveMultiListNode
{
public:
	IntrusiveMultiListNode(T& data);
	~IntrusiveMultiListNode();



	void unlink(IntrusiveMultiList<T>& list);
	void unlinkAll();

private:
	IntrusiveMultiListNode();
	IntrusiveMultiListNode(const IntrusiveMultiListNode&) = delete;
	IntrusiveMultiListNode& operator=(const IntrusiveMultiListNode&) = delete;

	struct LinkData
	{
		/**
		 * Base node belonging to the list itself
		 */
		LinkData()
			:prev(this)
			,next(this)
			,node(nullptr)
		{
		}
		/**
		 * Node belonging to elements
		 */
		LinkData(IntrusiveMultiListNode<T>& node, LinkData *prev, LinkData *next)
			:prev(prev)
			,next(next)
			,node(&node)
		{
		}
		LinkData(LinkData&&) = default;
		LinkData& operator=(LinkData&&) = default;


		typename IntrusiveMultiListNode<T>::LinkData *prev;
		typename IntrusiveMultiListNode<T>::LinkData *next;
		typename IntrusiveMultiListNode<T> *node;

	private:
		LinkData(const LinkData&) = delete;
		LinkData& operator=(const LinkData&) = delete;
	};
	std::map<const IntrusiveMultiList<T>*, LinkData> links;

	T * const data;


	friend class IntrusiveMultiList<T>;
};


template <typename T>
class IntrusiveMultiList
{
public:
	typedef T value_type;

	class Iterator
	{
	public:
		Iterator(typename IntrusiveMultiListNode<T>::LinkData& link);

		bool operator==(const Iterator& rhs) const;

		bool operator!=(const Iterator& rhs) const;

		T& operator*() const;

		T* operator->() const;

		Iterator& operator++();

		Iterator operator++(int);

		Iterator& operator--();

		Iterator operator--(int);

		//void deleteData() const;

	private:
		typename IntrusiveMultiListNode<T>::LinkData *link;
	};
	class ConstIterator
	{
	public:
		ConstIterator(typename const IntrusiveMultiListNode<T>::LinkData& link);

		bool operator==(const ConstIterator& rhs) const;

		bool operator!=(const ConstIterator& rhs) const;

		const T& operator*() const;

		const T* operator->() const;

		ConstIterator& operator++();

		ConstIterator operator++(int);

		ConstIterator& operator--();

		ConstIterator operator--(int);

	private:
		typename const IntrusiveMultiListNode<T>::LinkData *link;
	};

	IntrusiveMultiList();
	~IntrusiveMultiList();

	Iterator begin();
	Iterator end();
	ConstIterator begin() const;
	ConstIterator end() const;
	ConstIterator cbegin() const;
	ConstIterator cend() const;

	void clear();

	//void deleteAll();

	bool empty() const;

	void splice(IntrusiveMultiList<T>& donor);

	void insert(IntrusiveMultiListNode<T>& node);

	int size() const;

private:
	IntrusiveMultiList(const IntrusiveMultiList<T>&) = delete;
	IntrusiveMultiList<T>& operator=(const IntrusiveMultiList<T>&) = delete;


	typename IntrusiveMultiListNode<T>::LinkData head;

	friend class IntrusiveMultiListNode<T>;
};

/*
	----------------------------------
			LIST IMPLEMENTATION
	----------------------------------
*/

template <typename T>
IntrusiveMultiList<T>::IntrusiveMultiList()
	:head()
{
}

template <typename T>
IntrusiveMultiList<T>::~IntrusiveMultiList()
{
	this->clear();
}

template <typename T>
typename IntrusiveMultiList<T>::Iterator IntrusiveMultiList<T>::begin()
{
	return Iterator(*head.next);
}

template <typename T>
typename IntrusiveMultiList<T>::Iterator IntrusiveMultiList<T>::end()
{
	return Iterator(head);
}

template <typename T>
typename IntrusiveMultiList<T>::ConstIterator IntrusiveMultiList<T>::begin() const
{
	return cbegin();
}

template <typename T>
typename IntrusiveMultiList<T>::ConstIterator IntrusiveMultiList<T>::end() const
{
	return cend();
}

template <typename T>
typename IntrusiveMultiList<T>::ConstIterator IntrusiveMultiList<T>::cbegin() const
{
	return ConstIterator(*head.next);
}

template <typename T>
typename IntrusiveMultiList<T>::ConstIterator IntrusiveMultiList<T>::cend() const
{
	return ConstIterator(head);
}

template <typename T>
void IntrusiveMultiList<T>::clear()
{
	IntrusiveMultiListNode<T>::LinkData *temp = nullptr, *it = head.next;
	while (it != &head)
	{
		ASSERT(it->node != nullptr);
		temp = it->next;
		it->node->unlink(*this);
		it = temp;
	}
}

//template <typename T>
//void IntrusiveMultiList<T>::deleteAll()
//{
//	IntrusiveMultiListNode<T> *temp = nullptr, *it = head.next;
//	while (it != &head)
//	{
//		temp = it->next;
//		delete it->data;
//		it = temp;
//	}
//}

template <typename T>
bool IntrusiveMultiList<T>::empty() const
{
	return head.next == &head;
}

// TODO: implement with new intrusive list
//template <typename T>
//void IntrusiveMultiList<T>::splice(IntrusiveMultiList<T>& donor)
//{
//	if (!donor.empty())
//	{
//		head.prev->next donor.head.next;
//		
//		/*auto& donerLink = donor.head.links.at(&donor);
//		auto& link = head.links.at(this);
//		donorLink.next->prev = &link.head;
//		donerLink.head.prev->next = link.head.next;
//		link.head.next->prev = donerLink.head.prev;
//		link.head.next = donerLink.head.next;
//
//		donerLink.head.prev = donerLink.head.next = &donerLink.head*/
//
//		//auto& donerLink = donor.head.links.at(&donor);
//		//auto& link = head.links.at(this);
//		donor.head.links.at(&donor).next->prev = &head.links.at(this).head;
//		donor.head.links.at(&donor).head.prev->next = head.links.at(this).head.next;
//		head.links.at(this).head.next->prev = donor.head.links.at(&donor).head.prev;
//		head.links.at(this).head.next = donor.head.links.at(&donor).head.next;
//
//		donor.head.links.at(&donor).head.prev = donor.head.links.at(&donor).head.next = &donor.head.links.at(&donor).head;
//	}
//}

template <typename T>
void IntrusiveMultiList<T>::insert(IntrusiveMultiListNode<T>& node)
{
	ASSERT(node.data);//I really hope someone isn't inserting list heads...

	auto it = node.links.insert(std::make_pair(this, typename IntrusiveMultiListNode<T>::LinkData(node, head.prev, &head)));
	ASSERT(it.second);
	if (it.second)
	{
		typename IntrusiveMultiListNode<T>::LinkData& link = it.first->second;
		head.prev->next = &link;
		head.prev = &link;
	}
}

template <typename T>
int IntrusiveMultiList<T>::size() const
{
	int elems = 0;
	for (auto it = cbegin(), end = cend(); it != end; ++it)
	{
		++elems;
	}
	return elems;
}

/*
	----------------------------------
			NODE IMPLEMENTATION
	----------------------------------
*/

template <typename T>
IntrusiveMultiListNode<T>::IntrusiveMultiListNode(T& data)
	:data(&data)
{
}

template <typename T>
IntrusiveMultiListNode<T>::IntrusiveMultiListNode()
	:data(nullptr)
{
}

template <typename T>
IntrusiveMultiListNode<T>::~IntrusiveMultiListNode()
{
	this->unlinkAll();
}


template <typename T>
void IntrusiveMultiListNode<T>::unlink(IntrusiveMultiList<T>& list)
{
	const auto it = links.find(&list);
	ASSERT(it != links.end());
	if (it != links.end())
	{
		LinkData& link = it->second;
		if (link.prev || link.next)
		{
			ASSERT(link.next && link.prev);
			link.next->prev = link.prev;
			link.prev->next = link.next;
		}
		links.erase(it);
	}
}

template <typename T>
void IntrusiveMultiListNode<T>::unlinkAll()
{
	for (typename std::pair<const IntrusiveMultiList<T>* const, LinkData>& it : links)
	{
		LinkData& link = it.second;
		if (link.prev || link.next)
		{
			ASSERT(link.next && link.prev);
			link.next->prev = link.prev;
			link.prev->next = link.next;
		}
	}
	links.clear();
}




/*
	--------------------------------
		ITERATOR IMPLEMENTATION
	--------------------------------
*/
template <typename T>
IntrusiveMultiList<T>::Iterator::Iterator(typename IntrusiveMultiListNode<T>::LinkData& link)
	:link(&link)
{
}

template <typename T>
bool IntrusiveMultiList<T>::Iterator::operator==(const Iterator& rhs) const
{
	return link == rhs.link;
}

template <typename T>
bool IntrusiveMultiList<T>::Iterator::operator!=(const Iterator& rhs) const
{
	return link != rhs.link;
}

template <typename T>
T& IntrusiveMultiList<T>::Iterator::operator*() const
{
	ASSERT(link->node != nullptr);
	ASSERT(link->node->data != nullptr);//can't dereference past-end iterators
	return *(link->node->data);
}

template <typename T>
T* IntrusiveMultiList<T>::Iterator::operator->() const
{
	ASSERT(link->node != nullptr);
	ASSERT(link->node->data != nullptr);//can't dereference past-end iterators
	return link->node->data;
}

template <typename T>
typename IntrusiveMultiList<T>::Iterator& IntrusiveMultiList<T>::Iterator::operator++()
{
	link = link->next;
	return *this;
}

template <typename T>
typename IntrusiveMultiList<T>::Iterator IntrusiveMultiList<T>::Iterator::operator++(int)
{
	IntrusiveMultiList<T>::Iterator old(link*);
	link = link->next;
	return old;
}

template <typename T>
typename IntrusiveMultiList<T>::Iterator& IntrusiveMultiList<T>::Iterator::operator--()
{
	link = link->prev;
	return *this;
}

template <typename T>
typename IntrusiveMultiList<T>::Iterator IntrusiveMultiList<T>::Iterator::operator--(int)
{
	IntrusiveMultiList<T>::Iterator old(link*);
	link = link->prev;
	return old;
}

//template <typename T>
//void IntrusiveMultiList<T>::Iterator::deleteData() const
//{
//	ASSERT(node != nullptr);
//	delete &(node->data);
//}
/*			const iterator			*/
template <typename T>
IntrusiveMultiList<T>::ConstIterator::ConstIterator(typename const IntrusiveMultiListNode<T>::LinkData& link)
	:link(&link)
{
}

template <typename T>
bool IntrusiveMultiList<T>::ConstIterator::operator==(const ConstIterator& rhs) const
{
	return link == rhs.link;
}

template <typename T>
bool IntrusiveMultiList<T>::ConstIterator::operator!=(const ConstIterator& rhs) const
{
	return link != rhs.link;
}

template <typename T>
const T& IntrusiveMultiList<T>::ConstIterator::operator*() const
{
	ASSERT(link->node != nullptr);
	ASSERT(link->node->data != nullptr);//can't dereference past-end iterators
	return *(link->node->data);
}

template <typename T>
const T* IntrusiveMultiList<T>::ConstIterator::operator->() const
{
	ASSERT(link->node != nullptr);
	ASSERT(link->node->data != nullptr);//can't dereference past-end iterators
	return link->node->data;
}

template <typename T>
typename IntrusiveMultiList<T>::ConstIterator& IntrusiveMultiList<T>::ConstIterator::operator++()
{
	link = link->next;
	return *this;
}

template <typename T>
typename IntrusiveMultiList<T>::ConstIterator IntrusiveMultiList<T>::ConstIterator::operator++(int)
{
	IntrusiveMultiList<T>::Iterator old(link*);
	link = link->next;
	return old;
}

template <typename T>
typename IntrusiveMultiList<T>::ConstIterator& IntrusiveMultiList<T>::ConstIterator::operator--()
{
	link = link->prev;
	return *this;
}

template <typename T>
typename IntrusiveMultiList<T>::ConstIterator IntrusiveMultiList<T>::ConstIterator::operator--(int)
{
	IntrusiveMultiList<T>::Iterator old(link*);
	link = link->prev;
	return old;
}

}

#endif // SL_INTRUSIVEMULTILIST_HPP
