#ifndef SL_LINEARINTERPOLATOR_HPP
#define SL_LINEARINTERPOLATOR_HPP

#include <algorithm>
#include <map>

#include "Utility/Assert.hpp"
#include "Utility/Interpolator.hpp"

namespace sl
{

template <typename T, typename U = float>
class LinearInterpolator : public Interpolator<T, U>
{
public:
	LinearInterpolator();

	template <typename Container>
	LinearInterpolator(const Container& container);

	T interpolate(const U& x) const override;

	template <typename Container>
	void addPoints(const Container& container);

	void addPoint(const T& point, const U& value);

	void addPoint(T&& point, U&& value);

private:

	std::map<U, T> points;
};

template <typename T, typename U>
LinearInterpolator<T, U>::LinearInterpolator()
	:points()
{
}

template <typename T, typename U>
template <typename Container>
LinearInterpolator<T, U>::LinearInterpolator(const Container& container)
	:Interpolator<T, U>()
	,points()
{
	addPoints(container);
}

template <typename T, typename U>
T LinearInterpolator<T, U>::interpolate(const U& x) const
{
	if (points.empty())
	{
		ERROR("No defined points - cannot interpolate");
		return x;
	}
	if (points.size() == 1)
	{
		return points.begin()->second;
	}
	auto upper = points.upper_bound(x);
	std::pair<U, T> lowerEnd;
	std::pair<U, T> upperEnd;
	if (upper == points.end())
	{
		lowerEnd = *std::next(points.rbegin());
		upperEnd = *points.rbegin();
	}
	else if (x < points.begin()->first)
	{
		lowerEnd = *points.begin();
		upperEnd = *std::next(points.begin());
	}
	else
	{
		lowerEnd = *std::prev(upper);
		upperEnd = *upper;
	}
	
	//(lowerEnd.first <= x && x <= upperEnd.first);
	const U dist = upperEnd.first - lowerEnd.first;
	ASSERT(dist > 0.f);
	const T slope = (upperEnd.second - lowerEnd.second) * (1.f / dist);
	return slope * (x - lowerEnd.first) + lowerEnd.second;
}

template <typename T, typename U>
template <typename Container>
void LinearInterpolator<T, U>::addPoints(const Container& container)
{
	//points.insert(std::begin(container), std::end(container));
	for (const auto& point : container)
	{
		points.insert(point);
	}
}

template <typename T, typename U>
void LinearInterpolator<T, U>::addPoint(const T& point, const U& value)
{
	points.insert(std::make_pair(value, point));
}

template <typename T, typename U>
void LinearInterpolator<T, U>::addPoint(T&& point, U&& value)
{
	points.insert(std::make_pair(std::move(value), std::move(point)));
}

} // sl

#endif // SL_LINEARINTERPOLATOR_HPP