#ifndef SL_LOGGER_HPP
#define SL_LOGGER_HPP

#include <fstream>
#include <ostream>
#include <sstream>

namespace sl
{

class Logger
{
public:
	enum OutputLocation
	{
		Ignored,
		File,
		Console,
		Screen,
		Popup
	};

	enum Area
	{
		Core = 0,
		Dialogue,
		Graphics,
		GUI,
		NPC,
		Parsing,
		Pathfinding,
		Physics,
		Serialization,
		Town,
		TownGen,
		Utility,
		Temp
	};

	enum Priority
	{
		Useless = 0,
		Debug,
		Serious,
		UltraSerious
	};

	class Stream;

	Logger();


	/**
	 * Logs a message using the default logger in the given area if the prioty is sufficient
	 * @param area The area to log to
	 * @priority The priority of the message
	 * @return The appropriate ostream wrapper to log to
	 */
	static Stream log(Area area = Area::Temp, Priority priority = Priority::Useless);

	static void setDefaultLogger(Logger *logger);

		/**
	 * Logs a message in the given area if the prioty is sufficient
	 * @param area The area to log to
	 * @priority The priority of the message
	 * @return The appropriate ostream wrapper to log to
	 */
	Stream logTo(Area area = Area::Temp, Priority priority = Priority::Useless) const;


	/**
	 * Sets the output location for all Areas
	 * @param location The medium to output to
	 */
	void setOutputLocation(OutputLocation location);

	/**
	  * Sets the output location for a given area only
	  * @param area The Area to configure
	  * @param location The medium to output to
	  */
	void setOutputLocation(Area area, OutputLocation location);

	/**
	  * Sets the output location for only for a certain message priority
	  * @param priority The needed priority
	  * @param location The medium to output to
	  */
	void setOutputLocation(Priority priority, OutputLocation location);

	/**
	  * Sets the output location for a given area only and only for a certain message priority
	  * @param area The Area to configure
	  * @param priority The needed priority
	  * @param location The medium to output to
	  */
	void setOutputLocation(Area area, Priority priority, OutputLocation location);

	/**
	 * Sets the minimum priority for all areas
	 * @param priority The priority a message needs before it's displayed
	 */
	void setMinimumPriority(Priority priority);

	/**
	 * Sets the minimum priority for a given area
	 * @param area The area
	 * @param priority The priority a message needs before it's displayed
	 */
	void setMinimumPriority(Area area, Priority priority);

	/**
	 * Enables all logging globally (doesn't affect settings)
	 */
	void enable();

	/**
	 * Disables all logging globally (doesn't affect settings)
	 */
	void disable();

private:

	Logger(const Logger&);
	Logger& operator=(const Logger&);



	static const int AREAS = 11;
	static const int PRIORITIES = 4;
	static Logger *instance;

	Priority priorities[AREAS];
	OutputLocation outputs[PRIORITIES][AREAS];
	bool enabled;
};


/**
 * Class to be returned as a temporary object. Upon destruction it will do output-location-specific
 * stuff such as open popups/close files.
 */
class Logger::Stream
{
public:
	Stream(Logger::OutputLocation outputLocation);
	Stream(Stream&& rhs);
	~Stream();

	std::ostream& get();

	template <typename T>
	std::ostream& operator<<(const T& val)
	{
		return get() << val;
	}

private:
	//Stream(const Stream&);
	//Stream& operator=(const Stream&);
	//Stream&& operator= (Stream&&);

	/**
	 * An std::ostream which simply ignores everything
	 */
	class IgnoreStream : public std::ostream, private std::streambuf
	{
	public:
		IgnoreStream()
			:std::ostream(this)
		{
		}

		template <typename T>
		std::ostream& operator<<(const T& val) const
		{
			return *this;
		}
	};


	const Logger::OutputLocation outputLocation;
	IgnoreStream ignoreStream;
	std::ofstream fileStream;
	std::stringstream popupStream;
};

} // namespace sl

#endif // SL_LOGGER_HPP
