#ifndef SL_MATRIX_HPP
#define SL_MATRIX_HPP

#include <vector>

#include "Utility/Assert.hpp"

namespace sl
{

template <typename T>
class Matrix
{
public:
	Matrix();

	Matrix(int rows, int cols, const T& defaultValue = T());

	template <typename U, int M, int N>
	Matrix(const T arr[M][N]);

	const T& operator()(int row, int col) const;

	T& operator()(int row, int col);

	int getRows() const;

	int getCols() const;

	Matrix<T> operator+(const Matrix<T>& rhs) const;

	Matrix<T>& operator+=(const Matrix<T>& rhs);

	Matrix<T> operator*(const Matrix<T>& rhs) const;

	Matrix<T>& operator*=(const Matrix<T>& rhs);

	Matrix<T> pow(int n) const;

	Matrix<T>& exponentiate(int n);


private:

	int rows;
	int cols;
	std::vector<T> data;
};

template <typename T>
Matrix<T>::Matrix()
	:rows(0)
	,cols(0)
	,data()
{
}

template <typename T>
Matrix<T>::Matrix(int rows, int cols, const T& defaultValue)
	:rows(rows)
	,cols(cols)
	,data(rows * cols, defaultValue)
{
}


template <typename T>
template <typename U, int M, int N>
Matrix<T>::Matrix(const T arr[M][N])
	:rows(rows)
	,cols(cols)
	,data(std::begin(arr), std::end(arr))
{
}


template <typename T>
const T& Matrix<T>::operator()(int row, int col) const
{
	return data[row * rows + col];
}

template <typename T>
T& Matrix<T>::operator()(int row, int col)
{
	return data[row * rows + col];
}

template <typename T>
int Matrix<T>::getRows() const
{
	return rows;
}

template <typename T>
int Matrix<T>::getCols() const
{
	return cols;
}

template <typename T>
Matrix<T> Matrix<T>::operator+(const Matrix<T>& rhs) const
{
	Matrix<T> ret(*this);
	ret += rhs;
	return std::move(ret);
}

template <typename T>
Matrix<T>& Matrix<T>::operator+=(const Matrix<T>& rhs)
{
	ASSERT(rows == rhs.rows && cols == rhs.cols);
	const int n = rows * cols;
	for (int i = 0; i < n; ++i)
	{
		data[i] += rhs.data[i];
	}
}

template <typename T>
Matrix<T> Matrix<T>::operator*(const Matrix<T>& rhs) const
{
	ASSERT(cols == rhs.rows);
	Matrix<T> ret(rows, rhs.cols);
	for (int c = 0; c < rhs.cols; ++c)
	{
		for (int r = 0; r < rows; ++r)
		{
			T sum = T();
			for (int i = 0; i < cols; ++i)
			{
				sum += (*this)(r, i) * rhs(i, c);
			}
			ret(r, c) = sum;
		}
	}
	return std::move(ret);
}

template <typename T>
Matrix<T>& Matrix<T>::operator*=(const Matrix<T>& rhs)
{
	*this = *this * rhs;
	return *this;
}

template <typename T>
Matrix<T> Matrix<T>::pow(int n) const
{
	Matrix<T> ret(*this);
	ret.exponentiate(n);
	return std::move(ret);
}

template <typename T>
Matrix<T>& Matrix<T>::exponentiate(int n)
{
	ASSERT(rows == cols);
	Matrix<T> copy(*this);
	// @TODO optimize to get log-time exponentiation
	for (int i = 0; i < n; ++i)
	{
		*this *= copy;
	}
	return *this;
}

} // sl

#endif // SL_MATRIX_HPP
