/**
 * Due to the lack of std::function's move operators being able to move any wrapped functors (ie lambdas)
 * it is borderline useless for non-copyable lambdas. This is a simple wrapper that works around it and has
 * a similar interface.
 */
#ifndef SL_MOVEABLEFUNCTION_HPP
#define SL_MOVEABLEFUNCTION_HPP

#include <memory>

namespace sl
{

template <typename Return, typename... Arguments>
class MoveableFunction
{
	class Interface
	{
	public:
		virtual ~Interface() {}

		virtual Return operator()(Arguments... arguments) = 0;

		virtual Return operator()(Arguments... arguments) const = 0;
	};

	template <typename F>
	class Impl : public Interface
	{
	public:
		Impl(F&& f)
			:f(std::move(f))
		{
		}

		Return operator()(Arguments... arguments) override
		{
			return f(arguments...);
		}

		Return operator()(Arguments... arguments) const override
		{
			return f(arguments...);
		}

	private:
		F f;
	};
public:
	MoveableFunction()
		:f()
	{
	}

	//template <typename F>
	//MoveableFunction(F f)
	//	:f(new Impl<F>(std::move(f)))
	//{
	//}

	template <typename F>
	MoveableFunction(F&& f)
		:f(std::make_unique<Impl<F>>(std::move(f)))
	{
	}

	operator bool() const
	{
		return f.operator bool();
	}

	Return operator()(Arguments... arguments)
	{
		return (*f)(arguments...);
	}

	Return operator()(Arguments... arguments) const
	{
		return (*f)(arguments...);
	}

private:

	std::unique_ptr<Interface> f;
};

} // sl

#endif // SL_MOVEABLEFUNCTION_HPP