#ifndef SL_MULTIVARIATEVORONOI_HPP
#define SL_MULTIVARIATEVORONOI_HPP

#include "Utility/Vector.hpp"

#include <algorithm>
#include <tuple>
#include <vector>

namespace sl
{

/**
 * Computes neighbors of a voronoi cell around the cell that query lays within
 * If cullBehind=true, then additionally culls all voronoi neighbors that have dot products <= 0
 * with the vector between query's cell center and query. This is more of an optimization since doing it beforehand
 * could reduce the time a lot for the loops.
 * @param points All points that define a voronoi diagram in n-space
 * @param query Point to query into
 * @return 1st element is the index within points of voronoi cell query is in, then after the indices of the voronoi cells neighboring query 
 */
template <typename Vec, bool cullBehind = false>
std::vector<int> computeVoronoiNeighbors(const std::vector<Vec>& points, const Vec& query)
{
	using T = Vec::Field;
	// <Point, Index in points, Distance from query>
	const int n = points.size();
	std::vector<std::tuple<Vec, int, T>> pts;
	pts.reserve(n);
	for (int i = 0; i < n; ++i)
	{
		pts.push_back(std::forward_as_tuple(points[i], i, (query - points[i]).length()));
	}
	std::sort(pts.begin(), pts.end(), [](const auto& lhs, const auto& rhs) {
		return std::get<2>(lhs) < std::get<2>(rhs);
	});
	// skip the 1st one (cell you're in)
	const Vec& u = std::get<0>(pts.front());
	const T zero = T();
	std::vector<char> isCulled(n, 0);
	if constexpr (cullBehind)
	{
		const Vec uToQuery = query - u;
		for (int i = 1; i < n; ++i)
		{
			const Vec& v = std::get<0>(pts[i]);
			if (dot(uToQuery, v - u) <= 0)
			{
				isCulled[i] = 1;
			}
		}
	}
	for (int i = 1; i < n; ++i)
	{
		if (isCulled[i] == 0)
		{
			const Vec& v = std::get<0>(pts[i]);
			const Vec& vu = u - v;
			for (int j = 1; j < n; ++j)
			{
				if (i != j)
				{
					const Vec& w = std::get<0>(pts[j]);
					const Vec vw = w - v;
					if (dot(vu, vw) <= zero)
					{
						isCulled[j] = 1;
					}
				}
			}
		}
	}
	std::vector<int> output;
	for (int i = 0; i < n; ++i)
	{
		if (isCulled[i] == 0)
		{
			output.push_back(std::get<1>(pts[i]));
		}
	}
	return output;
}

} // sl

#endif // SL_MULTIVARIATEVORONOI_HPP