#ifndef SL_PROPORTIONLIST_HPP
#define SL_PROPORTIONLIST_HPP

#include <vector>

#include "Utility/Assert.hpp"
#include "Utility/Random.hpp"

namespace sl
{

template <typename T>
class ProportionList
{
public:
	class Element
	{
	public:
		Element(const T& data, double rate)
			:data(data)
			,rate(rate)
		{
		}
		Element(T&& data, double rate)
			:data(data)
			,rate(rate)
		{
		}

		T data;
		double rate;
	};

	//ProportionList();

	template <typename... Rest>
	ProportionList(Rest... rest);

	void insert(const T& element, double frequency);

	void insert(T&& element, double frequency);

	template <typename... Rest>
	void insert(const T& element, double frequency, Rest... rest);

	template <typename... Rest>
	void insert(T&& element, double frequency, Rest... rest);

	const T& get(Random& random) const;

	bool empty() const;

	std::size_t size() const;

	typename std::vector<Element>::const_iterator cbegin() const
	{
		return data.cbegin();
	}

	typename std::vector<Element>::const_iterator cend() const
	{
		return data.cend();
	}

	const T& operator[](std::size_t index) const;

private:
	void insert() {} // to avoid writing 3 constructors, so the one construct works as all 3 (no-arg / T&&/ const T&)

	typename std::vector<Element> data;
	double total;
};

template <typename T>
template <typename... Rest>
ProportionList<T>::ProportionList(Rest... rest)
	:total(0.0)
{
	insert(rest...);
}

template <typename T>
void ProportionList<T>::insert(const T& element, double frequency)
{
	data.push_back(Element(element, frequency));
	total += frequency;
	ASSERT(frequency > 0.0001);
}

template <typename T>
void ProportionList<T>::insert(T&& element, double frequency)
{
	data.push_back(Element(std::move(element), frequency));
	total += frequency;
	ASSERT(frequency > 0.0001);
}

template <typename T>
template <typename... Rest>
void ProportionList<T>::insert(const T& element, double frequency, Rest... rest)
{
	insert(element, frequency);
	insert(rest...);
}

template <typename T>
template <typename... Rest>
void ProportionList<T>::insert(T&& element, double frequency, Rest... rest)
{
	insert(element, frequency);
	insert(rest...);
}

template <typename T>
const T& ProportionList<T>::get(Random& random) const
{
	ASSERT(!empty());
	double n = random.randomFloat(total);
	for (const Element& e : data)
	{
		n -= e.rate;
		if (n <= 0.0001)
		{
			return e.data;
		}
	}
	ERROR("This is broken -- maybe increase the negligible value?");
	return data.front().data;
}

template <typename T>
bool ProportionList<T>::empty() const
{
	return data.empty();
}

template <typename T>
std::size_t ProportionList<T>::size() const
{
	return data.size();
}

template <typename T>
const T& ProportionList<T>::operator[](std::size_t index) const
{
	return data[index].data;
}

} // sl

#endif // SL_PROPORTIONLIST_HPP
