#include "Utility/Random.hpp"

#include <chrono>
#include <cmath>

#include "Utility/Assert.hpp"

namespace sl
{

Random* Random::instance = nullptr;

Random::Random()
	:generator(std::chrono::system_clock::now().time_since_epoch().count())
	,oldGeneration(false)
{
}

Random::Random(unsigned int seed)
	:generator(seed)
	,oldGeneration(false)
{
}

float Random::randomFloat(float min, float max)
{
	ASSERT(min <= max);
	return min + randomFloat(max - min);
}

int Random::randomInclusive(int min, int max)
{
	return min + generator() % (max - min + 1);
}

int Random::randomExclusive(int min, int max)
{
	return min + generator() % (max - min - 1) + 1;
}

float Random::randomFloat(float n)
{
	return n * (generator() % 0xFFFF) / (float) 0xFFFF;
}

int Random::random(int n)
{
	return n > 0 ? generator() % n : 0;
}

bool Random::chance(float n)
{
	return ((generator() % 0xFFFF) / (float) 0xFFFF) < n;
}

float Random::randomFromNormalDistribution(float meanValue, float standardDeviation)
{
	/*
		See http://en.wikipedia.org/wiki/Marsaglia_polar_method
		for a better explanation.
	*/

	if (oldGeneration)
	{
		oldGeneration = false;
		return meanValue + uglySlowLogarithmSquareRootShit * oldBValue * standardDeviation;
	}

	float a, b;	// The two uniformly distributed random numbers
	float s;	// The length formed between (a,b) and (0,0)
	// Generate two linearly distributed values within (-1, 1)
	// such that a^2 + b^2 < 1, basically it's generating coordinates within
	// a unit circle. x^2 + y^2 = r^2, so the radius must be < 1
	do
	{
		a = randomFloat(-1, 1);
		b = randomFloat(-1, 1);
		s = a * a + b * b;
	}
	while (s >= 1);

	// We can reuse this ugly thing for the next number generated
	uglySlowLogarithmSquareRootShit = sqrt((-2 * log(s)) / s);
	oldBValue = b;
	oldGeneration = true;


	// And now to adjust it to the given values
	return meanValue + uglySlowLogarithmSquareRootShit * a * standardDeviation;
}



/*	  Static methods	  */
Random& Random::get()
{
	if (!instance)
	{
		instance = new Random(time(0));
	}
	return *instance;
}

void Random::cleanup()
{
	if (instance)
	{
		delete instance;
		instance = nullptr;
	}
}

}
