/**
 * @section DESCRIPTION
 * A class that can be used to generate random numbers, all packaged up as one class.
 * Note: this class was designed before <random> was supported by my compiler so
 * that's why it implements its own standard deviation and stuff since it used
 * to be based on rand().
 */
#ifndef SL_RANDOM_HPP
#define SL_RANDOM_HPP

#include <map>
#include <random>
#include <initializer_list>

namespace sl
{

class Random
{
public:
	/**
	 * Constructs a Random object using the current time as a seed
	 */
	Random();
	/**
	 * Constructs a Random object with a custom seed
	 * @param seed The custom seed to use
	 */
	Random(unsigned int seed);


	/**
	 * Generates a value that will be normally distributed with respect to the given input.
	 * @param meanValue The mean of the distribution
	 * @param standardDeviation The standard deviation of the distribution
	 * @return The normally distributed random value generated
	 */
	float randomFromNormalDistribution(float meanValue, float standardDeviation);
	/**
	 * Generates a floating point number between two numbers
	 * @param min The minimum value
	 * @param max The maximum value
	 * @return The randomly generated value
	 */
	float randomFloat(float min, float max);
	/**
	 * Generates a random integer between two integers exclusive.
	 * min < return < max
	 * @param min The minimum value
	 * @param max The maximum value
	 * @return The random value generated.
	 */
	int randomExclusive(int min, int max);
	/**
	 * Generates a random integer between two integers inclusive.
	 * min <= return <= max
	 * @param min The minimum value
	 * @param max The maximum value
	 * @return The random value generated.
	 */
	int randomInclusive(int min, int max);
	/**
	 * Generates a random integer.
	 * 0 <= return < n
	 * @param n The upper bounds
	 * @return The randomly generated value
	 */
	int random(int n);
	/**
	 * Generates a random float.
	 * 0 <= return < n
	 * @param n The upper bounds
	 * @return The randomly generated value
	 */
	float randomFloat(float n);
	/**
	 * Simulates the chance of a probability.
	 * ie if n = 0.5, it will return true 50% of the time
	 * and if n = 0.1 it will return true 10% of the time, etc
	 * @param n The probability of returning true
	 * @return Whether or not the probability results in true
	 */
	bool chance(float n);
	/**
	 * Returns one of the inputs randomly
	 * @param il List of things of type T to choose from
	 * @return The randomly chosen element from il
	 */
	template <typename T>
	T choose(typename std::initializer_list<T> il)
	{
		return *(il.begin() + random(il.size()));
	}
	template <typename... Args>
	auto choose(Args... args) -> decltype(chooseVT(0, args...))
	{
		const int n = random(sizeof...(args));
		return chooseVT(n, args...);
	}
	template <typename T, template <typename, typename> class ContainerOfT>
	const T& choose(const ContainerOfT<T, std::allocator<T>>& container)
	{
		auto it = std::begin(container);
		std::advance(it, random(container.size()));
		return *it;
	}
	template <typename T, typename U>
	std::pair<U, T> choose(const std::map<U, T>& map)
	{
		auto it = std::begin(map);
		std::advance(it, random(map.size()));
		return *it;
	}
	/**
	 * Shuffles any STL container supporting random-access iterators
	 * Efficiency: O(n) - linear with number of elements to shuffle
	 * @param first A Random access iterator to the first element (ie .begin())
	 * @param last A random access iterator to the last element (ie .end())
	 */
	template <class RandomAccessIterator>
	void shuffle(RandomAccessIterator first, RandomAccessIterator last)
	{
		std::size_t n = (last - first);
		for (std::size_t i = 0; i < n; ++i)
		{
			std::swap(first[i], first[this->randomInclusive(i, n - 1)]);
		}
	}


	/**
	 * Get a singleton Random object that can be used for when a specific, lasting
	 * Random object isn't needed, or isn't possible. Good for when you're lazy.
	 * @return A reference to the singleton Random instance
	 */
	static Random& get();
	/**
	 * Deletes the singleton anonymous Random object if it exists
	 */
	static void cleanup();
private:
	template <typename T, typename... Args>
	T chooseVT(int n, const T& t, Args... args) const
	{
		return n > 0 ? chooseVT(n - 1, args...) : t;
	}
	template <typename T>
	T chooseVT(int n, const T& t) const
	{
		// ASSERT(n == 0);
		return t;
	}

	//!The random number generator that powers this class
	std::mt19937 generator;
	//!Whether we can reuse the next 2 variables for normal distribution
	bool oldGeneration;
	//!Used for more efficient normal distribution generation (since you generate 2 unscaled each time)
	float uglySlowLogarithmSquareRootShit;
	//!Used for more efficient normal distribution generation (since you generate 2 unscaled each time)
	float oldBValue;


	//!The singleton Random instance used for get()
	static Random *instance;
};

}

#endif
