/**
 * @section DESCRIPTION
 * A basic class to represent a rectangle. This is here to fill where I was using sf::Rect before.
 * I don't think much more of a description is really ncecessary here.
 */
#ifndef SL_RECT_HPP
#define SL_RECT_HPP

#include "Utility/Vector.hpp"

namespace sl
{

template <typename T>
struct Rect
{
	/**
	 * Constrcuts a rectangle, most likely (calls default constructors for T)
	 * at position 0,0 with 0 area.
	 */
	Rect();
	/**
	 * Constructs a rectangle with the given size at the default position
	 * @param width The width of the rectangle
	 * @param height The height of the rectangle
	 */
	Rect(const T& width, const T& height);
	/**
	 * Constucts a rectangle of the given size/loction
	 * @param left The left side's x position
	 * @param top The top side's y position
	 * @param width The width of the rectangle
	 * @param height The height of the rectangle
	 */
	Rect(const T& left, const T& top, const T& width, const T& height);
	/**
	 * Copy constructor. Makes a new rectangle of the same size/location as the input one
	 * @param other The Rect to copy
	 */
	Rect(const Rect<T>& other);
	/**
	 * Copies all data from rhs to this rectangle
	 * @param rhs The right hand side in the equation derp = rhs;
	 */
	inline Rect<T>& operator=(const Rect<T>& rhs);
	/**
	 * Compares if two Rects are equal
	 * @return true if they are equal, false otherwise
	 */
	inline bool operator==(const Rect<T>& rhs) const;
	/**
	 * Compares if two Rects are not equal
	 * @return false if they are equal, true otherwise
	 */
	inline bool operator!=(const Rect<T>& rhs) const;

	/**
	 * Returns whether or not two rectangles intersect each other.
	 * @param other The other rectangle to test againt.
	 * @return Whether or not they intersect at all.
	 */
	inline bool intersects(const Rect<T>& other) const;

	/**
	 * Returns whether or not a point is inside the bunds of the Rect
	 * @param x The x coordinate of the point to test
	 * @param y The y coordinate of the point to test
	 * @return Whether or not the point was within the Rect's bounds
	 */
	inline bool containsPoint(const T& x, const T& y) const;
	inline bool containsPoint(const Vector2<T>& pos) const;

	/**
	 * Returns the area of the rectangle.
	 * @return The area
	 */
	inline T area() const;

	/**
	 * Gets an intersection between two rectangles.
	 * @param other The other rectangle to check with
	 * @return The intersection if it exists, or a rect with 0 left/top/width/height otherwise
	 */
	inline Rect<T> getIntersection(const Rect<T>& other) const;

	/**
	 * Creates a new Rect that has been stretched and is centered in the center of the old one
	 * @param scalingFactor The factor by which the Rect is stretched
	 * @return The scaled Rect
	 */
	inline Rect<T> createTransformedRect(float scalingFactor) const;

	/**
	 * Creates a new Rect that has been stretched and is centered in the center of the old one
	 * @param horizScalingFactor The factor by which the Rect is stretched horizontally
	 * @param vertScalingFactor The factor by which the Rect is stretched vertically
	 * @return The scaled Rect
	 */
	inline Rect<T> createTransformedRect(float horizScalingFactor, float vertScalingFactor) const;

	/**
	 * Moves the rect by the given point vector
	 * @param point The offset to move the rect by
	 * @return The moved rect
	 */
	inline Rect<T>& operator+=(const Vector2<T>& point);

	/**
	 * Moves the rect by the given point vector negatively.
	 * @param point The offset to move the rect by
	 * @return The moved rect
	 */
	inline Rect<T>& operator-=(const Vector2<T>& point);

	inline Rect<T>& operator*=(const T& scalar);

	inline Rect<T>& operator/=(const T& scalar);

	/**
	 * Linearly maps from input relative to this, to a value relative to targetRect
	 * @param input. Input position (relative to this)
	 * @param targetRect The output region
	 * @return Position relative to targetRect
	 */
	Vector2<T> mapTo(const Vector2<T>& input, const Rect<T>& targetRect) const;

	template <typename U>
	inline Rect<U> to() const
	{
		return Rect<U>(static_cast<U>(left), static_cast<U>(top), static_cast<U>(width), static_cast<U>(height));
	}

	inline Vector2<T> topLeft() const
	{
		return Vector2<T>(left, top);
	}

	inline Vector2<T> topRight() const
	{
		return Vector2<T>(left + width, top);
	}

	inline Vector2<T> bottomLeft() const
	{
		return Vector2<T>(left, top + height);
	}

	inline Vector2<T> bottomRight() const
	{
		return Vector2<T>(left + width, top + height);
	}

	inline Vector2<T> size() const
	{
		return Vector2<T>(width, height);
	}

	inline T right() const
	{
		return left + width;
	}

	inline T bottom() const
	{
		return top + height;
	}

	//!The x position of the left side of the rectangle
	T left;
	//!The y position of the top side of the rectangle
	T top;
	//!The width of the rectangle
	T width;
	//!The height of the rectangle
	T height;
};

template <typename T>
Rect<T> operator+(const Rect<T>& lhs, const Vector2<T>& rhs);

template <typename T>
Rect<T> operator+(const Vector2<T>& lhs, const Rect<T>& rhs);

template <typename T>
Rect<T> operator-(const Rect<T>& lhs, const Vector2<T>& rhs);



template <typename T>
Rect<T>::Rect()
	:left()
	,top()
	,width()
	,height()
{
}

template <typename T>
Rect<T>::Rect(const T& width, const T& height)
	:left()
	,top()
	,width(width)
	,height(height)
{
}

template <typename T>
Rect<T>::Rect(const T& left, const T& top, const T& width, const T& height)
	:left(left)
	,top(top)
	,width(width)
	,height(height)
{
}

template <typename T>
Rect<T>::Rect(const Rect<T>& other)
	:left(other.left)
	,top(other.top)
	,width(other.width)
	,height(other.height)
{
}

template <typename T>
Rect<T>& Rect<T>::operator=(const Rect<T>& rhs)
{
	this->left = rhs.left;
	this->top = rhs.top;
	this->width = rhs.width;
	this->height = rhs.height;
	return *this;
}

template <typename T>
bool Rect<T>::operator==(const Rect<T>& rhs) const
{
	return this->left == rhs.left && this->top == rhs.top && this->width == rhs.width && this->height == rhs.height;
}

template <typename T>
bool Rect<T>::operator!=(const Rect<T>& rhs) const
{
	return this->left != rhs.left || this->top != rhs.top || this->width != rhs.width || this->height != rhs.height;
}

template <typename T>
bool Rect<T>::intersects(const Rect<T>& other) const
{
	return (left <= other.left + other.width && left + width >= other.left) && (top <= other.top + other.height && top + height >= other.top);
}

template <typename T>
bool Rect<T>::containsPoint(const T& x, const T& y) const
{
	return (x >= left && x < left + width && y >= top && y < top + height);
}

template <typename T>
inline bool Rect<T>::containsPoint(const Vector2<T>& pos) const
{
	return containsPoint(pos.x, pos.y);
}

template <typename T>
T Rect<T>::area() const
{
	return width * height;
}


template <typename T>
Rect<T> Rect<T>::getIntersection(const Rect<T>& other) const
{
	if (this->intersects(other))
	{
		const T x1 = left > other.left ? left : other.left;
		const T y1 = top > other.top ? top : left;
		const T x2 = (left + width) < (other.left + other.width) ?
					 (left + width) : (other.left + other.width);
		const T y2 = (top + height) < (other.top + other.height) ?
					 (top + height) : (other.top + other.height);
		return Rect<T>(x1, y1, x2 - x1, y2 - y1);
	}
	else
	{
		return Rect<T>(0, 0, 0, 0);
	}
}

template <typename T>
Rect<T> Rect<T>::createTransformedRect(float scalingFactor) const
{
	const float diff = scalingFactor - 1;
	return Rect<T>(left - diff / 2, top - diff / 2, scalingFactor * width, scalingFactor * height);
}

template <typename T>
Rect<T> Rect<T>::createTransformedRect(float horizScalingFactor, float vertScalingFactor) const
{
	return Rect<T>(left - (horizScalingFactor - 1) / 2, top - (vertScalingFactor - 1) / 2, horizScalingFactor * width, vertScalingFactor * height);
}

template <typename T>
Rect<T>& Rect<T>::operator+=(const Vector2<T>& point)
{
	left += point.x;
	top += point.y;
	return *this;
}

template <typename T>
Rect<T>& Rect<T>::operator-=(const Vector2<T>& point)
{
	left -= point.x;
	top -= point.y;
	return *this;
}

template <typename T>
Rect<T>& Rect<T>::operator*=(const T& scalar)
{
	left *= scalar;
	top *= scalar;
	width *= scalar;
	height *= scalar;
	return *this;
}

template <typename T>
Rect<T>& Rect<T>::operator/=(const T& scalar)
{
	left /= scalar;
	top /= scalar;
	width /= scalar;
	height /= scalar;
	return *this;
}

template <typename T>
Rect<T> operator+(const Rect<T>& lhs, const Vector2<T>& rhs)
{
	return Rect<int>(lhs.left + rhs.x, lhs.top + rhs.y, lhs.width, lhs.height);
}

template <typename T>
Rect<T> operator+(const Vector2<T>& lhs, const Rect<T>& rhs)
{
	return rhs + lhs;
}

template <typename T>
Rect<T> operator-(const Rect<T>& lhs, const Vector2<T>& rhs)
{
	return Rect<int>(lhs.left - rhs.x, lhs.top - rhs.y, lhs.width, lhs.height);
}

template <typename T>
static inline Rect<T> operator*(const Rect<T>& lhs, const T& rhs)
{
	Rect<T> out(lhs);
	out *= rhs;
	return out;
}

template <typename T>
static inline Rect<T> operator*(const T& lhs, const Rect<T>& rhs)
{
	return rhs * lhs;
}

template <typename T>
static inline Rect<T> operator/(const Rect<T>& lhs, const T& rhs)
{
	Rect<T> out(lhs);
	out /= rhs;
	return out;
}

template <typename T>
Vector2<T> Rect<T>::mapTo(const Vector2<T>& input, const Rect<T>& targetRect) const
{
	const float inputXRelative = ((float) input.x - left) / width;
	const float inputYRelative = ((float) input.y - top) / height;
	return Vector2<T>(targetRect.left + inputXRelative * targetRect.width, targetRect.top + inputYRelative * targetRect.height);
}

} // sl

#endif
