#ifndef SL_SPARSEVECTOR_HPP
#define SL_SPARSEVECTOR_HPP

#include <map>

#include "Utility/Vector.hpp"

namespace sl
{

template <typename Key, typename T>
class SparseVector
{
public:
	typedef T Field;
	SparseVector()
	{
	}
	SparseVector(const SparseVector<Key, T>& other)
		:data(other.data)
	{
	}
	SparseVector(SparseVector<Key, T>&& other)
		:data(std::move(other.data))
	{
	}

	/*	  Assignment operators		*/
	SparseVector<Key, T>& operator=(const SparseVector<Key, T>& rhs)
	{
		data = rhs.data;
		return *this;
	}
	SparseVector<Key, T>& operator=(SparseVector<Key, T>&& rhs)
	{
		data = std::move(rhs.data);
		return *this;
	}
	SparseVector<Key, T>& operator+=(const SparseVector& rhs)
	{
		for (const std::pair<Key, T>& dim : rhs.data)
		{
			data[dim.first] += dim.second;
		}
		return *this;
	}
	SparseVector<Key, T>& operator-=(const SparseVector& rhs)
	{
		for (const std::pair<Key, T>& dim : rhs.data)
		{
			data[dim.first] -= dim.second;
		}
		return *this;
	}
	SparseVector<Key, T>& operator*=(const T& rhs)
	{
		for (std::pair<Key, T>& dim : data)
		{
			dim.second *= rhs;
		}
		return *this;
	}
	SparseVector<Key, T>& operator/=(const T& rhs)
	{
		for (std::pair<Key, T>& dim : data)
		{
			dim.second /= rhs;
		}
		return *this;
	}

	/*	  Conditional operators	   */
	bool operator==(const SparseVector<Key, T>& rhs) const
	{
		return data == rhs.data;
	}
	bool operator!=(const SparseVector<Key, T>& rhs) const
	{
		return data != rhs.data;
	}
	bool operator<(const SparseVector<Key, T>& rhs) const
	{
		return data < rhs.data;
	}
	/*    Utility funtions         */
	template <typename U>
	SparseVector<Key, U> to() const
	{
		std::map<Key, U> convertedData;
		for (const std::pair<Key, T>& dim : data)
		{
			convertedData.insert(std::make_pair(dim.first, static_cast<U>(dim.second)));
		}
		return SparseVector<Key, U>(convertedData);
	}
	T length() const
	{
		return sqrtf(this->dot(*this));
	}
	T dot(const SparseVector<Key, T>& rhs) const
	{
		T r(0);
		for (const std::pair<Key, T>& dim : data)
		{
			auto it = rhs.data.find(dim.first);
			if (it != rhs.data.end())
			{
				r += dim.second * it->second;
			}
		}
		return r;
	}
	SparseVector<Key, T> normalized() const
	{
		return *this / length();
	}
	const T& operator[](const Key& key) const
	{
		return data.at(key);
	}
	T& operator[](const Key& key)
	{
		return data[key];
	}
private:
	std::map<Key, T> data;
};

/*	  Creation operators	  */
template <typename Key, typename T>
SparseVector<Key, T> operator+(const SparseVector<Key, T>& lhs, const SparseVector<Key, T>& rhs)
{
	return SparseVector<Key, T>(lhs) += rhs;
}

template <typename Key, typename T>
SparseVector<Key, T> operator-(const SparseVector<Key, T>& lhs, const SparseVector<Key, T>& rhs)
{
	return SparseVector<Key, T>(lhs) -= rhs;
}

template <typename Key, typename T>
static inline SparseVector<Key, T> operator*(const SparseVector<Key, T>& lhs, const T& rhs)
{
	SparseVector<Key, T> output(lhs);
	output *= rhs;
	return output;
}

template <typename Key, typename T>
static inline SparseVector<Key, T> operator*(const T lhs, const SparseVector<Key, T>& rhs)
{
	return rhs * lhs;
}

template <typename Key, typename T>
static inline SparseVector<Key, T> operator/(const SparseVector<Key, T>& lhs, const T& rhs)
{
	SparseVector<Key, T> output(lhs);
	output /= rhs;
	return output;
}

// just for style
template <typename Key, typename T>
static inline T dot(const SparseVector<Key, T>& lhs, const SparseVector<Key, T>& rhs)
{
	return lhs.dot(rhs);
}

template <typename Key, typename T>
static inline T dist(const SparseVector<Key, T>& lhs, const SparseVector<Key, T>& rhs)
{
	return (lhs - rhs).length();
}

template <typename Key, typename T>
static inline SparseVector<Key, T> proj(const SparseVector<Key, T>& src, const SparseVector<Key, T>& dst)
{
	return dst * (dot(src, dst) / (dot(dst, dst)));
}

} // sl

#endif // SL_SPARSEVECTOR_HPP