#include "Utility/Time/RecurringTimeIntervalGenerator.hpp"

#include "Utility/Assert.hpp"

#include <algorithm>

namespace sl
{

RecurringTimeIntervalGenerator::RecurringTimeIntervalGenerator(const RecurringTimeInterval& interval)
	:interval(interval)
{
}

std::vector<TimeInterval> RecurringTimeIntervalGenerator::generate(const GameTime& start, const GameTime& end) const
{
	ASSERT(start < end);
	std::vector<TimeInterval> times;

	const int startS = interval.getTimeOfDayStart().getTotalSeconds();
	const int endS = startS + interval.getDurationSeconds();
	GameDate iter(start.date);
	const bool startEndSameDay = start.date == end.date;
	// handle before edge-case
	if (start.time.getTotalSeconds() <= endS)
	{
		if (interval.isDateWithin(iter))
		{
			const int s = std::max(startS, start.time.getTotalSeconds());
			const int e = startEndSameDay
			            ? std::min(endS, end.time.getTotalSeconds())
			            : endS;
			if (e > s)
			{
				times.push_back(
					TimeInterval(
						GameTime(start.date, TimeOfDay(s)),
						GameTime(start.date, TimeOfDay(e))));
			}
		}
	}
	iter.advance(1);
	
	// repeating case for days not falling on start/end days (simple)
	const int middleDays = (end.date.getYear() - start.date.getYear()) * 365 + (end.date.getDay() - start.date.getDay()) - 2;
	for (int day = 0; day < middleDays; ++day)
	{
		if (interval.isDateWithin(iter))
		{
			times.push_back(
				TimeInterval(
					GameTime(iter, TimeOfDay(startS)),
					GameTime(iter, TimeOfDay(endS))));
		}
		iter.advance(1);
	}
	
	// then after edge-case
	if (end.time.getTotalSeconds() >= startS)
	{
		if (interval.isDateWithin(end.date))
		{
			const int s = startEndSameDay
			            ? std::max(startS, start.time.getTotalSeconds())
			            : startS;
			const int e = std::min(endS, end.time.getTotalSeconds());
			if (e > s)
			{
				times.push_back(
					TimeInterval(
						GameTime(end.date, TimeOfDay(s)),
						GameTime(end.date, TimeOfDay(e))));
			}
		}
	}
	
	return times;
}

bool RecurringTimeIntervalGenerator::isWithin(const GameTime& time) const
{
	return interval.isWithin(time);
}

} // sl