/**
 * @section DESCRIPTION
 * A basic 2D vector class to be used for positions or points or whatever.
 * Do I really have to write documentation for all these 1-3 line methods/operators?!
 */
#ifndef SL_VECTOR_HPP
#define SL_VECTOR_HPP

#include <cmath>

namespace sl
{

template <typename T, unsigned int N>
class Vector
{
public:
	typedef T Field;
	Vector()
		:head(T())
	{
	}

	template <typename... Args>
	Vector(const T& h, Args... t)
		:head(h)
		,tail(t...)
	{
	}

	template <typename U>
	Vector(const Vector<U, N>& other)
		:head(other.head)
		,tail(other.tail)
	{
	}

	/*	  Assignment operators		*/
	inline Vector<T, N>& operator=(const Vector<T, N>& rhs)
	{
		head = rhs.head;
		tail = rhs.tail;
		return *this;
	}
	inline Vector<T, N>& operator+=(const Vector<T, N>& rhs)
	{
		head += rhs.head;
		tail += rhs.tail;
		return *this;
	}
	Vector<T, N>& operator-=(const Vector<T, N>& rhs)
	{
		head -= rhs.head;
		tail -= rhs.tail;
		return *this;
	}
	Vector<T, N>& operator*=(const T& rhs)
	{
		head *= rhs;
		tail *= rhs;
		return *this;
	}
	Vector<T, N>& operator/=(const T& rhs)
	{
		head /= rhs;
		tail /= rhs;
		return *this;
	}

	/*	  Conditional operators	   */
	bool operator==(const Vector<T, N>& rhs) const
	{
		return head == rhs.head && tail == rhs.tail;
	}
	bool operator!=(const Vector<T, N>& rhs) const
	{
		return head != rhs.head || tail != rhs.tail;
	}
	bool operator<(const Vector<T, N>& rhs) const
	{
		return head < rhs.head || (head == rhs.head && tail < rhs.tail);
	}

	/*    Utility funtions         */
	template <typename U>
	Vector<U, N> to() const
	{
		return Vector<U, N>(*this);
	}
	T length() const
	{
		return sqrtf(this->dot(*this));
	}
	T dot(const Vector<T, N>& rhs) const
	{
		return head * rhs.head + tail.dot(rhs.tail);
	}
	Vector<T, N> normalized() const
	{
		return *this / length();
	}

	const T& operator[](int i) const
	{
		if (i <= 0)
		{
			return head;
		}
		return tail[i - 1];
	}

	T& operator[](int i)
	{
		if (i <= 0)
		{
			return head;
		}
		return tail[i - 1];
	}

	template <unsigned int K>
	const T& at() const
	{
		static_assert(K > 0 && K < N);
		return tail.at<K - 1>();
	}

	template <unsigned int K>
	T& at()
	{
		static_assert(K > 0 && K < N);
		return tail.at<K - 1>();
	}

	template <>
	const T& at<0>() const
	{
		return head;
	}

	template <>
	T& at<0>()
	{
		return head;
	}

private:
	T head;
	Vector<T, N - 1> tail;
};

template <typename T>
struct Vector<T, 2>
{
	typedef T Field;
	/*	  Constructors				*/
	Vector()
		:x(T())
		,y(T())
	{
	}
	Vector(const T& x, const T& y)
		:x(x)
		,y(y)
	{
	}
	template <typename U>
	Vector(const Vector<U, 2>& other)
		:x(static_cast<T>(other.x))
		,y(static_cast<T>(other.y))
	{
	}

	/*	  Assignment operators		*/
	inline Vector<T, 2>& operator=(const Vector<T, 2>& rhs)
	{
		x = rhs.x;
		y = rhs.y;
		return *this;
	}
	inline Vector<T, 2>& operator+=(const Vector<T, 2>& rhs)
	{
		x += rhs.x;
		y += rhs.y;
		return *this;
	}
	Vector<T, 2>& operator-=(const Vector<T, 2>& rhs)
	{
		x -= rhs.x;
		y -= rhs.y;
		return *this;
	}
	Vector<T, 2>& operator*=(const T& rhs)
	{
		x *= rhs;
		y *= rhs;
		return *this;
	}
	Vector<T, 2>& operator/=(const T& rhs)
	{
		x /= rhs;
		y /= rhs;
		return *this;
	}

	/*	  Conditional operators	   */
	bool operator==(const Vector<T, 2>& rhs) const
	{
		return x == rhs.x && y == rhs.y;
	}
	bool operator!=(const Vector<T, 2>& rhs) const
	{
		return x != rhs.x || y != rhs.y;
	}
	bool operator<(const Vector<T, 2>& rhs) const
	{
		return x < rhs.x || (x == rhs.x && y < rhs.y);
	}

	/*    Utility funtions         */
	template <typename U>
	Vector<U, 2> to() const
	{
		return Vector<U, 2>(static_cast<U>(x), static_cast<U>(y));
	}
	T length() const
	{
		return sqrtf(this->dot(*this));
	}
	T dot(const Vector<T, 2>& rhs) const
	{
		return x*rhs.x + y*rhs.y;
	}
	Vector<T, 2> normalized() const
	{
		return *this / length();
	}

	const T& operator[](int index) const
	{
		// TODO: assert? throw?
		return index == 1 ? y : x;
	}

	T& operator[](int index)
	{
		return index == 1 ? y : x;
	}

	template <unsigned int K>
	const T& at() const
	{
		static_assert(false);
		return x;
	}

	template <>
	const T& at<0>() const
	{
		return x;
	}

	template <>
	const T& at<1>() const
	{
		return y;
	}

	template <unsigned int K>
	T& at()
	{
		static_assert(false);
		return x;
	}

	template <>
	T& at<0>()
	{
		return x;
	}

	template <>
	T& at<1>()
	{
		return y;
	}

	T x;
	T y;
};

typedef Vector<float, 2> Vector2f;
typedef Vector<int, 2> Vector2i;
typedef Vector<unsigned int, 2> Vector2u;

template <typename T>
using Vector2 = Vector<T, 2>;


/*	  Creation operators	  */
template <typename T, unsigned int N>
Vector<T, N> operator+(const Vector<T, N>& lhs, const Vector<T, N>& rhs)
{
	return Vector<T, N>(lhs) += rhs;
}

template <typename T, unsigned int N>
Vector<T, N> operator-(const Vector<T, N>& lhs, const Vector<T, N>& rhs)
{
	return Vector<T, N>(lhs) -= rhs;
}

template <typename T, unsigned int N>
static inline Vector<T, N> operator*(const Vector<T, N>& lhs, const T& rhs)
{
	Vector<T, N> output(lhs);
	output *= rhs;
	return output;
}

template <typename T, unsigned int N>
static inline Vector<T, N> operator*(const T lhs, const Vector<T, N>& rhs)
{
	return rhs * lhs;
}

template <typename T, unsigned int N>
static inline Vector<T, N> operator/(const Vector<T, N>& lhs, const T& rhs)
{
	Vector<T, N> output(lhs);
	output /= rhs;
	return output;
}

// just for style
template <typename T, unsigned int N>
static inline T dot(const Vector<T, N>& lhs, const Vector<T, N>& rhs)
{
	return lhs.dot(rhs);
}

template <typename T, unsigned int N>
static inline T dist(const Vector<T, N>& lhs, const Vector<T, N>& rhs)
{
	return (lhs - rhs).length();
}

template <typename T, unsigned int N>
static inline Vector<T, N> proj(const Vector<T, N>& src, const Vector<T, N>& dst)
{
	return dst * (dot(src, dst) / (dot(dst, dst)));
}

/* type specializations for int/float mixing */
static inline Vector2f operator+(const Vector2i& lhs, const Vector2f& rhs)
{
	return lhs.to<float>() + rhs;
}

static inline Vector2f operator+(const Vector2f& lhs, const Vector2i& rhs)
{
	return lhs + rhs.to<float>();
}

static inline Vector2f operator*(const Vector2i& lhs, float rhs)
{
	return lhs.to<float>() * rhs;
}

static inline Vector2f operator*(float lhs, const Vector2i& rhs)
{
	return rhs.to<float>() * lhs;
}

static inline Vector2f operator/(const Vector2f& lhs, int rhs)
{
	return lhs * (1.f / rhs);
}

} // sl

#endif
